Package: polle
Title: Policy Learning
Version: 1.6.4
Authors@R: c(person(given = "Andreas",
                    family = "Nordland",
                    role = c("aut", "cre"),
                    email = "andreasnordland@gmail.com"),
             person(given = "Klaus",
                    family = "Holst",
                    role = c("aut"),
                    email = "klaus@holst.it",
                    comment = c(ORCID="0000-0002-1364-6789"))
             )
Description: Package for learning and evaluating (subgroup) policies via doubly robust loss functions. Policy learning methods include doubly robust blip/conditional average treatment effect learning and sequential policy tree learning. Methods for (subgroup) policy evaluation include doubly robust cross-fitting and online estimation/sequential validation. See Nordland and Holst (2026) <doi:10.18637/jss.v116.i04> for documentation and references.
License: Apache License (>= 2)
Encoding: UTF-8
Depends: R (>= 4.1), SuperLearner
Imports: data.table (>= 1.14.5), lava (>= 1.7.2.1), future.apply,
        progressr, methods, policytree (>= 1.2.0), survival, targeted
        (>= 0.6), DynTxRegime
Suggests: DTRlearn2, glmnet (>= 4.1-6), mets, mgcv, xgboost, knitr,
        ranger, rmarkdown, testthat (>= 3.0), ggplot2
RoxygenNote: 7.3.3
BugReports: https://github.com/AndreasNordland/polle/issues
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-16 20:45:24 UTC; oano
Author: Andreas Nordland [aut, cre],
  Klaus Holst [aut] (ORCID: <https://orcid.org/0000-0002-1364-6789>)
Maintainer: Andreas Nordland <andreasnordland@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-17 22:50:02 UTC
Built: R 4.5.2; ; 2026-05-18 00:39:32 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("+",policy_eval)
S3method(IC,policy_eval)
S3method(coef,policy_eval)
S3method(conditional,policy_eval)
S3method(estimate,policy_eval)
S3method(get_K,policy_data)
S3method(get_action_set,policy_data)
S3method(get_actions,policy_data)
S3method(get_event,history)
S3method(get_event,policy_data)
S3method(get_g_functions,policy_eval)
S3method(get_g_functions,policy_object)
S3method(get_history,policy_data)
S3method(get_history_names,policy_data)
S3method(get_id,history)
S3method(get_id,policy_data)
S3method(get_id_stage,history)
S3method(get_id_stage,policy_data)
S3method(get_n,policy_data)
S3method(get_policy,blip)
S3method(get_policy,drql)
S3method(get_policy,earl)
S3method(get_policy,owl)
S3method(get_policy,policy_eval)
S3method(get_policy,ptl)
S3method(get_policy,ql)
S3method(get_policy,rwl)
S3method(get_policy_actions,policy_eval)
S3method(get_policy_functions,blip)
S3method(get_policy_functions,drql)
S3method(get_policy_functions,policy_eval)
S3method(get_policy_functions,ptl)
S3method(get_policy_functions,ql)
S3method(get_policy_object,policy_eval)
S3method(get_q_functions,policy_eval)
S3method(get_q_functions,policy_object)
S3method(get_stage_action_sets,policy_data)
S3method(get_stage_data,policy_data)
S3method(get_utility,policy_data)
S3method(merge,policy_eval)
S3method(partial,policy_data)
S3method(plot,policy_data)
S3method(plot,policy_eval)
S3method(predict,QV_function)
S3method(predict,Q_function)
S3method(predict,SL.glm.formula)
S3method(predict,SL.nls)
S3method(predict,blip_function)
S3method(predict,c_cox)
S3method(predict,c_function)
S3method(predict,c_functions)
S3method(predict,c_no_censoring)
S3method(predict,g_empir)
S3method(predict,g_function)
S3method(predict,g_glm)
S3method(predict,g_glmnet)
S3method(predict,g_rf)
S3method(predict,g_sl)
S3method(predict,g_xgboost)
S3method(predict,m_function)
S3method(predict,nuisance_functions)
S3method(predict,q_degen)
S3method(predict,q_glm)
S3method(predict,q_glmnet)
S3method(predict,q_rf)
S3method(predict,q_sl)
S3method(predict,q_xgboost)
S3method(print,Q_function)
S3method(print,g_function)
S3method(print,nuisance_functions)
S3method(print,policy)
S3method(print,policy_data)
S3method(print,policy_eval)
S3method(print,policy_learn)
S3method(print,policy_object)
S3method(print,summary.policy_data)
S3method(subset_id,policy_data)
S3method(summary,policy_data)
S3method(summary,policy_eval)
S3method(vcov,policy_eval)
S3method(vcov,policy_eval_online)
export(All)
export(IC)
export(c_cox)
export(c_no_censoring)
export(conditional)
export(control_blip)
export(control_drql)
export(control_earl)
export(control_owl)
export(control_ptl)
export(control_rwl)
export(copy_policy_data)
export(estimate)
export(fit_c_functions)
export(fit_g_functions)
export(g_empir)
export(g_glm)
export(g_glmnet)
export(g_rf)
export(g_sl)
export(g_xgboost)
export(get_K)
export(get_action_set)
export(get_actions)
export(get_event)
export(get_g_functions)
export(get_history)
export(get_history_names)
export(get_id)
export(get_id_stage)
export(get_n)
export(get_policy)
export(get_policy_actions)
export(get_policy_functions)
export(get_policy_object)
export(get_q_functions)
export(get_stage_action_sets)
export(get_utility)
export(partial)
export(policy_data)
export(policy_def)
export(policy_eval)
export(policy_eval_online)
export(policy_learn)
export(q_glm)
export(q_glmnet)
export(q_rf)
export(q_sl)
export(q_xgboost)
export(sim_multi_stage)
export(sim_single_stage)
export(sim_single_stage_multi_actions)
export(sim_two_stage)
export(sim_two_stage_multi_actions)
export(subset_id)
import(SuperLearner)
import(data.table)
importFrom(SuperLearner,All)
importFrom(lava,IC)
importFrom(lava,estimate)
importFrom(methods,formalArgs)
importFrom(progressr,progressor)
importFrom(stats,.getXlevels)
importFrom(stats,addmargins)
importFrom(stats,aggregate)
importFrom(stats,as.formula)
importFrom(stats,binomial)
importFrom(stats,coef)
importFrom(stats,complete.cases)
importFrom(stats,delete.response)
importFrom(stats,gaussian)
importFrom(stats,glm)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,na.pass)
importFrom(stats,nls)
importFrom(stats,predict)
importFrom(stats,predict.glm)
importFrom(stats,reformulate)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(survival,Surv)
importFrom(survival,survfit)
importFrom(utils,capture.output)
importFrom(utils,tail)

<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![license](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/license/apache-2-0)
[![cran](https://www.r-pkg.org/badges/version-last-release/polle)](https://cranlogs.r-pkg.org/downloads/total/last-month/polle)
<!-- badges: end -->

# Policy Learning (`polle`) <img src="man/figures/logo.png" align="right" height="250" style="float:right; height:250px;"  alt="polle website" />

Package for evaluating user-specified finite stage policies and learning
optimal treatment policies via doubly robust loss functions. Policy
learning methods include doubly robust learning of the blip/conditional
average treatment effect and sequential policy tree learning. The
package also include methods for optimal subgroup analysis. See Nordland
and Holst (2026)
[doi:10.18637/jss.v116.i04](https://doi.org/10.18637/jss.v116.i04) for
documentation and references.

## ----lib, message = FALSE-----------------------------------------------------
library(polle)
library(data.table)

## ----simdata------------------------------------------------------------------
sim_single_stage_right_cens <- function(n = 2e3, zeta = c(0.7, 0.2), type = "right"){

  d <- sim_single_stage(n = n)
  pd <- policy_data(data = d,
                    action = "A",
                    covariates = c("Z", "L", "B"),
                    utility = "U")

  ld <- pd$stage_data

  ld[stage == 1, time := 1]
  ld[stage == 2, time := 2]

  ld[stage == 2, Z := d$Z]
  ld[stage == 2, L := d$L]
  ld[stage == 2, B := d$B]

  ## simulating the right censoring time
  ## only depending on the baseline covariate Z:
  C <- c(rexp(n, 1) / exp((-1) * cbind(1, as.numeric(d$Z)) %*% zeta))

  ld[stage == 1, time_c := C]
  ld[stage == 2, time_c := C]

  ld[, delta := time_c >= time]

  ld[delta == FALSE , event := 2]
  ld[delta == FALSE, A := NA]
  ld[delta == FALSE & stage == 2, U := NA]
  ld[delta == FALSE, U_A0 := 0]
  ld[delta == FALSE, U_A1 := 0]

  ld[ , tmp := shift(delta, fill = TRUE), by = list(id)]
  ld <- ld[tmp == TRUE, ]
  ld[ , time := pmin(time, time_c)]
  ld[ , time_c := NULL]
  ld[ , tmp := NULL]
  ld[ , delta := NULL]

  if (type == "interval"){
    ld[, time2 := time]
    ld[, time := shift(time, fill = 0), by = list(id)]
  }

  return(ld)
}

## ----simpd--------------------------------------------------------------------
set.seed(1)
ld <- sim_single_stage_right_cens(n = 5e2, type = "interval")
pd <- policy_data(data = ld, type = "long", action = "A", time = "time", time2 = "time2")

## ----pe-----------------------------------------------------------------------
pe <- policy_eval(
  policy_data = pd,
  policy = policy_def(1),
  m_model = q_glm(~ .),
  m_full_history = TRUE,
  c_models = list(c_cox(formula = ~ Z_1),
                  c_cox(formula = ~ Z_2*A_1)),
  c_full_history = TRUE
)
pe

## ----pl-----------------------------------------------------------------------
pe <- policy_eval(
  policy_data = pd,
  policy_learn = policy_learn(type = "blip", control = control_blip()),
  m_model = q_glm(~.),
  m_full_history = FALSE,
  c_models = c_cox(formula = ~ Z)
)
pe

SL.glm.formula <- function(Y, X, newX, obsWeights, family, id, formula=Y~., ...) {
  formula <- update(formula, Y~.)
  environment(formula) <- environment()
  fit <- glm(formula=formula, data=X, family=family, weights=obsWeights)
  pred <- predict(fit, newdata=newX, type="response")
  obj <- list(fit=fit)
  class(obj) <- c("SL.glm.formula")
  out <- list(pred = pred, fit = obj)
  return(out)
}
#' @export
predict.SL.glm.formula <- function(object, newdata, family, X = NULL, Y = NULL,...) {
	pred <- predict(object=object$fit, newdata=newdata, type="response")
	return(pred)
}

SL.nls <- function(Y, X, newX, obsWeights, family, id, formula, start,  ...) {
  formula <- as.formula(paste0("Y~", tail(as.character(formula), 1)))
  environment(formula) <- environment()
  fit <- nls(formula=formula, data=X, weights=obsWeights, start=start, ...)
  pred <- predict(fit, newdata=newX)
  obj <- list(fit=fit)
  class(obj) <- c("SL.nls")
  out <- list(pred = pred, fit = obj)
  return(out)
}
#' @export
predict.SL.nls <- function(object, newdata, family,
                           X = NULL, Y = NULL, ...) {
	pred <- predict(object=object$fit, newdata=newdata)
	return(pred)
}

am <- Vectorize(
  function(a, action_set) {
    am <- action_set %in% a
    return(am)
  },
  vectorize.args = "a"
)
action_matrix <- function(a, action_set) {
  t(am(a = a, action_set = action_set))
}

get_a_values <- function(a, action_set, values, na.rm = TRUE) {
  stopifnot(
    is.data.table(values),
    all(key(values) == c("id", "stage")),
    length(action_set) == ncol(values[, -c("id", "stage"), with = FALSE]),
    length(a) == nrow(values)
  )
  P <- action_matrix(a = a, action_set = action_set) * values[, -c("id", "stage"), with = FALSE]
  P <- apply(P, 1, sum, na.rm = na.rm)
  out <- data.table(values[, c("id", "stage"), with = FALSE], P = P)

  return(out)
}

ipw_weight <- function(D, G){
  stopifnot(
    is.vector(D) | is.matrix(D),
    is.vector(G) | is.matrix(G),
    all(dim(G) == dim(D)),
    all(is.na(D) == is.na(G))
  )

  ## replacing NA values with 1:
  D[is.na(D)] <- 1
  G[is.na(G)] <- 1

  if (is.vector(D)) {
    out <- D / G
  } else {
    out <- apply(D / G, 1, prod)
  }

  if (any(is.na(out))) {
    mes <- "The policy dictates actions with fitted probability 0. Unable to calculate inverse probability weights."
    stop(mes)
  }

  return(out)
}

colprod <- function(M) {
  if (is.vector(M)) {
    out <- M
  } else {
    out <- apply(M, 1, prod, na.rm = TRUE)
  }
  return(out)
}

remove_null_elements <- function(x) {
  Filter(Negate(is.null), x)
}

get_element <- function(x, name, check_name = TRUE) {
  if (!is.character(name) || length(name) != 1) {
    stop("'name' must be a character string.")
  }
  if (check_name == TRUE) {
    if (!(name %in% names(x))) {
      stop("named element does not exist.")
    }
  }
  ## element with exact match:
  y <- getElement(x, name = name)

  return(y)
}

get_col <- function(dt, col_name) {
  if (!data.table::is.data.table(dt)) {
    stop("Input must be a data.table")
  }
  if (!is.character(col_name) || length(col_name) != 1) {
    stop("col_name must be a single character string")
  }

  if (!col_name %in% names(dt)) {
    stop(sprintf("Column '%s' not found in data.table", col_name))
  }
  dt[[col_name]]
}

#' Create Random Folds of Indices
#'
#' @param number Integer. Number of folds to create
#' @param sample_size Integer. Total number of indices to split
#'
#' @return List of integer vectors containing indices, or NULL if number <= 1
#'
#' @examples
#' sample_folds(3, 10)
#' sample_folds(5, 100)
#' @noRd
sample_folds <- function(number, sample_size) {
  ## input checks:
  if (!is.numeric(number) || !is.numeric(sample_size)) {
    stop("Arguments must be numeric")
  }
  if (number < 0 || sample_size < 0) {
    stop("Arguments must be non-negative")
  }
  if (number %% 1 != 0 || sample_size %% 1 != 0) {
    stop("Arguments must be whole numbers")
  }
  if (sample_size < number) {
    stop ("sample_size is less than the number of folds")
  }

  if (number > 1) {
    folds <- split(sample(1:sample_size, sample_size), rep(1:number, length.out = sample_size))
    folds <- lapply(folds, sort)
  } else {
    folds <- NULL
  }
  return(folds)
}

testthat::test_that("IPW construction", {
  d <- c(0,1,1)
  g <- c(0.1,0.2,0.3)
  testthat::expect_equivalent(d/g, polle:::ipw_weight(d,g))
  testthat::expect_equivalent(d/(g*g), polle:::ipw_weight(cbind(d,d), cbind(g,g)))
  testthat::expect_error(polle:::ipw_weight(c(NA,1,1),g))
})

testthat::test_that("action_matrix", {
  A <- c(0,1,0,1)
  Aset <- c(0,1,2)
  res <- polle:::action_matrix(A, Aset)
  testthat::expect_true(ncol(res)==length(Aset))
  testthat::expect_true(nrow(res)==length(A))
  testthat::expect_true(all(!res[,3]))
  testthat::expect_true(all(res[,2]*1==A))
})

test_that("get_element", {
    x <- list(ab = "test", b = NULL, c = NA, d = vector())

    expect_equal(
        get_element(x, "ab"),
        "test"
    )

    expect_equal(
        get_element(x, "b"),
        NULL
    )

    expect_equal(
        get_element(x, "c"),
        NA
    )

    expect_equal(
        get_element(x, "d"),
        vector()
    )

    expect_error(
        get_element(x, "a"),
        "named element does not exist"
    )

    expect_error(
        get_element(x, 1),
        "'name' must be a character string"
    )

    expect_equal(
        get_element(x, "a", check_name = FALSE),
        NULL
    )
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
##' @importFrom lava IC estimate
##' @importFrom SuperLearner All
##' @importFrom stats predict gaussian binomial glm as.formula nls model.frame
##'   model.matrix reformulate predict.glm terms update .getXlevels addmargins
##'   coef delete.response complete.cases model.response na.pass aggregate vcov var
##' @importFrom methods formalArgs
##' @importFrom utils tail capture.output
##' @importFrom survival survfit Surv
##' @importFrom progressr progressor
##' @import data.table SuperLearner
## usethis namespace: end
NULL

##' @export
lava::IC

##' @export
lava::estimate

##' @export
SuperLearner::All

##' For internal use
##'
##' @title For internal use
##' @name SL.nls
##' @rdname internal
##' @keywords utilities
##' @export
##' @aliases
##' SL.nls SL.glm.formula

#' @title Control arguments for Residual Weighted Learning
#' @description \code{control_rwl} sets the default control arguments
#' for residual learning , \code{type = "rwl"}.
#' The arguments are passed directly to [DynTxRegime::rwl()] if not
#' specified otherwise.
#' @param moPropen Propensity model of class "ModelObj", see [modelObj::modelObj].
#' @param moMain Main effects outcome model of class "ModelObj".
#' @param regime An object of class [formula] specifying the design of the policy/regime.
#' @param fSet A function or NULL defining subset structure.
#' @param lambdas Numeric or numeric vector. Penalty parameter.
#' @param cvFolds Integer. Number of folds for cross-validation of the parameters.
#' \code{"logit"}, \code{"exp"}, \code{"hinge"}, \code{"sqhinge"}, \code{"huber"}.
#' @param kernel The options are \code{"linear"}, \code{"poly"}, \code{"radial"}.
#' @param kparam Numeric. Kernel parameter
#' @param responseType Character string. Options are \code{"continuous"},
#' \code{"binary"}, \code{"count"}.
#' @param verbose Integer.
#' @returns list of (default) control arguments.
#' @export
control_rwl <- function(moPropen,
                        moMain,
                        regime,
                        fSet = NULL,
                        lambdas = 2,
                        cvFolds = 0L,
                        kernel = "linear",
                        kparam = NULL,
                        responseType = "continuous",
                        verbose = 2L){
  control <- as.list(environment())
  return(control)
}

dyntxregime_rwl <- function(policy_data,
                            alpha,
                            L,
                            moPropen,
                            moMain,
                            regime,
                            fSet,
                            lambdas,
                            cvFolds,
                            kernel,
                            kparam,
                            responseType,
                            verbose,
                            ...){
  K <- get_K(policy_data)
  if (K != 1)
    stop("rwl is only implemented for single stage problems.")
  n <- get_n(policy_data)
  action_set <- get_action_set(policy_data)
  id_stage <- get_id_stage(policy_data)

  if (!(length(action_set) == 2))
    stop("rwl only works for a dichotomous action set.")

  if (alpha != 0)
    stop("alpha must be 0 when using rwl.")

  if (L != 1)
    stop("L must be 1 when using rwl (no cross-fitting).")
  if (any(get_element(policy_data, "cens_indicator")[["indicator"]])){
    stop("policy learning with type 'rwl' not implemented under right-censoring/missing outcomes.")
  }

  # getting the observed actions:
  actions <- get_actions(policy_data)

  # getting the observed (complete) utilities:
  utility <- get_utility(policy_data)

  H <- get_H(get_history(policy_data))
  A <- actions[["A"]]
  H <- cbind(A = A, H)

  rwl_object <- DynTxRegime::rwl(moPropen = moPropen,
                                 moMain = moMain,
                                 data = H,
                                 response = utility[["U"]],
                                 txName = "A",
                                 regime = regime,
                                 fSet = fSet,
                                 lambdas = lambdas,
                                 cvFolds = cvFolds,
                                 kernel = kernel,
                                 kparam = kparam,
                                 responseType = responseType,
                                 verbose = verbose)


  out <- list(
    rwl_object = rwl_object,
    action_set = action_set,
    K = K
  )
  class(out) <- c("rwl","policy_object","list")

  return(out)
}

#' @export
get_policy.rwl <- function(object, threshold = NULL){
  if (!(is.null(threshold) || identical(threshold, 0))) {
    stop("threshold is not implemented for rwl.")
  }
  rwl_object <- getElement(object, "rwl_object")
  action_set <- getElement(object, "action_set")
  K <- getElement(object, "K")

  policy <- function(policy_data){
    if (get_K(policy_data) != K)
      stop("The policy do not have the same number of stages as the policy data object.")

    H <- get_H(get_history(policy_data))

    pred <- DynTxRegime::optTx(rwl_object, H)
    policy_actions <- get_id_stage(policy_data)
    d <- NULL
    policy_actions[, d := pred$optimalTx]
    setkeyv(policy_actions, c("id", "stage"))

    return(policy_actions)
  }

  # setting class and attributes:
  policy <- new_policy(policy, name = "rwl")

  return(policy)
}

test_that("the polle implementation of rwl agrees with direct application of DynTxRegime::rwl in the single stage case.",{
  library("DynTxRegime")

  d1 <- sim_single_stage(200, seed=1)
  pd1 <- policy_data(d1,
                     action="A",
                     covariates = list("Z", "B", "L"),
                     utility="U")

  # direct application:
  moPropen1 <- buildModelObj(model = ~B+Z+L,
                             solver.method = 'glm',
                             solver.args = list('family'='binomial'),
                             predict.method = 'predict.glm',
                             predict.args = list(type='response'))

  moMain1 <- buildModelObj(model = ~B+Z+L,
                           solver.method = 'lm')

  set.seed(1)
  dir <- DynTxRegime::rwl(moPropen = moPropen1,
                          moMain = moMain1,
                          response = d1$U,
                          data = d1,
                          txName = "A",
                          lambdas = c(0.5, 1, 2),
                          regime = ~B+Z+L,
                          cvFolds = 3,
                          verbose = 0)

  # polle application:
  pl <- policy_learn(type = "rwl",
                     control = control_rwl(moPropen = moPropen1,
                                           moMain = moMain1,
                                           regime = ~B+Z+L,
                                           verbose = 0,
                                           lambdas = c(0.5, 1, 2),
                                           cvFolds = 3))
  set.seed(1)
  po <- pl(policy_data = pd1)

  # comparison
  expect_equal(
    dir@analysis@optimal@estimatedValue,
    po$rwl_object@analysis@optimal@estimatedValue
  )
})

check_actions <- function(actions, policy_data){
  ## checking the format of the actions data.table
  if (!is.data.table(actions))
    stop("actions must be a data.table.")
  if (!any("d" %in% colnames(actions)))
    stop("actions must have an action varible named 'd'.")
  if (!all(key(actions) == c("id", "stage")))
    stop("actions must have keys 'id' and 'stage'.")

  ## checking that the actions comply with the stage action sets
  K <- get_K(policy_data)
  stage_action_sets <- get_stage_action_sets(policy_data)
  stage <- NULL
  for(k_ in 1:K){
    if (!all(unlist(actions[stage == k_, "d"]) %in% stage_action_sets[[k_]])){
      mes <- "The policy actions does not comply with the stage action sets of the policy data object."
      stop(mes)
    }
  }
}

#' @rdname policy_eval
#' @export
coef.policy_eval <- function(object, ...) {
  return(get_element(object, "coef"))
}

#' @rdname policy_eval
#' @export
IC.policy_eval <- function(x, ...) {
  res <- cbind(get_element(x, "IC", check_name = FALSE))
  return(res)
}

#' @rdname policy_eval
#' @export
vcov.policy_eval <- function(object, ...) {
  ic <- IC(object)
  if (is.null(ic))
    return(NULL)
  n <- nrow(ic)
  return(crossprod(ic)/(n*n))
}

#' @rdname policy_eval
#' @export
print.policy_eval <- function(x,
                              digits = 4L,
                              width = 35L,
                              std.error = TRUE,
                              level = 0.95,
                              p.value = TRUE,
                              ...) {
  est <- estimate(x, level = level, ...)
  print(
    est,
    digits = digits,
    width = width,
    std.error = std.error,
    p.value = p.value,
    ...
  )
}


#' @rdname policy_eval
#' @export
summary.policy_eval <- function(object, ...) {
  lava::estimate(object, ...)
}

#' @rdname policy_eval
#' @export
estimate.policy_eval <- function(x,
                                 labels = get_element(x,
                                                      "name",
                                                      check_name = FALSE),
                                 level = 0.95,
                                 ...) {
  p <- length(coef(x))
  if (is.null(labels)) {
    target <- get_element(x, "target")
    if (p == 1) {
      labels <- target
    } else {
      labels <- paste0(target, seq(p))
    }
  }
  ic <- IC(x)
  if (is.null(ic)) {
    est <- lava::estimate(
                   NULL,
                   coef = coef(x),
                   vcov = vcov(x),
                   labels = labels,
                   ...
                 )
  } else {
    est <- lava::estimate(
                   NULL,
                   coef = coef(x),
                   IC = ic,
                   labels = labels,
                   level = level,
                   ...
                 )
  }
  return(est)
}

#' @rdname policy_eval
#' @export
"merge.policy_eval" <- function(x, y, ..., paired = TRUE) {
  dots <- list(...)
  idx <- names(dots) %in% formalArgs(lava::estimate.default)[-1]
  est_args <- list()
  if (length(idx)>0) {
    est_args <- dots[which(idx)]
    dots <- dots[-which(idx)]
  }
  m <- lapply(c(list(x, y), dots), function(p)
    do.call(estimate, c(list(p),est_args)))
  m <- do.call("merge", c(m, list(paired=paired)))
  return(m)
}

#' @rdname policy_eval
#' @export
"+.policy_eval" <- function(x,...) {
  merge(x, ...)
}

#' @title Conditional Policy Evaluation
#' @description \code{conditional()} is used to calculate the
#' policy value for each group defined by a given baseline variable.
#' @param object Policy evaluation object created by [policy_eval()].
#' @param policy_data Policy data object created by [policy_data()].
#' @param baseline Character string.
#' @returns object of inherited class 'estimate', see [lava::estimate.default].
#' The object is a list with elements 'coef' (policy value estimate for each
#' group) and 'IC' (influence curve estimate matrix).
#' @examples
#' library("polle")
#' library("data.table")
#' setDTthreads(1)
#' d <- sim_single_stage(n=2e3)
#' pd <- policy_data(d,
#'                   action = "A",
#'                   baseline = c("B"),
#'                   covariates = c("Z","L"),
#'                   utility = "U")
#'
#' # static policy:
#' p <- policy_def(1)
#'
#' pe <- policy_eval(pd,
#'                   policy = p)
#'
#' # conditional value for each group defined by B
#' conditional(pe, pd, "B")
#' @export
conditional <- function(object, policy_data, baseline)
  UseMethod("conditional")

#' @export
conditional.policy_eval <- function(object, policy_data, baseline) {
  policy_eval <- object

  if (!inherits(policy_eval, "policy_eval")) {
    stop("policy_eval must be of inherited class 'policy_eval'.")
  }
  if (!inherits(policy_data, "policy_data")) {
    stop("policy_data must be of inherited class 'policy_data'.")
  }
  if (!is.character(baseline) || length(baseline) != 1) {
    stop("baseline must be a single character.")
  }
  if (get_element(policy_eval, "target") != "value") {
    stop("only implemented for target = 'value'.")
  }

  baseline_data <- policy_data[["baseline_data"]]

  # checking IDs
  check <- all.equal(
    policy_eval[["id"]],
    baseline_data[["id"]]
  )
  if (!check) {
    stop("ID's does not match.")
  }

  # getting the doubly robust score:
  z <- IC(policy_eval) + coef(policy_eval)

  by <- baseline_data[, baseline, with = FALSE]
  agg <- aggregate(z, by = by, mean)
  coef <- agg[["V1"]]

  n <- get_n(policy_data)
  groups <- agg[[baseline]]
  IC <- matrix(0, nrow = nrow(baseline_data), ncol = length(groups))
  for (j in seq_along(coef)) {
    idx <- baseline_data[[baseline]] == groups[j]
    ic <- z[idx, ] - coef[j]
    IC[idx, j] <- ic / sum(idx) * n
  }
  est <- estimate(NULL,
    coef = coef,
    IC = cbind(IC),
    id = baseline_data[["id"]],
    labels = paste(baseline, groups, sep = ":")
  )
  return(est)
}

#' @export
get_g_functions.policy_eval <- function(object){
  getElement(object, "g_functions")
}

#' @export
get_q_functions.policy_eval <- function(object){
  getElement(object, "q_functions")
}

#' @export
get_policy.policy_eval <- function(object){
  po <- getElement(object, "policy_object")
  if (is.null(po)){
    mes <- "Learned policy is not available."
    stop(mes)
  }
  pf <- get_policy(po)
  return(pf)
}

#' @title Get Policy Actions
#'
#' @description \code{get_policy_actions()} extract the actions dictated by the
#' (learned and possibly cross-fitted) policy a every stage.
#' @param object Object of class [policy_eval].
#' @returns [data.table::data.table] with keys \code{id} and \code{stage} and action variable
#' \code{d}.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # defining a policy learner based on cross-fitted doubly robust Q-learning:
#' pl <- policy_learn(type = "drql",
#'                    control = control_drql(qv_models = list(q_glm(~C_1), q_glm(~C_1+C_2))),
#'                    full_history = TRUE,
#'                    L = 2) # number of folds for cross-fitting
#'
#' # evaluating the policy learner using 2-fold cross fitting:
#' pe <- policy_eval(type = "dr",
#'                    policy_data = pd,
#'                    policy_learn = pl,
#'                    q_models = q_glm(),
#'                    g_models = g_glm(),
#'                    M = 2) # number of folds for cross-fitting
#'
#' # Getting the cross-fitted actions dictated by the fitted policy:
#' head(get_policy_actions(pe))
#' @export
get_policy_actions <- function(object)
  UseMethod("get_policy_actions")

#' @export
get_policy_actions.policy_eval <- function(object){
  getElement(object, "policy_actions")
}

#' @export
get_policy_functions.policy_eval <- function(object, stage, ...){
  po <- getElement(object, "policy_object")
  if (is.null(po)){
    mes <- "Learned policy is not available."
    stop(mes)
  }
  pf <- get_policy_functions(po, stage = stage, ...)
  return(pf)
}

#' @title Plot histogram of the influence curve for a \code{policy_eval} object
#'
#' @param x Object of class [policy_eval]
#' @param ... Additional arguments
#' @examples
#' d <- sim_two_stage(2e3, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = "BB",
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#'
#' pe <- policy_eval(pd,
#'                   policy_learn = policy_learn())
#'
#' plot(pe)
#' @export
plot.policy_eval <- function(x, ...){
    ic <- IC(x)
    se <- sqrt(mean(ic^2))
    graphics::hist(ic,
                   xlab = "IC",
                   main = "Histogram of Influence Curve Terms",
                   prob = TRUE,
                   ...)
    graphics::curve(stats::dnorm(x, mean=0, sd=se),
                    add = TRUE,
                    lwd = 2,
                    col = "red")
    grDevices::dev.flush()

  invisible()
}

#' @title Control arguments for Outcome Weighted Learning
#' @description \code{control_owl()} sets the default control arguments
#' for backwards outcome weighted learning, \code{type = "owl"}.
#' The arguments are passed directly to [DTRlearn2::owl()] if not
#' specified otherwise.
#' @param policy_vars Character vector/string or list of character
#' vectors/strings. Variable names used to restrict the policy.
#' The names must be a subset of the history names, see get_history_names().
#' Not passed to \code{owl()}.
#' @param reuse_scales The history matrix passed to \code{owl()} is scaled
#' using [scale()] as advised. If \code{TRUE}, the scales of the history matrix
#' will be saved and reused when applied to (new) test data.
#' @param res.lasso If \code{TRUE} a lasso penalty is applied.
#' @param loss Loss function. The options are \code{"hinge"}, \code{"ramp"},
#' \code{"logit"}, \code{"logit.lasso"}, \code{"l2"}, \code{"l2.lasso"}.
#' @param kernel Type of kernel used by the support vector machine. The
#' options are \code{"linear"}, \code{"rbf"}.
#' @param augment If \code{TRUE} the outcomes are augmented.
#' @param c Regularization parameter.
#' @param sigma Tuning parameter.
#' @param s Slope parameter.
#' @param m Number of folds for cross-validation of the parameters.
#' @returns list of (default) control arguments.
#' @export
control_owl <- function(policy_vars = NULL,
                        reuse_scales = TRUE,
                        res.lasso = TRUE,
                        loss = 'hinge',
                        kernel = 'linear',
                        augment = FALSE,
                        c = 2^(-2:2),
                        sigma = c(0.03,0.05,0.07),
                        s = 2.^(-2:2),

                        m = 4){
  control <- as.list(environment())
  return(control)
}

dtrlearn2_owl <- function(policy_data,
                          alpha,
                          g_models, g_functions, g_full_history,
                          policy_vars, full_history,
                          L,
                          cross_fit_g_models, save_cross_fit_models,
                          future_args,
                          reuse_scales,
                          res.lasso, loss, kernel,
                          augment, c, sigma,
                          s, m,
                          ...){

  if ((is.null(g_models) & is.null(g_functions)))
    stop("Provide either g-models or g-functions.")

  K <- get_K(policy_data)
  n <- get_n(policy_data)
  action_set <- get_action_set(policy_data)
  stage_action_sets <- get_stage_action_sets(policy_data)
  id_stage <- get_id_stage(policy_data)

  if(!(length(unlist(unique(id_stage[,.N, by = "id"][, "N"]))) == 1))
    stop("owl is only implemented for a fixed number of stages.")

  for (k in seq_along(stage_action_sets)){
    if (length(stage_action_sets[[k]]) != 2)
      stop("owl only works for dichotomous stage action sets.")
  }

  if (alpha != 0)
    stop("alpha must be 0 when type = 'owl'")

  if (full_history == TRUE){
    if ((!is.list(policy_vars)) | (length(policy_vars) != K))
      stop("policy_vars must be a list of length K, when full_history = TRUE.")
  }
  if (any(get_element(policy_data, "cens_indicator")[["indicator"]])){
    stop("policy learning with type 'owl' not implemented under right-censoring/missing outcomes.")
  }

  # getting the observed actions:
  actions <- get_actions(policy_data)

  # getting the IDs:
  id <- get_id(policy_data)

  # getting the rewards
  rewards <- get_rewards(policy_data)

  # constructing the folds for cross-fitting
  if (L > 1){
    folds <- split(sample(1:n, n), rep(1:L, length.out = n))
  } else{
    folds <- NULL
  }

  # (cross-)fitting the g-functions:
  g_functions_cf <- NULL
  if (!is.null(folds) & cross_fit_g_models == TRUE){
    g_cf <- crossfit_function(
      policy_data = policy_data,
      fun = fit_g_functions,
      models = g_models,
      full_history = g_full_history,
      folds = folds,
      save_cross_fit_models = save_cross_fit_models,
      future_args = future_args
    )
    g_functions_cf <- getElement(g_cf, "functions")
    g_values <- getElement(g_cf, "values")
    valid_ids <- getElement(g_cf, "valid_ids")
    rm(g_cf)
  } else {
    if (is.null(g_functions)){
      g_functions <- fit_g_functions(policy_data,
                                     g_models = g_models,
                                     full_history = g_full_history)
    }
    g_values <- predict(g_functions, policy_data)
  }

  # fitting g-functions for determining new realistic actions:
  if (alpha > 0){
    if (is.null(g_functions)){
      g_functions <- fit_g_functions(policy_data,
                                     g_models = g_models,
                                     full_history = g_full_history)
    }
  } else{
    # g-functions are not saved if alpha == 0:
    g_functions <- NULL
  }

  # (n X K) matrix with entries U_{i,k+1}:
  stage <- NULL
  R <- as.matrix(dcast(rewards[stage != 1],
                       id~stage,
                       value.var = "U")[, -c("id"), with = FALSE])
  rm(stage)

  # (n X K) matrix with entries g_k(A_k, H_k)
  g_A_values <- get_a_values(a = actions$A, action_set = action_set, g_values)
  G <- as.matrix(dcast(g_A_values, id ~ stage, value.var = "P")[, -c("id"), with = FALSE])

  g_cols <- paste("g_", action_set, sep = "")
  X_designs <- list()
  X_scales <- list()
  X <- list()
  AA <- list()
  RR <- list()
  pi <- list()
  for (k in K:1){
    # getting the policy history
    policy_history_k <- get_history(policy_data, stage = k, full_history = full_history)
    if (full_history == TRUE){
      vars <- policy_vars[[k]]
    } else{
      vars <- policy_vars
    }
    H <- get_H(policy_history_k, vars = vars)
    if (is.null(policy_vars)){
      policy_vars <- names(H)
    }

    ### constructing the inputs for owl:
    # getting the design of the history (X) for owl:
    design_k <- get_design(formula = ~., data = H)
    x <- design_k$x
    if ((ncol(x) == 1))
      stop("DTRlearn2 has a bug. H must be a matrix with ncol(H) > 1.")
    design_k$x <- NULL
    X_designs[[k]] <- design_k
    # scaling the history
    x <- scale(x)
    X[[k]] <- x
    X_scales[[k]] <- attributes(x)[c("scaled:center", "scaled:scale")]

    stage_action_set <- stage_action_sets[[k]]
    # formatting the actions as {-1, 1}:
    aa <- A <- get_A(policy_history_k)
    aa[A == stage_action_set[1]] <- -1
    aa[A == stage_action_set[2]] <- 1
    aa <- as.numeric(aa)
    AA[[k]] <- aa

    # setting the rewards:
    RR[[k]] <- R[,k]

    # setting the action probabilities:
    pi[[k]] <- G[,k]
  }

  owl_object <- DTRlearn2::owl(H = X,
                               AA = AA,
                               RR = RR,
                               pi = pi,
                               K = K,
                               n = n,
                               res.lasso=res.lasso,
                               loss=loss,
                               kernel=kernel,
                               augment=augment,
                               c=c,
                               sigma=sigma,
                               s=s,
                               m=m)

  out <- list(
    owl_object = owl_object,
    reuse_scales = reuse_scales,
    X_scales = X_scales,
    X_designs = X_designs,
    g_functions = g_functions,
    g_functions_cf = g_functions_cf,
    full_history = full_history,
    policy_vars = policy_vars,
    action_set = action_set,
    stage_action_sets = stage_action_sets,
    K = K
  )
  class(out) <- c("owl","policy_object","list")

  return(out)
}

#' @export
get_policy.owl <- function(object, threshold = NULL){
  if (!(is.null(threshold) || identical(threshold, 0))) {
    stop("threshold is not implemented for owl.")
  }
  owl_object <- getElement(object, "owl_object")
  reuse_scales <- getElement(object, "reuse_scales")
  X_scales <- getElement(object, "X_scales")
  X_designs <- getElement(object, "X_designs")
  g_functions <- getElement(object, "g_functions")
  full_history <- getElement(object, "full_history")
  policy_vars <- getElement(object, "policy_vars")
  stage_action_sets <- getElement(object, "stage_action_sets")
  K <- getElement(object, "K")

  policy <- function(policy_data){
    if (get_K(policy_data) != K)
      stop("The policy do not have the same number of stages as the policy data object.")

    id_stage <- get_id_stage(policy_data)
    if(!(length(unlist(unique(id_stage[,.N, by = "id"][, "N"]))) == 1))
      stop("owl is only implemented for a fixed number of stages.")

    X <- list()
    for (k in K:1){
      # getting the policy history:
      policy_history_k <- get_history(policy_data, stage = k, full_history = full_history)

      if (full_history == TRUE){
        vars <- policy_vars[[k]]
      } else{
        vars <- policy_vars
      }
      # getting the design matrix:
      H <- get_H(policy_history_k, vars = vars)
      design <- X_designs[[k]]
      x <- apply_design(design, data = H)
      if (reuse_scales == TRUE){
        x <- scale(x,
                   center = X_scales[[k]]$`scaled:center`,
                   scale = X_scales[[k]]$`scaled:scale`)
      } else{
        x <- scale(x)
      }

      X[[k]] <- x
    }

    pred <- predict(owl_object, H = X, K = K)
    policy_actions <- get_id_stage(policy_data)
    stage <- NULL
    d <- NULL
    for (k in K:1){
      stage_action_set <- stage_action_sets[[k]]
      dd <- d_ <- pred$treatment[[k]]
      d_[dd == -1] <- stage_action_set[1]
      d_[dd == 1] <- stage_action_set[2]
      policy_actions[stage == k, d := d_]
    }
    rm(stage)
    setkeyv(policy_actions, c("id", "stage"))

    return(policy_actions)
  }
  class(policy) <- c("policy", "function")
  return(policy)
}

rql<- function(policy_data, alpha,
               g_models, g_functions, g_full_history,
               q_models, q_full_history,
               L,
               ...){
  K <- get_K(policy_data)
  n <- get_n(policy_data)
  action_set <- get_action_set(policy_data)
  stage_action_sets <- get_stage_action_sets(policy_data)

  # input checks:
  if (L != 1)
    stop("L must be 1 when type = 'ql' (no cross-fitting).")
  if (!is.null(g_functions)){
    if(!inherits(g_functions, what = "g_functions"))
      stop("g-functions must be of class 'g_functions'.")
  }
  if (is.list(q_models)){
    if (length(q_models) != K)
      stop("q_models must either be a list of length K or a single Q-model.")
  }
  if (any(get_element(policy_data, "cens_indicator")[["indicator"]])){
    stop("policy learning with type 'rql' not implemented under right-censoring/missing outcomes.")
  }

  # getting the IDs:
  id <- get_id(policy_data)

  # getting the observed (complete) utilities:
  utility <- get_utility(policy_data)

  # (n) vector with entries U_i:
  U <- utility$U

  # fitting the g-functions if alpha > 0:
  if (alpha != 0){
    if (is.null(g_functions)){
      g_functions <- fit_g_functions(policy_data,
                                     g_models = g_models,
                                     full_history = g_full_history)
    }
    g_values <- predict(g_functions, policy_data)
  }

  # (n X K+1) matrix with entries Q_k(H_{k,i}, d_k(H_{k,i})), Q_{K+1} = U:
  Q <- matrix(nrow = n, ncol = K+1)
  Q[, K+1] <- U

  q_cols <- paste("Q_", action_set, sep = "")
  g_cols <- paste("g_", action_set, sep = "")

  q_functions <- list()
  for (k in K:1){
    # fitting the Q-function:
    q_step_k <- q_step(
      policy_data = policy_data,
      k = k,
      full_history = q_full_history,
      Q = Q[, k+1],
      q_models = q_models
    )
    # getting the Q-function, Q-function values and the ID-index:
    q_functions[[k]] <- q_step_k$q_function
    q_values_k <- q_step_k$q_values
    idx_k <- q_step_k$idx_k

    if (alpha != 0){
      # getting the g-function values for each action:
      stage <- NULL
      g_values_k <- g_values[stage == k, ]
      rm(stage)

      # calculating the realistic actions:
      realistic_actions <- t(apply(g_values_k[, g_cols, with = FALSE], MARGIN = 1, function(x) x >= alpha))
      realistic_actions[which(realistic_actions == FALSE)] <- NA

      # getting the maximal realistic Q-function values:
      q_max_realistic_k <- apply(q_values_k[, q_cols, with = FALSE] * realistic_actions, MARGIN = 1, max, na.rm = TRUE)
    } else {
      q_max_realistic_k <- apply(q_values_k[, q_cols, with = FALSE], MARGIN = 1, max, na.rm = TRUE)
    }

    # inserting the maximal Q-function values in Q:
    Q[idx_k, k] <- q_max_realistic_k
    Q[!idx_k, k] <- Q[!idx_k, k+1]
  }
  names(q_functions) <- paste("stage_", 1:K, sep = "")
  class(q_functions) <- "nuisance_functions"
  attr(q_functions, "full_history") <- q_full_history

  # checking if deterministic rewards has been applied:
  deterministic_rewards <- policy_data[["colnames"]][["deterministic_rewards"]]
  event <- NULL
  indicator_deterministic_rewards <- !all(policy_data[["stage_data"]][event == 0,deterministic_rewards, with = FALSE] == 0)

  out <- list(
    q_functions = q_functions,
    g_functions = g_functions,
    action_set = action_set,
    stage_action_sets = stage_action_sets,
    alpha = alpha,
    K = K,
    indicator_deterministic_rewards = indicator_deterministic_rewards,
    deterministic_rewards = deterministic_rewards
  )
  out <- remove_null_elements(out)
  class(out) <- c("ql", "policy_object", "list")

  return(out)
}

#' @export
get_policy.ql <- function(object, threshold = NULL) {
  if (!(is.null(threshold) || identical(threshold, 0))) {
    stop("threshold is not implemented for ql.")
  }
  g_functions <- get_g_functions(object)
  q_functions <- get_q_functions(object)
  action_set <- getElement(object, "action_set")
  alpha <- getElement(object, "alpha")

  g_cols <- paste("g_", action_set, sep = "")
  q_cols <- paste("Q_", action_set, sep = "")

  policy <- function(policy_data) {
    # evaluating the Q-functions:
    q_values <- predict(q_functions, policy_data)

    if (alpha != 0) {
      # evaluating the g-functions:
      g_values <- predict(g_functions, policy_data)
      # calculating the realistic actions:
      realistic_actions <- t(apply(
        g_values[, g_cols, with = FALSE],
        MARGIN = 1,
        function(x) x >= alpha
      ))
      realistic_actions[which(realistic_actions == FALSE)] <- NA

      # getting the action with the maximal realistic Q-function value:
      idx <- apply(
        q_values[, q_cols, with = FALSE] * realistic_actions,
        MARGIN = 1,
        which.max
      )
      d <- action_set[idx]
    } else{
      # getting the action with the maximal Q-function value:
      idx <- apply(
        q_values[, q_cols, with = FALSE],
        MARGIN = 1,
        which.max
      )
      d <- action_set[idx]
    }

    # collecting the policy actions
    policy_actions <- get_id_stage(policy_data)
    policy_actions[, d := d]

    return(policy_actions)
  }

  # setting class and attributes:
  policy <- new_policy(policy, name = "ql")

  return(policy)
}

#' @rdname get_policy_functions
#' @export
get_policy_functions.ql <- function(object,
                                    stage,
                                    threshold = NULL,
                                    include_g_values = FALSE,
                                    ...) {
  K <- getElement(object, "K")
  check_stage(stage = stage, K = K)
  action_set <- getElement(object, "action_set")
  stage_action_sets <- getElement(object, "stage_action_sets")
  stage_action_set <- stage_action_sets[[stage]]
  rm(stage_action_sets)
  alpha <- getElement(object, "alpha")

  # getting the g-functions:
  g_functions <- getElement(object, "g_functions")
  if (!is.null(g_functions)) {
    if (length(g_functions) == K) {
      g_function <- g_functions[[stage]]
    } else {
      g_function <- g_functions[[1]]
    }
  }

  # getting the (residual) Q-functions:
  q_functions <- getElement(object, "q_functions")
  q_model <- getElement(q_functions[[stage]], "q_model")

  # getting the names of the deterministic reward variables:
  deterministic_rewards <- getElement(object, "deterministic_rewards")
  indicator_deterministic_rewards <- getElement(
    object,
    "indicator_deterministic_rewards"
  )

  stage_policy <- function(H) {
    if (is.data.frame(H)) {
      H <- as.data.table(H)
    }

    # adding zero valued deterministic reward columns if missing
    if (all(!(deterministic_rewards %in% colnames(H)))) {
      if (indicator_deterministic_rewards == TRUE) {
        mes <- paste0(
          "deterministic reward variables are missing in 'H' - the",
          " default value is set to zero."
        )
        warning(mes)
      }
      H[, deterministic_rewards] <- 0
    }
    if (!all(deterministic_rewards %in% colnames(H))) {
      tmp <- deterministic_rewards[!(deterministic_rewards %in% colnames(H))]
      H[, tmp] <- 0
      rm(tmp)
    }

    # (residual) Q-model predictions for the stage action set:
    q_values <- sapply(
      stage_action_set,
      function(a) predict(q_model, new_AH = cbind(A = a, H))
    )

    # adding the deterministic rewards
    q_values <- H[
      , deterministic_rewards[action_set %in% stage_action_set],
      with = FALSE
    ] + q_values
    colnames(q_values) <- stage_action_set

    # getting
    g_values <- NULL
    if (alpha == 0) {
      dd <- apply(q_values,
        MARGIN = 1,
        which.max
      )
      d <- stage_action_set[dd]
    } else {
      # evaluating the g-function:
      if (!all(g_function$H_names %in% names(H))) {
        mes <- paste(
          "H must contain the columns",
          paste(g_function$H_names, collapse = ","),
          "."
        )
        stop(mes)
      }
      g_values <- predict(g_function$g_model, new_H = H)

      # calculating the realistic actions:
      realistic_actions <- t(apply(
        g_values,
        MARGIN = 1, function(x) x >= alpha
      ))
      if (any(apply(realistic_actions, MARGIN = 1, sum) == 0)) {
        mes <- paste0(
          "Cases with no realistic actions occur.",
          " Consider resetting the alpha level."
        )
        stop(mes)
      }
      realistic_actions[which(realistic_actions == FALSE)] <- NA

      # getting the action with the maximal realistic Q-function value:
      dd <- apply(q_values * realistic_actions, MARGIN = 1, which.max)
      d <- stage_action_set[dd]
    }

    # including the g_values as attributes:
    if (include_g_values == TRUE) {
      attr(d, "g_values") <- g_values
    }

    return(d)
  }

  return(stage_policy)
}

new_policy_data <- function(stage_data,
                            baseline_data = NULL,
                            action_set = NULL,
                            stage_action_sets = NULL,
                            verbose) {
  ## checking and processing stage_data:
  if (missing(stage_data)) stage_data <- NULL
  if (is.null(stage_data)) stop("'stage_data' is missing or NULL.")
  if (is.data.frame(stage_data)) stage_data <- as.data.table(stage_data)
  if (!is.data.table(stage_data))
    stop("'stage_data' must be a data.table or a data.frame.")
  if (!all(c("id", "stage", "event", "A", "U") %in% colnames(stage_data)))
    stop("'stage_data' must contain id, stage, event, A (action) and U (utility/reward).")
  setcolorder(stage_data, c("id", "stage", "event", "time", "time2", "A"))

  ## setting and checking keys:
  setkeyv(stage_data, c("id", "stage"))
  if (any(is.na(stage_data[, "id"]))) stop("id has missing values.")
  if (any(is.na(stage_data[, "stage"]))) stop("stage has missing values.")
  if (anyDuplicated(stage_data, by = key(stage_data)) > 0)
    stop("The combination of id & stage must be unique.")
  stage <- NULL
  if (!all(stage_data[ , list("check" = all(stage == 1:.N)), by = "id"][,"check"]))
    stop("stage must be on the form 1, 2, ..., k.")
  rm(stage)

  ## checking the event variable:
  if (any(is.na(stage_data[,"event"]))) stop("event have missing values.")
  event <- NULL
  if (!all(stage_data[,list("check" = all(event == c(rep(0, times = (.N-1)), 1) | event == c(rep(0, times = (.N-1)), 2))),by = "id"][, "check"]))
    stop("event must be on the form 0,0,...,0,j (j in {1,2}).")
  rm(event)

  ## getting the dimensions:
  n <- length(unique(unlist(stage_data[,"id"])))
  stage <- NULL
  K <- stage_data[event == 0, list(max(stage))][[1]]
  rm(stage)

  time_indicator <- !all(is.na(stage_data[["time"]]))
  time2_indicator <- !all(is.na(stage_data[["time2"]]))
  ## checking the time and time2 variable:
  if (time_indicator == TRUE) {
    if (any(is.na(stage_data[["time"]]))) {
      stop("time has missing values")
    }
  }
  if (time2_indicator == TRUE) {
    if (any(is.na(stage_data[["time2"]]))) {
      stop("time2 has missing values")
    }
  }
  if ((time_indicator == TRUE) & (time2_indicator == TRUE)) {
    if (!all(stage_data[["time"]] <= stage_data[["time2"]])) {
      stop("time2 must be greater than or equal to time for all pairwise elements.")
    }
  }
  ## if time is provided, but time2 is not
  ## [time, time2) intervals are created by
  ## shifting time (with 0 as fill):
  if ((time_indicator == TRUE) & (time2_indicator == FALSE)){
    time <- time2 <- NULL
    stage_data[ , time2 := time]
    stage_data[ , time := shift(time, fill = 0), by = id]
  }

  ## checking the action variable (A):
  event <- NULL
  A <- NULL
  if (any(is.na(stage_data[event == 0, ][,"A"])))
    stop("'action' (A) has missing values.")
  if (!is.character(stage_data[,"A"])){
    stage_data[, A := as.character(A)]
  }
  rm(event, A)

  ## getting the global set of actions (A):
  obs_actions <- sort(unlist(unique(stage_data[,"A"])))
  if (!is.null(action_set)){
    action_set <- as.character(action_set)
    if (!all(obs_actions %in% action_set))
      stop("The given action set does not include all observed actions.")
    action_set <- sort(action_set)
  } else{
    action_set <- obs_actions
  }
  action_set <- unname(action_set)

  ## getting the action set at each stage:
  event <- NULL
  tmp <- stage_data[event == 0, list(set = list(sort(unique(A)))), stage]
  obs_stage_action_sets <- lapply(tmp$set, c)
  names(obs_stage_action_sets) <- paste("stage_", tmp$stage, sep = "")
  rm(event, tmp)
  if (!is.null(stage_action_sets)){
    if(length(stage_action_sets) != length(obs_stage_action_sets))
      stop("the given stage action set does not comply with the observed stage action sets.")
    for (k in seq_along(stage_action_sets)){
      if (!all(obs_stage_action_sets[[k]] %in% stage_action_sets[[k]]))
        stop("The given stage action sets do not include all observed stage actions.")
      stage_action_sets <- lapply(stage_action_sets, sort)
    }

  } else {
    stage_action_sets <- obs_stage_action_sets
  }

  ## checking the utility variable (U):
  if (!all(is.numeric(unlist(stage_data[,"U"]))))
    stop("'utility' U must be numeric.")
  ## for stage action events (event = 0) and termination events (event = 1)
  ## U should not be missing:
  if (any(is.na(stage_data[event %in% c(0,1),"U"])))
    stop("The utility variable U has missing values")
  if (any(is.na(stage_data[stage <= K,][event %in% c(2),"U"])))
    stop("The utility variable U has missing values")
  if (any(!is.na(stage_data[stage == (K+1) & (event %in% c(2)), "U"])))
    stop("The utility variable U must be NA for stage K+1 if event = 2.")

  ## checking the deterministic reward variables (U_A[.]):
  deterministic_reward_names <- paste("U_A", action_set, sep = "")
  missing_deterministic_reward_names <-
    deterministic_reward_names[!(deterministic_reward_names %in% names(stage_data))]
  if (length(missing_deterministic_reward_names) > 0){
    mes <- paste(missing_deterministic_reward_names, collapse = ", ")
    mes <- paste("Setting the deterministic rewards '", mes, "' to default value 0.", sep = "")
    if (verbose == TRUE){
      message(mes)
    }
    stage_data[(stage <= K), (missing_deterministic_reward_names) := 0]
  }
  if (!all(sapply(stage_data[, deterministic_reward_names, with = FALSE], function(col) is.numeric(col)))){
    mes <- paste(deterministic_reward_names, collapse = ", ")
    mes <- paste("'",mes, "'must be numeric.", sep = " ")
    stop(mes)
  }
  if(any(is.na(stage_data[stage <= K][event %in% c(0,2), deterministic_reward_names, with = FALSE])))
    stop("The deterministic reward variables have missing values.")

  ## coercing variables of type factor to type character in stage_data:
  sdf <- sapply(stage_data, function(x) is.factor(x))
  if (any(sdf)){
    f_names <- colnames(stage_data)[sdf]
    stage_data[, (f_names) := lapply(.SD, as.character), .SDcols = f_names]
    rm(f_names)
  }
  rm(sdf)

  ## getting the names of the state data (X_k):
  rn <- c("id", "stage", "event", "time", "time2", "A", "U", deterministic_reward_names)
  state_names <- names(stage_data)[!(names(stage_data) %in% rn)]

  ## checking and processing baseline_data:
  baseline_names <- NULL
  if (!is.null(baseline_data)){
    if (is.data.frame(baseline_data))
      baseline_data <- as.data.table(baseline_data)

    ## checking id:
    if (!all(c("id") %in% names(baseline_data))) stop("'baseline_data' must contain id.")
    if (any(is.na(baseline_data$id))) stop("'baseline_data' id has missing values")
    if (anyDuplicated(baseline_data, by = key(baseline_data)) > 0) stop("'baseline_data' id contains duplicates.")

    ## setting id as key:
    setkeyv(baseline_data, "id")

    ## comparing id:
    if (!all(unique(unlist(stage_data[, "id"])) == unlist(baseline_data[,"id"])))
      stop("'baseline_data' id must match 'stage_data' id.")

    ## coercing columns of type factor to type character in baseline_data:
    bdf <- sapply(baseline_data, function(x) is.factor(x))
    if (any(bdf)){
      f_names <- colnames(baseline_data)[bdf]
      baseline_data[, (f_names) := lapply(.SD, as.character), .SDcols = f_names]
      rm(f_names)
    }
    rm(bdf)

    ## getting the names of the baseline state data:
    baseline_names <- names(baseline_data)[!(names(baseline_data) %in% c("id"))]
  } else {
    id <- NULL
    baseline_data <- stage_data[, list(id = unique(id))]
    rm(id)
  }

  ## creating a right-censoring indicator for each stage (K+1):
  cens_indicator <- stage_data[, list(indicator = any(event == 2)), by = c("stage")]

  ## creating a terminal event indicator for each stage (K+1):
  terminal_indicator <- stage_data[, list(indicator = any(event == 1)), by = c("stage")]

  object <- list(
    stage_data = stage_data,
    baseline_data = baseline_data,
    colnames = list(
      state_names = state_names,
      deterministic_rewards = deterministic_reward_names,
      baseline_names = baseline_names
    ),
    action_set = action_set,
    stage_action_sets = stage_action_sets,
    dim = list(
      n = n,
      K = K
    ),
    cens_indicator = cens_indicator,
    terminal_indicator = terminal_indicator
  )

  class(object) <- "policy_data"

  return(object)
}

#' @title Create Policy Data Object
#'
#' @description  \code{policy_data()} creates a policy data object which
#' is used as input to [policy_eval()] and [policy_learn()] for policy
#' evaluation and data adaptive policy learning.
#'
#' @param data [data.frame] or [data.table::data.table]; see Examples.
#' @param baseline_data [data.frame] or [data.table::data.table]; see Examples.
#' @param type Character string. If "wide", \code{data} is considered to be on
#'   wide format. If "long", \code{data} is considered to be on long format; see
#'   Examples.
#' @param action Action variable name(s). Character vector or character string.
#'   \itemize{ \item A vector is valid for wide data. The length of the vector
#'   determines the number of stages (K). \item A string is valid for single
#'   stage wide data or long data. }
#' @param covariates Stage specific covariate name(s). Character vector or named
#'   list of character vectors. \itemize{ \item A vector is valid for single
#'   stage wide data or long data. \item A named list is valid for multiple
#'   stages wide data. Each element must be a character vector with length K.
#'   Each vector can contain NA elements, if a covariate is not available for
#'   the given stage(s). }
#' @param utility Utility/Reward variable name(s). Character string or vector.
#'   \itemize{ \item A string is valid for long data and wide data with a single
#'   final utility. \item A vector is valid for wide data with incremental
#'   rewards. Must have length K+1; see Examples. }
#' @param baseline Baseline covariate name(s). Character vector.
#' @param deterministic_rewards Deterministic reward variable name(s). Named
#'   list of character vectors of length K. The name of each element must be on
#'   the form "U_Aa" where "a" corresponds to an action in the action set.
#' @param id ID variable name. Character string.
#' @param stage Stage number variable name.
#' @param event Event indicator name.
#' @param time Character string
#' @param time2 Character string
#' @param action_set Character string. Action set across all stages.
#' @param verbose Logical. If TRUE, formatting comments are printed to the
#'   console.
#' @param digits Minimum number of digits to be printed.
#' @param x Object to be printed.
#' @param object Object of class [policy_data]
#' @param probs numeric vector (probabilities)
#' @param ... Additional arguments passed to print.
#' @details Each observation has the sequential form \deqn{O= {B, U_1, X_1, A_1,
#'   ..., U_K, X_K, A_K, U_{K+1}},} for a possibly stochastic number of stages
#'   K. \itemize{ \item \eqn{B} is a vector of baseline covariates. \item
#'   \eqn{U_k} is the reward at stage k (not influenced by the action
#'   \eqn{A_k}). \item \eqn{X_k} is a vector of state covariates summarizing the
#'   state at stage k. \item \eqn{A_k} is the categorical action at stage k. }
#'   The utility is given by the sum of the rewards, i.e., \eqn{U = \sum_{k =
#'   1}^{K+1} U_k}.
#' @returns \code{policy_data()} returns an object of class "policy_data". The
#'   object is a list containing the following elements:
#'   \item{\code{stage_data}}{[data.table::data.table] containing the id, stage
#'   number, event indicator, action (\eqn{A_k}), state covariates (\eqn{X_k}),
#'   reward (\eqn{U_k}), and the deterministic rewards.}
#'   \item{\code{baseline_data}}{[data.table::data.table] containing the id and
#'   baseline covariates (\eqn{B}).} \item{\code{colnames}}{List containing the
#'   state covariate names, baseline covariate names, and the deterministic
#'   reward variable names.} \item{\code{action_set}}{Sorted character vector
#'   describing the action set, i.e., the possible actions at all stages.}
#'   \item{\code{stage_action_sets}}{List of sorted character vectors describing
#'   the observed actions at each stage.} \item{\code{dim}}{List containing the
#'   number of observations (n) and the number of stages (K).}
#' @section S3 generics: The following S3 generic functions are available for an
#'   object of class \code{policy_data}: \describe{ \item{[partial()]}{ Trim the
#'   maximum number of stages in a \code{policy_data} object.}
#'   \item{[subset_id()]}{ Subset a a \code{policy_data} object on ID.}
#'   \item{[get_history()]}{ Summarize the history and action at a given stage.}
#'   \item{[get_history_names()]}{ Get history variable names.}
#'   \item{[get_actions()]}{ Get the action at every stage.}
#'   \item{[get_utility()]}{Get the utility.} \item{[plot()]}{Plot method.} }
#' @seealso [policy_eval()], [policy_learn()], [copy_policy_data()]
#' @examples
#' library("polle")
#' ### Single stage: Wide data
#' d1 <- sim_single_stage(n = 5e2, seed=1)
#' head(d1, 5)
#' # constructing policy_data object:
#' pd1 <- policy_data(d1,
#'                    action="A",
#'                    covariates=c("Z", "B", "L"),
#'                    utility="U")
#' pd1
#' # associated S3 methods:
#' methods(class = "policy_data")
#' head(get_actions(pd1), 5)
#' head(get_utility(pd1), 5)
#' head(get_history(pd1)$H, 5)
#'
#' ### Two stage: Wide data
#' d2 <- sim_two_stage(5e2, seed=1)
#' head(d2, 5)
#' # constructing policy_data object:
#' pd2 <- policy_data(d2,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd2
#' head(get_history(pd2, stage = 2)$H, 5) # state/Markov type history and action, (H_k,A_k).
#' head(get_history(pd2, stage = 2, full_history = TRUE)$H, 5) # Full history and action, (H_k,A_k).
#'
#' ### Multiple stages: Long data
#' d3 <- sim_multi_stage(5e2, seed = 1)
#' head(d3$stage_data, 10)
#' # constructing policy_data object:
#' pd3 <- policy_data(data = d3$stage_data,
#'                    baseline_data = d3$baseline_data,
#'                    type = "long",
#'                    id = "id",
#'                    stage = "stage",
#'                    event = "event",
#'                    action = "A",
#'                    utility = "U")
#' pd3
#' head(get_history(pd3, stage = 3)$H, 5) # state/Markov type history and action, (H_k,A_k).
#' head(get_history(pd3, stage = 2, full_history = TRUE)$H, 5) # Full history and action, (H_k,A_k).
#' @export
policy_data <- function(data, baseline_data,
                        type="wide",
                        action, covariates, utility,
                        baseline = NULL,
                        deterministic_rewards = NULL,
                        id = NULL, stage = NULL, event = NULL,
                        time = NULL, time2 = NULL,
                        action_set = NULL,
                        verbose = FALSE) {
  data <- check_data(data)
  data <- copy(data)
  if (!missing(baseline_data)){
    baseline_data <- check_data(baseline_data)
    baseline_data <- copy(baseline_data)
  }

  ## input checks:
  if(!(is.character(type) & (length(type == 1))))
    stop("'type' must be either \"wide\" or \"long\".")
  type <- tolower(type)

  ## format wide and long data:
  if (any(type %in% c("wide"))) { # wide data
    ## input warnings and checks:
    if (!missing(baseline_data)) {
      stop("When 'type'=wide' set 'baseline_data' to NULL and use 'baseline' instead.")
    }
    if (!is.null(time2)) {
      stop("time2 is not used when type = 'wide'. Please set to NULL.")
    }
    if (missing(covariates)) {
      covariates <- list()
    }

    # formatting the wide data:
    md <- melt_wide_data(
      data,
      id = id,
      action = action,
      time = time,
      covariates = covariates,
      utility = utility,
      baseline = baseline,
      deterministic_rewards = deterministic_rewards
    )

    # constructing new policy_data object:
    pd <- new_policy_data(
      stage_data = md$stage_data,
      baseline_data = md$baseline_data,
      action_set = action_set,
      verbose = verbose
    )
  } else if (any(type %in% "long")) { # long data
    ## input warnings and checks:
    if (!is.null(deterministic_rewards))
      warning("deterministic_rewards is not used when type = 'long'.")
    if (missing(covariates))
      covariates <- NULL
    if(!is.null(covariates))
      warning("covariates is not used when type = 'long'.")
    if (is.null(time) & !is.null(time2)) {
      stop("time2 provided, but time is NULL.")
    }

    ld <- format_long_data(
      data,
      baseline_data = baseline_data,
      id = id,
      action = action,
      stage = stage,
      event = event,
      time = time,
      time2 = time2,
      utility = utility,
      verbose = verbose
    )

    # constructing policy data object
    pd <- new_policy_data(
      stage_data = ld$stage_data,
      baseline_data = ld$baseline_data,
      action_set = action_set,
      verbose = verbose
    )

  } else{
    stop("'type' must be either \"wide\" or \"long\".")
  }
  return(pd)
}

check_data <- function(data){
  name <- deparse(substitute(data))

  if (is.data.frame(data) | is.matrix(data)){
    data <- as.data.table(data)
  }
  if (!is.data.table(data)){
    mes <- paste("'", name, "' must be a data.table.", sep = "")
    stop(mes)
  }
  if (any(duplicated(names(data)))){
    mes <- paste("'", name, "' has duplicated variable names.", sep = "")
    stop(mes)
  }

  return(data)
}


# Wide data ---------------------------------------------------------------

melt_wide_data <- function(wide_data,
                           id,
                           action,
                           time,
                           covariates,
                           baseline,
                           utility,
                           deterministic_rewards){
  ## checking the form of the variable inputs:
  ## id:
  if (!is.null(id)){
    if (!(is.character(id) & (length(id) == 1))){
      mes <- "'id' must be a character string."
      stop(mes)
    }
  }
  ## time:
  if (!is.null(time)){
    if (!(is.character(time) & (length(time) == 1))){
      mes <- "'time' must be a character string."
      stop(mes)
    }
  }
  ## action:
  if (is.list(action))
    action <- unlist(action)
  if (!is.vector(action) | !is.character(action))
    stop("'action' must be a vector or a list of type character.")
  # getting the number of stages:
  K <- length(action)
  if(!all(action %in% colnames(wide_data))){
    mes <- paste(action[!(action %in% colnames(wide_data))], collapse = ",")
    mes <- paste("Action variables ", mes, " not found in data.", sep = "")
    stop(mes)
  }
  ## getting the action set:
  action_set <- sort(unique(unlist(wide_data[ , action, with = FALSE])))
  ## covariates:
  if (!(is.list(covariates) | is.vector(covariates)))
    stop("'covariates' must be a character vector or a list of character vectors.")
  covariates <- lapply(covariates, function(covar){
    if (is.list(covar))
      covar <- unlist(covar)
    if (!is.vector(covar) | !is.character(covar))
      stop("'covariates' must be a character vector or a list of character vectors.")
    if(length(covar) != K){
      mes <- paste("Each element in 'covariates' must have length ", K, ".", sep = "")
      stop(mes)
    }
    return(covar)
  })
  if (is.null(names(covariates))){
    if (K == 1)
      names(covariates) <- covariates
    else
      stop("'covariates' must be a named list in case of multiple actions.")
  }
  ## baseline:
  if (!is.null(baseline)){
    baseline <- unlist(baseline)
    if (!is.character(baseline)){
      stop("'baseline' must be a character string or vector.")
    }
  }
  ## utility:
  if (is.list(utility))
    utility <- unlist(utility)
  if (!is.vector(utility) | !is.character(utility))
    stop("'utility' must be a vector or a list of type character.")
  if (length(utility) != 1){
    if (length(utility) != (K+1)){
      mes <- "'utility' must either be a character string or a character vector of length "
      mes <- paste(mes, (K+1), ".", sep = "")
      stop(mes)
    }
  }
  ## deterministic_rewards:
  if (!is.null(deterministic_rewards)){
    deterministic_rewards <- lapply(deterministic_rewards, function(du){
      if (is.list(du))
        du <- unlist(du)
      if (!is.vector(du) | !is.character(du))
        stop("'deterministic_rewards' must be a character vector or a list of character vectors.")
      if(length(du) != K){
        mes <- paste("Each element in 'deterministic_rewards' must have length ", K, ".", sep = "")
        stop(mes)
      }
      return(du)
    })
    deterministic_reward_names <- paste("U_A", action_set, sep = "")
    if ((!all(names(deterministic_rewards) %in% deterministic_reward_names)) | (!is.null(deterministic_rewards) & is.null(names(deterministic_rewards)))) {
      mes <- paste(deterministic_reward_names, collapse = "', '")
      mes <- paste("'deterministic_rewards' must be a named list with names in the set '", mes, "'.", sep = "")
      stop(mes)
    }
  }

  ## checking id:
  if (!is.null(id)){
    if(!(id %in% colnames(wide_data))){
      stop("invalid id variable (not in data)")
    }
    if (anyDuplicated(wide_data[[id]])){
      stop("'id' column in data must be unique.")
    }
  }

  ## checking time:
  if (!is.null(time)){
    if(!(time %in% colnames(wide_data))){
      stop("invalid time variable (not in data)")
    }
    if (anyDuplicated(wide_data[[time]])){
      stop("'time' column in data must be unique.")
    }
  }

  ## checking if 'data' contains all non-NA variable names:
  tmp <- unlist(c(id, action, covariates, baseline, utility, deterministic_rewards))
  tmp <- tmp[!is.na(tmp)]
  if (!all(tmp %in% names(wide_data))){
    mes <- tmp[!(tmp %in% names(wide_data))]
    mes <- paste(mes, collapse = "\", \"")
    mes <- paste("Variables not found in data: \"", mes, "\".", sep = "")
    stop(mes)
  }
  rm(tmp)

  ## checking for non-NA duplicates:
  tmp <- unlist(c(id, action, covariates, baseline, utility, deterministic_rewards))
  tmp <- tmp[!is.na(tmp)]
  if (anyDuplicated(tmp)>0){
    mes <- tmp[anyDuplicated(tmp)]
    mes <- paste(mes, collapse = "\", \"")
    mes <- paste("Duplicated variables: \"", mes, "\".", sep = "")
    stop(mes)
  }
  rm(tmp)

  ## checking for invalid variable names:
  if (any(c("event", "stage", "time", "time2") %in% names(covariates))){
    stop("'covariates' can not have named elements \"event\", \"stage\", \"time\", or \"time2\".")
  }

  ## setting default variables if missing:
  ## setting id if missing:
  if (is.null(id)){
    if (any("id" %in% colnames(wide_data))){
      stop("'data' has a variable id, but 'id' = NULL. Please set 'id' = \"id\" or change the name of the id variable.")
    }
    # setting id:
    wide_data[, id := 1:.N]
    id <- "id"
  }

  ## setting the rewards to 0 in case the final utility is provided:
  if (length(utility)==1) {
    for (i in seq(K)) {
      u <- paste0("_", utility[i], "_", i)
      stopifnot(!any(u %in% names(wide_data)))
      wide_data[, (u) := 0]
      utility <- c(u, utility)
    }
  }

  ## setting default time(s) at stages k <= K:
  if (!is.null(time)) {
    for (i in seq(K)) {
      tt <- paste0("_", time, "_", i)
      stopifnot(!any(tt %in% names(wide_data)))
      wide_data[, (tt) := 0]
      time <- c(tt, time)
    }
  }

  ## augmenting wide_data to handle NA covariates entries:
  for (l in seq_along(covariates)){
    na_idx <- is.na(covariates[[l]])
    if (any(na_idx)){
      if (all(na_idx)){
        mes <- names(covariates[l])
        mes <- paste("covariate", mes, 'is invalid.')
        stop(mes)
      }
      ## augmenting wide_data with NA column (of the same class):
      NA_col_name <- paste("_NA_", l, sep = "")
      wide_data[, c(NA_col_name) := wide_data[[covariates[[l]][!na_idx][[1]]]]]
      wide_data[, c(NA_col_name) := NA]
      ## updating the covariate names:
      covariates[[l]][na_idx] <- NA_col_name
    }
  }

  ## melting stage data:
  stage <- A <- U <- event <- NULL
  measure <- append(list("A" = action), covariates)
  if (!is.null(time)) {
    measure <- append(measure, list("time" = time))
  }
  measure <- append(measure, list("U" = utility))
  if (!is.null(deterministic_rewards)){
    measure <- append(measure, deterministic_rewards)
  }
  # selecting subset:
  sel <- unlist(c(id, action, time, covariates, utility, deterministic_rewards))
  sel <- sel[!is.na(sel)]
  stage_data <- subset(wide_data, select = sel)
  # converts to long data:
  stage_data <- melt(stage_data, id.vars = id, measure.vars = measure, variable.name = "stage")
  setnames(stage_data, id, "id")

  ## setting the format of stage and A:
  stage_data[ , stage := as.integer(as.character(stage))]
  stage_data[ , A := as.character(A)]
  # setting the event variable:
  stage_data[!is.na(A), event := 0]
  stage_data[is.na(A), event := 1]
  stage_data[(is.na(A) & is.na(U)), event := 2]

  ## setting time to NA if missing.
  ## setting time2 to NA
  if (is.null(time)) {
    set(stage_data, j = "time", value = as.numeric(NA))
  }
  set(stage_data, j = "time2", value = as.numeric(NA))

  # setting keys:
  setkeyv(stage_data, c("id", "stage"))
  rm(stage, A, U, event)

  ## getting baseline data:
  if (!is.null(baseline)){
    sel <- c(id, baseline)
    baseline_data <- subset(wide_data, select = sel)
    setnames(baseline_data, id, "id")
    setkeyv(baseline_data, "id")
  } else{
    baseline_data <- NULL
  }

  return(list(stage_data = stage_data, baseline_data = baseline_data))
}


# Long data ---------------------------------------------------------------

# sets default by reference:
set_default_name <- function(data, variable_name, default_name, create_if_null = FALSE){
  parsed_name <- deparse(substitute(variable_name))

  if (is.null(variable_name)){
    variable_name <- default_name
  }

  if (!(is.character(variable_name) & (length(variable_name) == 1))){
    mes <- parsed_name
    mes <- paste("'", mes, "' must be a character string.", sep = "")
    stop(mes)
  }

  if (!(variable_name %in% names(data))){
    if (create_if_null == TRUE) {
      data[ , c(variable_name) := as.numeric(NA)]
    } else {
      mes <- parsed_name
      mes <- paste("'", mes, "' is invalid.", sep = "")
      stop(mes)
    }
  }

  if ((default_name %in% names(data)) & (variable_name != default_name)){
    mes <- paste("'data' has a variable called \"", default_name, "\", but '", sep = "")
    mes <- paste(mes, parsed_name, "' = \"", variable_name, "\".", sep = "")
    mes <- paste(mes,  " Please remove or rename '", parsed_name, "'.", sep ="")
    stop(mes)
  }

  setnames(data, variable_name, default_name)
}

format_long_data <- function(long_data, baseline_data, id, action, stage, event, time, time2, utility, verbose){

  ## dealing with missing arguments:
  if (missing(action)){
    action <- NULL
  }
  if (missing(utility)){
    utility <- NULL
  }
  if (missing(baseline_data)){
    baseline_data <- NULL
  }

  ## checking for duplicates
  tmp <- unlist(c(id, action, stage, event, utility, time, time2))
  if (anyDuplicated(tmp)>0){
    mes <- tmp[anyDuplicated(tmp)]
    mes <- paste(mes, collapse = "\", \"")
    mes <- paste("Duplicated variables: \"", mes, "\".", sep = "")
    stop(mes)
  }
  rm(tmp)

  ### setting default variable names by reference in long_data:
  set_default_name(long_data, variable_name = id, default_name = "id")
  set_default_name(long_data, variable_name = action, default_name = "A")
  set_default_name(long_data, variable_name = stage, default_name = "stage")
  set_default_name(long_data, variable_name = event, default_name = "event")
  set_default_name(long_data, variable_name = utility, default_name = "U")
  set_default_name(long_data, variable_name = time, default_name = "time", create_if_null = TRUE)
  set_default_name(long_data, variable_name = time2, default_name = "time2", create_if_null = TRUE)

  ### searching for deterministic reward variables in long_data:
  action_set <- sort(unique(unlist(long_data[,"A", with = FALSE])))
  deterministic_reward_names <- paste("U_A", action_set, sep = "")
  drn <- names(long_data)[names(long_data) %in% deterministic_reward_names]
  if ((length(drn) > 0) & (verbose == TRUE)){
    mes <- paste("The variables ", paste(drn, collapse = ", "), " in 'data' are considered deterministic rewards.", sep = "")
    message(mes)
  }

  if (!is.null(baseline_data)){
    set_default_name(baseline_data, variable_name = id, default_name = "id")
  }

  out <- list(
    stage_data = long_data,
    baseline_data = baseline_data
  )
  return(out)
}

test_that("policy_data checks checks time and time2 input.", {

  ## long data
  ## 3 cases: no right censoring, right censored before/at stage 1 action,
  ## right censored utility outcome (at stage 2)
  ld <- data.table(
    id = c(1,1,2,3,3),
    stage = c(1,2,1,1,2),
    event = c(0,1,2,0,2),
    A = c("0", NA, NA, "1", NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3"),
    Z = c("A", NA, "A", "B", NA),
    L = c(1, 2, 2, 1, 3),
    time = c(1, 2, 0.5, 1, 1.5),
    U = c(0, 10, NA, 0, NA),
    U_A0 = c(0,0,NA,0,NA),
    U_A1 = c(0,0,NA,0,NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  ## baseline data:
  bd <- data.table(
    id = c(1,2,3),
    W = c("blue", "red", "blue")
  )
  setkey(bd, id)

  ## input checks:
  expect_error(
    pd <- policy_data(data = ld,
                      baseline_data = bd,
                      type = "long",
                      time = "time1"),
    "'data' has a variable called \"time\", but 'time' = \"time1\". Please remove or rename 'time'."
  )

  expect_error(
    policy_data(data = ld[stage == 1],
                action = "A",
                utility = "U",
                covariates = c("Z", "time"),
                id = "id",
                type = "wide",
                time = NULL),
    "'covariates' can not have named elements \"event\", \"stage\", \"time\", or \"time2\"."
  )

  ## checking time and time2 input:
  ld <- data.table(
    id = c(1,1,2,3,3),
    stage = c(1,2,1,1,2),
    event = c(0,1,2,0,2),
    A = c("0", NA, NA, "1", NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3"),
    Z = c("A", NA, "A", "B", NA),
    L = c(1, 2, 2, 1, 3),
    timetest = c(1, 2, 0.5, 1, 1.5),
    U = c(0, 10, 0, 0, NA),
    U_A0 = c(0,0,0,0,NA),
    U_A1 = c(0,0,0,0,NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  expect_error(
    pd <- policy_data(data = ld,
                      baseline_data = bd,
                      type = "long",
                      time = "timetest"),
    NA
  )
  expect_equal(
    pd$stage_data$time2,
    ld$timetest
  )
  expect_equal(
    pd$stage_data$time,
    c(0,1,0,0,1)
  )

  ## checking time and time2 input:
  ld <- data.table(
    id = c(1,1,2,3,3),
    stage = c(1,2,1,1,2),
    event = c(0,1,2,0,2),
    A = c("0", NA, NA, "1", NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3"),
    Z = c("A", NA, "A", "B", NA),
    L = c(1, 2, 2, 1, 3),
    time = c(1, 2, 0.5, NA, 1.5),
    U = c(0, 10,0, 0, NA),
    U_A0 = c(0,0,0,0,NA),
    U_A1 = c(0,0,0,0,NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  expect_error(
    pd <- policy_data(data = ld,
                      baseline_data = bd,
                      type = "long",
                      time = "time"),
    "time has missing values"
  )

  ld <- data.table(
    id = c(1,1,2,3,3),
    stage = c(1,2,1,1,2),
    event = c(0,1,2,0,2),
    A = c("0", NA, NA, "1", NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3"),
    Z = c("A", NA, "A", "B", NA),
    L = c(1, 2, 2, 1, 3),
    time = c(1, 2, 0.5, 1, 1.5),
    time2 = c(2, 2.2, 1, NA, 2),
    U = c(0, 10, 0, 0, NA),
    U_A0 = c(0,0,0,0,NA),
    U_A1 = c(0,0,0,0,NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  expect_error(
    pd <- policy_data(data = ld,
                      baseline_data = bd,
                      type = "long",
                      time = "time",
                      time2 = "time2"),
    "time2 has missing values"
  )

  ld <- data.table(
    id = c(1,1,2,3,3),
    stage = c(1,2,1,1,2),
    event = c(0,1,2,0,2),
    A = c("0", NA, NA, "1", NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3"),
    Z = c("A", NA, "A", "B", NA),
    L = c(1, 2, 2, 1, 3),
    time = c(1, 2, 0.5, 1, 1.5),
    time2 = c(2, 2.2, 1, 0.8, 2),
    U = c(0, 10, 0, 0, NA),
    U_A0 = c(0,0,0,0,NA),
    U_A1 = c(0,0,0,0,NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  expect_error(
    pd <- policy_data(data = ld,
                      baseline_data = bd,
                      type = "long",
                      time = "time",
                      time2 = "time2"),
    "time2 must be greater than or equal to time for all pairwise elements."
  )


})

test_that("policy_data formats right censored long data correctly in the single stage case.", {

  # long data
  ## 3 cases: no right censoring, right censored before/at stage 1 action,
  ## right censored utility outcome (at stage 2)
  ld <- data.table(
    id = c(1,1,2,3,3),
    stage = c(1,2,1,1,2),
    event = c(0,1,2,0,2),
    time = c(0, 1, 0, 0, 1),
    time2 = c(1, 2, 0.5, 1, 1.5),
    A = c("0", NA, NA, "1", NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3"),
    Z = c("A", NA, "A", "B", NA),
    L = c(1, 2, 2, 1, 3),
    U = c(0, 10, 0, 0, NA),
    U_A0 = c(0,0,0,0,NA),
    U_A1 = c(0,0,0,0,NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  ## baseline data:
  bd <- data.table(
    id = c(1,2,3),
    W = c("blue", "red", "blue")
  )
  setkey(bd, id)


  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    type = "long",
                    time = "time")

  ## stage_data and baseline_data should not change:
  expect_equal(ld, pd$stage_data)
  expect_equal(bd, pd$baseline_data)

})


test_that("policy_data formats right censored wide data correctly in the single stage case (single final outcome)", {

  ## stage 1 covariates and a single utility outcome
  ref_ld <- data.table(
    id = c(1,1,2,2,3,3),
    stage = c(1,2,1,2,1,2),
    event = c(0,1,0,2,0,2),
    time = c(0,0,0,0,0,0),
    time2 = c(0,2,0,1,0,1.5),
    A = c("0", NA, "1", NA, "1", NA),
    B = c("gr1", NA, "gr2", NA, "gr3", NA),
    Z = c("A", NA, "A", NA, "B", NA),
    L = c(1, NA, 2, NA, 3, NA),
    U = c(0, 10, 0, NA, 0, NA),
    U_A0 = c(0,NA,0,NA,0,NA),
    U_A1 = c(0,NA,0,NA,0,NA)
  )
  setkey(ref_ld, id,stage)
  setindex(ref_ld, event)

  wd <- data.table(
    id = c(1,2,3),
    A = c("0","1","1"),
    B = c("gr1","gr2","gr3"),
    Z = c("A","A","B"),
    L = c(1,2,3),
    U = c(10, NA, NA),
    time = c(2,1,1.5)
  )

  pd <- policy_data(data = wd,
                    type = "wide",
                    id = "id",
                    action = "A",
                    covariates = c("B", "Z", "L"),
                    utility = "U",
                    time = "time")

  expect_equal(
    pd$stage_data,
    ref_ld,
    check.attributes = FALSE
  )

  ## ## vector of covarites and utility
  ## ref_ld <- data.table(
  ##   id = c(1,1,2,2,3,3),
  ##   stage = c(1,2,1,2,1,2),
  ##   event = c(0,1,0,2,0,2),
  ##   time = c(0,0,0,0,0,0),
  ##   time2 = c(0,2,0,1,0,1.5),
  ##   A = c("0", NA, "1", NA, "1", NA),
  ##   B = c("gr1", NA, "gr2", NA, "gr3", NA),
  ##   Z = c("A", NA, "A", NA, "B", NA),
  ##   L = c(1, 4, 2, 5, 3, 6), # stage 2 values non-NA
  ##   U = c(5, 10, 7, NA, 3, NA), # stage 1 utility contributions
  ##   U_A0 = c(0,NA,0,NA,0,NA),
  ##   U_A1 = c(0,NA,0,NA,0,NA)
  ## )
  ## setkey(ref_ld, id,stage)
  ## setindex(ref_ld, event)

  ## wd <- data.table(
  ##   id = c(1,2,3),
  ##   A = c("0","1","1"),
  ##   B = c("gr1","gr2","gr3"),
  ##   Z = c("A","A","B"),
  ##   L = c(1,2,3),
  ##   L2 = c(3,5,6),
  ##   U = c(10, NA, NA),
  ##   time = c(2,1,1.5)
  ## )

  ## pd <- policy_data(data = wd,
  ##                   type = "wide",
  ##                   id = "id",
  ##                   action = "A",
  ##                   covariates = list(
  ##                     B = c("B", NA),
  ##                     Z = c("Z", NA),
  ##                     L = c("L", "L2")
  ##                   ),
  ##                   utility = "U",
  ##                   time = "time")

  ## expect_equal(
  ##   pd$stage_data,
  ##   ref_ld,
  ##   check.attributes = FALSE
  ## )

})

test_that("policy_g_functions returns a policy which selects the most probable action.",{
  d <- sim_single_stage(2e3, seed=1)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")


  g_functions <- fit_g_functions(pd, g_models = g_glm())
  pol <- polle:::policy_g_functions(g_functions = g_functions)

  expect_true(
    inherits(pol, "policy")
  )

  tmp <- predict(g_functions, pd)
  tmp2 <- as.character(as.numeric((tmp[,4] > tmp[,3])))

  expect_equal(
    pol(pd)$d,
    tmp2
  )
})

test_that("q_xgboost tunes parameters", {
  # data
  d <- sim_single_stage(200, seed=1)
  # policy data
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  set.seed(1)
  q_fun <- polle:::fit_Q_functions(pd,
                                   policy_actions = policy_def(1)(pd),
                                   q_models = q_xgboost(max_depth = c(2, 3),
                                                        learning_rate = c(0.3, 0.5),
                                                        nrounds = c(2, 4)))

  expect_equal(
    length(q_fun$stage_1$q_model$cv_res$cv),
    3 * 2 * 2 * 2 * 2
  )

})

test_that("q_xgboost gives the same result as plain xgboost",{
  library("xgboost")
  # data
  d <- sim_single_stage(200, seed=1)
  x <- d[,c("Z", "L", "B", "A")]
  y = d[["U"]]

  set.seed(1)
  bst <- xgboost::xgboost(x = x,
                          y = y,
                          max_depth = 2,
                          learning_rate = 1,
                          min_child_weight = 2,
                          nrounds = 2,
                          verbosity = 0)

  pred <- predict(bst, newdata = x)

  # policy data
  pd <- policy_data(d,
                    action = "A",
                    covariates = list("Z", "B", "L"),
                    utility = "U")

  set.seed(1)
  q_fun <- polle:::fit_Q_functions(pd,
                                   policy_actions = policy_def(1)(pd),
                                   q_models = q_xgboost(max_depth = 2,
                                                        learning_rate = 1,
                                                        nrounds = 2,
                                                        min_child_weight = 2))

  pred_polle <- predict(q_fun, pd)

  expect_equal(
    unname(pred),
    pred_polle[["Q_1"]] * (d$A == "1") + pred_polle[["Q_0"]] * (d$A == "0")
  )

})

test_that("get_policy.earl returns a policy", {
  library("DynTxRegime")

  d <- sim_single_stage(200, seed=1)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  moPropen1 <- buildModelObj(model = ~B+Z+L,
                             solver.method = 'glm',
                             solver.args = list('family'='binomial'),
                             predict.method = 'predict.glm',
                             predict.args = list(type='response'))

  pl <- policy_learn(type = "earl",
                     control = control_earl(moPropen = moPropen1,
                                            regime = ~B+Z+L,))
  expect_error({
    p <- get_policy(pl(pd, q_models = q_glm(), g_models = g_glm()))
  },
  NA
  )

  expect_true(
    inherits(p, what = "policy")
  )

  expect_error(
    p(pd),
    NA
  )
})

test_that("the polle implementation of earl agrees with direct application of DynTxRegime::earl in the single stage case.",{
  library("DynTxRegime")

  d1 <- sim_single_stage(200, seed=1)
  pd1 <- policy_data(d1,
                     action="A",
                     covariates = list("Z", "B", "L"),
                     utility="U")

  # direct application:
  moPropen1 <- buildModelObj(model = ~B+Z+L,
                             solver.method = 'glm',
                             solver.args = list('family'='binomial'),
                             predict.method = 'predict.glm',
                             predict.args = list(type='response'))

  moMain1 <- buildModelObj(model = ~B+Z+L,
                           solver.method = 'lm')

  moCont1 <- buildModelObj(model = ~B+Z+L,
                           solver.method = 'lm')

  set.seed(1)
  dir <- DynTxRegime::earl(moPropen = moPropen1,
                           moMain = moMain1,
                           moCont = moCont1,
                           response = d1$U,
                           data = d1,
                           txName = "A",
                           lambdas = c(0.5, 1, 2),
                           regime = ~B+Z+L,
                           cvFolds = 3,
                           verbose = 0)

  # polle application:
  pl <- policy_learn(type = "earl",
                     control = control_earl(moPropen = moPropen1,
                                            moMain = moMain1,
                                            moCont = moCont1,
                                            regime = ~B+Z+L,
                                            verbose = 0,
                                            lambdas = c(0.5, 1, 2),
                                            cvFolds = 3))
  set.seed(1)
  po <- pl(policy_data = pd1)

  # comparison
  expect_equal(
    dir@analysis@optimal@estimatedValue,
    po$earl_object@analysis@optimal@estimatedValue
  )
})

test_that("the polle implementation is robust in respect to the action set.",{
  d1 <- sim_single_stage(200, seed=1)
  d2 <- d1

  d2$A[d1$A == 0] <- "B"
  d2$A[d1$A == 1] <- "A"

  pd1 <- policy_data(d1,
                     action="A",
                     covariates = list("Z", "B", "L"),
                     utility="U")
  pd2 <- policy_data(d2,
                     action="A",
                     covariates = list("Z", "B", "L"),
                     utility="U")

  # earl specification
  moPropen1 <- buildModelObj(model = ~B+Z+L,
                             solver.method = 'glm',
                             solver.args = list('family'='binomial'),
                             predict.method = 'predict.glm',
                             predict.args = list(type='response'))

  moMain1 <- buildModelObj(model = ~B+Z+L,
                           solver.method = 'lm')

  moCont1 <- buildModelObj(model = ~B+Z+L,
                           solver.method = 'lm')

  pl <- policy_learn(type = "earl",
                     control = control_earl(moPropen = moPropen1,
                                            moMain = moMain1,
                                            moCont = moCont1,
                                            regime = ~B+Z+L,
                                            verbose = 0,
                                            lambdas = c(0.5, 1, 2),
                                            cvFolds = 3))
  set.seed(1)
  po1 <- pl(policy_data = pd1)
  set.seed(1)
  po2 <- pl(policy_data = pd1)

  expect_equal(
    po1$earl_object@analysis@optimal@estimatedValue,
    po2$earl_object@analysis@optimal@estimatedValue
  )

  expect_equal(
    po1$earl_object@analysis@optimal@optimalTx,
    po2$earl_object@analysis@optimal@optimalTx
  )

})

test_that("earl handles missing arguments", {
  library("DynTxRegime")

  d1 <- sim_single_stage(200, seed=1)
  pd1 <- policy_data(d1,
                     action="A",
                     covariates = list("Z", "B", "L"),
                     utility="U")

  pl <- policy_learn(type = "earl",
                     control = control_earl(regime = ~B+Z+L,
                                            verbose = 0,
                                            lambdas = c(0.5, 1, 2),
                                            cvFolds = 3))
  set.seed(1)
  # moPropen is required:
  expect_error(pl(policy_data = pd1))
})

test_that("q_sl with discreteSL = TRUE picks the learner with the lowest cvrisk.",{
  library("polle")
  d <- sim_single_stage(200, seed=1)
  d$BB <- sample(c("group 1", "group & 2", "group & 3"), size = 200, replace = TRUE)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L", "BB"),
                    utility="U")
  p <- policy_def(1)

  set.seed(1)
  qfun <- polle:::fit_Q_functions(pd,
                                  p(pd),
                                  q_sl(SL.library = c("SL.ranger", "SL.glm"),
                                       discreteSL = TRUE))

  expect_true(
    all(qfun$stage_1$q_model$model$coef %in% c(0,1)),
    sum(qfun$stage_1$q_model$model$coef) == 1
  )

})

test_that("fit_m_function has the expected output.", {

  ## right-censoring occur at stage 3:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  m <- fit_m_function(policy_data = pd, m_model = q_glm(~ x), full_history = FALSE)
  m_ref <- glm(y ~ x3, data = d[delta1 == 1 & delta2 == 1 &  delta3 == 1 ])

  expect_equal(
    unname(coef(m$m_model$model)),
    unname(coef(m_ref))
  )

  m <- fit_m_function(policy_data = pd, m_model = q_glm(~ x_1 + A_1 + x_2 + x_3), full_history = TRUE)
  m_ref <- glm(y ~ x1 + as.factor(a1) + x2 + x3, data = d[delta1 == 1 & delta2 == 1 &  delta3 == 1 ])

  expect_equal(
    unname(coef(m$m_model$model)),
    unname(coef(m_ref))
  )

  ## right-censoring does NOT occur at stage 3:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rep(1, times = 1e2) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  m <- fit_m_function(policy_data = pd, m_model = q_glm(~ x), full_history = FALSE)

  expect_equal(
    m,
    NULL
  )

  m <- fit_m_function(policy_data = pd, m_model = q_glm(~ x), full_history = TRUE)

  expect_equal(
    m,
    NULL
  )

})

test_that("fit_m_function handles missing covariates (throws and error message that is related to the m-model)", {

  ## right-censoring occur at stage 3:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  x3[1:10] <- as.numeric(NA)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  expect_error(
    m <- fit_m_function(policy_data = pd, m_model = q_glm(~ x), full_history = FALSE),
    "Error fitting m_model: NA/NaN/Inf in 'x' when calling 'q_glm' with formula:\nV_res ~ x"
  )


})

#' Policy Evaluation
#'
#' \code{policy_eval()} is used to estimate
#' the value of a given fixed policy
#' or a data adaptive policy (e.g. a policy
#' learned from the data). \code{policy_eval()}
#' is also used to estimate the average
#' treatment effect among the subjects who would
#' get the treatment under the policy.
#' @param policy_data Policy data object created by [policy_data()].
#' @param policy Policy object created by [policy_def()].
#' @param policy_learn Policy learner object created by [policy_learn()].
#' @param g_functions Fitted g-model objects, see [nuisance_functions].
#' Preferably, use \code{g_models}.
#' @param g_models List of action probability models/g-models for each stage
#' created by [g_empir()], [g_glm()], [g_rf()], [g_sl()] or similar functions.
#' Only used for evaluation if \code{g_functions} is \code{NULL}.
#' If a single model is provided and \code{g_full_history} is \code{FALSE},
#' a single g-model is fitted across all stages. If \code{g_full_history} is
#' \code{TRUE} the model is reused at every stage.
#' @param g_full_history If TRUE, the full history is used to fit each g-model.
#' If FALSE, the state/Markov type history is used to fit each g-model.
#' @param save_g_functions If TRUE, the fitted g-functions are saved.
#' @param q_functions Fitted Q-model objects, see [nuisance_functions].
#' Only valid if the Q-functions are fitted using the same policy.
#' Preferably, use \code{q_models}.
#' @param q_models Outcome regression models/Q-models created by
#' [q_glm()], [q_rf()], [q_sl()] or similar functions.
#' Only used for evaluation if \code{q_functions} is \code{NULL}.
#' If a single model is provided, the model is reused at every stage.
#' @param q_full_history Similar to g_full_history.
#' @param save_q_functions Similar to save_g_functions.
#' @param c_functions Fitted c-model/censoring probability model objects. Preferably, use \code{c_models}.
#' @param c_models List of right-censoring probability models, see [c_model].
#' @param c_full_history Similar to g_full_history.
#' @param save_c_functions Similar to save_g_functions.
#' @param m_function Fitted outcome model object for stage K+1. Preferably, use \code{m_model}.
#' @param m_model Outcome model for the utility at stage K+1. Only used if the final utility
#' contribution is missing/has been right-censored
#' @param m_full_history Similar to g_full_history.
#' @param save_m_function Similar to save_g_functions.
#' @param target Character string. Either "value" or "subgroup". If "value",
#' the target parameter is the policy value.
#' If "subgroup", the target parameter
#' is the average treatement effect among
#' the subgroup of subjects that would receive
#' treatment under the policy, see details.
#' "subgroup" is only implemented for \code{type = "dr"}
#' in the single-stage case with a dichotomous action set.
#' @param type Character string. Type of evaluation. Either \code{"dr"}
#' (doubly robust),
#' \code{"ipw"} (inverse propensity weighting),
#' or \code{"or"} (outcome regression).
#' @param cross_fit_type Character string.
#' Either "stacked", or "pooled", see details. (Only used if \code{M > 1} and target = "subgroup")
#' @param variance_type Character string. Either "pooled" (default),
#' "stacked" or "complete", see details. (Only used if \code{M > 1})
#' @param M Number of folds for cross-fitting.
#' @param nrep Number of repetitions of cross-fitting (estimates averaged over repeated cross-fittings)
#' @param min_subgroup_size Minimum number of observations in the evaluated subgroup (Only used if target = "subgroup").
#' @param future_args Arguments passed to [future.apply::future_apply()].
#' @param name Character string.
#' @param object,x,y Objects of class "policy_eval".
#' @param labels Name(s) of the estimate(s).
#' @param paired \code{TRUE} indicates that the estimates are based on
#' the same data sample.
#' @param digits Integer. Number of printed digits.
#' @param width Integer. Width of printed parameter name.
#' @param std.error Logical. Should the std.error be printed.
#' @param level Numeric. Level of confidence limits.
#' @param p.value Logical. Should the p.value for associated confidence level be printed.
#' @param ... Additional arguments.
#' @return \code{policy_eval()} returns an object of class "policy_eval".
#' The object is a list containing the following elements:
#' \item{\code{coef}}{Numeric vector. The estimated target parameter:
#' policy value or subgroup average treatment effect.}
#' \item{\code{IC}}{Numeric matrix. Estimated influence curve associated with
#' \code{coef}.}
#' \item{\code{type}}{Character string. The type of evaluation ("dr", "ipw",
#' "or").}
#' \item{\code{target}}{Character string. The target parameter ("value" or "subgroup")}
#' \item{\code{id}}{Character vector. The IDs of the observations.}
#' \item{\code{name}}{Character vector. Names for the each element in \code{coef}.}
#' \item{\code{coef_ipw}}{(only if \code{type = "dr"}) Numeric vector.
#' Estimate of \code{coef} based solely on inverse probability weighting.}
#' \item{\code{coef_or}}{(only if \code{type = "dr"}) Numeric vector.
#' Estimate of \code{coef} based solely on outcome regression.}
#' \item{\code{policy_actions}}{[data.table::data.table] with keys id and stage. Actions
#' associated with the policy for every observation and stage.}
#' \item{\code{policy_object}}{(only if \code{policy = NULL} and \code{M = 1})
#' The policy object returned by \code{policy_learn}, see [policy_learn].}
#' \item{\code{g_functions}}{(only if \code{M = 1}) The
#' fitted g-functions. Object of class "nuisance_functions".}
#' \item{\code{g_values}}{The fitted g-function values.}
#' \item{\code{q_functions}}{(only if \code{M = 1}) The
#' fitted Q-functions. Object of class "nuisance_functions".}
#' \item{\code{q_values}}{The fitted Q-function values.}
#' \item{\code{Z}}{(only if \code{target = "subgroup"})
#' Matrix with the doubly robust stage 1 scores for each action.}
#' \item{\code{subgroup_indicator}}{(only if \code{target = "subgroup"})
#' Logical matrix identifying subjects in the subgroup.
#' Each column represents a different subgroup threshold.}
#' \item{\code{cross_fits}}{(only if \code{M > 1}) List containing the
#' "policy_eval" object for every (validation) fold.}
#' \item{\code{folds}}{(only if \code{M > 1}) The (validation) folds used
#' for cross-fitting.}
#' \item{\code{cross_fit_type}}{Character string.}
#' \item{\code{variance_type}}{Character string.}
#' @section S3 generics:
#' The following S3 generic functions are available for an object of
#' class \code{policy_eval}:
#' \describe{
#' \item{[get_g_functions()]}{ Extract the fitted g-functions.}
#' \item{[get_q_functions()]}{ Extract the fitted Q-functions.}
#' \item{[get_policy()]}{ Extract the fitted policy object.}
#' \item{[get_policy_functions()]}{Extract the fitted policy function for a given stage.}
#' \item{[get_policy_actions()]}{ Extract the (fitted) policy actions.}
#' \item{[plot.policy_eval()]}{Plot diagnostics.}
#' }
#' @seealso [lava::IC], [lava::estimate.default].
#' @details
#' Each observation has the sequential form
#' \deqn{O= {B, U_1, X_1, A_1, ..., U_K, X_K, A_K, U_{K+1}},}
#' for a possibly stochastic number of stages K.
#' \itemize{
#'  \item \eqn{B} is a vector of baseline covariates.
#'  \item \eqn{U_k} is the reward at stage k
#' (not influenced by the action \eqn{A_k}).
#'  \item \eqn{X_k} is a vector of state
#' covariates summarizing the state at stage k.
#'  \item \eqn{A_k} is the categorical action
#' within the action set \eqn{\mathcal{A}} at stage k.
#' }
#' The utility is given by the sum of the rewards, i.e.,
#' \eqn{U = \sum_{k = 1}^{K+1} U_k}.
#'
#' A policy is a set of functions
#' \deqn{d = \{d_1, ..., d_K\},}
#' where \eqn{d_k} for \eqn{k\in \{1, ..., K\}}
#' maps \eqn{\{B, X_1, A_1, ..., A_{k-1}, X_k\}} into the
#' action set.
#'
#' Recursively define the Q-models (\code{q_models}):
#' \deqn{Q^d_K(h_K, a_K) = E[U|H_K = h_K, A_K = a_K]}
#' \deqn{Q^d_k(h_k, a_k) = E[Q_{k+1}(H_{k+1},
#' d_{k+1}(B,X_1, A_1,...,X_{k+1}))|H_k = h_k, A_k = a_k].}
#' If \code{q_full_history = TRUE},
#' \eqn{H_k = \{B, X_1, A_1, ..., A_{k-1}, X_k\}}, and if
#' \code{q_full_history = FALSE}, \eqn{H_k = \{B, X_k\}}.
#'
#' The g-models (\code{g_models}) are defined as
#' \deqn{g_k(h_k, a_k) = P(A_k = a_k|H_k = h_k).}
#' If \code{g_full_history = TRUE},
#' \eqn{H_k = \{B, X_1, A_1, ..., A_{k-1}, X_k\}}, and if
#' \code{g_full_history = FALSE}, \eqn{H_k = \{B, X_k\}}.
#' Furthermore, if \code{g_full_history = FALSE} and \code{g_models} is a
#' single model, it is assumed that \eqn{g_1(h_1, a_1) = ... = g_K(h_K, a_K)}.
#'
#' If \code{target = "value"} and \code{type = "or"}
#' \code{policy_eval()} returns the empirical estimate of
#' the value (\code{coef}):
#' \deqn{E\left[Q^d_1(H_1, d_1(\cdot))\right]}
#'
#' If \code{target = "value"} and \code{type = "ipw"} \code{policy_eval()}
#' returns the empirical estimates of
#' the value (\code{coef}) and influence curve (\code{IC}):
#' \deqn{E\left[\left(\prod_{k=1}^K I\{A_k = d_k(\cdot)\}
#' g_k(H_k, A_k)^{-1}\right) U\right].}
#' \deqn{\left(\prod_{k=1}^K I\{A_k =
#' d_k(\cdot)\} g_k(H_k, A_k)^{-1}\right) U -
#' E\left[\left(\prod_{k=1}^K
#' I\{A_k = d_k(\cdot)\} g_k(H_k, A_k)^{-1}\right) U\right].}
#'
#' If \code{target = "value"} and
#' \code{type = "dr"} \code{policy_eval} returns the empirical estimates of
#' the value (\code{coef}) and influence curve (\code{IC}):
#' \deqn{E[Z_1(d,g,Q^d)(O)],}
#' \deqn{Z_1(d, g, Q^d)(O) - E[Z_1(d,g, Q^d)(O)],}
#' where
#' \deqn{
#' Z_1(d, g, Q^d)(O) = Q^d_1(H_1 , d_1(\cdot)) +
#' \sum_{r = 1}^K \prod_{j = 1}^{r}
#' \frac{I\{A_j = d_j(\cdot)\}}{g_{j}(H_j, A_j)}
#' \{Q_{r+1}^d(H_{r+1} , d_{r+1}(\cdot)) - Q_{r}^d(H_r , d_r(\cdot))\}.
#' }
#'
#' If \code{target = "subgroup"}, \code{type = "dr"}, \code{K = 1},
#' and \eqn{\mathcal{A} = \{0,1\}}, \code{policy_eval()}
#' returns the empirical estimates of the subgroup average
#' treatment effect (\code{coef}) and influence curve (\code{IC}):
#' \deqn{E[Z_1(1,g,Q)(O) - Z_1(0,g,Q)(O) | d_1(\cdot) = 1],}
#' \deqn{\frac{1}{P(d_1(\cdot) = 1)} I\{d_1(\cdot) = 1\}
#' \Big\{Z_1(1,g,Q)(O) - Z_1(0,g,Q)(O) - E[Z_1(1,g,Q)(O)
#' - Z_1(0,g,Q)(O) | d_1(\cdot) = 1]\Big\}.}
#'
#' Applying \eqn{M}-fold cross-fitting using the \{M\} argument, let
#' \deqn{\mathcal{Z}_{1,m}(a) = \{Z_1(a, g_m, Q_m^d)(O): O\in \mathcal{O}_m \}.}
#'
#' If \code{target = "subgroup"}, \code{type = "dr"}, \code{K = 1},
#' \eqn{\mathcal{A} = \{0,1\}}, and \code{cross_fit_type = "pooled"},
#' \code{policy_eval()} returns the estimate \deqn{\frac{1}{{N^{-1} \sum_{i =
#' 1}^N I\{d(H_i) = 1\}}} N^{-1} \sum_{m=1}^M \sum_{(Z, H) \in \mathcal{Z}_{1,m}
#' \times \mathcal{H}_{1,m}} I\{d_1(H) = 1\} \left\{Z(1)-Z(0)\right\}} If
#' \code{cross_fit_type = "stacked"} the returned estimate is \deqn{M^{-1}
#' \sum_{m = 1}^M \frac{1}{{n^{-1} \sum_{h \in \mathcal{H}_{1,m}} I\{d(h) =
#' 1\}}} n^{-1} \sum_{(Z, H) \in \mathcal{Z}_{1,m} \times \mathcal{H}_{1,m}}
#' I\{d_1(H) = 1\} \left\{Z(1)-Z(0)\right\},} where for ease of notation we let
#' the integer \eqn{n} be the number of oberservations in each fold.
#' @references
#' van der Laan, Mark J., and Alexander R. Luedtke.
#' "Targeted learning of the mean outcome under an optimal dynamic treatment rule."
#' Journal of causal inference 3.1 (2015): 61-95.
#' \doi{10.1515/jci-2013-0022}
#' \cr \cr
#' Tsiatis, Anastasios A., et al. Dynamic
#' treatment regimes: Statistical methods for precision medicine. Chapman and
#' Hall/CRC, 2019. \doi{10.1201/9780429192692}.
#' \cr \cr
#' Victor Chernozhukov, Denis
#' Chetverikov, Mert Demirer, Esther Duflo, Christian Hansen, Whitney Newey,
#' James Robins, Double/debiased machine learning for treatment and structural
#' parameters, The Econometrics Journal, Volume 21, Issue 1, 1 February 2018,
#' Pages C1–C68, \doi{10.1111/ectj.12097}.
#' @export
#' @examples
#' library("polle")
#' ### Single stage:
#' d1 <- sim_single_stage(5e2, seed=1)
#' pd1 <- policy_data(d1,
#'                    action = "A",
#'                    covariates = list("Z", "B", "L"),
#'                    utility = "U")
#' pd1
#'
#' # defining a static policy (A=1):
#' pl1 <- policy_def(1)
#'
#' # evaluating the policy:
#' pe1 <- policy_eval(policy_data = pd1,
#'                    policy = pl1,
#'                    g_models = g_glm(),
#'                    q_models = q_glm(),
#'                    name = "A=1 (glm)")
#'
#' # summarizing the estimated value of the policy:
#' # (equivalent to summary(pe1)):
#' pe1
#' coef(pe1) # value coefficient
#' sqrt(vcov(pe1)) # value standard error
#'
#' # getting the g-function and Q-function values:
#' head(predict(get_g_functions(pe1), pd1))
#' head(predict(get_q_functions(pe1), pd1))
#'
#' # getting the fitted influence curve (IC) for the value:
#' head(IC(pe1))
#'
#' # evaluating the policy using random forest nuisance models:
#' set.seed(1)
#' pe1_rf <- policy_eval(policy_data = pd1,
#'                       policy = pl1,
#'                       g_models = g_rf(),
#'                       q_models = q_rf(),
#'                       name = "A=1 (rf)")
#'
#' # merging the two estimates (equivalent to pe1 + pe1_rf):
#' (est1 <- merge(pe1, pe1_rf))
#' coef(est1)
#' head(IC(est1))
#'
#' ### Two stages:
#' d2 <- sim_two_stage(5e2, seed=1)
#' pd2 <- policy_data(d2,
#'                    action = c("A_1", "A_2"),
#'                    covariates = list(L = c("L_1", "L_2"),
#'                                      C = c("C_1", "C_2")),
#'                    utility = c("U_1", "U_2", "U_3"))
#' pd2
#'
#' # defining a policy learner based on cross-fitted doubly robust Q-learning:
#' pl2 <- policy_learn(
#'    type = "drql",
#'    control = control_drql(qv_models = list(q_glm(~C_1),
#'                                            q_glm(~C_1+C_2))),
#'    full_history = TRUE,
#'    L = 2) # number of folds for cross-fitting
#'
#' # evaluating the policy learner using 2-fold cross fitting:
#' pe2 <- policy_eval(type = "dr",
#'                    policy_data = pd2,
#'                    policy_learn = pl2,
#'                    q_models = q_glm(),
#'                    g_models = g_glm(),
#'                    M = 2, # number of folds for cross-fitting
#'                    name = "drql")
#' # summarizing the estimated value of the policy:
#' pe2
#'
#' # getting the cross-fitted policy actions:
#' head(get_policy_actions(pe2))
policy_eval <- function(policy_data,
                        policy = NULL, policy_learn = NULL,
                        g_functions = NULL, g_models = g_glm(),
                        g_full_history = FALSE, save_g_functions = TRUE,
                        q_functions = NULL, q_models = q_glm(),
                        q_full_history = FALSE, save_q_functions = TRUE,
                        c_functions = NULL, c_models = NULL,
                        c_full_history = FALSE, save_c_functions = TRUE,
                        m_function = NULL, m_model = NULL,
                        m_full_history = FALSE, save_m_function = TRUE,
                        target = "value",
                        type = "dr",
                        cross_fit_type = "pooled",
                        variance_type = "pooled",
                        M = 1,
                        nrep = 1,
                        min_subgroup_size = 1,
                        future_args = list(future.seed=TRUE),
                        name = NULL
                        ) {
  ## argument input checks:
  args <- as.list(environment())
  args <- do.call(what = "policy_eval_input_checks", args)

  ## collecting the arguments to be passed on:
  args[["policy_data"]] <- NULL
  args[["M"]] <- NULL
  args[["train_block_size"]] <- NULL
  args[["future_args"]] <- NULL
  args[["variance_type"]] <- NULL
  args[["cross_fit_type"]] <- NULL
  args[["nrep"]] <- NULL

  if (M > 1) {
    if (nrep == 1){
      eval <- policy_eval_cross(
        args = args,
        policy_data = policy_data,
        M = M,
        cross_fit_type = cross_fit_type,
        variance_type = variance_type,
        future_args = future_args
      )
    } else {
      ## repeated cross-fitting
      eval <- policy_eval_rep(
        nrep = nrep,
        args = args,
        policy_data = policy_data,
        M = M,
        cross_fit_type = cross_fit_type,
        variance_type = variance_type,
        future_args = future_args
      )
    }

  } else {
    args[["train_policy_data"]] <- policy_data
    args[["valid_policy_data"]] <- policy_data
    eval <- do.call(what = policy_eval_type, args = args)
  }

  return(eval)
}

model_input_checks <- function(policy_data,
                               g_functions, g_models,
                               g_full_history, save_g_functions,
                               q_functions, q_models,
                               q_full_history, save_q_functions,
                               c_functions, c_models,
                               c_full_history, save_c_functions,
                               m_function, m_model,
                               m_full_history, save_m_function) {
  check_function <- function(func, class_name){
    if (!is.null(func) && !inherits(func, class_name)) {
      stop(paste(class_name, " must be of class '", class_name, "'.", sep=""))
    }
  }

  # check function classes:
  check_function(g_functions, "g_functions")
  check_function(q_functions, "q_functions")
  check_function(c_functions, "c_functions")
  check_function(m_function, "m_function")

  # check logical flags:
  logical_checks <- function(flag, name) {
    if (!(is.logical(flag) && length(flag) == 1)) {
      stop(paste(name, "must be TRUE or FALSE.", sep = " "))
    }
  }

  logical_checks(g_full_history, "g_full_history")
  logical_checks(q_full_history, "q_full_history")
  logical_checks(c_full_history, "c_full_history")
  logical_checks(m_full_history, "m_full_history")

  if ((is.null(g_functions) && is.null(g_models))) {
    stop("Provide either g_functions or g_models.")
  }
  if ((is.null(q_functions) && is.null(q_models))) {
    stop("Provide either q_functions or q_models.")
  }

  cens_indicator <- get_element(policy_data, "cens_indicator")
  terminal_indicator <- get_element(policy_data, "terminal_indicator")
  if (any(cens_indicator[["indicator"]])){
    if ((is.null(c_models) && is.null(c_functions))) {
      stop("Right-censoring events (event = 2) occur in the policy data. Please provide either c_functions or c_models.")
    }
  }
  missing <- cens_indicator[get("stage") == (get_K(policy_data)+1),][["indicator"]]
  if (missing == TRUE) {
    if ((is.null(m_model) && is.null(m_function))) {
      stop("Right-censoring events (event = 2) occur at stage K+1 in the policy data. Please provide either m_function or m_model.")
    }
  }
}

policy_eval_input_checks <- function(policy_data,
                                     policy, policy_learn,
                                     g_functions, g_models,
                                     g_full_history, save_g_functions,
                                     q_functions, q_models,
                                     q_full_history, save_q_functions,
                                     c_functions, c_models,
                                     c_full_history, save_c_functions,
                                     m_function, m_model,
                                     m_full_history, save_m_function,
                                     target,
                                     type,
                                     cross_fit_type,
                                     variance_type,
                                     M,
                                     nrep,
                                     min_subgroup_size,
                                     future_args = list(),
                                     name,
                                     train_block_size = 1){
  args <- as.list(environment())
  if (!inherits(policy_data, what = "policy_data"))
    stop("policy_data must be of inherited class 'policy_data'.")
  if (!is.null(policy)) {
    if (!inherits(policy, what = "policy")) {
      stop("policy must be of inherited class 'policy'.")
    }
  }

  model_input_checks(policy_data = policy_data,
                     g_functions = g_functions, g_models = g_models,
                     g_full_history = g_full_history, save_g_functions = save_g_functions,
                     q_functions = q_functions, q_models = q_models,
                     q_full_history = q_full_history, save_q_functions = save_q_functions,
                     c_functions = c_functions, c_models = c_models,
                     c_full_history = c_full_history, save_c_functions = save_c_functions,
                     m_function = m_function, m_model = m_model,
                     m_full_history = m_full_history, save_m_function = save_m_function)

  if ((is.null(policy) && is.null(policy_learn)) ||
      (!is.null(policy_learn) && !is.null(policy))) {
    stop("Provide either policy or policy_learn.")
  }
  if (is.null(policy) && !is.null(policy_learn)) {
    if (!inherits(policy_learn, what = "policy_learn")) {
      stop("policy_learn must be of inherited class 'policy_learn'.")
    }
  }

  if (!(is.numeric(M) && (length(M) == 1))) {
    stop("M must be an integer greater than 0.")
  }
  if (!(M %% 1 == 0)) {
    stop("M must be an integer greater than 0.")
  }
  if (M <= 0) {
    stop("M must be an integer greater than 0.")
  }
  if (!(min_subgroup_size %% 1 == 0)) {
    stop("min_subgroup_size must be an integer greater than 0.")
  }
  if (min_subgroup_size <= 0) {
    stop("min_subgroup_size must be an integer greater than 0.")
  }
  if (!is.list(future_args)) {
    stop("future_args must be a list.")
  }
  target <- tolower(target)
  if (length(target) != 1) {
    stop("target must be a character string.")
  }
  if (target %in% c(
                    "value",
                    "policy_value"
                  )) {
    target <- "value"
  } else if (target %in% c(
                           "subgroup",
                           "sub_effect",
                           "subeffect",
                           "subvalue",
                           "sub_value"
                         )) {
    target <- "subgroup"
  } else {
    stop("target must be either 'value' or 'subgroup'.")
  }
  if (!is.null(name)) {
    name <- as.character(name)
    if (target == "value") {
      if (length(name) != 1) {
        stop("name must be a character string.")
      }
    }
    if (target == "subgroup") {
      if (length(name) != 2) {
        stop("name must be a character vector of length 2 when target = 'subgroup'.")
      }
    }
  }
  type <- tolower(type)
  if (length(type) != 1) {
    stop("type must be a character string.")
  }
  if (type %in% c("dr", "aipw")) {
    type <- "dr"
  } else if (type %in% c("ipw")) {
    type <- "ipw"
  } else if (type %in% c("or", "q")) {
    type <- "or"
  } else {
    stop("type must be either 'dr', 'ipw' or  'or'.")
  }

  ## editing name:
  if (is.null(name)) {
    if (target == "value") {
      name <- "E[U(d)]"
    }

    if (target == "subgroup") {
      as <- get_action_set(policy_data)
      name1 <- paste0("E[U(", as[2], ")-U(", as[1], ")|d=", as[2], "]")
      name2 <- paste0("E[U(", as[2], ")-U(", as[1], ")|d=", as[1], "]")
      name <- c(name1, name2)
      rm(as, name1, name2)
    }
  }

  ## train_block_size used for online/sequential validation:
  if (!(is.numeric(train_block_size) && (length(train_block_size) == 1))) {
    stop("train_block_size must be an integer greater than 0.")
  }
  if (!(train_block_size %% 1 == 0)) {
    stop("train_block_size must be an integer greater than 0.")
  }
  if (train_block_size <= 0) {
    stop("train_block_size must be an integer greater than 0.")
  }

  ## editing args
  args[["target"]] <- target
  args[["type"]] <- type
  args[["name"]] <- name

  return(args)
}

policy_eval_object <- function(
    coef,
    IC,
    type,
    target,
    id,
    name,
    coef_ipw = NULL,
    coef_or = NULL,
    policy_actions = NULL,
    policy_object = NULL,
    g_functions = NULL,
    g_values = NULL,
    q_functions = NULL,
    q_values = NULL,
    c_functions = NULL,
    c_values = NULL,
    m_function = NULL,
    m_values = NULL,
    Z = NULL,
    subgroup_indicator = NULL,
    min_subgroup_size = NULL,
    cross_fits = NULL,
    folds = NULL,
    cross_fit_type = NULL,
    variance_type = NULL) {
  names(coef) <- name
  out <- as.list(environment())
  out <- remove_null_elements(out)
  class(out) <- c("policy_eval")

  return(out)
}

policy_eval_type <- function(target,
                             type,
                             train_policy_data,
                             valid_policy_data,
                             policy, policy_learn,
                             g_models, g_functions,
                             g_full_history, save_g_functions,
                             q_models, q_functions,
                             q_full_history, save_q_functions,
                             c_models, c_functions,
                             c_full_history, save_c_functions,
                             m_model, m_function,
                             m_full_history, save_m_function,
                             min_subgroup_size,
                             name) {
  ##
  ## training
  ##

  ## fitting the g-functions, the q-functions, c-functions, and the policy/policies:
  fits <- fit_functions(
    policy_data = train_policy_data,
    type = type,
    policy = policy, policy_learn = policy_learn,
    g_models = g_models, g_functions = g_functions,
    g_full_history = g_full_history,
    q_models = q_models, q_functions = q_functions,
    q_full_history = q_full_history,
    c_models = c_models, c_functions = c_functions,
    c_full_history = c_full_history,
    m_model = m_model, m_function = m_function,
    m_full_history = m_full_history
  )
  rm(train_policy_data)

  ## getting the fitted g-functions, Q-functions, and c-functions:
  g_functions <- get_element(fits, "g_functions")
  q_functions <- get_element(fits, "q_functions")
  c_functions <- get_element(fits, "c_functions")
  m_function <- get_element(fits, "m_function")

  ## getting the fitted policy/policies as a list:
  if (is.null(policy)) {
    policy <- get_policy(get_element(fits, "policy_object"))
  }
  if (inherits(policy, what = "policy")) {
    policy <- list(policy)
  }
  ## appending policy names:
  pol_names <- lapply(policy, function(x) attr(x, which = "name", exact = TRUE))
  pol_names <- unlist(pol_names)
  if (!is.null(pol_names)){
    name <- lapply(pol_names,
                   function(pn){
                     paste0(name, ": d=", pn)
                   })
    name <- unlist(name)
  }

  ##
  ## validating
  ##

  ## getting the id as a data.table:
  id <- data.table(id = get_id(valid_policy_data))
  setkeyv(id, "id")

  ## getting the g-function values (for events 0):
  g_values <- NULL
  if (!is.null(g_functions)) {
    g_values <- predict(g_functions, valid_policy_data, event_set = c(0))
  }

  ## getting the Q-function values (for events 0,2):
  q_values <- NULL
  if (!is.null(q_functions)) {
    q_values <- predict(q_functions, valid_policy_data, event_set = c(0,2))
  }

  ## getting the c-function values (for events 0,2):
  c_values <- NULL
  if (!is.null(c_functions)) {
    c_values <- predict(c_functions, valid_policy_data)
  }

  ## getting the m-function values:
  m_values <- NULL
  if (!is.null(m_function)) {
    m_values <- predict(m_function, valid_policy_data)
  }

  ## setting g-functions output:
  if (save_g_functions != TRUE) {
    g_functions <- NULL
  }
  ## setting Q-functions output:
  if (save_q_functions != TRUE) {
    q_functions <- NULL
  }

  ## setting the c-functions output:
  if (save_c_functions != TRUE) {
    c_functions <- NULL
  }

   ## setting the M-functions output:
  if (save_m_function != TRUE) {
    m_function <- NULL
  }

  ## getting the number of stages:
  K <- get_K(valid_policy_data)

  ## getting the action set and stage action set:
  action_set <- get_action_set(valid_policy_data)

  ## getting the observed actions:
  actions <- get_actions(valid_policy_data)

  ## getting the events:
  events <- get_event(valid_policy_data)

  ## getting the utility:
  utility <- get_utility(valid_policy_data)

  ## calculating the target estimate for each policy:
  if (length(policy) == 1) {
    ## getting the policy actions for both right-censoring and action events:
    policy_actions <- policy[[1]](valid_policy_data)

    ## checking that the policy actions comply with the stage action sets:
    check_actions(
      actions = policy_actions,
      policy_data = valid_policy_data
    )

    estimate_objects <- estimate_target(
      target = target,
      type = type,
      K = K,
      id = id,
      action_set = action_set,
      actions = actions,
      policy_actions = policy_actions,
      events = events,
      g_values = g_values,
      q_values = q_values,
      c_values = c_values,
      m_values = m_values,
      utility = utility,
      min_subgroup_size = min_subgroup_size
    )
    estimate_objects <- list(estimate_objects)
  } else {
    policy_actions <- NULL

    estimate_objects <- lapply(
      policy,
      function(p) {
        ## getting the policy actions:## getting the policy actions for both right-censoring and action events:
        policy_actions <- p(valid_policy_data)

        ## checking that the policy actions comply with the stage action sets:
        check_actions(
          actions = policy_actions,
          policy_data = valid_policy_data
        )

        out <- estimate_target(
          target = target,
          type = type,
          K = K,
          id = id,
          action_set = action_set,
          actions = actions,
          policy_actions = policy_actions,
          events = events,
          g_values = g_values,
          q_values = q_values,
          c_values = c_values,
          m_values = m_values,
          utility = utility,
          min_subgroup_size = min_subgroup_size
        )
        return(out)
      }
    )
  }

  subgroup_indicator <- NULL
  Z <- NULL
  if (target == "subgroup") {
    ## getting the doubly robust score for each treatment:
    Z <- get_element(estimate_objects[[1]], "Z")
    ## getting the subgroup indicator (if available):
    subgroup_indicator <- lapply(
      estimate_objects,
      function(eb) get_element(eb, "subgroup_indicator")
    )
    subgroup_indicator <- do.call(what = "cbind", subgroup_indicator)
  }

  ## collecting the target estimates:
  coef <- unlist(lapply(estimate_objects, function(eb) get_element(eb, "coef")))
  coef_ipw <- unlist(lapply(
    estimate_objects,
    function(eb) get_element(eb, "coef_ipw", check_name = FALSE)
  ))
  coef_or <- unlist(lapply(
    estimate_objects,
    function(eb) get_element(eb, "coef_or", check_name = FALSE)
  ))

  ## getting the influence curves:
  IC <- lapply(estimate_objects, function(eb) get_element(eb, "IC"))
  IC <- do.call(what = "cbind", IC)

  ## if (target == "subgroup" & length(coef) > 1) {
  ##   ## Reorder coefficients so all thresholds for d=1 are presented first
  ##   ## followied by d=0
  ##   idx <- seq_len(length(coef) / 2) * 2 - 1
  ##   coef <- coef[c(idx, idx + 1)]
  ##   IC <- IC[, c(idx, idx + 1)]
  ## }

  out <- policy_eval_object(
    coef = coef,
    IC = IC,
    type = type,
    target = target,
    coef_ipw = coef_ipw,
    coef_or = coef_or,
    id = get_id(valid_policy_data),
    policy_actions = policy_actions,
    policy_object = get_element(fits, "policy_object", check_name = FALSE),
    g_functions = g_functions,
    g_values = g_values,
    q_functions = q_functions,
    q_values = q_values,
    c_functions = c_functions,
    c_values = c_values,
    m_function = m_function,
    m_values = m_values,
    Z = Z,
    subgroup_indicator = subgroup_indicator,
    min_subgroup_size = min_subgroup_size,
    name = name
  )

  return(out)
}

policy_eval_cross <- function(args,
                              policy_data,
                              M,
                              cross_fit_type,
                              variance_type,
                              future_args
                              ) {

  ## check cross_fit_type input:
  cross_fit_type <- tolower(cross_fit_type)
  if (length(cross_fit_type) != 1) {
    stop("cross_fit_type must be a character string.")
  }
  if (!(cross_fit_type %in% c("pooled", "stacked"))) {
    stop("cross_fit_type must be either 'pooled' or 'stacked'.")
  }

  ## check varince_type input:
  variance_type <- tolower(variance_type)
  if (length(variance_type) != 1) {
    stop("variance_type must be a character string.")
  }
  if (!(variance_type %in% c("pooled", "stacked", "complete"))) {
    stop("variance_type must be either 'pooled', 'stacked' or 'complete'.")
  }

  type <- get_element(args, "type")
  target <- get_element(args, "target")
  min_subgroup_size <- get_element(args, "min_subgroup_size")
  n <- get_n(policy_data)
  id <- get_id(policy_data)
  ## policy_learn <- get_element(args, "policy_learn")
  ## if (!is.null(policy_learn)) {
  ##   n_thres <- length(get_element(attr(policy_learn, "pl_args"), "threshold"))
  ## } else {
  ##   n_thres <- 1
  ## }

  ## setting up the folds:
  folds <- split(sample(1:n, n), rep(1:M, length.out = n))
  folds <- lapply(folds, sort)
  names(folds) <- paste("fold_", 1:M, sep = "")

  ## cross-fitting the policy evaluation:
  prog <- progressor(along = folds)
  cross_args <- append(
    list(
      X = folds,
      FUN = policy_eval_fold,
      policy_data = policy_data,
      args = args,
      prog = prog
    ),
    future_args
  )

  cross_fits <- do.call(what = future.apply::future_lapply, cross_args)


  ## collecting the paramenter name(s):
  name <- get_element(cross_fits[[1]], "name")
  
  ## collecting the ids from each fold (unsorted):
  id <- unlist(lapply(
    cross_fits,
    function(x) get_element(x, "id")
  ), use.names = FALSE)

  ## getting the number of observations in each (training) fold (unsorted):
  n_folds <- unlist(lapply(
    cross_fits,
    function(x) length(get_element(x, "id"))
  ))

  ## collecting the cross-fitted policy actions (sorted):
  policy_actions <- lapply(
    cross_fits, function(x) get_element(x, "policy_actions", check_name = FALSE)
  )
  policy_actions <- rbindlist(policy_actions)
  if (nrow(policy_actions)>0) {
    setkey(policy_actions, "id", "stage")
  } else {
    policy_actions <- NULL
  }


  ## collecting the subgroup indicator (sorted):
  subgroup_indicator <- NULL
  if (target == "subgroup") {
    subgroup_indicator <- lapply(
      cross_fits, function(x) get_element(x, "subgroup_indicator")
    )
    subgroup_indicator <- do.call(what = "rbind", subgroup_indicator)
    subgroup_indicator <- subgroup_indicator[order(id), , drop = FALSE]
    ## subgroup_indicator <- cbind(subgroup_indicator, !subgroup_indicator)
    subgroup_size <- apply(subgroup_indicator,
                             MARGIN = 2,
                             sum)
  }

  ##
  ## collecting the cross-fitted doubly robust scores (sorted):
  ##

  ## target parameter: policy value
  if (target == "value") {
    ## vector of length n with elements Z_1(d):
    Zd <- lapply(
      cross_fits,
      function(x) {
        IC <- get_element(x, "IC")
        coef <- get_element(x, "coef")
        sweep(IC, MARGIN = 2, coef, FUN = "+")
      }
    )
    Zd <- do.call(what = "rbind", Zd)
    Zd <- Zd[order(id), , drop = FALSE]
  }

  ## target parameter: subgroup average treatment effect
  if (target == "subgroup") {
    Z <- lapply(
      cross_fits,
      function(x) {
        get_element(x, "Z")
      }
    )
    Z <- do.call(what = "rbind", Z)
    Z <- Z[order(id), , drop = FALSE]
  }

  ##
  ## calculating the cross-fitted estimate:
  ##

  ## target parameter: policy value
  if (target == "value") {
    coef <- apply(Zd, MARGIN = 2, mean)
  }

  ## target parameter: subgroup average treatment effect
  if (target == "subgroup") {
    if (cross_fit_type == "stacked") {
      coef <- lapply(
        cross_fits, function(x) get_element(x, "coef")
      )
      coef <- do.call(what = "rbind", coef)
      coef <- apply(
        coef,
        MARGIN = 2,
        function(x) {
          sum((n_folds / sum(n_folds)) * x)
        }
      )
    }
    if (cross_fit_type == "pooled") {
      ## calculating the subgroup average treatment effect estimate:
      coef <- apply(subgroup_indicator,
                    MARGIN = 2,
                    function(x) mean((Z[, 2] - Z[, 1])[x])
                    )
      coef[!(subgroup_size >= min_subgroup_size)] <- as.numeric(NA)

    }
    n_coef <- length(coef)
  }
  ## resetting the parameter name:
  names(coef) <- name

  ##
  ## calculating the influence curve for the cross-fitted estimator (sorted):
  ##

  ## target parameter: policy value
  if (target == "value") {
    if (variance_type == "stacked") {
      IC <- lapply(
        cross_fits, function(x) IC(x)
      )
      IC <- do.call(what = "rbind", IC)
      IC <- IC[order(id), , drop = FALSE]
    }
    if (variance_type == "pooled") {
      IC <- sweep(Zd, MARGIN = 2, coef)
    }
    if (variance_type == "complete") {
      args[["train_policy_data"]] <- policy_data
      args[["valid_policy_data"]] <- policy_data
      pe <- do.call(what = policy_eval_type, args = args)
      IC <- IC(pe)
      rm(pe)
    }
  }

  ## target parameter: subgroup average treatment effect
  if (target == "subgroup") {
    if (variance_type == "stacked") {
      IC <- lapply(
        cross_fits, function(x) IC(x)
      )
      IC <- do.call(what = "rbind", IC)
      IC <- IC[order(id), , drop = FALSE]
    }
    if (variance_type == "pooled") {
      tmp <- apply(subgroup_indicator,
                   MARGIN = 2,
                   function(x) mean((Z[, 2] - Z[, 1])[x])
                   )
      IC <- matrix(nrow = n, ncol = n_coef)
      for (j in 1:n_coef) {
        ic <- 1 / mean(subgroup_indicator[, j]) *
          subgroup_indicator[, j] * ((Z[, 2] - Z[, 1]) - tmp[j])
        IC[, j] <- ic
      }
      IC[ ,!(subgroup_size >= min_subgroup_size)] <- NA
      rm(ic, tmp)
    }
    if (variance_type == "complete") {
      args[["train_policy_data"]] <- policy_data
      args[["valid_policy_data"]] <- policy_data
      pe <- do.call(what = policy_eval_type, args = args)
      IC <- IC(pe)
      rm(pe)
    }
  }

  ## calculating the cross-fitted ipw value estimate
  ## (only if target = "value" and type = "dr"):
  coef_ipw <- NULL
  if (target == "value" && type == "dr") {
    coef_ipw <- lapply(
      cross_fits, function(x) get_element(x, "coef_ipw")
    )
    coef_ipw <- do.call(what = "rbind", coef_ipw)
    coef_ipw <- apply(
      coef_ipw,
      MARGIN = 2,
      function(x) {
        sum((n_folds / sum(n_folds)) * x)
      }
    )
  }


  ## calculating the cross-fitted outcome regression value estimate
  ## (only if target = "value" and type = "dr"):
  coef_or <- NULL
  if (target == "value" && type == "dr") {
    coef_or <- lapply(
      cross_fits, function(x) get_element(x, "coef_or")
    )
    coef_or <- do.call(what = "rbind", coef_or)
    coef_or <- apply(
      coef_or,
      MARGIN = 2,
      function(x) {
        sum((n_folds / sum(n_folds)) * x)
      }
    )
  }

  ## collecting the cross-fitted g- and Q-values (sorted):
  g_values <- lapply(
    cross_fits,
    function(x) get_element(x, "g_values", check_name = FALSE)
  )
  null_g_values <- unlist(lapply(g_values, is.null))
  if (!all(null_g_values)) {
    g_values <- data.table::rbindlist(g_values)
    setkey(g_values, "id", "stage")
  } else {
    g_values <- NULL
  }
  q_values <- lapply(
    cross_fits,
    function(x) get_element(x, "q_values", check_name = FALSE)
  )
  null_q_values <- unlist(lapply(q_values, is.null))
  if (!all(null_q_values)) {
    q_values <- data.table::rbindlist(q_values)
    setkey(q_values, "id", "stage")
  } else {
    q_values <- NULL
  }
  ## collecting the cross-fitted c- and m-values (sorted):
  c_values <- lapply(
    cross_fits,
    function(x) get_element(x, "c_values", check_name = FALSE)
  )
  null_c_values <- unlist(lapply(c_values, is.null))
  if (!all(null_c_values)) {
    c_values <- data.table::rbindlist(c_values)
    setkey(c_values, "id", "stage")
  } else {
    c_values <- NULL
  }
  m_values <- lapply(
    cross_fits,
    function(x) get_element(x, "m_values", check_name = FALSE)
  )
  null_m_values <- unlist(lapply(m_values, is.null))
  if (!all(null_m_values)) {
    m_values <- data.table::rbindlist(m_values)
    setkey(m_values, "id", "stage")
  } else {
    m_values <- NULL
  }

  ## sorting id:
  id <- id[order(id)]

  out <- policy_eval_object(
    coef = coef,
    IC = IC,
    type = type,
    target = target,
    coef_ipw = coef_ipw,
    coef_or = coef_or,
    id = id,
    policy_actions = policy_actions,
    g_values = g_values,
    q_values = q_values,
    c_values = c_values,
    m_values = m_values,
    cross_fits = cross_fits,
    folds = folds,
    name = name,
    variance_type = variance_type,
    cross_fit_type = cross_fit_type,
    subgroup_indicator = subgroup_indicator
  )

  return(out)
}

policy_eval_fold <- function(fold,
                             policy_data,
                             args,
                             prog) {
  K <- get_K(policy_data)
  id <- get_id(policy_data)

  train_id <- id[-fold]
  validation_id <- id[fold]

  ## training data:
  train_policy_data <- subset_id(policy_data, train_id)
  if (get_K(train_policy_data) != K) {
    stop("The number of stages varies accross the training folds.")
  }

  ## validation data:
  valid_policy_data <- subset_id(policy_data, validation_id)
  if (get_K(valid_policy_data) != K) {
    stop("The number of stages varies accross the validation folds.")
  }

  eval_args <- append(args,
                      list(valid_policy_data = valid_policy_data,
                           train_policy_data = train_policy_data))

  out <- do.call(what = "policy_eval_type", args = eval_args)

  ## progress:
  prog()

  return(out)
}

policy_eval_rep <- function(nrep,
                            args,
                            policy_data,
                            M,
                            cross_fit_type,
                            variance_type,
                            future_args){
  onerun <- function(...){
    pe <- policy_eval_cross(
      args = args,
      policy_data = policy_data,
      M = M,
      cross_fit_type = cross_fit_type,
      variance_type = variance_type,
      future_args = future_args
    )
    name <- get_element(pe, "name")
    out <- list(coef=coef(pe), IC=IC(pe), name = name)
    return(out)
  }

  ## repeated cross-fitting of the policy evaluation:
  prog <- progressor(steps = nrep)
  rep_args <- append(
    list(
      X = 1:nrep,
      FUN = onerun,
      prog = prog
    ),
    future_args
  )
  rep_fits <- do.call(what = future.apply::future_lapply, rep_args)

  name <- get_element(rep_fits[[1]], "name")

  coef <- lapply(rep_fits, function(x) get_element(x, "coef"))
  coef <- do.call(what = "rbind", coef)
  coef <- apply(coef, MARGIN = 2, mean)

  IC <- lapply(rep_fits, function(x) get_element(x, "IC"))
  IC <- Reduce("+", IC) / length(IC)

  eval <- policy_eval_object(
    coef = coef,
    IC = IC,
    type = get_element(args, "type"),
    target = get_element(args, "target"),
    id = get_id(policy_data),
    name = name,
    variance_type = variance_type,
    cross_fit_type = cross_fit_type,
    min_subgroup_size = get_element(args, "min_subgroup_size")
  )

  return(eval)

}

fit_m_function <- function(policy_data,
                           m_model,
                           full_history = FALSE) {
  K <- get_K(policy_data)

  # input checks:
  if (!is.null(m_model)){
    mes <- "m_model must be a single q_model."
    if (!inherits(m_model, "q_model")) {
      stop(mes)
    }
  }

  ## check if missing final outcomes occur (at stage K+1)
  missing <- get_element(policy_data, "cens_indicator")
  indicator <- stage <- NULL
  missing <- missing[get("stage") == (K+1), list(get("indicator"))]
  missing <- unlist(missing)

  if (missing == FALSE){
    m_function <- NULL
  } else {
    if (is.null(m_model)) {
      stop("right-censoring/missing (final) outcome occur at stage K+1, please provide an m_model.")
    }

    ## getting the IDs:
    id <- get_id(policy_data)

    ## getting the history at stage K+1:
    his <- get_history(policy_data,
                       stage = K+1,
                       full_history = full_history,
                       event_set = c(0,1,2))
    H <- get_H(his)[get_event(his) == 1, ]
    id_not_missing <- get_id(his)[get_event(his) == 1][["id"]]

    if(length(id_not_missing) == 0) {
      stop("Unable to fit m_model: all utility outcomes are missing")
    }

    ## getting the observed (complete) utility:
    utility <- get_utility(policy_data)
    ## vector with non-missing entries U_i:
    U <- utility$U[(id %in% id_not_missing)]
    stopifnot(any(!is.na(U)))
    stopifnot(length(U) == nrow(H))

    ## getting the historic rewards/utility contributions
    U_bar <- get_element(his, "U")[get("id") %in% id_not_missing, ][["U_bar"]]
    ## calculating the residual utility:
    V_res <- U - U_bar

    ## fitting the m-model:
    tryCatch({
      m_model <- m_model(AH = H, V_res = V_res)
    }, error = function(e) {
      stop("Error fitting m_model: ", e$message)
    }, warning = function(w) {
      warning("Warning in m_model fitting: ", w$message)
      ## Continue with the model despite warning
    })

    ## setting S3 class and attributes:
    m_function <- list(
      m_model = m_model,
      H_names = colnames(H),
      stage = K+1
    )
    class(m_function) <- "m_function"
    attr(m_function, "full_history") <- full_history
  }

  return(m_function)
}

#' @export
predict.m_function <- function(object, new_policy_data, ...) {
  K <- get_K(new_policy_data)
  ## check:
  if (get_element(object, "stage") != K+1){
    stop("The number of stages in m_function and new_policy_data does not match.")
  }

  ## getting the full history attribute:
  full_history <- attr(object, "full_history")

  ## creating the event history object:
  new_history <- get_history(new_policy_data,
                             full_history = full_history,
                             stage = K+1,
                             event_set = c(0,1,2))


  m_model <- getElement(object, "m_model")
  ## H_names <- getElement(object, "H_names")

  id_stage <- get_id_stage(new_history)
  new_H <- get_H(new_history)

  ## setting up output
  q_values <- id_stage
  ## getting residual predictions
  res_pred <- predict(m_model, new_AH = new_H)
  ## getting historic rewards
  U_bar <- get_element(new_history, "U")[["U_bar"]]

  set(q_values, j = "Q", value = U_bar + res_pred)

  return(q_values)
}

test_that("fit_functions handle multiple thresholds", {
    d <- sim_single_stage(200, seed = 1)
    pd <- policy_data(d,
        action = "A",
        covariates = list("Z", "B", "L"),
        utility = "U"
    )

    pl <- policy_learn(
        type = "blip",
        threshold = c(0, 1),
        control = control_blip()
    )

    expect_no_error(
      ff <- fit_functions(
        policy_data = pd,
        type = "dr",
        policy_learn = pl,
        g_models = g_glm(),
        g_full_history = FALSE,
        q_models = q_glm(),
        q_full_history = FALSE,
        c_full_history = FALSE,
        m_full_history = FALSE
      )
    )

    expect_is(ff$policy_object, class = "policy_object")
})

test_that("crossfit_functions has the expected outcome", {

  sim_two_stage_right_cens <- function(n = 1e4,
                                       par = c(gamma = 0.5,  beta = 1, zeta = 1),
                                       seed = NULL,
                                       action_model_1 = function(C_1, beta, ...)
                                         stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_1)),
                                       action_model_2 = function(C_2, beta, ...)
                                         stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_2)),
                                       deterministic_rewards = FALSE,
                                       cens_model = function(L, zeta, ...)
                                         stats::rbinom(n = NROW(L), size = 1, prob = lava::expit(zeta * abs(L)))) {
    d <- sim_two_stage(n = n)
    d$U <- d$U_1 + d$U_2 + d$U_3

    pd <- policy_data(data = d,
                      action = c("A_1", "A_2"),
                      baseline = c("B", "BB"),
                      covariates = list(L = c("L_1", "L_2"),
                                        C = c("C_1", "C_2")),
                      utility = "U")

    ld <- pd$stage_data
    ld[is.na(L), L := d[["L_3"]]]

    ## simulating the discrete right-censoring:
    delta <- rbinom(n = nrow(ld), size = 1, prob = cens_model(ld$L, zeta = par["zeta"]))

    ## adapting the data to the right-censoring process:
    ld[ , delta := delta]
    ld[ , delta := cumprod(delta), by = id]
    ld[ , tmp := cumsum(delta == 0), by = id]
    ld <- ld[tmp %in% c(0,1)]
    ld[ , tmp := NULL]
    ld[delta == 0, event := 2]
    ld[ , delta := NULL]
    ld[event == 2 & stage == 3, U := NA]
    ld[event == 2, A := NA]

    return(ld)
  }

  set.seed(1)
  ld <- sim_two_stage_right_cens(n = 1e2)
  pd <- policy_data(data = ld, type = "long", action = "A", time = "time", time2 = "time2")
  folds <- sample_folds(2, sample_size = 1e2)

  ## fit_g_functions:
  expect_no_error(
    gf <- crossfit_function(folds = folds,
                            policy_data = pd,
                            fun = fit_g_functions,
                            models = g_glm(~C),
                            full_history = FALSE,
                            save_cross_fit_models = TRUE)
  )

  sgf <- lapply(folds, function(f) {
    spd <- subset_id(pd, id = get_id(pd)[-f])
    sgf <- fit_g_functions(spd, g_models = g_glm(~C))
    return(sgf)
  })

  for (i in seq_along(sgf)) {
    tmp <- sgf[[i]]
    tmp2 <- gf$functions[[i]]

    expect_equal(tmp, tmp2)

    pred1 <- predict.nuisance_functions(tmp, new_policy_data =  subset_id(pd, id = folds[[i]]))
    pred2 <- gf$values[id %in% folds[[i]]]

    expect_equal(pred1, pred2)
  }
  rm(sgf, tmp, tmp2, pred1, pred2)


  ## fit_c_functions
  expect_no_error(
    cf <- crossfit_function(folds = folds,
                            policy_data = pd,
                            fun = fit_c_functions,
                            models = list(g_glm(~L), g_glm(~L), g_glm(~L)),
                            full_history = FALSE,
                            save_cross_fit_models = TRUE)
  )

  scf <- lapply(folds, function(f) {
    spd <- subset_id(pd, id = get_id(pd)[-f])
    scf <- fit_c_functions(spd, c_models = list(g_glm(~L), g_glm(~L), g_glm(~L)))
    return(scf)
  })

  for (i in seq_along(scf)) {
    tmp <- scf[[i]]
    tmp2 <- cf$functions[[i]]

    expect_equal(tmp, tmp2)

    pred1 <- predict.c_functions(tmp, new_policy_data =  subset_id(pd, id = folds[[i]]))
    pred2 <- cf$values[id %in% folds[[i]]]

    expect_equal(pred1, pred2)
  }
  rm(scf, cf, tmp, tmp2, pred1, pred2)

  ## fit_m_function
  expect_no_error(
    mf <- crossfit_function(folds = folds,
                            policy_data = pd,
                            fun = fit_m_function,
                            models = q_glm(~L),
                            full_history = FALSE,
                            save_cross_fit_models = TRUE)
  )

  smf <- lapply(folds, function(f) {
    spd <- subset_id(pd, id = get_id(pd)[-f])
    smf <- fit_m_function(spd, m_model = q_glm(~L))
    return(smf)
  })

  for (i in seq_along(smf)) {
    tmp <- smf[[i]]
    tmp2 <- mf$functions[[i]]

    expect_equal(tmp, tmp2)

    pred1 <- predict.m_function(tmp, new_policy_data =  subset_id(pd, id = folds[[i]]))
    pred2 <- mf$values[id %in% folds[[i]]]

    expect_equal(pred1, pred2)
  }


})

#' @title Nuisance Functions
#' @name nuisance_functions
#' @docType class
#'
#' @description The fitted g-functions and Q-functions are stored in an object
#' of class "nuisance_functions". The object is a list with a fitted model
#' object for every stage. Information on whether the full history or the
#' state/Markov-type history is stored as an attribute ("full_history").
#' @section S3 generics:
#' The following S3 generic functions are available for an object of class
#' \code{nuisance_functions}:
#' \describe{
#'   \item{\code{predict}}{ Predict the values of the g- or Q-functions based
#'   on a [policy_data] object.}
#'  }
#'
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # evaluating the static policy a=1:
#' pe <- policy_eval(policy_data = pd,
#'                   policy = policy_def(1, reuse = TRUE),
#'                   g_models = g_glm(),
#'                   q_models = q_glm())
#'
#' # getting the fitted g-functions:
#' (g_functions <- get_g_functions(pe))
#'
#' # getting the fitted Q-functions:
#' (q_functions <- get_q_functions(pe))
#'
#' # getting the fitted values:
#' head(predict(g_functions, pd))
#' head(predict(q_functions, pd))
NULL

#' Predict g-functions and Q-functions
#'
#' \code{predict()} returns the fitted values of the g-functions and
#' Q-functions when applied to a (new) policy data object.
#' @param object Object of class "nuisance_functions". Either \code{g_functions}
#' or \code{q_functions} as returned by [policy_eval()] or [policy_learn()].
#' @param new_policy_data Policy data object created by [policy_data()]. Action events (0),
#' terminal events (1), or right-censoring events (2).
#' @param event_set Subset of the events considered. Action events (0), terminal events (1), or
#' right-censoring events (2)
#' @param ... Additional arguments.
#' @returns [data.table::data.table] with keys \code{id} and \code{stage} and variables \code{g_a} or \code{Q_a} for
#' each action a in the actions set.
#' @examples
#' library("polle")
#' ### Single stage:
#' d <- sim_single_stage(5e2, seed=1)
#' pd <- policy_data(d, action="A", covariates=list("Z", "B", "L"), utility="U")
#' pd
#' # defining a static policy (A=1):
#' pl <- policy_def(1, name = "A=1")
#'
#' # doubly robust evaluation of the policy:
#' pe <- policy_eval(policy_data = pd,
#'                   policy = pl,
#'                   g_models = g_glm(),
#'                   q_models = q_glm())
#' # summarizing the estimated value of the policy:
#' pe
#'
#' # getting the fitted g-function values:
#' head(predict(get_g_functions(pe), pd))
#'
#' # getting the fitted Q-function values:
#' head(predict(get_q_functions(pe), pd))
#' @export
predict.nuisance_functions <- function(object, new_policy_data, event_set = c(0), ...){
  K <- get_K(new_policy_data)
  action_set <- get_action_set(new_policy_data)
  full_history <- attr(object, "full_history")

  if (length(object) == K){
    history <- lapply(1:K, function(s) get_history(new_policy_data,
                                                   stage = s,
                                                   full_history = full_history,
                                                   event_set = event_set))
    values <- mapply(history,
                     object,
                     FUN = function(h, f) predict(f, h),
                     SIMPLIFY = FALSE)
    values <- rbindlist(values)
    setkeyv(values, c("id", "stage"))
  } else if (length(object) == 1){
    ## state history across all stages:
    history <- get_history(new_policy_data,
                           stage = NULL,
                           full_history = FALSE,
                           event_set = event_set)
    values <- predict(object[[1]], history)
  } else{
    stop("Provide either 1 or K nuisance functions for evaluation.")
  }

  return(values)
}

#' @export
print.nuisance_functions <- function(x, ...){
  attr(x, "class") <- NULL
  print(x)
}

#' Get the history matrix H from a history object
#'
#' @param history Object of class "history".
#' @param vars Character vector. Subset of the history matrix.
#' @return [data.table::data.table] H
#' @noRd
get_H <- function(history, vars = NULL){

  H <- getElement(history, "H")
  stopifnot(!is.null(H))

  # collecting the data:
  H_names <- names(H)[!(names(H) %in% c("id", "stage"))]
  if (!is.null(vars)){
    if (is.character(vars)){
      if (!all(vars %in% H_names)){
        stop("Invalid selection of variables.")
      }
    } else
      stop("Selection of variables must be of type character.")
  } else
    vars <- H_names

  # vars dictates the ordering of the selected columns
  H <- H[, vars, with = FALSE]

  return(H)
}

#' Get the action vector A from a history object
#'
#' @param history Object of class "history".
#' @return Character vector.
#' @noRd
get_A <- function(history){
  A <- getElement(history, "A")

  action_name <- getElement(history, "action_name")
  stopifnot(!is.null(action_name))

  A <- A[[action_name]]

  return(A)
}

#' @export
get_event.history <- function(object){
  event <- get_element(object, "event")
  stopifnot(!is.null(event))
  event <- event[["event"]]
  return(event)
}

get_time <- function(history){
  time <- get_element(history, "time")
  stopifnot(!is.null(time))
  time <- time[["time"]]
  return(time)
}

get_time2 <- function(history){
  time <- get_element(history, "time")
  stopifnot(!is.null(time))
  time2 <- time[["time2"]]
  return(time2)
}

#' @export
get_id.history <- function(object){
  H <- getElement(object, "H")
  stopifnot(!is.null(H))
  id <- H[, "id"]
  stopifnot(!is.null(id))
  return(id)
}

#' @export
get_id_stage.history <- function(object, ...){
  H <- getElement(object, "H")
  stopifnot(!is.null(H))
  id_stage <- c("id", "stage")
  id_stage <- H[, id_stage, with = FALSE]

  return(id_stage)
}

test_that("predict.q_glmnet return a vector", {
  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
      action = "A",
      covariates = list("Z", "B", "L"),
      utility = "U"
      )

  his <- get_history(pd, stage = 1)
  Q <- get_utility(pd)$U

  qfun <- fit_Q_function(history = his, Q = Q, q_model = q_glmnet())

  qm <- qfun$q_model

  AH <- cbind(A = get_A(his), get_H(his))
  pred <- predict.q_glmnet(qm, new_AH = AH)

  expect_true(
      is.vector(pred)
  )
})

test_that("q_glmnet formats data correctly via the formula", {
    d1 <- sim_single_stage(200, seed = 1)
    d1$BB <- sample(c("group 1", "group & 2", "group & 3"), size = 200, replace = TRUE)
    pd1 <- policy_data(d1,
        action = "A",
        covariates = list("Z", "B", "L", "BB"),
        utility = "U"
    )

    expect_error(
        pe <- policy_eval(
            policy_data = pd1,
            policy_learn = policy_learn(type = "ql", alpha = 0.05),
            q_models = q_glmnet(formula = ~ A * .)
        ),
        NA
    )
})


test_that("q_glmnet handles missing covariates", {
    d <- sim_two_stage(2e3, seed = 1)
    d$C_1 <- NULL
    pd <- policy_data(d,
        action = c("A_1", "A_2"),
        baseline = c("BB", "B"),
        covariates = list(
            L = c("L_1", "L_2"),
            C = c(NA, "C_2")
        ), # C_1 is missing
        utility = c("U_1", "U_2", "U_3")
    )
    p <- policy_def(1, reuse = TRUE)

    expect_error(
        suppressWarnings({
            pe <- policy_eval(
                policy_data = pd,
                policy = p,
                type = "or",
                q_models = q_glmnet()
            )
        })
    )


    expect_error(
        policy_eval(
            policy_data = pd,
            policy = p,
            type = "or",
            q_models = q_glmnet(~ L + A)
        ),
        NA
    )
    expect_error(
        policy_eval(
            policy_data = pd,
            policy = p,
            type = "or",
            q_models = list(q_glmnet(~ L + A), q_glmnet())
        ),
        NA
    )

    d <- sim_two_stage(2e3, seed = 1)
    d$C_1[1:10] <- NA
    pd <- policy_data(d,
        action = c("A_1", "A_2"),
        baseline = c("BB", "B"),
        covariates = list(
            L = c("L_1", "L_2"),
            C = c("C_1", "C_2")
        ),
        utility = c("U_1", "U_2", "U_3")
    )

    expect_error(
        expect_warning(
            policy_eval(
                policy_data = pd,
                policy = p,
                type = "or",
                q_models = q_glmnet()
            ),
            "The regression variables C have missing NA values."
        )
    )
})

test_that("testing that q_glmnet returns the prediction associated with lambda.min", {

  par0 <- list(a = 0, b = 0.0, c = 0)
  sim_d <- function(n,
                    par = par0,
                    potential_outcomes = FALSE,
                    action_function = function(H){
                      n <- nrow(H)
                      rbinom(n = n, size = 1, prob = 0.5)
                    }) {
    W <- runif(n = n, min = -1, max = 1)
    L <- runif(n = n, min = -1, max = 1)
    H <- data.table(W = W, L = L)
    A <- as.numeric(action_function(H))
    U1 <- W + L + (par$c*W + par$a*L + par$b) # U^1
    U0 <- W + L # U^0
    U <- A * U1 + (1 - A) * U0 + rnorm(n = n)
    out <- data.table(W = W, L = L, A = A, U = U)
    if (potential_outcomes == TRUE) {
      out$U0 <- U0
      out$U1 <- U1
    }
    return(out)
  }

  set.seed(1)
  d<-sim_d(400)

  pd <- policy_data(
    d,
    action = "A",
    covariates = list("W", "L"),
    utility = "U"
  )
  his <- get_history(pd)

  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_glmnet(~.)),
    threshold=0
  )

  po <- pl(
    policy_data = pd,
    g_models = g_glm(~1),
    q_models = q_glmnet(~.*A)
  )

  qf <- get_q_functions(po)

  ## fitted values of the Q-function using s = "lambda.min"
  expect_equal(
    qf$stage_1$q_model$s,
    "lambda.min"
  )
  pred_lambda.min <- predict(qf, pd)

  ## predictions from a cv.glmnet object uses s = "lambda.1se" per default
  ## see https://cran.r-project.org/web/packages/glmnet/glmnet.pdf#page=59
  m_glmnet <- qf$stage_1$q_model$model
  ## applying the design (formula) to the action and history matrix:
  pred_data <- apply_design(design = qf$stage_1$q_model$design, data = cbind(A = "1", his$H))
  pred_lambda.min_ref <- predict(qf$stage_1$q_model$model, newx = pred_data, s = "lambda.min")

  expect_equal(
    pred_lambda.min$Q_1,
    pred_lambda.min_ref[,1] |> unname()
  )
})

estimate_target <- function(target = "value",
                            type,
                            K,
                            id,
                            action_set,
                            actions,
                            policy_actions,
                            events,
                            g_values,
                            q_values,
                            c_values,
                            m_values,
                            utility,
                            min_subgroup_size) {

  args <- as.list(environment())
  args[["target"]] <- NULL
  args[["type"]] <- NULL

  if (target == "value") {
    if (type == "dr") {
      out <- do.call(what = "dr_value", args)
    } else if (type == "ipw") {
      out <- do.call(what = "ipw_value", args)
    } else if (type == "or") {
      out <- do.call(what = "or_value", args)
    } else {
      stop()
    }
  } else if (target == "subgroup") {
    if (type == "dr") {
      out <- do.call(what = "dr_subgroup", args)
    } else {
      stop("target = 'subgroup' only implemented for type = 'dr'.")
    }
  } else {
    stop("unknown target argument.")
  }

  return(out)
}

#' Create event indicator matrix
#'
#' Converts event data to a binary matrix where events in the event_set are coded as 1,
#' all others (including NA) as 0.
#'
#' @param events data.table/data.frame containing id, stage, and event columns
#' @param event_set vector of valid event codes, subset of c(0,1,2,NA)
#'
#' @return matrix with binary indicators (0/1) for each stage. Dimensions (#id X #stage).
#' @noRd
event_matrix <- function(events, event_set) {
  if (!data.table::is.data.table(events)) {
    stop("events must be a data.table")
  }
  required_cols <- c("id", "stage", "event")
  if (!all(required_cols %in% names(events))) {
    stop("events must contain columns: id, stage, and event")
  }
  valid_events <- c(0, 1, 2, NA)
  if (!all(event_set %in% c(valid_events))) {
    stop("event_set must be a subset of c(0,1,2,NA)")
  }

  M <- dcast(events, id ~ stage, value.var = "event")
  id <- NULL
  M[, id := NULL]
  M[, names(M) := lapply(.SD, function(x) as.numeric(x %in% event_set))]

  ## replace NA values with 0:
  setnafill(M, fill = 0)
  M <- as.matrix(M)
  return(M)
}


#' Calculate censoring probability matrix
#'
#' @details the probability for right-censored/missing events is set to 1
#' (in order to avoid division by 0)
#'
#' @param c_values data.table containing id, stage, surv_time, and surv_time2
#' @param M matrix of missing event indicators
#' @param n integer, number of subjects
#' @param K integer, number of stages
#'
#' @return matrix of censoring probabilities. Dimensions (n X (K+1)).
#' @noRd
cens_prob_matrix <- function(c_values, M, n, K) {
  if (!is.null(c_values)){
    if (!data.table::is.data.table(c_values)) {
      stop("c_values must be a data.table")
    }
    required_cols <- c("id", "stage", "surv_time", "surv_time2")
    if (!all(required_cols %in% names(c_values))) {
      stop("c_values must contain columns: ",
           paste(required_cols, collapse = ", "))
    }

    surv_time <- dcast(c_values, id ~ stage,
                       value.var = "surv_time")[, -c("id"), with = FALSE]
    surv_time2 <- dcast(c_values, id ~ stage,
                        value.var = "surv_time2")[, -c("id"), with = FALSE]
    C <- as.matrix(surv_time2 / surv_time)
    ## set probabilities to 1 for missing events:
    C[M == 1] <- 1

    ## check for invalid values:
    if (any(is.na(C)) || any(is.infinite(C))) {
      warning("NA or infinite values detected in probability matrix")
    }

  } else {
    C <- matrix(nrow = n, ncol = K+1, 1)
  }

  return(C)
}

#' Create policy action indicator matrix
#'
#' @details right-censored/missing and terminal events are coded as 1.
#'
#' @param actions data.table with columns: id, stage, A (observed actions)
#' @param policy_actions data.table with columns: id, stage, d (policy actions)
#' @param id data.table/data.frame with subject IDs
#' @param D matrix of non-censoring/non-missing indicators
#' @param E matrix of terminal event indicators
#'
#' @return matrix of indicators where 1 indicates matching actions. Dimensions (#id X (K+1)).
#' @noRd
policy_action_indicator_matrix <- function(actions, policy_actions, id, D, E) {
  n <- nrow(id)
  K <- ncol(D) - 1

  if (!(is.null(actions) && is.null(policy_actions))) {
    ## input validation:
    if (!data.table::is.data.table(actions) ||
        !data.table::is.data.table(policy_actions)) {
      stop("actions and policy_actions must be data.tables")
    }
    required_cols_actions <- c("id", "stage", "A")
    required_cols_policy <- c("id", "stage", "d")
    if (!all(required_cols_actions %in% names(actions))) {
      stop("actions must contain columns: ",
           paste(required_cols_actions, collapse = ", "))
    }
    if (!all(required_cols_policy %in% names(policy_actions))) {
      stop("policy_actions must contain columns: ",
           paste(required_cols_policy, collapse = ", "))
    }

    wide_actions <- dcast(
      actions,
      id ~ stage,
      value.var = "A"
    )
    wide_actions <- merge(id, wide_actions, all.x = TRUE)
    IIA <- wide_actions[, -c("id"), with = FALSE]
    IIA <- as.matrix(IIA)
    stopifnot(
      nrow(IIA) == n,
      ncol(IIA) == K
    )

    wide_policy_actions <- dcast(
      policy_actions, id ~ stage,
      value.var = "d"
    )
    wide_policy_actions <- merge(id, wide_policy_actions, all.x = TRUE)
    IId <- wide_policy_actions[, -c("id"), with = FALSE]
    IId <- as.matrix(IId)
    stopifnot(
      nrow(IId) == n,
      ncol(IId) == K
    )

    II <- (IIA == IId)
    rm(IIA, IId)
  } else {
    II <- matrix(nrow = n, ncol = K)
  }
  ## set I(A_{K+1} = d_{K+1}) = 1:
  II <- cbind(II, 1)
  ## replace value with 1 for right-censored/missing events (D == 0):
  II[D == 0] <- 1
  ## replace value with 1 for terminal events (E == 1)
  II[E == 1] <- 1

  if (!(is.null(actions) && is.null(policy_actions))) {
    if (any(is.na(II))) {
      warning("Missing values detected in actions or policy actions")
    }
  }

  colnames(II) <- 1:ncol(II)

  return(II)
}

#' Calculate action probability matrix
#'
#' @details Right-censored/mising and terminal events are coded as 1.
#'
#' @param g_values data.table containing the action probability values
#' @param id data.table with subject IDs
#' @param D matrix of non-censoring indicators
#' @param E matrix of terminal event indicators
#' @param action_set vector of possible actions
#'
#' @return matrix of action probabilities. Dimensions (n X (K+1)).
#' @noRd
action_prob_matrix <- function(g_values, actions, id, D, E, action_set) {
  if (!data.table::is.data.table(g_values) ||
      !data.table::is.data.table(actions)) {
    stop("g_values and actions must be data.tables")
  }

  values <- get_a_values(
    a = actions[["A"]],
    action_set = action_set,
    values = g_values,
    na.rm = FALSE
  )
  G <- dcast(values, id ~ stage, value.var = "P")
  G <- merge(id, G, all.x = TRUE)
  id <- NULL
  G[ , id := NULL]
  G <- as.matrix(G)
  G <- cbind(G, 1)
  ## replace value with 1 for right-censoring/missing events:
  G[D == 0] <- 1
  ## replace value with 1 for terminal events:
  G[E == 1] <- 1

  if (any(is.na(G))) {
    warning("Missing values detected in the action probability")
  }

  colnames(G) <- 1:ncol(G)

  return(G)
}

#' Calculate policy action outcome regression matrix
#'
#' Creates an (n X (K+2)) matrix of Q-values for the given policy actions, handling both
#' terminal stages and right-censoring scenarios.
#'
#' @details
#' The matrix structure is as follows:
#' * Columns 1 to K: Auxiliary Q-function values for stages 1 to K. Missing values are replaced with 0.
#' * Column K+1: Final stage outcome regression values (m_values) or the observed
#' utility outcomes if no right-censoring at stage K+1 occur. Missing values are replaced with 0.
#' * Column K+2: Observed utility outcomes. Missing values are replaced with 0.
#'
#' Right-censored outcome regression values are replaced with 0.
#' Missing values due to terminal events are replaced with the observed utility outcomes.
#'
#' @param q_values data.table containing Q-function values
#' @param m_values data.table containing terminal stage values (or NULL)
#' @param U numeric vector of utility outcomes
#' @param policy_actions data.table containing policy decisions
#' @param id data.table with subject IDs
#' @param D matrix of non-censoring indicators
#' @param action_set vector of possible actions
#'
#' @return matrix of dimension (n X (K+2)) containing:
#'   * Q-values for the provided policy actions (cols 1:K)
#'   * Final stage outcome regression values (col K+1)
#'   * Observed utility outcomes (col K+2)
#' @noRd
policy_action_outcome_matrix <- function(q_values, m_values, U, policy_actions, id, D, action_set) {
  n <- nrow(id)
  K <- ncol(D) - 1

  if (!is.null(q_values)) {
    q_d_values <- get_a_values(
      a = policy_actions[["d"]],
      action_set = action_set,
      values = q_values
    )

    if (!all(complete.cases(q_d_values))) {
      stop("The policy dictates actions with incomplete (NA) Q-function values")
    }

    Q <- dcast(q_d_values, id ~ stage, value.var = "P")
    Q <- merge(id, Q, all.x = TRUE)
    Q[, id := NULL]
    Q <- as.matrix(Q)
  } else {
    Q <- matrix(nrow = n, ncol = K)
  }
  colnames(Q) <- 1:ncol(Q)

  ## add (n X 1) vector with values Q_{K+1}:
  ## Q_{K+1} = U if no right-censoring occur at stage K+1,
  ## i.e., when m_values is NULL
  if (is.null(m_values)) {
    Q <- cbind(Q, U)
  } else {
    M <- dcast(m_values, id ~ stage, value.var = "Q")
    M <- merge(id, M, all.x = TRUE)
    M[, id := NULL]
    Q <- cbind(Q, M)
  }
  ## add (n X 1) vector with values Q_{K+2} = U:
  Q <- cbind(Q, U)

  ## replace NA values with the observed utility outcome (possibly NA):
  Q <- apply(
    Q,
    MARGIN = 2,
    function(v) {
      v[is.na(v)] <- U[is.na(v)]
      return(v)
    }
  )
  ## replace values with 0 for missing outcome regression values
  Q[, -1][D == 0] <- 0

  return(Q)
}

dr_subgroup <- function(K,
                        id,
                        action_set,
                        actions,
                        policy_actions,
                        events,
                        g_values,
                        q_values,
                        c_values,
                        m_values,
                        utility,
                        min_subgroup_size) {
  if (K != 1) {
    mes <- paste0(
      "subgroup average treatment effect evaluation",
      " is not implemeted for multiple stages."
    )
    stop(mes)
  }
  if (length(action_set) != 2) {
    mes <- paste0(
      "subgroup average treatment effect evaluation is not ",
      "implemented for more than two actions."
    )
    stop(mes)
  }

  ##
  ## calculating the doubly robust score for each treatment
  ##

  ## (n X K+1) matrix with missing event indicators
  M <- event_matrix(events = events, event_set = c(NA))

  ## (n X K+1) matrix with terminal event indicators
  E <- event_matrix(events = events, event_set = c(1))

  ## (n X K+1) matrix with entries Delta_k:
  D <- event_matrix(events = events, event_set = c(0, 1))

  ## getting the utility vector U:
  U <- utility[["U"]]

  ## (n) vector indicating right-censored outcomes:
  censored_outcomes <- is.na(U)

  ## replace NA with 0 whenever D[,2] == 0:
  U[D[,2] == 0] <- 0

  ## (n X K+1) matrix with entries C_k(H_k):
  C <- cens_prob_matrix(c_values = c_values, M = M, n = nrow(id), K = K)

  ## (n X action_set) matrix with entries  I(A = a)
  II <- matrix(nrow = nrow(id), ncol = length(action_set))
  wide_actions <- dcast(actions, id ~ stage, value.var = "A")
  wide_actions <- merge(id, wide_actions, all.x = TRUE)
  IIA <- wide_actions[, -c("id"), with = FALSE]
  IIA <- as.matrix(IIA)
  for (j in seq_along(action_set)) {
    II[, j] <- (IIA == action_set[j])
  }
  rm(IIA)
  ## replace values with 0 whenever D[,1] == 0:
  II[D[,1] == 0,] <- 0

  ## (n X action_set) matrix with entries  g(a)
  g_values <- merge(policy_actions, g_values, all.x = TRUE)
  set(g_values, j = "d", value = NULL)
  G <- as.matrix(g_values[, -c("id", "stage"), with = FALSE])
  ## replace values with 1 whenever  D[,1] == 0:
  G[ D[,1] == 0,] <- 1

  ## (n X #actions) matrix with entries  Q(a)
  Q <- as.matrix(q_values[, -c("id", "stage"), with = FALSE])

  ## (n X 1) vector with values Q_{K+1}
  ## Q_{K+1} = U if no right-censoring occur at stage K+1,
  ## i.e., m_values is NULL:
  if (is.null(m_values)) {
    M <- U
  } else {
    M <- dcast(m_values, id ~ stage, value.var = "Q")
    M <- merge(id, M, all.x = TRUE)
    M <- M[, -c("id"), with = FALSE]
    M <- as.vector(as.matrix(M))
  }
  ## replace values with 0 whenever  D[,1] == 0:
  M[D[,1] == 0] <- 0

  # calculating the doubly robust (uncentralized) score for each treatment:
  Z <- Q +
    D[,1] / C[,1] * II / G * apply(Q, 2, function(q) M - q) +
    D[,1] / C[,1] * II / G * D[,2] / C[,2] * (U - M)
  colnames(Z) <- paste("Z_", action_set, sep = "")

  ## checks:
  stopifnot(
    !all(is.na(Z))
  )

  ## calculating the doubly robust blip score:
  blip <- Z[, 2] - Z[, 1]

  ## calculating the subgroup indicator for each treatment:
  subgroup_indicator <- (policy_actions[["d"]] == action_set[2])
  subgroup_indicator <- cbind(subgroup_indicator, !subgroup_indicator)
  subgroup_indicator <- unname(subgroup_indicator)

  res <- apply(
    subgroup_indicator,
    MARGIN = 2,
    FUN = function(si){
      ## calculating the subgroup average treatment effect:
      sate <- mean(blip[si])
      ## calculating the influence curve for the subgroup average treatment effect:
      IC <- 1 / mean(si) * si *
        (blip - sate)

      ss <- sum(si)
      if (ss < min_subgroup_size) {
        sate <- as.numeric(NA)
        IC <- rep(as.numeric(NA), length(si))
      } 

      out <- list(sate = sate, IC = IC)
    },
    simplify = FALSE
  )
  sate <- lapply(res, function(x) get_element(x, "sate")) |> unlist()
  IC <- lapply(res, function(x) get_element(x, "IC"))
  IC <- do.call(what = "cbind", IC)

  out <- list(
    coef = sate,
    IC = IC,
    Z = Z,
    subgroup_indicator = subgroup_indicator
  )

  return(out)
}

## doubly robust score for the policy value.
dr_value <- function(K,
                     id,
                     action_set,
                     actions,
                     policy_actions,
                     events,
                     g_values,
                     q_values,
                     c_values,
                     m_values,
                     utility,
                     ...) {
  ## (n) observed utility vector:
  U <- utility[["U"]]

  ## (n) vector indicating right-censored outcomes:
  censored_outcomes <- is.na(U)

  ## (n X K+1) matrix with missing event indicators
  M <- event_matrix(events = events, event_set = c(NA))

  ## (n X K+1) matrix with terminal event indicators
  E <- event_matrix(events = events, event_set = c(1))

  ## (n X K+1) matrix with entries Delta_k (non-censoring/non-missing indicators):
  D <- event_matrix(events = events, event_set = c(0, 1))

  ## (n X K+1) matrix with entries C_k(H_k) (non-censoring probabilities):
  C <- cens_prob_matrix(c_values = c_values, M = M, n = nrow(id), K = K)

  ## (n X K+1) matrix with entries I(d_k(H_k) = A_k)
  ## I(d_{K+1}(H_{K+1}) = A_{K+1}) = 1:
  II <- policy_action_indicator_matrix(actions = actions,
                                       policy_actions = policy_actions,
                                       D = D,
                                       E = E,
                                       id = id)

  ## (n X K+1) matrix with entries g_k(A_k, H_k) (action probabilities):
  ## g_{K+1} = 1
  G <- action_prob_matrix(g_values = g_values,
                          actions = actions,
                          id = id,
                          D = D,
                          E = E,
                          action_set = action_set)

  ## (n X (K+2)) matrix with columns Q_k(H_{k,i}, d_k(H_{k,i}))
  ## (outcome regression values):
  Q <- policy_action_outcome_matrix(q_values = q_values,
                                    m_values = m_values,
                                    U = U,
                                    policy_actions = policy_actions,
                                    id = id,
                                    D = D,
                                    action_set)

  ## calculating the doubly robust score:
  Zd <- Q[, 1]
  for (k in 1:(K+1)) {
    Zd <- Zd + ipw_weight(D[, 1:k], C[, 1:k]) * ipw_weight(II[, 1:k], G[, 1:k]) * (Q[, k + 1] - Q[, k])
  }

  ## calculating the IPW and OR scores:
  Zd_ipw <- ipw_weight(D[, 1:k], C[, 1:k]) * ipw_weight(II, G = G) * ifelse(is.na(U), 0, U)
  Zd_or <- Q[, 1]

  ##
  ## output checks
  ##

  stopifnot(
    all(!is.na(Zd)),
    all(!is.na(Zd_ipw)),
    all(!is.na(Zd_or))
  )

  out <- list(
    coef = mean(Zd),
    IC = Zd - mean(Zd),
    coef_ipw = mean(Zd_ipw),
    coef_or = mean(Zd_or)
  )

  return(out)
}

or_value <- function(K,
                     action_set,
                     actions,
                     policy_actions,
                     q_values,
                     ...) {

  ##
  ## calculating the outcome regression score
  ##

  # (n X K) matrix with entries Q_k(d_k(H_k), H_k)
  q_d_values <- get_a_values(
    a = policy_actions$d,
    action_set = action_set,
    values = q_values
  )
  if (!all(complete.cases(q_d_values))) {
    stop("The policy dictates actions with incomplete (NA) Q-function values")
  }
  Q <- dcast(q_d_values, id ~ stage, value.var = "P")[, -c("id"), with = FALSE]
  Q <- as.matrix(Q)

  Zd_or <- Q[, 1]

  ##
  ## output checks
  ##

  stopifnot(
    all(!is.na(Zd_or))
  )

  out <- list(
    coef = mean(Zd_or),
    IC = NULL
  )
  return(out)
}

ipw_value <- function(K,
                      action_set,
                      actions,
                      policy_actions,
                      g_values,
                      utility,
                      ...) {
  ##
  ## calculating the inverse probability score
  ##

  # (n) vector with entries U_i:
  U <- utility$U

  # (n X K) matrix with entries I(d_k(H_k) = A_k):
  IIA <- as.matrix(dcast(
    actions, id ~ stage,
    value.var = "A"
  )[, -c("id"), with = FALSE])
  IId <- as.matrix(dcast(
    policy_actions, id ~ stage,
    value.var = "d"
  )[, -c("id"), with = FALSE])
  II <- (IIA == IId)
  rm(IIA, IId)

  # (n X K) matrix with entries g_k(d_k(H_k), H_k):
  g_d_values <- get_a_values(
    a = policy_actions$d,
    action_set = action_set,
    g_values
  )
  G <- as.matrix(dcast(
    g_d_values,
    id ~ stage,
    value.var = "P"
  )[, -c("id"), with = FALSE])

  Zd_ipw <- ipw_weight(II, G = G) * U

  ##
  ## output checks
  ##

  stopifnot(
    !all(is.na(Zd_ipw))
  )

  out <- list(
    coef = mean(Zd_ipw),
    IC = Zd_ipw - mean(Zd_ipw)
  )
  return(out)
}

#' Online/Sequential Policy Evaluation
#'
#' \code{policy_eval_online()} is used to estimate
#' the value of a given fixed policy
#' or a data adaptive policy (e.g. a policy
#' learned from the data). \code{policy_eval_online()}
#' is also used to estimate the subgroup average
#' treatment effect as defined by the (learned) policy.
#' The evaluation is based on a online/sequential validation
#' estimation scheme making the estimation approach valid for a
#' non-converging policy under no heterogenuous treatment effect
#' (exceptional law), see details.
#' @param policy_data Policy data object created by [policy_data()].
#' @param policy Policy object created by [policy_def()].
#' @param policy_learn Policy learner object created by [policy_learn()].
#' @param g_functions Fitted g-model objects, see [nuisance_functions].
#' Preferably, use \code{g_models}.
#' @param g_models List of action probability models/g-models for each stage
#' created by [g_empir()], [g_glm()], [g_rf()], [g_sl()] or similar functions.
#' Only used for evaluation if \code{g_functions} is \code{NULL}.
#' If a single model is provided and \code{g_full_history} is \code{FALSE},
#' a single g-model is fitted across all stages. If \code{g_full_history} is
#' \code{TRUE} the model is reused at every stage.
#' @param g_full_history If TRUE, the full history is used to fit each g-model.
#' If FALSE, the state/Markov type history is used to fit each g-model.
#' @param save_g_functions If TRUE, the fitted g-functions are saved.
#' @param q_functions Fitted Q-model objects, see [nuisance_functions].
#' Only valid if the Q-functions are fitted using the same policy.
#' Preferably, use \code{q_models}.
#' @param q_models Outcome regression models/Q-models created by
#' [q_glm()], [q_rf()], [q_sl()] or similar functions.
#' Only used for evaluation if \code{q_functions} is \code{NULL}.
#' If a single model is provided, the model is reused at every stage.
#' @param q_full_history Similar to g_full_history.
#' @param save_q_functions Similar to save_g_functions.
#' @param c_functions Fitted c-model/censoring probability model objects. Preferably, use \code{c_models}.
#' @param c_models List of right-censoring probability models, see [c_model].
#' @param c_full_history Similar to g_full_history.
#' @param save_c_functions Similar to save_g_functions.
#' @param m_function Fitted outcome model object for stage K+1. Preferably, use \code{m_model}.
#' @param m_model Outcome model for the utility at stage K+1. Only used if the final utility
#' contribution is missing/has been right-censored
#' @param m_full_history Similar to g_full_history.
#' @param save_m_function Similar to save_g_functions.
#' @param target Character string. Either "value" or "subgroup". If "value",
#' the target parameter is the policy value.
#' If "subgroup", the target parameter
#' is the subgroup average treatement effect given by the policy, see details.
#' "subgroup" is only implemented for \code{type = "dr"}
#' in the single-stage case with a dichotomous action set.
#' @param M Number of folds for online estimation/sequential validation excluding the initial training block, see details.
#' @param train_block_size Integer. Size of the initial training block only
#' used for training of the policy and nuisance models, see details.
#' @param name Character string.
#' @param min_subgroup_size Minimum number of observations in the evaluated subgroup (Only used if target = "subgroup").
#' @return \code{policy_eval_online()} returns an object of inherited class "policy_eval_online", "policy_eval".
#' The object is a list containing the following elements:
#' \item{\code{coef}}{Numeric vector. The estimated target parameters:
#' policy value or subgroup average treatment effect.}
#' \item{\code{vcov}}{Numeric vector. The estimated squared standard deviation associated with
#' \code{coef}.}
#' \item{\code{target}}{Character string. The target parameter ("value" or "subgroup")}
#' \item{\code{id}}{Character vector. The IDs of the observations.}
#' \item{\code{name}}{Character vector. Names for the each element in \code{coef}.}
#' \item{\code{train_sequential_index}}{list of indexes used for training at each step.}
#' \item{\code{valid_sequential_index}}{list of indexes used for validation at each step.}
#' @details
#' ## Setup
#' Each observation has the sequential form
#' \deqn{O= {B, U_1, X_1, A_1, ..., U_K, X_K, A_K, U_{K+1}},}
#' for a possibly stochastic number of stages K.
#' \itemize{
#'  \item \eqn{B} is a vector of baseline covariates.
#'  \item \eqn{U_k} is the reward at stage k
#' (not influenced by the action \eqn{A_k}).
#'  \item \eqn{X_k} is a vector of state
#' covariates summarizing the state at stage k.
#'  \item \eqn{A_k} is the categorical action
#' within the action set \eqn{\mathcal{A}} at stage k.
#' }
#' The utility is given by the sum of the rewards, i.e.,
#' \eqn{U = \sum_{k = 1}^{K+1} U_k}.
#'
#' A (subgroup) policy is a set of functions
#' \deqn{d = \{d_1, ..., d_K\},}
#' where \eqn{d_k} for \eqn{k\in \{1, ..., K\}}
#' maps a subset or function \eqn{V_1} of \eqn{\{B, X_1, A_1, ..., A_{k-1}, X_k\}} into the
#' action set (or set of subgroups).
#'
#' Recursively define the Q-models (\code{q_models}):
#' \deqn{Q^d_K(h_K, a_K) = \mathbb{E}[U|H_K = h_K, A_K = a_K]}
#' \deqn{Q^d_k(h_k, a_k) = \mathbb{E}[Q_{k+1}(H_{k+1},
#' d_{k+1}(V_{k+1}))|H_k = h_k, A_k = a_k].}
#' If \code{q_full_history = TRUE},
#' \eqn{H_k = \{B, X_1, A_1, ..., A_{k-1}, X_k\}}, and if
#' \code{q_full_history = FALSE}, \eqn{H_k = \{B, X_k\}}.
#'
#' The g-models (\code{g_models}) are defined as
#' \deqn{g_k(h_k, a_k) = \mathbb{P}(A_k = a_k|H_k = h_k).}
#' If \code{g_full_history = TRUE},
#' \eqn{H_k = \{B, X_1, A_1, ..., A_{k-1}, X_k\}}, and if
#' \code{g_full_history = FALSE}, \eqn{H_k = \{B, X_k\}}.
#' Furthermore, if \code{g_full_history = FALSE} and \code{g_models} is a
#' single model, it is assumed that \eqn{g_1(h_1, a_1) = ... = g_K(h_K, a_K)}.
#'
#' ## Target parameters
#' If \code{target = "value"}, \code{policy_eval_online}
#' returns the estimates of
#' the value, i.e., the expected potential utility under the policy, (\code{coef}):
#' \deqn{\mathbb{E}[U^{(d)}]}
#' If \code{target = "subgroup"}, \code{K = 1}, \eqn{\mathcal{A} = \{0,1\}},
#' and \eqn{d_1(V_1) \in \{s_1, s_2\} },  \code{policy_eval()}
#' returns the estimates of the subgroup average
#' treatment effect (\code{coef}):
#' \deqn{\mathbb{E}[U^{(1)} - U^{(0)}| d_1(\cdot) = s]\quad s\in \{s_1,s_2\},}
#'
#' ## Online estimation/sequential validation
#' Estimation of the target parameter is based online estimation/sequential
#' validation using the doubly robust score. The following figure illustrate
#' online estimation using \code{M = 5} steps and an initial training block of
#' size \code{train_block_size = }\eqn{l}.
#'
#' ![](online-validation.jpg "Online estimaiton scheme")
#'
#' Step 1:
#'
#' The \eqn{n} observations are randomly ordered. In step 1,
#' the first \eqn{\{1,...,l\}} observations, highlighted in teal/blue, are used to fit the
#' Q-models, g-models, the policy (if using the \code{policy_learn} argument), and other required models.
#' We denote the collection of these fitted models as \eqn{P}.
#' The remaining observations are split into \code{M} blocks of size \eqn{m = (n-l)/M}, which
#' we for simplicity assume to be a whole number. In step 1, the target
#' parameter is estimated using the associated doubly robust score \eqn{Z(P)}
#' evaluated on the first validation fold
#' highlighted in pink \eqn{\{l+1,...,l+m\}}:
#'
#' \deqn{
#'  \frac{\sum_{i = l+1}^{l+m}  {\widehat \sigma_{i}^{-1}}
#'  Z(\widehat P_i)(O_i)}
#'  {\sum_{i = l+1}^{l+m} \widehat \sigma_{i}^{-1}},
#' }
#' where \eqn{\widehat P_i} for \eqn{i \in \{l+1,...,l+m\}}
#' refer to the fitted models trained on \eqn{\{1,...,l\}}, and \eqn{\widehat \sigma_i}
#' is the insample estimate for the standard deviation based on the training observations \eqn{\{1,...,l\}}.
#' We will later give an exact expression for \eqn{\widehat \sigma_i} for each target parameter.
#' Note that \eqn{\widehat \sigma_i} is constant for \eqn{i \in \{l+1,...,l+m\}}, but it will be
#' convenient to keep the same index for \eqn{\widehat \sigma}.
#'
#' Step 2 to M:
#'
#' In step 2, observations with index \eqn{\{1,...,l+m\}} are used to fit the model collection \eqn{P},
#' as well as the insample estimate for the standard deviation. For \eqn{i \in \{l+m+1,...,l+2m\}} these are
#' denoted as \eqn{\widehat P_i, \widehat \sigma_i}.
#' This sequential model fitting is repeated for all \code{M}
#' steps and the updated online estimator is given by
#'
#' \deqn{
#'  \frac{\sum_{i = l+1}^{n}  {\widehat \sigma_{i}^{-1}}
#'  Z(\widehat P_i)(O_i)}
#'  {\sum_{i = l+1}^n \widehat \sigma_{i}^{-1}},
#' }
#'
#' with an associated standard error estimate given by
#'
#' \deqn{\frac{\left(\frac{1}{n-l}\sum_{i = l+1}^n \widehat \sigma_{i}^{-1}\right)^{-1}}{\sqrt{n-l}}.}
#'
#' ## Doubly robust scores
#' \code{target = "value"}:
#'
#' For a policy value target the doubly robust score is given by
#' \deqn{
#' Z(d, g, Q^d)(O) = Q^d_1(H_1 , d_1(V_1)) +
#' \sum_{r = 1}^K \prod_{j = 1}^{r}
#' \frac{I\{A_j = d_j(\cdot)\}}{g_{j}(H_j, A_j)}
#' \{Q_{r+1}^d(H_{r+1} , d_{r+1}(V_1)) - Q_{r}^d(H_r , d_r(V_1))\}.
#' }
#' The influence function(/curve) for the associated onestep etimator is
#' \deqn{Z(d, g, Q^d)(O) - \mathbb{E}[Z(d,g, Q^d)(O)],}
#' which is used to estimate the insample stadard deviation. For example,
#' in step 2, i.e., for \eqn{i \in \{l+m+1,...,l+2m\}}
#' \deqn{
#'  \widehat \sigma_i^2 = \frac{1}{l+m}\sum_{j=1}^{l+m} \left(Z(\widehat d_i,\widehat Q_i,\widehat{g}_i)(O_j) - \frac{1}{l+m}\sum_{r=1}^{l+m} Z(\widehat d_i,\widehat Q_i,\widehat{g}_i)(O_r) \right)^2
#' }
#'
#' \code{target = "subgroup"}:
#'
#' For a subgroup average treatment effect target,
#' where \code{K = 1} (single-stage),
#' \eqn{\mathcal{A} = \{0,1\}} (binary treatment), and
#' \eqn{d_1(V_1) \in \{s_1, s_2\}} (dichotomous subgroup policy) the
#' doubly robust score is given by
#'
#' \deqn{
#' Z(d,g,Q, D) = \frac{I\{d_1(\cdot) = s\}}{D}
#' \Big\{Z_1(1,g,Q)(O) - Z_1(0,g,Q)(O) \Big\}.}
#' \deqn{
#' Z_1(a, g, Q)(O) = Q_1(H_1 , a) +
#' \frac{I\{A = a\}}{g_1(H_1, a)}
#' \{U - Q_{1}(H_1 , a)\},
#' }
#' where \eqn{D} is \eqn{\mathbb{P}(d_1(V_1) = s)}.
#'
#' The associated onestep/estimating equation estimator has influence function
#' \deqn{\frac{ I\{d_1(\cdot) = s\}}{D}
#' \Big\{Z_1(1,g,Q)(O) - Z_1(0,g,Q)(O) - E[Z_1(1,g,Q)(O)
#' - Z_1(0,g,Q)(O) | d_1(\cdot) = s]\Big\},}
#' which is used to estimate the standard deviation \eqn{\widehat \sigma}.
#'
#' @references
#' Luedtke, Alexander R, and Mark J van der Laan.
#' “STATISTICAL INFERENCE FOR THE MEAN OUTCOME
#' UNDER A POSSIBLY NON-UNIQUE OPTIMAL TREATMENT STRATEGY.”
#' Annals of statistics vol. 44,2 (2016): 713-742.
#' \doi{10.1214/15-AOS1384}
#' @export
policy_eval_online <- function(policy_data,
                               policy = NULL, policy_learn = NULL,
                               g_functions = NULL, g_models = g_glm(),
                               g_full_history = FALSE, save_g_functions = TRUE,
                               q_functions = NULL, q_models = q_glm(),
                               q_full_history = FALSE, save_q_functions = TRUE,
                               c_functions = NULL, c_models = NULL,
                               c_full_history = FALSE, save_c_functions = TRUE,
                               m_function = NULL, m_model = NULL,
                               m_full_history = FALSE, save_m_function = TRUE,
                               target = "value",
                               M = 4,
                               train_block_size = get_n(policy_data) / 5,
                               name = NULL,
                               min_subgroup_size = 1) {
  ## setting type to "dr":
  type <- "dr"

   ## argument input checks:
  args <- as.list(environment())
  args <- do.call(what = "policy_eval_input_checks", args)

  ## collecting the arguments to be passed on:
  args[["policy_data"]] <- NULL
  args[["M"]] <- NULL
  args[["train_block_size"]] <- NULL
  args[["future_args"]] <- NULL
  args[["variance_type"]] <- NULL
  args[["cross_fit_type"]] <- NULL
  args[["nrep"]] <- NULL

  eval <- policy_eval_on(args = args,
                         policy_data = policy_data,
                         train_block_size = train_block_size,
                         M = M)

  return(eval)
}


policy_eval_on <- function(args,
                           policy_data,
                           train_block_size,
                           M) {
  n <- get_n(policy_data)
  id <- get_id(policy_data)
  K <- get_K(policy_data)
  target <- get_element(args, "target")

  ## random index of the data (by id):
  random_index <- sample(1:n, n)

  ## setting up the sequential training and validation folds:
  sequential_index <- random_index[(train_block_size +1):n]
  sequential_index <- split(sequential_index, rep(1:M, length.out = length(sequential_index)))

  train_sequential_index <- list(random_index[1:train_block_size])
  train_sequential_index <- append(train_sequential_index, sequential_index[-M])
  if (M > 1){
    tmp <- train_sequential_index[[1]]
    for (m in 2:length(train_sequential_index)) {
      tmp <- c(tmp, train_sequential_index[[m]])
      train_sequential_index[[m]] <- tmp
    }
    rm(tmp)
  }
  train_sequential_index <- lapply(train_sequential_index, sort)
  valid_sequential_index <- sequential_index
  valid_sequential_index <- lapply(valid_sequential_index, sort)

  sequential_fits <- lapply(seq_along(train_sequential_index), function(m){
    ## training and validation ids:
    train_id <- id[train_sequential_index[[m]]]
    validation_id <- id[valid_sequential_index[[m]]]

    ## training data:
    train_policy_data <- subset_id(policy_data, train_id)
    if (get_K(train_policy_data) != K) {
      stop("The number of stages varies accross the training folds.")
    }

    ## validation data:
    valid_policy_data <- subset_id(policy_data, validation_id)
    if (get_K(valid_policy_data) != K) {
      stop("The number of stages varies accross the validation folds.")
    }


    eval_args <- append(args,
                        list(valid_policy_data = valid_policy_data,
                             train_policy_data = train_policy_data))

    valid_pe <- do.call(what = "policy_eval_type", args = eval_args)

    ## getting the in-sample variance estimate
    args[["g_models"]] <- NULL
    args[["q_models"]] <- NULL

    args[["g_functions"]] <- get_g_functions(valid_pe)
    args[["q_functions"]] <- get_q_functions(valid_pe)

    policy <- get_policy(valid_pe)
    if (is.null(policy)){
      policy <- get_element(args, "policy")
    }
    if (inherits(policy, what = "policy")) {
      policy <- list(policy)
    }

    insample_parameters <- lapply(policy, function(p){
      args[["policy"]] <- p

      eval_args <- append(args,
                          list(valid_policy_data = train_policy_data,
                               train_policy_data = train_policy_data))

      pe <- do.call(what = "policy_eval_type", args = eval_args)

      ic <- IC(pe)
      sigma2 <- var(ic)*(nrow(ic)-1)/nrow(ic)
      sigma2 <- diag(sigma2)
      names(sigma2) <- get_element(pe, "name")

      out <- list(sigma2 = sigma2)

      if (target == "subgroup"){
        subgroup_indicator <- get_element(pe, "subgroup_indicator")
        subgroup_prob <- apply(
          subgroup_indicator,
          MARGIN = 2,
          FUN = mean,
          simplify = FALSE
        )
        subgroup_prob <- unlist(subgroup_prob)
        names(subgroup_prob) <- get_element(pe, "name")
        out[["subgroup_prob"]] <- subgroup_prob
      }

      return(out)
    })

    insample_sigma2 <- lapply(insample_parameters, function(x) get_element(x, "sigma2"))
    insample_sigma2 <- unlist(insample_sigma2)

    insample_subgroup_prob <- lapply(insample_parameters, function(x) get_element(x, "subgroup_prob", check_name = FALSE))
    insample_subgroup_prob <- unlist(insample_subgroup_prob)

    out <- valid_pe
    out[["insample_sigma2"]] <- insample_sigma2
    out[["insample_subgroup_prob"]] <- insample_subgroup_prob

    return(out)
  })

  target <- get_element(args, "target")
  if (target == "value"){
    online_onestep_terms <- lapply(sequential_fits, function(x){
      IC <- get_element(x, "IC")
      coef <- get_element(x, "coef")
      Z <- apply(IC, MARGIN = 1, function(x) x + coef, simplify = FALSE)
      Z <- do.call(what = "rbind", Z)

      sigma <- unname(sqrt(get_element(x, "insample_sigma2")))

      scaled_Z <- apply(Z, MARGIN = 1, function(x) x / sigma, simplify = FALSE)
      scaled_Z <- do.call(what = "rbind", scaled_Z)

      Gamma <- Z * 0 + 1
      Gamma <- apply(Gamma, MARGIN = 1, function(x) x / sigma, simplify = FALSE)
      Gamma <- do.call(what = "rbind", Gamma)

      out <- list(scaled_Z = scaled_Z, Gamma = Gamma)
      return(out)
    })

    scaled_Z <- lapply(online_onestep_terms, function(x) get_element(x, "scaled_Z"))
    scaled_Z <- do.call(what = "rbind", scaled_Z)

    Gamma <- lapply(online_onestep_terms, function(x) get_element(x, "Gamma"))
    Gamma <- do.call(what = "rbind", Gamma)

    coef <- colSums(scaled_Z) / colSums(Gamma)
    vcov <- colMeans(Gamma)^(-2)/nrow(Gamma)
  } else if (target == "subgroup") {
    online_onestep_terms <- lapply(sequential_fits, function(x){
      subgroup_indicator <- get_element(x, "subgroup_indicator")
      Z <- get_element(x, "Z")
      subgroup_prob <- get_element(x, "insample_subgroup_prob")
      sigma <- unname(sqrt(get_element(x, "insample_sigma2")))

      blip <- Z[, 2] - Z[, 1]
      D <- apply(
        subgroup_indicator,
        MARGIN = 2,
        FUN = function(si){
          blip * si
        },
        simplify = FALSE
      )
      D <- do.call(what = "cbind", D)
      D <- apply(D, MARGIN = 1, function(x) x / subgroup_prob, simplify = FALSE)
      D <- do.call(what = "rbind", D)

      scaled_D <- apply(D, MARGIN = 1, function(x) x / sigma, simplify = FALSE)
      scaled_D <- do.call(what = "rbind", scaled_D)

      Gamma <- D * 0 + 1
      Gamma <- apply(Gamma, MARGIN = 1, function(x) x / sigma, simplify = FALSE)
      Gamma <- do.call(what = "rbind", Gamma)

      out <- list(scaled_D = scaled_D, Gamma = Gamma)
      return(out)
    })

    scaled_D <- lapply(online_onestep_terms, function(x) get_element(x, "scaled_D"))
    scaled_D <- do.call(what = "rbind", scaled_D)

    Gamma <- lapply(online_onestep_terms, function(x) get_element(x, "Gamma"))
    Gamma <- do.call(what = "rbind", Gamma)

    coef <- colSums(scaled_D) / colSums(Gamma)
    vcov <- colMeans(Gamma)^(-2)/nrow(Gamma)

  } else {
    mes <- "policy_eval_online only implemented for target = 'value' or 'subgroup'."
    stop(mes)
  }

  out <- policy_eval_online_object(
    coef = coef,
    vcov = vcov,
    type = get_element(args, "type"),
    target = target,
    id = id,
    train_sequential_index = unname(train_sequential_index),
    valid_sequential_index = unname(valid_sequential_index),
    name = get_element(sequential_fits[[1]], "name")
  )

  return(out)
}

policy_eval_online_object <- function(coef,
                                      vcov,
                                      type,
                                      target,
                                      id,
                                      name,
                                      train_sequential_index,
                                      valid_sequential_index) {
  names(coef) <- name
  out <- as.list(environment())
  out <- remove_null_elements(out)
  class(out) <- c("policy_eval_online", "policy_eval")

  return(out)
}

#' @rdname policy_eval
#' @export
vcov.policy_eval_online <- function(object, ...) {
  vcov <- get_element(object, "vcov")
  tmp <- matrix(NA, ncol = length(vcov), nrow = length(vcov))
  diag(tmp) <- vcov
  vcov <- tmp
  return(vcov)
}

new_c_model <- function(c_model){
  class(c_model) <- c("c_model", "function")
  return(c_model)
}

update_c_formula <- function(formula, time, time2, event, H) {
  tt <- terms(formula, data = H)
  if (length(attr(tt, "term.labels")) == 0){
    formula <- Surv(time = time, time2 = time2, event == 2) ~ 1
  } else {
    response <- "Surv(time = time, time2 = time2, event == 2)"
    formula <- reformulate(attr(tt, "term.labels"), response = response)
  }
  environment(formula)$time <- time
  environment(formula)$time2 <- time2
  environment(formula)$event <- event
  return(formula)
}

#' @title c_model class object
#'
#' @description
#' Provides constructors for right-censoring models.
#' Main constructors include:
#' * `c_cox()`: Cox proportional hazards model
#' * `c_no_censoring()`: Model for scenarios without censoring
#' The constructors are used as input for [policy_eval()] and [policy_learn()].
#'
#' @param formula An object of class [formula] specifying the design matrix for
#' the right-censoring model. Use [get_history_names()] to view the available
#' variable names.
#' @param offset offsets for Cox model, see [mets::phreg()]
#' @param weights weights for Cox score equations, see [mets::phreg()]
#' @param ... Additional arguments passed to the model.
#' @details
#' \code{c_cox()} is a wrapper of [mets::phreg()] (Cox proportional hazard model).\cr
#' @returns
#' A c-model object (function) with arguments:
#' * event: censoring events
#' * time: start time
#' * time2: end time
#' * H: history matrix
#' @seealso [get_history_names()].
#' @docType class
#' @name c_model
NULL

## Cox model interface

#' @rdname c_model
#' @export
c_cox <- function(formula = ~.,
                  offset = NULL,
                  weights = NULL,
                  ...) {
  formula <- as.formula(formula)
  dotdotdot <- list(...)

  if (!requireNamespace("mets")) stop("Package 'mets' required")

  c_cox <- function(event, time, time2, H) {
    ## setting the reponse of the formula to Surv(time = time, time2 = time2, event == 2):
    ## note that time, time2 and event are added to the formula environment
    formula <- update_c_formula(
        formula = formula,
        time = time,
        time2 = time2,
        event = event,
        H = H
    )
    args_cox <- append(
        list(
            formula = formula,
            data = cbind(H, time, time2, event),
            offset = offset,
            weights = weights
        ),
        dotdotdot
    )
    model <- tryCatch(do.call(mets::phreg, args = args_cox),
                      error = function(e) e
                      )
    if (inherits(model, "error")) {
      model$message <-
        paste0(model$message, " when calling 'c_cox' with formula:\n",
               format(formula))
      stop(model)
    }

    model$call <- NULL

    m <- list(model = model)
    class(m) <- c("c_cox")
    return(m)
  }

  ## setting class:
  c_cox <- new_c_model(c_cox)

  return(c_cox)
}

#' @export
predict.c_cox <- function(object, new_H, new_time, ...){
  model <- get_element(object, "model")

  ch <- targeted::cumhaz(object = model, newdata = new_H, times = new_time, individual.time = TRUE, ...)
  surv <- as.vector(get_element(ch, "surv"))
  return(surv)
}

#' @rdname c_model
#' @export
c_no_censoring <- function(){

  c_no_censoring <- function(event, time, time2, H){
    model <- list()
    class(model) <- c("c_no_censoring")
    return(model)
  }

  ## setting class:
  c_no_censoring <- new_c_model(c_no_censoring)

  return(c_no_censoring)
}

#' @export
predict.c_no_censoring <- function(object, new_H, new_time, ...){
  n <- length(new_time)
  surv <- rep(1, times = n)
  return(surv)
}

#' Fit Single g-function
#'
#' \code{fit_g_function} is used to fit a single g-model.
#' @param history History object created by [get_history()]
#' @param g_model Propensity model/g-model created by [g_glm()], [g_rf()], [g_sl()] or similar functions.
#' @returns Object of class "g_function".
#' @examples
#' library("polle")
#' ### Simulating two-stage policy data
#' d <- sim_two_stage(2e3, seed=1, par=par0)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' history <- get_history(pd, stage = 1)
#' head(history$H)
#'
#' g_function <- fit_g_function(
#'   history = get_history(pd, stage = 1),
#'   g_model = g_glm()
#' )
#' g_function
#' @noRd
fit_g_function <- function(history, g_model){
  action_set <- getElement(history, "action_set")
  stage_action_set <- getElement(history, "stage_action_set")
  if(is.null(stage_action_set)){
    stage_action_set <- action_set
  }
  stage <- getElement(history, "stage")

  # getting the action (A) and the model matrix (H):
  A <- get_A(history)
  H <- get_H(history)

  # checking that all actions in the stage action set occur:
  if (!all(stage_action_set == sort(unique(A)))){
    if (!is.null(stage)){
      mes <- "Not all actions in the stage action set occur at stage"
      mes <- paste(mes, paste(stage, collapse = ", "))
      mes <- paste(mes, ". Unable to fit g-function.", sep = "")
    } else{
      ## mes <- "Not all actions in the action set occur. Unable to fit g-function."
    }
    stop(mes)
  }

  # fitting the model:
  g_model <- g_model(A = A, H = H, action_set = stage_action_set)

  g_function <- list(
    g_model = g_model,
    H_names = colnames(H),
    action_set = action_set,
    stage_action_set = stage_action_set,
    stage = stage
  )
  class(g_function) <- "g_function"

  return(g_function)
}

#' @export
print.g_function <- function(x, ...){
  y <- x$g_model
  y$action_set <- NULL
  attr(y,"class") <- NULL

  print(y)
}

#' @export
predict.g_function <- function(object, new_history, ...){
  g_model <- getElement(object, "g_model")
  H_names <- getElement(object, "H_names")
  # action set of the new history object:
  action_set <- getElement(new_history, "action_set")
  # stage action set of the fitted g-function:
  stage_action_set <- getElement(object, "stage_action_set")
  stage <- getElement(object, "stage")

  id_stage <- get_id_stage(new_history)
  new_H <- get_H(new_history)

  if (ncol(new_H) == 0) {
    ## preventing empty data.table
    new_H <- data.table(dummy = rep(as.numeric(NA),nrow(id_stage)))
  }

  # checks
  if(!all(stage_action_set %in% action_set))
    stop("The fitted stage action set is not a subset of the new action set.")
  if (!all(H_names %in% names(new_H)))
    stop("new_history does not contain the same column names as the original history.")

  g_values <- predict(g_model, new_H = new_H)
  if (!all(g_values >= 0) |
      !all(g_values <= 1)){
    stop("Not all g-model values are within [0,1].")
  }

  tmp <- matrix(data = 0, nrow = nrow(id_stage), ncol = length(action_set))
  colnames(tmp) <- action_set
  tmp[, stage_action_set] <- g_values
  g_values <- tmp; rm(tmp)
  colnames(g_values) <- paste("g", action_set, sep = "_")

  if (!all(complete.cases(g_values))){
    if(!is.null(stage)){
      mes <- paste("The g-function predictions at stage ",
                   stage,
                   " have missing values.",
                   sep = "")
    } else {
      mes <- "The g-function predictions have missing values."
    }
    stop(mes)
  }

  if (!all(abs(apply(g_values, sum, MARGIN = 1) - 1) < 1e-12)){
    if(!is.null(stage)){
      mes <- paste("The g-function predictions at stage ",
                   stage,
                   " do not sum to 1.",
                   sep = "")
    } else {
      mes <- "The g-function predictions do not sum to 1."
    }
    stop(mes)
  }


  # including the id's and stage number(s)
  g_values <- data.table(id_stage, g_values)
  setkeyv(g_values, c("id", "stage"))

  return(g_values)
}

#' Fit g-functions
#'
#' \code{fit_g_functions} is used to fit a list of g-models.
#' @param policy_data Policy data object created by [policy_data()].
#' @param g_models List of action probability models/g-models for each stage
#' created by [g_empir()], [g_glm()], [g_rf()], [g_sl()] or similar functions.
#' @param full_history If TRUE, the full history is used to fit each g-model.
#' If FALSE, the single stage/"Markov type" history is used to fit each g-model.
#' @examples
#' library("polle")
#' ### Simulating two-stage policy data
#' d <- sim_two_stage(2e3, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # fitting a single g-model across all stages:
#' g_functions <- fit_g_functions(policy_data = pd,
#'                                g_models = g_glm(),
#'                                full_history = FALSE)
#' g_functions
#'
#' # fitting a g-model for each stage:
#' g_functions <- fit_g_functions(policy_data = pd,
#'                                g_models = list(g_glm(), g_glm()),
#'                                full_history = TRUE)
#' g_functions
#' @export
fit_g_functions <- function(policy_data, g_models, full_history = FALSE){
  K <- get_K(policy_data)

  # input checks:
  if (!(is.logical(full_history) & (length(full_history) == 1)))
    stop("full_history must be TRUE or FALSE")
  if (is.null(g_models))
    stop("Please provide g_models.")
  mes <- "g_models must be a single g_model or a list of K g_models's."
  if (is.list(g_models)){
    tmp <- all(unlist(lapply(g_models, function(gm) inherits(gm, "g_model"))))
    if (!tmp)
      stop(mes)
    rm(tmp)
    if (length(g_models) != K)
      stop(mes)
  } else{
    if (!inherits(g_models, "g_model"))
      stop(mes)
    if (full_history == TRUE)
      stop("full_history must be FALSE when a single g-model is provided.")
  }

  # fitting the g-models:
  if (is.list(g_models)){
    history <- lapply(1:K,
                      function(s) get_history(policy_data,
                                              stage = s,
                                              full_history = full_history))
    g_functions <- mapply(history,
                          g_models,
                          FUN = function(h, gm) fit_g_function(history = h,
                                                               g_model = gm),
                          SIMPLIFY = FALSE)
    names(g_functions) <- paste("stage_", 1:K, sep = "")
  } else{
    history <- get_history(policy_data, stage = NULL, full_history = FALSE)
    g_functions <- list(all_stages = fit_g_function(history, g_models))
  }

  class(g_functions) <- c("g_functions", "nuisance_functions")
  attr(g_functions, "full_history") <- full_history

  return(g_functions)
}

#' @title Get g-functions
#'
#' @description \code{get_g_functions()} returns a list of (fitted) g-functions
#' associated with each stage.
#' @param object Object of class [policy_eval] or [policy_object].
#' @returns List of class [nuisance_functions].
#' @seealso [predict.nuisance_functions]
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # evaluating the static policy a=1 using inverse propensity weighting
#' # based on a GLM model at each stage
#' pe <- policy_eval(type = "ipw",
#'                   policy_data = pd,
#'                   policy = policy_def(1, reuse = TRUE, name = "A=1"),
#'                   g_models = list(g_glm(), g_glm()))
#' pe
#'
#' # getting the g-functions
#' g_functions <- get_g_functions(pe)
#' g_functions
#'
#' # getting the fitted g-function values
#' head(predict(g_functions, pd))
#' @export
get_g_functions <- function(object)
  UseMethod("get_g_functions")

#' Create Policy Learner
#'
#' \code{policy_learn()} is used to specify a policy
#' learning method (Q-learning,
#' doubly robust Q-learning, policy tree
#' learning and outcome weighted learning).
#' Evaluating the policy learner returns a policy object.
#' @param type Type of policy learner method:
#' \itemize{
#'   \item \code{"ql"}: Quality/Q-learning.
#'   \item \code{"drql"}: Doubly Robust Q-learning.
#'   \item \code{"blip"}:
#' Doubly Robust blip-learning (only for dichotomous actions).
#'   \item \code{"ptl"}: Policy Tree Learning.
#'   \item \code{"owl"}: Outcome Weighted Learning.
#'   \item \code{"earl"}:
#' Efficient Augmentation and Relaxation Learning (only single stage).
#'   \item \code{"rwl"}: Residual Weighted Learning (only single stage).
#' }
#' @param control List of control arguments.
#' Values (and default values) are set using
#' \code{control_{type}()}. Key arguments include:\cr
#' [control_drql()]:\cr
#' \itemize{
#'   \item \code{qv_models}:
#' Single element or list of V-restricted Q-models created
#' by [q_glm()], [q_rf()], [q_sl()] or similar functions.
#' }
#' [control_blip()]:\cr
#' \itemize{
#'   \item \code{blip_models}:
#' Single element or list of V-restricted blip-models created
#' by [q_glm()], [q_rf()], [q_sl()] or similar functions.
#' }
#' [control_ptl()]: \cr
#' \itemize{
#'   \item \code{policy_vars}: Character vector/string or list of character
#' vectors/strings. Variable names used to construct
#' the V-restricted policy tree.
#' The names must be a subset of the history names, see get_history_names().
#'   \item \code{hybrid}: If \code{TRUE},
#' [policytree::hybrid_policy_tree()] is used to
#' fit a policy tree.
#'   \item \code{depth}:
#' Integer or integer vector. The depth of the fitted policy
#' tree for each stage.
#' }
#' [control_owl()]: \cr
#' \itemize{
#'   \item \code{policy_vars}: As in \code{control_ptl()}.
#'   \item \code{loss}: Loss function.
#' The options are \code{"hinge"}, \code{"ramp"},
#' \code{"logit"}, \code{"logit.lasso"}, \code{"l2"}, \code{"l2.lasso"}.
#'   \item \code{kernel}: Type of kernel
#' used by the support vector machine. The
#' options are \code{"linear"}, \code{"rbf"}.
#'    \item \code{augment}:  If \code{TRUE} the outcomes are augmented.
#' }
#' [control_earl()]/[control_rwl()]: \cr
#' \itemize{
#'   \item \code{moPropen}:
#' Propensity model of class "ModelObj", see [modelObj::modelObj].
#'   \item \code{moMain}: Main effects outcome model of class "ModelObj".
#'   \item \code{moCont} Contrast outcome model of class "ModelObj".
#'   \item \code{regime}:
#' An object of class [formula] specifying the design of the policy.
#'   \item \code{surrogate}:
#' The surrogate 0-1 loss function. The options are
#' \code{"logit"}, \code{"exp"},
#' \code{"hinge"}, \code{"sqhinge"}, \code{"huber"}.
#'   \item \code{kernel}: The options are
#' \code{"linear"}, \code{"poly"}, \code{"radial"}.
#' }
#' @param alpha Probability threshold for determining realistic actions.
#' @param threshold Numeric vector, thresholds for not
#' choosing the reference action at stage 1.
#' @param full_history If \code{TRUE}, the full
#' history is used to fit each policy function (e.g. QV-model,
#' policy tree). If FALSE, the single stage/
#' "Markov type" history is used to fit each policy function.
#' @param L Number of folds for cross-fitting nuisance models.
#' @param cross_fit_g_models If \code{TRUE}, the g-models will not be
#' cross-fitted even if L > 1.
#' @param cross_fit_c_models If \code{TRUE}, the c-models will not be
#' cross-fitted even if L > 1.
#' @param save_cross_fit_models If \code{TRUE}, the cross-fitted
#' models will be saved.
#' @param future_args Arguments passed to [future.apply::future_apply()].
#' @param name Character string.
#' @param x Object of class "policy_object" or "policy_learn".
#' @param ... Additional arguments passed to print.
#' @returns Function of inherited class \code{"policy_learn"}.
#' Evaluating the function on a [policy_data] object returns an object of
#' class [policy_object]. A policy object is a list containing all or
#' some of the following elements:
#' \describe{
#' \item{\code{q_functions}}{Fitted Q-functions. Object
#'                           of class "nuisance_functions".}
#' \item{\code{g_functions}}{Fitted g-functions. Object
#'                           of class "nuisance_functions".}
#' \item{\code{action_set}}{Sorted character vector describing
#'                          the action set, i.e.,
#'                          the possible actions at each stage.}
#' \item{\code{alpha}}{Numeric. Probability threshold to determine
#'                     realistic actions.}
#' \item{\code{K}}{Integer. Maximal number of stages.}
#' \item{\code{qv_functions}}{(only if \code{type = "drql"}) Fitted
#' V-restricted Q-functions. Contains a fitted model for each stage and action.}
#' \item{\code{ptl_objects}}{(only if \code{type = "ptl"}) Fitted V-restricted
#' policy trees. Contains a [policytree::policy_tree] for each stage.}
#' \item{\code{ptl_designs}}{(only if \code{type = "ptl"}) Specification of the
#' V-restricted design matrix for each stage}
#' }
#' @section S3 generics:
#' The following S3 generic functions are available for an object of
#' class "policy_object":
#' \describe{
#' \item{[get_g_functions()]}{Extract the fitted g-functions.}
#' \item{[get_q_functions()]}{Extract the fitted Q-functions.}
#' \item{[get_policy()]}{Extract the fitted policy object.}
#' \item{[get_policy_functions()]}{Extract the fitted policy function for
#'                                 a given stage.}
#' \item{[get_policy_actions()]}{Extract the (fitted) policy actions.}
#' }
#' @references
#' Doubly Robust Q-learning (\code{type = "drql"}): Luedtke, Alexander R., and
#' Mark J. van der Laan. "Super-learning of an optimal dynamic treatment rule."
#' The international journal of biostatistics 12.1 (2016): 305-332.
#' \doi{10.1515/ijb-2015-0052}.\cr
#' \cr
#' Policy Tree Learning (\code{type = "ptl"}): Zhou, Zhengyuan, Susan Athey,
#' and Stefan Wager. "Offline multi-action policy learning: Generalization and
#' optimization." Operations Research (2022). \doi{10.1287/opre.2022.2271}.\cr
#' \cr
#' (Augmented) Outcome Weighted Learning: Liu, Ying, et al. "Augmented
#' outcome‐weighted learning for estimating optimal dynamic treatment regimens."
#' Statistics in medicine 37.26 (2018): 3776-3788. \doi{10.1002/sim.7844}.
#' @seealso [policy_eval()]
#' @examples
#' library("polle")
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("BB"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' ### V-restricted (Doubly Robust) Q-learning
#'
#' # specifying the learner:
#' pl <- policy_learn(
#'   type = "drql",
#'   control = control_drql(qv_models = list(q_glm(formula = ~ C_1 + BB),
#'                                           q_glm(formula = ~ L_1 + BB))),
#'   full_history = TRUE
#' )
#'
#' # evaluating the learned policy
#' pe <- policy_eval(policy_data = pd,
#'                   policy_learn = pl,
#'                   q_models = q_glm(),
#'                   g_models = g_glm())
#' pe
#' # getting the policy object:
#' po <- get_policy_object(pe)
#' # inspecting the fitted QV-model for each action strata at stage 1:
#' po$qv_functions$stage_1
#' head(get_policy(pe)(pd))
#' @export
policy_learn <- function(type = "blip",
                         control = control_blip(),
                         alpha = 0,
                         threshold = NULL,
                         full_history = FALSE,
                         L = 1,
                         cross_fit_g_models = TRUE,
                         cross_fit_c_models = TRUE,
                         save_cross_fit_models = FALSE,
                         future_args = list(future.seed = TRUE),
                         name = type) {
  # input checks:
  if (length(type) != 1 || !is.character(type)) {
    stop("type must be a character string.")
  }
  if (!(is.numeric(alpha) && (length(alpha) == 1))) {
    stop("alpha must be numeric and in [0, 0.5).")
  }
  if (!(alpha >= 0 && alpha < 0.5)) {
    stop("alpha must be numeric and in [0, 0.5).")
  }
  threshold_indicator <- !is.null(threshold)
  if (threshold_indicator == TRUE) {
    if (!(is.numeric(threshold) && (length(threshold) >= 1))) {
      stop("threshold must be numeric (vector).")
    }
    threshold <- unique(sort(threshold))
  }
  if (isFALSE(threshold_indicator)) {
    threshold <- 0
  }
  if (!(is.logical(full_history) && (length(full_history) == 1))) {
    stop("full_history must be TRUE or FALSE")
  }
  if (!(is.numeric(L) && (length(L) == 1))) {
    stop("L must be an integer greater than 0.")
  }
  if (!(L %% 1 == 0)) {
    stop("L must be an integer greater than 0.")
  }
  if (L <= 0) {
    stop("L must be an integer greater than 0.")
  }
  if (!(is.logical(cross_fit_g_models) && (length(cross_fit_g_models) == 1))) {
    stop("cross_fit_g_models must be TRUE or FALSE")
  }
  if (!(is.logical(save_cross_fit_models) &&
    (length(save_cross_fit_models) == 1))) {
    stop("save_cross_fit_models must be TRUE or FALSE")
  }
  if (!is.null(name)) {
    name <- as.character(name)
    if (length(name) != 1) {
      stop("name must be a character string.")
    }
    if (length(threshold) > 1 || any(threshold != 0)) {
      name <- paste(name, "(eta=", threshold, ")", sep = "")
    }
  }

  if (!is.list(future_args)) {
    stop("future_args must be a list.")
  }

  pl_args <- list(
    alpha = alpha,
    threshold = threshold,
    L = L,
    cross_fit_g_models = cross_fit_g_models,
    cross_fit_c_models = cross_fit_c_models,
    save_cross_fit_models = save_cross_fit_models,
    future_args = future_args,
    full_history = full_history
  )

  type <- tolower(type)
  if (type %in% c("ql", "rql", "q_learning", "q-learning")) {
    call <- "rql"
  } else if (type %in% c("drql", "rqvl", "qvl", "qv_learning", "qv-learning")) {
    call <- "drql"
  } else if (type %in% c("ptl", "policytree", "policy_tree")) {
    if (!requireNamespace("policytree")) {
      mes <- paste0(
        "The policytree package is required ",
        "to perform value searching using trees."
      )
      stop(mes)
    }
    call <- "ptl"
  } else if (type %in% c("owl", "bowl")) {
    if (!requireNamespace("DTRlearn2")) {
      mes <- paste0(
        "The DTRlearn2 package is required to perform value ",
        "searching using outcome-weighted learning."
      )
      stop(mes)
    }
    call <- "dtrlearn2_owl"
  } else if (type %in% c("earl")) {
    call <- "dyntxregime_earl"
  } else if (type %in% c("rwl")) {
    call <- "dyntxregime_rwl"
  } else if (type %in% c("drb", "blip")) {
    call <- "blip"
  } else {
    mes <- paste0(
      "Unknown type of policy learner. Use 'ql', 'drql', ",
      "'ptl', 'owl', 'earl' or 'rwl'."
    )
    stop(mes)
  }

  if ((threshold_indicator == TRUE) && (!(call %in% c("blip", "ptl")))) {
    mes <- paste0(
      "The threshold argument is only implemented for type 'blip' or 'ptl'. ",
      "Please set threshold = NULL."
    )
    stop(mes)
  }

  args <- append(pl_args, control)
  pl <- pl(call = call, args = args)
  class(pl) <- c("policy_learn", "function")
  attr(pl, "type") <- type
  attr(pl, "pl_args") <- pl_args
  attr(pl, "name") <- name

  return(pl)
}

pl <- function(call, args) {
  function(policy_data,
           g_models = NULL, g_functions = NULL, g_full_history = FALSE,
           q_models, q_full_history = FALSE,
           c_models = NULL, c_functions = NULL, c_full_history = FALSE,
           m_model = NULL, m_function = NULL, m_full_history = FALSE) {
    ## evaluating the learner:
    eval_args <- as.list(environment())
    args <- append(args, eval_args)
    do.call(what = call, args)
  }
}

#' @rdname policy_learn
#' @name policy_object
NULL

#' @rdname policy_learn
#' @export
print.policy_learn <- function(x, ...) {
  cat("Policy learner with arguments:")

  cp <- args(x)
  cp <- capture.output(print.function(cp))
  cp <- capture.output(print.function(x))
  cp <- paste(cp, collapse = "")
  cp <- gsub(" ", "", cp, fixed = TRUE)
  cp <- gsub(",", ", ", cp, fixed = TRUE)
  n_start <- 10
  n_end <- gregexpr(")", cp)[[1]][1]
  cp <- substr(cp, n_start, n_end - 1)

  cat("\n")
  cat(paste(strwrap(cp, 60), collapse = "\n"))
  cat("\n")
}

#' Fit Single c-function
#'
#' \code{fit_c_function} is used to fit a single right-censoring model.
#' \code{c_model} can eiter be of inherited class c_model or g_model for
#' discrete right-censoring.
#' @param history History object created by [get_history()]
#' @param c_model Censoring model, c-model/g-model, created by ?? or similar functions.
#' @returns Object of class "c_function".
#' @noRd
fit_c_function <- function(history, c_model){
  stage <- get_element(history, "stage", check_name = FALSE)
  c_model_class <- class(c_model)

  ## getting the event, time and the model matrix (H):
  event <- get_event(history)
  time <- get_time(history)
  time2 <- get_time2(history)
  H <- get_H(history)

  ## checking if right-censoring occur
  no_censoring <- all(event != 2)

  ## overwriting the c-model if no censoring occur:
  if (no_censoring == TRUE) {
    c_model <- c_no_censoring()
    c_model_class <- class(c_model)
  }

  ## fitting the model:
  if (inherits(c_model, "g_model")) {
    A <- as.numeric(!(event == c(2)))
    tryCatch({
      c_model <- c_model(A = A, H = H, action_set = c(0,1))
    }, error = function(e) {
      stop("Error fitting c_model: ", e$message)
    }, warning = function(w) {
      warning("Warning in c_model fitting: ", w$message)
    })
  } else {
    c_model <- c_model(event = event, time = time, time2 = time2, H = H)
  }

  c_function <- list(
    c_model = c_model,
    c_model_class = c_model_class,
    H_names = colnames(H),
    stage = stage
  )
  class(c_function) <- "c_function"

  return(c_function)
}

#' @export
predict.c_function <- function(object, new_history, ...){
  c_model <- get_element(object, "c_model")
  c_model_class <- get_element(object, "c_model_class")
  H_names <- get_element(object, "H_names")

  id_stage <- get_id_stage(new_history)
  new_H <- get_H(new_history)

  ## time points for evaluation of the c-model:
  new_time <- get_time(new_history)
  new_time2 <- get_time2(new_history)

  if ("g_model" %in% c_model_class) {
    surv <- matrix(predict(c_model, new_H = new_H)[, 2])
    surv <- cbind(1, surv)
    colnames(surv) <- c("surv_time", "surv_time2")
  } else {
    surv_time <- predict(c_model, new_H = new_H, new_time = new_time)
    surv_time2 <- predict(c_model, new_H = new_H, new_time = new_time2)
    surv <- cbind(surv_time, surv_time2)
    colnames(surv) <- c("surv_time", "surv_time2")
  }

  if (!all(complete.cases(surv))){
    mes <- "The c-function predictions have missing values."
    stop(mes)
  }

  if (!all((surv) >= 0 & (surv <= 1))){
    mes <- "The c-function predictions are not in [0,1]."
    stop(mes)
  }

  # including the id's and stage number(s)
  c_values <- data.table(id_stage, surv)
  setkeyv(c_values, c("id", "stage"))

  return(c_values)
}

#' Fit Censoring Functions
#'
#' Fits right-censoring models for each stage or a single model across all stages.
#'
#' @param policy_data Policy data object created by [policy_data()]
#' @param c_models Single c_model or list of K+1 c_models
#' @param full_history Logical; use full history (TRUE) or Markov-type history (FALSE)
#'
#' @return List of fitted censoring functions with class "c_functions"
#'
#' @details
#' The function handles two scenarios:
#' * Multiple models: One model per stage (length K+1)
#' * Single model: Same model applied across all stages
#'
#' @export
fit_c_functions <- function(policy_data, c_models, full_history = FALSE){
  K <- get_K(policy_data)

  # input checks:
  if (!(is.logical(full_history) & (length(full_history) == 1)))
    stop("full_history must be TRUE or FALSE")
  if (is.null(c_models))
    stop("Please provide c_models.")

  mes <- "c_models must be a single c_model (or g_model) or a list of K+1 c_model's (or g_model's)."
  if (is.list(c_models)) {
    ## updating class:
    c_models <- lapply(c_models, function(cm) {
      if (inherits(cm, "g_model")) {
        class(cm) <- c("c_model", class(cm))
      }
      return(cm)
    })
    tmp <- all(unlist(lapply(c_models, function(cm) inherits(cm, "c_model"))))
    if (!tmp)
      stop(mes)
    rm(tmp)
    if (length(c_models) != (K+1))
      stop(mes)
  } else {
    if (inherits(c_models, "g_model")) {
      ## updating class:
      class(c_models) <- c("c_model", class(c_models))
    }
    if (!inherits(c_models, "c_model"))
      stop(mes)
    if (full_history == TRUE)
      stop("full_history must be FALSE when a single c-model is provided.")
  }

  # fitting the c-models:
  if (is.list(c_models)){
    history <- lapply(1:(K+1),
                      function(s) get_history(policy_data,
                                              stage = s,
                                              full_history = full_history,
                                              event_set = c(0,1,2)))
    c_functions <- mapply(history,
                          c_models,
                          FUN = function(h, cm) fit_c_function(history = h,
                                                               c_model = cm),
                          SIMPLIFY = FALSE)
    names(c_functions) <- paste("stage_", 1:(K+1), sep = "")
  } else{
    history <- get_history(policy_data,
                           stage = NULL,
                           full_history = FALSE,
                           event_set = c(0,1,2))
    c_functions <- list(all_stages = fit_c_function(history, c_models))
  }

  class(c_functions) <- c("c_functions")
  attr(c_functions, "full_history") <- full_history

  return(c_functions)
}

#' @export
predict.c_functions <- function(object, new_policy_data, ...){
  K <- get_K(new_policy_data)
  full_history <- attr(object, "full_history")

  if (length(object) == (K+1)){
    history <- lapply(1:(K+1), function(s) get_history(new_policy_data,
                                                       stage = s,
                                                       full_history = full_history,
                                                       event_set = c(0,1,2)))
    values <- mapply(history,
                     object,
                     FUN = function(h, f) predict(f, h),
                     SIMPLIFY = FALSE)
    values <- rbindlist(values)
    setkeyv(values, c("id", "stage"))
  } else if (length(object) == 1){
    ## state history across all stages 1,...,K+1:
    history <- get_history(new_policy_data,
                           stage = NULL,
                           full_history = FALSE,
                           event_set = c(0,1,2))
    values <- predict(object[[1]], history)
  } else{
    stop("Provide either 1 or K+1 c-functions for evaluation.")
  }

  return(values)
}

sim_single_stage_right_cens <- function(n = 5e2, zeta = c(0.7, 0.2), type = "right"){

  d <- sim_single_stage(n = n)
  pd <- policy_data(data = d,
                    action = "A",
                    covariates = c("Z", "L", "B"),
                    utility = "U")

  ld <- pd$stage_data

  ld[stage == 1, time := 1]
  ld[stage == 2, time := 2]

  ld[stage == 2, Z := d$Z]
  ld[stage == 2, L := d$L]
  ld[stage == 2, B := d$B]

  ## simulating the right censoring time
  ## only depending on the baseline covariate Z:
  C <- c(rexp(n, 1) / exp((-1) * cbind(1, as.numeric(d$Z)) %*% zeta))

  ld[stage == 1, time_c := C]
  ld[stage == 2, time_c := C]

  ld[, delta := time_c >= time]

  ld[delta == FALSE , event := 2]
  ld[delta == FALSE, A := NA]
  ld[delta == FALSE & stage == 2, U := NA]
  ld[delta == FALSE & stage == 2, U_A0 := NA]
  ld[delta == FALSE & stage == 2, U_A1 := NA]

  ld[ , tmp := shift(delta, fill = TRUE), by = list(id)]
  ld <- ld[tmp == TRUE, ]
  ld[ , time := pmin(time, time_c)]
  ld[ , time_c := NULL]
  ld[ , tmp := NULL]
  ld[ , delta := NULL]

  if (type == "interval"){
    ld[, time2 := time]
    ld[, time := shift(time, fill = 0), by = list(id)]
  }

  return(ld)
}


test_that("c_cox returns the expected output", {

  library("mets")

  set.seed(1)
  ld <- sim_single_stage_right_cens(n = 5e2, type = "interval")
  pd <- policy_data(data = ld, type = "long", action = "A", time = "time", time2 = "time2")

  ## state history across all stages:
  his <- get_history(pd, event_set = c(0,1,2))
  fcf <- fit_c_function(
    history = his,
    c_model = c_cox(formula = ~ Z)
  )
  ref_cox <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z, data = ld)
  ref_cox$call <- NULL

  fcf2 <- fit_c_functions(pd, c_models = c_cox(formula = ~ Z))$all_stages$c_model$model

  expect_equal(
    coef(fcf$c_model$model),
    coef(ref_cox)
  )
  expect_equal(
    coef(fcf$c_model$model),
    coef(fcf2)
  )

  surv <- predict.c_function(fcf, new_history = his)

  ref_surv_time <- as.vector(predict(ref_cox, newdata = ld, times = ld$time, individual.time = TRUE)$surv)
  ref_surv_time2 <- as.vector(predict(ref_cox, newdata = ld, times = ld$time2, individual.time = TRUE)$surv)

  expect_equal(
    surv$surv_time,
    ref_surv_time
  )
  expect_equal(
    surv$surv_time2,
    ref_surv_time2
  )

  ## state history for a given stage
  his <- get_history(pd, event_set = c(0,1,2), stage = 2)
  fcf <- fit_c_function(
    history = his,
    c_model = c_cox(formula = ~ Z)
  )
  ref_cox <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z, data = ld[stage == 2,])
  ref_cox$call <- NULL

  fcf2 <- fit_c_functions(pd, c_models = list(c_cox(formula = ~ Z),c_cox(formula = ~ Z)))$stage_2$c_model$model

  expect_equal(
    coef(fcf$c_model$model),
    coef(ref_cox)
  )
  expect_equal(
    coef(fcf$c_model$model),
    coef(fcf2)
  )

  surv <- predict.c_function(fcf, new_history = his)

  ref_surv_time <- as.vector(predict(ref_cox, newdata = ld[stage == 2,], times = ld[stage == 2,]$time, individual.time = TRUE)$surv)
  ref_surv_time2 <- as.vector(predict(ref_cox, newdata = ld[stage == 2,], times = ld[stage == 2,]$time2, individual.time = TRUE)$surv)

  expect_equal(
    surv$surv_time,
    ref_surv_time
  )
  expect_equal(
    surv$surv_time2,
    ref_surv_time2
  )

  his <- get_history(pd, event_set = c(0,1,2), stage = 1)
  fcf <- fit_c_function(
    history = his,
    c_model = c_cox(formula = ~ Z)
  )
  ref_cox <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z, data = ld[stage == 1,])
  ref_cox$call <- NULL

  fcf2 <- fit_c_functions(pd, c_models = list(c_cox(formula = ~ Z),c_cox(formula = ~ Z)))$stage_1$c_model$model

  expect_equal(
    coef(fcf$c_model$model),
    coef(ref_cox)
  )
  expect_equal(
    coef(fcf$c_model$model),
    coef(fcf2)
  )

  surv <- predict.c_function(fcf, new_history = his)

  ref_surv_time <- as.vector(predict(ref_cox, newdata = ld[stage == 1,], times = ld[stage == 1,]$time, individual.time = TRUE)$surv)
  ref_surv_time2 <- as.vector(predict(ref_cox, newdata = ld[stage == 1,], times = ld[stage == 1,]$time2, individual.time = TRUE)$surv)

  expect_equal(
    surv$surv_time,
    ref_surv_time
  )
  expect_equal(
    surv$surv_time2,
    ref_surv_time2
  )

  ## full history
  his <- get_history(pd, event_set = c(0,1,2), stage = 2, full_history = TRUE)
  fcf <- fit_c_function(
    history = his,
    c_model = c_cox(formula = ~ Z_1 * A_1)
  )
  ref_cox <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z*tmp, data = ld[, tmp := shift(A), by = id][stage == 2,])
  ref_cox$call <- NULL

  fcf2 <- fit_c_functions(pd, full_history = TRUE, c_models = list(c_cox(formula = ~ Z_1),c_cox(formula = ~ Z_1 * A_1)))$stage_2$c_model$model

  expect_equal(
    unname(coef(fcf$c_model$model)),
    unname(coef(ref_cox))
  )
  expect_equal(
    unname(coef(fcf$c_model$model)),
    unname(coef(fcf2))
  )

  surv <- predict.c_function(fcf, new_history = his)

  ref_surv_time <- as.vector(predict(ref_cox, newdata = ld[, tmp := shift(A), by = id][stage == 2,], times = ld[stage == 2,]$time, individual.time = TRUE)$surv)
  ref_surv_time2 <- as.vector(predict(ref_cox, newdata = ld[, tmp := shift(A), by = id][stage == 2,], times = ld[stage == 2,]$time2, individual.time = TRUE)$surv)

  expect_equal(
    surv$surv_time,
    ref_surv_time
  )
  expect_equal(
    surv$surv_time2,
    ref_surv_time2
  )

})

#' @title Simulate Single-Stage Data
#'
#' @param n Number of observations.
#' @param par Named vector with distributional parameters.
#' \itemize{
#'  \item{} \code{k}: \eqn{\kappa}
#'  \item{} \code{d}: \eqn{\delta}
#'  \item{} \code{a}: \eqn{\alpha}
#'  \item{} \code{b}: \eqn{\beta}
#'  \item{} \code{c}: \eqn{\gamma}
#'  \item{} \code{p}: \eqn{\pi}
#' }
#' @param action_model Function used to specify the action/treatment probability (logit link).
#' @param utility_model Function used to specify the conditional mean utility.
#' @param seed Integer.
#' @param return_model If TRUE, the [lava::lvm] model is returned.
#' @param ... Additional arguments passed to [lava::lvm()].
#' @returns data.frame with n rows and columns Z, L, B, A, and U.
#' @details
#' \code{sim_single_stage} samples \code{n} iid observation
#' \eqn{O = (B, Z, L, A, U)} with the following distribution:
#' \deqn{
#' B \sim Bernoulli(\pi)\\
#' Z, L \sim Uniform([0,1])\\
#' A\mid Z,L,B \sim Bernoulli(expit\{\kappa Z^{-2}(Z+L-1) + \delta B)\\
#' U\mid Z,L,A \sim \mathcal{N}(Z+L+A\cdot\{\gamma Z + \alpha L + \beta\}, 1)
#' }
#' @export
sim_single_stage <- function(n=1e4,
                             par=c(k = .1,  d = .5, a = 1, b = -2.5, c = 3, p = .3),
                             action_model = function(Z, L, B, k, d){
                               k*(Z + L - 1)*Z^(-2) + d*(B==1)
                             },
                             utility_model = function(Z, L, A, a, b, c){
                               Z + L + A*(c*Z + a*L + b)
                             },
                             seed=NULL,
                             return_model=FALSE,
                             ...) {
  if (!is.null(seed)) set.seed(seed)
  m <- lava::lvm(...)
  lava::parameter(m) <- ~ a + b + c + d + k + p
  lava::distribution(m, ~Z+L+B+A) <- list(lava::uniform.lvm(),
                                          lava::uniform.lvm(),
                                          lava::binomial.lvm("identity"),
                                          lava::binomial.lvm("logit"))
  lava::regression(m, ~B) <- function(p){p}
  lava::regression(m, ~A) <- action_model
  lava::regression(m, ~U) <- utility_model

  if (return_model) return(m)
  lava::sim(m, n, p=par)
}

#' @title Simulate Single-Stage Multi-Action Data
#'
#' @param n Number of observations.
#' @param seed Integer.
#' @returns data.frame with n rows and columns z, x, a, and u.
#' @details
#' \code{sim_single_stage_multi_actions} samples \code{n} iid observation
#' \eqn{O = (z, x, a, u)} with the following distribution:
#' \deqn{
#' z, x \sim Uniform([0,1])\\
#' \tilde a \sim \mathcal{N}(0,1)\\
#' a \mid \tilde a \sim
#' \begin{cases}
#' 0 \quad if \quad \tilde a < -1\\
#' 1 \quad if \quad \tilde a -1 \leq a < 0.5\\
#' 2 \quad otherwise
#' \end{cases}\\
#' u \mid z, x \sim \mathcal{N}(x + z + I\{a=2\}(x-0.5) + I\{a=1\}(x^2 + z -0.5), 1)
#' }
#' @export
sim_single_stage_multi_actions <- function(n = 1e3, seed = NULL){
  if (!is.null(seed)) set.seed(seed)

  z <- stats::runif(n)
  x <- stats::runif(n)
  a0 <- stats::rnorm(mean = x, n = n)
  afun <- Vectorize(function(y){
    if (y<(-1)) return(0)
    else if (y<0.5) return(1)
    else return(2)
  })
  a <- afun(a0)
  u <- stats::rnorm(
    mean = x + z + (a==2)*(x-0.5) + (a==1)*(x*x+z-0.5),
    n = n
  )

  d <- data.frame(
    z = z,
    x = x,
    a = a,
    u = u
  )
  return(d)
}


#' @title Simulate Two-Stage Data
#'
#' @param n Number of observations.
#' @param par Named vector with distributional parameters.
#' \itemize{
#'  \item{} \code{gamma}: \eqn{\gamma}
#'  \item{} \code{beta}: \eqn{\beta}
#' }
#' @param seed Integer.
#' @param action_model_1 Function used to specify the action/treatment at stage 1.
#' @param action_model_2 Function used to specify the action/treatment at stage 2.
#' @param deterministic_rewards Logical. If TRUE, the deterministic reward
#' contributions are returned as well (columns U_1_A0, U_1_A1, U_2_A0, U_2_A1).
#' @details
#' \code{sim_two_stage} samples \code{n} iid observation
#' \eqn{O} with the following distribution:
#' \eqn{BB} is a random categorical variable with levels \code{group1},
#' \code{group2}, and \code{group3}. Furthermore,
#' \deqn{
#' B \sim \mathcal{N}(0,1)\\
#' L_{1} \sim \mathcal{N}(0, 1)\\
#' C_{1} \mid L_{1} \sim \mathcal{N}(L_1, 1)\\
#' A_1 \mid C_1 \sim Bernoulli(expit(\beta C_1))\\
#' L_{2} \sim \mathcal{N} (0, 1)\\
#' C_{2} \mid A_1, L_1 \sim \mathcal{N}(\gamma L_1 + A_1, 1)\\
#' A_2 \mid C_2 \sim Bernoulli(expit(\beta C_2))\\
#' L_{3} \sim \mathcal{N} (0, 1)
#' }
#' The rewards are calculated as
#' \deqn{
#' U_1 = L_1\\
#' U_2 = A_1\cdot C_1 + L_2 \\
#' U_3 = A_2\cdot C_2 + L_3.
#' }
#' @returns [data.table::data.table] with n rows and columns B, BB, L_1, C_1, A_1, L_2, C_2,
#' A_2, L_3, U_1, U_2, U_3 (,U_1_A0, U_1_A1, U_2_A0, U_2_A1).
#' @export
sim_two_stage <- function(n = 1e4,
                          par = c(gamma = 0.5,  beta = 1),
                          seed = NULL,
                          action_model_1 = function(C_1, beta, ...)
                            stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_1)),
                          action_model_2 = function(C_2, beta, ...)
                            stats::rbinom(n = NROW(C_2), size = 1, prob = lava::expit(beta * C_2)),
                          deterministic_rewards = FALSE){
  if (!is.null(seed)) set.seed(seed)

  gamma <- getElement(par, "gamma")
  beta <- getElement(par, "beta")

  # baseline
  B <- stats::rnorm(n = n)
  BB <- sample(c("group1", "group2", "group3"), n, replace = TRUE)

  # stage 1
  L_1 <- stats::rnorm(n = n)
  C_1 <- L_1 + stats::rnorm(n = n)
  U_1 <- L_1
  A_1 <- action_model_1(C_1 = C_1, L_1 = L_1, beta = beta, gamma = gamma)
  U_1_0 <- C_1 * 0
  U_1_1 <- C_1

  # stage 2
  L_2 <- stats::rnorm(n = n, 0, 1)
  C_2 <- (gamma * L_1 + A_1) + stats::rnorm(n = n)
  U_2 <- A_1 * C_1 + L_2
  A_2 <- action_model_2(C_1 = C_1, L_1 = L_1, C_2 = C_2, L_2 = L_2, A_1 = A_1, beta = beta, gamma = gamma)
  U_2_0 <- C_2 * 0
  U_2_1 <- C_2

  L_3 <- stats::rnorm(n = n)
  U_3 <- A_2 * C_2 + L_3

  data <- data.table(
    B = B,
    BB = BB,
    L_1 = L_1,
    C_1 = C_1,
    A_1 = A_1,
    L_2 = L_2,
    C_2 = C_2,
    A_2 = A_2,
    L_3 = L_3,
    U_1 = U_1,
    U_2 = U_2,
    U_3 = U_3
  )
  if (deterministic_rewards == TRUE){
    U_1_A0 <- U_1_A1 <- U_2_A0 <- U_2_A1 <- NULL
    data[, U_1_A0 := U_1_0]
    data[, U_1_A1 := U_1_1]
    data[, U_2_A0 := U_2_0]
    data[, U_2_A1 := U_2_1]
  }

  return(data)
}

#' @title Simulate Two-Stage Multi-Action Data
#'
#' @param n Number of observations.
#' @param par Named vector with distributional parameters.
#' \itemize{
#'  \item{} \code{gamma}: \eqn{\gamma}
#'  \item{} \code{beta}: \eqn{\beta}
#'  \item{} \code{prob}: \eqn{p}
#' }
#' @param seed Integer.
#' @param action_model_1 Function used to specify the dichotomous
#' action/treatment at stage 1.
#' @details
#' \code{sim_two_stage_multi_actions} samples \code{n} iid observation
#' \eqn{O} with the following distribution:
#' \eqn{BB} is a random categorical variable with levels \code{group1},
#' \code{group2}, and \code{group3}. Furthermore,
#' \deqn{
#' B \sim \mathcal{N}(0,1)\\
#' L_{1} \sim \mathcal{N}(0, 1)\\
#' C_{1} \mid L_{1} \sim \mathcal{N}(L_1, 1)\\
#' P(A_1='yes'\mid C_1) =  expit(\beta C_1)\\
#' P(A_1='no'\mid C_1) = 1 - P(A_1='yes' \mid C_1)\\
#' L_{2} \sim \mathcal{N} (0, 1)\\
#' C_{2} \mid A_1, L_1 \sim \mathcal{N}(\gamma L_1 + A_1, 1)\\
#' P(A_2='yes') = p_1\\
#' P(A_2='no') = p_2\\
#' P(A_2='default') = p_3\\
#' L_{3} \sim \mathcal{N} (0, 1)
#' }
#' The rewards are calculated as
#' \deqn{
#' U_1 = L_1\\
#' U_2 = A_1\cdot C_1 + L_2 \\
#' U_3 = A_2\cdot C_2 + L_3.
#' }
#' @returns [data.table::data.table] with n rows and columns B, BB, L_1, C_1, A_1, L_2, C_2,
#' A_2, L_3, U_1, U_2, U_3.
#' @export
sim_two_stage_multi_actions <- function(n=1e3,
                                        par=list(gamma = 0.5,
                                                 beta = 1,
                                                 prob = c(0.2, 0.4, 0.4)),
                                        seed=NULL,
                                        action_model_1 = function(C_1, beta, ...)
                                          stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_1))){
  if (!is.null(seed)) set.seed(seed)

  gamma <- getElement(par, "gamma")
  beta <- getElement(par, "beta")
  prob <- getElement(par, "prob")

  # baseline
  B <- stats::rnorm(n = n)
  BB <- sample(c("group1", "group2", "group3"), n, replace = TRUE)

  # stage 1
  L_1 <- stats::rnorm(n = n)
  C_1 <- L_1 + stats::rnorm(n = n)
  U_1 <- L_1
  a_1 <- action_model_1(C_1 = C_1, L_1 = L_1, beta = beta, gamma = gamma)
  A_1 <- ifelse(a_1, "yes", "no")
  U_1_0 <- C_1 * 0
  U_1_1 <- C_1

  # stage 2
  L_2 <- stats::rnorm(n = n, 0, 1)
  C_2 <- (gamma * L_1 + a_1) + stats::rnorm(n = n)
  U_2 <- a_1 * C_1 + L_2
  A_2 <- sample(c("yes", "no", "default"), prob = prob, size = n, replace = TRUE)
  U_2_0 <- C_2 * 0
  U_2_1 <- C_2

  L_3 <- stats::rnorm(n = n)
  U_3 <- (A_2 == "yes") * C_2 + L_3

  data <- data.table(
    B = B,
    BB = BB,
    L_1 = L_1,
    C_1 = C_1,
    A_1 = A_1,
    L_2 = L_2,
    C_2 = C_2,
    A_2 = A_2,
    L_3 = L_3,
    U_1 = U_1,
    U_2 = U_2,
    U_3 = U_3
  )

  return(data)

}


sim_multi_stage_obs <- function(a,
                                tau,
                                gamma,
                                alpha,
                                sigma,
                                beta,
                                psi,
                                xi,
                                ...){
  a_fun <- a
  stage_vec <- vector("numeric")
  entry_vec <- vector("numeric")
  exit_vec <- vector("numeric")
  event_vec <- vector("numeric")
  a_vec <- vector("numeric")

  # latent variable
  w <- stats::rnorm(n=1)

  # baseline variable
  b <- stats::rbinom(n = 1, size = 1, prob = xi)

  # stage specific variables
  x_vec <- vector("numeric")
  x_lead_vec <- vector("numeric")

  stage <- 1
  entry_vec <- c(entry_vec, 0)
  t <- 0
  x_lead <- 0

  while (t<tau){
    x <- stats::rnorm(n = 1, mean = alpha[1] + (alpha[2] * t) + (alpha[3]*t^2) + (alpha[4] * x_lead) + (alpha[5] * b), sd = sigma)
    a <- a_fun(stage = stage, t = t, x = x, x_lead = x_lead, b = b, beta = beta)

    exit_vec <- c(exit_vec, t)
    stage_vec <- c(stage_vec, stage)
    event_vec <- c(event_vec, 0)
    x_vec <- c(x_vec, x)
    a_vec <- c(a_vec, a)
    x_lead_vec <- c(x_lead_vec, x_lead)

    if (a == 1){
      entry_vec <- c(entry_vec, t)
    }
    # the time increment comes from an exponential distribution with mean exp(gamma[1] + gamma[2] * x + gamma[3] * w)
    # remember that mean(t_increment)  = 1 / rate
    rate <- exp((gamma[1] + gamma[2] * x + gamma[3] * w))
    t_increment <- if(a == 1) stats::rexp(n = 1, rate = rate) else Inf

    t <- t + t_increment + psi # minimum time increment psi
    stage <- stage + 1
    x_lead <- x
  }

  if (a == 0){
    stage_vec <- c(stage_vec, stage)
    entry_vec <- c(entry_vec, last(exit_vec))
    exit_vec <- c(exit_vec, last(exit_vec))
    event_vec <- c(event_vec, 1)
    a_vec <- c(a_vec, NA)
    x_vec <- c(x_vec, NA)
    x_lead_vec <- c(x_lead_vec, NA)
  }

  if (a == 1){
    stage_vec <- c(stage_vec, stage)
    exit_vec <- c(exit_vec, tau)
    event_vec <- c(event_vec, 2)
    a_vec <- c(a_vec, NA)
    x_vec <- c(x_vec, 0)
    x_lead_vec <- c(x_lead_vec, 0)
  }

  stage_data <- matrix(c(stage_vec, entry_vec, exit_vec, event_vec, a_vec, x_vec, x_lead_vec), ncol = 7, byrow = FALSE)
  colnames(stage_data) <- c("stage", "entry", "exit", "event", "A", "X", "X_lead")

  baseline_data <- matrix(b, ncol = 1) # cbind(b, w)
  colnames(baseline_data) <- c("B") # c("B", "W")

  return(list(stage_data = stage_data, baseline_data = baseline_data))
}

#' @title Simulate Multi-Stage Data
#'
#' @param n Number of observations.
#' @param par Named list with distributional parameters.
#' \itemize{
#'  \item{} \code{tau}: \eqn{\tau}
#'  \item{} \code{gamma}: \eqn{\gamma}
#'  \item{} \code{alpha}: \eqn{\alpha}
#'  \item{} \code{beta}: \eqn{\beta}
#'  \item{} \code{psi}: \eqn{\psi}
#'  \item{} \code{xi}: \eqn{\xi}
#' }
#' @param seed Integer.
#' @param a Function used to specify the action/treatment at every stage.
#' @details
#' \code{sim_multi_stage} samples \code{n} iid observation
#' \eqn{O} with the following distribution:
#' \deqn{
#' W \sim \mathcal{N}(0, 1)\\
#' B \sim Ber(\xi)
#' }
#' For \eqn{k\geq 1} let
#' \deqn{
#' (T_k - T_{k-1})| X_{k-1}, A_{k-1}, W \sim
#' \begin{cases}
#' Exp\Big\{\exp\left(\gamma^T [1, X_{k-1}, W] \right) \Big\} + \psi \quad A_{k-1} = 1\\
#' \infty \quad A_{k-1} = 0
#' \end{cases}\\
#' X_{k}\mid T_k, X_{k-1}, B \sim
#' \begin{cases}
#' \mathcal{N}\left\{ \alpha^T [1, T_k, T^2_k, X_{k-1}, B], 1\right\} \quad T_k < \infty \\
#' 0 \quad T_k = \infty
#' \end{cases}\\
#' A_k \mid X_k, T_k \sim
#' \begin{cases}
#' Ber\left\{ expit\left(\beta^T[1, T_{k}^2, X_k] \right)\right\} \quad T_k < \infty\\
#' 0 \quad T_k = \infty,
#' \end{cases}
#' }
#' Note that \eqn{\psi} is the minimum increment.
#' @returns list with elements \code{stage_data} ([data.table::data.table]) and
#'  \code{baseline_data} ([data.table::data.table]).
#' @export
sim_multi_stage <- function(n,
                            par = list(tau = 10,
                                       gamma = c(0, -0.2, 0.3),
                                       alpha = c(0, 0.5, 0.2, -0.5, 0.4),
                                       beta = c(3, -0.5, -0.5),
                                       psi = 1,
                                       xi = 0.3),
                            a = function(t, x, beta, ...){
                              prob <- lava::expit(beta[1] + (beta[2] * t^2) + (beta[3] * x))
                              stats::rbinom(n = 1, size = 1, prob = prob)
                            },
                            seed = NULL){
  par <- append(par, list(sigma = 1))

  if (!is.null(seed)) set.seed(seed)

  if (!is.null(a)){
    par <- append(par, list(a = a))
  }

  l <- sapply(
    1:n,
    function(id){
      d <- do.call(what = "sim_multi_stage_obs", par)
      stage_data <- d$stage_data
      baseline_data <- d$baseline_data

      stage_data <- cbind(id = id, stage_data)
      baseline_data <- cbind(id = id, baseline_data)

      return(list(stage_data = stage_data, baseline_data = baseline_data))
    },
    simplify = "array"
  )
  stage_data <- do.call(what  = "rbind", l["stage_data",])
  stage_data <- as.data.table(stage_data)
  U <- exit <- entry <- A <- X <- event <- U_A0 <- U_A1 <-  NULL

  stage_data[, U := (exit - entry) + shift(ifelse(!is.na(A), -X * A, 0), fill = 0)]
  stage_data[event %in% c(0), U_A0 := 0]
  stage_data[event %in% c(0), U_A1 := -X]
  stage_data[event == 2, event := 1]
  stage_data[, A := as.character(A)]

  setnames(stage_data, "exit", "t")
  stage_data[, entry := NULL]
  setcolorder(stage_data, c("id", "stage", "event"))
  setindex(stage_data, NULL)

  id <- NULL
  baseline_data <- as.data.table(do.call(what  = "rbind", l["baseline_data",]))
  baseline_data[, id := as.numeric(id)]

  return(list(stage_data = stage_data, baseline_data = baseline_data))
}

test_that("the implementation of owl agrees with direct application of DTRlearn2::owl in the single stage case.",{
  library("DTRlearn2")

  d1 <- sim_single_stage(200, seed=1)
  pd1 <- policy_data(d1,
                     action="A",
                     covariates = list("Z", "B", "L"),
                     utility="U")

  H <- scale(d1[, c("B", "Z", "L")])
  AA <- 2*d1[, "A"]-1
  RR <- d1[, "U"]

  pi_model <- glm(A~B+Z+L, data = d1, family = binomial())
  pi <- predict(pi_model, type = "response")
  pi <- pi * d1$A + (1-pi) * (1-d1$A)

  set.seed(1)
  owl1 <- owl(H = H, AA = AA, RR = RR, n = nrow(d1), K = 1, pi = pi)
  owl1_dd <- predict(owl1, H = H, K = 1)$treatment[[1]]
  owl1_d <- unname(as.character(unlist((owl1_dd + 1)/2)))

  pl <- policy_learn(type = "owl", control = control_owl())

  set.seed(1)
  owl2 <- pl(policy_data = pd1, g_models = g_glm())
  owl2_d <- unname(unlist(get_policy(owl2)(pd1)[, "d"]))
  expect_equal(owl1_d,owl2_d)
})

test_that("the implementation of owl agrees with direct application of DTRlearn2::owl in the two stage case.",{
  d <- sim_two_stage(200, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  H1 <- d[, c("B", "L_1", "C_1")]
  H2 <- d[, c("B", "L_2", "C_2")]
  H <- list(H1, H2)
  H <- lapply(H, scale)
  AA <- list(2*d$A_1-1, 2*d$A_2-1)

  pi1_model <- glm(A_1~B+C_1+L_1, data = d, family = binomial())
  pi1 <- predict(pi1_model, type = "response")
  pi1 <- pi1 * d$A_1 + (1-pi1) * (1-d$A_1)
  pi2_model <- glm(A_2~B+C_2+L_2, data = d, family = binomial())
  pi2 <- predict(pi2_model, type = "response")
  pi2 <- pi2 * d$A_2 + (1-pi2) * (1-d$A_2)
  pi <- list(pi1, pi2)

  RR <- list(d$U_2, d$U_3)

  library("DTRlearn2")
  set.seed(1)
  owl1 <- owl(H = H, AA = AA, RR = RR, n = nrow(d), K = 2, pi = pi)
  owl_pred <- predict(owl1, H = H, K = 2)
  owl1_dd1 <- owl_pred$treatment[[1]]
  owl1_dd2 <- owl_pred$treatment[[2]]
  owl1_d1 <- unname(as.character(unlist((owl1_dd1 + 1)/2)))
  owl1_d2 <- unname(as.character(unlist((owl1_dd2 + 1)/2)))

  pl <- policy_learn(type = "owl",
                     control = control_owl(policy_vars = c("B", "L", "C"),
                                           reuse_scales = TRUE))
  set.seed(1)
  owl2 <- pl(policy_data = pd,
             g_models = list(g_glm(~B + L + C), g_glm(~B + L + C)))

  tmp <- get_policy(owl2)(pd)
  owl2_d1 <- unname(unlist(tmp[stage == 1, "d", with = FALSE]))
  owl2_d2 <- unname(unlist(tmp[stage == 2, "d", with = FALSE]))

  expect_equal(
    unname(owl2$owl_object$pi[[1]]),
    unname(pi[[1]]),
    tolerance = 1e-10
  )
  expect_equal(
    unname(owl2$owl_object$pi[[2]]),
    unname(pi[[2]]),
    tolerance = 1e-10
  )
  expect_equal(owl1_d1,owl2_d1)
  expect_equal(owl1_d2,owl2_d2)

  # varying stage action sets:
  d2 <- copy(d)
  d2$A_1[d2$A_1 == 1] <- "t1"
  d2$A_1[d2$A_1 == 0] <- "t2"
  d2$A_2[d2$A_2 == 1] <- "t3"
  d2$A_2[d2$A_2 == 0] <- "t4"
  pd2 <- policy_data(d2,
                     action = c("A_1", "A_2"),
                     baseline = c("BB", "B"),
                     covariates = list(L = c("L_1", "L_2"),
                                       C = c("C_1", "C_2")),
                     utility = c("U_1", "U_2", "U_3"))

  set.seed(1)
  owl3 <- pl(policy_data = pd2,
             g_models = list(g_glm(~B + L + C), g_glm(~B + L + C)))
  tmp3 <- get_policy(owl3)(pd2)
  owl3_d1 <- unname(unlist(tmp3[stage == 1, "d", with = FALSE]))
  owl3_d2 <- unname(unlist(tmp3[stage == 2, "d", with = FALSE]))

  owl1_d1 <- unname(owl1_dd1[,1])
  owl1_d1[owl1_d1 == -1] <- "t2"
  owl1_d1[owl1_d1 == 1] <- "t1"
  owl1_d2 <- unname(owl1_dd2[,1])
  owl1_d2[owl1_d2 == -1] <- "t4"
  owl1_d2[owl1_d2 == 1] <- "t3"

  expect_equal(owl1_d1,owl3_d1)
  expect_equal(owl1_d2,owl3_d2)
})

test_that("policy_learn with type owl runs as intended", {
  d <- sim_two_stage(200, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  pl <- policy_learn(type = "owl",
                     control = control_owl(),
                     L = 2,
                     alpha = 0,
                     save_cross_fit_models = TRUE)

  po <- pl(pd,
           g_models = g_glm(),
           q_models = q_glm())

  expect_true(
    is.null(po$g_function)
  )
  expect_true(
    !is.null(po$g_functions_cf)
  )

  expect_true(
    is.data.table(get_policy(po)(pd))
  )
  expect_true(
    all(complete.cases(get_policy(po)(pd)))
  )

})

test_that("input to policy_learn with type owl handles incorrect input in a multi-stage setup.",{
  d <- sim_multi_stage(200, seed = 1)
   ## changing administrative right-censoring events to terminal events:
  d$stage_data[event == 2, U := 10]
  d$stage_data[event == 2, event := 1]

  # constructing policy_data object:
  pd <- policy_data(data = d$stage_data,
                    baseline_data = d$baseline_data,
                    type = "long",
                    id = "id",
                    stage = "stage",
                    event = "event",
                    action = "A",
                    utility = "U")
  pd <- partial(pd, 3)

  pl <- policy_learn(type = "owl")

  expect_error(policy_eval(policy_data = pd,
                           policy_learn = pl), "owl is only implemented for a fixed number of stages.")
})

test_that("get_policy.ql returns a policy", {
  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
    action = "A",
    covariates = list("Z", "B", "L"),
    utility = "U"
  )

  pl <- policy_learn(type = "ql")
  expect_error(
    {
      p <- get_policy(pl(pd, q_models = q_glm()))
    },
    NA
  )
  expect_true(
    inherits(p, what = "policy")
  )
  expect_error(
    p(pd),
    NA
  )
})

test_that("get_policy_functions.ql agrees with get_policy.ql", {
  d <- sim_two_stage(500, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  # not full history:
  pl <- policy_learn(type = "ql")
  po <- pl(pd, q_models = q_glm())

  expect_error(
    get_policy_functions(po),
    "stage argument is missing."
  )

  pf_1 <- get_policy_functions(po, stage = 1)
  his_1 <- get_history(pd, stage = 1,
                       full_history = FALSE)
  H1 <- his_1$H[, -c("id", "stage"), with = FALSE]
  pa_1 <- pf_1(H1)

  pf_2 <- get_policy_functions(po, stage = 2)
  his_2 <- get_history(pd, stage = 2,
                       full_history = FALSE)
  H2 <- his_2$H[, -c("id", "stage"), with = FALSE]
  H2 <- cbind(H2, his_2$U[, -c("id", "stage", "U_bar"), with = FALSE])
  pa_2 <- pf_2(H2)

  ## benchmark
  pa <- get_policy(po)(pd)

  expect_equal(
    pa[stage ==1]$d,
    pa_1
  )
  expect_equal(
    pa[stage ==2]$d,
    pa_2
  )

  # full history:
  pl <- policy_learn(type = "ql")
  po <- pl(pd, q_models = q_glm(), q_full_history = TRUE)
  pf_1 <- get_policy_functions(po, stage = 1)
  his_1 <- get_history(pd, stage = 1,
                       full_history = TRUE)
  H1 <- his_1$H[, -c("id", "stage"), with = FALSE]
  H1 <- cbind(H1, his_1$U[, -c("id", "stage", "U_bar"), with = FALSE])
  pa_1 <- pf_1(H1)

  pf_2 <- get_policy_functions(po, stage = 2)
  his_2 <- get_history(pd, stage = 2,
                       full_history = TRUE)
  H2 <- his_2$H[, -c("id", "stage"), with = FALSE]
  H2 <- cbind(H2, his_2$U[, -c("id", "stage", "U_bar"), with = FALSE])
  pa_2 <- pf_2(H2)

  ## benchmark
  pa <- get_policy(po)(pd)

  expect_equal(
    pa[stage ==1]$d,
    pa_1
  )
  expect_equal(
    pa[stage ==2]$d,
    pa_2
  )
})

test_that("get_policy_functions.ql handles deterministic rewards", {
  d <- sim_two_stage(500, seed=1, deterministic_rewards = TRUE)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"),
                    deterministic_rewards = list(
                      U_A0 = c("U_1_A0", "U_2_A0"),
                      U_A1 = c("U_1_A1", "U_2_A1")
                    ))

  # not full history:
  pl <- policy_learn(type = "ql")
  po <- pl(pd, q_models = q_glm())

  pf_1 <- get_policy_functions(po, stage = 1)
  his_1 <- get_history(pd, stage = 1,
                       full_history = FALSE)
  H1 <- his_1$H[, -c("id", "stage"), with = FALSE]
  H1 <- cbind(H1, his_1$U[, -c("id", "stage", "U_bar"), with = FALSE])
  pa_1 <- pf_1(H1)

  pf_2 <- get_policy_functions(po, stage = 2)
  his_2 <- get_history(pd, stage = 2,
                       full_history = FALSE)
  H2 <- his_2$H[, -c("id", "stage"), with = FALSE]
  H2 <- cbind(H2, his_2$U[, -c("id", "stage", "U_bar"), with = FALSE])
  pa_2 <- pf_2(H2)

  ## benchmark
  pa <- get_policy(po)(pd)

  expect_equal(
    pa[stage ==1]$d,
    pa_1
  )
  expect_equal(
    pa[stage ==2]$d,
    pa_2
  )

  # full history:
  pl <- policy_learn(type = "ql")
  po <- pl(pd, q_models = q_glm(), q_full_history = TRUE)
  pf_1 <- get_policy_functions(po, stage = 1)
  his_1 <- get_history(pd, stage = 1,
                       full_history = TRUE)
  H1 <- his_1$H[, -c("id", "stage"), with = FALSE]
  H1 <- cbind(H1, his_1$U[, -c("id", "stage", "U_bar"), with = FALSE])
  pa_1 <- pf_1(H1)

  pf_2 <- get_policy_functions(po, stage = 2)
  his_2 <- get_history(pd, stage = 2,
                       full_history = TRUE)
  H2 <- his_2$H[, -c("id", "stage"), with = FALSE]
  H2 <- cbind(H2, his_2$U[, -c("id", "stage", "U_bar"), with = FALSE])
  pa_2 <- pf_2(H2)

  ## benchmark
  pa <- get_policy(po)(pd)

  expect_equal(
    pa[stage ==1]$d,
    pa_1
  )
  expect_equal(
    pa[stage ==2]$d,
    pa_2
  )
})

test_that("get_policy_functions.ql handles partially missing deterministic rewards", {
  d <- sim_two_stage(500, seed=1, deterministic_rewards = TRUE)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"),
                    deterministic_rewards = list(
                      U_A0 = c("U_1_A0", "U_2_A0"),
                      U_A1 = c("U_1_A1", "U_2_A1")
                    ))

  # not full history:
  pl <- policy_learn(type = "ql")
  po <- pl(pd, q_models = q_glm())

  pf_1 <- get_policy_functions(po, stage = 1)
  his_1 <- get_history(pd, stage = 1,
                       full_history = FALSE)
  H1 <- his_1$H[, -c("id", "stage"), with = FALSE]
  H1 <- cbind(H1, his_1$U[, -c("id", "stage", "U_bar", "U_A0"), with = FALSE])
  pa_1 <- pf_1(H1)

  pf_2 <- get_policy_functions(po, stage = 2)
  his_2 <- get_history(pd, stage = 2,
                       full_history = FALSE)
  H2 <- his_2$H[, -c("id", "stage"), with = FALSE]
  H2 <- cbind(H2, his_2$U[, -c("id", "stage", "U_bar", "U_A0"), with = FALSE])
  pa_2 <- pf_2(H2)

  ## benchmark
  pa <- get_policy(po)(pd)

  expect_equal(
    pa[stage ==1]$d,
    pa_1
  )
  expect_equal(
    pa[stage ==2]$d,
    pa_2
  )

  # full history:
  pl <- policy_learn(type = "ql")
  po <- pl(pd, q_models = q_glm(), q_full_history = TRUE)
  pf_1 <- get_policy_functions(po, stage = 1)
  his_1 <- get_history(pd, stage = 1,
                       full_history = TRUE)
  H1 <- his_1$H[, -c("id", "stage"), with = FALSE]
  H1 <- cbind(H1, his_1$U[, -c("id", "stage", "U_bar", "U_A0"), with = FALSE])
  pa_1 <- pf_1(H1)

  pf_2 <- get_policy_functions(po, stage = 2)
  his_2 <- get_history(pd, stage = 2,
                       full_history = TRUE)
  H2 <- his_2$H[, -c("id", "stage"), with = FALSE]
  H2 <- cbind(H2, his_2$U[, -c("id", "stage", "U_bar", "U_A0"), with = FALSE])
  pa_2 <- pf_2(H2)

  ## benchmark
  pa <- get_policy(po)(pd)

  expect_equal(
    pa[stage ==1]$d,
    pa_1
  )
  expect_equal(
    pa[stage ==2]$d,
    pa_2
  )
})

test_that("get_policy_functions.ql handles realistic actions", {
  d <- sim_two_stage(500, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  # not full history:
  pl <- policy_learn(alpha = 0.1, type = "ql")
  po <- pl(pd, q_models = q_glm(), g_models = g_glm())

  pf_1 <- get_policy_functions(po, stage = 1)
  his_1 <- get_history(pd, stage = 1,
                       full_history = FALSE)
  H1 <- his_1$H[, -c("id", "stage"), with = FALSE]
  pa_1 <- pf_1(H1)

  pf_2 <- get_policy_functions(po, stage = 2)
  his_2 <- get_history(pd, stage = 2,
                       full_history = FALSE)
  H2 <- his_2$H[, -c("id", "stage"), with = FALSE]
  H2 <- cbind(H2, his_2$U[, -c("id", "stage", "U_bar", "U_A0"), with = FALSE])
  pa_2 <- pf_2(H2)

  ## benchmark
  pa <- get_policy(po)(pd)

  expect_equal(
    pa[stage ==1]$d,
    pa_1
  )
  expect_equal(
    pa[stage ==2]$d,
    pa_2
  )

  # full history:
  pl <- policy_learn(alpha = 0.1, type = "ql")
  po <- pl(pd, q_models = q_glm(), q_full_history = TRUE, g_models = g_glm(), g_full_history = FALSE)
  pf_1 <- get_policy_functions(po, stage = 1)
  his_1 <- get_history(pd, stage = 1,
                       full_history = TRUE)
  H1 <- his_1$H[, -c("id", "stage"), with = FALSE]
  H1 <- cbind(H1, his_1$U[, -c("id", "stage", "U_bar", "U_A0"), with = FALSE])
  H1$L <- H1$L_1
  H1$C <- H1$C_1
  pa_1 <- pf_1(H1)

  pf_2 <- get_policy_functions(po, stage = 2)
  his_2 <- get_history(pd, stage = 2,
                       full_history = TRUE)
  H2 <- his_2$H[, -c("id", "stage"), with = FALSE]
  H2 <- cbind(H2, his_2$U[, -c("id", "stage", "U_bar", "U_A0"), with = FALSE])
  H2$L <- H2$L_2
  H2$C <- H2$C_2
  pa_2 <- pf_2(H2)

  ## benchmark
  pa <- get_policy(po)(pd)

  expect_equal(
    pa[stage ==1]$d,
    pa_1
  )
  expect_equal(
    pa[stage ==2]$d,
    pa_2
  )
})

test_that("get_policy_functions.ql handles varying stage action set", {
  d <- sim_two_stage_multi_actions(n=500, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  # not full history:
  pl <- policy_learn(alpha = 0.05, type = "ql")
  po <- pl(pd, q_models = q_glm(), g_models = list(g_glm(), g_empir()))

  pf_1 <- get_policy_functions(po, stage = 1)
  his_1 <- get_history(pd, stage = 1,
                       full_history = FALSE)
  H1 <- his_1$H[, -c("id", "stage"), with = FALSE]
  pa_1 <- pf_1(H1)

  pf_2 <- get_policy_functions(po, stage = 2)
  his_2 <- get_history(pd, stage = 2,
                       full_history = FALSE)
  H2 <- his_2$H[, -c("id", "stage"), with = FALSE]
  pa_2 <- pf_2(H2)

  ## benchmark
  pa <- get_policy(po)(pd)

  expect_equal(
    pa[stage ==1]$d,
    pa_1
  )
  expect_equal(
    pa[stage ==2]$d,
    pa_2
  )
})


test_that("policy_learn with type = 'ql' works as intended",{
  d <- sim_two_stage(200, seed=1)

  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  ql <- policy_learn(type = "ql",
                     alpha = 0,
                     L = 1)

  expect_error(
    ql(policy_data = pd, q_models = q_glm()),
    NA
  )
  expect_error(
    ql(policy_data = pd, q_models = list(q_glm(), q_glm())),
    NA
  )

  ql <- policy_learn(type = "ql",
                     alpha = 0.1,
                     L = 1)

  expect_error(
    ql(policy_data = pd, q_models = q_glm()),
    "Please provide g_models."
  )

  expect_error({
    po <- ql(policy_data = pd, q_models = q_glm(), g_glm())
  }, NA)

  expect_error({
    p <- get_policy(po)
  }, NA)

  expect_s3_class(p(pd), class = "data.table")
})

test_that("policy_learn with type ql handles varying action sets", {
  d <- sim_two_stage_multi_actions(n = 1e2)
  pd <- policy_data(
    data = d,
    action = c("A_1", "A_2"),
    baseline = c("B", "BB"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )
  pl <- policy_learn(type = "ql")

  expect_error(
    po <- pl(pd, q_models = q_glm()),
    NA
  )
  expect_error(
    pa <- get_policy(po)(pd),
    NA
  )
  expect_error(
    pe <- policy_eval(pd,
      policy = get_policy(po),
      g_models = list(g_glm(), g_rf())
    ),
    NA
  )
})

sim_single_stage_right_cens <- function(n = 2e3, zeta = c(0.7, 0.2), type = "right"){

  d <- sim_single_stage(n = n)
  pd <- policy_data(data = d,
                    action = "A",
                    covariates = c("Z", "L", "B"),
                    utility = "U")

  ld <- pd$stage_data

  ld[stage == 1, time := 1]
  ld[stage == 2, time := 2]

  ld[stage == 2, Z := d$Z]
  ld[stage == 2, L := d$L]
  ld[stage == 2, B := d$B]

  ## simulating the right censoring time
  ## only depending on the baseline covariate Z:
  C <- c(rexp(n, 1) / exp((-1) * cbind(1, as.numeric(d$Z)) %*% zeta))

  ld[stage == 1, time_c := C]
  ld[stage == 2, time_c := C]

  ld[, delta := time_c >= time]

  ld[delta == FALSE , event := 2]
  ld[delta == FALSE, A := NA]
  ld[delta == FALSE & stage == 2, U := NA]
  ld[delta == FALSE & stage == 2, U_A0 := NA]
  ld[delta == FALSE & stage == 2, U_A1 := NA]

  ld[ , tmp := shift(delta, fill = TRUE), by = list(id)]
  ld <- ld[tmp == TRUE, ]
  ld[ , time := pmin(time, time_c)]
  ld[ , time_c := NULL]
  ld[ , tmp := NULL]
  ld[ , delta := NULL]

  if (type == "interval"){
    ld[, time2 := time]
    ld[, time := shift(time, fill = 0), by = list(id)]
  }

  return(ld)
}

sim_two_stage_right_cens <- function(n = 1e4,
                                     par = c(gamma = 0.5,  beta = 1, zeta = 1),
                                     seed = NULL,
                                     action_model_1 = function(C_1, beta, ...)
                                       stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_1)),
                                     action_model_2 = function(C_2, beta, ...)
                                       stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_2)),
                                     deterministic_rewards = FALSE,
                                     cens_model_1 = function(L, zeta, ...)
                                       stats::rbinom(n = NROW(L), size = 1, prob = lava::expit(zeta * abs(L))),
                                     cens_model_2 = function(L, zeta, ...)
                                       stats::rbinom(n = NROW(L), size = 1, prob = lava::expit(zeta * abs(L))),
                                     cens_model_3 = function(L, zeta, ...)
                                       stats::rbinom(n = NROW(L), size = 1, prob = lava::expit(zeta * abs(L)))) {

  d <- sim_two_stage(n = n)
  d$U <- d$U_1 + d$U_2 + d$U_3

  pd <- policy_data(data = d,
                    action = c("A_1", "A_2"),
                    baseline = c("B", "BB"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = "U")

  ld <- pd$stage_data
  ld[is.na(L), L := d[["L_3"]]]

  ## simulating the discrete right-censoring:
  delta_1 <- rbinom(n = nrow(ld[stage == 1]), size = 1, prob = cens_model_1(ld[stage == 1]$L, zeta = par["zeta"]))
  delta_2 <- rbinom(n = nrow(ld[stage == 2]), size = 1, prob = cens_model_2(ld[stage == 2]$L, zeta = par["zeta"]))
  delta_3 <- rbinom(n = nrow(ld[stage == 3]), size = 1, prob = cens_model_3(ld[stage == 3]$L, zeta = par["zeta"]))

  ## adapting the data to the right-censoring process:
  ld[stage == 1 , delta := delta_1]
  ld[stage == 2 , delta := delta_2]
  ld[stage == 3 , delta := delta_3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  return(ld)
}

test_that("fit_c_functions return the expected output", {

  set.seed(1)
  ld <- sim_single_stage_right_cens(n = 5e2, type = "interval")
  pd <- policy_data(data = ld, type = "long", action = "A", time = "time", time2 = "time2")

  ## single c-model:

  expect_error(
    fcf <- fit_c_functions(
      policy_data = pd,
      c_models = c_cox(formula = ~ Z, time = "time", time_2 = "time_2"),
      full_history = FALSE
    ),
    NA
  )

  ref_model <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z, data = ld)

  expect_equal(
    coef(ref_model),
    coef(fcf$all_stages$c_model$model)
  )

  ## c-model for each stage, state history:
  expect_error(
    fcf <- fit_c_functions(
      policy_data = pd,
      c_models = list(
        c_cox(formula = ~ Z, time = "time", time_2 = "time_2"),
        c_cox(formula = ~ Z, time = "time", time_2 = "time_2")
      ),
      full_history = FALSE
    ),
    NA
  )

  ref_model1 <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z, data = ld[stage == 1, ])
  ref_model2 <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z, data = ld[stage == 2, ])

  expect_equal(
    coef(ref_model1),
    coef(fcf$stage_1$c_model$model)
  )

  expect_equal(
    coef(ref_model2),
    coef(fcf$stage_2$c_model$model)
  )

  ## c-model for each stage, full history:
  expect_error(
    fcf <- fit_c_functions(
      policy_data = pd,
      c_models = list(
        c_cox(formula = ~ Z_1, time = "time", time_2 = "time_2"),
        c_cox(formula = ~ Z_2, time = "time", time_2 = "time_2")
      ),
      full_history = TRUE
    ),
    NA
  )

  ref_model1 <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z, data = ld[stage == 1, ])
  ref_model2 <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z, data = ld[stage == 2, ])

  expect_equal(
    unname(coef(ref_model1)),
    unname(coef(fcf$stage_1$c_model$model))
  )

  expect_equal(
    unname(coef(ref_model2)),
    unname(coef(fcf$stage_2$c_model$model))
  )


})

test_that("fit_c_functions(): g_model's passed as the argument c_models are fitted and predicted correctly for discrete right-censoring:", {

  set.seed(1)
  ld <- sim_two_stage_right_cens(n = 500)
  ld[ , absL := abs(L)]
  pd <- policy_data(data = ld, type = "long")

  ## single c_model:
  ref_model <- glm(I(event != 2) ~ absL, data = ld, family = binomial())
  tmp <- fit_c_functions(policy_data = pd,
                         c_models = g_glm(~ absL),
                         full_history = FALSE)
  expect_equal(
    coef(ref_model),
    coef(tmp$all_stages$c_model$model)
  )

  ref_predict <- predict(ref_model, newdata = ld, type = "response")
  ref_predict <- cbind(1, ref_predict)
  colnames(ref_predict) <- c("surv_time", "surv_time2")
  tmp <- predict(tmp, new_policy_data = pd)[ , 3:4] |> as.matrix()

  expect_equal(
    ref_predict,
    tmp,
    check.attributes = FALSE
  )

  ## multiple g_model's:
  ref_model1 <- glm(I(event != 2) ~ absL, data = ld[stage == 1], family = binomial())
  ref_model2 <- glm(I(event != 2) ~ absL, data = ld[stage == 2], family = binomial())
  ref_model3 <- glm(I(event != 2) ~ absL, data = ld[stage == 3], family = binomial())

  tmp <- fit_c_functions(policy_data = pd,
                         c_models = list(g_glm(~ absL), g_glm(~ absL), g_glm(~ absL)),
                         full_history = FALSE)

  ref_model <- list(ref_model1, ref_model2, ref_model3)

  for (i in seq_along(ref_model)) {
    expect_equal(
      coef(ref_model[[i]]),
      coef(tmp[[i]]$c_model$model)
    )
  }

  for (i in seq_along(ref_model)) {
    ref_predict <- predict(ref_model[[i]], newdata = ld[stage == i], type = "response")
    ref_predict <- cbind(1, ref_predict)
    colnames(ref_predict) <- c("surv_time", "surv_time2")
    pred <- predict(tmp, new_policy_data = pd)[ stage == i, 3:4] |> as.matrix()
    expect_equal(
      ref_predict,
      pred,
      check.attributes = FALSE
    )
  }

})

test_that("fit_c_functions() related errors to fitting the c-model.", {

  set.seed(1)
  ld <- sim_two_stage_right_cens(n = 500)
  ld[ , absL := abs(L)]
  ld$absL[1:10] <- as.numeric(NA)
  pd <- policy_data(data = ld, type = "long")

  expect_error(
    fit_c_functions(policy_data = pd,
                    c_models = list(g_glm(~ absL), g_glm(~ absL), g_glm(~ absL)),
                    full_history = FALSE),
    "Error fitting c_model: NA/NaN/Inf in 'x' when calling 'g_glm' with formula:\n~absL"
  )

})

test_that("fit_c_functions(): returns default right-censoring model for stages where no right-censoring occur,", {

  set.seed(1)
  ld <- sim_two_stage_right_cens(n = 500, cens_model_2 = function(L, ...) rep(1, length(L))) # no right-censoring at stage 2
  ld[ , absL := abs(L)]
  pd <- policy_data(data = ld, type = "long")

  expect_no_error(
    tmp <- fit_c_functions(policy_data = pd,
                           c_models = list(g_glm(~ absL), g_glm(~ absL), g_glm(~ absL)),
                           full_history = FALSE)
  )

  pred <- predict(tmp, new_policy_data = pd)

  ref_model1 <- glm(I(event != 2) ~ absL, data = ld[stage == 1], family = binomial())
  ref_model3 <- glm(I(event != 2) ~ absL, data = ld[stage == 3], family = binomial())

  ref_predict1 <- predict(ref_model1, newdata = ld[stage == 1], type = "response")
  ref_predict3 <- predict(ref_model3, newdata = ld[stage == 3], type = "response")

  ref <- cbind(get_id_stage(pd, event_set = c(0,1,2)))
  ref[, surv_time := 1]
  ref[stage == 1, surv_time2 := ref_predict1]
  ref[stage == 2, surv_time2 := 1]
  ref[stage == 3, surv_time2 := ref_predict3]
  setkey(ref, id, stage)
  attr(ref, "index") <- NULL

  expect_equal(
    pred,
    ref
  )

})

sim_single_stage_right_cens <- function(n = 2e3, zeta = c(0.7, 0.2), type = "right"){

  d <- sim_single_stage(n = n)
  pd <- policy_data(data = d,
                    action = "A",
                    covariates = c("Z", "L", "B"),
                    utility = "U")

  ld <- pd$stage_data

  ld[stage == 1, time := 1]
  ld[stage == 2, time := 2]

  ld[stage == 2, Z := d$Z]
  ld[stage == 2, L := d$L]
  ld[stage == 2, B := d$B]

  ## simulating the right censoring time
  ## only depending on the baseline covariate Z:
  C <- c(rexp(n, 1) / exp((-1) * cbind(1, as.numeric(d$Z)) %*% zeta))

  ld[stage == 1, time_c := C]
  ld[stage == 2, time_c := C]

  ld[, delta := time_c >= time]

  ld[delta == FALSE , event := 2]
  ld[delta == FALSE, A := NA]
  ld[delta == FALSE & stage == 2, U := NA]
  ld[delta == FALSE, U_A0 := 0]
  ld[delta == FALSE, U_A1 := 0]

  ld[ , tmp := shift(delta, fill = TRUE), by = list(id)]
  ld <- ld[tmp == TRUE, ]
  ld[ , time := pmin(time, time_c)]
  ld[ , time_c := NULL]
  ld[ , tmp := NULL]
  ld[ , delta := NULL]

  if (type == "interval"){
    ld[, time2 := time]
    ld[, time := shift(time, fill = 0), by = list(id)]
  }

  return(ld)
}

sim_two_stage_right_cens <- function(n = 1e4,
                                     par = c(gamma = 0.5,  beta = 1, zeta = 1),
                                     seed = NULL,
                                     action_model_1 = function(C_1, beta, ...)
                                       stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_1)),
                                     action_model_2 = function(C_2, beta, ...)
                                       stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_2)),
                                     deterministic_rewards = FALSE,
                                     cens_model = function(L, zeta, ...)
                                       stats::rbinom(n = NROW(L), size = 1, prob = lava::expit(zeta * abs(L)))) {

  d <- sim_two_stage(n = n)
  d$U <- d$U_1 + d$U_2 + d$U_3

  pd <- policy_data(data = d,
                    action = c("A_1", "A_2"),
                    baseline = c("B", "BB"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = "U")

  ld <- pd$stage_data
  ld[is.na(L), L := d[["L_3"]]]

  ## simulating the discrete right-censoring:
  delta <- rbinom(n = nrow(ld), size = 1, prob = cens_model(ld$L, zeta = par["zeta"]))

  ## adapting the data to the right-censoring process:
  ld[ , delta := delta]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  return(ld)
}

test_that("policy_eval checks input under right-censoring", {

  ## right-censoring at every stage:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## policy:
  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ## g-functions:
  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ## c-functions:
  cf <- fit_c_functions(policy_data = pd,
                        c_models = g_glm(~1))

  ## m-function:
  mf <- fit_m_function(policy_data = pd,
                       m_model = q_glm(~x))

  ## missing c_models and c_functions input:
  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      q_models = list(q_degen(var = "x"), q_degen(var = "x")),
      m_model = q_degen(var = "x"),
      g_functions = gf
    ),
    "Right-censoring events \\(event = 2\\) occur in the policy data\\. Please provide either c_functions or c_models\\.$"
  )

  ## incorrect c_functions input:
  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      q_models = list(q_degen(var = "x"), q_degen(var = "x")),
      m_model = q_degen(var = "x"),
      c_functions = gf,
      g_functions = gf
    ),
    "c_functions must be of class 'c_functions'\\."
  )

  ## incorrect c_models input:
  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      q_models = list(q_degen(var = "x"), q_degen(var = "x")),
      m_model = q_degen(var = "x"),
      c_models = q_glm(~1),
      g_functions = gf
    ),
    "c_models must be a single c_model \\(or g_model\\) or a list of K\\+1 c_model's \\(or g_model's\\)\\."
  )

  ## missing m_model and m_function input:
  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      q_models = list(q_degen(var = "x"), q_degen(var = "x")),
      c_functions = cf,
      g_functions = gf
    ),
    "Right-censoring events \\(event = 2\\) occur at stage K\\+1 in the policy data. Please provide either m_function or m_model\\."
  )

})

test_that("policy_eval with target 'value' has the expected output for the fixed two-stage case under right-censoring and single final utility outcome.", {

  ## right-censoring at every stage:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## policy:
  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ## g-functions:
  gf <- fit_g_functions(policy_data = pd,
                  g_models = list(g_glm(~1), g_glm(~1)))
  tmp <- predict(gf, pd)

  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  g3 <- ld[stage == 2 & event == 0, mean(A == "3")]
  g4 <- ld[stage == 2 & event == 0, mean(A == "4")]
  expect_equal(
    c(g1, g2, g3, g4),
    apply(tmp[, 3:6, with = FALSE], 2, function(x) max(unique(x))) |> unname()
  )
  rm(tmp)

  ## c-functions:
  cf <- fit_c_functions(policy_data = pd,
                        c_models = g_glm(~1))
  tmp <- predict(cf, pd)
  c1 <- 1-ld[, list(mean(event == 2))][[1]]
  expect_equal(
    unique(tmp$surv_time),
    1
  )
  expect_equal(
    tmp$surv_time2 |> unique(),
    c1
  )

  Z <- d$x1 +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * (d$x2 - d$x1) +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d$p2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) * (d$x3 - d$x2) +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d$p2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) *
    (d$delta3 == 1) / c1 * (d$y - d$x3)

  ref_pe <- mean(Z)
  ref_IC <-  Z - ref_pe

  ref_pe_or <- mean(d$x1)

  ref_pe_ipw <- mean((d$a1 == d$p1) / 0.5 * (d$m == 0) * (d$a2 == d$p2) / 0.5 *  d$y)

  ##
  ## no cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    m_model = q_degen(var = "x"),
    g_functions = gf,
    c_functions = cf
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  ##
  ## cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    m_model = q_degen(var = "x"),
    g_functions = gf,
    c_functions = cf,
    M = 2
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  ## right-censoring NOT occuring at stage K+1:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rep(1, times = 1e2) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## policy:
  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ## g-functions:
  gf <- fit_g_functions(policy_data = pd,
                  g_models = list(g_glm(~1), g_glm(~1)))
  tmp <- predict(gf, pd)

  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  g3 <- ld[stage == 2 & event == 0, mean(A == "3")]
  g4 <- ld[stage == 2 & event == 0, mean(A == "4")]
  expect_equal(
    c(g1, g2, g3, g4),
    apply(tmp[, 3:6, with = FALSE], 2, function(x) max(unique(x))) |> unname()
  )
  rm(tmp)

  ## c-functions:
  cf <- fit_c_functions(policy_data = pd,
                        c_models = g_glm(~1))
  tmp <- predict(cf, pd)
  c1 <- 1-ld[, list(mean(event == 2))][[1]]
  expect_equal(
    unique(tmp$surv_time),
    1
  )
  expect_equal(
    tmp$surv_time2 |> unique(),
    c1
  )

  Z <- d$x1 +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * (d$x2 - d$x1) +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d$p2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) * (d$y - d$x2)

  ref_pe <- mean(Z)
  ref_IC <-  Z - ref_pe

  ref_pe_or <- mean(d$x1)

  ref_pe_ipw <- mean((d$a1 == d$p1) / 0.5 * (d$m == 0) * (d$a2 == d$p2) / 0.5 *  d$y)

  ##
  ## no cross-fitting
  ##

  mf <- fit_m_function(policy_data = pd, m_model = q_glm(~x))
  expect_null(mf)

  ## m-model input:
  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    m_model = q_degen(var = "x"), # m-values should be ignored
    g_functions = gf,
    c_functions = cf
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  ## no m_model input:
  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    g_functions = gf,
    c_functions = cf
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  ## right-censoring ONLY occuring at stage K+1:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rep(1, times = 1e2) # stage 1 non-missing indicator
  delta2 <- rep(1, times = 1e2) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## policy:
  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ## g-functions:
  gf <- fit_g_functions(policy_data = pd,
                  g_models = list(g_glm(~1), g_glm(~1)))
  tmp <- predict(gf, pd)

  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  g3 <- ld[stage == 2 & event == 0, mean(A == "3")]
  g4 <- ld[stage == 2 & event == 0, mean(A == "4")]
  expect_equal(
    c(g1, g2, g3, g4),
    apply(tmp[, 3:6, with = FALSE], 2, function(x) max(unique(x))) |> unname()
  )
  rm(tmp)

  ## c-functions:
  cf <- fit_c_functions(policy_data = pd,
                        c_models = g_glm(~1))
  tmp <- predict(cf, pd)
  c1 <- 1-ld[, list(mean(event == 2))][[1]]
  expect_equal(
    unique(tmp$surv_time),
    1
  )
  expect_equal(
    tmp$surv_time2 |> unique(),
    c1
  )

  Z <- d$x1 +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * (d$x2 - d$x1) +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d$p2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) * (d$x3 - d$x2) +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d$p2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) *
    (d$delta3 == 1) / c1 * (d$y - d$x3)

  ref_pe <- mean(Z)
  ref_IC <-  Z - ref_pe

  ref_pe_or <- mean(d$x1)

  ref_pe_ipw <- mean((d$a1 == d$p1) / 0.5 * (d$m == 0) * (d$a2 == d$p2) / 0.5 *  d$y)

  ##
  ## no cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    m_model = q_degen(var = "x"),
    g_functions = gf,
    c_functions = cf
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

})

test_that("policy_eval returns the expected c_functions output when using c-models.", {

  set.seed(1)
  ld <- sim_single_stage_right_cens(n = 5e2, type = "interval")
  pd <- policy_data(data = ld, type = "long", action = "A", time = "time", time2 = "time2")

  ## single c-model and m-model:

  pe <- policy_eval(
    policy_data = pd,
    policy = policy_def(1),
    m_model = q_glm(~.),
    m_full_history = FALSE,
    c_models = c_cox(formula = ~ Z)
  )

  library(mets)
  ref_model <- phreg(Surv(time = time, time2 = time2, event = (event == 2)) ~ Z, data = ld)

  expect_equal(
    coef(ref_model),
    coef(pe$c_functions$all_stages$c_model$model)
  )

})


test_that("policy_eval with target 'subgroup' has the correct output under right-censoring.", {

  ## right-censoring at every stage:
  set.seed(1)
  z <- 1:1e2*1.0
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  z2 <- runif(n = 1e2, min = -1, max = 1)
  d <- data.table(z = z, a = a, y = y, p = p, delta1 = delta1, delta2 = delta2, z2 = z2)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p"),
    utility = c("y")
  )

  ld <- pd$stage_data
  ld[stage == 2, z := z2]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 2, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  p <- policy_def(function(p) p, name = "p")

  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  c1 <- 1-ld[, list(mean(event == 2))][[1]]

  ref_Z <- cbind(
    d$z +
    (d$delta1 == 1) / c1 * (d$a == 1) / g1 * (d$z2 - d$z) +
    (d$delta1 == 1) / c1 * (d$delta2 == 1) / c1 * (d$a == 1) / g1 *(d$y - d$z2),
    d$z +
    (d$delta1 == 1) / c1 * (d$a == 2) / g2 * (d$z2 - d$z) +
    (d$delta1 == 1) / c1 * (d$delta2 == 1) / c1 * (d$a == 2) / g2 *(d$y - d$z2)
  )
  ref_blip <- ref_Z[, 2] - ref_Z[, 1]
  ref_sub <- mean(ref_blip[d$p == 2])
  ref_sub_comp <- mean(ref_blip[d$p == 1])
  ref_IC <- 2 * (d$p == 2) * (ref_blip - ref_sub)
  ref_IC_comp <- 2 * (d$p == 1) * (ref_blip - ref_sub_comp)

  ## g-functions:
  gf <- fit_g_functions(policy_data = pd,
                        g_models = list(g_glm(~1)))

  ## c-functions:
  cf <- fit_c_functions(policy_data = pd,
                        c_models = g_glm(~1))

  ##
  ## no cross-fitting
  ##

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p,
    q_models = polle:::q_degen(var = "z"),
    m_model = polle:::q_degen(var = "z"),
    g_functions = gf,
    c_functions = cf
  )

  expect_equal(
    coef(sub) |> unname(),
    c(ref_sub, ref_sub_comp)
  )

  expect_equal(
    IC(sub),
    cbind(ref_IC, ref_IC_comp) |> unname()
  )

  expect_equal(
    names(sub$coef),
    c("E[U(2)-U(1)|d=2]: d=p", "E[U(2)-U(1)|d=1]: d=p")
  )

  ##
  ## cross-fitting
  ##

  ## in each training, the empirical propensity is no longer 0.5
  ## instead a g_model is fitted on the complete data:
  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p,
    q_models = polle:::q_degen(var = "z"),
    m_model = polle:::q_degen(var = "z"),
    g_functions = gf,
    c_functions = cf,
    M = 2
  )

  expect_equal(
    coef(sub) |> unname(),
    c(ref_sub, ref_sub_comp)
  )

  expect_equal(
    IC(sub),
    cbind(ref_IC, ref_IC_comp) |> unname()
  )

  ## right-censoring NOT occuring at stage 2:
  set.seed(1)
  z <- 1:1e2*1.0
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rep(1, times = 1e2) # stage 2 non-missing indicator
  z2 <- runif(n = 1e2, min = -1, max = 1)
  d <- data.table(z = z, a = a, y = y, p = p, delta1 = delta1, delta2 = delta2, z2 = z2)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p"),
    utility = c("y")
  )

  ld <- pd$stage_data
  ld[stage == 2, z := z2]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 2, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  p <- policy_def(function(p) p, name = "p")

  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  c1 <- 1-ld[, list(mean(event == 2))][[1]]

  ref_Z <- cbind(
    d$z +
    (d$delta1 == 1) / c1 * (d$a == 1) / g1 * (d$y - d$z),
    d$z +
    (d$delta1 == 1) / c1 * (d$a == 2) / g2 * (d$y - d$z)
  )
  ref_blip <- ref_Z[, 2] - ref_Z[, 1]
  ref_sub <- mean(ref_blip[d$p == 2])
  ref_sub_comp <- mean(ref_blip[d$p == 1])
  ref_IC <- 2 * (d$p == 2) * (ref_blip - ref_sub)
  ref_IC_comp <- 2 * (d$p == 1) * (ref_blip - ref_sub_comp)

  ## g-functions:
  gf <- fit_g_functions(policy_data = pd,
                        g_models = list(g_glm(~1)))

  ## c-functions:
  cf <- fit_c_functions(policy_data = pd,
                        c_models = g_glm(~1))

  ##
  ## no cross-fitting
  ##

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p,
    q_models = polle:::q_degen(var = "z"),
    m_model = polle:::q_degen(var = "z"),
    g_functions = gf,
    c_functions = cf
  )

  expect_equal(
    coef(sub) |> unname(),
    c(ref_sub, ref_sub_comp)
  )

  expect_equal(
    IC(sub),
    cbind(ref_IC, ref_IC_comp) |> unname()
  )

  expect_equal(
    names(sub$coef),
    c("E[U(2)-U(1)|d=2]: d=p", "E[U(2)-U(1)|d=1]: d=p")
  )

  ## no m_model input:
  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p,
    q_models = polle:::q_degen(var = "z"),
    g_functions = gf,
    c_functions = cf
  )

  expect_equal(
    coef(sub) |> unname(),
    c(ref_sub, ref_sub_comp)
  )

  expect_equal(
    IC(sub),
    cbind(ref_IC, ref_IC_comp) |> unname()
  )

  expect_equal(
    names(sub$coef),
    c("E[U(2)-U(1)|d=2]: d=p", "E[U(2)-U(1)|d=1]: d=p")
  )

  ## right-censoring NOT occuring at stage 1:
  set.seed(1)
  z <- 1:1e2*1.0
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  delta1 <- rep(1, times = 1e2)  # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  z2 <- runif(n = 1e2, min = -1, max = 1)
  d <- data.table(z = z, a = a, y = y, p = p, delta1 = delta1, delta2 = delta2, z2 = z2)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p"),
    utility = c("y")
  )

  ld <- pd$stage_data
  ld[stage == 2, z := z2]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 2, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  p <- policy_def(function(p) p, name = "p")

  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  c1 <- 1-ld[, list(mean(event == 2))][[1]]

  ref_Z <- cbind(
    d$z +
    (d$delta1 == 1) / c1 * (d$a == 1) / g1 * (d$z2 - d$z) +
    (d$delta1 == 1) / c1 * (d$delta2 == 1) / c1 * (d$a == 1) / g1 *(d$y - d$z2),
    d$z +
    (d$delta1 == 1) / c1 * (d$a == 2) / g2 * (d$z2 - d$z) +
    (d$delta1 == 1) / c1 * (d$delta2 == 1) / c1 * (d$a == 2) / g2 *(d$y - d$z2)
  )
  ref_blip <- ref_Z[, 2] - ref_Z[, 1]
  ref_sub <- mean(ref_blip[d$p == 2])
  ref_sub_comp <- mean(ref_blip[d$p == 1])
  ref_IC <- 2 * (d$p == 2) * (ref_blip - ref_sub)
  ref_IC_comp <- 2 * (d$p == 1) * (ref_blip - ref_sub_comp)

  ## g-functions:
  gf <- fit_g_functions(policy_data = pd,
                        g_models = list(g_glm(~1)))

  ## c-functions:
  cf <- fit_c_functions(policy_data = pd,
                        c_models = g_glm(~1))

  ##
  ## no cross-fitting
  ##

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p,
    q_models = polle:::q_degen(var = "z"),
    m_model = polle:::q_degen(var = "z"),
    g_functions = gf,
    c_functions = cf
  )

  expect_equal(
    coef(sub) |> unname(),
    c(ref_sub, ref_sub_comp)
  )

  expect_equal(
    IC(sub),
    cbind(ref_IC, ref_IC_comp) |> unname()
  )

  expect_equal(
    names(sub$coef),
    c("E[U(2)-U(1)|d=2]: d=p", "E[U(2)-U(1)|d=1]: d=p")
  )

})

test_that("policy_eval() handles sparse censoring.", {

  ## right-censoring for a single subject at stage 2.
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 1) # stage 1 non-missing indicator
  delta2 <- c(0, rep(1, 99)) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 1) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## policy:
  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ## cross-fitting
  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      q_models = list(q_degen(var = "x"), q_degen(var = "x")),
      g_functions = gf,
      c_models = g_empir(),
      M = 2
    )
  )

  expect_equal(
    pe$c_values$surv_time2 |> unique() |> sort(),
    c(1-1/149, 1)
  )

})

test_that("policy_eval() handles abundant censoring.", {

  ## right-censoring for full but 1 subject at stage 3.
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 1) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 1) # stage 2 non-missing indicator
  delta3 <- c(1, rep(0, 99))  # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## policy:
  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ## no cross-fitting
  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      q_models = list(q_degen(var = "x"), q_degen(var = "x")),
      g_functions = gf,
      c_models = g_empir(),
      m_model = polle:::q_degen(var = "x"),
      M = 1
    )
  )


  ## cross-fitting
  ## plan(sequential, split = TRUE) # browser() will work in future.apply
  expect_error(
    suppressWarnings(
      pe <- policy_eval(
        policy_data = pd,
        policy = p,
        q_models = list(q_degen(var = "x"), q_degen(var = "x")),
        g_functions = gf,
        c_models = g_empir(),
        m_model = polle:::q_degen(var = "x"),
        M = 2
      )
    ),
    "Unable to fit m_model: all utility outcomes are missing"
  )


   ## right-censoring for full but 1 subject at stage 2.
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 1) # stage 1 non-missing indicator
  delta2 <- c(1, rep(0, 99)) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 1)  # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  fpd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

})

test_that("policy_eval with target 'value' has the expected output
for the two-stage case under right-censoring and terminal events.", {

  ## right-censoring at every stage, but no death:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  u1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  u2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  u3 <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  p1 = p1,
                  p2 = p2,
                  u1 = u1,
                  u2 = u2,
                  u3 = u3,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)

  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("u1", "u2", "u3")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## policy:
  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ## g-functions:
  gf <- fit_g_functions(policy_data = pd,
                        g_models = list(g_glm(~1), g_glm(~1)))
  tmp <- predict(gf, pd)

  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  g3 <- ld[stage == 2 & event == 0, mean(A == "3")]
  g4 <- ld[stage == 2 & event == 0, mean(A == "4")]
  expect_equal(
    c(g1, g2, g3, g4),
    apply(tmp[, 3:6, with = FALSE], 2, function(x) max(unique(x))) |> unname()
  )
  rm(tmp)

  ## c-functions:
  cf <- fit_c_functions(policy_data = pd,
                        c_models = g_glm(~1))
  tmp <- predict(cf, pd)
  c1 <- 1-ld[, list(mean(event == 2))][[1]]
  expect_equal(
    unique(tmp$surv_time),
    1
  )
  expect_equal(
    tmp$surv_time2 |> unique(),
    c1
  )

  Z <- (d$x1 + d$u1) +
    (d$delta1 == 1) / c1 *
    (d$a1 == d$p1) / (g1 * (d$a1 == "1") +
                      g2 *
                      (d$a1 == "2")) * ((d$x2 + d$u2) - d$x1) +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") +
                                             g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 *
    (d$a2 == d$p2) / (g3 * (d$a2 == "3") +
                      g4 * (d$a2 == "4")) * (d$x3 - d$x2) +
    (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") +
                                             g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d$p2) / (g3 * (d$a2 == "3") +
                                             g4 * (d$a2 == "4")) *
    (d$delta3 == 1) / c1 * (d$u3 - d$x3)

  ref_pe <- mean(Z)
  ref_IC <-  Z - ref_pe

  ##
  ## no cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    m_model = q_degen(var = "x"),
    g_functions = gf,
    c_functions = cf
  )

  ## stage 1 Q-values
  expect_true(
    all(pe$q_values[stage == 1]$Q_1 == (d$x1 + d$u1)),
    all(pe$q_values[stage == 1]$Q_2 == (d$x1 + d$u1))
  )

  ## stage 2 Q-values
  expect_true(
    all(pe$q_values[stage == 2]$Q_3 == (d$x2 + d$u1 + d$u2)[d$delta1 == 1]),
    all(pe$q_values[stage == 2]$Q_4 == (d$x2 + d$u1 + d$u2)[d$delta1 == 1])
  )

  ## stage 3 Q-values
  expect_true(
    all(pe$m_values$Q == (d$x3 + d$u1 + d$u2)[d$delta1 == 1 & d$delta2 == 1])
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  rm(list = ls())
  ## right-censoring at every stage, and death at stage 2:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  u1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  u3 <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 3 non-missing indicator
  death2 <- rbinom(n = 1e2, size = 1, prob = 0.2) # stage 2 death indicator
  u2 <- runif(n = 1e2, min = -1, max = 1) - death2 * runif(n = 1e2, min = 0, max = 1)
  d <- data.table(
    x1 = x1,
    a1 = a1,
    x2 = x2,
    a2 = a2,
    x3 = x3,
    p1 = p1,
    p2 = p2,
    u1 = u1,
    u2 = u2,
    u3 = u3,
    delta1 = delta1,
    delta2 = delta2,
    delta3 = delta3,
    death2 = death2
  )

  pd <- policy_data(
      data = d,
      action = c("a1", "a2"),
      covariates = list(
          x = c("x1", "x2"),
          p = c("p1", "p2")
      ),
      utility = c("u1", "u2", "u3")
  )

  ## policy:
  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta := delta3]

  ld[, death := 0]
  ld[stage == 2, death := death2]

  ld[, delta := cumprod(delta), by = id]
  ld[, nondeath := cumprod(!death), by = id]

  ld[, tmp := cumsum(delta == 0), by = id]
  ld[, tmp2 := cumsum(nondeath == 0), by = id]

  ld <- ld[tmp %in% c(0, 1)]
  ld <- ld[tmp2 %in% c(0, 1)]

  ld[, tmp := NULL]
  ld[, tmp2 := NULL]

  ld[delta == 0, event := 2]
  ld[death == 1, event := 1] # death before censoring

  ld[, delta := NULL]
  ld[, death := NULL]
  ld[, nondeath := NULL]

  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]
  ld[event == 1, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## g-values
  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  g3 <- ld[stage == 2 & event == 0, mean(A == "3")]
  g4 <- ld[stage == 2 & event == 0, mean(A == "4")]

  ## c-values:
  c1 <- 1-ld[, list(mean(event == 2))][[1]]

  Z <- (
    ## Q1
    (d$x1 + d$u1) +
    ## stage 1 dr score
    (d$death2 == 0) * (d$delta1 == 1) / c1 * (d$a1 == d$p1) /
    (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * ((d$x2 + d$u2) - d$x1) +
    (d$death2 == 1) * (d$delta1 == 1) / c1 * (d$a1 == d$p1) /
    (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * (d$u2 - d$x1) +
    ## stage 2 dr score
    (d$death2 == 0) * (d$delta1 == 1) / c1 * (d$a1 == d$p1) /
    (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 *
    (d$a2 == d$p2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) * (d$x3 - d$x2) +
    ## stage 3 dr score
    (d$death2 == 0) * (d$delta1 == 1) / c1 * (d$a1 == d$p1) / (g1 * (d$a1 == "1") +
                                                               g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d$p2) / (g3 * (d$a2 == "3") +
                                             g4 * (d$a2 == "4")) *
    (d$delta3 == 1) / c1 * (d$u3 - d$x3)
  )

  ref_pe <- mean(Z)
  ref_IC <-  Z - ref_pe

  ##
  ## no cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    m_model = q_degen(var = "x"),
    g_models =  list(g_glm(~1), g_glm(~1)),
    c_models = g_glm(~1)
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
      IC(pe),
      matrix(ref_IC)
  )

  rm(list =ls())
  ## right-censoring at every stage, and death at stage 1 and 2:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  u1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  u3 <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 3 non-missing indicator
  death1 <- rbinom(n = 1e2, size = 1, prob = 0.1) # stage 1 death indicator
  death2 <- rbinom(n = 1e2, size = 1, prob = 0.2) # stage 2 death indicator
  u2 <- runif(n = 1e2, min = -1, max = 1) - death2 * runif(n = 1e2, min = 0, max = 1)
  d <- data.table(
    x1 = x1,
    a1 = a1,
    x2 = x2,
    a2 = a2,
    x3 = x3,
    p1 = p1,
    p2 = p2,
    u1 = u1,
    u2 = u2,
    u3 = u3,
    delta1 = delta1,
    delta2 = delta2,
    delta3 = delta3,
    death1 = death1,
    death2 = death2
  )

  pd <- policy_data(
      data = d,
      action = c("a1", "a2"),
      covariates = list(
          x = c("x1", "x2"),
          p = c("p1", "p2")
      ),
      utility = c("u1", "u2", "u3")
  )

  ## policy:
  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta := delta3]

  ld[, death := 0]
  ld[stage == 1, death := death1]
  ld[stage == 2, death := death2]

  ld[, delta := cumprod(delta), by = id]
  ld[, nondeath := cumprod(!death), by = id]

  ld[, tmp := cumsum(delta == 0), by = id]
  ld[, tmp2 := cumsum(nondeath == 0), by = id]

  ld <- ld[tmp %in% c(0, 1)]
  ld <- ld[tmp2 %in% c(0, 1)]

  ld[, tmp := NULL]
  ld[, tmp2 := NULL]

  ld[delta == 0, event := 2]
  ld[death == 1, event := 1] # death before censoring

  ld[, delta := NULL]
  ld[, death := NULL]
  ld[, nondeath := NULL]

  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]
  ld[event == 1, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## g-values
  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  g3 <- ld[stage == 2 & event == 0, mean(A == "3")]
  g4 <- ld[stage == 2 & event == 0, mean(A == "4")]

  ## c-values:
  c1 <- 1-ld[, list(mean(event == 2))][[1]]

  ## Q1
  Z <- (d$death1 == 0) * (d$x1 + d$u1)
  Z <- Z + (d$death1 == 1) * (d$u1)
  ## stage 1 dr score
  Z <- Z + (d$death1 == 0) * (d$death2 == 0) * (d$delta1 == 1) / c1 * (d$a1 == d$p1) /
    (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * ((d$x2 + d$u2) - d$x1)
  Z <- Z + (d$death1 == 0) * (d$death2 == 1) * (d$delta1 == 1) / c1 * (d$a1 == d$p1) /
    (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * (d$u2 - d$x1)
  ## stage 2 dr score
  Z <- Z + (d$death1 == 0) * (d$death2 == 0) * (d$delta1 == 1) / c1 *
    (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 *
    (d$a2 == d$p2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) * (d$x3 - d$x2)
    ## stage 3 dr score
  Z <- Z + (d$death1 == 0) * (d$death2 == 0) *
    (d$delta1 == 1) / c1 *
    (d$a1 == d$p1) / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 *
    (d$a2 == d$p2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) *
    (d$delta3 == 1) / c1 * (d$u3 - d$x3)

  ref_pe <- mean(Z)
  ref_IC <-  Z - ref_pe

  ##
  ## no cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    m_model = q_degen(var = "x"),
    g_models =  list(g_glm(~1), g_glm(~1)),
    c_models = g_glm(~1)
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
      IC(pe),
      matrix(ref_IC)
  )
})

test_that("get_policy.drql returns a policy", {
  d <- sim_single_stage(200, seed=1)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  pl <- policy_learn(type = "drql",
                     control = control_drql())
  expect_error({
    p <- get_policy(pl(pd, q_models = q_glm(), g_models = g_glm()))
  },
  NA
  )
  expect_true(
    inherits(p, what = "policy")
  )
  expect_true(
    is.data.table(p(pd))
  )

  pl <- policy_learn(type = "drql",
                     control = control_drql(
                       qv_models = q_sl()
                     ))
  expect_error({
    p <- get_policy(pl(pd, q_models = q_glm(), g_models = g_glm()))
  },
  NA
  )
  expect_true(
    inherits(p, what = "policy")
  )
  expect_true(
    is.data.table(p(pd))
  )

})

test_that("policy_learn with type = 'drql' checks input", {
  d <- sim_two_stage(200, seed=1)

  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))
  qv <- policy_learn(type = "drql",
                     control = control_drql())

  gfun <- fit_g_functions(pd, g_models = g_glm(), full_history = FALSE)
  gfun2 <- fit_g_functions(pd, g_models = list(g_glm(), g_glm()), full_history = FALSE)

  expect_error(
    qv(policy_data = pd, q_models = q_glm()),
    "Please provide g_models."
  )
  expect_error(
    qv(policy_data = pd, q_models = q_glm(), g_functions = g_glm()),
    "g-functions must be of class 'g_functions'."
  )
  expect_error(
    qv(policy_data = pd, q_models = list(), g_models = g_glm()),
    "q_models must either be a list of length K or a single Q-model."
  )
  expect_error(
    qv(policy_data = pd,
       q_models = list(q_glm(), q_glm(),q_glm()),
       g_models = g_glm()),
    "q_models must either be a list of length K or a single Q-model."
  )

})

test_that("policy_learn with type drql works as intended", {
  d <- sim_two_stage(200, seed = 1)

  pd <- policy_data(d,
    action = c("A_1", "A_2"),
    baseline = c("BB", "B"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )

  qv <- policy_learn(
    type = "drql",
    control = control_drql()
  )
  gfun <- fit_g_functions(pd, g_models = g_glm(), full_history = FALSE)
  gfun2 <- fit_g_functions(
    pd,
    g_models = list(g_glm(), g_glm()),
    full_history = FALSE
  )

  expect_error(
    qv(policy_data = pd, g_models = g_glm(), q_models = q_glm()),
    NA
  )
  expect_error(
    qv(policy_data = pd, g_functions = gfun, q_models = q_glm()),
    NA
  )
  expect_error(
    qv(policy_data = pd, g_functions = gfun2, q_models = q_glm()),
    NA
  )
  expect_error(
    policy_eval(policy_data = pd, policy_learn = qv),
    NA
  )
  expect_error(
    policy_eval(policy_data = pd, policy_learn = qv, g_functions = gfun),
    NA
  )
  expect_error(
    policy_eval(policy_data = pd, policy_learn = qv, g_functions = gfun2),
    NA
  )

  qv <- policy_learn(
    type = "drql",
    control = control_drql(qv_models = q_glm(formula = Y ~ .))
  )
  expect_error(
    policy_eval(policy_data = pd, policy_learn = qv),
    NA
  )

  qv <- policy_learn(
    type = "drql",
    control = control_drql(qv_models = q_glm(formula = Y ~ BB))
  )

  expect_error(
    policy_eval(policy_data = pd, policy_learn = qv),
    NA
  )

  qv <- policy_learn(
    type = "drql",
    control = control_drql(qv_models = q_glm(formula = ~X))
  )

  expect_error(
    policy_eval(policy_data = pd, policy_learn = qv),
    "Error in qv_model: variable 'X' is not found in data when calling 'q_glm' with formula: ~X"
  )

  qv <- policy_learn(
    type = "drql",
    control = control_drql(qv_models = q_glm(formula = Y ~ X))
  )
  expect_error(
    policy_eval(policy_data = pd, policy_learn = qv),
    "Error in qv_model: variable 'X' is not found in data when calling 'q_glm' with formula: ~X"
  )

  # q_glm formula default is A * (.), and A is not used when fitting the
  # QV-model.
  qv <- policy_learn(
    type = "drql",
    control = control_drql(qv_models = q_glm())
  )

  expect_error(
    policy_eval(policy_data = pd, policy_learn = qv),
    "Error in qv_model: variable 'A' is not found in data when calling 'q_glm' with formula: ~A \\+ L \\+ C \\+ BB \\+ B \\+ A:L \\+ A:C \\+ A:BB \\+ A:B"
  )

})

test_that("policy_learn with type drql handles varying action sets", {
  d <- sim_two_stage_multi_actions(n = 1e2)
  pd <- policy_data(data = d,
                    action = c("A_1", "A_2"),
                    baseline = c("B", "BB"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  pl <- policy_learn(type = "drql",
                     control = control_drql())
  expect_error(
    po <- pl(pd, q_models = q_glm(), g_models = list(g_glm(), g_rf())),
    NA
  )
  expect_error(
    pa <- get_policy(po)(pd),
    NA
  )

  expect_error(
    pf1 <- get_policy_functions(po, stage = 1),
    NA
  )
  H1 <- get_history(pd, stage = 1, full_history = FALSE)$H[, -c("id", "stage"), with = FALSE]
  expect_error(
    pa1 <- pf1(H1),
    NA
  )
  expect_equal(
    pa[stage == 1,]$d,
    pa1
  )

  expect_error(
    pf2 <- get_policy_functions(po, stage = 2),
    NA
  )
  H2 <- get_history(pd, stage = 2, full_history = FALSE)$H[, -c("id", "stage"), with = FALSE]
  expect_error(
    pa2 <- pf2(H2),
    NA
  )
  expect_equal(
    pa[stage == 2,]$d,
    pa2
  )

  expect_error(
    pe <- policy_eval(pd, policy = get_policy(po),
                      g_models = list(g_glm(), g_rf())),
    NA
  )

  # realistic policy:
  pl <- policy_learn(type = "drql",
                     control = control_drql(),
                     alpha = 0.4)
  expect_error(
    po <- pl(pd, q_models = q_glm(), g_models = list(g_glm(), g_rf())),
    "Cases with no realistic actions occur. Consider resetting the alpha level."
  )

  # realistic policy:
  pl <- policy_learn(type = "drql",
                     control = control_drql(),
                     alpha = 0.2)
  expect_error(
    po <- pl(pd, q_models = q_glm(), g_models = list(g_glm(), g_rf())),
    NA
  )
  expect_error(
    pa <- get_policy(po)(pd),
    NA
  )

  expect_error(
    pf1 <- get_policy_functions(po, stage = 1),
    NA
  )
  H1 <- get_history(pd, stage = 1, full_history = FALSE)$H[, -c("id", "stage"), with = FALSE]
  expect_error(
    pa1 <- pf1(H1),
    NA
  )
  expect_equal(
    pa[stage == 1,]$d,
    pa1
  )

  expect_error(
    pf2 <- get_policy_functions(po, stage = 2),
    NA
  )
  H2 <- get_history(pd, stage = 2, full_history = FALSE)$H[, -c("id", "stage"), with = FALSE]
  expect_error(
    pa2 <- pf2(H2),
    NA
  )
  expect_equal(
    pa[stage == 2,]$d,
    pa2
  )

  expect_error(
    pe <- policy_eval(pd, policy = get_policy(po),
                      g_models = list(g_glm(), g_rf())),
    NA
  )

  # cross-fitting
  # realistic policy:
  pl <- policy_learn(type = "drql",
                     control = control_drql(),
                     alpha = 0.2,
                     L = 2)
  expect_error(
    po <- pl(pd, q_models = q_glm(), g_models = list(g_glm(), g_rf())),
    NA
  )
  expect_error(
    pa <- get_policy(po)(pd),
    NA
  )

  expect_error(
    pf1 <- get_policy_functions(po, stage = 1),
    NA
  )
  H1 <- get_history(pd, stage = 1, full_history = FALSE)$H[, -c("id", "stage"), with = FALSE]
  expect_error(
    pa1 <- pf1(H1),
    NA
  )
  expect_equal(
    pa[stage == 1,]$d,
    pa1
  )

  expect_error(
    pf2 <- get_policy_functions(po, stage = 2),
    NA
  )
  H2 <- get_history(pd, stage = 2, full_history = FALSE)$H[, -c("id", "stage"), with = FALSE]
  expect_error(
    pa2 <- pf2(H2),
    NA
  )
  expect_equal(
    pa[stage == 2,]$d,
    pa2
  )

  expect_error(
    pe <- policy_eval(pd, policy = get_policy(po),
                      g_models = list(g_glm(), g_rf()),
                      M = 2),
    NA
  )

})

test_that("policy_learn with type = 'drql' uses the cross_fit_g_models argument", {
  d <- sim_two_stage_multi_actions(n = 1e2)
  pd <- policy_data(
    data = d,
    action = c("A_1", "A_2"),
    baseline = c("B", "BB"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )

  pl <- policy_learn(
    type = "drql",
    cross_fit_g_models = FALSE,
    L = 2,
    alpha = 0.1,
    control = control_drql()
  )

  gfun <- fit_g_functions(pd,
    list(g_empir(), g_empir()),
    full_history = FALSE
  )

  po <- pl(pd,
    g_models = list(g_empir(), g_empir()),
    q_models = q_glm()
  )

  expect_equal(
    po$g_functions,
    gfun
  )
})



test_that("policy_learn with type = 'drql' saves cross-fitted models.", {
  d <- sim_two_stage(n = 1e2)
  pd <- policy_data(
    data = d,
    action = c("A_1", "A_2"),
    baseline = c("B", "BB"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )

  pl <- policy_learn(
    type = "drql",
    cross_fit_g_models = FALSE,
    L = 2,
    alpha = 0.1,
    save_cross_fit_models = FALSE,
    control = control_drql()
  )

  po <- pl(pd,
    g_models = list(g_empir(), g_empir()),
    q_models = q_glm()
  )

  expect_true(
    is.null(unlist(po$q_functions_cf))
  )

  pl <- policy_learn(
    type = "drql",
    cross_fit_g_models = FALSE,
    L = 2,
    alpha = 0.1,
    save_cross_fit_models = TRUE,
    control = control_drql()
  )

  po <- pl(pd,
    g_models = list(g_empir(), g_empir()),
    q_models = q_glm()
  )

  expect_true(
    !is.null(unlist(po$q_functions_cf))
  )
})


test_that("policy_learn with type drql handles multiple stages with varying stage action sets", {

  d <- sim_multi_stage(300, seed = 1)
  ## changing administrative right-censoring events to terminal events:
  d$stage_data[event == 2, U := 10]
  d$stage_data[event == 2, event := 1]

  # constructing policy_data object:
  pd <- policy_data(data = d$stage_data,
                    baseline_data = d$baseline_data,
                    type = "long",
                    id = "id",
                    stage = "stage",
                    event = "event",
                    action = "A",
                    utility = "U")

  pd3 <- partial(pd, 3)

  pl <- policy_learn(type = "drql",
                     control = control_drql(
                       qv_models = q_sl(cvControl = SuperLearner.CV.control(V = 10L))
                     ),
                     alpha = 0.1,
                     L = 2)

  set.seed(1)
  suppressWarnings(
    pe <- policy_eval(policy_data = pd3,
                      policy_learn = pl,
                      g_models = list(g_glm(),g_glm(), g_glm()))
  )

  # renaming actions at each stage:
  d2_stage_data <- copy(d$stage_data)
  d2_stage_data[stage == 1, A := ifelse(A == 1, "t2", "t1")]
  d2_stage_data[stage == 2, A := ifelse(A == 1, "t4", "t3")]
  d2_stage_data[stage == 3, A := ifelse(A == 1, "t6", "t5")]
  d2_stage_data[, U_At1 := 0]
  d2_stage_data[stage == 1, U_At1 := U_A0]
  d2_stage_data[, U_At2 := 0]
  d2_stage_data[stage == 1, U_At2 := U_A1]
  d2_stage_data[, U_At3 := 0]
  d2_stage_data[stage == 2, U_At3 := U_A0]
  d2_stage_data[, U_At4 := 0]
  d2_stage_data[stage == 2, U_At4 := U_A1]
  d2_stage_data[, U_At5 := 0]
  d2_stage_data[stage == 3, U_At5 := U_A0]
  d2_stage_data[, U_At6 := 0]
  d2_stage_data[stage == 3, U_At6 := U_A1]
  d2_stage_data[, U_A0 := NULL]
  d2_stage_data[, U_A1 := NULL]

  pd3_2 <- policy_data(data = d2_stage_data,
                       baseline_data = d$baseline_data,
                       type = "long",
                       id = "id",
                       stage = "stage",
                       event = "event",
                       action = "A",
                       utility = "U")
  pd3_2 <- partial(pd3_2, K = 3)

  set.seed(1)
  suppressWarnings(
    pe2 <- policy_eval(policy_data = pd3_2,
                       policy_learn = pl,
                       g_models = list(g_glm(),g_glm(), g_glm()))
  )
  expect_equal(
    pe$value_estimate,
    pe2$value_estimate
  )
  expect_equal(
    pe$IC,
    pe2$IC
  )
})

test_that("policy_eval with target = 'subgroup' checks inputs.", {
    d <- sim_single_stage(1e2, seed = 1)
    pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")
    p <- policy_def(1)

    expect_error(
        policy_eval(
            policy_data = pd,
            policy = p,
            target = "test"
        ),
        "target must be either 'value' or 'subgroup'."
    )

    d <- sim_single_stage_multi_actions(1e2, seed = 1)
    pd <- policy_data(d, action = "a", covariates = c("z"), utility = "u")
    p <- policy_def(1)

    expect_error(
        policy_eval(
            policy_data = pd,
            policy = p,
            g_models = g_empir(),
            target = "subgroup"
        ),
        "subgroup average treatment effect evaluation is not implemented for more than two actions."
    )

    d <- sim_two_stage(1e2, seed = 1)
    pd <- policy_data(
        data = d,
        action = c("A_1", "A_2"),
        covariates = list(L = c("L_1", "L_2")),
        utility = "U_3"
    )
    p <- policy_def(1, reuse = TRUE)

    expect_error(
        policy_eval(
            policy_data = pd,
            policy = p,
            g_models = g_empir(),
            target = "subgroup"
        ),
        "subgroup average treatment effect evaluation is not implemeted for multiple stages."
    )
})


test_that("policy_eval with target 'subgroup' agrees with targeted::cate.", {
    n <- 1e3
    Z <- rnorm(n = n)
    A <- rbinom(size = 1, n = n, prob = 0.5)
    U <- rnorm(mean = Z + Z * A, n = n)
    d <- data.table(Z = Z, A = A, U = U)
    rm(Z, A, U)
    pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")
    p <- policy_def(function(Z) (Z > 0) * 1)

    ##
    ## no cross-fitting:
    ##

    pe <- policy_eval(
        policy_data = pd,
        policy = p,
        g_models = g_glm(~1),
        q_models = q_glm(~ A * Z),
        target = "sub_effect",
        M = 1
    )

    ## implementation from the targeted package:
    d$d <- p(pd)$d

    ca <- targeted::cate(
        cate.model = ~ factor(d) - 1,
        response.model = U ~ A * Z,
        propensity.model = A ~ 1,
        data = d,
        second.order = FALSE,
        nfolds = 1,
    )

    expect_equal(
        unname(coef(pe)),
        unname(coef(ca)[c("factor(d)1", "factor(d)0")])
    )

    expect_equal(
        IC(pe) |> unname(),
      IC(ca)[, c("factor(d)1", "factor(d)0"), drop=FALSE] |>
      unname()
    )

    ##
    ## cross-fitting: pooled estimate and variance
    ##

    set.seed(1)
    pe <- policy_eval(
        policy_data = pd,
        policy = p,
        g_models = g_glm(~1),
        q_models = q_glm(~ A * Z),
        target = "sub_effect",
        cross_fit_type = "pooled",
        variance_type = "pooled",
        M = 2
    )

    set.seed(1)
    ca <- targeted::cate(
        cate.model = ~ factor(d) - 1,
        response.model = U ~ A * Z,
        propensity.model = A ~ 1,
        second.order = FALSE,
        data = d,
        nfolds = 2,
    )

    expect_equal(
        unname(coef(pe)),
        unname(coef(ca)[c("factor(d)1", "factor(d)0")])
    )

    expect_equal(
      IC(pe),
      IC(ca)[, c("factor(d)1", "factor(d)0")] |>
      unname()
    )

})

test_that("policy_eval with target 'sub_effect' has the correct outputs: test1.", {
  test_output <- function(pe) {
    ## value_estimate
    expect_true(
      !is.null(coef(pe)) && is.numeric(coef(pe))
    )

    ## IC in group d=1 should be zero for Z < 0
    expect_true(
      all(IC(pe)[d$Z <= 0, 1] == 0)
    )
    ## ## ... and for the other subgroup:
    ## expect_true(
    ##     all(IC(pe)[d$Z >0, 2] == 0)
    ## )

    ## id
    expect_true(
      all(pe$id == 1:1e2)
    )

    ## type
    expect_equal(
      pe$type,
      "dr"
    )

    ## target
    expect_equal(
      pe$target,
      "subgroup"
    )

    ## name
    expect_equal(
      pe$name,
      names(pe$coef)
    )
    expect_equal(
      pe$name,
      c("E[U(1)-U(0)|d=1]: d=test", "E[U(1)-U(0)|d=0]: d=test")
    )
  }

  d <- sim_single_stage(1e2, seed = 1)
  d$A <- c(rep(0, 50), rep(1, 50))
  pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")
  p <- policy_def(function(Z) (Z > 0) * 1, name = "test")

  ## no cross-fitting
  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      target = "subgroup"
    )
  )
  test_output(pe)

  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      target = "subgroup",
      name = c("group1", "group2")
    )
  )
  expect_equal(
    names(coef(pe)),
    c("group1: d=test", "group2: d=test")
  )

  ## cross-fitting: stacked estimator
  set.seed(1)
  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      target = "subgroup",
      M = 2,
      cross_fit_type = "stacked",
      variance_type = "stacked"
    )
  )
  test_output(pe)

  ## cross-fitting: pooled estimator
  set.seed(1)
  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      target = "sub_effect",
      M = 2,
      cross_fit_type = "pooled",
      variance_type = "pooled"
    )
  )
  test_output(pe)

  ## cross-fitting: pooled estimator, complete variance estimate
  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      target = "subgroup",
      M = 2,
      cross_fit_type = "pooled",
      variance_type = "complete"
    )
  )
  test_output(pe)
})

test_that("policy_eval with target 'subgroup' has the correct outputs: test2.", {
    z <- 1:1e2
    a <- c(rep(1, 50), rep(2, 50))
    y <- a * 2
    p <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
    d <- data.table(z = z, a = a, y = y, p = p)
    rm(a, z, y)
    pd <- policy_data(
        data = d,
        action = "a",
        covariates = c("z", "p"),
        utility = c("y")
    )

    ## his <- get_history(pd, stage = 1)
    ## qfun <- fit_Q_function(history = his, Q = d$y, q_degen(var = "z"))
    ## predict.Q_function(qfun, new_history = his)

    p <- policy_def(function(p) p, name = "p")

    ref_Z <- cbind(
        (d$a == 1) / 0.5 * (d$y - d$z) + d$z,
        (d$a == 2) / 0.5 * (d$y - d$z) + d$z
    )
    ref_blip <- ref_Z[, 2] - ref_Z[, 1]
    ref_sub <- mean(ref_blip[d$p == 2])
    ref_sub_comp <- mean(ref_blip[d$p == 1])
    ref_IC <- 2 * (d$p == 2) * (ref_blip - ref_sub)
    ref_IC_comp <- 2 * (d$p == 1) * (ref_blip - ref_sub_comp)

    ##
    ## no cross-fitting
    ##

    sub <- policy_eval(
        target = "subgroup",
        policy_data = pd,
        policy = p,
        q_models = polle:::q_degen(var = "z"),
        g_models = g_glm(~1)
    )

    expect_equal(
        coef(sub) |> unname(),
        c(ref_sub, ref_sub_comp)
    )

    expect_equal(
        IC(sub),
        cbind(ref_IC, ref_IC_comp) |> unname()
    )

    expect_equal(
      names(sub$coef),
      c("E[U(2)-U(1)|d=2]: d=p", "E[U(2)-U(1)|d=1]: d=p")
    )

    ##
    ## cross-fitting
    ##

    ## in each training, the empirical propensity is no longer 0.5
    ## instead a g_model is fitted on the complete data:
    gf <- fit_g_functions(pd, g_models = g_glm(~1))

    sub <- policy_eval(
        target = "sub_effect",
        policy_data = pd,
        policy = p,
        q_models = polle:::q_degen(var = "z"),
        g_functions = gf,
        M = 2,
    )

    expect_equal(
        coef(sub) |> unname(),
        c(ref_sub, ref_sub_comp)
    )

    expect_equal(
        IC(sub),
        cbind(ref_IC, ref_IC_comp) |> unname()
    )
})

test_that("policy_eval with target 'subgroup' returns NA when no subjects are in the subgroup.", {

  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p1 <- c(rep(1, 50), rep(2, 50))
  p2 <- c(rep(1, 100))

  d <- data.table(z = z, a = a, y = y, p1 = p1, p2 = p2)
  rm(a, z, y, p1, p2)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p1", "p2"),
    utility = c("y")
  )

  ref_Z <- cbind(
  (d$a == 1) / 0.5 * (d$y - d$z) + d$z,
  (d$a == 2) / 0.5 * (d$y - d$z) + d$z
  )
  ref_blip <- ref_Z[, 2] - ref_Z[, 1]


  p <- policy_def(1)

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p,
    q_models = polle:::q_degen(var = "z"),
    g_models = g_glm(~1)
  )

  expect_equal(
    coef(sub) |> unname(),
    c(as.numeric(NA), mean(ref_blip))
  )

  expect_equal(
    unname(IC(sub)[,1,drop=FALSE]),
    cbind(rep(as.numeric(NA), 1e2))
  )

  expect_no_error(
    tmp <- capture.output(print(sub))
  )

  ## cross-fitting

  ## in each training, the empirical propensity is no longer 0.5
  ## instead a g_model is fitted on the complete data:
  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  set.seed(1)
  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p,
    q_models = polle:::q_degen(var = "z"),
    g_functions = gf,
    M = 2
  )

  expect_equal(
    coef(sub) |> unname(),
    c(as.numeric(NA), mean(ref_blip))
  )

  expect_equal(
    unname(IC(sub)[,1,drop=FALSE]),
    cbind(rep(as.numeric(NA), 1e2))
  )

  expect_no_error(
    tmp <- capture.output(print(sub))
  )

  set.seed(1)
  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p,
    q_models = polle:::q_degen(var = "z"),
    g_functions = gf,
    M = 2,
    cross_fit_type = "stacked",
    variance_type = "stacked"
  )

  expect_equal(
    coef(sub) |> unname(),
    c(as.numeric(NA), mean(ref_blip))
  )

  expect_equal(
    unname(IC(sub)[,1,drop=FALSE]),
    cbind(rep(as.numeric(NA), 1e2))
  )

  expect_no_error(
    tmp <- capture.output(print(sub))
  )

})

test_that("policy_eval with target 'subgroup' works with policy_learning with multiple thresholds.", {

  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p1 <- c(rep(1, 50), rep(2, 50))
  p2 <- c(rep(1, 100))

  d <- data.table(z = z, a = a, y = y, p1 = p1, p2 = p2)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p1", "p2"),
    utility = c("y")
  )

  ref_Z <- cbind(
  (d$a == 1) / 0.5 * (d$y - d$z) + d$z,
  (d$a == 2) / 0.5 * (d$y - d$z) + d$z
  )
  ref_blip <- ref_Z[, 2] - ref_Z[, 1]

  ref_sub2_eta50 <- mean(ref_blip[d$p1 == 2])
  ref_IC2_eta50 <- 2 * (d$p1 == 2) * (ref_blip - ref_sub2_eta50)
  ref_sub1_eta50 <- mean(ref_blip[d$p1 == 1])
  ref_IC1_eta50 <- 2 * (d$p1 == 1) * (ref_blip - ref_sub1_eta50)

  ref_sub1_eta101 <- mean(ref_blip[d$p2 == 1])
  ref_IC1_eta101 <- (d$p2 == 1) * (ref_blip - ref_sub1_eta101)

  pl <- policy_learn(
    type = "blip",
    threshold = c(50, 101),
    control = control_blip(blip_models = polle:::q_degen(var = "z"))
  )

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy_learn = pl,
    q_models = polle:::q_degen(var = "z"),
    g_models = g_glm(~1)
  )

  expect_no_error(
    tmp <- capture.output(print(sub))
  )

  expect_equal(
    sub$name,
    names(sub$coef)
  )

  expect_equal(
    sub$name,
    c("E[U(2)-U(1)|d=2]: d=blip(eta=50)", "E[U(2)-U(1)|d=1]: d=blip(eta=50)",
      "E[U(2)-U(1)|d=2]: d=blip(eta=101)", "E[U(2)-U(1)|d=1]: d=blip(eta=101)")
  )

  expect_equal(
    coef(sub) |> unname(),
    c(ref_sub2_eta50, ref_sub1_eta50, as.numeric(NA), ref_sub1_eta101)
  )

  expect_equal(
    IC(sub),
    cbind(ref_IC2_eta50, ref_IC1_eta50, rep(as.numeric(NA), 1e2), ref_IC1_eta101) |> unname()
  )

  expect_equal(
    sub$subgroup_indicator,
    cbind(p1 == 2, p1 == 1, p2 == 2, p2 == 1)
  )

  ## cross-fitting:

  ## in each training, the empirical propensity is no longer 0.5
  ## instead a g_model is fitted on the complete data:
  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  set.seed(1)
  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy_learn = pl,
    q_models = polle:::q_degen(var = "z"),
    g_functions = gf,
    cross_fit_type = "pooled",
    variance_type = "pooled",
    M = 2)

  expect_no_error(
    tmp <- capture.output(print(sub))
  )

  expect_equal(
    sub$name,
    names(sub$coef)
  )

  expect_equal(
    sub$name,
    c("E[U(2)-U(1)|d=2]: d=blip(eta=50)", "E[U(2)-U(1)|d=1]: d=blip(eta=50)",
      "E[U(2)-U(1)|d=2]: d=blip(eta=101)", "E[U(2)-U(1)|d=1]: d=blip(eta=101)")
  )

  expect_equal(
    coef(sub) |> unname(),
    c(ref_sub2_eta50, ref_sub1_eta50, as.numeric(NA), ref_sub1_eta101)
  )

  expect_equal(
    IC(sub),
    cbind(ref_IC2_eta50, ref_IC1_eta50, rep(as.numeric(NA), 1e2), ref_IC1_eta101) |> unname()
  )

  expect_equal(
    sub$subgroup_indicator,
    cbind(p1 == 2, p1 == 1, p2 == 2, p2 == 1)
  )

  ## stacked:

  set.seed(1)
  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy_learn = pl,
    q_models = polle:::q_degen(var = "z"),
    g_functions = gf,
    M = 2,
    cross_fit_type = "stacked",
    variance_type = "stacked"
  )

  expect_no_error(
    tmp <- capture.output(print(sub))
  )

  expect_equal(
    sub$name,
    names(sub$coef)
  )

  expect_equal(
    sub$name,
    c("E[U(2)-U(1)|d=2]: d=blip(eta=50)", "E[U(2)-U(1)|d=1]: d=blip(eta=50)",
      "E[U(2)-U(1)|d=2]: d=blip(eta=101)", "E[U(2)-U(1)|d=1]: d=blip(eta=101)")
  )

  ref_sub2_eta50_fold1 <- mean(ref_blip[sub$folds[[1]]][d$p1[sub$folds[[1]]] == 2])
  ref_sub2_eta50_fold2 <- mean(ref_blip[sub$folds[[2]]][d$p1[sub$folds[[2]]] == 2])
  ref_sub1_eta50_fold1 <- mean(ref_blip[sub$folds[[1]]][d$p1[sub$folds[[1]]] == 1])
  ref_sub1_eta50_fold2 <- mean(ref_blip[sub$folds[[2]]][d$p1[sub$folds[[2]]] == 1])

  ref_sub2_eta101_fold1 <- mean(ref_blip[sub$folds[[1]]][d$p2[sub$folds[[1]]] == 2])
  ref_sub2_eta101_fold2 <- mean(ref_blip[sub$folds[[2]]][d$p2[sub$folds[[2]]] == 2])
  ref_sub1_eta101_fold1 <- mean(ref_blip[sub$folds[[1]]][d$p2[sub$folds[[1]]] == 1])
  ref_sub1_eta101_fold2 <- mean(ref_blip[sub$folds[[2]]][d$p2[sub$folds[[2]]] == 1])

  expect_equal(
    c(
      mean(c(ref_sub2_eta50_fold1, ref_sub2_eta50_fold2)),
      mean(c(ref_sub1_eta50_fold1, ref_sub1_eta50_fold2)),
      mean(c(ref_sub2_eta101_fold1, ref_sub2_eta101_fold2)),
      mean(c(ref_sub1_eta101_fold1, ref_sub1_eta101_fold2))

    ),
    coef(sub) |> unname()
  )

})

test_that("policy_eval with target 'subgroup' returns NA when the subgroup count is below the minimum.", {

  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p1 <- c(rep(1, 4), rep(2, 96))
  p2 <- c(rep(2, 4), rep(1, 96))
  d <- data.table(z = z, a = a, y = y, p1 = p1, p2 = p2)
  rm(a, z, y, p1, p2)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p1", "p2"),
    utility = c("y")
  )

  p1 <- policy_def(function(p1) p1, name = "p1") # 4 observations in the subgroup d = 1
  p2 <- policy_def(function(p2) p2, name = "p2") # 4 observations in the subgroup d = 2

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p1,
    q_models = polle:::q_degen(var = "z"),
    g_models = g_glm(~1)
  )

  expect_equal(
    coef(sub) |> is.na() |> unname(),
    c(FALSE, FALSE)
  )

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p1,
    q_models = polle:::q_degen(var = "z"),
    g_models = g_glm(~1),
    min_subgroup_size = 5
  )

  expect_equal(
    coef(sub) |> is.na() |> unname(),
    c(FALSE, TRUE)
  )

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p2,
    q_models = polle:::q_degen(var = "z"),
    g_models = g_glm(~1),
    min_subgroup_size = 5
  )

  expect_equal(
    coef(sub) |> is.na() |> unname(),
    c(TRUE, FALSE)
  )

  ## cross-fitting

  ## in each training, the empirical propensity is no longer 0.5
  ## instead a g_model is fitted on the complete data:
  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p2,
    q_models = polle:::q_degen(var = "z"),
    g_functions = gf,
    min_subgroup_size = 5,
    M = 2
  )

  expect_equal(
    coef(sub) |> is.na() |> unname(),
    c(TRUE, FALSE)
  )

  sub <- policy_eval(
    target = "subgroup",
    policy_data = pd,
    policy = p2,
    q_models = polle:::q_degen(var = "z"),
    g_functions = gf,
    min_subgroup_size = 5,
    M = 2,
    cross_fit_type = "stacked"
  )

  expect_equal(
    coef(sub) |> is.na() |> unname(),
    c(TRUE, FALSE)
  )

})

test_that("policy_eval runs with policy_learn using quantile_prob_thres ", {

  d1 <- sim_single_stage(1e3, seed=1)
  pd1 <- policy_data(d1, action = "A", covariates = c("Z"), utility = "U")

  qs <- c(0.25, 0.5, 0.75)

  learn1 <- policy_learn(type = "blip",
                         control = control_blip(
                           blip_models = q_glm(~ .),
                           quantile_prob_threshold = qs
                         ))
  expect_error(
    pe1 <- policy_eval(
      policy_data = pd1,
      policy_learn = learn1,
      target = "subgroup",
      g_models = g_glm(~ 1),
      q_models = q_glm(~ A * (.)),
      M = 2
    ),
    NA
  )

})

test_that("get_policy.ptl returns a policy", {
  d <- sim_single_stage(200, seed=1)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  pl <- policy_learn(type = "ptl", control = control_ptl())
  expect_error({
    p <- get_policy(pl(pd, q_models = q_glm(), g_models = g_glm()))
  },
  NA
  )

  expect_true(
    inherits(p, what = "policy")
  )

  expect_error(
    p(pd),
    NA
  )
})

test_that("policy_learn with type ptl works as intended", {
  d <- sim_two_stage(200, seed = 1)

  pd <- policy_data(d,
    action = c("A_1", "A_2"),
    baseline = c("BB", "B"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )

  pl <- policy_learn(
    type = "ptl",
    control = control_ptl(),
    alpha = 0,
    L = 1
  )

  expect_error(
    {
      po <- pl(
        policy_data = pd,
        g_models = g_glm(),
        q_models = q_glm()
      )
    },
    NA
  )

  expect_s3_class(
    get_policy(po)(pd),
    "data.table"
  )


  expect_error(
    {
      pl <- policy_learn(
        type = "ptl",
        control = control_ptl(
          policy_vars = c("L", "BB"),
          depth = 3,
          hybrid = TRUE
        ),
        alpha = 0,
        L = 1
      )
    },
    NA
  )

  expect_error(
    {
      po <- pl(
        policy_data = pd,
        g_models = g_glm(),
        q_models = q_glm()
      )
    },
    NA
  )

  expect_s3_class(get_policy(po)(pd), "data.table")
})


test_that("policy_learn with type ptl handles varying action sets", {
  d <- sim_two_stage_multi_actions(n = 1e2)
  pd <- policy_data(
    data = d,
    action = c("A_1", "A_2"),
    baseline = c("B", "BB"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )

  pl <- policy_learn(
    type = "ptl",
    control = control_ptl()
  )
  expect_error(
    po <- pl(pd, q_models = q_glm(), g_models = list(g_glm(), g_rf())),
    NA
  )
  expect_error(
    pa <- get_policy(po)(pd),
    NA
  )

  expect_error(
    pe <- policy_eval(pd,
      policy = get_policy(po),
      g_models = list(g_glm(), g_rf())
    ),
    NA
  )

  # realistic actions
  d2 <- copy(d)
  d2$A_2[d2$A_2 == "no"] <- "yes"
  pd2 <- policy_data(
    data = d2,
    action = c("A_1", "A_2"),
    baseline = c("B", "BB"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )

  pl2 <- policy_learn(
    type = "ptl",
    control = control_ptl(),
    alpha = 0.45,
    L = 2
  )
  expect_error(
    po2 <- pl2(pd2, q_models = q_glm(), g_models = list(g_glm(), g_rf())),
    NA
  )
  expect_error(
    pa2 <- get_policy(po2)(pd2),
    NA
  )
  expect_error(
    pe2 <- policy_eval(pd2,
      policy_learn = pl2,
      g_models = list(g_glm(), g_rf())
    ),
    NA
  )

  expect_error(
    pf2_1 <- get_policy_functions(po2, stage = 1),
    NA
  )
  H1 <- get_history(pd2, stage = 1)$H
  expect_error(
    pfa2_1 <- pf2_1(H1),
    NA
  )
  expect_equal(
    pa2[stage == 1, ]$d,
    pfa2_1
  )

  expect_error(
    pf2_2 <- get_policy_functions(po2, stage = 2),
    NA
  )
  H1 <- get_history(pd2, stage = 2)$H
  expect_error(
    pfa2_2 <- pf2_2(H1),
    NA
  )
  expect_equal(
    pa2[stage == 2, ]$d,
    pfa2_2
  )
})

test_that("policy_learn with type = 'ptl' works with thresholds.", {

  d <- sim_two_stage(200, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  pl <- policy_learn(
    type = "ptl",
    control = control_ptl(),
    threshold = c(0, 0.5, 1)
  )

  expect_no_error(
    po <- pl(
      policy_data = pd,
      g_models = g_glm(),
      q_models = q_glm()
    )
  )

  expect_no_error(
    pf <- get_policy(po)
  )

  expect_equal(
    lapply(pf, function(x) attr(x, "name")) |> unlist(),
    paste0("ptl(eta=", c(0,0.5,1), ")")
  )

  expect_error(
    pf <- get_policy(po, threshold = 2),
    "Invalid threshold. Choose between 0, 0.5, 1"
  )

  expect_no_error(
    pf <- get_policy(po, threshold = 0.5)
  )

  expect_true(
    is.data.table(pf(pd)),
    all(pf(pd)$d %in% get_action_set(pd))
  )

  expect_error(
    pf <- get_policy_functions(po, stage = 1),
    "Choose threshold"
  )

  expect_no_error(
    pf <- get_policy_functions(po, stage = 1, threshold = 1)
  )
})

test_that("get_policy and get_policy_functions agree with type ptl and a non-zero threshold argument,", {
  d <- sim_two_stage(200, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  pl <- policy_learn(
    type = "ptl",
    threshold = 1,
    control = control_ptl()
  )
  po <- pl(pd, q_models = q_glm(), g_models = g_glm())


  ## pred <- predict.blip_function(po$blip_functions$stage_1, new_history = his)
  ## hist(x = pred$blip, breaks = 100)

  d <- get_policy(po)(pd)[stage == 1]$d

  pf <- get_policy_functions(po, stage = 1)
  his <- get_history(pd, stage = 1)
  H <- get_H(his)
  dd <- pf(H)

  expect_equal(
    d,
    dd
  )

  d <- get_policy(po)(pd)[stage == 2]$d
  pf <- get_policy_functions(po, stage = 2)
  his <- get_history(pd, stage = 2)
  H <- get_H(his)
  dd <- pf(H)

  expect_equal(
    d,
    dd
  )
})

fit_Q_function <- function(history, Q, q_model) {
  action_set <- getElement(history, "action_set")
  stage_action_set <- getElement(history, "stage_action_set")
  stage <- getElement(history, "stage")
  deterministic_rewards <- getElement(history, "deterministic_rewards")

  # getting the action (A) and the model matrix (H):
  A <- get_A(history)
  H <- get_H(history)
  AH <- cbind(A, H)

  # checking that the dimensions fit:
  if (nrow(AH) != length(Q))
    stop("Unable to fit Q-function.")

  # checking that all actions in the stage action set occur:
  if (!all(stage_action_set == sort(unique(A)))) {
    mes <- "Not all stage actions occur at stage"
    mes <- paste(mes, paste(stage, collapse = ", "))
    mes <- paste(mes, ". Unable to fit Q-function.", sep = "")
    stop(mes)
  }

  ## getting the historic rewards
  U <- getElement(history, "U")
  ## calculating the residual (fitted) values
  U_A <- apply(
    action_matrix(a = A, action_set = action_set) * U[, deterministic_rewards, with = FALSE],
    MARGIN = 1,
    sum
  )
  V_res <- unlist(Q - U[, "U_bar"] - U_A)

  ## removing missing outcomes (censored/coarsened)
  missing_ <- is.na(Q)
  V_res <- V_res[missing_ == FALSE]
  AH <- AH[missing_ == FALSE, ]

  ## fitting the (residual) Q-model
  q_model <- q_model(V_res = V_res, AH = AH)

  q_function <- list(
    q_model = q_model,
    AH_names = colnames(AH),
    action_set = action_set,
    stage_action_set = stage_action_set,
    stage = stage
  )
  class(q_function) <- "Q_function"

  return(q_function)
}

#' @export
print.Q_function <- function(x, ...) {
  y <- x$q_model
  attr(y, "class") <- NULL
  print(y)
}

#' @export
predict.Q_function <- function(object, new_history, ...) {
  q_model <- getElement(object, "q_model")
  AH_names <- getElement(object, "AH_names")
  ## action set of the new history object
  action_set <- getElement(new_history, "action_set")
  ## stage action set of the fitted Q-function
  stage_action_set <- getElement(object, "stage_action_set")
  stage <- getElement(object, "stage")
  ## deterministic rewards of the new history object
  deterministic_rewards <- getElement(new_history, "deterministic_rewards")

  id_stage <- get_id_stage(new_history)
  new_H <- get_H(new_history)
  if (ncol(new_H) == 0) {
    ## preventing empty data.table
    new_H <- data.table(dummy = rep(as.numeric(NA), nrow(id_stage)))
  }

  new_AH_names <- c("A", names(new_H))

  ## checks
  if(!all(stage_action_set %in% action_set))
    stop("The fitted stage action set is not a subset of the new action set.")
  if (!all(AH_names %in% new_AH_names))
    stop("new_history does not contain the same column names as the original history.")

  ## getting the historic rewards
  U <- new_history$U

  ## getting the residual predictions for the stage action set
  residual_q_predictions <- sapply(
    stage_action_set,
    function(a) predict(q_model, new_AH = cbind(A = a, new_H))
  )
  tmp <- matrix(data = NA, nrow = nrow(id_stage), ncol = length(action_set))
  colnames(tmp) <- action_set
  tmp[, stage_action_set] <- residual_q_predictions
  q_values <- tmp; rm(tmp)

  # adding the historic utilities and deterministic rewards
  q_values <- U$U_bar + U[, deterministic_rewards, with = FALSE] + q_values
  colnames(q_values) <- paste("Q", action_set, sep = "_")

  if (!all(complete.cases(q_values[, action_set %in% stage_action_set, with = FALSE]))){
    if(!is.null(stage)){
      mes <- paste("The Q-function predictions at stage ",
                   stage,
                   " have missing values.",
                   sep = "")
    } else {
      mes <- "The Q-function predictions have missing values."
    }
    stop(mes)
  }

  # including the IDs and stage number
  q_values <- data.table(id_stage, q_values)
  setkeyv(q_values, c("id", "stage"))

  return(q_values)
}

q_step <- function(policy_data, k, full_history, Q, q_models) {

  if (is.null(q_models))
    stop("Please provide q_models.")

  stage <- NULL
  id <- get_id(policy_data)
  id_k <- get_id_stage(policy_data)[stage == k][["id"]]
  idx_k <- (id %in% id_k)
  rm(stage)

  if (is.list(q_models)){
    q_model <- q_models[[k]]
  } else{
    q_model <- q_models
  }

  ## getting the Q-function history for each action event:
  q_history <- get_history(policy_data,
                           stage = k,
                           full_history = full_history)
  ## fitting the Q-function:
  q_function <- fit_Q_function(q_history, Q = Q[idx_k], q_model = q_model)
  ## getting the Q-function values for each action and right-censoring event:
  q_history <- get_history(policy_data,
                           stage = k,
                           full_history = full_history,
                           event_set = c(0,2))
  q_values <- predict(q_function, new_history = q_history)
  ## getting the ID-index for each action and right-censoring event:
  idx_k <- (id %in% q_values[["id"]])

  out <- list(
    q_function = q_function,
    q_values = q_values,
    idx_k = idx_k
  )
  return(out)
}

q_step_cf <- function(folds,
                      policy_data,
                      k,
                      full_history,
                      Q,
                      q_models,
                      save_cross_fit_models,
                      future_args) {
  id <- get_id(policy_data)
  K <- get_K(policy_data)

  future_args <- append(
    future_args,
    list(
      X = folds,
      FUN = function(f) {
        train_id <- id[-f]
        train_policy_data <- subset_id(policy_data, train_id)
        train_Q <- Q[-f]
        if (train_policy_data$dim$K != K) {
          stop("The number of stages varies accross the training folds.")
        }
        train_q_step <- q_step(train_policy_data,
          k = k,
          full_history = full_history,
          Q = train_Q,
          q_models = q_models
        )
        train_q_function <- train_q_step$q_function

        valid_id <- id[f]
        valid_policy_data <- subset_id(policy_data, valid_id)
        valid_history <- get_history(valid_policy_data,
                                     stage = k,
                                     full_history = full_history,
                                     event_set = c(0,2))
        valid_values <- predict(train_q_function, valid_history)

        if (save_cross_fit_models == FALSE) {
          train_q_function <- NULL
        }

        out <- list(
          train_q_function = train_q_function,
          valid_values = valid_values
        )
        return(out)
      }
    )
  )
  q_step_cf <- do.call(what = future.apply::future_lapply, future_args)
  q_step_cf <- simplify2array(q_step_cf)

  q_functions_cf <- q_step_cf["train_q_function", ]
  q_values <- q_step_cf["valid_values", ]

  q_values <- rbindlist(q_values)
  setkeyv(q_values, c("id", "stage"))

  ## getting the ID-index for each Q-value
  ## (action and right-censoring events):
  idx_k <- (id %in% q_values[["id"]])

  out <- list(
    q_functions_cf = q_functions_cf,
    q_values = q_values,
    idx_k = idx_k
  )
  return(out)
}

#' Fit Q-functions
#'
#' \code{fit_Q_functions} is used to fit a list of Q-models
#' @noRd
#' @param policy_data Policy data object created by [policy_data()].
#' @param policy_actions Policy actions, see [policy_def].
#' @param q_models Outcome regression models/Q-models created by [q_glm()], [q_rf()], [q_sl()] or similar functions.
#' @param full_history If TRUE, the full history is used to fit each Q-model. If FALSE, the single stage/"Markov type" history is used to fit each Q-model.
#' @param m_function Function of class m_function to impute missing outcomes.
#' @examples
#' library("polle")
#' ### Simulating two-stage policy data
#' d <- sim_two_stage(2e3, seed = 1)
#' pd <- policy_data(d,
#'   action = c("A_1", "A_2"),
#'   covariates = list(
#'     L = c("L_1", "L_2"),
#'     C = c("C_1", "C_2")
#'   ),
#'   utility = c("U_1", "U_2", "U_3")
#' )
#' pd
#'
#' # Defining a static policy (A=1)
#' pl <- policy_def(1, reuse = TRUE)
#'
#' # fitting a Q-model for each stage:
#' q_functions <- fit_Q_functions(
#'   policy_data = pd,
#'   policy_actions = pl(pd),
#'   q_models = list(q_glm(), q_glm()),
#'   full_history = TRUE
#' )
#' q_functions
fit_Q_functions <- function(policy_data,
                            policy_actions,
                            q_models,
                            full_history = FALSE,
                            m_function = NULL) {
  K <- get_K(policy_data)
  n <- get_n(policy_data)
  action_set <- get_action_set(policy_data)

  # input checks:
  if (is.null(q_models)) {
    stop("Please provide q_models.")
  }

  mes <- "q_models must be a single q_models or a list of K q_models's."
  if (is.list(q_models)) {
    tmp <- all(unlist(lapply(q_models, function(qm) inherits(qm, "q_model"))))
    if (!tmp) {
      stop(mes)
    }
    rm(tmp)
    if (length(q_models) != K) {
      stop(mes)
    }
  } else {
    if (!inherits(q_models, "q_model")) {
      stop(mes)
    }
    if (full_history == TRUE) {
      stop("full_history must be FALSE when a single q-model is provided.")
    }
  }

  # getting the IDs:
  id <- get_id(policy_data)

  ## (n X K+1) matrix with entries
  ## k-column = Q_k(H_{k,i}, d_k(H_{k,i})), k = 1,...,K
  ## K+1-column = U (if no missing final outcomes) or Q_{K+1}(H_{K+1})
  Q <- matrix(nrow = n, ncol = K + 1)
  if (is.null(m_function) == TRUE){
    ## getting the observed (complete) utility:
    utility <- get_utility(policy_data)
    ## (n) vector with entries U_i:
    U <- utility$U
    Q[, K + 1] <- U
  } else {
    Q_K1 <- predict(m_function, new_policy_data = policy_data)
    Q[(id %in% Q_K1[["id"]]), K + 1] <- Q_K1[["Q"]]
  }

  # fitting the Q-functions:
  q_functions <- list()
  for (k in K:1) {
    q_step_k <- q_step(
      policy_data = policy_data,
      k = k,
      full_history = full_history,
      Q = Q[, k + 1],
      q_models = q_models
    )
    # getting the Q-function, Q-function values and the ID-index
    q_functions[[k]] <- q_step_k$q_function
    q_values_k <- q_step_k$q_values
    idx_k <- q_step_k$idx_k

    # getting the Q-function values under the policy for each right-censoring and action event:
    d_k <- merge(q_values_k, policy_actions)[["d"]]
    q_d_values_k <- get_a_values(a = d_k, action_set = action_set, values = q_values_k)$P

    # inserting the Q-function values under the policy in Q:
    Q[idx_k, k] <- q_d_values_k
    Q[!idx_k, k] <- Q[!idx_k, k + 1]
  }

  # setting names, classes and attributes:
  names(q_functions) <- paste("stage_", 1:K, sep = "")
  class(q_functions) <- c("q_functions", "nuisance_functions")
  attr(q_functions, "full_history") <- full_history

  return(q_functions)
}

#' @title Get Q-functions
#'
#' @description \code{get_q_functions()} returns a list of (fitted) Q-functions
#' associated with each stage.
#' @param object Object of class [policy_eval] or [policy_object].
#' @returns List of class [nuisance_functions].
#' @seealso [predict.nuisance_functions]
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed = 1)
#' pd <- policy_data(d,
#'   action = c("A_1", "A_2"),
#'   baseline = c("B"),
#'   covariates = list(
#'     L = c("L_1", "L_2"),
#'     C = c("C_1", "C_2")
#'   ),
#'   utility = c("U_1", "U_2", "U_3")
#' )
#' pd
#'
#' # evaluating the static policy a=1 using outcome regression
#' # based on a GLM model at each stage.
#' pe <- policy_eval(
#'   type = "or",
#'   policy_data = pd,
#'   policy = policy_def(1, reuse = TRUE, name = "A=1"),
#'   q_models = list(q_glm(), q_glm())
#' )
#' pe
#'
#' # getting the Q-functions
#' q_functions <- get_q_functions(pe)
#'
#' # getting the fitted g-function values
#' head(predict(q_functions, pd))
#' @export
get_q_functions <- function(object) {
  UseMethod("get_q_functions")
}

test_that("policy_eval returns g and Q-function values.", {
  d <- sim_single_stage(1e2, seed = 1)
  pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")
  p <- policy_def(1)

  pe <- policy_eval(pd, policy = p)

  expect_true(is.data.table(pe$g_values))
  expect_true(nrow(pe$g_values) == 1e2)
  expect_true(all(names(pe$g_values) == c("id", "stage", "g_0", "g_1")))

  expect_true(is.data.table(pe$q_values))
  expect_true(nrow(pe$q_values) == 1e2)
  expect_true(all(names(pe$q_values) == c("id", "stage", "Q_0", "Q_1")))

  pe <- policy_eval(pd, policy = p, type = "ipw")
  expect_true(!is.null(pe$g_values))
  expect_true(is.null(pe$q_values))

  pe <- policy_eval(pd, policy = p, type = "or")
  expect_true(is.null(pe$g_values))
  expect_true(!is.null(pe$q_values))

  ##
  ## cross-fitting
  ##

  pe <- policy_eval(pd, policy = p, M = 2)
  expect_true(is.data.table(pe$g_values))
  expect_true(nrow(pe$g_values) == 1e2)
  expect_true(all(names(pe$g_values) == c("id", "stage", "g_0", "g_1")))
  expect_true(all(key(pe$g_values) == c("id", "stage")))

  expect_true(is.data.table(pe$q_values))
  expect_true(nrow(pe$q_values) == 1e2)
  expect_true(all(names(pe$q_values) == c("id", "stage", "Q_0", "Q_1")))
  expect_true(all(key(pe$q_values) == c("id", "stage")))

  pe <- policy_eval(pd, policy = p, type = "ipw", M = 2)
  expect_true(!is.null(pe$g_values))
  expect_true(is.null(pe$q_values))
})

test_that(
  "policy_eval do not save g and Q-functions when
save_g_functions = FALSE and save_q_functions = FALSE.",
  {
    d <- sim_single_stage(1e2, seed = 1)
    pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")
    p <- policy_def(1)

    pe <- policy_eval(pd, policy = p)
    expect_true(!is.null(pe$g_functions))
    expect_true(!is.null(pe$q_functions))

    pe <- policy_eval(pd, policy = p, save_g_functions = FALSE)
    expect_true(is.null(pe$g_functions))
    expect_true(!is.null(pe$q_functions))

    pe <- policy_eval(pd, policy = p, save_q_functions = FALSE)
    expect_true(!is.null(pe$g_functions))
    expect_true(is.null(pe$q_functions))

    pe <- policy_eval(pd, policy = p, save_g_functions = FALSE, M = 2)
    expect_true(is.null(pe$cross_fits[[1]]$g_functions))

    pe <- policy_eval(pd, policy = p, save_q_functions = FALSE, M = 2)
    expect_true(is.null(pe$cross_fits[[1]]$q_functions))
  }
)

test_that("policy_eval checks inputs.", {
  d <- sim_single_stage(1e2, seed = 1)
  pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")

  p <- policy_def(1)
  pl <- policy_learn()

  ## policy_data
  expect_error(
    policy_eval(
      policy_data = d,
      policy = p
    ),
    "policy_data must be of inherited class 'policy_data'."
  )

  ## policy
  expect_error(
    policy_eval(
      policy_data = pd
    ),
    "Provide either policy or policy_learn."
  )

  expect_error(
    policy_eval(
      policy_data = pd,
      policy = function() 1
    ),
    "policy must be of inherited class 'policy'."
  )

  ## type
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p,
      type = "test"
    ),
    "type must be either 'dr', 'ipw' or  'or'."
  )
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p,
      type = NULL
    ),
    "type must be a character string."
  )
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p,
      type = c("test", "test")
    ),
    "type must be a character string."
  )

  ## policy_learn
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = NULL,
      policy_learn = p
    ),
    "policy_learn must be of inherited class 'policy_learn'."
  )
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p,
      policy_learn = p
    ),
    "Provide either policy or policy_learn."
  )

  ## g-models
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p,
      g_model = g_glm()
    ),
    NA
  )
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p,
      g_models = list(g_glm())
    ),
    NA
  )
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p,
      g_model = function() 1
    ),
    "g_models must be a single g_model or a list of K g_models's."
  )

  gfun <- fit_g_functions(pd, g_glm(), FALSE)
  qfun <- fit_Q_functions(pd, q_models = q_glm(), policy_actions = p(pd), m_function = NULL)

  ## g-functions
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p,
      g_functions = gfun
    ),
    NA
  )
  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      g_functions = qfun
    ),
    "g_functions must be of class 'g_functions'."
  )

  ## q-functions
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p,
      q_functions = qfun
    ),
    NA
  )
  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      q_functions = gfun
    ),
    "q_functions must be of class 'q_functions'."
  )

  ## M: cross-fitting
  mm <- list("test", TRUE, c(1, 2), -1, 0, 1.2)
  lapply(mm, function(m) {
    expect_error(
      policy_eval(
        policy_data = pd,
        policy = p,
        M = m
      )
    )
  })
  mm <- list(1, 2)
  lapply(mm, function(m) {
    expect_error(
      policy_eval(
        policy_data = pd,
        policy = p,
        M = m
      ),
      NA
    )
  })
  rm(mm)

  ## names
  nn <- list("test", NULL, 1, TRUE)
  lapply(nn, function(n) {
    expect_error(
      policy_eval(
        policy_data = pd,
        policy = p,
        name = n
      ),
      NA
    )
  })

  nn <- list(c(1, 2), list())
  lapply(nn, function(n) {
    expect_error(
      policy_eval(
        policy_data = pd,
        policy = p,
        name = n
      )
    )
  })
})

test_that("policy_eval handles combinations of nuisance models and functions correctly.", {

  d <- sim_single_stage(1e2, seed = 1)
  pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")
  p <- policy_def(1)

  ## g-models and g-functions:
  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  expect_error(
    pe <- policy_eval(pd, policy = p, g_functions = NULL, g_models = NULL),
    "Provide either g_functions or g_models."
  )

  expect_no_error(
    pe <- policy_eval(pd, policy = p, g_functions = gf, g_models = g_glm(~Z))
  )
  ## if a g-function is provided, the g-model is ignored
  expect_equal(
    get_g_functions(pe),
    gf
  )
  expect_equal(
    pe$g_values,
    predict.nuisance_functions(gf, new_policy_data = pd)
  )

  ## cross-fitting
  expect_no_error(
    pe <- policy_eval(pd, policy = p, g_functions = gf, g_models = g_glm(~Z), M = 2)
  )
  ## if a g-function is provided, the g-model is ignored
  expect_equal(
    pe$cross_fits$fold_1$g_functions,
    gf
  )
  expect_equal(
    pe$g_values,
    predict.nuisance_functions(gf, new_policy_data = pd)
  )
  rm(gf, pe)

  ## q-models and q-functions:
  qf <- fit_Q_functions(pd, policy_actions = p(pd), q_models = q_glm(~1))

  expect_error(
    pe <- policy_eval(pd, policy = p, q_functions = NULL, q_models = NULL),
    "Provide either q_functions or q_models."
  )

  expect_no_error(
    pe <- policy_eval(pd, policy = p, q_functions = qf, q_models = q_glm(~Z))
  )

  ## if a g-function is provided, the g-model is ignored
  expect_equal(
    get_q_functions(pe),
    qf
  )

  expect_equal(
    pe$q_values,
    predict.nuisance_functions(qf, new_policy_data = pd)
  )

  ## cross-fitting
  expect_no_error(
    pe <- policy_eval(pd, policy = p, q_functions = qf, q_models = q_glm(~Z), M = 2)
  )

  ## if a g-function is provided, the g-model is ignored
  expect_equal(
    pe$cross_fits$fold_1$q_functions,
    qf
  )
  expect_equal(
    pe$q_values,
    predict.nuisance_functions(qf, new_policy_data = pd)
  )

})

test_that("policy_eval runs on a subset of the data with missing actions from the action set.", {
  d1 <- sim_single_stage(1e2, seed = 1)
  pd1 <- policy_data(d1, action = "A", covariates = c("Z"), utility = "U")

  pd2 <- subset_id(pd1, id = get_id(pd1)[d1$A == "0"])

  ## ipw
  expect_error(
    pe1_ipw <- policy_eval(pd1, policy = policy_def(1), type = "ipw"),
    NA
  )
  expect_error(
    pe2_ipw <- policy_eval(
      policy_data = pd2,
      g_functions = get_g_functions(pe1_ipw),
      policy = policy_def(1),
      type = "ipw"
    ),
    NA
  )
  expect_equal(
    coef(pe2_ipw) |> unname(),
    0
  )
  expect_equal(
    IC(pe2_ipw),
    matrix(rep(0, get_n(pd2)))
  )

  ## or
  expect_error(
    pe1_or <- policy_eval(pd1, policy = policy_def(1), type = "or"),
    NA
  )
  expect_error(
    pe2_or <- policy_eval(
      policy_data = pd2,
      q_functions = get_q_functions(pe1_or),
      policy = policy_def(1),
      type = "or"
    ),
    NA
  )
  expect_equal(
    IC(pe2_or),
    NULL
  )
})

test_that("policy_eval handles varying stage action sets.", {
  d <- sim_two_stage_multi_actions(n = 1e2)
  pd <- policy_data(
    data = d,
    action = c("A_1", "A_2"),
    baseline = c("B", "BB"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )

  p <- policy_def(
    c("yes", "no")
  )
  expect_error(
    pe <- policy_eval(pd,
      policy = p,
      type = "dr",
      g_models = list(g_glm(), g_rf())
    ),
    NA
  )

  p <- policy_def(
    c("default", "default")
  )
  expect_error(
    policy_eval(pd,
      policy = p,
      type = "dr",
      g_models = list(g_glm(), g_rf())
    ),
    "The policy actions does not comply with the stage action sets of the policy data object."
  )
})


test_that("policy_eval with target = 'value' runs when cross-fitting.", {
  d <- sim_single_stage(1e2, seed = 1)
  pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")
  p <- policy_def(1, name = 1)

  set.seed(1)
  expect_no_error(
    pe_pooled <- policy_eval(
      policy_data = pd,
      policy = p,
      M = 2,
      variance_type = "pooled"
    )
  )

  expect_true(
    is.numeric(coef(pe_pooled))
  )

  set.seed(1)
  expect_no_error(
    pe_stacked <- policy_eval(
      policy_data = pd,
      policy = p,
      M = 2,
      variance_type = "stacked"
    )
  )

  expect_equal(
    coef(pe_pooled),
    coef(pe_stacked)
  )

  expect_true(
    (pe_pooled |> estimate() |> vcov())
    != (pe_stacked |> estimate() |> vcov())
  )

  set.seed(1)
  expect_no_error(
    pe_pooled <- policy_eval(
      policy_data = pd,
      policy = p,
      M = 2,
      variance_type = "complete"
    )
  )
})

test_that("policy_eval with target = 'value' agrees with targeted::cate in the single-stage case.", {
  d <- sim_single_stage(1e2, seed = 1)
  pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")
  p <- policy_def(1)

  ## set.seed(1)
  ## ca_dml1 <- targeted::cate(
  ##   treatment = A ~ 1,
  ##   response_model = U ~ A * Z,
  ##   propensity_model = A ~ Z,
  ##   type = "dml1",
  ##   data = d,
  ##   nfolds = 2,
  ##   contrast = 1
  ##   )

  ## set.seed(1)
  ## pe_dml1 <- policy_eval(
  ##   policy_data = pd,
  ##   policy = p,
  ##   M = 2,
  ##   variance_type = "stacked",
  ##   )

  set.seed(1)
  ca_dml2 <- targeted::cate(
    cate.model = A ~ 1,
    response.model = U ~ A * Z,
    propensity.model = A ~ Z,
    data = d,
    nfolds = 2,
    second.order = FALSE,
    contrast = 1
  )

  set.seed(1)
  pe_dml2 <- policy_eval(
    policy_data = pd,
    policy = p,
    M = 2,
    variance_type = "pooled",
  )

  expect_equal(
    coef(ca_dml2$estimate)["(Intercept)"] |> unname(),
    coef(pe_dml2) |> unname()
  )

  expect_equal(
    vcov(ca_dml2$estimate)["(Intercept)", "(Intercept)"] |> unname(),
    vcov(estimate(pe_dml2))[1, 1] |> unname()
  )
})


test_that("policy_eval with target 'value' has the correct outputs in the single-stage case.", {
  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  d <- data.table(z = z, a = a, y = y, p = p)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p"),
    utility = c("y")
  )

  ## his <- get_history(pd, stage = 1)
  ## qfun <- fit_Q_function(history = his, Q = d$y, q_degen(var = "z"))
  ## predict.Q_function(qfun, new_history = his)

  p <- policy_def(function(p) p, name = "test")

  ref_pe <- mean((d$a == d$p) / 0.5 * (d$y - d$z) + d$z)
  ref_IC <- (d$a == d$p) / 0.5 * (d$y - d$z) + d$z - ref_pe

  ##
  ## no cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = q_degen(var = "z"),
    g_models = g_glm(~1)
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  ##
  ## cross-fitting: default (pooled)
  ##

  ## in each training, the empirical propensity is no longer 0.5
  ## instead a g_model is fitted on the complete data:
  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = q_degen(var = "z"),
    g_functions = gf,
    M = 2,
    )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  ##
  ## cross-fitting: stacked variance (IC)
  ##

  set.seed(1)
  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = q_degen(var = "z"),
    g_functions = gf,
    M = 2,
    variance_type = "stacked"
  )

  set.seed(1)
  n <- nrow(d)
  M <- 2
  folds <- split(sample(1:n, n), rep(1:M, length.out = n))
  folds <- lapply(folds, sort)

  ref_IC <- numeric(n)
  for (f in folds) {
    d_fold <- d[f, ]
    ref_pe_fold <- mean((d_fold$a == d_fold$p) / 0.5 * (d_fold$y - d_fold$z) +
      d_fold$z)
    ref_IC[f] <- (d_fold$a == d_fold$p) / 0.5 * (d_fold$y - d_fold$z) +
      d_fold$z - ref_pe_fold
  }

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  ##
  ## cross-fitting: complete variance
  ##

  set.seed(1)
  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = q_degen(var = "z"),
    g_functions = gf,
    M = 2,
    variance_type = "complete"
  )

  ref_pe <- mean((d$a == d$p) / 0.5 * (d$y - d$z) + d$z)
  ref_IC <- (d$a == d$p) / 0.5 * (d$y - d$z) + d$z - ref_pe

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )
})

test_that("policy_eval passes nuisance models to policy_learn correctly", {
  d <- sim_single_stage(1e2, seed = 1)
  pd <- policy_data(d, action = "A", covariates = c("Z"), utility = "U")

  ## g-models and g-functions
  pl <- policy_learn(type = "blip", control = control_blip(), alpha = 0.05)
  gf <- fit_g_functions(pd, g_models = g_glm(~1))
  expect_no_error(
    pe <- policy_eval(pd,
                      policy_learn = pl,
                      g_functions = gf,
                      g_models = g_glm(~Z)) # g_models are ignored with no cross-fitting
  )
  po <- get_policy_object(pe)
  expect_equal(
    po$g_functions,
    gf
  )

  pl <- policy_learn(type = "blip", control = control_blip(), alpha = 0.05, L = 2, save_cross_fit_models = TRUE)
  expect_no_error(
    pe <- policy_eval(pd,
                      policy_learn = pl,
                      g_functions = gf,     # g_functions are used within policy_eval
                      g_models = g_glm(~Z)) # g_models are used within policy_learn
  )

  po <- get_policy_object(pe)
  expect_equal(
    gf,
    get_g_functions(pe)
  )
  expect_equal(
    predict.nuisance_functions(gf, new_policy_data = pd),
    pe$g_values
  )
  tmp <- glm(A ~ Z, family = binomial, data = d[-po$folds[[1]],])
  tmp2 <- po$g_functions_cf[[1]]$all_stages
  expect_equal(
    coef(tmp) |> unname(),
    coef(tmp2$g_model$model) |> unname()
  )

  ## q-models and q-functions
  pl <- policy_learn(type = "blip", control = control_blip(), alpha = 0.05)
  qf <- fit_Q_functions(pd, policy_actions = policy_def(1)(pd), q_models = q_glm(~A))
  expect_no_error(
    pe <- policy_eval(pd,
                      policy_learn = pl,
                      q_functions = qf,
                      q_models = q_glm(~A+Z)) # q_models are NOT ignored even when a Q-function is provided.
  )
  po <- get_policy_object(pe)
  expect_true(
    !identical(
      po$q_functions$stage_1$q_model$model |> coef() |> unname(),
      qf$stage_1$q_model$model |> coef() |> unname()
    )
  )

})

test_that("policy_eval() using policy_learn() has the correct output", {
  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p <- (z > 75) + 1
  d <- data.table(z = z, a = a, y = y, p = p)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p"),
    utility = c("y")
  )

  ref_pe <- mean((d$a == d$p) / 0.5 * (d$y - d$z) + d$z)
  ref_IC <- matrix((d$a == d$p) / 0.5 * (d$y - d$z) + d$z - ref_pe)

  pl <- policy_learn(
    type = "blip",
    threshold = 75,
    control = control_blip(blip_models = q_degen(var = "z"))
  )

  pe <- policy_eval(
    policy_data = pd,
    policy_learn = pl,
    q_models = q_degen(var = "z"),
    g_models = g_glm(~1)
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=blip(eta=75)")
  )

  expect_equal(
    ref_pe,
    coef(pe) |> unname()
  )
  expect_equal(
    IC(pe),
    ref_IC
  )
})

test_that("policy_eval with target 'value' has the correct outputs for the fixed two-stage case and a single final utility outcome.", {

  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  y = y,
                  p1 = p1,
                  p2 = p2)
  rm(x1, a1, x2, a2, y)

  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  ## his <- get_history(pd, stage = 1)
  ## qfun <- fit_Q_function(history = his, Q = d$y, q_degen(var = "z"))
  ## predict.Q_function(qfun, new_history = his)

  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ref_pe <- mean(
    d$x1 +
    (d$a1 == d$p1) / 0.5 * (d$x2 - d$x1) +
    (d$a1 == d$p1) / 0.5 * (d$a1 == d$p1) / 0.5 * (d$y - d$x2)
  )

  ref_IC <- d$x1 +
    (d$a1 == d$p1) / 0.5 * (d$x2 - d$x1) +
    (d$a1 == d$p1) / 0.5 * (d$a1 == d$p1) / 0.5 * (d$y - d$x2 ) -
    ref_pe

  ##
  ## no cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    g_models = list(g_glm(~1), g_glm(~1))
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

})

test_that("policy_eval with target 'value' has the correct outputs for the stochastic two-stage case and a single final utility outcome.", {

  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  m <- rbinom(n = 1e2, size = 1, prob = 0.5) # stage 2 missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  m = m)
  rm(x1, a1, x2, a2, y)

  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage == 2, m := m]
  ld[, m := nafill(m, fill = 0)]
  ld <- ld[m == 0, ]
  ld[ , stage := 1:.N, by = id]
  setkey(ld, id, stage)
  ld[ , m := NULL]

  pd <- policy_data(data = ld, type = "long")

  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ref_pe <- mean(d$x1 +
                 (d$a1 == d$p1) / 0.5 * (((d$m == 0) * d$x2 + (d$m == 1) * d$y) - d$x1) +
                 (d$m == 0) * (d$a1 == d$p1) / 0.5 * (d$a2 == d$p2) / 0.5 * (d$y - d$x2))

  ref_IC <-  (d$x1 +
              (d$a1 == d$p1) / 0.5 * (((d$m == 0) * d$x2 + (d$m == 1) * d$y) - d$x1) +
              (d$m == 0) * (d$a1 == d$p1) / 0.5 * (d$a2 == d$p2) / 0.5 * (d$y - d$x2) -
              ref_pe)

  ref_pe_or <- mean(d$x1)

  ref_pe_ipw <- mean((d$a1 == d$p1) / 0.5 * (d$m == 0) * (d$a2 == d$p2) / 0.5 *  d$y)

  ##
  ## no cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    g_functions = gf
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  expect_equal(
    get_element(pe, "coef_or"),
    ref_pe_or
  )

  expect_equal(
    get_element(pe, "coef_ipw"),
    ref_pe_ipw
  )

  ##
  ## cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    g_functions = gf,
    M = 2
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  expect_equal(
    get_element(pe, "coef_or"),
    ref_pe_or
  )

  expect_equal(
    get_element(pe, "coef_ipw"),
    ref_pe_ipw
  )


})


test_that("policy_eval() return estimates for multiple policies associated with multiple thresholds.", {

  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p1 <- (z > 28) + 1
  p2 <- (z > 76) + 1
  d <- data.table(z = z, a = a, y = y, p1 = p1, p2 = p2)
  rm(a, z, y, p1, p2)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z"),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  ref_pe1 <- mean((d$a == d$p1) / 0.5 * (d$y - d$z) + d$z)
  ## pooled IC
  ref_IC1 <- matrix((d$a == d$p1) / 0.5 * (d$y - d$z) + d$z - ref_pe1)

  ref_pe2 <- mean((d$a == d$p2) / 0.5 * (d$y - d$z) + d$z)
  ref_IC2 <- matrix((d$a == d$p2) / 0.5 * (d$y - d$z) + d$z - ref_pe2)

  ref_pe <- c(ref_pe1, ref_pe2)
  ref_IC <- cbind(ref_IC1, ref_IC2)

  p1 <- policy_def(28)
  p0 <- policy_def(76)
  expect_error(
    policy_eval(
      policy_data = pd,
      policy = "test"
    ),
    "policy must be of inherited class 'policy'."
  )
  rm(p1, p0)

  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_degen(var = "z")),
    threshold = c(28, 76)
  )

  expect_no_error({
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = pl,
      q_models = q_degen(var = "z"),
      g_functions = gf
    )
  })
  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=blip(eta=28)", "E[U(d)]: d=blip(eta=76)")
  )

  ##
  ## cross-fitting
  ##

  ## cross fit estimate types
  expect_no_error({
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = pl,
      M = 3,
      cross_fit_type = "pooled",
      q_models = q_degen(var = "z"),
      g_functions = gf
    )
  })

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=blip(eta=28)", "E[U(d)]: d=blip(eta=76)")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_no_error({
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = pl,
      M = 3,
      cross_fit_type = "stacked",
      q_models = q_degen(var = "z"),
      g_functions = gf
    )
  })
  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )
  
  ## variance types
  expect_no_error({
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = pl,
      M = 3,
      q_models = q_degen(var = "z"),
      g_functions = gf
    )
  })

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=blip(eta=28)", "E[U(d)]: d=blip(eta=76)")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_no_error({
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = pl,
      M = 3,
      variance_type = "stacked",
      q_models = q_degen(var = "z"),
      g_functions = gf
    )
  })

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=blip(eta=28)", "E[U(d)]: d=blip(eta=76)")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_no_error({
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = pl,
      M = 3,
      variance_type = "pooled",
      q_models = q_degen(var = "z"),
      g_functions = gf
    )
  })
  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

})

test_that("conditional.policy_eval agrees with targeted::cate", {
  n <- 1e3
  B <- rbinom(n = n, size = 1, prob = 0.5)
  Z <- rnorm(n = n)
  A <- rbinom(size = 1, n = n, prob = 0.5)
  U <- rnorm(mean = Z + Z * A, n = n)
  d <- data.table(Z = Z, A = A, U = U, B = B)
  rm(Z, A, U, B)
  pd <- policy_data(d, action = "A", covariates = c("Z"), baseline = ("B"), utility = "U")
  p1 <- policy_def(1)
  p0 <- policy_def(0)

  ##
  ## no cross-fitting:
  ##

  pe1 <- policy_eval(
    policy_data = pd,
    policy = p1,
    g_models = g_glm(~1),
    q_models = q_glm(~ A * Z),
    target = "value",
    M = 1
  )
  pe0 <- policy_eval(
    policy_data = pd,
    policy = p0,
    g_models = g_glm(~1),
    q_models = q_glm(~ A * Z),
    target = "value",
    M = 1
  )

  cond_pe1 <- conditional(pe1, pd, baseline = "B")
  cond_pe0 <- conditional(pe0, pd, baseline = "B")

  est <- merge(cond_pe1, cond_pe0)
  est <- estimate(est, function(p) p[1] - p[3])

  ## implementation from the targeted package:
  library(targeted)

  ca <- targeted::cate(
    cate.model = A ~ factor(B) - 1,
    response.model = U ~ A * Z,
    propensity.model = A ~ 1,
    second.order = FALSE,
    data = d,
    nfolds = 1,
  )

  expect_equal(
    unname(coef(est)),
    unname(coef(ca)["factor(B)0"])
  )

  expect_equal(
    IC(est) |> matrix(),
    ca$estimate$IC[, "factor(B)0"] |> unname() |> matrix()
  )
})

test_that("policy_eval with target = 'value' runs when performing repeated cross-fitting.", {

  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  d <- data.table(z = z, a = a, y = y, p = p)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p"),
    utility = c("y")
  )
  p <- policy_def(function(p) p, name = "test")

  ref_pe <- mean((d$a == d$p) / 0.5 * (d$y - d$z) + d$z)
  ref_IC <- (d$a == d$p) / 0.5 * (d$y - d$z) + d$z - ref_pe

  gf <- fit_g_functions(pd, g_models = g_glm(~1))
  set.seed(1)
  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      M = 2,
      nrep = 2,
      q_models = q_degen(var = "z"),
      g_functions = gf
    )
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p1 <- (z > 28) + 1
  p2 <- (z > 76) + 1
  d <- data.table(z = z, a = a, y = y, p1 = p1, p2 = p2)
  rm(a, z, y, p1, p2)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z"),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  ref_pe1 <- mean((d$a == d$p1) / 0.5 * (d$y - d$z) + d$z)
  ## pooled IC
  ref_IC1 <- matrix((d$a == d$p1) / 0.5 * (d$y - d$z) + d$z - ref_pe1)

  ref_pe2 <- mean((d$a == d$p2) / 0.5 * (d$y - d$z) + d$z)
  ref_IC2 <- matrix((d$a == d$p2) / 0.5 * (d$y - d$z) + d$z - ref_pe2)

  ref_pe <- c(ref_pe1, ref_pe2)
  ref_IC <- cbind(ref_IC1, ref_IC2)


  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_degen(var = "z")),
    threshold = c(28, 76)
  )

  expect_no_error({
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = pl,
      q_models = q_degen(var = "z"),
      g_functions = gf,
      nrep = 2,
      M = 2
    )
  })

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    ref_IC
  )
})

test_that("policy_eval() runs without covariates.", {

  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  d <- data.table(a = a, y = y)
  rm(a, y)
  pd <- policy_data(
    data = d,
    action = "a",
    utility = c("y")
  )
  p <- policy_def("2", name = "test")

  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      q_models = q_glm(~A),
      g_models = g_glm(~1)
    )
  )

  ref_pe <- mean((d$a == "2") / 0.5 * d$y)
  ref_IC <- matrix((d$a == "2") / 0.5 * (d$y - 4) + 4 - ref_pe)

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    ref_IC,
    tolerance = 1e-14
  )

  expect_no_error(
    pe <- policy_eval(
      policy_data = pd,
      policy = p,
      q_models = q_glm(~1),
      g_models = g_glm(~1)
    )
  )

  ref_pe <- mean((d$a == "2") / 0.5 * d$y)
  ref_IC <- matrix((d$a == "2") / 0.5 * (d$y - 3) + 3 - ref_pe)

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    ref_IC,
    tolerance = 1e-14
  )
})

check_formula <- function(formula, data, call = "") {
  tt <- terms(formula, data = data)
  if (length(attr(tt, "term.labels")) > 0) {
    formula <- reformulate(attr(tt, "term.labels"), response = NULL)
    tt <- terms(formula, data = data)
    all_vars <- all.vars(tt)
    data_names <- colnames(data)
    if (!all(all_vars %in% data_names)) {
      missing_vars <- all_vars[!all_vars %in% data_names]

      if (length(missing_vars) > 0) {

        formula_str <- deparse(formula, width.cutoff = 500L)
        formula_str <- paste(formula_str, collapse = "")


        if (length(missing_vars) == 1) {
          missing_text <- paste0("variable '", missing_vars, "' is")
        } else {
          missing_text <- paste0("variables '",
                                 paste(missing_vars, collapse = "', '"),
                                 "' are")
        }

        call_text <- if (nchar(call) > 0) paste0("when calling '", call, "' ") else ""

        error_msg <- paste0(missing_text, " not found in data ",
                            call_text, "with formula: ", formula_str)


        err <- simpleError(error_msg)
        err$missing_variables <- missing_vars
        err$available_variables <- data_names
        err$formula <- formula
        err$call_context <- call

        stop(err)
      }
    }
  }
}

update_g_formula <- function(formula, A, H) {
  if (!inherits(formula, "formula")) {
    stop("'formula' must be a formula object")
  }

  action_set <- sort(unique(A))
  if (length(action_set) != 2)
    stop("g_glm requires exactly two levels.")
  AA_ <- A
  AA_[A == action_set[1]] <- 0
  AA_[A == action_set[2]] <- 1
  AA_ <- as.numeric(AA_)
  tt <- terms(formula, data = H)

  if (length(attr(tt, "term.labels")) == 0)
    formula <- AA_ ~ 1
  else
    formula <- reformulate(attr(tt, "term.labels"), response = "AA_")

  if ("AA_" %in% colnames(H)) {
    stop("Variable name 'AA_' is not allowed in the history H. ",
         "Please rename this column.")
  }

  environment(formula)$AA_ <- AA_
  attr(formula, "action_set") <- action_set
  attr(formula, "response") <- "AA_"

  return(formula)
}

check_missing_regressor <- function(formula, data){
  data <- as.data.table(data)
  tt <- terms(formula, data = data)
  tt <- terms(reformulate(attr(tt, "term.labels"), response = NULL),
              data = data)
  v <- all.vars(tt)
  cols_na <- names(which(colSums(is.na(data[,v, with = FALSE])) > 0))
  if (length(cols_na) > 0){
    mes <- paste(cols_na, collapse = ", ")
    mes <- paste("The regression variables ", mes, " have missing (NA) values.", sep = "")
    warning(mes)
  }
}

new_g_model <- function(g_model){
  class(g_model) <- c("g_model", "function")
  return(g_model)
}

# g_model documentation -----------------------------------------------------------

#' @title g_model class object
#'
#' @description  Use \code{g_glm()}, \code{g_empir()},
#' \code{g_glmnet()}, \code{g_rf()}, \code{g_sl()}, \code{g_xgboost} to construct
#' an action probability model/g-model object.
#' The constructors are used as input for [policy_eval()] and [policy_learn()].
#'
#' @param formula An object of class [formula] specifying the design matrix for
#' the propensity model/g-model. Use [get_history_names()] to view the available
#' variable names.
#' @param family A description of the error distribution and link function to
#' be used in the model.
#' @param model (Only used by \code{g_glm}) If \code{FALSE} model frame will
#' not be saved.
#' @param na.action (Only used by \code{g_glm}) A function which indicates what
#' should happen when the data contain NAs, see [na.pass].
#' @param alpha (Only used by \code{g_glmnet}) The elastic net mixing parameter
#' between 0 and 1. alpha equal to 1 is the lasso penalty, and alpha equal
#' to 0 the ridge penalty.
#' @param s (Only used by \code{g_glmnet}) Value(s) of the penalty parameter
#' lambda at which predictions are required, see [glmnet::predict.glmnet()].
#' @param num.trees (Only used by \code{g_rf}) Number of trees.
#' @param mtry (Only used by \code{g_rf}) Number of variables to possibly split
#'  at in each node.
#' @param cv_args (Only used by \code{g_rf} and \code{g_xgboost}) Cross-validation parameters.
#' Only used if multiple hyper-parameters are given. \code{K} is the number
#' of folds and
#' \code{rep} is the number of replications.
#' @param SL.library (Only used by \code{g_sl}) Either a character vector of prediction algorithms or
#' a list containing character vectors, see [SuperLearner::SuperLearner].
#' @param env (Only used by \code{g_sl}) Environment containing the learner functions. Defaults to the calling environment.
#' @param onlySL (Only used by \code{g_sl}) Logical. If TRUE, only saves and computes predictions
#' for algorithms with non-zero coefficients in the super learner object.
#' @param objective (Only used by \code{g_xgboost}) specify the learning
#' task and the corresponding learning objective, see [xgboost::xgboost].
#' @param nrounds (Only used by \code{g_xgboost}) max number of boosting iterations.
#' @param max_depth (Only used by \code{g_xgboost}) maximum depth of a tree.
#' @param learning_rate (Only used by \code{g_xgboost}) learning rate.
#' @param nthread (Only used by \code{g_xgboost}) number of threads.
#' @param ... Additional arguments passed to [glm()], [glmnet::glmnet],
#' [ranger::ranger] or [SuperLearner::SuperLearner].
#' @details
#' \code{g_glm()} is a wrapper of [glm()] (generalized linear model).\cr
#' \code{g_empir()} calculates the empirical probabilities within the groups
#' defined by the formula.\cr
#' \code{g_glmnet()} is a wrapper of [glmnet::glmnet()] (generalized linear model via
#' penalized maximum likelihood).\cr
#' \code{g_rf()} is a wrapper of [ranger::ranger()] (random forest).
#' When multiple hyper-parameters are given, the
#' model with the lowest cross-validation error is selected.\cr
#' \code{g_sl()} is a wrapper of [SuperLearner::SuperLearner] (ensemble model).\cr
#' \code{g_xgboost()} is a wrapper of [xgboost::xgboost].
#' @returns g-model object: function with arguments 'A'
#' (action vector), 'H' (history matrix) and 'action_set'.
#' @seealso [get_history_names()], [get_g_functions()].
#' @docType class
#' @name g_model
#' @examples
#' library("polle")
#' ### Two stages:
#' d <- sim_two_stage(2e2, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # available state history variable names:
#' get_history_names(pd)
#' # defining a g-model:
#' g_model <- g_glm(formula = ~B+C)
#'
#' # evaluating the static policy (A=1) using inverse propensity weighting
#' # based on a state glm model across all stages:
#' pe <- policy_eval(type = "ipw",
#'                   policy_data = pd,
#'                   policy = policy_def(1, reuse = TRUE),
#'                  g_models = g_model)
#' # inspecting the fitted g-model:
#' get_g_functions(pe)
#'
#' # available full history variable names at each stage:
#' get_history_names(pd, stage = 1)
#' get_history_names(pd, stage = 2)
#'
#' # evaluating the same policy based on a full history
#' # glm model for each stage:
#' pe <- policy_eval(type = "ipw",
#'                    policy_data = pd,
#'                    policy = policy_def(1, reuse = TRUE),
#'                    g_models = list(g_glm(~ L_1 + B),
#'                                    g_glm(~ A_1 + L_2 + B)),
#'                    g_full_history = TRUE)
#' # inspecting the fitted g-models:
#' get_g_functions(pe)
NULL


# empirical (group) probability -------------------------------------------

calculate_prop_table <- function(data, formula){
  tt <- terms(formula, data = data)
  # grouping variables:
  v <- all.vars(tt)
  N_ <- N_group_ <- empir_prob <- A <-  NULL
  tab <- data[ , list(N_ = .N), by = c("A", v)]
  tab[, N_group_ := sum(N_), by = v]
  tab <- tab[, empir_prob := N_ / N_group_][order(A), ]
  tab[, c("N_","N_group_") := NULL]

  return(list(tab=tab, v=v))
}

#' @rdname g_model
#' @export
g_empir <- function(formula = ~1, ...) {
  formula <- as.formula(formula)
  environment(formula) <- NULL

  g_empir <- function(A, H, action_set){
    check_formula(formula = formula, data = H, call = "g_empir")
    data <- cbind(A, H)
    tmp <- calculate_prop_table(data, formula = formula)
    tab <- tmp[["tab"]]
    v <- tmp[["v"]]
    out <- list(tab=tab, action_set=action_set, v=v)
    class(out) <- c("g_empir")
    return(out)
  }
  # setting class:
  g_empir <- new_g_model(g_empir)
  return(g_empir)
}
#' @export
predict.g_empir <- function(object, new_H, ...){
  tab <- object[["tab"]]
  action_set <- object[["action_set"]]
  v <- object[["v"]]

  if (length(v) == 0){
    A <- NULL
    probs <- tab[order(A),]$empir_prob
    probs <- matrix(probs,
                    nrow = nrow(new_H),
                    ncol = length(action_set),
                    byrow = TRUE)
  } else{
    probs <- matrix(0,
                    nrow = nrow(new_H),
                    ncol = length(action_set),
                    byrow = TRUE)
    new_H <- cbind(id = 1:nrow(new_H), new_H)
    for (j in seq_along(action_set)){
      A_ <- action_set[j]
      id <- A <- NULL
      tmp <- merge(new_H, tab[A == A_,], all.x = TRUE)[order(id)]
      tmp[is.na(tmp)] <- 0
      probs[,j] <- tmp[["empir_prob"]]
    }
  }

  return(probs)
}

# glm interface --------------------------------------

#' @rdname g_model
#' @export
g_glm <- function(formula = ~.,
                  family = "binomial",
                  model = FALSE,
                  na.action = na.pass,
                  ...) {
  formula <- as.formula(formula)
  dotdotdot <- list(...)

  g_glm <- function(A, H, action_set){
    check_formula(formula = formula, data = H, call = "g_glm")
    formula <- update_g_formula(formula, A, H)
    args_glm <- append(list(formula = formula,
                            data = H,
                            family = family,
                            model = model,
                            na.action = na.action),
                       dotdotdot)

    model <- tryCatch(do.call(what = "glm", args = args_glm),
                      error = function(e) e)
    if (inherits(model, "error")) {
      formula <- delete.response(terms(formula))
      model$message <-
        paste0(model$message, " when calling 'g_glm' with formula:\n",
               format(formula))
      stop(model)
    }

    model$call <- NULL

    m <- list(model = model)
    class(m) <- c("g_glm")
    return(m)
  }

  ## setting class:
  g_glm <- new_g_model(g_glm)

  return(g_glm)
}
#' @export
predict.g_glm <- function(object, new_H, ...){
  model <- getElement(object, "model")
  fit <- predict.glm(object = model, newdata = new_H, type = "response")
  probs <- cbind((1-fit), fit)
  return(probs)
}


# glmnet (Elastic Net) interface ------------------------------------------

#' @rdname g_model
#' @export
g_glmnet <- function(formula = ~.,
                     family = "binomial",
                     alpha = 1,
                     s = "lambda.min", ...) {
  if (!requireNamespace("glmnet")) stop("Package 'glmnet' required")
  formula <- as.formula(formula)
  dotdotdot <- list(...)
  g_glmnet <- function(A, H, action_set){
    check_formula(formula = formula, data = H, call = "g_glmnet")
    formula <- update_g_formula(formula, A, H)
    y <- get_response(formula, data=H)
    des <- get_design(formula, data=H)
    if (ncol(des$x)<2)
      stop("g_glmnet requires a model matrix with two or more columns.")

    args_glmnet <- list(y = y,  x = des$x, family = family, alpha = alpha)
    args_glmnet <- append(args_glmnet, dotdotdot)

    model <- tryCatch(do.call(glmnet::cv.glmnet, args = args_glmnet),
                      error = function(e) e
    )
    if (inherits(model, "error")) {
      formula <- delete.response(terms(formula))
      model$message <-
        paste0(model$message, " when calling 'g_glmnet' with formula:\n",
               format(formula))
      stop(model)
    }

    model$call <- NULL
    des$x <- NULL

    m  <- list(model = model,
               s = s,
               design = des,
               action_set = action_set)
    class(m) <- c("g_glmnet")
    return(m)
  }
  # setting class:
  g_glmnet <- new_g_model(g_glmnet)

  return(g_glmnet)
}

#' @export
predict.g_glmnet <- function(object, new_H, ...) {
  design <- getElement(object, "design")
  model <- getElement(object, "model")
  s <- getElement(object, "s")

  newx <- apply_design(design, data = new_H)
  fit <- predict(model,  newx = newx, type = "response", s = s)
  probs <- cbind((1-fit), fit)
  return(probs)
}

# ranger (Random Forest) interface ----------------------------------------

perf_ranger_prob <- function(fit, data,  ...) {
  score_class(predict(fit, data=as.matrix(data[,-1,drop=FALSE]),
                      num.threads=1)$predictions,
              data[,1], ...)
}

#' @rdname g_model
#' @export
g_rf <- function(formula = ~.,
                 num.trees=c(500),
                 mtry=NULL,
                 cv_args=list(nfolds=5, rep=1),
                 ...) {
  if (!requireNamespace("ranger")) stop("Package 'ranger' required")
  formula <- as.formula(formula)
  environment(formula) <- NULL
  dotdotdot <- list(...)
  hyper_par <- expand.list(num.trees=num.trees, mtry=mtry)
  rf_args <- function(p) {
    list(probability=TRUE, num.threads=1,
         num.trees=p$num.trees, mtry=p$mtry)
  }
  ml <- lapply(hyper_par, function(p)
    function(data) {
      rf_args <- append(rf_args(p), list(y=data[,1], x=as.matrix(data[,-1,drop=FALSE])))
      rf_args <- append(rf_args, dotdotdot)
      do.call(ranger::ranger, args=rf_args)
    })

  g_rf <- function(A, H, action_set) {
    A <- factor(A, levels = action_set)
    check_formula(formula = formula, data = H, call = "g_rf")
    des <- get_design(formula, data=H)
    data <- data.frame(A, des$x)
    res <- NULL; best <- 1
    if (length(ml)>1) {
      res <- tryCatch(targeted::cv(ml, data=data, perf=perf_ranger_prob,
                               nfolds=cv_args$nfolds, rep=cv_args$rep),
                      error=function(...) NULL)
      best <- if (is.null(res)) 1 else which.min(summary(res))
    }
    if (!is.null(res)) {
      model <- res$fit[[best]]
    } else {
      model <- ml[[best]](data)
    }

    model$call <- NULL
    des$x <- NULL

    m <- list(model = model,
              rf_args = rf_args(hyper_par[[best]]),
              num.trees=num.trees[best],
              design = des,
              action_set = action_set)
    class(m) <- c("g_rf")
    return(m)
  }
  # setting class:
  g_rf <- new_g_model(g_rf)

  return(g_rf)
}

#' @export
predict.g_rf <- function(object, new_H, ...){
  model <- getElement(object, "model")
  design <- getElement(object, "design")

  new_data <- apply_design(design = design, data = new_H)
  pr <- predict(model, data=new_data, num.threads=1)$predictions
  return(pr)
}

# SuperLearner interface --------------------------------------------------

#' @rdname g_model
#' @export
g_sl <- function(formula = ~ .,
                 SL.library=c("SL.mean", "SL.glm"),
                 family=binomial(),
                 env = parent.frame(),
                 onlySL = TRUE,
                 ...) {
  if (!requireNamespace("SuperLearner"))
    stop("Package 'SuperLearner' required.")
  formula <- as.formula(formula)
  force(SL.library)
  force(env)
  dotdotdot <- list(...)
  g_sl <- function(A, H, action_set) {
    A <- as.numeric(factor(A, levels=action_set))-1
    check_formula(formula = formula, data = H, call = "g_sl")
    des <- get_design(formula, data=H)
    sl_args <- append(list(Y=A,
                           X=as.data.frame(des$x),
                           family=family,
                           SL.library=SL.library,
                           env = env),
                      dotdotdot)
    model <- do.call(SuperLearner::SuperLearner, sl_args)

    model$call <- NULL
    if(all(model$coef == 0))
      stop("In g_sl(): All metalearner coefficients are zero.")
    if(onlySL == TRUE){
      model$fitLibrary[model$coef == 0] <- NA
    }
    des$x <- NULL

    m <- list(model = model,
              design = des,
              onlySL = onlySL,
              action_set = action_set)

    class(m) <- c("g_sl")
    return(m)
  }
  # setting class:
  g_sl <- new_g_model(g_sl)

  return(g_sl)
}

#' @export
predict.g_sl <- function(object, new_H, ...) {
  model <- getElement(object, "model")
  design <- getElement(object, "design")
  onlySL <- getElement(object, "onlySL")
  newdata <- apply_design(design = design, data = new_H)
  newdata <- as.data.frame(newdata)
  pr <- predict(model,
                newdata=newdata,
                onlySL = onlySL)$pred
  pr <- cbind((1-pr), pr)
  return(pr)
}

# sl3 (SuperLearner) interface ----

# g_sl3 <- function(formula = ~ ., learner, folds=5, ...) {
#   if (!requireNamespace("sl3"))
#     stop("Package 'sl3' required.")
#   force(formula)
#   dotdotdot <- list(...)
#   if (missing(learner)) {
#     lrn_enet <- sl3::make_learner(sl3::Lrnr_glmnet, alpha=0.5, outcome_type="binomial")
#     lrn_rf <- sl3::make_learner(sl3::Lrnr_ranger, num.trees = 500, outcome_type="binomial")
#     lrn_mean <- sl3::make_learner(sl3::Lrnr_mean, outcome_type="binomial")
#     lrn_stack <- sl3::Stack$new(lrn_enet, lrn_rf, lrn_mean)
#     learner <- sl3::make_learner(sl3::Lrnr_sl, learners = lrn_stack, outcome_type="binomial")
#   }
#   g_sl3 <- function(A, H, action_set) {
#     A <- factor(A, levels=action_set)
#     des <- get_design(formula, data=H)
#     X <- data.frame(des$x)
#     colnames(X) <- gsub("[^[:alnum:]]", "_", colnames(X))
#     tsk <- sl3::make_sl3_Task(data=cbind(A, X),
#                               outcome_type="categorical",
#                               outcome="A",
#                               outcome_levels=action_set,
#                               covariates=names(X),
#                               folds=folds)
#     fit <- learner$train(tsk)
#     m <- with(des, list(fit = fit,
#                         xlevels = xlevels,
#                         terms = terms,
#                         action_set = action_set))
#     class(m) <- c("g_sl3")
#     return(m)
#   }
#   class(g_sl3) <- c("g_model", "function")
#   return(g_sl3)
# }
#
# predict.g_sl3 <- function(object, new_H, ...) {
#   mf <- with(object, model.frame(terms, data=new_H, xlev = xlevels,
#                                  drop.unused.levels=FALSE))
#   newx <- as.data.frame(model.matrix(mf, data=new_H, xlev = object$xlevels))
#   tsk <- sl3::make_sl3_Task(data=newx, covariates=names(newx))
#   pr <- object$fit$predict(tsk)
#   return(pr)
# }


# xgboost interface -----------------------------------------------------------------

#' @rdname g_model
#' @export
g_xgboost <- function(formula = ~ .,
                      objective = "binary:logistic",
                      nrounds,
                      max_depth = 6,
                      learning_rate = NULL,
                      nthread = 1,
                      cv_args = list(nfolds = 3, rep = 1),
                      ...) {
  if (!requireNamespace("xgboost")) stop("Package 'xgboost' required")
  formula <- as.formula(formula)
  environment(formula) <- NULL
  dotdotdot <- list()

  ml <- function(formula = A ~ .,
                 objective,
                 nrounds,
                 max_depth,
                 learning_rate,
                 nthread,
                 dotdotdot) {

    targeted::learner$new(formula,
                          info = "xgBoost",
                          estimate = function(x, y) {
                            xgb_args <- list(
                              x = quote(x), # avoid embedding data in call
                              y = quote(y),
                              objective = objective,
                              nrounds = nrounds,
                              max_depth = max_depth,
                              learning_rate = learning_rate,
                              nthread = nthread,
                              verbosity = 0,
                              print_every = 0
                            )
                            xgb_args <- append(xgb_args, dotdotdot)
                            do.call(xgboost::xgboost, args = xgb_args)
                          })
  }

  ml_args <- expand.list(
    nrounds = nrounds,
    max_depth = max_depth,
    learning_rate = learning_rate
  )
  cv_par <- ml_args
  ml_args <- lapply(ml_args, function(p) {
    p <- append(p,
                list(
                  objective = objective,
                  dotdotdot = dotdotdot,
                  nthread = nthread
                ))
    return(p)
  })
  ml_models <- lapply(ml_args, function(par) do.call(ml, par))

  g <- function(A, H, action_set) {
    # checks:
    if (length(action_set) != 2)
      stop("g_xgboost in only implemented for a dichotomous action set.")

    # formatting data:
    A <- factor(A, levels = action_set)
    A <- as.numeric(A) - 1
    A <- as.logical(A)
    check_formula(formula = formula, data = H, call = "g_xgboost")
    des <- get_design(formula, data=H)
    data <- data.frame(A, des$x)

    # cross-validating models
    cv_res <- NULL
    if (length(ml_models)>1){
      cv_res <- tryCatch(targeted::cv(ml_models, data, nfolds = cv_args$nfolds, rep = cv_args$rep),
        error = function(e) e
        )
      if (inherits(cv_res, "error")) {
        formula <- delete.response(terms(formula))
        cv_res$message <-
          paste0(cv_res$message, " when calling 'g_xgboost' with formula:\n",
                 format(formula))
        stop(cv_res)
      }
      cv_res$names <- unlist(lapply(cv_par, function(x) paste(paste(names(x), x, sep = ":"), collapse = ",")))
      ml_args_best <- ml_args[[which.min(coef(cv_res)[, 1])]]
      model <- do.call(ml, ml_args_best)
      model <- model$estimate(data)
    } else {
      model <- do.call(ml, ml_args[[1]])
      model <- model$estimate(data)
    }

    # setting model output
    des$x <- NULL
    m <- list(model = model,
              design = des,
              action_set = action_set,
              cv_res = cv_res)
    class(m) <- c("g_xgboost")
    return(m)
  }
  # setting class:
  g <- new_g_model(g)

  return(g)
}

#' @export
predict.g_xgboost <- function(object, new_H, ...){
  model <- getElement(object, "model")
  design <- getElement(object, "design")
  new_data <- apply_design(design = design, data = new_H)

  fit <- predict(model, newdata = new_data)
  probs <- cbind((1-fit), fit)

  return(probs)
}

test_that("calculate_prop_table", {
  data <- data.table(
    A = c("yes", "no", "yes", "no", "yes", "yes"),
    B = c("a", "a", "a", "a", "b", "b"),
    BB = c("c","c","c","d","d", "d")
  )

  expect_equal(
    calculate_prop_table(data, formula = ~ 1)[["tab"]],
    data.table(A = c("no", "yes"), empir_prob = c(2/6, 4/6))
  )

  expect_equal(
    calculate_prop_table(data, formula = ~ B)[["tab"]],
    data.table(A = c("no", "yes", "yes"), B = c("a", "a", "b"), empir_prob = c(1/2, 1/2, 1))
  )

})

test_that("g_empir predictions in a single stage setting", {
  n <- 1e2
  d <- sim_single_stage(n = n, seed = 1)
  set.seed(1)
  d$BB <- sample(x = c("group1", "group2"), prob = c(0.5, 0.5), replace = TRUE, size = n)
  d$BBB <- d$A

  pd <- policy_data(d,
                    action = "A",
                    covariates = c("Z", "L", "B", "BB", "BBB"),
                    utility = "U")

  # no groups
  g_functions <- fit_g_functions(pd,
                                 g_models = g_empir(),
                                 full_history = FALSE)
  pred <- predict(g_functions, new_policy_data = pd)
  ref <- pd$stage_data[event == 0, list(p = .N / n) ,by = "A"][order(A),]$p
  ref <- matrix(rep(ref, times = n), ncol = 2, byrow = TRUE)
  expect_equal(
    ref,
    unname(as.matrix(pred[, c("g_0", "g_1"), with = FALSE]))
  )

  # single grouping variable
  g_functions <- fit_g_functions(pd,
                                 g_models = g_empir(~B),
                                 full_history = FALSE)
  g_functions2 <- fit_g_functions(pd,
                                 g_models = g_glm(~B),
                                 full_history = FALSE)
  pred <- predict(g_functions, new_policy_data = pd)
  pred2 <- predict(g_functions2, new_policy_data = pd)
  ref <- pd$stage_data[event == 0, list(p = .N) ,by = c("A", "B")]
  ref[, N_B := sum(p), by = "B"]
  ref[, p := p / N_B]
  ref[, N_B := NULL]
  H <- get_history(pd)$H
  ref <- cbind(
    merge(H, ref[A==0,], by = "B", all.x = TRUE)[order(id, stage),]$p,
    merge(H, ref[A==1,], by = "B", all.x = TRUE)[order(id, stage),]$p
  )
  expect_equal(
    ref,
    unname(as.matrix(pred[, c("g_0", "g_1"), with = FALSE]))
  )
  expect_equal(
    unname(as.matrix(pred[, c("g_0", "g_1"), with = FALSE])),
    unname(as.matrix(pred2[, c("g_0", "g_1"), with = FALSE]))
  )

  # two grouping variables
  g_functions <- fit_g_functions(pd,
                                 g_models = g_empir(~B + BB),
                                 full_history = FALSE)
  pred <- predict(g_functions, new_policy_data = pd)
  tmp <- cbind(pred, B = d$B, BB = d$BB) # incorrect
  ref <- pd$stage_data[event == 0, list(p = .N) ,by = c("A", "B", "BB")]
  ref[, N_group := sum(p), by = c("B", "BB")]
  ref[, p := p / N_group]
  ref[, N_group := NULL]
  ref <- cbind(
    merge(H, ref[A==0,], by = c("B", "BB"), all.x = TRUE)[order(id, stage),]$p,
    merge(H, ref[A==1,], by = c("B", "BB"), all.x = TRUE)[order(id, stage),]$p
  )
  expect_equal(
    ref,
    unname(as.matrix(pred[, c("g_0", "g_1"), with = FALSE]))
  )

  # degenerate distribution
  g_functions <- fit_g_functions(pd,
                                 g_models = g_empir(~B + BBB),
                                 full_history = FALSE)
  pred <- predict(g_functions, new_policy_data = pd)
  ref <- cbind(as.numeric(d$BBB == "0"), as.numeric(d$BBB == "1"))
  expect_equal(
    ref,
    unname(as.matrix(pred[, c("g_0", "g_1"), with = FALSE]))
  )

})

test_that("g_empir predictions in a two stage setting", {
  set.seed(1)
  d <- sim_two_stage_multi_actions(n = 1e2)
  pd <- policy_data(data = d,
                    action = c("A_1", "A_2"),
                    baseline = c("B", "BB"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))
  n <- get_n(pd)

  # no groups
  g_functions <- fit_g_functions(pd,
                                 g_models = list(g_empir(), g_empir()),
                                 full_history = FALSE)
  pred <- predict(g_functions, new_policy_data = pd)

  ref1 <- pd$stage_data[event == 0 & stage == 1, list(p = .N / n) ,by = c("A")][order(A),]$p
  ref2 <- pd$stage_data[event == 0 & stage == 2, list(p = .N / n) ,by = c("A")][order(A),]$p
  ref <- matrix(rep(c(0,ref1, ref2), times = n), ncol = 3, byrow = TRUE)
  expect_equal(
    ref,
    unname(as.matrix(pred[, c("g_default", "g_no", "g_yes"), with = FALSE]))
  )

  # single grouping variable
  g_functions <- fit_g_functions(pd,
                                 g_models = list(g_empir(~BB), g_empir(~BB)),
                                 full_history = FALSE)
  pred <- predict(g_functions, new_policy_data = pd)
  ref <- merge(pd$stage_data, pd$baseline_data)[event == 0, list(p = .N) , by = c("stage","A","BB")]
  ref[, N_B := sum(p), by = c("stage", "BB")]
  ref[, p := p / N_B]
  ref[, N_B := NULL]
  H <- get_history(pd)$H
  ref <- cbind(
    merge(H, ref[A=="default",], by = c("stage","BB"), all.x = TRUE)[order(id, stage),]$p,
    merge(H, ref[A=="no",], by = c("stage","BB"), all.x = TRUE)[order(id, stage),]$p,
    merge(H, ref[A=="yes",], by = c("stage","BB"), all.x = TRUE)[order(id, stage),]$p
  )
  ref[is.na(ref)] <- 0
  expect_equal(
    ref,
    unname(as.matrix(pred[, c("g_default", "g_no", "g_yes"), with = FALSE]))
  )

})

library("data.table")

test_that("update_g_formula", {

  d <- sim_single_stage(1e2, seed = 1)
  d$AA_ <- ifelse(d$A == 0, "p", "t")
  pd <- policy_data(d, action = "A", covariates = c("Z", "AA_"), utility = "U")

  p1 <- policy_def(1)

  expect_error(
    policy_eval(pd,
                policy = p1,
                g_models = g_glm(~Z),
                q_models = q_glm(~Z)),
    "Variable name 'AA_' is not allowed in the history H. Please rename this column."
  )


  expect_error(
    policy_eval(pd,
                policy = p1,
                g_models = g_glm(~ Z+ AA_),
                q_models = q_glm(~ Z)),
    "Variable name 'AA_' is not allowed in the history H. Please rename this column."
  )
})


test_that("predict.g_functions checks the action set",{
  d <- sim_two_stage_multi_actions(n = 1e2)

  pd <- policy_data(data = d,
                    action = c("A_1", "A_2"),
                    baseline = c("B", "BB"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  expect_error(
    gfit <- fit_g_functions(pd, g_models = list(g_glm(), g_rf()), full_history = TRUE),
    NA
  )

  d2 <- copy(d)
  d2$A_1[d2$A_1 == "no"] <- "maybe"

  pd2 <- policy_data(data = d2,
                     action = c("A_1", "A_2"),
                     baseline = c("B", "BB"),
                     covariates = list(L = c("L_1", "L_2"),
                                       C = c("C_1", "C_2")),
                     utility = c("U_1", "U_2", "U_3"))

  expect_error(
    predict(gfit, pd2),
    "factor A_1 has new levels maybe"
  )

  expect_equal(
    {
      gval <- predict.g_function(gfit[[1]],
                                 get_history(pd2, stage = 1,
                                             full_history = TRUE))
      names(gval)
    },
    c("id", "stage", paste("g_", pd2$action_set, sep = ""))
  )

  d3 <- copy(d2)
  d3$A_2[d3$A_2 == "no"] <- "maybe"

  pd3 <- policy_data(data = d3,
                     action = c("A_1", "A_2"),
                     baseline = c("B", "BB"),
                     covariates = list(L = c("L_1", "L_2"),
                                       C = c("C_1", "C_2")),
                     utility = c("U_1", "U_2", "U_3"))

  expect_error(
    predict(gfit, pd3),
    "The fitted stage action set is not a subset of the new action set."
  )

})

test_that("fit_g_functions handles varying stage-action sets", {
  d <- sim_two_stage_multi_actions(n = 1e2)
  expect_error(
    pd <- policy_data(data = d,
                      action = c("A_1", "A_2"),
                      baseline = c("B", "BB"),
                      covariates = list(L = c("L_1", "L_2"),
                                        C = c("C_1", "C_2")),
                      utility = c("U_1", "U_2", "U_3")),
    NA
  )

  expect_error(
    gfit <- fit_g_functions(pd, g_models = list(g_glm(), g_rf()), full_history = TRUE),
    NA
  )
  expect_error(
    predict(gfit, pd),
    NA
  )

  expect_error(
    gfit <- fit_g_functions(pd, g_models = list(g_glm(), g_rf()), full_history = FALSE),
    NA
  )
  expect_error(
    predict(gfit, pd),
    NA
  )

  expect_error(
    gfit <- fit_g_functions(pd, g_models = g_glm(), full_history = FALSE),
    "g_glm requires exactly two levels."
  )

  set.seed(1)
  folds <- list(c(1:30), 31:get_n(pd))
  expect_error(
    gfit <- crossfit_function(folds = folds,
                               policy_data = pd,
                               fun = fit_g_functions,
                               models = list(g_glm(), g_rf()),
                               full_history = FALSE,
                               save_cross_fit_models = TRUE),
    NA
  )
})

test_that("fit_g_function_cf saves the cross-fitted models",{
  d <- sim_two_stage_multi_actions(n = 1e2)
  expect_error(
    pd <- policy_data(data = d,
                      action = c("A_1", "A_2"),
                      baseline = c("B", "BB"),
                      covariates = list(L = c("L_1", "L_2"),
                                        C = c("C_1", "C_2")),
                      utility = c("U_1", "U_2", "U_3")),
    NA
  )

  set.seed(1)
  folds <- list(c(1:30), 31:get_n(pd))
  expect_error(
    gfit <- crossfit_function(folds,
                              policy_data = pd,
                              fun = fit_g_functions,
                              models = list(g_glm(), g_rf()),
                              full_history = FALSE,
                              save_cross_fit_models = TRUE),
    NA
  )

  expect_true(
    all(!unlist(lapply(gfit$functions, is.null)))
  )

  set.seed(1)
  folds <- list(c(1:30), 31:get_n(pd))
  expect_error(
    gfit <- crossfit_function(folds,
                              policy_data = pd,
                              fun = fit_g_functions,
                              models = list(g_glm(), g_rf()),
                              full_history = FALSE,
                              save_cross_fit_models = FALSE),
    NA
  )
  expect_true(
    all(unlist(lapply(gfit$functions, is.null)))
  )

})

test_that("g_models checks formula input", {

  d <- sim_two_stage(1e2, seed = 1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(
                      L = c("L_1", "L_2"),
                      C = c("C_1", "C_2")
                    ),
                    utility = c("U_1", "U_2", "U_3")
                    )

  p_dynamic <- policy_def(
    policy_functions = list(
      function(L_1) (L_1 > 0) * 1,
      function(C_2) (C_2 > 0) * 1
    ),
    reuse = FALSE
  )

  expect_error(
    policy_eval(policy_data = pd,
                policy = p_dynamic,
                g_models = g_glm(formula = A ~ X)),
    "variable 'X' is not found in data when calling 'g_glm' with formula: ~X"
  )

  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p_dynamic,
      g_models = g_sl(formula = a ~ X)),
    "variable 'X' is not found in data when calling 'g_sl' with formula: ~X"
  )

  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p_dynamic,
      g_models = g_rf(formula = ~X)),
    "variable 'X' is not found in data when calling 'g_rf' with formula: ~X"
  )

  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p_dynamic,
      g_models = g_glmnet(formula = ~X)),
    "variable 'X' is not found in data when calling 'g_glmnet' with formula: ~X"
  )

  expect_error(
    policy_eval(
      policy_data = pd,
      policy = p_dynamic,
      g_models = g_empir(formula = ~X)),
    "variable 'X' is not found in data when calling 'g_empir' with formula: ~X"
  )

})


test_that("g_rf runs:", {
  d1 <- sim_single_stage(200, seed = 1)
  d1$BB <- sample(c("group 1", "group & 2", "group & 3"), size = 200, replace = TRUE)
  pd1 <- policy_data(d1,
    action = "A",
    covariates = list("Z", "B", "L", "BB"),
    utility = "U"
  )

  expect_no_error(
    pe <- policy_eval(
      policy_data = pd1,
      policy_learn = policy_learn(type = "ql", alpha = 0.05),
      g_models = g_rf(formula = ~.),
      g_full_history = FALSE,
      type = "ipw"
    )
  )

  expect_is(pe, class = "policy_eval")
})


test_that("g_sl formats data correctly via the formula", {
  d1 <- sim_single_stage(200, seed = 1)
  d1$BB <- sample(c("group 1", "group & 2", "group & 3"), size = 200, replace = TRUE)
  pd1 <- policy_data(d1,
    action = "A",
    covariates = list("Z", "B", "L", "BB"),
    utility = "U"
  )

  library("SuperLearner")
  env <- as.environment("package:SuperLearner")
  expect_error(
    policy_eval(
      policy_data = pd1,
      policy_learn = policy_learn(type = "ql", alpha = 0.05),
      g_models = g_sl(formula = ~., env = env),
      g_full_history = FALSE,
      q_models = q_glm()
    ),
    NA
  )
})

test_that("g_sl can find user-defined learners", {
  d1 <- sim_single_stage(200, seed = 1)
  d1$BB <- sample(c("group 1", "group & 2", "group & 3"), size = 200, replace = TRUE)
  pd1 <- policy_data(d1,
    action = "A",
    covariates = list("Z", "B", "L", "BB"),
    utility = "U"
  )

  library("SuperLearner")
  env <- as.environment("package:SuperLearner")
  env <- new.env(parent = env)
  env$SL.test <- function(Y, X, newX, family, obsWeights, model = TRUE, ...) {
    if (is.matrix(X)) {
      X = as.data.frame(X)
    }
    fit.glm <- glm(Y ~ .,
      data = X, family = family, weights = obsWeights,
      model = model
    )
    if (is.matrix(newX)) {
      newX = as.data.frame(newX)
    }
    pred <- predict(fit.glm, newdata = newX, type = "response")
    fit <- list(object = fit.glm)
    class(fit) <- "SL.test"
    out <- list(pred = pred, fit = fit)
    return(out)
  }
  env$predict.SL.test <- function(object, newdata, ...) {
    if (is.matrix(newdata)) {
      newdata = as.data.frame(newdata)
    }
    pred <- predict(
      object = object$object, newdata = newdata,
      type = "response"
    )
    pred
  }

  g_model <- g_sl(
    formula = ~.,
    SL.library = "SL.test",
    env = env
  )
  his <- get_history(pd1)

  expect_error(
    {
      g_model(
        H = polle:::get_H(his),
        A = polle:::get_A(his),
        action_set = pd1$action_set
      )
    },
    NA
  )
})

test_that("g_glmnet formats data correctly via the formula",{
  d1 <- sim_single_stage(200, seed=1)
  d1$BB <- sample(c("group 1", "group & 2", "group & 3"), size = 200, replace = TRUE)
  pd1 <- policy_data(d1,
                     action="A",
                     covariates = list("Z", "B", "L", "BB"),
                     utility="U")

  expect_error(
    pe <- policy_eval(
      policy_data = pd1,
      policy_learn = policy_learn(type = "ql", alpha = 0.05),
      g_models = g_glmnet(formula = ~.),
      g_full_history = FALSE,
      q_models = q_glm()
    ),
    NA
  )
})


# missing values ----------------------------------------------------------

test_that("g_glm handles missing covariates", {
  d <- sim_two_stage(2e3, seed=1)
  d$C_1 <- NULL
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c(NA, "C_2")), # C_1 is missing
                    utility = c("U_1", "U_2", "U_3"))
  p <- policy_def(1, reuse = TRUE)

  expect_error(
    policy_eval(policy_data = pd,
                policy = p,
                type = "ipw"),
    "NA/NaN/Inf in 'x'"
  )
  expect_error(
    policy_eval(policy_data = pd,
                policy = p,
                g_models = list(g_glm(), g_glm()),
                type = "ipw"),
    "NA/NaN/Inf in 'x'"
  )

  expect_error(
    policy_eval(policy_data = pd,
                policy = p,
                g_models = g_glm(~L),
                q_models = q_glm(~L)),
    NA
  )
  expect_error(
    pe <- policy_eval(policy_data = pd,
                policy = p,
                g_models = list(g_glm(~L), g_glm()),
                q_models = list(q_glm(~L), q_glm())),
    NA
  )

  d <- sim_two_stage(2e3, seed=1)
  d$C_1[1:10] <- NA
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")), # C_1 is missing
                    utility = c("U_1", "U_2", "U_3"))
  expect_error(
    policy_eval(policy_data = pd,
                policy = p,
                type = "ipw"),
    "NA/NaN/Inf in 'x'"
  )
})

test_that("g_glmnet handles missing covariates", {
  d <- sim_two_stage(2e3, seed=1)
  d$C_1 <- NULL
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c(NA, "C_2")), # C_1 is missing
                    utility = c("U_1", "U_2", "U_3"))
  p <- policy_def(1, reuse = TRUE)

  expect_error(
    suppressWarnings({
      pe <- policy_eval(policy_data = pd,
                        policy = p,
                        type = "ipw",
                        g_models = list(g_glmnet(), g_glmnet()))
    })
  )

  expect_error(
    policy_eval(policy_data = pd,
                policy = p,
                type = "ipw",
                g_models = g_glmnet(~L + B)),
    NA
  )
  expect_error(
    policy_eval(policy_data = pd,
                policy = p,
                type = "ipw",
                g_models = list(g_glmnet(~L + B), g_glmnet())),
    NA
  )
  d <- sim_two_stage(2e3, seed=1)
  d$C_1[1:10] <- NA
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")), # C_1 is missing
                    utility = c("U_1", "U_2", "U_3"))
  expect_error(
    expect_warning(
      policy_eval(policy_data = pd,
                  policy = p,
                  type = "ipw",
                  g_models = g_glmnet()),
      "The regression variables C have missing NA values."
    )
  )

})

test_that("g_xgboost gives the same result as plain xgboost",{
  # data
  d <- sim_single_stage(1000, seed=1)
  x <- d[,c("Z", "L", "B")]
  y <- as.logical(d[["A"]])

  set.seed(1)
  bst <- xgboost::xgboost(x = x, y = y,
                 max_depth = 2, learning_rate = 1, nrounds = 2,
                 objective = "binary:logistic",
                 verbosity = 0)

  pred <- predict(bst, newdata = x)

  # policy data
  pd <- policy_data(d,
                    action = "A",
                    covariates = list("Z", "B", "L"),
                    utility = "U")

  set.seed(1)
  g_fun <- fit_g_functions(pd,
                           g_models = g_xgboost(max_depth = 2,
                                                learning_rate = 1,
                                                nrounds = 2))

  pred_polle <- predict(g_fun, pd)

  expect_equal(
    unname(pred),
    pred_polle[["g_1"]]
  )

})

## test_that("g_xgboost gives the same result as SL.xgboost", {
##   # policy data
##   d <- sim_single_stage(1000, seed=1)
##   pd <- policy_data(d,
##                     action="A",
##                     covariates = list("Z", "B", "L"),
##                     utility="U")

##   g_fun <- fit_g_functions(pd,
##                            g_models = g_xgboost(max_depth = 2,
##                                                 learning_rate = 1,
##                                                 nrounds = 1))
##   pred_1 <- predict(g_fun, pd)

##   xgboost_tune <- list(max_depth = 2,
##                        shrinkage = 1, # eta
##                        ntrees  = 1) # nrounds
##   sl_learner <- create.Learner("SL.xgboost",tune = xgboost_tune)
##   g_fun_sl <- fit_g_functions(pd,
##                               g_models = g_sl(SL.library = sl_learner$names, env = environment()))
##   pred_2 <- predict(g_fun_sl, pd)

##   expect_equal(
##     pred_1,
##     pred_2
##   )

## })

test_that("g_xgboost tunes parameters",{
  # data
  d <- sim_single_stage(1000, seed=1)
  # policy data
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  set.seed(1)
  g_fun <- fit_g_functions(pd,
                           g_models = g_xgboost(max_depth = 2,
                                                learning_rate = c(0.3,1),
                                                nrounds = c(2,4)))

  expect_equal(
    length(g_fun$all_stages$g_model$cv_res$cv),
    3 * 2 * 2 * 2
  )

})

new_policy <- function(fun, name){
  attr(fun, "name") <- name
  class(fun) <- c("policy", "function")

  return(fun)
}

#' @title Policy-class
#' @name policy
#'
#' @description  A function of inherited class "policy" takes a policy
#' data object as input and returns the policy actions for every observation
#' for every (observed) stage.
#' @section S3 generics:
#' The following S3 generic functions are available for an object of class
#' \code{policy}:
#' \describe{
#'   \item{\code{print}}{Baisc print function}
#' }
#' @details A policy can either be defined directly by the user using
#' [policy_def] or a policy can be fitted using [policy_learn]
#' (or [policy_eval]). [policy_learn] returns a [policy_object] from which
#' the policy can be extracted using [get_policy].
#' @returns [data.table::data.table] with keys \code{id} and \code{stage} and
#' action variable \code{d}.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#'
#' # defining a dynamic policy:
#' p <- policy_def(
#'   function(L) (L>0)*1,
#'   reuse = TRUE
#' )
#' p
#' head(p(pd), 5)
#'
#' # V-restricted (Doubly Robust) Q-learning:
#' # specifying the learner:
#' pl <- policy_learn(type = "drql",
#'                    control = control_drql(qv_models = q_glm(formula = ~ C)))
#'
#' # fitting the policy (object):
#' po <- pl(policy_data = pd,
#'          q_models = q_glm(),
#'          g_models = g_glm())
#'
#' p <- get_policy(po)
#' p
#'
#' head(p(pd))
#' @docType class
NULL

#' Define Policy
#'
#' \code{policy_def} returns a function of class [policy].
#' The function input is a [policy_data] object and it returns a [data.table::data.table]
#'  with keys \code{id} and \code{stage} and action variable \code{d}.
#'
#' @param policy_functions A single function/character string or a list of
#' functions/character strings. The list must have the same length as the number
#' of stages.
#' @param full_history If TRUE, the full history at each stage is used as
#' input to the policy functions.
#' @param reuse If TRUE, the policy function is reused at every stage.
#' @param name Character string.
#' @param natural_action Logical. If TRUE, the natural/observed action
#' variable will be passeed to the policy functions.
#' @param stage_number Logical. If TRUE, the stage number is passed to the
#' policy functions.
#' @returns Function of class \code{"policy"}. The function takes a
#' [policy_data] object as input and returns a [data.table::data.table]
#' with keys \code{id} and \code{stage} and action variable \code{d}.
#' @seealso [get_history_names()], [get_history()].
#' @examples
#' library("polle")
#' ### Single stage"
#' d1 <- sim_single_stage(5e2, seed=1)
#' pd1 <- policy_data(d1, action="A", covariates=list("Z", "B", "L"), utility="U")
#' pd1
#'
#' # defining a static policy (A=1):
#' p1_static <- policy_def(1)
#'
#' # applying the policy:
#' p1_static(pd1)
#'
#' # defining a dynamic policy:
#' p1_dynamic <- policy_def(
#'   function(Z, L) ((3*Z + 1*L -2.5)>0)*1
#' )
#' p1_dynamic(pd1)
#'
#' ### Two stages:
#' d2 <- sim_two_stage(5e2, seed = 1)
#' pd2 <- policy_data(d2,
#'                   action = c("A_1", "A_2"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#'
#' # defining a static policy (A=0):
#' p2_static <- policy_def(0,
#'                         reuse = TRUE)
#' p2_static(pd2)
#'
#' # defining a reused dynamic policy:
#' p2_dynamic_reuse <- policy_def(
#'   function(L) (L > 0)*1,
#'   reuse = TRUE
#' )
#' p2_dynamic_reuse(pd2)
#'
#' # defining a dynamic policy for each stage based on the full history:
#' # available variable names at each stage:
#' get_history_names(pd2, stage = 1)
#' get_history_names(pd2, stage = 2)
#'
#' p2_dynamic <- policy_def(
#'   policy_functions = list(
#'     function(L_1) (L_1 > 0)*1,
#'     function(L_1, L_2) (L_1 + L_2 > 0)*1
#'   ),
#'   full_history = TRUE
#' )
#' p2_dynamic(pd2)
#' @export
policy_def <- function(policy_functions,
                       full_history = FALSE,
                       reuse = FALSE,
                       name = NULL,
                       natural_action = FALSE,
                       stage_number = FALSE){
  force(policy_functions)
  force(full_history)
  force(reuse)

  ## input checks
  if (!(is.logical(full_history) & (length(full_history) == 1)))
    stop("full_history must be TRUE or FALSE")
  if (!(is.logical(reuse) & (length(reuse) == 1)))
    stop("reuse must be TRUE or FALSE")
  if (full_history == TRUE & reuse == TRUE)
    stop("full_history must be FALSE when reuse is TRUE.")
  if (!is.null(name)){
    name <- as.character(name)
    if (length(name) != 1)
      stop("name must be a character string.")
  }
  if (reuse == TRUE &
      inherits(policy_functions, what = c("list",
                                          "numeric",
                                          "character",
                                          "logical",
                                          "factor",
                                          "integer"))){
    if (length(policy_functions) != 1)
      stop("When reuse is TRUE, policy_functions must be a single function or a constant.")
  }

  policy <- function(policy_data){
    if(!inherits(policy_data, what = "policy_data")){
      stop("The policy input is not of inherited class policy_data.")
    }
    action_set <- get_action_set(policy_data)
    K <- get_K(policy_data)

    if (reuse == TRUE){
      ## policy_functions as a list:
      policy_functions <- replicate(K, policy_functions)
    }
    ## policy_functions as list (or vector):
    policy_functions <- c(policy_functions)

    if (length(policy_functions) != K)
      stop("policy_functions must be a list of length K.")

    policy_functions <- lapply(
      policy_functions,
      function(pf){
        if (inherits(pf, what = "function")){
          pf <- dynamic_policy(fun = pf,
                               natural_action = natural_action,
                               stage_number = stage_number)
        } else {
          pf <- static_policy(action = pf)
        }
        return(pf)
      }
    )

    stage_histories <- lapply(
      1:K,
      function(k) get_history(policy_data,
                              stage = k,
                              full_history = full_history,
                              event_set = c(0,2)) # action and right-censoring events
    )
    policy_actions <- mapply(
      function(sp, sh) sp(sh),
      policy_functions,
      stage_histories,
      SIMPLIFY = FALSE
    )
    policy_actions <- rbindlist(policy_actions)
    setkeyv(policy_actions, c("id", "stage"))

    if (!all(unlist(policy_actions[, "d"]) %in% action_set)){
      mes <- "The policy actions does not comply with the action set of the policy data object."
      warning(mes)
    }

    return(policy_actions)
  }

  # setting class and attributes:
  policy <- new_policy(policy, name = name)

  return(policy)
}

#' @export
print.policy <- function(x, ...) {

  cat("Policy with argument(s)")

  cp <- args(x)
  cp <- capture.output(print.function(cp))
  cp <- capture.output(print.function(x))
  cp <- paste(cp, collapse = "")
  cp <- gsub(" ", "", cp, fixed = TRUE)
  cp <- gsub(",", ", ", cp, fixed = TRUE)
  n_start <- 10
  n_end <- gregexpr(")", cp)[[1]][1]
  cp <- substr(cp,n_start, n_end-1)

  cat("\n")
  cat(cp)
  cat("\n")

}


#' @title Define Static Policy
#'
#' @description \code{static_policy} defines a static policy with a given
#' action.
#' @param action Character string. The static action to be applied.
#' @param name Name of the policy.
#' @returns function with arguments \code{history}. Specifically,
#' \code{history} is a history object, see [history]. When evaluated, the
#' function returns a [data.table::data.table] with keys \code{id} and \code{stage} and
#' action variable \code{d}.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # getting a history object for stage 2:
#' his <- get_history(pd, stage = 2)
#'
#' # applying the static policy at the given stage:
#' head(static_policy(action = 1)(his))
#' @noRd
static_policy <- function(action, name=paste0("a=",action)) {
  action <- as.character(action)
  if (length(action) != 1)
    stop("the action argument in static_policy must be a single character or string.")

  f <- function(history) {
    d <- NULL
    pol <- get_id_stage(history)
    pol[, d := action]
    return(pol)
  }
  return(structure(f, name=name))
}

#' @title Define Dynamic Policy
#'
#' @description \code{dynamic_policy} defines a dynamic policy given by a
#' user-specifed function.
#' @param fun Function with arguments associated with variables a given history
#' object, see example. The function must return a vector of character strings.
#' @param natural_action Logical
#' @returns function with arguments \code{history}. Specifically,
#' \code{history} is a history object, see [history]. When evaluated, the
#' function returns a [data.table::data.table] with keys \code{id} and \code{stage} and
#' action variable \code{d}.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # getting the state/Markov-type history object for stage 2:
#' his <- get_history(pd, stage = 2)
#'
#' # avaible variable names:
#' get_history_names(pd)
#' colnames(his$H)
#'
#' # applying the dynamic policy at stage 2:
#' head(dynamic_policy(fun = function(C) C>1)(his))
#'
#' # getting the full history object for stage 2:
#' his <- get_history(pd, stage = 2, full_history = TRUE)
#'
#' # avaible variable names:
#' get_history_names(pd, stage = 2)
#' colnames(his$H)
#'
#' # applying the dynamic policy at stage 2:
#' head(dynamic_policy(fun = function(C_1, C_2) (C_1>1) & (C_2>1))(his))
#' @noRd
dynamic_policy <- function(fun, natural_action, stage_number) {

  if (!"..." %in% names(formals(fun))) {
    formals(fun) <- c(formals(fun), alist(...=))
  }

  f <- function(history){
    args <- get_H(history)
    if (isTRUE(natural_action)){
      args <- cbind(A = get_A(history), args)
    }
    if (isTRUE(stage_number)) {
      stage <- get_id_stage(history)[["stage"]]
      args <- cbind(stage = stage, args)
    }
    action <- do.call(what = "fun", args = args)
    action <- as.character(action)
    d <- NULL
    pol <- get_id_stage(history)
    pol[, d := action]
    return(pol)
  }
  return(f)
}

policy_g_functions <- function(g_functions, name = "pgf"){
  force(g_functions)

  policy <- function(policy_data){
    action_set <- get_action_set(policy_data)
    g_cols <- paste("g_", action_set, sep = "")
    g_values <- predict.nuisance_functions(g_functions,
                                           policy_data,
                                           event_set = c(0,2))

    dd <- apply(
      g_values[ , g_cols, with = FALSE],
      MARGIN = 1,
      which.max
    )
    d <- action_set[dd]

    policy_actions <- cbind(get_id_stage(policy_data), d = d)
    setkeyv(policy_actions, c("id", "stage"))

    return(policy_actions)
  }

  policy <- new_policy(fun = policy, name = name)

  return(policy)
}

get_response <- function(formula, ...) {
  if (!is.null(attr(formula, "response"))) {
    y <- get(attr(formula, "response"), envir=environment(formula))
  } else {
    y <- model.response(model.frame(formula, ...))
  }
  return(y)
}

get_design <- function(formula, data) {
  tt <- terms(formula, data=data)
  attr(tt, "intercept") <- 1
  tt <- delete.response(tt)
  op <- options(na.action = "na.pass")

  mf <- tryCatch(model.frame(tt, data=data, drop.unused.levels = TRUE),
                    error = function(e) e
  )
  if (inherits(mf, "error")) {
    formula <- delete.response(terms(formula))
    mf$message <-
      paste0(mf$message, " when calling model.frame with formula:\n",
             format(formula))
    stop(mf)
  }

  xlevels <- .getXlevels(tt, mf)
  x <- model.matrix(mf, data=data)
  options(op)
  idx_inter <- which(colnames(x) == "(Intercept)")
  if (length(idx_inter)>0)
    x <- x[,-idx_inter, drop = FALSE]
  colnames(x) <- gsub("[^[:alnum:].]", "_", colnames(x))
  return(list(terms=tt, xlevels=xlevels, x=x))
}

apply_design <- function(design, data){
  terms <- getElement(design, "terms")
  xlevels <- getElement(design, "xlevels")

  op <- options(na.action = "na.pass")
  mf <- model.frame(terms,
                    data=data,
                    xlev = xlevels,
                    drop.unused.levels=FALSE)
  newx <- model.matrix(mf, data=data, xlev = xlevels)
  options(op)

  idx_inter <- which(colnames(newx) == "(Intercept)")
  if (length(idx_inter)>0)
    newx <- newx[,-idx_inter, drop = FALSE]

  colnames(newx) <- gsub("[^[:alnum:].]", "_", colnames(newx))
  return(newx)
}

supp_warnings <- function(expr, mess, fun) {
  if(!is.character(mess))
    stop()
  if(!is.character(fun))
    stop()
  if(length(mess) != length(fun))
    stop()

  eval.parent(
    substitute(
      withCallingHandlers(expr, warning = function (w) {
        mess_ <- mess
        fun_ <- fun
        cm   <- conditionMessage(w)
        cc <- conditionCall(w)
        cond_cc <- FALSE
        if (is.call(cc) & length(as.character(cc))>0){
          cc <- as.character(cc)[[1]]
          cond_cc <- (cc == fun_)
        }
        cond_cm <- (cm == mess_)
        if (any(cond_cm & cond_cc))
          tryInvokeRestart("muffleWarning")
      })
    )
  )
}

test_that("get_design() and apply_design passes dots",{

  df <- data.table(x.1=c(1,2),y=c(3,5), "z:1" =c(3,6), "_w_" = c(1,2))

  des <- get_design(formula = y ~ ., data = df)

  expect_equal(
    colnames(des$x),
    c("x.1", "_z_1_", "__w__")
  )

  ades <- apply_design(design = des, data = df)

  expect_equal(
    colnames(ades),
    c("x.1", "_z_1_", "__w__")
  )

})

library(testthat)

test_that("supp_warnings blocks specific warnings from specific functions",{

  expect_warning(
    supp_warnings(
      expr = {
        warning("warn", call. = FALSE)
      },
      mess = "warn",
      fun = "test"
    ),
    "warn"
  )

  test <- function(){
    warning("warn", call. = TRUE)
  }

  expect_no_warning(
    supp_warnings(
      expr = test(),
      mess = "warn",
      fun = "test"
    )
  )

  test <- function(){
    warning("warn2", call. = FALSE)
  }

  expect_warning(
    supp_warnings(
      expr = test(),
      mess = "warn",
      fun = "test"
    )
  )


  test2 <- function(){
    warning("warn2", call. = TRUE)
  }
  test <- function(){
    warning("warn", call. = TRUE)
    test2()

  }

  expect_no_warning(
    supp_warnings(
      expr = test(),
      mess = c("warn", "warn2"),
      fun = c("test", "test2")
    )
  )

  expect_warning(
    supp_warnings(
      expr = test(),
      mess = c("warn", "warn2"),
      fun = c("test", "test3")
    )
  )

})

#' @rdname policy_learn
#' @export
print.policy_object <- function(x, ...){
  cat("Policy object with list elements:")

  cp <- paste(names(x), collapse = ", ")
  cp <- paste(cp, collapse = "")
  cp <- gsub(" ", "", cp, fixed = TRUE)
  cp <- gsub(",", ", ", cp, fixed = TRUE)

  cat("\n")
  cat(paste(strwrap(cp, 60), collapse="\n"))

    cat("\n")
  cat("Use 'get_policy' to get the associated policy.")
  cat("\n")
}

#' @title Get Policy Object
#'
#' @description Extract the fitted policy object.
#' @param object Object of class [policy_eval].
#' @returns Object of class [policy_object].
#' @examples
#' library("polle")
#' ### Single stage:
#' d1 <- sim_single_stage(5e2, seed=1)
#' pd1 <- policy_data(d1, action="A", covariates=list("Z", "B", "L"), utility="U")
#' pd1
#'
#'
#' # evaluating the policy:
#' pe1 <- policy_eval(policy_data = pd1,
#'                    policy_learn = policy_learn(type = "drql",
#'                                                control = control_drql(qv_models = q_glm(~.))),
#'                    g_models = g_glm(),
#'                    q_models = q_glm())
#'
#' # extracting the policy object:
#' get_policy_object(pe1)
#' @export
get_policy_object <- function(object)
  UseMethod("get_policy_object")

#' @export
get_policy_object.policy_eval <- function(object){
  po <- getElement(object, "policy_object")
  return(po)
}

#' @title Get Policy
#'
#' @description \code{get_policy} extracts the policy from a policy object
#' or a policy evaluation object The policy is a function which take a
#' policy data object as input and returns the policy actions.
#' @param object Object of class [policy_object] or [policy_eval].
#' @param threshold Numeric vector.
#' Thresholds for the first stage policy function.
#' @returns function of class [policy].
#' @examples
#' library("polle")
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed = 1)
#' pd <- policy_data(d,
#'   action = c("A_1", "A_2"),
#'   baseline = c("BB"),
#'   covariates = list(
#'     L = c("L_1", "L_2"),
#'     C = c("C_1", "C_2")
#'   ),
#'   utility = c("U_1", "U_2", "U_3")
#' )
#' pd
#'
#' ### V-restricted (Doubly Robust) Q-learning
#'
#' # specifying the learner:
#' pl <- policy_learn(
#'   type = "drql",
#'   control = control_drql(qv_models = q_glm(formula = ~C))
#' )
#'
#' # fitting the policy (object):
#' po <- pl(
#'   policy_data = pd,
#'   q_models = q_glm(),
#'   g_models = g_glm()
#' )
#'
#' # getting and applying the policy:
#' head(get_policy(po)(pd))
#'
#' # the policy learner can also be evaluated directly:
#' pe <- policy_eval(
#'   policy_data = pd,
#'   policy_learn = pl,
#'   q_models = q_glm(),
#'   g_models = g_glm()
#' )
#'
#' # getting and applying the policy again:
#' head(get_policy(pe)(pd))
#' @export
get_policy <- function(object, threshold = NULL) {
  UseMethod("get_policy")
}

#' @export
get_policy.policy_eval <- function(object, threshold = NULL) {
  po <- get_policy_object(object)
  if (is.null(po)) {
    return(NULL)
  }
  p <- get_policy(po, threshold = threshold)
  return(p)
}

#' Get Policy Functions
#'
#' \code{get_policy_functions()} returns a function defining the policy at
#' the given stage. \code{get_policy_functions()} is useful when implementing
#' the learned policy.
#'
#' @param object Object of class "policy_object" or "policy_eval",
#' see [policy_learn] and [policy_eval].
#' @param stage Integer. Stage number.
#' @param threshold  Numeric, threshold for not
#' choosing the reference action at stage 1.
#' @param ... Additional arguments.
#' @param include_g_values If TRUE, the g-values are included as an attribute.
#' @returns Functions with arguments:
#' \describe{
#' \item{\code{H}}{[data.table::data.table] containing the variables needed to evaluate the policy (and g-function).}
#' }
#' @examples
#' library("polle")
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = "BB",
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' ### Realistic V-restricted Policy Tree Learning
#' # specifying the learner:
#' pl <- policy_learn(type = "ptl",
#'                    control = control_ptl(policy_vars = list(c("C_1", "BB"),
#'                                                             c("L_1", "BB"))),
#'                    full_history = TRUE,
#'                    alpha = 0.05)
#'
#' # evaluating the learner:
#' pe <- policy_eval(policy_data = pd,
#'                   policy_learn = pl,
#'                   q_models = q_glm(),
#'                   g_models = g_glm())
#'
#' # getting the policy function at stage 2:
#' pf2 <- get_policy_functions(pe, stage = 2)
#' args(pf2)
#'
#' # applying the policy function to new data:
#' set.seed(1)
#' L_1 <- rnorm(n = 10)
#' new_H <- data.frame(C = rnorm(n = 10),
#'                     L = L_1,
#'                     L_1 = L_1,
#'                     BB = "group1")
#' d2 <- pf2(H = new_H)
#' head(d2)
#' @export
get_policy_functions <- function(object, stage, threshold, ...) {
  UseMethod("get_policy_functions")
}

#' @export
get_g_functions.policy_object <- function(object) {
  getElement(object, "g_functions")
}

#' @export
get_q_functions.policy_object <- function(object) {
  getElement(object, "q_functions")
}

check_stage <- function(stage, K) {
  if (missing(stage)) {
    stop("stage argument is missing.")
  }
  is_int <- stage %% 1 == 0
  if (!(is_int && (length(stage) == 1) && (stage >= 1) && (stage <= K))) {
    stop("stage must be a positive integer less than or equal to K.")
  }
}

test_that("policy_learn with type 'blip' has the expected output for the fixed two-stage case under right-censoring and single final utility outcome.", {

  ## right-censoring at every stage:
  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  delta1 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 1 non-missing indicator
  delta2 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 2 non-missing indicator
  delta3 <- rbinom(n = 1e2, size = 1, prob = 0.8) # stage 3 non-missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  delta1 = delta1,
                  delta2 = delta2,
                  delta3 = delta3)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage ==3, x := x3]

  ld[stage == 1, delta:= delta1]
  ld[stage == 2, delta:= delta2]
  ld[stage == 3, delta:= delta3]
  ld[ , delta := cumprod(delta), by = id]
  ld[ , tmp := cumsum(delta == 0), by = id]
  ld <- ld[tmp %in% c(0,1)]
  ld[ , tmp := NULL]
  ld[delta == 0, event := 2]
  ld[ , delta := NULL]
  ld[event == 2 & stage == 3, U := NA]
  ld[event == 2, A := NA]

  pd <- policy_data(data = ld, type = "long")

  ## policy_learn(type = "blip")):
  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_glm(~x))
  )

  ## missing c-models and m-model:
  expect_error(
    po <- pl(pd, q_models = q_glm(), g_models = g_glm())
  )

  ## g-functions:
  gf <- fit_g_functions(policy_data = pd,
                        g_models = list(g_glm(~1), g_glm(~1)))
  g1 <- ld[stage == 1 & event == 0, mean(A == "1")]
  g2 <- ld[stage == 1 & event == 0, mean(A == "2")]
  g3 <- ld[stage == 2 & event == 0, mean(A == "3")]
  g4 <- ld[stage == 2 & event == 0, mean(A == "4")]

  ## c-functions:
  cf <- fit_c_functions(policy_data = pd,
                        c_models = g_glm(~1))
  c1 <- unique(predict(cf, new_policy_data = pd)$surv_time2)

  ##
  ## no cross-fitting
  ##

  expect_no_error(
    po <- pl(pd, q_models = list(q_glm(~x), q_degen(var = "x")), g_functions = gf, c_functions = cf, m_model = q_degen(var = "x"))
  )

  ## stage 2

  d2 <- d[delta1 == 1]

  Z2 <- cbind(
    d2$x2 +
    (d2$delta2 == 1) / c1 * (d2$a2 == "3") / (g3 * (d2$a2 == "3") + g4 * (d2$a2 == "4")) * (d2$x3 - d2$x2) +
    (d2$delta2 == 1) / c1 * (d2$delta3 == 1) / c1 * (d2$a2 == "3") / (g3 * (d2$a2 == "3") + g4 * (d2$a2 == "4")) * (d2$y - d2$x3),
     d2$x2 +
    (d2$delta2 == 1) / c1 * (d2$a2 == "4") / (g3 * (d2$a2 == "3") + g4 * (d2$a2 == "4")) * (d2$x3 - d2$x2) +
    (d2$delta2 == 1) / c1 * (d2$delta3 == 1) / c1 * (d2$a2 == "4") / (g3 * (d2$a2 == "3") + g4 * (d2$a2 == "4")) * (d2$y - d2$x3)
  )

  blip2 <- Z2[,2] - Z2[,1]

  blip_model2 <- glm(blip2 ~ d2$x2)

  expect_equal(
    po$blip_functions$stage_2$blip_model$model |> coef() |> unname(),
    blip_model2 |> coef() |> unname()
  )

  dd2 <- ifelse(predict(blip_model2, newdata = data.frame(x2 = d2$x2)) > 0, "4", "3")

  expect_equal(
    unname(dd2),
    get_policy(po)(pd)[stage == 2]$d
  )

  pf2 <- get_policy_functions(po, stage = 2)
  tmp <- pf2(get_history(pd, stage = 2, event_set = c(0,2))$H)

  expect_equal(tmp, unname(dd2))

  tmp <- glm(x2 ~ x1, data = d2)
  q1 <- predict(tmp, newdata = d, type = "response")
  rm(tmp)

  ## stage 1

  d2 <- d$delta1
  d2[d2 == 1] <- dd2

  Z1 <- cbind(
    q1 +
    (d$delta1 == 1) / c1 * (d$a1 == "1") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * (d$x2 - q1) +
    (d$delta1 == 1) / c1 * (d$a1 == "1") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) * (d$x3 - d$x2) +
    (d$delta1 == 1) / c1 * (d$a1 == "1") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) *
    (d$delta3 == 1) / c1 * (d$y - d$x3),
    q1 +
    (d$delta1 == 1) / c1 * (d$a1 == "2") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * (d$x2 - q1) +
    (d$delta1 == 1) / c1 * (d$a1 == "2") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) * (d$x3 - d$x2) +
    (d$delta1 == 1) / c1 * (d$a1 == "2") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) *
    (d$delta3 == 1) / c1 * (d$y - d$x3)
  )

  blip1 <- Z1[,2] - Z1[,1]

  blip_model1 <- glm(blip1 ~ d$x1)

  expect_equal(
    po$blip_functions$stage_1$blip_model$model |> coef() |> unname(),
    blip_model1 |> coef() |> unname()
  )

  d1 <- ifelse(predict(blip_model1, newdata = data.frame(x1 = d$x1)) > 0, "2", "1")

  expect_equal(
    unname(d1),
    get_policy(po)(pd)[stage == 1]$d
  )

  ##
  ## cross-fitting
  ##

  ## policy_learn(type = "blip")):
  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_glm(~x)),
    L = 2,
    cross_fit_g_models = FALSE,
    cross_fit_c_models = FALSE,
    save_cross_fit_models = TRUE
  )

  expect_no_error(
    po <- pl(pd,
             q_models = list(q_glm(~x), q_degen(var = "x")),
             g_functions = gf,
             c_functions = cf,
             m_model = q_degen(var = "x"))
  )

  ## stage 2

  d2 <- d[delta1 == 1]

  Z2 <- cbind(
    d2$x2 +
    (d2$delta2 == 1) / c1 * (d2$a2 == "3") / (g3 * (d2$a2 == "3") + g4 * (d2$a2 == "4")) * (d2$x3 - d2$x2) +
    (d2$delta2 == 1) / c1 * (d2$delta3 == 1) / c1 * (d2$a2 == "3") / (g3 * (d2$a2 == "3") + g4 * (d2$a2 == "4")) * (d2$y - d2$x3),
     d2$x2 +
    (d2$delta2 == 1) / c1 * (d2$a2 == "4") / (g3 * (d2$a2 == "3") + g4 * (d2$a2 == "4")) * (d2$x3 - d2$x2) +
    (d2$delta2 == 1) / c1 * (d2$delta3 == 1) / c1 * (d2$a2 == "4") / (g3 * (d2$a2 == "3") + g4 * (d2$a2 == "4")) * (d2$y - d2$x3)
  )

  blip2 <- Z2[,2] - Z2[,1]

  blip_model2 <- glm(blip2 ~ d2$x2)

  expect_equal(
    po$blip_functions$stage_2$blip_model$model |> coef() |> unname(),
    blip_model2 |> coef() |> unname()
  )

  dd2 <- ifelse(predict(blip_model2, newdata = data.frame(x2 = d2$x2)) > 0, "4", "3")

  expect_equal(
    unname(dd2),
    get_policy(po)(pd)[stage == 2]$d
  )

  pf2 <- get_policy_functions(po, stage = 2)
  tmp <- pf2(get_history(pd, stage = 2, event_set = c(0,2))$H)

  expect_equal(tmp, unname(dd2))

  ## cross-fitted stage 1 Q-function
  q1 <- rep(as.numeric(NA), 1e2)
  q1[po$folds[[1]]] <- predict(glm(x2 ~ x1, data = d[!po$folds[[1]],][delta1 == 1]), type = "response", newdata = d[po$folds[[1]],])
  q1[po$folds[[2]]] <- predict(glm(x2 ~ x1, data = d[-po$folds[[2]],][delta1 == 1]), type = "response", newdata = d[po$folds[[2]],])

  ## stage 1

  d2 <- d$delta1
  d2[d2 == 1] <- dd2

  Z1 <- cbind(
    q1 +
    (d$delta1 == 1) / c1 * (d$a1 == "1") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * (d$x2 - q1) +
    (d$delta1 == 1) / c1 * (d$a1 == "1") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) * (d$x3 - d$x2) +
    (d$delta1 == 1) / c1 * (d$a1 == "1") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) *
    (d$delta3 == 1) / c1 * (d$y - d$x3),
    q1 +
    (d$delta1 == 1) / c1 * (d$a1 == "2") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) * (d$x2 - q1) +
    (d$delta1 == 1) / c1 * (d$a1 == "2") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) * (d$x3 - d$x2) +
    (d$delta1 == 1) / c1 * (d$a1 == "2") / (g1 * (d$a1 == "1") + g2 * (d$a1 == "2")) *
    (d$delta2 == 1) / c1 * (d$a2 == d2) / (g3 * (d$a2 == "3") + g4 * (d$a2 == "4")) *
    (d$delta3 == 1) / c1 * (d$y - d$x3)
  )

  blip1 <- Z1[,2] - Z1[,1]

  blip_model1 <- glm(blip1 ~ d$x1)

  expect_equal(
    po$blip_functions$stage_1$blip_model$model |> coef() |> unname(),
    blip_model1 |> coef() |> unname()
  )

  d1 <- ifelse(predict(blip_model1, newdata = data.frame(x1 = d$x1)) > 0, "2", "1")

  expect_equal(
    unname(d1),
    get_policy(po)(pd)[stage == 1]$d
  )

  d1 <- ifelse(predict(blip_model1, newdata = data.frame(x1 = d$x1)) > 0, "2", "1")

  expect_equal(
    unname(d1),
    get_policy(po)(pd)[stage == 1]$d
  )
})

#' @rdname policy_data
#' @export
print.policy_data <- function(x, digits = 2, ...){
  s <- summary(x, probs=NULL)
  cat(
    paste("Policy data with n = ", s$n,
          " observations and maximal K = ", s$K,
          " action stages.", sep = "")
  )
  cat("\n\n")

  cat("Count of observed actions at a given stage:")

  cat("\n")

  print(s$tab)

  cat("\n")

  if(get_element(s, "cens_indicator") == TRUE){
    cat("Count of observations right-censored at a given stage (not the cumulative count):")

    cat("\n")

    print(get_element(s, "cens_tab"), row.names = FALSE)

    cat("\n")
  }

  bas_var <- paste("Baseline covariates: ", s$baseline_covar, sep = "")
  cat(paste(strwrap(bas_var, 60), collapse="\n"))

  cat("\n")

  state_var <- paste("State covariates: ", s$stage_covar, sep = "")
  cat(paste(strwrap(state_var, 60), collapse="\n"))

  cat("\n")
  mean_utility <- mean(get_utility(x)$U, na.rm = TRUE)
  mean_utility <- round(mean_utility, digits = digits)

  mes <-  paste("Average utility: ", mean_utility, sep = "")
  if(get_element(s, "cens_indicator") == TRUE){
    mes <-  paste("Average complete case utility: ", mean_utility, sep = "")
  }
  cat(
   mes
  )
  cat("\n")
}

#' @rdname policy_data
#' @export
summary.policy_data <- function(object, probs=seq(0, 1, .25), ...) {
  K <- get_K(object)
  n <- get_n(object)
  stage <- event <- NULL  # R-check glob. var.
  action_set <- get_action_set(object)
  stage_data <- get_element(object, "stage_data")
  st <- stage_data[event == 0,][, c("stage", "A"), with = FALSE]
  st$A <- factor(st$A, levels = action_set)
  colnames(st) <- c("stage", "action")
  tab <- table(st)
  stable <- addmargins(tab, 2, FUN = list(n = sum))
  bc <- paste(object$colnames$baseline_names, collapse = ", ")
  sc <- paste(object$colnames$state_names, collapse = ", ")
  stagedist <- list()
  cens_indicator <-  any(get_element(object, "cens_indicator")[["indicator"]])
  cens_tab <- NULL
  if (!cens_indicator) {
  dt <- get_cum_rewards(object)
  if (!is.null(probs))
    for (i in seq_len(K+1)) {
      ss <- dt[stage==i]
      val <- stats::quantile(ss[["U"]], probs = probs, ...)
      stagedist <- append(stagedist, list(val))
    }
  } else {
    cens_st <- stage_data[event == 2,][, c("stage"), with = FALSE]
    colnames(cens_st) <- c("stage")
    cens_tab <- cens_st[ , list(n = .N), by = "stage"]
    cens_tab <- cens_tab[order(stage),]
    cens_tab <- as.data.frame(cens_tab)
  }
  res <- list(n=n, K=K, tab=stable, stagedist=stagedist,
              baseline_covar=bc, stage_covar=sc,
              cens_indicator = cens_indicator, cens_tab = cens_tab)
  class(res) <- "summary.policy_data"
  res
}

get_cum_rewards <- function(policy_data, policy=NULL) {
  K <- get_K(policy_data)
  n <- get_n(policy_data)
  A <- U <- id <- stage <- NULL  # R-check glob. var.
  dt <- policy_data$stage_data[, c("id", "stage", "event", "A", "U")]
  setkeyv(dt, c("id", "stage"))
  dt[, U:=cumsum(U), by=id]

  count <- 0
  policy_group <- pol_ind <- NULL # R-check glob. var.
  dt[, policy_group:=0]
  if (!is.null(policy)) {
    if (!is.list(policy)) policy <- list(policy)
    for (pol in policy) {
      count <- count+1
      d <- merge(dt, pol(policy_data), all.x = TRUE)
      d[, pol_ind := all(d == A, na.rm = TRUE), by = "id"]
      dt[d[["pol_ind"]] == TRUE,
         policy_group:= count,
         by=id]
    }
    dt <- subset(dt, policy_group>0)

    default_lab <- paste("policy_", 1:length(policy), sep = "")
    lab <- lapply(policy, function(pol) attributes(pol)[["name"]])
    for (j in seq_along(lab)){
      if(is.null(lab[[j]]))
        lab[[j]] <- default_lab[[j]]
    }
    lab <- unlist(lab)
    dt[,policy_group:=lab[policy_group]]
  }
  dt[, policy_group := as.character(policy_group)]
  dt[policy_group == "0", policy_group := "all"]
  return(dt)
}

#' Plot policy data for given policies
#'
#' @param x Object of class [policy_data]
#' @param policy An object or list of objects of class [policy]
#' @param which A subset of the numbers 1:2
#' \itemize{
#'  \item{1} Spaghetti plot of the cumulative rewards
#'  \item{2} Plot of the policy actions for a given stage
#' }
#' @param stage Stage number for plot 2
#' @param history_variables character vector of length 2 for plot 2
#' @param jitter numeric
#' @param ... Additional arguments
#' @examples
#' library("polle")
#' library("data.table")
#' setDTthreads(1)
#' d3 <- sim_multi_stage(2e2, seed = 1)
#' pd3 <- policy_data(data = d3$stage_data,
#'                    baseline_data = d3$baseline_data,
#'                    type = "long",
#'                    id = "id",
#'                    stage = "stage",
#'                    event = "event",
#'                    action = "A",
#'                    utility = "U")
#' pd3 <- partial(pd3, K = 3)
#'
#' # specifying two static policies:
#' p0 <- policy_def(c(1,1,0), name = "p0")
#' p1 <- policy_def(c(1,0,0), name = "p1")
#'
#' plot(pd3)
#' plot(pd3, policy = list(p0, p1))
#'
#' # learning and plotting a policy:
#'  pe3 <- policy_eval(pd3,
#'                     policy_learn = policy_learn(),
#'                     q_models = q_glm(formula = ~t + X + X_lead))
#' plot(pd3, list(get_policy(pe3), p0))
#'
#' # plotting the recommended actions at a specific stage:
#' plot(pd3, get_policy(pe3),
#'      which = 2,
#'      stage = 2,
#'      history_variables = c("t","X"))
#' @export
plot.policy_data <- function(x,
                             policy=NULL,
                             which = c(1),
                             stage = 1,
                             history_variables = NULL,
                             jitter=.05,
                             ...) {
  policy_data <- x
  if (!is.numeric(which) || any(which < 1) || any(which > 2))
    stop("'which' must be in 1:2")
  show <- rep(FALSE, 2)
  show[which] <- TRUE

  if (!is.null(stage)){
    if (!(is.numeric(stage) & (length(stage) == 1)))
      stop("stage must be an integer greater than 0.")
    if (!(stage %% 1 == 0))
      stop("stage must be an integer greater than 0.")
    if (stage<=0)
      stop("stage must be an integer greater than 0.")
  }
  stage_ <- stage

  if (show[1L]) {
    dt <- get_cum_rewards(policy_data, policy = policy)
    if (nrow(dt) == 0)
      stop("No observations comply with the given policy/policies.")
    lev <- unique(dt[["policy_group"]])
    id <- stage <- NULL  # R-check glob. var.
    dots <- list(...)
    pargs <- c("col","lty","pch")
    for (p in pargs) {
      if (is.null(dots[[p]])) {
        dots[[p]] <- seq_len(length(policy)+1)
      }
      dots[[p]] <- rep(dots[[p]], length.out=length(policy)+1)
    }
    ylab <- "Cumulative reward"
    xlab <- "Stage"
    legend <- "topleft"
    for (p in c("xlab","ylab","legend")) {
      if (!is.null(dots[[p]])) {
        assign(p, dots[[p]])
        dots[[p]] <- NULL
      }
    }
    plot(U ~ stage, data=dt, type="n", xlab="", ylab="", axes=FALSE)
    for (i in seq_along(lev)) {
      di <- subset(dt, dt$policy_group==lev[i])
      wide <- t(dcast(di,
                      id ~ stage, value.var="U")[,-1])
      args <- dots
      for (p in pargs)
        args[[p]] <- args[[p]][i]

      do.call(graphics::points, c(list(U ~ base::jitter(as.numeric(stage), jitter),
                                       data=di),
                                  args))
      do.call(graphics::matlines, c(list(wide), args))
    }
    graphics::title(xlab=xlab, ylab=ylab)
    if (length(lev)>0 && !is.null(legend)) {
      graphics::legend(legend, legend=lev,
                       col=dots$col,
                       pch=dots$pch,
                       lty=dots$lty)
    }
    graphics::box()
    graphics::axis(2)
    graphics::axis(1, at=1:get_K(x))
    grDevices::dev.flush()
  }
  if (show[2L]){
    if (is.null(policy_data) | !inherits(policy_data, "policy_data"))
      stop("For plot 2 please provide policy_data.")
    K <- get_K(policy_data)
    if (!(stage_ <= K))
      stop("stage must be lower than or equal to the maximal number of stages.")
    if (is.null(history_variables))
      stop("For plot 2 please provide history_variables (character vector of length 2).")
    if (!(is.character(history_variables) & length(history_variables)==2))
      stop("For plot 2 please provide history_variables (character vector of length 2).")

    plot_data <- data.table::merge.data.table(
      policy(policy_data)[stage ==stage_],
      get_history(policy_data)[["H"]][stage ==stage_]
    )

    if (!all(history_variables %in% colnames(plot_data)))
      stop("Invalid history_variables.")

    d <- as.factor(plot_data[["d"]])
    main <- paste("Plot of policy actions at stage ", stage_, sep = "")
    xx <- plot_data[[history_variables[1]]]
    if (is.character(xx))
      xx <- factor(xx, levels = sort(unique(xx)))
    yy <- plot_data[[history_variables[2]]]
    if (is.character(yy))
      yy <- factor(yy, levels = sort(unique(yy)))
    graphics::plot.default(
      x = xx,
      xlab = history_variables[1],
      xaxt = 'n',
      y = yy,
      yaxt = 'n',
      ylab = history_variables[2],
      main = main,
      col=d,
      ...)
    graphics::legend('topright', legend = levels(d), col = 1:8, cex = 0.8, pch = 1)
    if (is.factor(xx)){
      graphics::axis(1, at=(1:length(levels(xx))),labels=levels(xx))
    } else {
      graphics::axis(1)
    }
    if (is.factor(yy)){
      graphics::axis(2, at=(1:length(levels(yy))),labels=levels(yy))
    } else {
      graphics::axis(2)
    }
    grDevices::dev.flush()
  }

  invisible()
}

#' @export
print.summary.policy_data <- function(x, ...) {
  cat(paste("Policy data with n = ", x$n,
            " observations and maximal K = ",
            x$K, " stages.\n", sep = ""))
  cat("\n")
  print(x$tab)
  cat("\n")
  cat(paste("Baseline covariates: ", x$baseline_covar, sep = ""))
  cat("\n")
  cat(paste("Stage covariates: ", x$stage_covar, sep = ""))
  cat("\n")

  cat("\n")
  for (i in 1:(x$K+1)) {
    cat("Cumulative reward distribution at stage ",i,":\n", sep="")
    print(x$stagedist[[i]])
    cat("\n")
  }
}

#' Copy Policy Data Object
#'
#' Objects of class [policy_data] contains elements of
#' class [data.table::data.table].
#' \code{data.table} provide functions that operate on objects by reference.
#' Thus, the \code{policy_data} object is not copied when modified by reference,
#' see examples. An explicit copy can be made by \code{copy_policy_data}. The
#' function is a wrapper of [data.table::copy()].
#' @param object Object of class [policy_data].
#' @return Object of class [policy_data].
#' @export
#' @examples
#' library("polle")
#' ### Single stage case: Wide data
#' d1 <- sim_single_stage(5e2, seed=1)
#' head(d1, 5)
#' # constructing policy_data object:
#' pd1 <- policy_data(d1,
#'                    action="A",
#'                    covariates=c("Z", "B", "L"),
#'                    utility="U")
#' pd1
#'
#' # True copy
#' pd2 <- copy_policy_data(pd1)
#' # manipulating the data.table by reference:
#' pd2$baseline_data[, id := id + 1]
#' head(pd2$baseline_data$id - pd1$baseline_data$id)
#'
#' # False copy
#' pd2 <- pd1
#' # manipulating the data.table by reference:
#' pd2$baseline_data[, id := id + 1]
#' head(pd2$baseline_data$id - pd1$baseline_data$id)
copy_policy_data <- function(object){

  object$stage_data <- copy(object$stage_data)
  object$baseline_data <- copy(object$baseline_data)

  return(object)
}

partial_stage_data <- function(stage_data, K, deterministic_rewards){
  required_names <- c("id", "stage", "event", "A", "U")

  # filtering stage_data rows up till stage K:
  stage <- NULL
  stage_data_K <- stage_data[stage <= K, ]

  # filtering the residual stage data rows for stages above K:
  stage_data_res <- stage_data[stage > K, required_names, with = FALSE]

  # summarizing residual stage data as a single row:
  event <- U <- NULL
  stage_data_res_sum <- stage_data_res[
    ,
    list(
      stage = min(stage), # min(stage) is K + 1
      event = max(event), # max(event) is the event at stage K*, which is either 1 or 2
      U = sum(U) # sum(U) is the sum of the utility contributions from stage K+1 to K*
    ),
    by = "id"
  ]

  # updated action set:
  action_set <- unname(sort(unlist(unique(stage_data_K[,"A"]))))
  deterministic_rewards_K <- paste("U_A", action_set, sep = "")

  # removing excess deterministic reward variables:
  tmp <- setdiff(deterministic_rewards, deterministic_rewards_K)
  if (length(tmp) > 0)
    stage_data_K[, (tmp) := NULL]

  stage_data_res_sum[, (deterministic_rewards_K) := NA]
  # binding stage_data_K with stage_data_res_sum
  stage_data <- rbindlist(list(stage_data_K, stage_data_res_sum),
                          fill = TRUE,
                          use.names = TRUE)

  # setting keys and index
  setkeyv(stage_data, c("id", "stage"))
  setindexv(stage_data, "event")

  out <- list(
    stage_data = stage_data,
    action_set = action_set,
    deterministic_rewards = deterministic_rewards_K
  )

  return(out)
}

#' Trim Number of Stages
#'
#' \code{partial} creates a partial policy data object by trimming
#'  the maximum number of stages in the policy data object to a fixed
#'  given number.
#' @param object Object of class [policy_data].
#' @param K Maximum number of stages.
#' @return Object of class [policy_data].
#' @examples
#' library("polle")
#' ### Multiple stage case
#' d <- sim_multi_stage(5e2, seed = 1)
#' # constructing policy_data object:
#' pd <- policy_data(data = d$stage_data,
#'                    baseline_data = d$baseline_data,
#'                    type = "long",
#'                    id = "id",
#'                    stage = "stage",
#'                    event = "event",
#'                    action = "A",
#'                    utility = "U")
#' pd
#' # Creating a partial policy data object with 3 stages
#' pd3 <- partial(pd, K = 3)
#' pd3
#' @export
partial <- function(object, K)
  UseMethod("partial")

#' @export
partial.policy_data <- function(object, K){
  ## input checks:
  if (!is.numeric(K))
    stop("K must be an integer greater than or equal to 1.")
  if (!((K %% 1 == 0) & (K>0)))
    stop("K must be an integer greater than or equal to 1.")

  ## copy object to avoid reference issues in data.table
  object <- copy_policy_data(object)

  object_K <- get_K(object)
  if(K >= object_K)
    return(object)

  ## partial not implemented for right-censored data:
  cens_indicator <- get_element(object, "cens_indicator")
  if (any(get_col(cens_indicator, "indicator"))) {
    stop("partial.policy_data() not implemented for right-censored data.")
  }

  ## transforming the stage data:
  psd <- partial_stage_data(
    stage_data = object[["stage_data"]],
    K = K,
    deterministic_rewards = object[["colnames"]][["deterministic_rewards"]]
  )

  ## updating the object
  object[["stage_data"]] <- psd[["stage_data"]]
  object[["dim"]][["K"]] <- K
  object[["stage_action_sets"]] <- object[["stage_action_sets"]][1:K]
  object[["action_set"]] <- psd[["action_set"]]
  object[["colnames"]][["deterministic_rewards"]] <- psd[["deterministic_rewards"]]

  return(object)
}

#' Subset Policy Data on ID
#'
#' \code{subset_id} returns a policy data object containing the given IDs.
#' @param object Object of class [policy_data].
#' @param id character vectors of IDs.
#' @param preserve_action_set If TRUE, the action sets must be preserved.
#' @returns Object of class [policy_data].
#' @examples
#' library("polle")
#' ### Single stage:
#' d <- sim_single_stage(5e2, seed=1)
#' # constructing policy_data object:
#' pd <- policy_data(d, action="A", covariates=list("Z", "B", "L"), utility="U")
#' pd
#'
#' # getting the observation IDs:
#' get_id(pd)[1:10]
#'
#' # subsetting on IDs:
#' pdsub <- subset_id(pd, id = 250:500)
#' pdsub
#' get_id(pdsub)[1:10]
#' @export
subset_id <- function(object, id, preserve_action_set = TRUE)
  UseMethod("subset_id")

#' @export
subset_id.policy_data <- function(object, id, preserve_action_set = TRUE){
  if (!all(id %in% get_id(object)))
    stop("Invalid subset of IDs.")
  id_ <- id; rm(id)

  action_set <- NULL
  stage_action_sets <- NULL
  if (preserve_action_set == TRUE){
    action_set <- object$action_set
    stage_action_sets <- object$stage_action_sets
  }

  spd <- new_policy_data(
    stage_data = object$stage_data[id %in% id_],
    baseline_data = object$baseline_data[id %in% id_],
    action_set = action_set,
    stage_action_sets = stage_action_sets
  )

  return(spd)
}

#' Get the full history for a given stage
#'
#' @noRd
#' @param object Object of class [policy_data()].
#' @param stage stage number.
#' @param event_set Integer vector. Subset of c(0,1,2).
#' @return Object of class "history".
full_history <- function(object, stage, event_set){
  K <- get_K(object)
  object_colnames <- get_element(object, "colnames")
  stage_data <- get_element(object, "stage_data")
  state_names <- get_element(object_colnames, "state_names")
  baseline_data <- get_element(object, "baseline_data")
  baseline_names <- get_element(object_colnames, "baseline_names")
  action_set <- get_action_set(object)
  stage_action_sets <- get_stage_action_sets(object)
  deterministic_rewards <- get_element(object_colnames, "deterministic_rewards")
  stage_ <- stage
  rm(stage)

  if (stage_ > (K+1)) {
    stop("The stage number must be lower or equal to K+1.")
  }

  stage_action_set <- NULL
  if (stage_ <= K) {
    stage_action_set <- stage_action_sets[[stage_]]
  }

  ## default values
  U_bar <- event <- id <- NULL
  ## getting the history at the given stage:
  AH_names <- c("id", "stage", "A", state_names)
  ## filtering rows up till the given stage number:
  stage <- NULL
  AH <- stage_data[event %in% event_set][stage <= stage_, AH_names, with = FALSE]
  ## filtering observations with an action at the given stage:
  AH <- AH[, if(any(stage == stage_)) .SD, id]
  ## checking if empty:
  if (nrow(AH) == 0) {
    stop("empty history for the given stage and event_set.")
  }

  ## transforming the data from long to wide format:
  AH <- dcast(AH, id ~ stage, value.var = AH_names[-c(1, 2)])
  if (length(AH_names[-c(1, 2)]) == 1) {
    colnames(AH)[-1] <- paste(AH_names[-c(1, 2)], colnames(AH)[-1], sep = "_")
  }

  ## inserting stage column:
  AH[, stage := stage_]
  ## merging the stage specific histories and the the baseline data by reference:
  if(length(baseline_names) > 0){
    AH[baseline_data, (baseline_names) := mget(paste0('i.', baseline_names))]
  }
  ## setting key and column order
  setkeyv(AH, c("id", "stage"))
  setcolorder(AH, neworder = c("id", "stage"))

  ## getting the actions and history:
  action_name <- paste("A", stage_, sep = "_")
  A_names <- c("id", "stage", action_name)
  A <- AH[ , A_names, with = FALSE]
  H <- AH[, c(action_name) := NULL]

  ## getting the deterministic utility contributions:
  U <- stage_data[stage <= stage_][, U_bar := sum(U * (event != 1), na.rm = TRUE), id]
  U <- U[event %in% event_set][stage == stage_,]
  U_names <- c("id", "stage", "U_bar", deterministic_rewards)
  U <- U[ , U_names, with = FALSE]

  ## getting the event at the given stage:
  event <-  stage_data[event %in% event_set][stage == stage_, c("id", "stage", "event"), with = FALSE]

  ## getting the time at the given stage:
  time <- stage_data[event %in% event_set][stage == stage_, c("id", "stage", "time", "time2"), with = FALSE]

  history <- list(
    H = H,
    A = A,
    event = event,
    time = time,
    U = U,
    action_name = action_name,
    deterministic_rewards = deterministic_rewards,
    action_set = action_set,
    stage_action_set = stage_action_set,
    stage = stage_
  )
  history <- remove_null_elements(history)
  class(history) <- "history"

  return(history)
}

#' Get the state/Markov-type history for a given stage
#'
#' @noRd
#' @param object object of class [policy_data()].
#' @param stage stage number.
#' @param event_set Integer vector. Subset of c(0,1,2).
#' @return Object of class "history".
stage_state_history <- function(object, stage, event_set){
  K <- get_K(object)
  stage_data <- get_element(object, "stage_data")
  state_names <- get_element(object, "colnames") |> get_element("state_names")
  baseline_data <- get_element(object, "baseline_data")
  baseline_names <- get_element(object,"colnames") |> get_element("baseline_names")
  action_set <- get_action_set(object)
  stage_action_sets <- get_stage_action_sets(object)
  deterministic_rewards <- get_element(object, "colnames") |>
    get_element("deterministic_rewards")
  stage_ <- stage; rm(stage)

  if (stage_ > (K+1)) {
    stop("The stage number must be lower or equal to K+1.")
  }

  stage_action_set <- NULL
  if (stage_ <= K) {
    stage_action_set <- stage_action_sets[[stage_]]
  }
  action_name <- "A"
  ## default values:
  U_bar <- event <- id <- NULL
  ## getting the state names:
  AH_names <- c("id", "stage", "A", state_names)
  ## filtering observations with the given events at the given stage:
  AH <- stage_data[event %in% event_set][stage == stage_, AH_names, with = FALSE]
  if (nrow(AH) == 0) {
    stop("empty history for the given stage and event_set.")
  }
  ## merging the state history and the the baseline data:
  if(length(baseline_names) > 0){
    AH[baseline_data, (baseline_names) := mget(paste0('i.', baseline_names))]
  }
  ## constructing history H and action A at the given stage:
  A_names <- c("id", "stage", "A")
  A <- AH[,  A_names, with = FALSE]
  H <- AH[, c("A") := NULL]

  ## summing the past utility contributions execept for terminal events:
  U <- stage_data[stage <= stage_][, U_bar := sum(U * (event != 1), na.rm = TRUE), by = "id"]
  U <- U[event %in% event_set][stage == stage_,]
  U_names <- c("id", "stage", "U_bar", deterministic_rewards)
  U <- U[, U_names, with = FALSE]
  ## getting the event at the given stage:
  event <-  stage_data[event %in% event_set][stage == stage_, c("id", "stage", "event"), with = FALSE]
  ## getting the time at the given stage:
  time <- stage_data[event %in% event_set][stage == stage_, c("id", "stage", "time", "time2"), with = FALSE]

  history <- list(
    H = H,
    A = A,
    event = event,
    time = time,
    U = U,
    action_name = action_name,
    deterministic_rewards = deterministic_rewards,
    action_set = action_set,
    stage_action_set = stage_action_set,
    stage = stage_
  )
  history <- remove_null_elements(history)
  class(history) <- "history"

  return(history)
}

#' Get the state/Markov-type history for all stages
#'
#' @noRd
#' @param object object of class [policy_data()].
#' @param event_set Integer vector. Subset of c(0,1,2).
#' @return Object of class "history".
state_history <- function(object, event_set){
  stage_data <- get_element(object, "stage_data")
  state_names <- get_element(object, "colnames") |> get_element("state_names")
  baseline_data <- get_element(object, "baseline_data")
  baseline_names <- get_element(object, "colnames") |> get_element("baseline_names")
  action_set <- get_element(object, "action_set")

  action_name <- "A"
  ## default values:
  event <- NULL
  ## constructing history H and action A at the given stage:
  AH_names <- c("id", "stage", "A", state_names)
  AH <- stage_data[event %in% event_set][, AH_names, with = FALSE]
  ## merging the stage specific histories and the the baseline data by reference:
  if(length(baseline_names) > 0){
    AH[baseline_data, (baseline_names) := mget(paste0('i.', baseline_names))]
  }
  A_names <- c("id", "stage", "A")
  A <- AH[ , A_names, with = FALSE]
  H <- AH[, c("A") := NULL]

  ## getting the event at the given stage:
  event <-  stage_data[event %in% event_set][ , c("id", "stage", "event"), with = FALSE]

  ## getting the time at the given stage:
  time <- stage_data[event %in% event_set][ , c("id", "stage", "time", "time2"), with = FALSE]

  history <- list(
    H = H,
    A = A,
    event = event,
    time = time,
    action_name = action_name,
    action_set = action_set
  )
  history <- remove_null_elements(history)
  class(history) <- "history"

  return(history)
}

get_stage_data <- function(object)
  UseMethod("get_stage_data")

#' @export
get_stage_data.policy_data <- function(object){
  out <- getElement(object, "stage_data")
  return(out)
}

#' @name history
#' @rdname get_history
NULL

#' Get History Object
#'
#' \code{get_history} summarizes the history and action at a given stage from a
#' [policy_data] object.
#' @param object Object of class [policy_data].
#' @param stage Stage number. If NULL, the state/Markov-type history across
#' all stages is returned.
#' @param full_history Logical. If TRUE, the full history is returned
#' If FALSE, only the state/Markov-type history is returned.
#' @param event_set Integer vector. Subset of c(0,1,2).
#' @details
#' Each observation has the sequential form
#' \deqn{O= {B, U_1, X_1, A_1, ..., U_K, X_K, A_K, U_{K+1}},}
#' for a possibly stochastic number of stages K.
#' \itemize{
#'  \item{} \eqn{B} is a vector of baseline covariates.
#'  \item{} \eqn{U_k} is the reward at stage k (not influenced by the action \eqn{A_k}).
#'  \item{} \eqn{X_k} is a vector of state covariates summarizing the state at stage k.
#'  \item{} \eqn{A_k} is the categorical action at stage k.
#' }
#' @returns Object of class [history]. The object is a list
#' containing the following elements:
#' \item{\code{H}}{[data.table::data.table] with keys id and stage and with variables
#'                 \{\eqn{B}, \eqn{X_k}\} (state history) or
#'                 \{\eqn{B}, \eqn{X_1}, \eqn{A_1}, ..., \eqn{X_k}\}
#'                 (full history), see details.}
#' \item{\code{A}}{[data.table::data.table] with keys id and stage and variable \eqn{A_k}, see
#'                 details.}
#' \item{action_name}{Name of the action variable in \code{A}.}
#' \item{action_set}{Sorted character vector defining the action set.}
#' \item{U}{(If \code{stage} is not NULL) [data.table::data.table] with keys id and stage
#'          and with variables U_bar and U_Aa for every a in the actions set.
#'          U_bar is the accumulated rewards up till and including the given
#'          stage, i.e., \eqn{\sum_{j=1}^k U_j}. U_Aa is the deterministic
#'          reward of action a.}
#' @examples
#' library("polle")
#' ### Single stage:
#' d1 <- sim_single_stage(5e2, seed=1)
#' # constructing policy_data object:
#' pd1 <- policy_data(d1, action="A", covariates=list("Z", "B", "L"), utility="U")
#' pd1
#'
#' # In the single stage case, set stage = NULL
#' h1 <- get_history(pd1)
#' head(h1$H)
#' head(h1$A)
#'
#' ### Two stages:
#' d2 <- sim_two_stage(5e2, seed=1)
#' # constructing policy_data object:
#' pd2 <- policy_data(d2,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd2
#' # getting the state/Markov-type history across all stages:
#' h2 <- get_history(pd2)
#' head(h2$H)
#' head(h2$A)
#'
#' # getting the full history at stage 2:
#' h2 <- get_history(pd2, stage = 2, full_history = TRUE)
#' head(h2$H)
#' head(h2$A)
#' head(h2$U)
#'
#' # getting the state/Markov-type history at stage 2:
#' h2 <- get_history(pd2, stage = 2, full_history = FALSE)
#' head(h2$H)
#' head(h2$A)
#'
#' ### Multiple stages
#' d3 <- sim_multi_stage(5e2, seed = 1)
#' # constructing policy_data object:
#' pd3 <- policy_data(data = d3$stage_data,
#'                    baseline_data = d3$baseline_data,
#'                    type = "long",
#'                    id = "id",
#'                    stage = "stage",
#'                    event = "event",
#'                    action = "A",
#'                    utility = "U")
#' pd3
#'
#' # getting the full history at stage 2:
#' h3 <- get_history(pd3, stage = 2, full_history = TRUE)
#' head(h3$H)
#' # note that not all observations have two stages:
#' nrow(h3$H) # number of observations with two stages.
#' get_n(pd3) # number of observations in total.
#' @export
get_history <- function(object, stage = NULL, full_history = FALSE, event_set = c(0))
  UseMethod("get_history")

#' @export
get_history.policy_data <- function(object,
                                    stage = NULL,
                                    full_history = FALSE,
                                    event_set = c(0)) {
  # input checks:
  if (!is.logical(full_history) | (length(full_history) != 1))
    stop("full_history must be TRUE or FALSE")
  if (!is.null(stage)){
    if (!(is.numeric(stage) & (length(stage) == 1)))
      stop("stage must be an integer greater than 0.")
    if (!(stage %% 1 == 0))
      stop("stage must be an integer greater than 0.")
    if (stage<=0)
      stop("stage must be an integer greater than 0.")
  }
  mes <- "event_set must be a subset of c(0,1,2)."
  if (!is.vector(event_set)) {
    stop(mes)
  }
  if (!all(event_set %in% c(0,1,2))) {
    stop(mes)
  }

  if (full_history == TRUE){
    if (is.null(stage)) stop("Please provide a stage number.")
    his <- full_history(object, stage = stage, event_set = event_set)
  } else{
    if (is.null(stage)){
      his <- state_history(object, event_set = event_set)
    } else{
      his <- stage_state_history(object, stage = stage, event_set = event_set)
    }
  }
  return(his)
}

#' Get history variable names
#'
#' \code{get_history_names()} returns the state covariate names of the history data
#' table for a given stage. The function is useful when specifying
#' the design matrix for [g_model] and [q_model] objects.
#' @param object Policy data object created by [policy_data()].
#' @param stage Stage number. If NULL, the state/Markov-type history variable
#' names are returned.
#' @return Character vector.
#' @examples
#' library("polle")
#' ### Multiple stages:
#' d3 <- sim_multi_stage(5e2, seed = 1)
#' pd3 <- policy_data(data = d3$stage_data,
#'                    baseline_data = d3$baseline_data,
#'                    type = "long",
#'                    id = "id",
#'                    stage = "stage",
#'                    event = "event",
#'                    action = "A",
#'                    utility = "U")
#' pd3
#' # state/Markov type history variable names (H):
#' get_history_names(pd3)
#' # full history variable names (H_k) at stage 2:
#' get_history_names(pd3, stage = 2)
#' @export
get_history_names <- function(object, stage)
  UseMethod("get_history_names")

#' @export
get_history_names.policy_data <- function(object, stage = NULL){
  if (is.null(stage)){
    history <- get_history(object, full_history = FALSE)
  } else{
    history <- get_history(object, stage = stage, full_history = TRUE)
  }
  H <- getElement(history, "H")
  stopifnot(!is.null(H))
  history_names <- names(H)[!(names(H) %in% c("id", "stage"))]
  return(history_names)
}


#' Get the Utility
#'
#' \code{get_utility()} returns the utility, i.e., the sum of the rewards,
#' for every observation in the policy data object.
#'
#' @param object Object of class [policy_data].
#' @returns [data.table::data.table] with key id and numeric variable U.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' # constructing policy_data object:
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # getting the utility:
#' head(get_utility(pd))
#'@export
get_utility <- function(object)
  UseMethod("get_utility")

#' @export
get_utility.policy_data <- function(object){
  stage_data <- get_stage_data(object)
  U <- stage_data[, list(U = sum(get("U")), event2 = any(get("event") == 2)), by = "id"]
  U[get("event2") == TRUE, U := NA]
  set(U, j = "event2", value = NULL)
  return(U)
}

get_rewards <- function(object){
  stage_data <- get_stage_data(object)
  R <- stage_data[, c("id", "stage", "U"), with = FALSE]
  return(R)
}

#' Get Actions
#'
#' \code{get_actions} returns the actions at every stage for every observation
#' in the policy data object.
#' @param object Object of class [policy_data].
#' @returns [data.table::data.table] with keys id and stage and character variable A.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' # constructing policy_data object:
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # getting the actions:
#' head(get_actions(pd))
#' @export
get_actions <- function(object)
  UseMethod("get_actions")

#' @export
get_actions.policy_data <- function(object){
  stage_data <- get_stage_data(object)
  event <- NULL
  actions <- stage_data[event == 0, c("id", "stage", "A"), with = FALSE]
  return(actions)
}

#' Get IDs
#'
#' \code{get_id} returns the ID for every observation in the policy data object.
#' @param object Object of class [policy_data] or [history].
#' @returns Character vector.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' # constructing policy_data object:
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # getting the IDs:
#' head(get_id(pd))
#' @export
get_id <- function(object)
  UseMethod("get_id")

#' @export
get_id.policy_data <- function(object){
  id <- unique(object$stage_data$id)
  return(id)
}

#' Get IDs and Stages
#'
#' \code{get_id_stage} returns the ID and stage number differnt types of events.
#' @param object Object of class [policy_data] or [history].
#' @param event_set Integer vector. Subset of c(0,1,2).
#' @returns [data.table::data.table] with keys id and stage.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' # constructing policy_data object:
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # getting the IDs and stages:
#' head(get_id_stage(pd))
#' @export
get_id_stage <- function(object, event_set = c(0))
  UseMethod("get_id_stage")

#' @export
get_id_stage.policy_data <- function(object, event_set = c(0)){

  mes <- "event_set must be a subset of c(0,1,2)."
  if (!is.vector(event_set)) {
    stop(mes)
  }
  if (!all(event_set %in% c(0,1,2))) {
    stop(mes)
  }

  stage_data <- get_stage_data(object)
  id_stage_names <- c("id", "stage")
  id_stage <- stage_data[get("event") %in% event_set, ][, id_stage_names, with = FALSE]

  return(id_stage)
}

#' Get Maximal Stages
#'
#' \code{get_K} returns the maximal number of stages for the observations in
#' the policy data object.
#' @param object Object of class [policy_data].
#' @returns Integer.
#' @examples
#' d <- sim_multi_stage(5e2, seed = 1)
#' pd <- policy_data(data = d$stage_data,
#'                    baseline_data = d$baseline_data,
#'                    type = "long",
#'                    id = "id",
#'                    stage = "stage",
#'                    event = "event",
#'                    action = "A",
#'                    utility = "U")
#' pd
#' # getting the maximal number of stages:
#' get_K(pd)
#' @export
get_K <- function(object)
  UseMethod("get_K")

#' @export
get_K.policy_data <- function(object){
  K <- object$dim$K
  return(K)
}

#' Get Number of Observations
#'
#' \code{get_n} returns the number of observations in
#' the policy data object.
#' @param object Object of class [policy_data].
#' @returns Integer.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' # constructing policy_data object:
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # getting the number of observations:
#' get_n(pd)
#' @export
get_n <- function(object)
  UseMethod("get_n")

#' @export
get_n.policy_data <- function(object){
  n <- object$dim$n
  return(n)
}

#' Get Action Set
#'
#' \code{get_action_set} returns the action set, i.e., the possible
#' actions at each stage for the policy data object.
#' @param object Object of class [policy_data].
#' @returns Character vector.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage(5e2, seed=1)
#' # constructing policy_data object:
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # getting the actions set:
#' get_action_set(pd)
#' @export
get_action_set <- function(object)
  UseMethod("get_action_set")

#' @export
get_action_set.policy_data <- function(object){
  action_set <- object$action_set
  return(action_set)
}

#' Get Stage Action Sets
#'
#' \code{get_stage_action_sets} returns the action sets at each stage, i.e.,
#' the possible actions at each stage for the policy data object.
#' @param object Object of class [policy_data].
#' @returns List of character vectors.
#' @examples
#' ### Two stages:
#' d <- sim_two_stage_multi_actions(5e2, seed=1)
#' # constructing policy_data object:
#' pd <- policy_data(d,
#'                   action = c("A_1", "A_2"),
#'                   baseline = c("B"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd
#'
#' # getting the stage actions set:
#' get_stage_action_sets(pd)
#' @export
get_stage_action_sets <- function(object)
  UseMethod("get_stage_action_sets")

#' @export
get_stage_action_sets.policy_data <- function(object){
  stage_action_sets <- getElement(object, "stage_action_sets")
  return(stage_action_sets)
}

#' Get event indicators
#'
#' \code{get_event} returns the event indicators for given IDs
#' stages
#' @param object Object of class [policy_data] or [history].
#' @returns [data.table::data.table] with keys id and stage.
#' @export
get_event <- function(object)
  UseMethod("get_event")

#' @export
get_event.policy_data <- function(object){
  stage_data <- get_stage_data(object)
  events <- stage_data[, c("id", "stage", "event"), with = FALSE]
  return(events)
}

test_that("check_data fails if not given a data.table with unique variables.", {
  data <- c("A", "B")
  expect_error(check_data(data), "'data' must be a data.table.")
  baseline_data <- c("A", "B")
  expect_error(check_data(baseline_data), "'baseline_data' must be a data.table.")

  data <- data.table(id = 1:10, id = 10:1)
  expect_error(check_data(data), "'data' has duplicated variable names.")
})

test_that("policy_data checks inputs",{
  d <- sim_single_stage(10, seed=1)

  expect_error(
    policy_data(d, action = 1),
    "'action' must be a vector or a list of type character."
  )
  expect_error(
    policy_data(d, action = TRUE),
    "'action' must be a vector or a list of type character."
  )
  expect_error(
    policy_data(d, action = "a"),
    "Action variables a not found in data."
  )
  expect_error(
    policy_data(d, action = c("A", "a")),
    "Action variables a not found in data."
  )
  expect_error(
    policy_data(d, action = c("a1", "a")),
    "Action variables a1,a not found in data."
  )

  expect_error(
    policy_data(d, action = c("A"), covariates = 1),
    "'covariates' must be a character vector or a list of character vectors."
  )

  expect_error(
    policy_data(d,
                action = c("A"),
                covariates = list(X="a",Y="b"),
                utility = "U"),
    "Variables not found in data: \"a\", \"b\"."
  )
  expect_error(
    policy_data(d,
                action = c("A"),
                covariates = "a",
                utility = "U"),
    "Variables not found in data: \"a\"."
  )
  expect_error(
    policy_data(d,
                action = c("A"),
                covariates = c("Z"),
                utility = "U"),
    NA
  )
  expect_error(
    policy_data(c(1,2,3),
                action = c("A"),
                covariates = c("Z"),
                utility = "U"),
    "'data' must be a data.table."
  )
  expect_error(
    policy_data(d,
                type = "test",
                action = c("A"),
                covariates = c("Z"),
                utility = "U"),
    "'type' must be either \"wide\" or \"long\"."
  )
  expect_error(
    suppressWarnings({
      policy_data(d,
                  type = "long",
                  action = c("A"),
                  covariates = c("Z"),
                  utility = "U")
    })
  )
})

test_that("policy_data() checks id for type = 'wide'", {
  ## long data
  ## 3 cases: no right censoring, right censored before/at stage 1 action,
  ## right censored utility outcome (at stage 2)
  ld <- data.table(
    id = c(1,1,2,3,3),
    stage = c(1,2,1,1,2),
    event = c(0,1,2,0,2),
    A = c("0", NA, NA, "1", NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3"),
    Z = c("A", NA, "A", "B", NA),
    L = c(1, 2, 2, 1, 3),
    time = c(1, 2, 0.5, 1, 1.5),
    U = c(0, 10, NA, 0, NA),
    U_A0 = c(0,0,NA,0,NA),
    U_A1 = c(0,0,NA,0,NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  expect_error(
    policy_data(data = ld,
                action = "A",
                utility = "U",
                covariates = c("Z"),
                id = "id",
                type = "wide"),
    "'id' column in data must be unique."
  )
})

test_that("policy_data handles varying actions set",{
  d <- sim_two_stage_multi_actions(n = 1e4)
  expect_error(
    pd <- policy_data(data = d,
                      action = c("A_1", "A_2"),
                      baseline = c("B", "BB"),
                      covariates = list(L = c("L_1", "L_2"),
                                        C = c("C_1", "C_2")),
                      utility = c("U_1", "U_2", "U_3")),
    NA
  )

  expect_equal(
    pd$action_set,
    sort(c("yes", "no", "default"))
  )

  sas <- list(c("no", "yes"), c("default","no", "yes"))
  names(sas) <- c("stage_1", "stage_2")
  expect_equal(
    pd$stage_action_sets,
    sas
  )

  expect_equal(
    partial(pd, K=2)$stage_action_sets,
    sas
  )

  expect_equal(
    partial(pd, K=1)$stage_action_sets,
    sas[1]
  )

  expect_equal(
    get_stage_action_sets(subset_id(pd, id = 1:2)),
    sas
  )

})

test_that("policy_data() handles varying sets of/missing covariates in a given stage",{
  n <- 20
  set.seed(1)
  library("polle")
  library("data.table")
  d <- data.table(id = 1:n,
                  Y_1 = rnorm(n),
                  X_1 = rnorm(n),
                  Z_1 = rnorm(n),
                  W_1 = rep("test", n),
                  A_1 = rbinom(n = n, size = 1, prob = .5),
                  X_2 = rnorm(n),
                  Z_2 = rnorm(n),
                  Y_2 = rnorm(n),
                  A_2 = rbinom(n = n, size = 1, prob = .5),
                  U = rnorm(n))

  expect_error(
    pd <- policy_data(data = d,
                      id = "id",
                      action = c("A_1"),
                      covariates = c("X_1", NA),
                      utility = "U"),
    "covariate NA is invalid."
  )

  pd <- policy_data(data = d,
                    id = "id",
                    action = c("A_1", "A_2"),
                    covariates = list(X = c("X_1", "X_2"),
                                      Y = c(NA, "Y_2")),
                    utility = "U")
  expect_equal(
    polle:::get_stage_data.policy_data(pd)$Y,
    unlist(lapply(d$Y_2, function(x) c(NA, x, NA)))
  )

  pd <- policy_data(data = d,
                    id = "id",
                    action = c("A_1", "A_2"),
                    covariates = list(X = c("X_1", "X_2"),
                                      Y = c("Y_1", NA)),
                    utility = "U")
  expect_equal(
    polle:::get_stage_data.policy_data(pd)$Y,
    unlist(lapply(d$Y_1, function(x) c(x, NA, NA)))
  )

  pd <- policy_data(data = d,
                    id = "id",
                    action = c("A_1", "A_2"),
                    covariates = list(X = c("X_1", "X_2"),
                                      W = c("W_1", NA)),
                    utility = "U")
  expect_equal(
    polle:::get_stage_data.policy_data(pd)$W,
    unlist(lapply(d$W_1, function(x) c(x, NA, NA)))
  )

  expect_error(
    policy_data(data = d,
                id = "id",
                action = c("A_1", "A_2"),
                covariates = list(X = c("X_1", "X_2"),
                                  W = c(NA, NA)),
                utility = "U"),
    "'covariates' must be a character vector or a list of character vectors."
  )

  pd <- policy_data(data = d,
                    id = "id",
                    action = c("A_1", "A_2"),
                    covariates = list(X = c("X_1", "X_2"),
                                      Y = c(NA, "Y_2"),
                                      W = c("W_1", NA)),
                    utility = "U")
  expect_equal(
    polle:::get_stage_data.policy_data(pd)$W,
    unlist(lapply(d$W_1, function(x) c(x, NA, NA)))
  )
  expect_equal(
    polle:::get_stage_data.policy_data(pd)$Y,
    unlist(lapply(d$Y_2, function(x) c(NA, x, NA)))
  )

})

test_that("policy_data() melts wide data correctly in a two stage case.", {
  wide_data <- data.table(
    B = c("gr1", "gr2"),
    Z_1 = c("A", "B"),
    L_1 = c(1,2),
    Z_2 = c("C", "D"),
    L_2 = c(3,4),
    treat_1 = c(0,1),
    treat_2 = c(1,0),
    outcome = c(10, 5)
  )

  target_stage_data <- data.table(
    id = c(1,1,1,2,2,2),
    stage = c(1,2,3,1,2,3),
    event = c(0,0,1,0,0,1),
    time = as.numeric(c(NA,NA,NA,NA,NA,NA)),
    time2 = as.numeric(c(NA,NA,NA,NA,NA,NA)),
    A = as.character(c(0,1,NA,1,0,NA)),
    Z = c("A","C",NA,"B","D",NA),
    L = c(1,3,NA,2,4,NA),
    U = c(0,0,10,0,0,5),
    U_A0 = c(0,0,NA,0,0,NA),
    U_A1 = c(0,0,NA,0,0,NA)
  )
  setkey(target_stage_data, id, stage)
  setindex(target_stage_data, event)

  pd <- policy_data(
    data = wide_data,
    action = c("treat_1", "treat_2"),
    covariates = list(
      Z = c("Z_1", "Z_2"),
      L = c("L_1", "L_2")
    ),
    baseline = "B",
    utility = "outcome"
  )
  setindex(pd$stage_data, event)
  expect_equal(pd$stage_data, target_stage_data, check.attributes = TRUE)

  # including deterministic rewards:
  wide_data <- data.table(
    B = c("gr1", "gr2"),
    Z_1 = c("A", "B"),
    L_1 = c(1,2),
    Z_2 = c("C", "D"),
    L_2 = c(3,4),
    treat_1 = c(0,1),
    treat_2 = c(1,0),
    outcome_1 = c(10, 5),
    outcome_2 = c(7, 3),
    outcome_3 = c(-2, 1),
    outcome_1_A0 = c(1.5, 2.5),
    outcome_2_A0 = c(3.5, 4.5)
  )

  target_stage_data <- data.table(
    id = c(1,1,1,2,2,2),
    stage = c(1,2,3,1,2,3),
    event = c(0,0,1,0,0,1),
    time = rep(as.numeric(NA), 6),
    time2 = rep(as.numeric(NA), 6),
    A = as.character(c(0,1,NA,1,0,NA)),
    Z = c("A","C",NA,"B","D",NA),
    L = c(1,3,NA,2,4,NA),
    U = c(10,7,-2,5,3,1),
    U_A0 = c(1.5,3.5,NA,2.5,4.5,NA),
    U_A1 = c(0,0,NA,0,0,NA)
  )
  setkey(target_stage_data, id, stage)
  setindex(target_stage_data, event)

  pd <- policy_data(
    data = wide_data,
    action = c("treat_1", "treat_2"),
    covariates = list(
      Z = c("Z_1", "Z_2"),
      L = c("L_1", "L_2")
    ),
    baseline = "B",
    utility = c("outcome_1", "outcome_2", "outcome_3"),
    deterministic_rewards = list(U_A0 = c("outcome_1_A0", "outcome_2_A0"))
  )
  setindex(pd$stage_data, event)
  expect_equal(pd$stage_data, target_stage_data, check.attributes = TRUE)

  # invalid inputs:
  args <- list(
    data = wide_data,
    action = c("treat_1", "treat_2"),
    covariates = list(
      Z = c("Z_1", "Z_2"),
      L = c("L_1", "L_2")
    ),
    baseline = "B",
    utility = c("outcome_1", "outcome_2", "outcome_3"),
    deterministic_rewards = list(U_A0 = c("outcome_1_A0", "outcome_2_A0"))
  )

  args_copy <- args
  args_copy$covariates$Z <- c("Z_1", "Z_1")
  expect_error(do.call(what = "policy_data", args_copy), "Duplicated variables: \"Z_1\".")

  args_copy <- args
  args_copy$covariates$Z <- c("Z_1")
  expect_error(do.call(what = "policy_data", args_copy), "Each element in 'covariates' must have length 2.")
})

test_that("policy_data() melts wide data correctly in a single stage case.", {

  wide_data <- data.table(
    B = c("gr1", "gr2"),
    Z = c("A", "B"),
    L = c(1, 2),
    treat = c(0,1),
    outcome = c(10, 5)
  )

  target_stage_data <- data.table(
    id = c(1,1,2,2),
    stage = c(1,2,1,2),
    event = c(0,1,0,1),
    time = rep(as.numeric(NA), 4),
    time2 = rep(as.numeric(NA), 4),
    A = c("0", NA, "1", NA),
    B = c("gr1", NA, "gr2", NA),
    Z = c("A", NA, "B", NA),
    L = c(1, NA, 2, NA),
    U = c(0, 10, 0, 5),
    U_A0 = c(0,NA,0,NA),
    U_A1 = c(0,NA,0,NA)
  )
  setkey(target_stage_data, id, stage)
  setindex(target_stage_data, event)

  ## except equal:
  pd <- policy_data(data = wide_data,
                    action = "treat",
                    covariates = c("B", "Z", "L"),
                    utility = "outcome")
  setindex(pd$stage_data, event)
  expect_equal(pd$stage_data, target_stage_data, check.attributes = TRUE)
  rm(pd)

  ## duplicate variable names:
  expect_error(
    policy_data(data = wide_data,
                action = "treat",
                covariates = c("B", "Z", "Z"),
                utility = "outcome"),
    "Duplicated variables: \"Z\"."
  )
  expect_error(
    policy_data(data = wide_data,
                action = "treat",
                covariates = c("B", "Z", "treat"),
                utility = "outcome"),
    "Duplicated variables: \"treat\"."
  )
  expect_error(
    policy_data(data = wide_data,
                action = "treat",
                covariates = c("B", "Z", "L"),
                utility = "treat"),
    "Duplicated variables: \"treat\"."
  )

  # invalid variable names:
  wide_data_copy <- copy(wide_data)
  wide_data_copy$stage <- 1
  wide_data_copy$event <- 0
  expect_error(
    policy_data(data = wide_data_copy,
                action = "treat",
                covariates = c("B", "Z", "stage"),
                utility = "outcome"),
    "'covariates' can not have named elements \"event\", \"stage\", \"time\", or \"time2\"."
  )
  expect_error(
    policy_data(data = wide_data_copy,
                action = "treat",
                covariates = c("B", "Z", "event"),
                utility = "outcome"),
    "'covariates' can not have named elements \"event\", \"stage\", \"time\", or \"time2\"."
  )

  ## id input
  wide_data_copy <- copy(wide_data)
  wide_data_copy$id <- c(1,2)
  expect_error(
    policy_data(data = wide_data_copy,
                action = "treat",
                covariates = c("B", "Z", "L"),
                utility = "outcome"),
    "'data' has a variable id, but 'id' = NULL. Please set 'id' = \"id\" or change the name of the id variable."
  )
  pd <- policy_data(data = wide_data_copy,
                    action = "treat",
                    covariates = c("B", "Z", "L"),
                    utility = "outcome",
                    id = "id")
  setindex(pd$stage_data, event)
  expect_equal(
    pd$stage_data,
    target_stage_data,
    check.attributes = TRUE
  )
  wide_data_copy$id <- NULL
  wide_data_copy$ID <- c(1,2)

  pd <- policy_data(data = wide_data_copy,
                action = "treat",
                covariates = c("B", "Z", "L"),
                utility = "outcome",
                id = "ID")
  setindex(pd$stage_data, event)
  expect_equal(
    pd$stage_data,
    target_stage_data,
    check.attributes = TRUE
  )

  pd <- policy_data(data = wide_data,
                    action = "treat",
                    covariates = c("B", "Z", "L"),
                    utility = "outcome")
  expect_equal(
    pd$colnames$state_names,
    c("B", "Z", "L")
  )

  ## expected baseline
  target_baseline <- data.table(id = c(1,2), B = c("gr1", "gr2"))
  setkey(target_baseline, id)
  pd <- policy_data(data = wide_data,
                    action = "treat",
                    covariates = c("Z", "L"),
                    utility = "outcome",
                    baseline = "B")
  expect_equal(
    pd$baseline_data,
    pd$baseline_data
  )
  expect_equal(
    pd$colnames$baseline_names,
    c("B")
  )
  rm(pd)

  ## deterministic rewards
  wide_data_copy <- copy(wide_data)
  wide_data_copy$reward_A0 <- 2
  wide_data_copy$reward_A1 <- 1
  expect_error(
    policy_data(data = wide_data_copy,
                action = "treat",
                covariates = c("B", "Z", "L"),
                utility = "outcome",
                deterministic_rewards = list("reward_A0", "reward_A1")),
    "'deterministic_rewards' must be a named list with names in the set 'U_A0', 'U_A1'."
  )
  expect_error(
    policy_data(data = wide_data_copy,
                action = "treat",
                covariates = c("B", "Z", "L"),
                utility = "outcome", deterministic_rewards = list(U_0 = "reward_A0", U_1 = "reward_A1")),
    "'deterministic_rewards' must be a named list with names in the set 'U_A0', 'U_A1'."
  )
  pd <- policy_data(data = wide_data_copy,
                    action = "treat",
                    covariates = c("B", "Z", "L"),
                    utility = "outcome",
                    deterministic_rewards = list(U_A0 = "reward_A0", U_A1 = "reward_A1"))
  expect_identical(
    pd$stage_data[, c("U_A0", "U_A1"), with = FALSE],
    data.table(U_A0 = c(2,NA,2,NA), U_A1 = c(1,NA,1,NA))
  )
})

test_that("policy_data() formats long data correctly for a single stage case.", {

  ## long data:
  ld <- data.table(
    id = c(1,1,2,2),
    stage = c(1,2,1,2),
    event = c(0,1,0,1),
    time = rep(as.numeric(NA), 4),
    time2 = rep(as.numeric(NA), 4),
    A = c("0", NA, "1", NA),
    B = c("gr1", NA, "gr2", NA),
    Z = c("A", NA, "B", NA),
    L = c(1, NA, 2, NA),
    U = c(0, 10, 0, 5),
    U_A0 = rep(0, 4),
    U_A1 = rep(0, 4)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  ## baseline data:
  bd <- data.table(
    id = c(1,2),
    W = c("blue", "red")
  )
  setkey(bd, id)

  ## correct
  expect_error(
    policy_data(data = ld, baseline_data = bd, type = "long"),
    NA
  )
  pd <- policy_data(data = ld, baseline_data = bd, type = "long")
  expect_equal(ld, pd$stage_data)
  expect_equal(bd, pd$baseline_data)

  ## invalid variable names
  expect_error(
    policy_data(data = ld, type = "long", id = 1),
    "'id' must be a character string."
  )
  expect_error(
    policy_data(data = ld, type = "long", id = "ID"),
    "'id' is invalid."
  )
  expect_error(
    policy_data(data = ld, type = "long", action = "AA"),
    "'action' is invalid."
  )

  # setting new names:
  ld_copy <- copy(ld)
  setnames(ld_copy, "id", "ID")
  setnames(ld_copy, "stage", "k")
  setnames(ld_copy, "event", "evt")
  setnames(ld_copy, "A", "action")
  setnames(ld_copy, "U", "reward")

  bd_copy <- copy(bd)
  setnames(bd_copy, "id", "ID")

  ## correct variable names:
  expect_error(
    policy_data(
      data = ld_copy,
      baseline_data = bd_copy,
      type = "long",
      action = "action",
      id = "ID",
      stage = "k",
      event = "evt",
      utility = "reward"
    ),
    NA
  )
  ## comparison with original:
  expect_identical(
    policy_data(
      data = ld_copy,
      baseline_data = bd_copy,
      type = "long",
      action = "action",
      id = "ID",
      stage = "k",
      event = "evt",
      utility = "reward"
    ),
    pd
  )

  # preventing duplicated id variable:
  ld_copy$id <- c(3,3,4,4)
  expect_error(
    policy_data(data = ld_copy, type = "long", id = "ID"),
    "'data' has a variable called \"id\", but 'id' = \"ID\". Please remove or rename 'id'."
  )
  ld_copy$id <- NULL

  # preventing duplicated action variable:
  ld_copy$A <- c(0,NA,0,NA)
  expect_error(
    policy_data(
      data = ld_copy,
      type = "long",
      action = "action",
      id = "ID",
      stage = "k",
      event = "evt",
      utility = "U"
    ),
    "'data' has a variable called \"A\", but 'action' = \"action\". Please remove or rename 'action'."
  )
  ld_copy$A <- NULL

  # missing variable with non-matching default\
  expect_error(
    policy_data(
      data = ld_copy,
      type = "long",
      action = "action",
      id = "ID",
      stage = "k",
      event = "evt"
    ),
    "'utility' is invalid."
  )

})

test_that("policy_data() handles missing values.", {
  # long data:
  ld <- data.table(
    id = c(1,1,2,2),
    stage = c(1,2,1,2),
    event = c(0,1,0,1),
    A = c(0, NA, 1, NA),
    B = c("gr1", NA, "gr2", NA),
    Z = c("A", NA, "B", NA),
    L = c(1, NA, 2, NA),
    U = c(0, 10, 0, 5),
    U_A0 = rep(0, 4),
    U_A1 = rep(0, 4)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,2),
    W = c("blue", "red")
  )
  setkey(bd, id)

  # correct
  expect_error(
    policy_data(data = ld, baseline_data = bd, type = "long"),
    NA
  )

  # missing values:
  ld$A <- c(NA , NA, 1, NA)
  expect_error(policy_data(data = ld, baseline_data = bd, type = "long"))
  ld$A <- c(0, NA, 1, NA)

  # missing utility
  ld$U <- c(NA, 10, 0, 5)
  expect_error(
    policy_data(data = ld, baseline_data = bd, type = "long"),
    "The utility variable U has missing values"
  )
  ld$U <- c(0, 10, 0, 5)

  # allowing for missing covariates
  ld$Z <- c(NA, NA, "B", NA)
  expect_error(
    policy_data(data = ld, baseline_data = bd, type = "long"),
    NA
  )
})

test_that("policy_data() runs without covariates.", {

  ## long data without covariates:
  ld <- data.table(
    id = c(1,1,2,2),
    stage = c(1,2,1,2),
    event = c(0,1,0,1),
    A = c(0, NA, 1, NA),
    U = c(0, 10, 0, 5),
    U_A0 = rep(0, 4),
    U_A1 = rep(0, 4)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  expect_error(
    pd <- policy_data(data = ld,  type = "long"),
    NA
  )

  ## wide data without covariates:
  wide_data <- data.table(
    treat = c(0,1),
    outcome = c(10, 5)
  )

  expect_error(
    pd <- policy_data(data = wide_data, action = "treat", utility = "outcome", covariates = list()),
    NA
  )

})

test_that("get_history() checks input", {
  d <- sim_single_stage(1e2, seed=1)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  # stage
  ss <- list("test", logical(0), c(TRUE, TRUE), -1, 1.1, 0)
  lapply(ss, function(ss){
    expect_error(
      get_history(pd,
                  stage = ss),
      "stage must be an integer greater than 0."
    )

  })
  ss <- list(1)
  lapply(ss, function(ss){
    expect_error(
      get_history(pd,
                  stage = ss),
      NA
    )

  })
  expect_error(
    get_history(pd,
                stage = 2)
  )


  # full_history
  fh <- list("test", logical(0), c(TRUE, TRUE))
  lapply(fh, function(fh){
    expect_error(
      get_history(pd,
                  full_history = fh)
    )

  })
  fh <- list(TRUE, TRUE)
  lapply(fh, function(fh){
    expect_error(
      get_history(pd,
                  stage = 1,
                  full_history = fh),
      NA
    )

  })
  rm(fh)
})

test_that("full_history() returns the history associated with the action process when event_set = c(0).", {

  # long data
  ld <- data.table(
    id = c(1,1,3,3,4,4),
    stage = c(1,2,1,2,1,2),
    event = c(0,1,0,1,0,1),
    A = c("0", NA, "1", NA,"0", NA),
    B = c("gr1","gr1", "gr3", "gr3", "gr4", "gr4"),
    Z = c("A", NA, "B", NA, "B", NA),
    L = c(1, 2, 1, 3, 3,4),
    U = c(0.1, 10, 0.2, 0.3, 0.4, 0.5),
    U_A0 = c(0,NA,0,NA, 0, NA),
    U_A1 = c(0,NA,0,NA, 0, NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,3,4),
    W = c("blue", "blue", "red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    type = "long")


  ## stage 1:
  fh <- full_history(pd, stage = 1, event_set = c(0))

  ref <- data.table(
    id = c(1,3,4),
    stage = c(1,1,1),
    B_1 = c("gr1", "gr3", "gr4"),
    Z_1 = c("A", "B", "B"),
    L_1 = c(1,1,3),
    W = c("blue", "blue", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(1,1,1),
    A_1 = c("0","1","0")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$A,
    ref
  )

  expect_equal(
    fh$action_name,
    "A_1"
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(1,1,1),
    U_bar = c(0.1, 0.2, 0.4),
    U_A0 = c(0,0,0),
    U_A1 = c(0,0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$U,
    ref
  )

  ## stage 2
  expect_error(
    fh <- full_history(pd, stage = 2, event_set = c(0)),
    "empty history for the given stage and event_set."
  )

})

test_that("full_history() returns the history associated with the event process when event_set = c(0,2).", {

  # long data
  ## 5 cases: no right censoring, right censored before/at stage 1 action,
  ## 2Xright censored utility outcome (at stage 2), terminal at stage 1
  ld <- data.table(
    id = c(1,1,2,3,3,4,4,5),
    stage = c(1,2,1,1,2,1,2,1),
    event = c(0,1,2,0,2,0,2,1),
    A = c("0", NA, NA, "1", NA,"0", NA, NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3", "gr4", "gr4", "gr1"),
    Z = c("A", NA, "A", "B", NA, "B", NA, "B"),
    L = c(1,2,2,1,3,3,4,6),
    time = c(1, 2, 0.5,1,1.5,1,1.2,1.3),
    U = c(0.1, 0.2,0.3,0.4,NA,0.6,NA,0.4),
    U_A0 = c(0,0,0,0,0,0,0,0),
    U_A1 = c(0,0,0,0,0,0,0,0)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,2,3,4,5),
    W = c("blue","red","blue","red","red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    time = c("time"),
                    type = "long")


  ## stage 1:
  fh <- full_history(pd, stage = 1, event_set = c(0,2))

  ref <- data.table(
    id = c(1,2,3,4),
    stage = c(1,1,1,1),
    B_1 = c("gr1", "gr2", "gr3", "gr4"),
    Z_1 = c("A", "A", "B", "B"),
    L_1 = c(1,2,1,3),
    W = c("blue", "red", "blue", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4),
    stage = c(1,1,1,1),
    event = c(0,2,0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4),
    stage = c(1,1,1,1),
    time = c(0,0,0,0),
    time2 = c(1,0.5,1,1)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4),
    stage = c(1,1,1,1),
    U_bar = c(0.1, 0.3, 0.4, 0.6),
    U_A0 = c(0,0,0,0),
    U_A1 = c(0,0,0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$U,
    ref
  )

  ## stage 2:
  fh <- full_history(pd, stage = 2, event_set = c(0,2))

  ref <- data.table(
    id = c(3,4),
    stage = c(2,2),
    A_1 = c("1","0"),
    B_1 = c("gr3", "gr4"),
    B_2 = c("gr3", "gr4"),
    Z_1 = c("B", "B"),
    Z_2 = as.character(c(NA, NA)),
    L_1 = c(1,3),
    L_2 = c(3,4),
    W = c("blue", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(3,4),
    stage = c(2,2),
    event = c(2,2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(3,4),
    stage = c(2,2),
    time = c(1,1),
    time2 = c(1.5,1.2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

  ref <- data.table(
    id = c(3,4),
    stage = c(2,2),
    U_bar = as.numeric(c(0.4, 0.6)),
    U_A0 = c(0,0),
    U_A1 = c(0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$U,
    ref
  )

})

test_that("full_history() returns the history associated with the event process when event_set = c(0,1,2)", {

  # long data
  ## 5 cases: no right censoring, right censored before/at stage 1 action,
  ## 2Xright censored utility outcome (at stage 2), terminal at stage 1
  ld <- data.table(
    id = c(1,1,2,3,3,4,4,5),
    stage = c(1,2,1,1,2,1,2,1),
    event = c(0,1,2,0,2,0,2,1),
    A = c("0", NA, NA, "1", NA,"0", NA, NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3", "gr4", "gr4", "gr1"),
    Z = c("A", NA, "A", "B", NA, "B", NA, "B"),
    L = c(1,2,2,1,3,3,4,6),
    time = c(1, 2, 0.5,1,1.5,1,1.2,1.3),
    U = c(0.1, 0.2,0.3,0.4,NA,0.6,NA,0.4),
    U_A0 = c(0,0,0,0,0,0,0,0),
    U_A1 = c(0,0,0,0,0,0,0,0)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,2,3,4,5),
    W = c("blue","red","blue","red","red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    time = c("time"),
                    type = "long")


  ## stage 1:
  fh <- full_history(pd, stage = 1, event_set = c(0,1,2))

  ref <- data.table(
    id = c(1,2,3,4,5),
    stage = c(1,1,1,1,1),
    B_1 = c("gr1", "gr2", "gr3", "gr4","gr1"),
    Z_1 = c("A", "A", "B", "B","B"),
    L_1 = c(1,2,1,3,6),
    W = c("blue", "red", "blue", "red","red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4,5),
    stage = c(1,1,1,1,1),
    event = c(0,2,0,0,1)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4,5),
    stage = c(1,1,1,1,1),
    time = c(0,0,0,0,0),
    time2 = c(1,0.5,1,1,1.3)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4,5),
    stage = c(1,1,1,1,1),
    U_bar = c(0.1, 0.3, 0.4, 0.6, 0),
    U_A0 = c(0,0,0,0,0),
    U_A1 = c(0,0,0,0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$U,
    ref
  )

  ## stage 2:
  fh <- full_history(pd, stage = 2, event_set = c(0,1,2))

  ref <- data.table(
    id = c(1,3,4),
    stage = c(2,2,2),
    A_1 = c("0","1","0"),
    B_1 = c("gr1","gr3", "gr4"),
    B_2 = c("gr1","gr3", "gr4"),
    Z_1 = c("A", "B", "B"),
    Z_2 = as.character(c(NA, NA, NA)),
    L_1 = c(1,1,3),
    L_2 = c(2,3,4),
    W = c("blue", "blue", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(2,2,2),
    event = c(1,2,2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(2,2,2),
    time = c(1,1,1),
    time2 = c(2,1.5,1.2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(2,2,2),
    U_bar = c(0.1, 0.4, 0.6),
    U_A0 = c(0,0,0),
    U_A1 = c(0,0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$U,
    ref
  )

})


test_that("stage_state_history() returns the history associated with the action process when event_set = c(0)", {

  # long data
  ld <- data.table(
    id = c(1,1,3,3,4,4),
    stage = c(1,2,1,2,1,2),
    event = c(0,1,0,1,0,1),
    A = c("0", NA, "1", NA,"0", NA),
    B = c("gr1","gr1", "gr3", "gr3", "gr4", "gr4"),
    Z = c("A", NA, "B", NA, "B", NA),
    L = c(1, 2, 1, 3, 3,4),
    time = c(1, 2, 1, 1.5, 1, 1.2),
    U = c(0.1, 10,0.2, 0.3, 0.4, 0.5),
    U_A0 = c(0,0,0,NA, 0, NA),
    U_A1 = c(0,0,0,NA, 0, NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,3,4),
    W = c("blue", "blue", "red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    time = "time",
                    type = "long")

  ## stage 1:
  fh <- stage_state_history(pd, stage = 1, event_set = c(0))

  ref <- data.table(
    id = c(1,3,4),
    stage = c(1,1,1),
    B = c("gr1", "gr3", "gr4"),
    Z = c("A", "B", "B"),
    L = c(1,1,3),
    W = c("blue", "blue", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(1,1,1),
    A = c("0","1","0")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$A,
    ref
  )

  expect_equal(
    fh$action_name,
    "A"
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(1,1,1),
    U_bar = c(0.1, 0.2, 0.4),
    U_A0 = c(0,0,0),
    U_A1 = c(0,0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$U,
    ref
  )

  ## stage 2
  expect_error(
    fh <- stage_state_history(pd, stage = 2, event_set = c(0)),
    "empty history for the given stage and event_set."
  )

})

test_that("stage_state_history() returns the history associated with the event process when event_set = c(0,2)", {

  # long data
  ## 5 cases: no right censoring, right censored before/at stage 1 action,
  ## 2Xright censored utility outcome (at stage 2), terminal at stage 1
  ld <- data.table(
    id = c(1,1,2,3,3,4,4,5),
    stage = c(1,2,1,1,2,1,2,1),
    event = c(0,1,2,0,2,0,2,1),
    A = c("0", NA, NA, "1", NA,"0", NA, NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3", "gr4", "gr4", "gr1"),
    Z = c("A", NA, "A", "B", NA, "B", NA, "B"),
    L = c(1,2,2,1,3,3,4,6),
    time = c(1, 2, 0.5,1,1.5,1,1.2,1.3),
    U = c(0.1, 0.2,0.3,0.4,NA,0.6,NA,0.4),
    U_A0 = c(0,0,0,0,0,0,0,0),
    U_A1 = c(0,0,0,0,0,0,0,0)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,2,3,4,5),
    W = c("blue","red","blue","red","red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    time = c("time"),
                    type = "long")

  ## stage 1:
  fh <- stage_state_history(pd, stage = 1, event_set = c(0,2))

  ref <- data.table(
    id = c(1,2,3,4),
    stage = c(1,1,1,1),
    B = c("gr1", "gr2", "gr3", "gr4"),
    Z = c("A", "A", "B", "B"),
    L = c(1,2,1,3),
    W = c("blue", "red", "blue", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4),
    stage = c(1,1,1,1),
    event = c(0,2,0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4),
    stage = c(1,1,1,1),
    time = c(0,0,0,0),
    time2 = c(1,0.5,1,1)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4),
    stage = c(1,1,1,1),
    U_bar = c(0.1, 0.3, 0.4, 0.6),
    U_A0 = c(0,0,0,0),
    U_A1 = c(0,0,0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$U,
    ref
  )

  ## stage 2
  fh <- stage_state_history(pd, stage = 2, event_set = c(0,2))

  ref <- data.table(
    id = c(3,4),
    stage = c(2,2),
    B = c("gr3", "gr4"),
    Z = as.character(c(NA, NA)),
    L = c(3,4),
    W = c("blue", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(3,4),
    stage = c(2,2),
    event = c(2,2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(3,4),
    stage = c(2,2),
    time = c(1,1),
    time2 = c(1.5,1.2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

  ref <- data.table(
    id = c(3,4),
    stage = c(2,2),
    U_bar = as.numeric(c(0.4, 0.6)),
    U_A0 = c(0,0),
    U_A1 = c(0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$U,
    ref
  )

})

test_that("stage_state_history() returns the history associated with the event process when event_set = c(0,1,2)", {

  # long data
  ## 5 cases: no right censoring, right censored before/at stage 1 action,
  ## 2Xright censored utility outcome (at stage 2), terminal at stage 1
  ld <- data.table(
    id = c(1,1,2,3,3,4,4,5),
    stage = c(1,2,1,1,2,1,2,1),
    event = c(0,1,2,0,2,0,2,1),
    A = c("0", NA, NA, "1", NA,"0", NA, NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3", "gr4", "gr4", "gr1"),
    Z = c("A", NA, "A", "B", NA, "B", NA, "B"),
    L = c(1,2,2,1,3,3,4,6),
    time = c(1, 2, 0.5,1,1.5,1,1.2,1.3),
    U = c(0.1, 0.2,0.3,0.4,NA,0.6,NA,0.4),
    U_A0 = c(0,0,0,0,0,0,0,0),
    U_A1 = c(0,0,0,0,0,0,0,0)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,2,3,4,5),
    W = c("blue","red","blue","red","red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    time = c("time"),
                    type = "long")

  ## stage 1:
  fh <- stage_state_history(pd, stage = 1, event_set = c(0,1,2))

  ref <- data.table(
    id = c(1,2,3,4,5),
    stage = c(1,1,1,1,1),
    B = c("gr1", "gr2", "gr3", "gr4", "gr1"),
    Z = c("A", "A", "B", "B", "B"),
    L = c(1,2,1,3,6),
    W = c("blue", "red", "blue", "red","red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4,5),
    stage = c(1,1,1,1,1),
    event = c(0,2,0,0,1)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,4,5),
    stage = c(1,1,1,1,1),
    time = c(0,0,0,0,0),
    time2 = c(1,0.5,1,1,1.3)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

  ## stage 2
  fh <- stage_state_history(pd, stage = 2, event_set = c(0,1,2))

  ref <- data.table(
    id = c(1,3,4),
    stage = c(2,2,2),
    B = c("gr1","gr3", "gr4"),
    Z = as.character(c(NA,NA, NA)),
    L = c(2,3,4),
    W = c("blue","blue", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(2,2,2),
    event = c(1,2,2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(2,2,2),
    time = c(1,1,1),
    time2 = c(2,1.5,1.2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(2,2,2),
    U_bar = as.numeric(c(0.1,0.4, 0.6)),
    U_A0 = c(0,0,0),
    U_A1 = c(0,0,0)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$U,
    ref
  )

})


test_that("state_history() returns the history associated with the action process when event_set = c(0)", {

  # long data
  ld <- data.table(
    id = c(1,1,3,3,4,4),
    stage = c(1,2,1,2,1,2),
    event = c(0,1,0,1,0,1),
    A = c("0", NA, "1", NA,"0", NA),
    B = c("gr1","gr1", "gr3", "gr3", "gr4", "gr4"),
    Z = c("A", NA, "B", NA, "B", NA),
    L = c(1, 2, 1, 3, 3,4),
    time = c(1, 2, 1, 1.5, 1, 1.2),
    U = c(0.2, 10, 0.3, 0, 0.4, 0),
    U_A0 = c(0,0,0,NA, 0, NA),
    U_A1 = c(0,0,0,NA, 0, NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,3,4),
    W = c("blue", "blue", "red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    time = "time",
                    type = "long")


  ## stage 1:
  fh <- state_history(pd, event_set = c(0))

  ref <- data.table(
    id = c(1,3,4),
    stage = c(1,1,1),
    B = c("gr1","gr3","gr4"),
    Z = c("A", "B", "B"),
    L = c(1, 1, 3),
    W = c("blue", "blue", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,3,4),
    stage = c(1,1,1),
    A = c("0","1","0")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$A,
    ref
  )

  expect_equal(
    fh$action_name,
    "A"
  )

  expect_equal(
    fh$U,
    NULL
  )

})




test_that("state_history() returns the history associated with the event process when type = 'censoring'", {

  # long data
  ## 5 cases: no right censoring, right censored before/at stage 1 action,
  ## 2Xright censored utility outcome (at stage 2), terminal at stage 1
  ld <- data.table(
    id = c(1,1,2,3,3,4,4,5),
    stage = c(1,2,1,1,2,1,2,1),
    event = c(0,1,2,0,2,0,2,1),
    A = c("0", NA, NA, "1", NA,"0", NA, NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3", "gr4", "gr4", "gr1"),
    Z = c("A", NA, "A", "B", NA, "B", NA, "B"),
    L = c(1,2,2,1,3,3,4,6),
    time = c(1, 2, 0.5,1,1.5,1,1.2,1.3),
    U = c(0.1, 0.2,0.3,0.4,NA,0.6,NA,0.4),
    U_A0 = c(0,0,0,0,0,0,0,0),
    U_A1 = c(0,0,0,0,0,0,0,0)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,2,3,4,5),
    W = c("blue","red","blue","red","red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    time = c("time"),
                    type = "long")


  fh <- state_history(pd, event_set = c(0,2))

  ref <- data.table(
    id = c(1,2,3,3,4,4),
    stage = c(1,1,1,2,1,2),
    B = c("gr1","gr2", "gr3", "gr3", "gr4", "gr4"),
    Z = c("A", "A", "B", NA, "B", NA),
    L = c(1, 2, 1, 3, 3,4),
    W = c("blue","red", "blue", "blue", "red", "red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,3,4,4),
    stage = c(1,1,1,2,1,2),
    event = c(0,2,0,2,0,2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(1,2,3,3,4,4),
    stage = c(1,1,1,2,1,2),
    time = c(0,0,0,1,0,1),
    time2 = c(1,0.5,1,1.5,1,1.2)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

})

test_that("state_history() returns the history associated with the event process when type = 'all'", {

  # long data
  ## 5 cases: no right censoring, right censored before/at stage 1 action,
  ## 2Xright censored utility outcome (at stage 2), terminal at stage 1
  ld <- data.table(
    id = c(1,1,2,3,3,4,4,5),
    stage = c(1,2,1,1,2,1,2,1),
    event = c(0,1,2,0,2,0,2,1),
    A = c("0", NA, NA, "1", NA,"0", NA, NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3", "gr4", "gr4", "gr1"),
    Z = c("A", NA, "A", "B", NA, "B", NA, "B"),
    L = c(1,2,2,1,3,3,4,6),
    time = c(1, 2, 0.5,1,1.5,1,1.2,1.3),
    U = c(0.1, 0.2,0.3,0.4,NA,0.6,NA,0.4),
    U_A0 = c(0,0,0,0,0,0,0,0),
    U_A1 = c(0,0,0,0,0,0,0,0)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,2,3,4,5),
    W = c("blue","red","blue","red","red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld,
                    baseline_data = bd,
                    time = c("time"),
                    type = "long")


  fh <- state_history(pd, event_set = c(0,1,2))

  ref <- data.table(
    id = c(1,1,2,3,3,4,4,5),
    stage = c(1,2,1,1,2,1,2,1),
    B = c("gr1","gr1", "gr2", "gr3", "gr3", "gr4", "gr4","gr1"),
    Z = c("A", NA, "A", "B", NA, "B", NA,"B"),
    L = c(1, 2, 2, 1, 3, 3,4,6),
    W = c("blue", "blue", "red", "blue", "blue", "red", "red","red")
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$H,
    ref
  )

  ref <- data.table(
    id = c(1,1,2,3,3,4,4,5),
    stage = c(1,2,1,1,2,1,2,1),
    event = c(0,1,2,0,2,0,2,1)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$event,
    ref
  )

  ref <- data.table(
    id = c(1,1,2,3,3,4,4,5),
    stage = c(1,2,1,1,2,1,2,1),
    time = c(0,1,0,0,1,0,1,0),
    time2 = c(1,2,0.5,1,1.5,1,1.2,1.3)
  )
  setkey(ref, id, stage)

  expect_equal(
    fh$time,
    ref
  )

  expect_equal(
    fh$U,
    NULL
  )

})

test_that("get_utility returns NA for right-censored observations", {
  sim_single_stage_right_cens <- function(n = 2e3, zeta = c(0.7, 0.2), type = "right"){

    d <- sim_single_stage(n = n)
    pd <- policy_data(data = d,
                      action = "A",
                      covariates = c("Z", "L", "B"),
                      utility = "U")

    ld <- pd$stage_data

    ld[stage == 1, time := 1]
    ld[stage == 2, time := 2]

    ld[stage == 2, Z := d$Z]
    ld[stage == 2, L := d$L]
    ld[stage == 2, B := d$B]

    ## simulating the right censoring time
    ## only depending on the baseline covariate Z:
    C <- c(rexp(n, 1) / exp((-1) * cbind(1, as.numeric(d$Z)) %*% zeta))

    ld[stage == 1, time_c := C]
    ld[stage == 2, time_c := C]

    ld[, delta := time_c >= time]

    ld[delta == FALSE , event := 2]
    ld[delta == FALSE, A := NA]
    ld[delta == FALSE & stage == 2, U := NA]
    ld[delta == FALSE, U_A0 := 0]
    ld[delta == FALSE, U_A1 := 0]

    ld[ , tmp := shift(delta, fill = TRUE), by = list(id)]
    ld <- ld[tmp == TRUE, ]
    ld[ , time := pmin(time, time_c)]
    ld[ , time_c := NULL]
    ld[ , tmp := NULL]
    ld[ , delta := NULL]

    if (type == "interval"){
      ld[, time2 := time]
      ld[, time := shift(time, fill = 0), by = list(id)]
    }

    return(ld)
  }

  set.seed(1)
  ld <- sim_single_stage_right_cens(n = 5e2, type = "interval")
  pd <- policy_data(data = ld, type = "long", action = "A", time = "time", time2 = "time2")

  util <- get_utility(pd)

  expect_equal(
    c("id", "U"),
    colnames(util)
  )

  expect_equal(
    key(util),
    "id"
  )

  expect_equal(
    ld[ , list(na = any(event == 2)), by = id][["na"]],
    is.na(get_utility(pd)[["U"]])
  )

  expect_equal(
    ld[ , list(na = any(event == 2), U = sum(U)), by = id][na == FALSE,][["U"]],
    get_utility(pd)[["U"]][!is.na(get_utility(pd)[["U"]])]
  )

})

test_that("a policy_data object prints as expected under right censoring occur.", {

  # long data
  ## 4 cases: no right censoring, right censored before/at stage 1 action,
  ## 2Xright censored utility outcome (at stage 2)
  ld <- data.table(
    id = c(1,1,2,3,3,4,4),
    stage = c(1,2,1,1,2,1,2),
    event = c(0,1,2,0,2,0,2),
    A = c("0", NA, NA, "1", NA,"0", NA),
    B = c("gr1","gr1", "gr2", "gr3", "gr3", "gr4", "gr4"),
    Z = c("A", NA, "A", "B", NA, "B", NA),
    L = c(1, 2, 2, 1, 3, 3,4),
    time = c(1, 2, 0.5, 1, 1.5, 1, 1.2),
    U = c(0, 10, 0, 0, NA, 0, NA),
    U_A0 = c(0,0,0,0,NA, 0, NA),
    U_A1 = c(0,0,0,0,NA, 0, NA)
  )
  setkey(ld, id, stage)
  setindex(ld, event)

  # baseline data:
  bd <- data.table(
    id = c(1,2,3,4),
    W = c("blue", "red", "blue", "red")
  )
  setkey(bd, id)

  pd <- policy_data(data = ld, baseline_data = bd, type = "long")

  invisible(
    capture.output(
      expect_error(
        print(pd),
        NA
      )
    )
  )

})


test_that("plot.policy_data() does not fail for right-censored data", {

  sim_two_stage_right_cens <- function(n = 2e2,
                                       par = c(gamma = 0.5,  beta = 1, zeta = 1),
                                       seed = NULL,
                                       action_model_1 = function(C_1, beta, ...)
                                         stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_1)),
                                       action_model_2 = function(C_2, beta, ...)
                                         stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_2)),
                                       deterministic_rewards = FALSE,
                                       cens_model = function(L, zeta, ...)
                                         stats::rbinom(n = NROW(L), size = 1, prob = lava::expit(zeta * abs(L)))) {
    d <- sim_two_stage(n = n)
    d$U <- d$U_1 + d$U_2 + d$U_3

    pd <- policy_data(data = d,
                      action = c("A_1", "A_2"),
                      baseline = c("B", "BB"),
                      covariates = list(L = c("L_1", "L_2"),
                                        C = c("C_1", "C_2")),
                      utility = "U")

    ld <- pd$stage_data
    ld[is.na(L), L := d[["L_3"]]]

    ## simulating the discrete right-censoring:
    delta <- rbinom(n = nrow(ld), size = 1, prob = cens_model(ld$L, zeta = par["zeta"]))

    ## adapting the data to the right-censoring process:
    ld[ , delta := delta]
    ld[ , delta := cumprod(delta), by = id]
    ld[ , tmp := cumsum(delta == 0), by = id]
    ld <- ld[tmp %in% c(0,1)]
    ld[ , tmp := NULL]
    ld[delta == 0, event := 2]
    ld[ , delta := NULL]
    ld[event == 2 & stage == 3, U := NA]
    ld[event == 2, A := NA]

    return(ld)
  }

  ld <- sim_two_stage_right_cens()

  pd <- policy_data(data = ld, type = "long")
  p1 <- policy_def(1, reuse = TRUE)
  p0 <- policy_def(0, reuse = TRUE)

  pdf(NULL)
  expect_no_error(
    plot(pd, which = 1, policy = list(p1, p0))

  )
  dev.off()

})

test_that("partial.policy_data() handles right-censoring", {

  sim_two_stage_right_cens <- function(n = 2e2,
                                       par = c(gamma = 0.5,  beta = 1, zeta = 1),
                                       seed = NULL,
                                       action_model_1 = function(C_1, beta, ...)
                                         stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_1)),
                                       action_model_2 = function(C_2, beta, ...)
                                         stats::rbinom(n = NROW(C_1), size = 1, prob = lava::expit(beta * C_2)),
                                       deterministic_rewards = FALSE,
                                       cens_model = function(L, zeta, ...)
                                         stats::rbinom(n = NROW(L), size = 1, prob = lava::expit(zeta * abs(L)))) {
    d <- sim_two_stage(n = n)
    d$U <- d$U_1 + d$U_2 + d$U_3

    pd <- policy_data(data = d,
                      action = c("A_1", "A_2"),
                      baseline = c("B", "BB"),
                      covariates = list(L = c("L_1", "L_2"),
                                        C = c("C_1", "C_2")),
                      utility = "U")

    ld <- pd$stage_data
    ld[is.na(L), L := d[["L_3"]]]

    ## simulating the discrete right-censoring:
    delta <- rbinom(n = nrow(ld), size = 1, prob = cens_model(ld$L, zeta = par["zeta"]))

    ## adapting the data to the right-censoring process:
    ld[ , delta := delta]
    ld[ , delta := cumprod(delta), by = id]
    ld[ , tmp := cumsum(delta == 0), by = id]
    ld <- ld[tmp %in% c(0,1)]
    ld[ , tmp := NULL]
    ld[delta == 0, event := 2]
    ld[ , delta := NULL]
    ld[event == 2 & stage == 3, U := NA]
    ld[event == 2, A := NA]

    return(ld)
  }

  ld <- sim_two_stage_right_cens()

  pd <- policy_data(data = ld, type = "long")

  expect_no_error(
    pd2 <- partial(pd, K = 2)
  )

  expect_equal(pd, pd2)

  expect_error(
    pd1 <- partial(pd, K = 1),
    "partial.policy_data\\(\\) not implemented for right-censored data."
  )

})

test_that("the action set is preserved when subsetting",{
  d1 <- sim_single_stage(10, seed=1)
  pd1 <- policy_data(d1, action = "A", covariates = c("Z"), utility = "U")

  expect_error(
    pd2 <- subset_id(pd1, id = get_id(pd1)[d1$A == "0"]),
    NA
  )

  expect_equal(
    get_action_set(pd1),
    get_action_set(pd2)
  )

  invisible(capture.output(
    expect_error(
      print(pd2),
      NA
    )
  ))

})

test_that("policy_def checks the action set",{
  d <- sim_single_stage(2e3, seed=1)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  p <- policy_def("test")

  expect_warning(
    p(pd),
    "The policy actions does not comply with the action set of the policy data object."
  )

  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U",
                    action_set = c("1", "0", "test"))

  expect_error(
    p(pd),
    NA
  )

})

# Single stage ------------------------------------------------------------

test_that("policy_def handles static policies (single stage).",{
  d <- sim_single_stage(2e3, seed=1)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  p <- policy_def(1)
  expect_equal(
    p(pd)[["d"]],
    rep("1", get_n(pd))
  )

  p <- policy_def("1")
  expect_equal(
    p(pd)[["d"]],
    rep("1", get_n(pd))
  )

  p <- policy_def(as.factor("1"))
  expect_equal(
    p(pd)[["d"]],
    rep("1", get_n(pd))
  )

  p <- policy_def(TRUE)
  expect_warning(
    p(pd)[["d"]],
    "The policy actions does not comply with the action set of the policy data object."
  )

  p <- policy_def(c(1))
  expect_equal(
    p(pd)[["d"]],
    rep("1", get_n(pd))
  )

  p <- policy_def(list(1))
  expect_equal(
    p(pd)[["d"]],
    rep("1", get_n(pd))
  )

  p <- policy_def(list(2))
  expect_warning(
    p(pd)[["d"]],
    "The policy actions does not comply with the action set of the policy data object."
  )

  p <- list(
    policy_def(c(0,1)),
    policy_def(c("0","1")),
    policy_def(as.factor(c("0","1"))),
    policy_def(list("0","1"))
  )
  lapply(p,function(p){
    expect_error(
      p(pd),
      "policy_functions must be a list of length K."
    )
  })

  ll <- list(
    c(0,1),
    c("0","1"),
    as.factor(c("0","1")),
    list("0","1")
  )
  lapply(ll,function(l){
    expect_error(
      policy_def(ll, reuse = TRUE),
      "When reuse is TRUE, policy_functions must be a single function or a constant."
    )
  })

})

test_that("policy_def handles dynamic policies (single stage).",{
  d <- sim_single_stage(2e3, seed=1)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  fun <- function(L) (L>0)*1
  res <- do.call(what = function(L,...) as.character(fun(L)), d)

  p <- policy_def(fun, name = "a=(L>0)")
  expect_equal(
    p(pd)[["d"]],
    res
  )
  expect_equal(
    attr(p, "name"),
    "a=(L>0)"
  )

  p <- policy_def(fun, reuse = TRUE)
  expect_equal(
    p(pd)[["d"]],
    res
  )

  expect_error(
    policy_def(fun, reuse = TRUE, full_history = TRUE),
    "full_history must be FALSE when reuse is TRUE."
  )

  p <- policy_def(list(fun), reuse = TRUE)
  expect_equal(
    p(pd)[["d"]],
    res
  )

  expect_error(
    policy_def(list(fun, fun), reuse = TRUE),
    "When reuse is TRUE, policy_functions must be a single function or a constant."
  )
})

# Two stages ------------------------------------------------------------

test_that("policy_def handles static policies (two stages).",{
  d <- sim_two_stage(2e3, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  a <- list("1",
            1,
            list(1),
            c(1))
  tmp <- lapply(
    a,
    function(a){
      p <- policy_def(a, reuse = TRUE)
      expect_equal(p(pd)[["d"]],
                   rep("1", get_n(pd) * get_K(pd)))
    }
  )

  a <- list(c("1", "1"),
            c(1,1),
            list(1,1),
            as.factor(c(1,1)))
  tmp <- lapply(
    a,
    function(a){
      p <- policy_def(a, reuse = FALSE)
      expect_equal(p(pd)[["d"]],
                   rep("1", get_n(pd) * get_K(pd)))
    }
  )

  p <- policy_def(c(0,1), reuse = FALSE)
  expect_equal(
    p(pd)[["d"]],
    as.character(rep(c(0,1), times = get_n(pd)))
  )

  p <- policy_def(2, reuse = TRUE)
  expect_warning(
    p(pd),
    "The policy actions does not comply with the action set of the policy data object."
  )

  p <- policy_def(c(1,2), reuse = FALSE)
  expect_warning(
    p(pd),
    "The policy actions does not comply with the action set of the policy data object."
  )

  p <- policy_def(c(1,1,1), reuse = FALSE)
  expect_error(
    p(pd),
    "policy_functions must be a list of length K."
  )

})

test_that("policy_def handles dynamic policies (two stages).",{
  d <- sim_two_stage(2e3, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  fun <- function(L) (L>0)*1
  res <- do.call(pd[["stage_data"]][event == 0, ], what = function(L,...) as.character(fun(L)))

  p <- policy_def(fun, name = "a=(L>0)")
  expect_error(
    p(d),
    'The policy input is not of inherited class policy_data.'
  )
  expect_error(
    p(pd),
    "policy_functions must be a list of length K."
  )

  p <- policy_def(fun, reuse = TRUE, name = "a=(L>0)")
  expect_equal(
    p(pd)[["d"]],
    res
  )
  expect_equal(
    attr(p, "name"),
    "a=(L>0)"
  )

  expect_error(
    policy_def(fun, reuse = TRUE, full_history = TRUE),
    "full_history must be FALSE when reuse is TRUE."
  )

  p <- policy_def(list(fun, fun), reuse = FALSE, name = "a=(L>0)")
  expect_equal(
    p(pd)[["d"]],
    res
  )

  fun <- list(function(L) (L>0)*1, function(C) (C>0)*1)
  res <- do.call(pd[["stage_data"]][event == 0, ], what = function(L, C, stage,...){
    (stage == 1) * as.numeric((L>0)) +
      (stage == 2) * as.numeric((C>0))
  })
  res <- as.character(res)

  p <- policy_def(fun, reuse = FALSE, name = "a=(L>0)")
  expect_equal(
    p(pd)[["d"]],
    res
  )

  fun <- list(function(L_1) (L_1>0)*1, function(C_2) (C_2>0)*1)
  p <- policy_def(fun, reuse = FALSE, name = "a=(L>0)", full_history = TRUE)
  expect_equal(
    p(pd)[["d"]],
    res
  )

  fun <- list(function(L_2) (L_2>0)*1, function(C_2) (C_2>0)*1)
  p <- policy_def(fun, reuse = FALSE, name = "a=(L>0)", full_history = TRUE)
  expect_error(
    p(pd)[["d"]],
    "argument \"L_2\" is missing, with no default"
  )

})

# Stochastic number of stages ---------------------------------------------

test_that("policy_def handles a stochastic number of stages", {

  set.seed(1)
  d <- sim_multi_stage(50, seed = 1)
  # constructing policy_data object:
  pd <- policy_data(data = d$stage_data,
                    baseline_data = d$baseline_data,
                    type = "long",
                    id = "id",
                    stage = "stage",
                    event = "event",
                    action = "A",
                    utility = "U")
  pd <- partial(pd, K = 3)

  fun <- function(X, ...) as.character((X>0)*1)
  res <- do.call(pd[["stage_data"]][stage <= 3][event %in% c(0,2),], what = fun)

  p <- policy_def(function(X) as.character((X>0)*1), reuse = TRUE)

  expect_equal(
    p(pd)[["d"]],
    res
  )

  p <- policy_def(replicate(get_K(pd), function(X) (X>0)*1), reuse = FALSE)
  expect_equal(
    p(pd)[["d"]],
    res
  )

})

test_that("policy_def gets access to the naturally chosen access and the stage number", {
  d <- sim_two_stage(2e3, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  p <- policy_def(function(A){A},
                  reuse = TRUE,
                  natural_action = TRUE)

  pa <- p(pd)
  a <- get_actions(pd)
  setnames(a, new = "d", old = "A")

  expect_equal(
    a,
    pa
  )

  p <- policy_def(function(A, L){ifelse(L>0, A, "1")},
                  reuse = TRUE,
                  natural_action = TRUE)

  pa <- p(pd)
  his <- get_history(pd)$H
  A <- get_actions(pd)$A
  ref_pa <-  ifelse(his$L > 0, A, "1")
  expect_equal(
    ref_pa,
    pa$d
  )

  p <- policy_def(function(A, stage){ifelse(stage == 1, A, "1")},
                  reuse = TRUE,
                  natural_action = TRUE,
                  stage_number = TRUE)
  pa <- p(pd)
  his <- get_history(pd)$H
  A <- get_actions(pd)$A
  ref_pa <-  ifelse(his$stage == 1, A, "1")

  expect_equal(
    ref_pa,
    pa$d
  )

})

fit_QV_function <- function(history, Z, qv_model, valid_ids) {
  action_set <- getElement(history, "action_set")
  stage_action_set <- getElement(history, "stage_action_set")
  stage <- getElement(history, "stage")
  H <- get_H(history)
  stage_id <- unlist(get_id.history(history),
    use.names = FALSE
  )

  folds <- NULL
  if (!is.null(valid_ids)) {
    # adapting the validation ids to the current stage:
    stage_valid_ids <- lapply(
      valid_ids,
      function(vi) {
        vi[vi %in% stage_id]
      }
    )
    folds <-
      lapply(
        stage_valid_ids,
        function(vi) {
          which(stage_id %in% vi)
        }
      )
  }

   ## fitting the QV-model:
  qv_model <- apply(
    Z[, stage_action_set],
    MARGIN = 2,
    function(z) {
      model <- tryCatch({
        qv_model(V_res = z, AH = H, folds = folds)
      }, error = function(e) {
        stop("Error in qv_model: ", conditionMessage(e))
      })
      return(model)
    }
  )
  names(qv_model) <- paste("QV_", stage_action_set, sep = "")

  qv_function <- list(
    qv_model = qv_model,
    action_set = action_set,
    stage_action_set = stage_action_set,
    stage = stage
  )
  class(qv_function) <- "QV_function"

  return(qv_function)
}

#' @export
predict.QV_function <- function(object, new_history, ...) {
  id_stage <- get_id_stage(new_history)
  new_H <- get_H(new_history)
  # action set of the new history object
  action_set <- getElement(new_history, "action_set")
  # stage action set of the fitted QV-function
  stage_action_set <- getElement(object, "stage_action_set")
  qv_model <- getElement(object, "qv_model")

  qv_values <- matrix(NA, nrow = nrow(new_H), ncol = length(action_set))
  colnames(qv_values) <- paste("QV_", action_set, sep = "")

  qv_values[, action_set %in% stage_action_set] <- sapply(
    qv_model,
    function(qvm) {
      predict(qvm, new_H)
    }
  )
  qv_values <- data.table(id_stage, qv_values)
  setkeyv(qv_values, c("id", "stage"))

  return(qv_values)
}

#' @title Control arguments for doubly robust Q-learning
#' @description \code{control_drql} sets the default control arguments
#' for doubly robust Q-learning, \code{type = "drql"}.
#' @param qv_models Single element or list of V-restricted Q-models created
#' by [q_glm()], [q_rf()], [q_sl()] or similar functions.
#' @returns list of (default) control arguments.
#' @export
control_drql <- function(qv_models = q_glm(~.)) {
  control <- as.list(environment())
  return(control)
}

# note: if L > 1 and alpha > 0 then the g_functions fitted on the complete data
# is used to determine future realistic actions.
drql <- function(policy_data,
                 alpha,
                 g_models, g_functions, g_full_history,
                 q_models, q_full_history,
                 qv_models, full_history,
                 L, cross_fit_g_models,
                 save_cross_fit_models, future_args,
                 ...) {
  K <- get_K(policy_data)
  n <- get_n(policy_data)
  action_set <- get_action_set(policy_data)
  stage_action_sets <- get_stage_action_sets(policy_data)

  # input checks:
  if (!is.null(g_functions)) {
    if (!inherits(g_functions, what = "g_functions")) {
      stop("g-functions must be of class 'g_functions'.")
    }
  }
  if (is.list(q_models)) {
    if (length(q_models) != K) {
      stop("q_models must either be a list of length K or a single Q-model.")
    }
  }
  if (is.null(qv_models)) {
    stop("qv_models are missing.")
  }
  if (is.list(qv_models)) {
    if (length(qv_models) != K) {
      stop("qv_models must either be a list of length K or a single QV-model.")
    }
  }
  if (any(get_element(policy_data, "cens_indicator")[["indicator"]])){
    stop("policy learning with type 'drql' not implemented under right-censoring/missing outcomes.")
  }

  # getting the observed actions:
  actions <- get_actions(policy_data)

  # getting the observed (complete) utilities:
  utility <- get_utility(policy_data)

  # constructing the folds for cross-fitting:
  if (L > 1) {
    folds <- split(sample(1:n, n), rep(1:L, length.out = n))
    folds <- lapply(folds, sort)
  } else {
    folds <- NULL
  }

  # (cross-)fitting the g-functions:
  g_functions_cf <- NULL
  valid_ids <- NULL
  if (!is.null(folds) && cross_fit_g_models == TRUE) {
    g_cf <- crossfit_function(
      policy_data = policy_data,
      fun = fit_g_functions,
      models = g_models,
      full_history = g_full_history,
      folds = folds,
      save_cross_fit_models = save_cross_fit_models,
      future_args = future_args
    )
    g_functions_cf <- getElement(g_cf, "functions")
    g_values <- getElement(g_cf, "values")
    valid_ids <- getElement(g_cf, "valid_ids")
    rm(g_cf)
  } else {
    if (is.null(g_functions)) {
      g_functions <- fit_g_functions(policy_data,
        g_models = g_models,
        full_history = g_full_history
      )
    }
    g_values <- predict(g_functions, policy_data)
  }

  # fitting g-functions for determining new realistic actions:
  if (alpha > 0) {
    if (is.null(g_functions)) {
      g_functions <- fit_g_functions(policy_data,
        g_models = g_models,
        full_history = g_full_history
      )
    }
  } else {
    # g-functions are not saved if alpha == 0:
    g_functions <- NULL
  }

  # (n) vector with entries U_i:
  U <- utility$U
  # (n X K) matrix with entries I(d_k(H_k) = A_k):
  II <- matrix(nrow = n, ncol = K)
  # (n X K) matrix with entries g_k(A_k, H_k)
  g_A_values <- get_a_values(a = actions$A,
                             action_set = action_set,
                             g_values)
  G <- dcast(g_A_values, id ~ stage, value.var = "P")[, -c("id"), with = FALSE]
  G <- as.matrix(G)
  # (n X K+1) matrix with entries Q_k(H_{k,i}, d_k(H_{k,i})), Q_{K+1} = U:
  Q <- matrix(nrow = n, ncol = K+1)
  Q[, K+1] <- U
  # (n X K) matrix with entries d_k(H_k) (including unrealistic actions)
  D <- matrix(nrow = n, ncol = K)

  g_cols <- paste("g_", action_set, sep = "")
  q_cols <- paste("Q_", action_set, sep = "")
  qv_cols <- paste("QV_", action_set, sep = "")

  q_functions <- list()
  q_functions_cf <- list()
  qv_functions <- list()

  for (k in K:1) {
    if (is.null(folds)) {
      q_step_k <- q_step(
        policy_data = policy_data,
        k = k,
        full_history = q_full_history,
        Q = Q[, k + 1],
        q_models = q_models
      )
      # getting the Q-function, Q-function values and the ID-index:
      q_functions[[k]] <- q_step_k$q_function
      q_values_k <- q_step_k$q_values
      idx_k <- q_step_k$idx_k
    } else {
      q_step_cf_k <- q_step_cf(
        folds = folds,
        policy_data = policy_data,
        k = k,
        full_history = q_full_history,
        Q = Q[, k + 1],
        q_models = q_models,
        save_cross_fit_models = save_cross_fit_models,
        future_args = future_args
      )
      q_functions_cf[[k]] <- getElement(q_step_cf_k, "q_functions_cf")
      q_values_k <- getElement(q_step_cf_k, "q_values")
      idx_k <- getElement(q_step_cf_k, "idx_k")
      rm(q_step_cf_k)
    }

    # getting the action matrix for stage k:
    stage <- NULL
    A_k <- actions[stage == k, ]$A
    IA_k <- action_matrix(A_k, action_set)
    rm(stage)

    # calculating the Z-matrix
    Z_1 <- Q_k <- as.matrix(q_values_k[, q_cols, with = FALSE])
    Z_2 <- (IA_k / G[idx_k, k]) * (Q[idx_k, k + 1] - Q_k)
    Z_3 <- 0
    if (k != K) {
      for (r in (k + 1):K) {
        Z_3 <- Z_3 + ipw_weight(II[idx_k, (k + 1):r], G[idx_k, (k + 1):r]) *
          (Q[idx_k, r + 1] - Q[idx_k, r])
      }
      Z_3 <- (IA_k / G[idx_k, k]) * Z_3
    }
    Z <- Z_1 + Z_2 + Z_3
    colnames(Z) <- action_set

    # getting the history for the QV model:
    qv_history_k <- get_history(policy_data,
      stage = k,
      full_history = full_history
    )
    # fitting the QV-function:
    if (is.list(qv_models)) {
      qv_model_k <- qv_models[[k]]
    } else {
      qv_model_k <- qv_models
    }
    qv_function_k <- fit_QV_function(qv_history_k,
      Z = Z,
      qv_model = qv_model_k,
      valid_ids = valid_ids
    )
    qv_functions[[k]] <- qv_function_k
    # getting the QV-function values:
    qv_values_k <- predict(qv_function_k, new_history = qv_history_k)

    if (alpha != 0) {
      # getting the g-function values for each action:
      g_values_k <- g_values[stage == k, ]

      # calculating the realistic actions:
      realistic_actions <- t(apply(g_values_k[, g_cols, with = FALSE],
        MARGIN = 1,
        function(x) x >= alpha
      ))
      # checking that a realistic action exists:
      if (any(apply(realistic_actions, MARGIN = 1, sum) == 0)) {
        mes <- paste0(
          "Cases with no realistic actions occur.",
          " Consider resetting the alpha level."
        )
        stop(mes)
      }
      realistic_actions[which(realistic_actions == FALSE)] <- NA

      # getting the action with the maximal realistic QV-function value:
      dd <- apply(
        qv_values_k[, qv_cols, with = FALSE] * realistic_actions,
        MARGIN = 1,
        which.max
      )
    } else {
      dd <- apply(qv_values_k[, qv_cols, with = FALSE], MARGIN = 1, which.max)
    }

    d <- action_set[dd]

    q_d_k <- get_a_values(a = d, action_set = action_set, q_values_k)$P
    Q[idx_k, k] <- q_d_k
    Q[!idx_k, k] <- Q[!idx_k, k + 1]
    II[idx_k, k] <- (A_k == d)
    II[!idx_k, k] <- TRUE
    D[idx_k, k] <- d
    G[!idx_k, k] <- TRUE
  }

  if (length(q_functions) > 0) {
    class(q_functions) <- "nuisance_functions"
    attr(q_functions, "full_history") <- q_full_history
    names(q_functions) <- paste("stage_", 1:K, sep = "")
  } else {
    q_functions <- NULL
  }
  if (length(q_functions_cf) == 0) {
    q_functions_cf <- NULL
  }
  class(qv_functions) <- "nuisance_functions"
  attr(qv_functions, "full_history") <- full_history
  names(qv_functions) <- paste("stage_", 1:K, sep = "")

  out <- list(
    qv_functions = qv_functions,
    q_functions = q_functions,
    q_functions_cf = q_functions_cf,
    g_functions = g_functions,
    g_functions_cf = g_functions_cf,
    action_set = action_set,
    stage_action_sets = stage_action_sets,
    alpha = alpha,
    K = K,
    folds = folds
  )
  out <- remove_null_elements(out)
  class(out) <- c("drql", "policy_object", "list")

  return(out)
}

#' @rdname get_policy_functions
#' @export
get_policy_functions.drql <- function(object,
                                      stage,
                                      threshold = NULL,
                                      include_g_values = FALSE,
                                      ...) {
  K <- getElement(object, "K")
  check_stage(stage = stage, K = K)
  stage_action_sets <- getElement(object, "stage_action_sets")
  stage_action_set <- stage_action_sets[[stage]]
  rm(stage_action_sets)
  alpha <- getElement(object, "alpha")
  g_functions <- getElement(object, "g_functions")
  if (!is.null(g_functions)) {
    if (length(g_functions) == K) {
      g_function <- g_functions[[stage]]
    } else {
      g_function <- g_functions[[1]]
    }
  }
  qv_functions <- getElement(object, "qv_functions")
  qv_model <- getElement(qv_functions[[stage]], "qv_model")

  stage_policy <- function(H) {
    qv_values <- sapply(
      qv_model,
      function(qvm) {
        predict(qvm, H)
      }
    )

    g_values <- NULL
    if (alpha == 0) {
      dd <- apply(qv_values, MARGIN = 1, which.max)
      d <- stage_action_set[dd]
    } else{
      # evaluating the g-function:
      if (!all(g_function$H_names %in% names(H))) {
        mes <- paste(
          "H must contain the columns",
          paste(g_function$H_names, collapse = ","),
          "."
        )
        stop(mes)
      }
      g_values <- predict(g_function$g_model, new_H = H)

      # calculating the realistic actions:
      realistic_actions <- t(apply(
        g_values,
        MARGIN = 1, function(x) x >= alpha
      ))
      if (any(apply(realistic_actions, MARGIN = 1, sum) == 0)) {
        stop("Cases with no realistic actions occur. Consider resetting the alpha level.")
      }
      realistic_actions[which(realistic_actions == FALSE)] <- NA

      # getting the action with the maximal realistic Q-function value:
      dd <- apply(qv_values * realistic_actions, MARGIN = 1, which.max)
      d <- stage_action_set[dd]
    }

    # including the g_values as attributes:
    if (include_g_values == TRUE)
      attr(d, "g_values") <- g_values

    return(d)
  }

  return(stage_policy)
}

#' @export
get_policy.drql <- function(object, threshold = NULL) {
  if (!(is.null(threshold) || identical(threshold, 0))) {
    stop("threshold is not implemented for drql.")
  }
  g_functions <- get_g_functions(object)
  qv_functions <- getElement(object, "qv_functions")

  action_set <- getElement(object, "action_set")
  K <- getElement(object, "K")
  alpha <- getElement(object, "alpha")

  g_cols <- paste("g_", action_set, sep = "")
  qv_cols <- paste("QV_", action_set, sep = "")

  policy <- function(policy_data) {
    if (get_K(policy_data) != K) {
      stop("The policy do not have the same number of stages as the policy data object.")
    }
    # evaluating the Q-functions:
    qv_values <- predict(qv_functions, policy_data)

    if (alpha != 0) {
      # evaluating the g-functions:
      g_values <- predict(g_functions, policy_data)
      # calculating the realistic actions:
      realistic_actions <- t(apply(
        g_values[, g_cols, with = FALSE],
        MARGIN = 1,
        function(x) x >= alpha
      ))
      if (any(apply(realistic_actions, MARGIN = 1, sum) == 0)) {
        stop("Cases with no realistic actions occur. Consider resetting the alpha level.")
      }
      realistic_actions[which(realistic_actions == FALSE)] <- NA

      # getting the action with the maximal realistic Q-function value:
      dd <- apply(
        qv_values[, qv_cols, with = FALSE] * realistic_actions,
        MARGIN = 1,
        which.max
      )
      d <- action_set[dd]
    } else {
      # getting the action with the maximal Q-function value:
      dd <- apply(
        qv_values[, qv_cols, with = FALSE],
        MARGIN = 1,
        which.max
      )
      d <- action_set[dd]
    }

    # collecting the policy actions
    policy_actions <- get_id_stage(policy_data)
    policy_actions[, d := d]

    return(policy_actions)
  }

  # setting class and attributes:
  policy <- new_policy(policy, name = "drql")

  return(policy)
}

# plot function
plot_policy_object.drql <- function(
    policy_object,
    policy_data,
    id,
    continuous_variable_1,
    range_1,
    continuous_variable_2,
    range_2,
    stage = NULL,
    granularity = 20
){
  id_ <- id
  action_set <- get_action_set(policy_data)
  policy_data <- subset_id(policy_data,
                           id = id_,
                           preserve_action_set = FALSE)
  stage_ <- stage; rm(stage)

  # constructing the grid for the continuous variables:
  seq_variable_1 <- seq(from = range_1[1],
                        to = range_1[2],
                        length.out = granularity)
  seq_variable_2 <- seq(from = range_2[1],
                        to = range_2[2],
                        length.out = granularity)
  grid <- data.table::CJ(seq_variable_1, seq_variable_2)
  names(grid) <- c(continuous_variable_1, continuous_variable_2)

  # getting the number of stages for the given observation:
  K <- get_K(policy_data)

  # setting the default value of 'stage' (all stages):
  if (is.null(stage_))
    stage_ <- 1:K

  # taking the intersection of the observed stages and the input stages:
  stage_ <- intersect(1:K, stage_)
  k <- length(stage_)

  # getting the g-functions
  g_functions <- get_g_functions(policy_object)
  g_full_history <- attr(g_functions, "full_history")

  plot_data <- list()
  plot_data_2 <- list()
  for (s_ in 1:k){

    # full history:
    qv_functions <- getElement(policy_object, "qv_functions")
    full_history <- attr(qv_functions, "full_history")
    stopifnot(full_history == g_full_history)

    # geetting the history matrix
    his <- get_history(policy_data,
                       stage = stage_[s_],
                       full_history = full_history)

    H <- his[["H"]]

    # getting the observed action
    A <- his[["A"]]
    A <- merge(A, H)

    # checks
    stopifnot(
      continuous_variable_1 %in% names(H),
      continuous_variable_2 %in% names(H)
    )

    H <- H[rep(1, each = granularity^2)]
    H[,(continuous_variable_1) := NULL]
    H[,(continuous_variable_2) := NULL]
    H <- cbind(H, grid)

    # getting the policy actions
    pf <- get_policy_functions(policy_object,
                               stage = stage_[s_],
                               include_g_values = TRUE)
    d <- pf(H)

    # getting the realistic actions
    g_values <- attr(d, "g_values")
    colnames(g_values) <- paste("g_", action_set, sep = "")
    alpha <- getElement(policy_object, "alpha")
    realistic_actions <- t(apply(g_values, MARGIN = 1, function(x) x >= alpha))
    realistic <- apply(realistic_actions, 1, prod)

    H <- cbind(H, d = d, realistic = realistic, g_values)

    plot_data[[s_]] <- H
    plot_data_2[[s_]] <- A
  }
  plot_data <- data.table::rbindlist(plot_data)
  plot_data$realistic <- as.factor(as.logical(plot_data$realistic))
  plot_data$d <- factor(plot_data$d, levels = action_set)

  plot_data_2 <- data.table::rbindlist(plot_data_2)
  plot_data_2$A <- factor(plot_data_2$A, levels = action_set)

  # plot of the policy actions:
  p1 <- ggplot2::ggplot(plot_data) +
    ggplot2::geom_point(ggplot2::aes_string(x = {{continuous_variable_1}},
                                            y = {{continuous_variable_2}},
                                            color = "d",
                                            alpha = "realistic")) +
    ggplot2::geom_point(data = plot_data_2, ggplot2::aes_string(x = {{continuous_variable_1}},
                                                                y = {{continuous_variable_2}},
                                                                fill = "A"),
                        size = 2,
                        color = "black",
                        pch=21) +
    ggplot2::scale_alpha_discrete(range = c(0.35, 1)) +
    ggplot2::scale_fill_discrete(drop=FALSE) +
    ggplot2::facet_wrap(~stage, labeller = "label_both") +
    ggplot2::theme_bw()

  return(p1)
}

update_q_formula <- function(formula, data, V_res) {
  tt <- terms(formula, data = data)
  if (length(attr(tt, "term.labels")) == 0) {
    formula <- V_res ~ 1
  } else {
    formula <- reformulate(attr(tt, "term.labels"), response = "V_res")
  }
  return(formula)
}

new_q_model <- function(q_model) {
  class(q_model) <- c("q_model", "function")
  return(q_model)
}

# Q-model documentation ---------------------------------------------------

#' @title q_model class object
#'
#' @description  Use \code{q_glm()}, \code{q_glmnet()}, \code{q_rf()}, and \code{q_sl()} to construct
#' an outcome regression model/Q-model object.
#' The constructors are used as input for [policy_eval()] and [policy_learn()].
#'
#' @param formula An object of class [formula] specifying the design matrix for
#' the outcome regression model/Q-model at the given stage. The action at the
#' given stage is always denoted 'A', see examples. Use
#' [get_history_names()] to see the additional
#' available variable names.
#' @param family A description of the error distribution and link function to
#' be used in the model.
#' @param model (Only used by \code{q_glm}) If \code{FALSE} model frame will
#' not be saved.
#' @param na.action (Only used by \code{q_glm}) A function which indicates what
#' should happen when the data contain NAs, see [na.pass].
#' @param alpha (Only used by \code{q_glmnet}) The elasticnet mixing parameter
#' between 0 and 1. alpha equal to 1 is the lasso penalty, and alpha equal
#' to 0 the ridge penalty.
#' @param s (Only used by \code{q_glmnet}) Value(s) of the penalty parameter
#' lambda at which predictions are required, see [glmnet::predict.glmnet()].
#' @param num.trees (Only used by \code{q_rf}) Number of trees.
#' @param mtry (Only used by \code{q_rf}) Number of variables to possibly split
#'  at in each node.
#' @param cv_args (Only used by \code{q_rf}) Cross-validation parameters.
#' Only used if multiple hyper-parameters are given. \code{K} is the number
#' of folds and
#' \code{rep} is the number of replications.
#' @param SL.library (Only used by \code{q_sl}) Either a character vector of
#' prediction algorithms or a list containing character vectors,
#' see [SuperLearner::SuperLearner].
#' @param env (Only used by \code{q_sl}) Environment containing the learner
#' functions. Defaults to the calling environment.
#' @param onlySL (Only used by \code{q_sl}) Logical. If TRUE, only saves and computes predictions
#' for algorithms with non-zero coefficients in the super learner object.
#' @param discreteSL (Only used by \code{q_sl}) If TRUE, select the model with
#' the lowest cross-validated risk.
#' @param objective (Only used by \code{q_xgboost}) specify the learning
#' task and the corresponding learning objective, see [xgboost::xgboost].
#' @param nrounds (Only used by \code{q_xgboost}) max number of boosting iterations.
#' @param max_depth (Only used by \code{q_xgboost}) maximum depth of a tree.
#' @param learning_rate (Only used by \code{q_xgboost}) learning rate.
#' @param nthread (Only used by \code{q_xgboost}) number of threads.
#' @param ... Additional arguments passed to [glm()], [glmnet::glmnet],
#' [ranger::ranger] or [SuperLearner::SuperLearner].
#' @details
#' \code{q_glm()} is a wrapper of [glm()] (generalized linear model).\cr
#' \code{q_glmnet()} is a wrapper of [glmnet::glmnet()] (generalized linear model via
#' penalized maximum likelihood).\cr
#' \code{q_rf()} is a wrapper of [ranger::ranger()] (random forest).
#' When multiple hyper-parameters are given, the
#' model with the lowest cross-validation error is selected.\cr
#' \code{q_sl()} is a wrapper of [SuperLearner::SuperLearner] (ensemble model).
#' \code{q_xgboost()} is a wrapper of [xgboost::xgboost].
#' @returns q_model object: function with arguments 'AH'
#' (combined action and history matrix) and 'V_res' (residual value/expected
#' utility).
#' @docType class
#' @name q_model
#' @seealso [get_history_names()], [get_q_functions()].
#' @examples
#' library("polle")
#' ### Single stage case
#' d1 <- sim_single_stage(5e2, seed=1)
#' pd1 <- policy_data(d1,
#'                    action="A",
#'                    covariates=list("Z", "B", "L"),
#'                    utility="U")
#' pd1
#'
#' # available history variable names for the outcome regression:
#' get_history_names(pd1)
#'
#' # evaluating the static policy a=1 using inverse
#' # propensity weighting based on the given Q-model:
#' pe1 <- policy_eval(type = "or",
#'                    policy_data = pd1,
#'                    policy = policy_def(1, name = "A=1"),
#'                    q_model = q_glm(formula = ~A*.))
#' pe1
#'
#' # getting the fitted Q-function values
#' head(predict(get_q_functions(pe1), pd1))
#'
#' ### Two stages:
#' d2 <- sim_two_stage(5e2, seed=1)
#' pd2 <- policy_data(d2,
#'                   action = c("A_1", "A_2"),
#'                   covariates = list(L = c("L_1", "L_2"),
#'                                     C = c("C_1", "C_2")),
#'                   utility = c("U_1", "U_2", "U_3"))
#' pd2
#'
#' # available full history variable names at each stage:
#' get_history_names(pd2, stage = 1)
#' get_history_names(pd2, stage = 2)
#'
#' # evaluating the static policy a=1 using outcome
#' # regression based on a glm model for each stage:
#' pe2 <- policy_eval(type = "or",
#'             policy_data = pd2,
#'             policy = policy_def(1, reuse = TRUE, name = "A=1"),
#'             q_model = list(q_glm(~ A * L_1),
#'                            q_glm(~ A * (L_1 + L_2))),
#'             q_full_history = TRUE)
#' pe2
#'
#' # getting the fitted Q-function values
#' head(predict(get_q_functions(pe2), pd2))
NULL

# glm interface --------------------------------------

#' @rdname q_model
#' @export
q_glm <- function(formula = ~ A*.,
                  family = gaussian(),
                  model = FALSE,
                  na.action = na.pass,
                  ...){
  formula <- as.formula(formula)
  dotdotdot <- list(...)

  q_glm <- function(AH, V_res, ...) {
    data <- AH
    check_formula(formula = formula, data = data, call = "q_glm")
    formula <- update_q_formula(formula = formula, data = data, V_res = V_res)
    args_glm <- list(
      formula = formula,
      data = data,
      family = family,
      model = model,
      na.action = na.action
    )

    args_glm <- append(args_glm, dotdotdot)
    model <- tryCatch(
      do.call(what = "glm", args = args_glm),
      error = function(e) e
    )
    if (inherits(model, "error")) {
      model$message <-
        paste0(model$message, " when calling 'q_glm' with formula:\n",
               format(formula))
      stop(model)
    }

    model$call <- NULL
    m <- list(model = model)

    class(m) <- c("q_glm")
    return(m)
  }
  q_glm <- new_q_model(q_glm)
  return(q_glm)
}
#' @export
predict.q_glm <- function(object, new_AH, ...){
  model <- getElement(object, "model")
  pred <- predict(model, newdata = new_AH, type = "response")
  return(pred)
}

# glmnet (Elastic Net) interface ------------------------------------------

#' @rdname q_model
#' @export
q_glmnet <- function(formula = ~ A*.,
                     family = "gaussian",
                     alpha = 1,
                     s = "lambda.min",
                     ...) {
  if (!requireNamespace("glmnet"))
    stop("Package 'glmnet' required.")
  formula <- as.formula(formula)
  dotdotdot <- list(...)

  q_glmnet <- function(AH, V_res, ...) {
    check_formula(formula = formula, data = AH, call = "q_glmnet")
    des <- get_design(formula, data=AH)
    y <- V_res
    args_glmnet <- c(list(y = y, x = des$x,
                        family = family, alpha = alpha), dotdotdot)

    model <- tryCatch(do.call(glmnet::cv.glmnet, args = args_glmnet),
                      error = function(e) e
    )
    if (inherits(model, "error")) {
      model$message <-
        paste0(model$message, " when calling 'q_glmnet' with formula:\n",
               format(formula))
      stop(model)
    }

    model$call <- NULL
    des$x <- NULL

    m  <- list(model = model,
               s = s,
               design = des)

    class(m) <- c("q_glmnet")
    return(m)
  }
  q_glmnet <- new_q_model(q_glmnet)
  return(q_glmnet)
}
#' @export
predict.q_glmnet <- function(object, new_AH, ...) {
  design <- getElement(object, "design")
  model <- getElement(object, "model")
  s <- getElement(object, "s")

  newx <- apply_design(design, data = new_AH)
  pred <- predict(
    model,
    newx = newx,
    type = "response",
    s = s
  )
  pred <- as.vector(pred)
  return(pred)
}

# ranger (Random Forest) interface ----------------------------------------

perf_ranger <- function(fit, data,  ...) {
  y <- as.numeric(data[, 1])
  x <- data[, -1, drop=FALSE]
  mean((predict(fit, data=x, num.threads=1) - y)^2)^.5
}

#' @rdname q_model
#' @export
q_rf <- function(formula = ~.,
                 num.trees=c(250, 500, 750), mtry=NULL,
                 cv_args=list(nfolds=3, rep=1), ...) {
  if (!requireNamespace("ranger")) stop("Package 'ranger' required.")
  formula <- as.formula(formula)
  dotdotdot <- list(...)
  hyper_par <- expand.list(num.trees=num.trees, mtry=mtry)
  rf_args <- function(p) {
    list(num.threads=1, num.trees=p$num.trees, mtry=p$mtry)
  }
  ml <- lapply(hyper_par, function(p)
    function(data) {
      rf_args <- append(rf_args(p), list(y=data[, 1],
                                         x=as.matrix(data[, -1, drop=FALSE])))
      rf_args <- append(rf_args, dotdotdot)
      do.call(ranger::ranger, args=rf_args)
    })

  q_rf <- function(AH, V_res, ...) {
    check_formula(formula = formula, data = AH, call = "q_rf")
    des <- get_design(formula, data=AH)
    data <- data.frame(V_res, des$x)
    res <- NULL; best <- 1
    if (length(ml)>1) {
      res <- tryCatch(targeted::cv(ml, data=data, perf=perf_ranger,
                               nfolds=cv_args$nfolds, rep=cv_args$rep),
                      error=function(...) NULL)
      best <- if (is.null(res)) 1 else which.min(summary(res))
    }
    if (!is.null(res)) {
      model <- res$fit[[best]]
    } else {
      model <- ml[[best]](data)
    }

    model$call <- NULL
    des$x <- NULL

    m <- list(model = model,
              rf_args = rf_args(hyper_par[[best]]),
              num.trees=num.trees[best],
              design = des)
    class(m) <- c("q_rf")
    return(m)
  }
  q_rf <- new_q_model(q_rf)
  return(q_rf)
}
#' @export
predict.q_rf <- function(object, new_AH, ...) {
  model <- getElement(object, "model")
  design <- getElement(object, "design")

  new_data <- apply_design(design = design, data = new_AH)
  pr <- predict(model, data=new_data, num.threads=1)$predictions
  return(pr)
}

# SuperLearner interface --------------------------------------------------

#' @rdname q_model
#' @export
q_sl <- function(formula = ~ .,
                 SL.library=c("SL.mean", "SL.glm"),
                 env = parent.frame(),
                 onlySL = TRUE,
                 discreteSL = FALSE,
                 ...){
  if (!requireNamespace("SuperLearner"))
    stop("Package 'SuperLearner' required.")
  formula <- as.formula(formula)
  force(SL.library)
  force(env)
  dotdotdot <- list(...)
  q_sl <- function(AH, V_res, folds = NULL, ...) {
    check_formula(formula = formula, data = AH, call = "q_sl")
    des <- get_design(formula, data=AH)
    if (missing(V_res) || is.null(V_res))
      V_res <- get_response(formula, data=AH)
    args_SL <- list(Y = as.numeric(V_res),
                    X = as.data.frame(des$x),
                    SL.library = SL.library,
                    env = env)
    args_SL <- append(args_SL, dotdotdot)
    if (!is.null(folds)){
      # given folds, the cvControl argument is overwritten
      cvControl <- SuperLearner.CV.control(
        V = length(folds),
        shuffle = FALSE,
        validRows = folds
      )
      args_SL[["cvControl"]] <- cvControl
    }
    model <- do.call(SuperLearner::SuperLearner, args = args_SL)
    model$call <- NULL
    if(all(model$coef == 0)){
      warning("In q_sl(): All metalearner coefficients are zero. Selecting the learner with the lowest cvrisk.")
      min_idx <- which.min(model$cvRisk)
      coef_ <- model$coef * 0
      coef_[min_idx] <- 1
      model$coef <- coef_
    }

    if (discreteSL == TRUE){
      min_idx <- which.min(model$cvRisk)
      coef_ <- model$coef * 0
      coef_[min_idx] <- 1
      model$coef <- coef_
    }
    if(onlySL == TRUE){
      model$fitLibrary[model$coef == 0] <- NA
    }

    des$x <- NULL
    m <- list(model = model,
              design = des,
              onlySL = onlySL)
    class(m) <- c("q_sl")
    return(m)
  }
  q_sl <- new_q_model(q_sl)
  return(q_sl)
}
#' @export
predict.q_sl <- function(object, new_AH, ...) {
  model <- getElement(object, "model")
  design <- getElement(object, "design")
  onlySL <- getElement(object, "onlySL")
  newdata <- apply_design(design = design, data = new_AH)
  newdata <- as.data.frame(newdata)
  pred <- predict(
         model,
         newdata = newdata,
         onlySL = onlySL
  )$pred[, 1]
  return(pred)
}

# xgboost interface -----------------------------------------------------------------

#' @rdname q_model
#' @export
q_xgboost <- function(formula = ~.,
                      objective = "reg:squarederror",
                      nrounds,
                      max_depth = 6,
                      learning_rate = NULL,
                      nthread = 1,
                      cv_args=list(nfolds=3, rep=1),
                      ...) {
  if (!requireNamespace("xgboost")) stop("Package 'xgboost' required")
  formula <- as.formula(formula)
  environment(formula) <- NULL
  dotdotdot <- list(...)

  ml <- function(formula = V_res ~ .,
                 objective,
                 nrounds,
                 max_depth,
                 learning_rate,
                 nthread,
                 dotdotdot) {

    targeted::learner$new(formula,
                          info = "xgBoost",
                          estimate = function(x, y) {
                            xgb_args <- list(
                              x = quote(x), # avoid embedding data in call
                              y = quote(y),
                              objective = objective,
                              nrounds = nrounds,
                              max_depth = max_depth,
                              learning_rate = learning_rate,
                              nthread = nthread,
                              verbosity = 0,
                              print_every = 0
                            )
                            xgb_args <- append(xgb_args, dotdotdot)
                            do.call(xgboost::xgboost, args = xgb_args)
                          })
  }

  ml_args <- expand.list(
    nrounds = nrounds,
    max_depth = max_depth,
    learning_rate = learning_rate
  )
  cv_par <- ml_args
  ml_args <- lapply(ml_args, function(p) {
    p <- append(p,
                list(
                  objective = objective,
                  dotdotdot = dotdotdot,
                  nthread = nthread
                ))
    return(p)
  })
  ml_models <- lapply(ml_args, function(par) do.call(ml, par))

  q <- function(AH, V_res, ...) {
    ## formatting data:
    check_formula(formula = formula, data = AH, call = "q_xgboost")
    des <- get_design(formula, data=AH)
    if (missing(V_res) || is.null(V_res))
      V_res <- get_response(formula, data=AH)
    data <- data.frame(V_res, des$x)

    cv_res <- NULL
    if (length(ml_models)>1){
      cv_res <- tryCatch(targeted::cv(ml_models, data, nfolds=cv_args$nfolds, rep = cv_args$rep),
                         error = function(e) e)
      if (inherits(cv_res, "error")) {
        cv_res$message <-
          paste0(cv_res$message, " when calling 'q_xgboost' with formula:\n",
                 format(formula))
        stop(cv_res)
      }
      cv_res$names <- unlist(lapply(cv_par, function(x) paste(paste(names(x), x, sep = ":"), collapse = ",")))
      ml_args_best <- ml_args[[which.min(coef(cv_res)[, 1])]]
      model <- do.call(ml, ml_args_best)
      model <- model$estimate(data)
    } else {
      model <- do.call(ml, ml_args[[1]])
      model <- tryCatch(model$estimate(data),
                        error = function(e) e)
      if (inherits(model, "error")) {
        model$message <-
          paste0(model$message, " when calling 'q_xgboost' with formula:\n",
                 format(formula))
        stop(model)
      }
    }

    # setting model output
    des$x <- NULL
    m <- list(model = model,
              design = des,
              cv_res = cv_res,
              cv_par = cv_par)
    class(m) <- c("q_xgboost")
    return(m)
  }
  # setting class:
  q <- new_q_model(q)

  return(q)
}
#' @export
predict.q_xgboost <- function(object, new_AH, ...){
  model <- getElement(object, "model")
  design <- getElement(object, "design")
  new_data <- apply_design(design = design, data = new_AH)

  pred <- predict(model, newdata = new_data)
  return(pred)
}

## used for testing:
q_degen <- function(var) {
  force(var)
  q_degen <- function(AH, V_res, ...) {
    m <- list(var = var)
    class(m) <- "q_degen"
    return(m)
  }
  q_degen <- new_q_model(q_degen)
  return(q_degen)
}

#' @export
predict.q_degen <- function(object, new_AH, ...) {
  var <- getElement(object, "var")
  pred <- unname(unlist(new_AH[, var, with = FALSE]))
  return(pred)
}

test_that("fit_q_functions handles varying stage-action sets", {

  d <- sim_two_stage_multi_actions(n = 1e2)
  expect_error(
    pd <- policy_data(data = d,
                      action = c("A_1", "A_2"),
                      baseline = c("B", "BB"),
                      covariates = list(L = c("L_1", "L_2"),
                                        C = c("C_1", "C_2")),
                      utility = c("U_1", "U_2", "U_3")),
    NA
  )

  p <- policy_def(
    c("yes", "no")
  )

  expect_error(
    qfit <- fit_Q_functions(pd, policy_actions = p(pd), q_models = list(q_glm(), q_glm()), full_history = FALSE, m_function = NULL),
    NA
  )
  expect_error(
    tmp <- predict(qfit, pd),
    NA
  )

  tmp2 <- predict(qfit$stage_1$q_model$model,
                  newdata = cbind(A = "yes", get_history(pd, stage = 1)$H[1,]))
  tmp2 <- tmp2 + get_history(pd, stage = 1)$U$U_bar[1]
  expect_equal(unname(tmp2), unname(unlist(tmp[1, "Q_yes"])))

  tmp2 <- predict(qfit$stage_1$q_model$model,
                  newdata = cbind(A = "no", get_history(pd, stage = 1)$H[1,]))
  tmp2 <- tmp2 + get_history(pd, stage = 1)$U$U_bar[1]
  expect_equal(unname(tmp2), unname(unlist(tmp[1, "Q_no"])))

})

test_that("q_models checks formula input", {
  d <- sim_two_stage(2e3, seed=1)
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))

  p_dynamic <- policy_def(
    policy_functions = list(function(L_1) (L_1>0)*1,
                            function(C_2) (C_2>0)*1),
    reuse = FALSE
  )

  ## glm
  expect_error(policy_eval(policy_data = pd,
                           policy = p_dynamic,
                           q_models = q_glm(formula = Y~X)),
               "variable 'X' is not found in data when calling 'q_glm' with formula: ~X")
  expect_no_error(policy_eval(policy_data = pd,
                           policy = p_dynamic,
                           q_models = q_glm(formula = ~ns(C))))

  ## xgboost
  expect_error(policy_eval(policy_data = pd,
                           policy = p_dynamic,
                           q_models = q_xgboost(formula = Y~X, nrounds = 2)),
               "variable 'X' is not found in data when calling 'q_xgboost' with formula: ~X")

})

test_that("q_models checks formula input 2", {

  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
                    action = "A",
                    covariates = list("Z", "B", "L"),
                    utility = "U"
                    )

  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_glm(formula = ~ A*.))
  )
  expect_error(
    po <- pl(pd, q_models = q_glm(), g_models = g_glm()),
    "Error in blip_model: variable 'A' is not found in data when calling 'q_glm' with formula: ~A \\+ Z \\+ B \\+ L \\+ A:Z \\+ A:B \\+ A:L"
  )

  ## A in global environment (formula defined in the global environment):
  A <- d$A

  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_glm(formula = ~ A*.))
  )
  expect_error(
    po <- pl(pd, q_models = q_glm(), g_models = g_glm()),
    "Error in blip_model: variable 'A' is not found in data when calling 'q_glm' with formula: ~A \\+ Z \\+ B \\+ L \\+ A:Z \\+ A:B \\+ A:L"
  )

})

test_that("q_rf formats data correctly via the formula",{
  d1 <- sim_single_stage(200, seed=1)
  d1$BB <- sample(c("group 1", "group & 2", "group & 3"), size = 200, replace = TRUE)
  pd1 <- policy_data(d1,
                     action="A",
                     covariates = list("Z", "B", "L", "BB"),
                     utility="U")

  expect_error(
    pe <- policy_eval(
      policy_data = pd1,
      policy_learn = policy_learn(type = "ql", alpha = 0.05),
      g_models = g_glm(),
      g_full_history = FALSE,
      q_models = q_rf()
    ),
    NA
  )
})

test_that("q_sl formats data correctly via the formula",{
  d1 <- sim_single_stage(200, seed=1)
  d1$BB <- sample(c("group 1", "group & 2", "group & 3"), size = 200, replace = TRUE)
  pd1 <- policy_data(d1,
                     action="A",
                     covariates = list("Z", "B", "L", "BB"),
                     utility="U")

  expect_error(
    pe <- policy_eval(
      policy_data = pd1,
      policy_learn = policy_learn(type = "ql", alpha = 0.05),
      g_models = g_glm(),
      g_full_history = FALSE,
      q_models = q_sl()
    ),
    NA
  )

})

test_that("q_sl can find user-defined learners",{
  library("polle")
  d <- sim_single_stage(200, seed=1)
  d$BB <- sample(c("group 1", "group & 2", "group & 3"), size = 200, replace = TRUE)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L", "BB"),
                    utility="U")
  p <- policy_def(1)

  env <- as.environment("package:SuperLearner")
  env <- new.env(parent = env)
  with(env,{
    SL.test <- function (Y, X, newX, family, obsWeights, model = TRUE, ...){
      if (is.matrix(X)) {
        X = as.data.frame(X)
      }
      fit.glm <- glm(Y ~ ., data = X, family = family, weights = obsWeights,
                     model = model)
      if (is.matrix(newX)) {
        newX = as.data.frame(newX)
      }
      pred <- predict(fit.glm, newdata = newX, type = "response")
      fit <- list(object = fit.glm)
      class(fit) <- "SL.glm" # SL.test
      out <- list(pred = pred, fit = fit)
      return(out)
    }
  })

  expect_error(
    qfun <- polle:::fit_Q_functions(pd,
                                    p(pd),
                                    q_sl(SL.library = "SL.test", env = env),
                                    m_function = NULL),
    NA)

})

test_that("q_glm and q_sl(SL.library('SL.glm')) are (almost) equivalent",{
  library("SuperLearner")
  d1 <- sim_single_stage(200, seed=1)
  d1$A <- as.character(d1$A)
  d1$BB <- sample(c("group 1", "group_2", "G & 4"), size = nrow(d1), replace = TRUE)

  q1 <- q_glm(formula = ~.)
  q2 <- q_sl(formula = ~., SL.library = "SL.glm")

  q1 <- q1(AH = d1[,c("A", "B", "Z", "L", "BB")], V_res = d1$U)
  q2 <- q2(AH = d1[,c("A", "B", "Z", "L", "BB")], V_res = d1$U)

  # names are different
  expect_equal(
    unname(coef(q1$model)),
    unname(coef(q2$model$fitLibrary$SL.glm_All$object))
  )

  q1 <- q_glm(formula = ~.)
  q2 <- q_sl(formula = ~., SL.library = "SL.glm")
  d1$B <- as.character(d1$B)
  pd1 <- policy_data(d1, action = "A", covariates = c("Z", "L", "B", "BB"), utility = "U")
  pe1 <- policy_eval(policy_data = pd1,
              policy = policy_def(1),
              type = "or",
              q_models = q1)
  pe2 <- policy_eval(policy_data = pd1,
                     policy = policy_def(1),
                     type = "or",
                     q_models = q2)

  # names are different
  expect_equal(
    unname(coef(pe1$q_functions$stage_1$q_model$glm_model)),
    unname(coef( pe2$q_functions$stage_1$q_model$fit$fitLibrary$SL.glm_All$object))
  )
})


# missing data ------------------------------------------------------------


test_that("q_glm handles missing covariates", {
  d <- sim_two_stage(2e3, seed=1)
  d$C_1 <- NULL
  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c(NA, "C_2")), # C_1 is missing
                    utility = c("U_1", "U_2", "U_3"))
  p <- policy_def(1, reuse = TRUE)

  expect_error(
    policy_eval(policy_data = pd,
                policy = p),
    "NA/NaN/Inf in 'x'"
  )
  expect_error(
    policy_eval(policy_data = pd,
                policy = p,
                type = "or",
                q_models = q_glm(~L)),
    NA
  )
  expect_error(
    policy_eval(policy_data = pd,
                policy = p,
                type = "or",
                q_models = list(q_glm(~L), q_glm())),
    NA
  )
})

test_that("policy_learn with type blip works as intended", {

  d <- sim_two_stage(200, seed=1)

  pd <- policy_data(d,
                    action = c("A_1", "A_2"),
                    baseline = c("BB", "B"),
                    covariates = list(L = c("L_1", "L_2"),
                                      C = c("C_1", "C_2")),
                    utility = c("U_1", "U_2", "U_3"))


  # default settings:
  pl <- policy_learn(type = "blip",
                     control = control_blip())
  expect_true({
    po <- pl(
      policy_data = pd,
      g_models = g_glm(),
      q_models = q_glm()
    )

    inherits(po, "policy_object")
  })

  # realistic policy:
  pl <- policy_learn(type = "blip",
                     control = control_blip(),
                     alpha = 0.2)
  expect_true({
    po <- pl(
      policy_data = pd,
      g_models = g_glm(),
      q_models = q_glm()
    )

    inherits(po, "policy_object")
  })

  # predictions from the blip model:
  expect_true({
    blip <-predict(po$blip_functions, pd)
    !any(is.na(blip$blip))
  })
  expect_true({
    all(is.numeric(blip$blip))
  })

  # folds passed to the blip model:
  pl <- policy_learn(type = "blip",
                     L = 2,
                     control = control_blip(
                       blip_models = q_sl()
                     ))
  expect_true({
    po <- pl(
      policy_data = pd,
      g_models = g_glm(),
      q_models = q_glm()
    )
    t1 <- inherits(po, "policy_object")
    t2 <- po$blip_functions$stage_1$blip_model$model$cvControl$V == 2
    t3 <- all.equal(po$folds, po$blip_functions$stage_1$blip_model$model$cvControl$validRows)

    all(c(t1, t2, t3))
  })

})

test_that("get_policy_functions.blip returns a list of policy functions", {
  d <- sim_two_stage(200, seed = 1)
  d$A_2 <- paste(d$A_2, "test", sep = "")
  d$A_1 <- as.character(d$A_1)
  pd <- policy_data(d,
    action = c("A_1", "A_2"),
    baseline = c("BB", "B"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )

  pl <- policy_learn(
    type = "blip",
    control = control_blip()
  )
  po <- pl(
    policy_data = pd,
    g_models = list(g_glm(), g_glm()),
    q_models = q_glm()
  )

  expect_error(
    pf1 <- get_policy_functions(po, stage = 1),
    NA
  )
  his <- get_history(pd, stage = 1)
  H <- get_H(his)

  expect_true(
    all(pf1(H) %in% get_stage_action_sets(pd)[[1]])
  )

  # realistic action set at level alpha:
  pl <- policy_learn(
    type = "blip",
    control = control_blip(),
    alpha = 0.2
  )
  po <- pl(
    policy_data = pd,
    g_models = list(g_glm(), g_glm()),
    q_models = q_glm()
  )

  expect_error(
    pf1 <- get_policy_functions(po, stage = 1),
    NA
  )
  his <- get_history(pd, stage = 1)
  H <- get_H(his)

  expect_true(
    all(pf1(H) %in% get_stage_action_sets(pd)[[1]])
  )

  expect_error(
    pf2 <- get_policy_functions(po, stage = 2),
    NA
  )
  his <- get_history(pd, stage = 2)
  H <- get_H(his)

  expect_true(
    all(pf2(H) %in% get_stage_action_sets(pd)[[2]])
  )

  expect_error(
    get_policy_functions(po),
    "stage argument is missing"
  )
})


test_that("get_policy.blip returns a policy", {

  d <- sim_two_stage(200, seed = 1)
  d$A_2 <- paste(d$A_2, "test", sep = "")
  d$A_1 <- as.character(d$A_1)
  pd <- policy_data(d,
    action = c("A_1", "A_2"),
    baseline = c("BB", "B"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = c("U_1", "U_2", "U_3")
  )

  pl <- policy_learn(
    type = "blip",
    control = control_blip()
  )
  po <- pl(
    policy_data = pd,
    g_models = list(g_glm(), g_glm()),
    q_models = q_glm()
  )

  expect_true({
    p <- get_policy(po)
    inherits(p, "policy")
  })

  expect_true(
    is.data.table(p(pd))
  )

  # realistic action set at level alpha:
  pl <- policy_learn(
    type = "blip",
    control = control_blip(),
    alpha = 0.2
  )
  po <- pl(
    policy_data = pd,
    g_models = list(g_glm(), g_glm()),
    q_models = q_glm()
  )

  expect_true({
    p <- get_policy(po)
    inherits(p, "policy")
  })

  pa <- p(pd)
  expect_true(is.data.table(pa))
  expect_true(all(pa$d %in% get_action_set(pd)))
  expect_true(all(key(pa) == c("id", "stage")))
  expect_true(inherits(pa$d, "character"))

  # comparing get_policy_function and get_policy:
  pf1 <- get_policy_functions(po, stage = 1)
  his1 <- get_history(pd, stage = 1)
  H1 <- get_H(his1)

  expect_equal(
    pf1(H1),
    pa[stage == 1, ]$d
  )

  pf2 <- get_policy_functions(po, stage = 2)
  his2 <- get_history(pd, stage = 2)
  H2 <- get_H(his2)

  expect_equal(
    pf2(H2),
    pa[stage == 2, ]$d
  )
})

test_that("policy_learn with type blip is persistent in the single stage case.", {
  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
    action = "A",
    covariates = list("Z", "B", "L"),
    utility = "U"
  )

  pl <- policy_learn(
    type = "blip",
    control = control_blip()
  )
  po <- pl(pd, q_models = q_glm(), g_models = g_glm())

  ## his <- get_history(pd)
  ## pred <- predict.blip_function(po$blip_functions$stage_1, new_history = his)
  ## hist(x = pred$blip, breaks = 100)

  d <- get_policy(po)(pd)$d

  fix <- c(
    1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0,
    0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0,
    1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
    0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1,
    0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1,
    1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0,
    1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0,
    0, 0
  )
  fix <- as.character(fix)

  expect_equal(
    d,
    fix
  )
})

test_that("policy_learn with type blip passes the threshold argument in the single-stage case,", {
  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
    action = "A",
    covariates = list("Z", "B", "L"),
    utility = "U"
  )

  pl <- policy_learn(
    type = "blip",
    threshold = c(1, 0.5),
    control = control_blip()
  )
  po <- pl(pd, q_models = q_glm(), g_models = g_glm())

  expect_equal(
    po$threshold,
    c(0.5, 1)
  )

  pl <- policy_learn(
    type = "blip",
    threshold = c(3),
    control = control_blip()
  )
  po2 <- pl(pd, q_models = q_glm(), g_models = g_glm())

  ## the choosen treshold(s) should not influence the blip-function:
  expect_equal(
    po$blip_functions,
    po2$blip_functions
  )
})

test_that("get_policy.blip uses the threshold argument", {
  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
    action = "A",
    covariates = list("Z", "B", "L"),
    utility = "U"
  )

  pl <- policy_learn(
    type = "blip",
    threshold = 1,
    control = control_blip()
  )
  po <- pl(pd, q_models = q_glm(), g_models = g_glm())

  his <- get_history(pd)
  pred <- predict.blip_function(po$blip_functions$stage_1, new_history = his)
  ## hist(x = pred$blip, breaks = 100)

  d <- get_policy(po)(pd)$d

  dd <- c("0", "1")[(pred$blip > 1) + 1]
  dd0 <- c("0", "1")[(pred$blip > 0) + 1]

  expect_equal(
    d,
    dd
  )

  d <- get_policy(po, threshold = 1)(pd)$d

  expect_equal(
    d,
    dd
  )

  d <- get_policy(po, threshold = 0)(pd)$d

  expect_true(
    any(!(d == dd))
  )

  d <- get_policy(po, threshold = c(0, 1))[[2]](pd)$d

  expect_equal(
    d,
    dd
  )

  d <- get_policy(po, threshold = c(0, 1))[[1]](pd)$d

  expect_equal(
    d,
    dd0
  )
})

test_that("policy_learn with type blip uses the threshold argument in the two-stage case,", {
  d <- sim_two_stage(n = 1e2, seed = 1)
  d$U <- d$U_1 + d$U_2 + d$U_3 - 2 * d$A_2
  pd <- policy_data(d,
    action = c("A_1", "A_2"),
    covariates = list(
      L = c("L_1", "L_2"),
      C = c("C_1", "C_2")
    ),
    utility = "U"
  )

  pl <- policy_learn(
    type = "blip",
    threshold = 1,
    control = control_blip()
  )
  po <- pl(pd, q_models = q_glm(), g_models = g_glm())

  ## first stage:
  his <- get_history(pd, stage = 1)
  pred <- predict.blip_function(po$blip_functions$stage_1, new_history = his)
  ## hist(x = pred$blip, breaks = 100)
  dd <- c("0", "1")[(pred$blip > 1) + 1] # use the threshold in the first stage
  d <- get_policy(po)(pd)[stage == 1]$d

  expect_equal(
    d,
    dd
  )

  ## second stage:
  his <- get_history(pd, stage = 2)
  pred <- predict.blip_function(po$blip_functions$stage_2, new_history = his)
  ## hist(x = pred$blip, breaks = 100)
  dd <- c("0", "1")[(pred$blip > 0) + 1] # usual optimal policy in the second stage
  d <- get_policy(po)(pd)[stage == 2]$d

  expect_equal(
    d,
    dd
  )
})

test_that("get_policy and get_policy_functions agree with type blip and a non-zero threshold argument,", {
  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
    action = "A",
    covariates = list("Z", "B", "L"),
    utility = "U"
  )

  pl <- policy_learn(
    type = "blip",
    threshold = 1,
    control = control_blip()
  )
  po <- pl(pd, q_models = q_glm(), g_models = g_glm())


  ## pred <- predict.blip_function(po$blip_functions$stage_1, new_history = his)
  ## hist(x = pred$blip, breaks = 100)

  d <- get_policy(po)(pd)$d
  pf <- get_policy_functions(po, stage = 1)
  his <- get_history(pd)
  H <- get_H(his)
  dd <- pf(H)

  expect_equal(
    d,
    dd
  )

  d <- get_policy(po, threshold = 0)(pd)$d
  pf <- get_policy_functions(po, stage = 1, threshold = 0)
  his <- get_history(pd)
  H <- get_H(his)
  dd <- pf(H)

  expect_equal(
    d,
    dd
  )

  expect_error(
    get_policy_functions(po, threshold = c(0, 1)),
    "Set single threshold"
  )

})

test_that("get_policy.blip() returns multiple policies when given multiple thresholds.", {
  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
                    action = "A",
                    covariates = list("Z", "B", "L"),
                    utility = "U"
                    )

  pl <- policy_learn(
    type = "blip",
    threshold = c(0, 1, 2),
    control = control_blip()
  )
  po <- pl(pd, q_models = q_glm(), g_models = g_glm())

  pol_list <- get_policy(po)

  expect_equal(
    lapply(pol_list, function(x) attr(x, "name")) |> unlist(),
    paste0("blip(eta=", c(0,1,2), ")")
  )

  expect_true(
    is.list(pol_list),
    length(pol_list) == 3,
    all(unlist(lapply(pol_list, function(x) inherits(x, what = "policy"))))
  )

  expect_no_error(
    pol_list[[1]](pd)
  )
  expect_no_error(
    pol_list[[2]](pd)
  )
  expect_no_error(
    pol_list[[3]](pd)
  )


  his <- get_history(pd)
  pred <- predict.blip_function(po$blip_functions$stage_1, new_history = his)
  ## hist(x = pred$blip, breaks = 100)

  d <- pol_list[[2]](pd)$d
  dd <- c("0", "1")[(pred$blip > 1) + 1]

  expect_equal(
    d,
    dd
  )
})

test_that("policy_learn with type blip uses the quantile_threshold argument.", {

  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
    action = "A",
    covariates = list("Z", "B", "L"),
    utility = "U"
    )

  ## policy value

  pl <- policy_learn(
    type = "blip",
    threshold = c(1),
    control = control_blip(quantile_prob_threshold = c(0.25, 0.5))
  )
  po <- pl(pd, q_models = q_glm(), g_models = g_glm())
  pe <- policy_eval(
    policy_data = pd,
    policy_learn = pl
  )

  pl_ref <- policy_learn(
    type = "blip",
    threshold = pe$policy_object$threshold,
    control = control_blip()
  )
  pe_ref <- policy_eval(
    policy_data = pd,
    policy_learn = pl_ref
  )

  expect_equal(
    estimate(pe),
    estimate(pe_ref)
  )

  pl <- policy_learn(
    type = "blip",
    threshold = c(1),
    control = control_blip(quantile_prob_threshold = c(0.25, 1.1))
  )

  expect_error(
    policy_eval(policy_data = pd, policy_learn = pl),
    "quantile_prob_threshold must be numeric vector with values in"
  )

})

test_that("policy_learn with type 'blip' has the expected output for a fixed two-stage case with a single final utility outcome.", {

  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  x3 <- runif(n = 1e2, min = -1, max = 1)
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  x3 = x3,
                  y = y,
                  p1 = p1,
                  p2 = p2)


  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ##
  ## no cross-fitting
  ##

  ## policy_learn(type = "blip")):
  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_glm(~x))
  )

  expect_no_error(
    po <- pl(pd,
             q_models = list(q_glm(~x), q_degen(var = "x")),
             g_functions = gf)
  )

  ## stage 2

  Z2 <- cbind(
    d$x2 +
    (d$a2 == "3") / (0.5 * (d$a2 == "3") + 0.5 * (d$a2 == "4")) * (d$y - d$x2),
    d$x2 +
    (d$a2 == "4") / (0.5 * (d$a2 == "3") + 0.5 * (d$a2 == "4")) * (d$y - d$x2)
  )

  blip2 <- Z2[,2] - Z2[,1]

  blip_model2 <- glm(blip2 ~ d$x2)

  expect_equal(
    po$blip_functions$stage_2$blip_model$model |> coef() |> unname(),
    blip_model2 |> coef() |> unname()
  )

  d2 <- ifelse(predict(blip_model2, newdata = data.frame(x2 = d$x2)) > 0, "4", "3")

  expect_equal(
    unname(d2),
    get_policy(po)(pd)[stage == 2]$d
  )

  q1 <- predict(glm(d$x2 ~ d$x1), type = "response")

  ## stage 1

  Z1 <- cbind(
    q1 +
    (d$a1 == "1") / (0.5 * (d$a1 == "1") + 0.5 * (d$a1 == "2")) * (d$x2 - q1) +
    (d$a1 == "1") / (0.5 * (d$a1 == "1") + 0.5 * (d$a1 == "2")) *
    (d$a2 == d2) / (0.5 * (d$a2 == "3") + 0.5 * (d$a2 == "4")) * (d$y - d$x2),
    q1 +
    (d$a1 == "2") / (0.5 * (d$a1 == "1") + 0.5 * (d$a1 == "2")) * (d$x2 - q1) +
    (d$a1 == "2") / (0.5 * (d$a1 == "1") + 0.5 * (d$a1 == "2")) *
    (d$a2 == d2) / (0.5 * (d$a2 == "3") + 0.5 * (d$a2 == "4")) * (d$y - d$x2)
  )
  blip1 <- Z1[,2] - Z1[,1]

  blip_model1 <- glm(blip1 ~ d$x1)

  expect_equal(
    po$blip_functions$stage_1$blip_model$model |> coef() |> unname(),
    blip_model1 |> coef() |> unname()
  )

  d1 <- ifelse(predict(blip_model1, newdata = data.frame(x1 = d$x1)) > 0, "2", "1")

  expect_equal(
    unname(d1),
    get_policy(po)(pd)[stage == 1]$d
  )

  rm(list = ls()[!(ls() %in% c("pd", "d", "gf"))])

  ##
  ## cross-fitting
  ##

  ## policy_learn(type = "blip")):
  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_glm(~x)),
    L = 2,
    cross_fit_g_models = FALSE
  )

  expect_no_error(
    po <- pl(pd,
             q_models = list(q_glm(~x), q_degen(var = "x")),
             g_functions = gf)
  )

  ## stage 2

  Z2 <- cbind(
    d$x2 +
    (d$a2 == "3") / (0.5 * (d$a2 == "3") + 0.5 * (d$a2 == "4")) * (d$y - d$x2),
    d$x2 +
    (d$a2 == "4") / (0.5 * (d$a2 == "3") + 0.5 * (d$a2 == "4")) * (d$y - d$x2)
  )

  blip2 <- Z2[,2] - Z2[,1]

  blip_model2 <- glm(blip2 ~ d$x2)

  expect_equal(
    po$blip_functions$stage_2$blip_model$model |> coef() |> unname(),
    blip_model2 |> coef() |> unname()
  )

  d2 <- ifelse(predict(blip_model2, newdata = data.frame(x2 = d$x2)) > 0, "4", "3")

  expect_equal(
    unname(d2),
    get_policy(po)(pd)[stage == 2]$d
  )

  q1 <- rep(as.numeric(NA), 1e2)
  q1[po$folds[[1]]] <- predict(glm(x2 ~ x1, data = d[-po$folds[[1]],]), type = "response", newdata = d[po$folds[[1]],])
  q1[po$folds[[2]]] <- predict(glm(x2 ~ x1, data = d[-po$folds[[2]],]), type = "response", newdata = d[po$folds[[2]],])

  ## stage 1

  Z1 <- cbind(
    q1 +
    (d$a1 == "1") / (0.5 * (d$a1 == "1") + 0.5 * (d$a1 == "2")) * (d$x2 - q1) +
    (d$a1 == "1") / (0.5 * (d$a1 == "1") + 0.5 * (d$a1 == "2")) *
    (d$a2 == d2) / (0.5 * (d$a2 == "3") + 0.5 * (d$a2 == "4")) * (d$y - d$x2),
    q1 +
    (d$a1 == "2") / (0.5 * (d$a1 == "1") + 0.5 * (d$a1 == "2")) * (d$x2 - q1) +
    (d$a1 == "2") / (0.5 * (d$a1 == "1") + 0.5 * (d$a1 == "2")) *
    (d$a2 == d2) / (0.5 * (d$a2 == "3") + 0.5 * (d$a2 == "4")) * (d$y - d$x2)
  )
  blip1 <- Z1[,2] - Z1[,1]

  blip_model1 <- glm(blip1 ~ d$x1)

  expect_equal(
    po$blip_functions$stage_1$blip_model$model |> coef() |> unname(),
    blip_model1 |> coef() |> unname()
  )

  d1 <- ifelse(predict(blip_model1, newdata = data.frame(x1 = d$x1)) > 0, "2", "1")

  expect_equal(
    unname(d1),
    get_policy(po)(pd)[stage == 1]$d
  )

})


test_that("policy_learn with type 'blip' has the correct outputs for the stochastic two-stage case and a single final utility outcome.", {

  set.seed(1)
  x1 <- runif(n = 1e2, min = -1, max = 1)
  a1 <- c(rep(1, 50), rep(2, 50))
  x2 <- runif(n = 1e2, min = -1, max = 1)
  a2 <- c(rep(3, 50), rep(4, 50))
  y <- a1 * x1 + a2 * x2 + runif(n = 1e2, min = -1, max = 1)
  p1 <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  p2 <- c(rep(3, 25), rep(4, 25), rep(3, 25), rep(4, 25))
  m <- rbinom(n = 1e2, size = 1, prob = 0.5) # stage 2 missing indicator
  d <- data.table(x1 = x1,
                  a1 = a1,
                  x2 = x2,
                  a2 = a2,
                  y = y,
                  p1 = p1,
                  p2 = p2,
                  m = m)
  rm(x1, a1, x2, a2, y)

  pd <- policy_data(
    data = d,
    action = c("a1", "a2"),
    covariates = list(x = c("x1", "x2"),
                      p = c("p1", "p2")),
    utility = c("y")
  )

  gf <- fit_g_functions(pd, g_models = list(g_glm(~1), g_glm(~1)))

  ld <- pd$stage_data
  ld[stage == 2, m := m]
  ld[, m := nafill(m, fill = 0)]
  ld <- ld[m == 0, ]
  ld[ , stage := 1:.N, by = id]
  setkey(ld, id, stage)
  ld[ , m := NULL]

  pd <- policy_data(data = ld, type = "long")

  ##
  ## no cross-fitting
  ##

  ## policy_learn(type = "blip")):
  pl <- policy_learn(
    type = "blip",
    control = control_blip(blip_models = q_glm(~x))
  )

  expect_no_error(
    po <- pl(pd,
             q_models = list(q_glm(~x), q_degen(var = "x")),
             g_functions = gf)
  )

  ## stage2

  p <- policy_def(function(p) p, name = "test", reuse = TRUE)

  ref_pe <- mean(d$x1 +
                 (d$a1 == d$p1) / 0.5 * (((d$m == 0) * d$x2 + (d$m == 1) * d$y) - d$x1) +
                 (d$m == 0) * (d$a1 == d$p1) / 0.5 * (d$a2 == d$p2) / 0.5 * (d$y - d$x2))

  ref_IC <-  (d$x1 +
              (d$a1 == d$p1) / 0.5 * (((d$m == 0) * d$x2 + (d$m == 1) * d$y) - d$x1) +
              (d$m == 0) * (d$a1 == d$p1) / 0.5 * (d$a2 == d$p2) / 0.5 * (d$y - d$x2) -
              ref_pe)

  ref_pe_or <- mean(d$x1)

  ref_pe_ipw <- mean((d$a1 == d$p1) / 0.5 * (d$m == 0) * (d$a2 == d$p2) / 0.5 *  d$y)

  ##
  ## no cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    g_functions = gf
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  expect_equal(
    get_element(pe, "coef_or"),
    ref_pe_or
  )

  expect_equal(
    get_element(pe, "coef_ipw"),
    ref_pe_ipw
  )

  ##
  ## cross-fitting
  ##

  pe <- policy_eval(
    policy_data = pd,
    policy = p,
    q_models = list(q_degen(var = "x"), q_degen(var = "x")),
    g_functions = gf,
    M = 2
  )

  expect_equal(
    pe$name,
    names(coef(pe))
  )
  expect_equal(
    pe$name,
    c("E[U(d)]: d=test")
  )

  expect_equal(
    coef(pe) |> unname(),
    ref_pe
  )

  expect_equal(
    IC(pe),
    matrix(ref_IC)
  )

  expect_equal(
    get_element(pe, "coef_or"),
    ref_pe_or
  )

  expect_equal(
    get_element(pe, "coef_ipw"),
    ref_pe_ipw
  )


})

#' @title Control arguments for Efficient Augmentation and Relaxation Learning
#' @description \code{control_earl} sets the default control arguments
#' for efficient augmentation and relaxation learning , \code{type = "earl"}.
#' The arguments are passed directly to [DynTxRegime::earl()] if not
#' specified otherwise.
#' @param moPropen Propensity model of class "ModelObj", see [modelObj::modelObj].
#' @param moMain Main effects outcome model of class "ModelObj".
#' @param moCont Contrast outcome model of class "ModelObj".
#' @param regime An object of class [formula] specifying the design of the policy/regime.
#' @param iter Maximum number of iterations for outcome regression.
#' @param fSet A function or NULL defining subset structure.
#' @param lambdas Numeric or numeric vector. Penalty parameter.
#' @param cvFolds Integer. Number of folds for cross-validation of the parameters.
#' @param surrogate The surrogate 0-1 loss function. The options are
#' \code{"logit"}, \code{"exp"}, \code{"hinge"}, \code{"sqhinge"}, \code{"huber"}.
#' @param kernel The options are \code{"linear"}, \code{"poly"}, \code{"radial"}.
#' @param kparam Numeric. Kernel parameter
#' @param verbose Integer.
#' @returns list of (default) control arguments.
#' @export
control_earl <- function(moPropen,
                         moMain,
                         moCont,
                         regime,
                         iter = 0L,
                         fSet = NULL,
                         lambdas = 0.5,
                         cvFolds = 0L,
                         surrogate = "hinge",
                         kernel = "linear",
                         kparam = NULL,
                         verbose = 0L){
  control <- as.list(environment())
  return(control)
}

dyntxregime_earl <- function(policy_data,
                             alpha,
                             L,
                             moPropen,
                             moMain,
                             moCont,
                             regime,
                             iter,
                             fSet,
                             lambdas,
                             cvFolds,
                             surrogate,
                             kernel,
                             kparam,
                             verbose,
                             ...){
  K <- get_K(policy_data)
  if (K != 1)
    stop("earl is only implemented for single stage problems.")
  n <- get_n(policy_data)
  action_set <- get_action_set(policy_data)
  id_stage <- get_id_stage(policy_data)

  if (!(length(action_set) == 2))
    stop("earl only works for a dichotomous action set.")

  if (alpha != 0)
    stop("alpha must be 0 when using earl")

  if (L != 1)
    stop("L must be 1 when using earl (no cross-fitting)")
  if (any(get_element(policy_data, "cens_indicator")[["indicator"]])){
    stop("policy learning with type 'earl' not implemented under right-censoring/missing outcomes.")
  }

  # getting the observed actions:
  actions <- get_actions(policy_data)

  # getting the observed (complete) utilities:
  utility <- get_utility(policy_data)

  H <- get_H(get_history(policy_data))

  A <- actions[["A"]]
  H <- cbind(A = A, H)


  earl_object <- DynTxRegime::earl(moPropen = moPropen,
                                   moMain = moMain,
                                   moCont = moCont,
                                   data = H,
                                   response = utility[["U"]],
                                   txName = "A",
                                   regime = regime,
                                   iter = iter,
                                   fSet = fSet,
                                   lambdas = lambdas,
                                   cvFolds = cvFolds,
                                   surrogate = surrogate,
                                   kernel = kernel,
                                   kparam = kparam,
                                   verbose = verbose)


  out <- list(
    earl_object = earl_object,
    action_set = action_set,
    K = K
  )
  class(out) <- c("earl","policy_object","list")

  return(out)
}

#' @export
get_policy.earl <- function(object, threshold = NULL){
  if (!(is.null(threshold) || identical(threshold, 0))) {
    stop("threshold is not implemented for earl.")
  }
  earl_object <- getElement(object, "earl_object")
  action_set <- getElement(object, "action_set")
  K <- getElement(object, "K")

  policy <- function(policy_data){
    if (get_K(policy_data) != K)
      stop("The policy do not have the same number of stages as the policy data object.")

    H <- get_H(get_history(policy_data))

    pred <- DynTxRegime::optTx(earl_object, H)
    policy_actions <- get_id_stage(policy_data)
    d <- NULL
    policy_actions[, d := pred$optimalTx]
    setkeyv(policy_actions, c("id", "stage"))

    return(policy_actions)
  }

  # setting class and attributes:
  policy <- new_policy(policy, name = "earl")

  return(policy)
}

score_class <- function(prob, obs, levels=NULL, weights = NULL,
                       metric = c("brier", "logscore"), messages = 0) {
    if (is.null(levels) && length(colnames(prob))==length(levels))
      levels <- colnames(prob)
    if (is.factor(obs) && is.null(levels)) {
      levels <- levels(obs)
    }
    if (is.null(levels)) stop("missing definition of levels")
    cl <- levels
    if (is.factor(obs)) {
      cl.obs <- levels(obs)
    } else {
      cl.obs <- unique(obs)
    }
    newcl <- which(!cl.obs %in% cl)
    if (length(newcl)) {
        if (messages > 0)
            warning("new classes among observations")
        temp <- array(0, dim = c(nrow(prob), length(newcl)))
        colnames(temp) <- cl.obs[newcl]
        prob <- cbind(prob, temp)
    }
    y <- outer(obs, levels, "==")
    Bi <- apply((prob - y)^2, 1, sum)
    li <- apply(log(prob) * y, 1, function(x) sum(x[is.finite(x)],
        na.rm = TRUE))
    if (!is.null(weights)) {
        B <- stats::weighted.mean(Bi, w = weights, na.rm = TRUE)
        L <- stats::weighted.mean(li, w = weights, na.rm = TRUE)
    } else {
        B <- mean(Bi, na.rm = TRUE)
        L <- mean(li, na.rm = TRUE)
    }
    if (tolower(metric[1])%in%"brier") return(B)
    return(L)
}


expand.list <- function(...) {
  dots <- list(...)
  nam <- names(dots)
  nulls <- c()
  for (i in seq_along(dots)) {
    if (is.null(dots[[i]])) {
      dots[[i]] <- NA
      nulls <- c(nulls, i)
    }
  }
  names(dots) <- nam
  dat <- do.call(expand.grid, c(dots, list(KEEP.OUT.ATTRS = FALSE)))
  lapply(seq(NROW(dat)),
         function(i) {
           res <- as.list(dat[i,])
           if (length(nulls)>0) res[nulls] <- list(NULL)
           res
           })
}

fit_blip_function <- function(history, Z, blip_model, valid_ids) {
  action_set <- getElement(history, "action_set")
  stage_action_set <- getElement(history, "stage_action_set")
  stage <- getElement(history, "stage")
  H <- get_H(history)
  stage_id <- unlist(get_id.history(history),
    use.names = FALSE
  )

  folds <- NULL
  if (!is.null(valid_ids)) {
    # adapting the validation ids to the current stage:
    stage_valid_ids <- lapply(
      valid_ids,
      function(vi) {
        vi[vi %in% stage_id]
      }
    )
    folds <-
      lapply(
        stage_valid_ids,
        function(vi) {
          which(stage_id %in% vi)
        }
      )
  }

  # fitting the blip-model:
  z <- Z[, stage_action_set]
  blip <- z[, 2] - z[, 1]

  blip_model <- tryCatch({
    blip_model(AH = H, V_res = blip, folds = folds)
  }, error = function(e) {
    stop("Error in blip_model: ", conditionMessage(e))
  })

  blip_function <- list(
    blip_model = blip_model,
    action_set = action_set,
    stage_action_set = stage_action_set,
    stage = stage
  )
  class(blip_function) <- "blip_function"

  return(blip_function)
}

#' @export
predict.blip_function <- function(object, new_history, ...) {
  id_stage <- get_id_stage(new_history)
  new_H <- get_H(new_history)
  blip_model <- getElement(object, "blip_model")
  blip <- predict(blip_model, new_H)
  blip <- data.table(id_stage, blip = as.vector(unlist(blip)))
  setkeyv(blip, c("id", "stage"))

  return(blip)
}

#' @title Control arguments for doubly robust blip-learning
#' @description \code{control_blip} sets the default control arguments
#' for doubly robust blip-learning, \code{type = "blip"}.
#' @param blip_models Single element or list of V-restricted blip-models created
#' by [q_glm()], [q_rf()], [q_sl()] or similar functions.
#' @param quantile_prob_threshold Numeric vector. Quantile probabilities
#' for adaptively setting the threshold for the conditional average treatment effect.
#' @returns list of (default) control arguments.
#' @export
control_blip <- function(blip_models = q_glm(~.),
                         quantile_prob_threshold = NULL)
{
  control <- as.list(environment())
  return(control)
}

blip <- function(policy_data,
                 alpha,
                 threshold,
                 quantile_prob_threshold,
                 g_models, g_functions, g_full_history,
                 q_models, q_full_history,
                 c_models, c_functions, c_full_history,
                 m_model, m_function, m_full_history,
                 blip_models, full_history,
                 L,
                 cross_fit_g_models,
                 cross_fit_c_models,
                 save_cross_fit_models, future_args,
                 ...) {
  K <- get_K(policy_data)
  n <- get_n(policy_data)
  action_set <- get_action_set(policy_data)
  stage_action_sets <- get_stage_action_sets(policy_data)

  ## input checks:
  model_input_checks(
    policy_data = policy_data,
    g_models = g_models,
    g_functions = g_functions,
    g_full_history = g_full_history,
    q_models = q_models,
    q_functions = NULL,
    q_full_history = q_full_history,
    c_models = c_models,
    c_functions = c_functions,
    c_full_history = c_full_history,
    m_model = m_model,
    m_function = m_function,
    m_full_history = m_full_history
  )
  if(is.null(blip_models))
    stop("blip_models are missing.")
  if (is.list(blip_models)) {
    if (length(blip_models) != K) {
      mes <- paste0(
        "blip_models must either be a list of length K ",
        "or a single blip-model."
      )
      stop(mes)
    }
  }
  if (!all(lapply(stage_action_sets, length) == 2)) {
    mes <- paste0(
      "policy_learn with type 'blip' only works ",
      "for dichotomous stage action sets."
    )
    stop(mes)
  }
  if (!is.null(quantile_prob_threshold)) {
    if (!(is.numeric(quantile_prob_threshold) && (length(quantile_prob_threshold) >= 1))) {
      stop("quantile_prob_threshold must be numeric vector with values in [0,1].")
    }
    if (!all(quantile_prob_threshold >= 0 & quantile_prob_threshold <= 1 )){
      stop("quantile_prob_threshold must be numeric vector with values in [0,1].")
    }
    quantile_prob_threshold <- unique(sort(quantile_prob_threshold))
  }

  ##
  ## collecting values as data.table's:
  ##

  ## getting the id as a data.table:
  id <- data.table(id = get_id(policy_data))
  setkeyv(id, "id")

  ## getting the observed actions as a data.table:
  actions <- get_actions(policy_data)

  ## getting the events:
  events <- get_event(policy_data)

  ## getting the observed (complete) utilities as a data.table:
  utility <- get_utility(policy_data)

  # sampling the folds for cross-fitting:
  folds <- sample_folds(number = L, sample_size = n)

  ## (cross-)fitting the g-functions:
  valid_ids <- NULL
  g_functions_cf <- NULL
  if (!is.null(folds) && isTRUE(cross_fit_g_models)) {
    g_cf <- crossfit_function(
      policy_data = policy_data,
      fun = fit_g_functions,
      models = g_models,
      full_history = g_full_history,
      folds = folds,
      save_cross_fit_models = save_cross_fit_models,
      future_args = future_args
    )
    g_functions_cf <- getElement(g_cf, "functions")
    g_values <- getElement(g_cf, "values")
    valid_ids <- getElement(g_cf, "valid_ids")
    rm(g_cf)
  } else {
    if (is.null(g_functions)) {
      g_functions <- fit_g_functions(policy_data,
                                     g_models = g_models,
                                     full_history = g_full_history)
    }
    g_values <- predict(g_functions, policy_data)
  }

  ## fitting g-functions for determining new realistic actions:
  if (alpha > 0) {
    if (is.null(g_functions)) {
      g_functions <- fit_g_functions(policy_data,
                                     g_models = g_models,
                                     full_history = g_full_history)
    }
  } else {
    ## g-functions are not saved if alpha == 0:
    g_functions <- NULL
  }

  ## (cross-)fitting the c-functions:
  c_functions_cf <- NULL
  c_values <- NULL
  if (!is.null(folds) && isTRUE(cross_fit_c_models) && !is.null(c_models)) {
    c_cf <- crossfit_function(
      policy_data = policy_data,
      fun = fit_c_functions,
      models = c_models,
      full_history = c_full_history,
      folds = folds,
      save_cross_fit_models = save_cross_fit_models,
      future_args = future_args
    )
    c_functions_cf <- getElement(c_cf, "functions")
    c_values <- getElement(c_cf, "values")
    rm(c_cf)
  } else {
    if (is.null(c_functions) && !is.null(c_models)) {
      c_functions <- fit_c_functions(policy_data,
                                     c_models = c_models,
                                     full_history = c_full_history)
    }
    if (!is.null(c_functions)) {
      c_values <- predict(c_functions, policy_data)
    }
  }

  ## (cross-)fitting m-function
  m_functions_cf <- NULL
  m_values <- NULL
  if (!is.null(folds) && !is.null(m_model)) {
    m_cf <- crossfit_function(
      policy_data = policy_data,
      fun = fit_m_function,
      models = m_model,
      full_history = m_full_history,
      folds = folds,
      save_cross_fit_models = save_cross_fit_models,
      future_args = future_args
    )
    m_functions_cf <- getElement(m_cf, "functions")
    m_values <- getElement(m_cf, "values")
    rm(m_cf)
  } else {
    if (is.null(m_function) && !is.null(m_model)) {
      m_function <- fit_m_function(policy_data,
                                   m_model = m_model,
                                   full_history = m_full_history)
    }
    if (!is.null(m_function)) {
      m_values <- predict(m_function, policy_data)
    }
  }

  ##
  ## transforming values into vectors and matrices:
  ##

  ## (n) observed utility vector U:
  U <- utility[["U"]]

  ## (n) vector indicating right-censored outcomes:
  censored_outcomes <- is.na(U)

  ## (n X K+1) matrix with entries Delta_k (cumulative non-censoring indicators):
  D <- event_matrix(events = events, event_set = c(0,1))

  ## (n X K+1) matrix with missing event indicators
  M <- event_matrix(events = events, event_set = c(NA))

  ## (n X K+1) matrix with terminal event indicators
  E <- event_matrix(events = events, event_set = c(1))

  ## (n X K+1) matrix with entries C_k(H_k) (non-censoring probabilities):
  C <- cens_prob_matrix(c_values = c_values,
                        M = M,
                        n = nrow(id),
                        K = K)
  rm(M)

  ## (n X (K+1)) matrix with columns g_k(A_k, H_k)
  G <- action_prob_matrix(g_values = g_values,
                          actions = actions,
                          id = id,
                          D = D,
                          E = E,
                          action_set = action_set)

  ## (n X K+1) matrix with entries I(d_k(H_k) = A_k):
  II <- policy_action_indicator_matrix(actions = NULL,
                                       policy_actions = NULL,
                                       id = id,
                                       D = D,
                                       E = E)

  ## (n X K+2) matrix with entries Q_k(H_{k,i}, d_k(H_{k,i})), Q_{K+2} = U:
  Q <- policy_action_outcome_matrix(q_values = NULL,
                                    m_values = m_values,
                                    U = U,
                                    policy_actions = NULL,
                                    id = id,
                                    D = D,
                                    action_set = action_set)

  g_cols <- paste("g_", action_set, sep = "")
  q_cols <- paste("Q_", action_set, sep = "")

  q_functions <- list()
  q_functions_cf <- list()
  blip_functions <- list()

  for (k in K:1) {
    if (is.null(folds)) {
      q_step_k <- q_step(
        policy_data = policy_data,
        k = k,
        full_history = q_full_history,
        Q = Q[, k + 1],
        q_models = q_models
      )
      # getting the Q-function, Q-function values and the ID-index:
      q_functions[[k]] <- q_step_k$q_function
      q_values_k <- q_step_k$q_values
      idx_k <- q_step_k$idx_k
    } else {
      q_step_cf_k <- q_step_cf(
        folds = folds,
        policy_data = policy_data,
        k = k,
        full_history = q_full_history,
        Q = Q[, k + 1],
        q_models = q_models,
        save_cross_fit_models = save_cross_fit_models,
        future_args = future_args
      )
      q_functions_cf[[k]] <- getElement(q_step_cf_k, "q_functions_cf")
      q_values_k <- getElement(q_step_cf_k, "q_values")
      idx_k <- getElement(q_step_cf_k, "idx_k")
      rm(q_step_cf_k)
    }

    ## action matrix for stage k:
    stage <- NULL
    actions_k <- actions[stage == k, ]
    actions_k <- merge(q_values_k[, .SD, .SDcols = key(q_values_k)], actions_k, all.x = TRUE)
    A_k <- actions_k$A
    IA_k <- action_matrix(A_k, action_set)
    rm(stage)

    # calculating the Z-matrix
    Z_1 <- Q_k <- as.matrix(q_values_k[, q_cols, with = FALSE])
    Z_2 <- (D[idx_k, k] / C[idx_k, k]) * (IA_k / G[idx_k, k]) * (Q[idx_k, k + 1] - Q_k)
    Z_3 <- 0

    for (r in (k + 1):(K+1)) {
      Z_3 <- Z_3 + ipw_weight(D[idx_k, (k + 1):r], C[idx_k, (k + 1):r]) *
        ipw_weight(II[idx_k, (k + 1):r], G[idx_k, (k + 1):r]) *
        (Q[idx_k, r + 1] - Q[idx_k, r])
    }
    Z_3 <- (D[idx_k, k] / C[idx_k, k]) * (IA_k / G[idx_k, k]) * Z_3

    Z <- Z_1 + Z_2 + Z_3
    colnames(Z) <- action_set

    # getting the history for the blip model:
    blip_history_k <- get_history(policy_data,
      stage = k,
      full_history = full_history,
      event_set = c(0,2)
    )
    # fitting the blip-function:
    if (is.list(blip_models)) {
      blip_model_k <- blip_models[[k]]
    } else {
      blip_model_k <- blip_models
    }
    blip_function_k <- fit_blip_function(blip_history_k,
                                         Z = Z,
                                         blip_model = blip_model_k,
                                         valid_ids = valid_ids)
    blip_functions[[k]] <- blip_function_k

    # getting the blip-function values:
    blip_k <- predict(blip_function_k, new_history = blip_history_k)

    ## calculating the the threshold associated with a given
    ## quantile of the fitted blip values
    quantile_threshold <- NULL
    if (!is.null(quantile_prob_threshold)){
      if (k == 1){
        quantile_threshold <- stats::quantile(x = blip_k$blip, probs = quantile_prob_threshold)
      }
      ## overwriting the threshold argument:
      threshold <- unname(quantile_threshold)
    }

    # getting the stage action with a positive blip:
    stage_action_set <- stage_action_sets[[k]]
    if (alpha != 0) {
      # getting the g-function values for each action:
      g_values_k <- g_values[stage == k, ]

      # calculating the realistic actions:
      realistic_actions <- t(apply(g_values_k[, g_cols, with = FALSE],
        MARGIN = 1,
        function(x) x >= alpha
      ))
      realistic_actions <- realistic_actions[, action_set %in% stage_action_set]
      # checking that a realistic action exists:
      if (any(apply(realistic_actions, MARGIN = 1, sum) == 0)) {
        mes <- paste0(
          "Cases with no realistic actions occur. ",
          "Consider resetting the alpha level."
        )
        stop(mes)
      }

      # calculating the optimal actions:
      dd <- (blip_k$blip > 0) + 1

      # overruling unrealistic actions:
      realistic <- apply(realistic_actions, 1, prod)
      dd <- dd * realistic +
        (realistic == 0) * apply(realistic_actions, 1, which.max)
    } else {
      dd <- (blip_k$blip > 0) + 1
    }
    d <- stage_action_set[dd]

    q_d_k <- get_a_values(a = d, action_set = action_set, q_values_k)$P
    Q[idx_k, k] <- q_d_k
    ii <- (A_k == d)
    ii[is.na(A_k)] <- 1
    II[idx_k, k] <- ii
  }


  if (length(q_functions) > 0) {
    class(q_functions) <- "nuisance_functions"
    attr(q_functions, "full_history") <- q_full_history
    names(q_functions) <- paste("stage_", 1:K, sep = "")
  } else{
    q_functions <- NULL
  }
  if (length(q_functions_cf) == 0) {
    q_functions_cf <- NULL
  }
  class(blip_functions) <- "nuisance_functions"
  attr(blip_functions, "full_history") <- full_history
  names(blip_functions) <- paste("stage_", 1:K, sep = "")

  out <- list(
    blip_functions = blip_functions,
    q_functions = q_functions,
    q_functions_cf = q_functions_cf,
    g_functions = g_functions,
    g_functions_cf = g_functions_cf,
    action_set = action_set,
    stage_action_sets = stage_action_sets,
    alpha = alpha,
    threshold = threshold,
    quantile_prob_threshold = quantile_prob_threshold,
    K = K,
    folds = folds,
    Z_1 = Z_1
  )
  out <- remove_null_elements(out)
  class(out) <- c("blip", "policy_object", "list")

  return(out)
}

#' @rdname get_policy_functions
#' @export
get_policy_functions.blip <- function(object,
                                      stage,
                                      threshold = NULL,
                                      include_g_values = FALSE,
                                      ...) {
  threshold_selection <- get_element(object, "threshold")
  threshold <- set_threshold(
    threshold = threshold,
    selection = threshold_selection,
    overwrite = TRUE
  )
  if (length(threshold) > 1) {
    mes <- "Set single threshold."
    stop(mes)
  }
  K <- get_element(object, "K")
  check_stage(stage = stage, K = K)
  stage_action_sets <- get_element(object, "stage_action_sets")
  stage_action_set <- stage_action_sets[[stage]]
  rm(stage_action_sets)
  alpha <- get_element(object, "alpha")
  g_functions <- get_element(object, "g_functions", check_name = FALSE)
  if (!is.null(g_functions)) {
    if (length(g_functions) == K) {
      g_function <- g_functions[[stage]]
    } else {
      g_function <- g_functions[[1]]
    }
  }
  blip_functions <- getElement(object, "blip_functions")
  blip_model <- getElement(blip_functions[[stage]], "blip_model")

  stage_policy <- function(H) {
    blip <- predict(blip_model, H)
    g_values <- NULL

    ## setting a possibly non-zero threshold for the first stage:
    th <- 0
    if (stage == 1) {
      th <- threshold
    }

    if (alpha == 0) {
      dd <- (blip > th) + 1
      d <- stage_action_set[dd]
    } else {
      # evaluating the g-function:
      if (!all(g_function$H_names %in% names(H))) {
        mes <- paste(
          "H must contain the columns",
          paste(g_function$H_names, collapse = ","),
          "."
        )
        stop(mes)
      }
      # g-function values on the stage action set:
      g_values <- predict(g_function$g_model, new_H = H)

      # calculating the realistic actions:
      realistic_actions <- t(
        apply(g_values,
          MARGIN = 1,
          function(x) x >= alpha
        )
      )

      if (any(apply(realistic_actions, MARGIN = 1, sum) == 0)) {
        mes <- paste0(
          "Cases with no realistic actions occur. ",
          "Consider resetting the alpha level."
        )
        stop(mes)
      }

      # calculating the optimal actions:
      dd <- (blip > th) + 1
      realistic <- apply(realistic_actions, 1, prod)
      dd <- dd * realistic +
        (realistic == 0) * apply(realistic_actions, 1, which.max)
      d <- stage_action_set[dd]
    }

    # including the g_values as attributes:
    if (include_g_values == TRUE) {
      attr(d, "g_values") <- g_values
    }

    return(d)
  }

  return(stage_policy)
}

set_threshold <- function(threshold = NULL, selection, overwrite = FALSE) {
  if (is.null(threshold)) {
    return(selection)
  }

  ## check threshold:
  if (!(is.numeric(threshold))) {
    stop("threshold must be numeric (vector).")
  }
  threshold <- unique(sort(threshold))

  if (overwrite == FALSE) {
    if (!all(threshold %in% selection)) {
      mes <- paste(selection, collapse = ", ")
      mes <- paste("Invalid threshold. Choose between ", mes, ".", sep = "")
      stop(mes)
    }
  }
  out <- threshold
  return(out)
}

#' @export
get_policy.blip <- function(object, threshold = NULL) {
  blip_functions <- get_element(object, "blip_functions")
  g_functions <- get_g_functions(object)
  action_set <- get_element(object, "action_set")
  g_cols <- paste("g_", action_set, sep = "")
  stage_action_sets <- get_element(object, "stage_action_sets")
  K <- get_element(object, "K")
  alpha <- get_element(object, "alpha")
  threshold_selection <- get_element(object, "threshold")
  threshold <- set_threshold(
    threshold = threshold,
    selection = threshold_selection,
    overwrite = TRUE
  )

  ## returning a policy for each threshold:
  policy_list <- lapply(
    threshold,
    function(th) {
      policy <- function(policy_data) {
        if (get_K(policy_data) != K) {
          mes <- paste0(
            "The policy do not have the same number of ",
            "stages as the policy data object."
          )
          stop(mes)
        }
        ## evaluating the blip-functions:
        blip_values <- predict(blip_functions, policy_data, event_set = c(0,2))

        ## getting the stage actions (sa) associated with the blip:
        sa <- as.data.table(do.call(what = "rbind", stage_action_sets))
        colnames(sa) <- c("ref_action", "alt_action")
        sa <- cbind(stage = 1:K, sa)
        setkeyv(sa, "stage")

        if (alpha != 0) {
          ## evaluating the g-functions:
          g_values <- predict(g_functions, policy_data)
          ## calculating the realistic actions:
          realistic_actions <- t(apply(
            g_values[, g_cols, with = FALSE],
            MARGIN = 1,
            function(x) x >= alpha
          ))
          if (any(apply(realistic_actions, MARGIN = 1, sum) == 0)) {
            mes <- paste0(
              "Cases with no realistic actions occur. ",
              "Consider resetting the alpha level."
            )
            stop(mes)
          }

          realistic <- apply(
            realistic_actions,
            MARGIN = 1, function(x) action_set[x], simplify = FALSE
          )
          realistic_indicator <- unlist(
            lapply(realistic, function(x) length(x) == 2)
          )
          realistic[realistic_indicator] <- NA
          realistic <- unlist(realistic)

          ## getting the optimal stage action:
          blip_values <- merge(blip_values, sa, by = "stage")
          setkeyv(blip_values, c("id", "stage"))
          alt_action <- ref_action <- ri <- d <- r <- NULL
          blip_values[, "d" := ifelse(blip > 0, alt_action, ref_action)]
          stage <- NULL
          blip_values[
            stage == 1,
            "d" := ifelse(blip > th, alt_action, ref_action)
          ]
          blip_values[, "ri" := realistic_indicator]
          blip_values[, "r" := realistic]
          blip_values[, "d" := ifelse(ri, d, r)]
        } else {
          ## getting the optimal stage action:
          blip_values <- merge(blip_values, sa, by = "stage")
          setkeyv(blip_values, c("id", "stage"))
          blip_values[, "d" := ifelse(blip > 0, alt_action, ref_action)]
          stage <- NULL
          blip_values[
            stage == 1,
            "d" := ifelse(blip > th, alt_action, ref_action)
          ]
        }

        ## collecting the policy actions
        policy_actions <- blip_values[, c("id", "stage", "d")]

        return(policy_actions)
      }

      ## setting class and attributes:
      name <- paste0("blip(eta=", round(th, 3), ")")
      policy <- new_policy(policy, name = name)

      return(policy)
    }
  )

  if (length(policy_list) == 1) {
    out <- policy_list[[1]]
  } else {
    out <- policy_list
  }

  return(out)
}

# plot function
plot_policy_object.blip <- function(
    policy_object,
    policy_data,
    id,
    continuous_variable_1,
    range_1,
    continuous_variable_2,
    range_2,
    stage = NULL,
    granularity = 20) {
  id_ <- id
  action_set <- get_action_set(policy_data)
  policy_data <- subset_id(policy_data,
    id = id_,
    preserve_action_set = FALSE
  )
  stage_ <- stage
  rm(stage)

  # constructing the grid for the continuous variables:
  seq_variable_1 <- seq(
    from = range_1[1],
    to = range_1[2],
    length.out = granularity
  )
  seq_variable_2 <- seq(
    from = range_2[1],
    to = range_2[2],
    length.out = granularity
  )
  grid <- data.table::CJ(seq_variable_1, seq_variable_2)
  names(grid) <- c(continuous_variable_1, continuous_variable_2)

  # getting the number of stages for the given observation:
  K <- get_K(policy_data)

  # setting the default value of 'stage' (all stages):
  if (is.null(stage_)) {
    stage_ <- 1:K
  }

  # taking the intersection of the observed stages and the input stages:
  stage_ <- intersect(1:K, stage_)
  k <- length(stage_)

  # getting the g-functions
  g_functions <- get_g_functions(policy_object)
  g_full_history <- attr(g_functions, "full_history")

  plot_data <- list()
  plot_data_2 <- list()
  for (s_ in 1:k) {
    # full history:
    blip_functions <- getElement(policy_object, "blip_functions")
    full_history <- attr(blip_functions, "full_history")
    stopifnot(full_history == g_full_history)

    # getting the history matrix
    his <- get_history(policy_data,
      stage = stage_[s_],
      full_history = full_history
    )

    H <- his[["H"]]

    # getting the observed action
    A <- his[["A"]]
    A <- merge(A, H)

    # checks
    stopifnot(
      continuous_variable_1 %in% names(H),
      continuous_variable_2 %in% names(H)
    )

    H <- H[rep(1, each = granularity^2)]
    H[, (continuous_variable_1) := NULL]
    H[, (continuous_variable_2) := NULL]
    H <- cbind(H, grid)

    # getting the policy actions
    pf <- get_policy_functions(policy_object,
      stage = stage_[s_],
      include_g_values = TRUE
    )
    d <- pf(H)

    # getting the realistic actions
    g_values <- attr(d, "g_values")
    colnames(g_values) <- paste("g_", action_set, sep = "")
    alpha <- getElement(policy_object, "alpha")
    realistic_actions <- t(apply(g_values, MARGIN = 1, function(x) x >= alpha))
    realistic <- apply(realistic_actions, 1, prod)

    H <- cbind(H, d = d, realistic = realistic, g_values)

    plot_data[[s_]] <- H
    plot_data_2[[s_]] <- A
  }
  plot_data <- data.table::rbindlist(plot_data)
  plot_data$realistic <- as.factor(as.logical(plot_data$realistic))
  plot_data$d <- factor(plot_data$d, levels = action_set)

  plot_data_2 <- data.table::rbindlist(plot_data_2)
  plot_data_2$A <- factor(plot_data_2$A, levels = action_set)

  # plot of the policy actions:
  p1 <- ggplot2::ggplot(plot_data) +
    ggplot2::geom_point(ggplot2::aes_string(
      x = {{ continuous_variable_1 }},
      y = {{ continuous_variable_2 }},
      color = "d",
      alpha = "realistic"
    )) +
    ggplot2::geom_point(
      data = plot_data_2,
      ggplot2::aes_string(
        x = {{ continuous_variable_1 }},
        y = {{ continuous_variable_2 }},
        fill = "A"
      ),
      size = 2,
      color = "black",
      pch = 21
    ) +
    ggplot2::scale_alpha_discrete(range = c(0.35, 1)) +
    ggplot2::scale_fill_discrete(drop = FALSE) +
    ggplot2::facet_wrap(~stage, labeller = "label_both") +
    ggplot2::theme_bw()

  return(p1)
}

test_that("policy_eval_online() uses valid indexes for training and validation:", {

  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  d <- data.table(z = z, a = a, y = y, p = p)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p"),
    utility = c("y")
  )
  p <- policy_def(function(p) p, name = "test")
  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  set.seed(45234)
  pe <- policy_eval_online(
    policy_data = pd,
    policy = p,
    target = "value",
    g_functions = gf,
    q_models = q_degen(var = "z"),
    train_block_size = 60,
    M = 2
  )

  train_sequential_index <- pe$train_sequential_index
  validation_sequential_index <- pe$valid_sequential_index

  ## expecting accumulating training indexes:
  expect_true(
    all(train_sequential_index[[1]] %in% train_sequential_index[[2]])
  )

  ## expecting that the validation indexes are included in the training indexes:
  expect_true(
  all(sort(c(train_sequential_index[[1]], validation_sequential_index[[1]])) == train_sequential_index[[2]])
  )

  ## expecting that all the data is used:
  expect_true(
    all(sort(c(train_sequential_index[[2]], validation_sequential_index[[2]])) == 1:100)
  )

})

test_that("policy_eval_online has the expected output for target = 'value' in the single stage case:", {

  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  d <- data.table(z = z, a = a, y = y, p = p)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p"),
    utility = c("y")
  )
  p <- policy_def(function(p) p, name = "test")
  gf <- fit_g_functions(pd, g_models = g_glm(~1))
  ref_Z <- (d$a == d$p) / 0.5 * (d$y - d$z) + d$z

  ## single policy

  set.seed(1)
  pe <- policy_eval_online(
    policy_data = pd,
    policy = p,
    target = "value",
    g_functions = gf,
    q_models = q_degen(var = "z"),
    train_block_size = 60,
    M = 2
  )

  train_sequential_index <- pe$train_sequential_index
  valid_sequential_index <- pe$valid_sequential_index

  ## standard deviation estimate for each training sample:
  train_sigma <- lapply(
    train_sequential_index,
    function(x){
      tmp <- ref_Z[x]
      mean((tmp - mean(tmp))^2)
    }
  ) |> unlist() |> sqrt()

  valid_count <- lapply(
    valid_sequential_index,
    length
  ) |> unlist()
  train_sigma_inverse_sum <- sum(valid_count / train_sigma)

  ## the doubly robust score evaluated on each of the validation samples:
  valid_score <- lapply(
    valid_sequential_index,
    function(x){
      sum(ref_Z[x])
    }
  ) |> unlist()

  ## reference online estimate:
  ref_coef <- sum(train_sigma^(-1) * valid_score) / train_sigma_inverse_sum
  names(ref_coef) <- "E[U(d)]: d=test"


  expect_equal(
    coef(pe),
    ref_coef
  )

  ## reference squared standard error estimate:
  ref_vcov <- (mean(train_sigma^(-1))^{-1}/sqrt(100-60))^2


  expect_equal(
    ref_vcov,
    vcov(pe)[[1]]
  )

  ## multiple policies

  pl <- policy_learn(
    type = "blip",
    threshold = c(25, 75),
    control = control_blip(blip_models = q_degen(var = "z"))
  )

  ## reference doubly robust score:
  ref_Z_25 <- (d$a == c(rep(1, 25), rep(2, 75))) / 0.5 * (d$y - d$z) + d$z
  ref_Z_75 <- (d$a == c(rep(1, 75), rep(2, 25))) / 0.5 * (d$y - d$z) + d$z

  set.seed(1)
  pe <- policy_eval_online(
    policy_data = pd,
    policy_learn = pl,
    target = "value",
    g_functions = gf,
    q_models = q_degen(var = "z"),
    train_block_size = 60,
    M = 2
  )

  train_sequential_index <- pe$train_sequential_index
  valid_sequential_index <- pe$valid_sequential_index

  ## standard deviation estimate for each training sample
  train_sigma <- lapply(
    train_sequential_index,
    function(x){
      tmp25 <- ref_Z_25[x]
      tmp75 <- ref_Z_75[x]
      out <- c(
        mean((tmp25 - mean(tmp25))^2),
        mean((tmp75 - mean(tmp75))^2)
      )
      out <- sqrt(out)
    }
  ) |> do.call(what = "rbind")

  valid_count <- lapply(
    valid_sequential_index,
    length
  ) |> unlist()

  ## sum of the inverse standard deviaiton estimates
  train_sigma_inverse_sum <-  1 / train_sigma *
    matrix(rep(valid_count,2), byrow = TRUE, ncol = length(valid_count))
  train_sigma_inverse_sum <- colSums(train_sigma_inverse_sum)

  ## the doubly robust score evaluated on each of the validation samples:
  valid_score <- lapply(
    valid_sequential_index,
    function(x){
      c(sum(ref_Z_25[x]),sum(ref_Z_75[x]))
    }
  ) |> do.call(what = "rbind")

  ## reference online estimate:
  ref_coef <- colSums(train_sigma^(-1) * valid_score) / train_sigma_inverse_sum
  names(ref_coef) <- c("E[U(d)]: d=blip(eta=25)", "E[U(d)]: d=blip(eta=75)")

  expect_equal(
    coef(pe),
    ref_coef
  )

  ## reference squared standard error estimate:
  ref_vcov <- (colMeans(train_sigma^(-1))^{-1}/sqrt(100-60))^2
  tmp <- matrix(ncol = length(ref_vcov), nrow = length(ref_vcov))
  diag(tmp) <- ref_vcov
  ref_vcov <- tmp

  expect_equal(
    ref_vcov,
    vcov(pe)
  )

})

test_that("policy_eval_sequential has the expected output in the single stage case for target = 'subgroup'", {

  z <- 1:1e2
  a <- c(rep(1, 50), rep(2, 50))
  y <- a * 2
  p <- c(rep(1, 25), rep(2, 25), rep(1, 25), rep(2, 25))
  d <- data.table(z = z, a = a, y = y, p = p)
  rm(a, z, y)
  pd <- policy_data(
    data = d,
    action = "a",
    covariates = c("z", "p"),
    utility = c("y")
  )
  p <- policy_def(function(p) p, name = "test")
  gf <- fit_g_functions(pd, g_models = g_glm(~1))

  ref_Z_1 <- (d$a == 1) / 0.5 * (d$y - d$z) + d$z
  ref_Z_2 <- (d$a == 2) / 0.5 * (d$y - d$z) + d$z
  ref_blip <- ref_Z_2 - ref_Z_1

  ## single policy

  pe <- policy_eval_online(
    policy_data = pd,
    policy = p,
    target = "subgroup",
    g_functions = gf,
    q_models = q_degen(var = "z"),
    train_block_size = 60,
    M = 2
  )

  train_sequential_index <- pe$train_sequential_index
  valid_sequential_index <- pe$valid_sequential_index

  ## standard deviation estimate for each training sample:
  train_sigma <- lapply(
    train_sequential_index,
    function(x){
      blip <- ref_blip[x]
      s1 <- (d$p == 1)[x]
      s2 <- (d$p == 2)[x]

      est_1 <- mean(blip[s1])
      est_2 <- mean(blip[s2])

      var_1 <- mean((s1 / mean(s1) * (blip - est_1))^2)
      var_2 <- mean((s2 / mean(s2) * (blip - est_2))^2)

      return(sqrt(c(var_1, var_2)))
    }
  ) |> do.call(what = "rbind")

  train_subgroup_prob <- lapply(
    train_sequential_index,
    function(x){
      s1 <- (d$p == 1)[x]
      s2 <- (d$p == 2)[x]
      return(c(mean(s1), mean(s2)))
    }
  ) |> do.call(what = "rbind")

  valid_count <- lapply(
    valid_sequential_index,
    length
  ) |> unlist()

  ## sum of the inverse standard deviaiton estimates
  train_sigma_inverse_sum <-  1 / train_sigma *
    matrix(rep(valid_count,2), byrow = TRUE, ncol = length(valid_count))
  train_sigma_inverse_sum <- colSums(train_sigma_inverse_sum)

  ## the doubly robust score evaluated on each of the validation samples:
  valid_score <- lapply(
    valid_sequential_index,
    function(x){
      blip <- ref_blip[x]
      s1 <- (d$p == 1)[x]
      s2 <- (d$p == 2)[x]

      return(
        c(
          sum(s1 * blip),
          sum(s2 * blip)
        )
      )
    }
  ) |> do.call(what = "rbind")
  valid_score <- valid_score / train_subgroup_prob

  ## reference online estimate:
  ref_coef <- colSums(train_sigma^(-1) * valid_score) / train_sigma_inverse_sum
  ref_coef <- rev(ref_coef)
  names(ref_coef) <- c("E[U(2)-U(1)|d=2]: d=test", "E[U(2)-U(1)|d=1]: d=test")

  expect_equal(
    coef(pe),
    ref_coef
  )

  ## reference squared standard error estimate:
  ref_vcov <- (colMeans(train_sigma^(-1))^{-1}/sqrt(100-60))^2
  tmp <- matrix(ncol = length(ref_vcov), nrow = length(ref_vcov))
  diag(tmp) <- rev(ref_vcov)
  ref_vcov <- tmp

  expect_equal(
    ref_vcov,
    vcov(pe)
  )

})

test_that("policy_learn checks input", {

  expect_error(policy_learn(type = "test"),
               "Unknown type of policy learner. Use 'ql', 'drql', 'ptl', 'owl', 'earl' or 'rwl'.")

  expect_error(policy_learn(type = c("test", "test")),
               "type must be a character string.")


  d <- sim_single_stage(n=1e2)
  pd <- policy_data(d,
                    action="A",
                    covariates = list("Z", "B", "L"),
                    utility="U")

  # alpha
  expect_error(
    policy_eval(pd,
                policy_learn =  policy_learn(type = "ql",
                                             alpha = "test"))
  )
  expect_error(
    policy_eval(pd,
                policy_learn =  policy_learn(type = "ql",
                                             alpha = c(0,1)))
  )
  expect_error(
    policy_eval(pd,
                policy_learn =  policy_learn(type = "ql",
                                             alpha = NULL))
  )
  expect_error(
    policy_eval(pd,
                policy_learn =  policy_learn(type = "ql",
                                             alpha = -0.1))
  )

  expect_error(
    policy_eval(pd,
                policy_learn =  policy_learn(type = "ql",
                                             alpha = 0)),
    NA
  )
  expect_error(
    policy_eval(pd,
                policy_learn =  policy_learn(type = "ql",
                                             alpha = 0.1)),
    NA
  )

  # full_history
  expect_error(
    policy_eval(pd,
                policy_learn =  policy_learn(type = "ql",
                                             full_history = TRUE)),
    NA
  )
  expect_error(
    policy_eval(pd,
                policy_learn =  policy_learn(type = "ql",
                                             full_history = FALSE)),
    NA
  )
  expect_error(
    policy_eval(pd,
                policy_learn =  policy_learn(type = "ql",
                                             full_history = "test")),
    "full_history must be TRUE or FALSE"
  )

  # L: number of folds
  expect_error(
    policy_learn(type = "ql",
                 L = "test")
  )
  expect_error(
    policy_learn(type = "ql",
                 L = 0)
  )
  expect_error(
    policy_learn(type = "ql",
                 L = -1)
  )
  expect_error(
    policy_learn(type = "ql",
                 L = 1.1)
  )
  expect_error(
    policy_learn(type = "ql",
                 L = 1),
    NA
  )
  expect_error(
    policy_learn(type = "ql",
                 L = 2),
    NA
  )

  # cross_fit_g_models
  expect_error(
    policy_learn(type = "ql",
                                 cross_fit_g_models = TRUE),
    NA
  )
  expect_error(
    policy_learn(type = "ql",
                                 cross_fit_g_models = FALSE),
    NA
  )
  expect_error(
    policy_learn(type = "ql",
                                 cross_fit_g_models = "test"),
    "cross_fit_g_models must be TRUE or FALSE"
  )

  # cross_fit_g_models
  expect_error(
    policy_learn(type = "ql",
                 save_cross_fit_models = TRUE),
    NA
  )
  expect_error(
    policy_learn(type = "ql",
                 save_cross_fit_models = FALSE),
    NA
  )
  expect_error(
    policy_learn(type = "ql",
                 save_cross_fit_models = "test"),
    "save_cross_fit_models must be TRUE or FALSE"
  )

  # future_args
  fa <- list(
    "test",
    NULL,
    1,
    TRUE
  )
  lapply(
    fa, function(fa){
      expect_error(
        policy_learn(type = "ql",
                     future_args =  fa)
      )
    })
  fa <- list(
    list(),
    list(future_seed = TRUE)
  )
  lapply(
    fa, function(fa){
      expect_error(
        policy_learn(type = "ql",
                     future_args =  fa),
        NA
      )
    })

  # name
  nn <- list(1, TRUE, "test", NULL)
  lapply(
    nn, function(name){
      expect_error(
        policy_learn(type = "ql",
                     name = name),
        NA
      )
    })

  nn <- list(c(1,1))
  lapply(
    nn, function(name){
      expect_error(
        policy_learn(type = "ql",
                     name = name)
      )
    })

})

test_that("policy_learn returns an error if type != 'blip' or type != 'ptl and threshold != 0.", {
  d <- sim_single_stage(200, seed = 1)
  pd <- policy_data(d,
    action = "A",
    covariates = list("Z", "B", "L"),
    utility = "U"
  )

  expect_no_error(
    policy_learn(type = "ptl", control = control_ptl(), threshold = NULL)
  )

  expect_no_error(
    policy_learn(type = "ptl", control = control_ptl(), threshold = 1)
  )

  expect_no_error(
    policy_learn(type = "blip", control = control_blip(), threshold = 1)
  )

  expect_error(
    policy_learn(type = "drql", control = control_drql(), threshold = 1),
    ".only implemented for type 'blip' or 'ptl'"
  )
})

test_that("set_threshold", {
  expect_equal(
    set_threshold(selection = c(1, 2, 3)),
    c(1, 2, 3)
  )

  expect_equal(
    set_threshold(threshold = c(1, 3), selection = c(1, 2, 3)),
    c(1, 3)
  )

  expect_equal(
    set_threshold(threshold = c(3, 1), selection = c(1, 2, 3)),
    c(1, 3)
  )

  expect_equal(
    set_threshold(threshold = c(), selection = c(1, 2, 3)),
    c(1, 2, 3)
  )

  expect_error(
    set_threshold(threshold = c(1.5, 3), selection = c(1, 2, 3)),
    "Invalid threshold. Choose between 1, 2, 3."
  )

  expect_equal(
    set_threshold(threshold = c(3, 1.5), selection = c(1, 2, 3), overwrite = TRUE),
    c(1.5, 3)
  )
})

test_that("policy_learn() does not run with missing outcomes for drql, rql, ptl, earl, rwl, owl.", {

  sim_single_stage_right_cens <- function(n = 2e3, zeta = c(0.7, 0.2), type = "right"){

    d <- sim_single_stage(n = n)
    pd <- policy_data(data = d,
                      action = "A",
                      covariates = c("Z", "L", "B"),
                      utility = "U")

    ld <- pd$stage_data

    ld[stage == 1, time := 1]
    ld[stage == 2, time := 2]

    ld[stage == 2, Z := d$Z]
    ld[stage == 2, L := d$L]
    ld[stage == 2, B := d$B]

    ## simulating the right censoring time
    ## only depending on the baseline covariate Z:
    C <- c(rexp(n, 1) / exp((-1) * cbind(1, as.numeric(d$Z)) %*% zeta))

    ld[stage == 1, time_c := C]
    ld[stage == 2, time_c := C]

    ld[, delta := time_c >= time]

    ld[delta == FALSE , event := 2]
    ld[delta == FALSE, A := NA]
    ld[delta == FALSE & stage == 2, U := NA]
    ld[delta == FALSE, U_A0 := 0]
    ld[delta == FALSE, U_A1 := 0]

    ld[ , tmp := shift(delta, fill = TRUE), by = list(id)]
    ld <- ld[tmp == TRUE, ]
    ld[ , time := pmin(time, time_c)]
    ld[ , time_c := NULL]
    ld[ , tmp := NULL]
    ld[ , delta := NULL]

    if (type == "interval"){
      ld[, time2 := time]
      ld[, time := shift(time, fill = 0), by = list(id)]
    }

    return(ld)
  }

  set.seed(1)
  ld <- sim_single_stage_right_cens(n = 5e2, type = "interval")
  pd <- policy_data(data = ld, type = "long", action = "A", time = "time", time2 = "time2")

  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = policy_learn(type = "drql", control = control_drql()),
      m_model = q_glm(~.),
      m_full_history = FALSE,
      c_models = c_cox(formula = ~ Z, time = "time", time_2 = "time_2"),
      ),
    "policy learning with type 'drql' not implemented under right-censoring/missing outcomes."
  )

  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = policy_learn(type = "rql", control = list()),
      m_model = q_glm(~.),
      m_full_history = FALSE,
      c_models = c_cox(formula = ~ Z, time = "time", time_2 = "time_2"),
      ),
    "policy learning with type 'rql' not implemented under right-censoring/missing outcomes."
  )

  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = policy_learn(type = "ptl", control = control_ptl()),
      m_model = q_glm(~.),
      m_full_history = FALSE,
      c_models = c_cox(formula = ~ Z, time = "time", time_2 = "time_2"),
      ),
    "policy learning with type 'ptl' not implemented under right-censoring/missing outcomes."
  )

  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = policy_learn(type = "earl", control = control_earl()),
      m_model = q_glm(~.),
      m_full_history = FALSE,
      c_models = c_cox(formula = ~ Z, time = "time", time_2 = "time_2"),
      ),
    "policy learning with type 'earl' not implemented under right-censoring/missing outcomes."
  )

  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = policy_learn(type = "rwl", control = control_rwl()),
      m_model = q_glm(~.),
      m_full_history = FALSE,
      c_models = c_cox(formula = ~ Z, time = "time", time_2 = "time_2"),
      ),
    "policy learning with type 'rwl' not implemented under right-censoring/missing outcomes."
  )

  expect_error(
    pe <- policy_eval(
      policy_data = pd,
      policy_learn = policy_learn(type = "owl", control = control_owl()),
      m_model = q_glm(~.),
      m_full_history = FALSE,
      c_models = c_cox(formula = ~ Z, time = "time", time_2 = "time_2"),
      ),
    "policy learning with type 'owl' not implemented under right-censoring/missing outcomes."
  )

})

#' @title Control arguments for Policy Tree Learning
#' @description \code{control_ptl} sets the default control arguments
#' for doubly robust policy tree learning, \code{type = "ptl"}.
#' The arguments are passed directly to [policytree::policy_tree()] (or
#' [policytree::hybrid_policy_tree()]) if not specified otherwise.
#' @param policy_vars Character vector/string or list of character
#' vectors/strings. Variable names used to
#' construct the V-restricted policy tree.
#' The names must be a subset of the history names, see get_history_names().
#' Not passed to \code{policy_tree()}.
#' @param hybrid If \code{TRUE}, [policytree::hybrid_policy_tree()] is used to
#' fit a policy tree. Not passed to \code{policy_tree()}.
#' @param depth Integer or integer vector. The depth of the fitted policy
#' tree for each stage.
#' @param search.depth (only used if \code{hybrid = TRUE}) Integer or integer
#' vector. Depth to look ahead when splitting at each stage.
#' @param split.step Integer or integer vector. The number of possible splits
#' to consider when performing policy tree search at each stage.
#' @param min.node.size Integer or integer vector. The smallest terminal node
#' size permitted at each stage.
#' @returns list of (default) control arguments.
#' @export
control_ptl <- function(policy_vars = NULL,
                        hybrid = FALSE,
                        depth = 2,
                        search.depth = 2,
                        split.step = 1,
                        min.node.size = 1) {
  control <- as.list(environment())
  return(control)
}

ptl <- function(policy_data,
                alpha,
                threshold,
                g_models, g_functions, g_full_history,
                q_models, q_full_history,
                policy_vars, full_history,
                L,
                cross_fit_g_models, save_cross_fit_models,
                future_args,
                depth, split.step, min.node.size, hybrid, search.depth,
                ...) {
  K <- get_K(policy_data)
  n <- get_n(policy_data)
  action_set <- get_action_set(policy_data)
  stage_action_sets <- get_stage_action_sets(policy_data)

  # input checks:
  if (alpha > 0) {
    for (k in seq_along(stage_action_sets)) {
      if (length(stage_action_sets[[k]]) != 2) {
        mes <- paste0(
          "realistic policy tree learning is only implemented for ",
          "dichotomous stage action sets. Use ",
          "doubly robust Q-learning (drql) instead."
        )
        stop(mes)
      }
    }
  }
  if ((is.null(g_models) && is.null(g_functions))) {
    stop("Provide either g-models or g-functions.")
  }
  if (is.list(q_models)) {
    if (length(q_models) != K) {
      stop("q_models must either be a list of length K or a single Q-model.")
    }
  }
  if (full_history == TRUE) {
    if ((!is.list(policy_vars)) || (length(policy_vars) != K)) {
      stop("policy_vars must be a list of length K, when full_history = TRUE.")
    }
  }
  if (is.list(depth)) {
    depth <- unlist(depth)
  }
  if (length(depth) == 1) {
    depth <- rep(depth, K)
  } else {
    if (length(depth) != K || !all(depth %% 1 == 0)) {
      stop("depth must be an integer vector of length K.")
    }
  }
  if (is.list(split.step)) {
    split.step <- unlist(split.step)
  }
  if (length(split.step) == 1) {
    split.step <- rep(split.step, K)
  } else {
    if (length(split.step) != K || !all(split.step %% 1 == 0)) {
      stop("split.step must be an integer vector of length K.")
    }
  }
  if (is.list(min.node.size)) {
    min.node.size <- unlist(min.node.size)
  }
  if (length(min.node.size) == 1) {
    min.node.size <- rep(min.node.size, K)
  } else {
    if (length(min.node.size) != K || !all(min.node.size %% 1 == 0)) {
      stop("min.node.size must be an integer vector of length K.")
    }
  }
  if (is.list(search.depth)) {
    search.depth <- unlist(search.depth)
  }
  if (length(search.depth) == 1) {
    search.depth <- rep(search.depth, K)
  } else {
    if (length(search.depth) != K || !all(search.depth %% 1 == 0)) {
      stop("search.depth must be an integer vector of length K.")
    }
  }
  if (any(get_element(policy_data, "cens_indicator")[["indicator"]])){
    stop("policy learning with type 'ptl' not implemented under right-censoring/missing outcomes.")
  }

  # getting the observed actions:
  actions <- get_actions(policy_data)

  # getting the observed (complete) utilities:
  utility <- get_utility(policy_data)

  # constructing the folds for cross-fitting
  if (L > 1) {
    folds <- split(sample(1:n, n), rep(1:L, length.out = n))
  } else {
    folds <- NULL
  }

  # (cross-)fitting the g-functions:
  g_functions_cf <- NULL
  if (!is.null(folds) && cross_fit_g_models == TRUE) {
    g_cf <- crossfit_function(
      policy_data = policy_data,
      fun = fit_g_functions,
      models = g_models,
      full_history = g_full_history,
      folds = folds,
      save_cross_fit_models = save_cross_fit_models,
      future_args = future_args
    )
    g_functions_cf <- getElement(g_cf, "functions")
    g_values <- getElement(g_cf, "values")
    rm(g_cf)
  } else {
    if (is.null(g_functions)) {
      g_functions <- fit_g_functions(policy_data,
        g_models = g_models,
        full_history = g_full_history
      )
    }
    g_values <- predict(g_functions, policy_data)
  }

  # fitting g-functions for determining new realistic actions:
  if (alpha > 0) {
    if (is.null(g_functions)) {
      g_functions <- fit_g_functions(policy_data,
        g_models = g_models,
        full_history = g_full_history
      )
    }
  } else {
    # g-functions are not saved if alpha == 0:
    g_functions <- NULL
  }

  # (n) vector with entries U_i:
  U <- utility$U
  # (n X K) matrix with entries I(d_k(H_k) = A_k):
  II <- matrix(nrow = n, ncol = K)
  # (n X K) matrix with entries g_k(A_k, H_k)
  g_A_values <- get_a_values(a = actions$A, action_set = action_set, g_values)
  G <- dcast(g_A_values, id ~ stage, value.var = "P")[, -c("id"), with = FALSE]
  G <- as.matrix(G)
  # (n X K+1) matrix with entries Q_k(H_{k,i}, d_k(H_{k,i})), Q_{K+1} = U:
  Q <- matrix(nrow = n, ncol = K+1)
  Q[, K+1] <- U
  # (n X K) matrix with entries d_k(H_k) (including unrealistic actions)
  D <- matrix(nrow = n, ncol = K)

  q_cols <- paste("Q_", action_set, sep = "")
  ptl_objects <- list()
  ptl_designs <- list()
  q_functions <- list()
  q_functions_cf <- list()
  for (k in K:1){
    if (is.null(folds)) {
      q_step_k <- q_step(
        policy_data = policy_data,
        k = k,
        full_history = q_full_history,
        Q = Q[, k + 1],
        q_models = q_models
      )
      # getting the Q-function, Q-function values and the ID-index:
      q_functions[[k]] <- getElement(q_step_k, "q_function")
      q_values_k <- getElement(q_step_k, "q_values")
      idx_k <- getElement(q_step_k, "idx_k")
    } else {
      q_step_cf_k <- q_step_cf(
        folds = folds,
        policy_data = policy_data,
        k = k,
        full_history = q_full_history,
        Q = Q[, k + 1],
        q_models = q_models,
        save_cross_fit_models = save_cross_fit_models,
        future_args = future_args
      )
      q_functions_cf[[k]] <- getElement(q_step_cf_k, "q_functions_cf")
      q_values_k <- getElement(q_step_cf_k, "q_values")
      idx_k <- getElement(q_step_cf_k, "idx_k")
      rm(q_step_cf_k)
    }

    # getting the stage action set
    stage_action_set_k <- stage_action_sets[[k]]

    # getting the action matrix for stage k:
    stage <- NULL
    A_k <- unlist(actions[stage == k, "A"])
    IA_k <- action_matrix(A_k, action_set)
    rm(stage)

    # calculating Gamma (Z-matrix), see ?policytree::policy_tree
    Gamma_1 <- Q_k <- as.matrix(q_values_k[, q_cols, with = FALSE])
    Gamma_2 <- (IA_k / G[idx_k, k]) * (Q[idx_k, k+1] - Q_k)
    Gamma_3 <- 0
    if (k != K) {
      for (r in (k + 1):K) {
        Gamma_3 <- Gamma_3 +
          ipw_weight(II[idx_k, (k + 1):r], G[idx_k, (k + 1):r]) *
          (Q[idx_k, r + 1] - Q[idx_k, r])
      }
      Gamma_3 <- (IA_k / G[idx_k, k]) * Gamma_3
    }
    Gamma <- Gamma_1 + Gamma_2 + Gamma_3
    colnames(Gamma) <- action_set
    Gamma <- Gamma[, stage_action_set_k]

    # getting the policy history
    policy_history_k <- get_history(policy_data,
                                    stage = k,
                                    full_history = full_history)
    if (full_history == TRUE) {
      vars <- policy_vars[[k]]
    } else {
      vars <- policy_vars
    }
    H <- get_H(policy_history_k, vars = vars)
    if (is.null(policy_vars)) {
      policy_vars <- names(H)
    }

    # design matrix for policy_tree:
    design_k <- get_design(formula = ~., data = H)
    X <- design_k$x
    design_k$x <- NULL
    ptl_designs[[k]] <- design_k

    # tuning parameters for policy_tree:
    depth_k <- depth[[k]]
    split.step_k <- split.step[[k]]
    min.node.size_k <- min.node.size[[k]]
    search.depth_k <- search.depth[[k]]

    if (k != 1) {
      if (hybrid == FALSE) {
        ptl_k <- policytree::policy_tree(
          X = X,
          Gamma = Gamma,
          depth = depth_k,
          split.step = split.step_k,
          min.node.size = min.node.size_k
        )
      } else if (hybrid == TRUE) {
        ptl_k <- policytree::hybrid_policy_tree(
          X = X,
          Gamma = Gamma,
          depth = depth_k,
          split.step = split.step_k,
          min.node.size = min.node.size_k,
          search.depth = search.depth_k
        )
      }
      ptl_objects[[k]] <- ptl_k
      dd <- predict(ptl_k, X)
      d <- stage_action_set_k[dd]

      if (alpha != 0) {
        # getting the g-function values for each action:
        g_values_k <- g_values[stage == k, ]

        g_stage_cols <- paste("g_", stage_action_set_k, sep = "")
        # modifying the policy to only recommend realistic actions
        d[(g_values_k[, g_stage_cols[1], with = FALSE] < alpha)] <-
          stage_action_set_k[2]
        d[(g_values_k[, g_stage_cols[2], with = FALSE] < alpha)] <-
          stage_action_set_k[1]
      }

      q_d_k <- get_a_values(a = d, action_set = action_set, q_values_k)[["P"]]
      Q[idx_k, k] <- q_d_k
      Q[!idx_k, k] <- Q[!idx_k, k + 1]
      II[idx_k, k] <- (A_k == d)
      II[!idx_k, k] <- TRUE
      D[idx_k, k] <- d
      G[!idx_k, k] <- TRUE
    } else {
      ptl_k_threshold <- lapply(
        threshold,
        function(th) {
          gamma <- Gamma
          ## adding the threshold to the reference action:
          gamma[, 1] <- gamma[, 1] + th

          if (hybrid == FALSE) {
            ptl_k <- policytree::policy_tree(
              X = X,
              Gamma = gamma,
              depth = depth_k,
              split.step = split.step_k,
              min.node.size = min.node.size_k
            )
          } else if (hybrid == TRUE) {
            ptl_k <- policytree::hybrid_policy_tree(
              X = X,
              Gamma = gamma,
              depth = depth_k,
              split.step = split.step_k,
              min.node.size = min.node.size_k,
              search.depth = search.depth_k
            )
          }
          return(ptl_k)
        }
      )
      tmp <- paste0("threshold_", threshold)
      names(ptl_k_threshold) <- tmp
      ptl_objects[[k]] <- ptl_k_threshold
    }
  }

  # setting outputs:
  if (length(q_functions) > 0) {
    class(q_functions) <- "nuisance_functions"
    attr(q_functions, "full_history") <- q_full_history
    names(q_functions) <- paste("stage_", 1:K, sep = "")
  } else {
    q_functions <- NULL
  }
  if (length(q_functions_cf) == 0) {
    q_functions_cf <- NULL
  }
  names(ptl_objects) <- paste("stage_", 1:K, sep = "")

  out <- list(
    ptl_objects = ptl_objects,
    ptl_designs = ptl_designs,
    full_history = full_history,
    policy_vars = policy_vars,
    g_functions = g_functions,
    g_functions_cf = g_functions_cf,
    q_functions = q_functions,
    q_functions_cf = q_functions_cf,
    action_set = action_set,
    stage_action_sets = stage_action_sets,
    alpha = alpha,
    threshold = threshold,
    K = K,
    folds = folds
  )
  out <- remove_null_elements(out)
  class(out) <- c("ptl", "policy_object")

  return(out)
}

#' @export
get_policy.ptl <- function(object, threshold = NULL) {
  stage_action_sets <- get_element(object, "stage_action_sets")
  K <- get_element(object, "K")
  full_history <- get_element(object, "full_history")
  ptl_objects <- get_element(object, "ptl_objects")
  ptl_designs <- get_element(object, "ptl_designs")
  policy_vars <- get_element(object, "policy_vars")
  g_functions <- get_element(object, "g_functions", check_name = FALSE)
  alpha <- get_element(object, "alpha")
  threshold_selection <- get_element(object, "threshold")
  threshold <- set_threshold(
    threshold = threshold,
    selection = threshold_selection,
    overwrite = FALSE
  )
  threshold_idx <- which(threshold_selection %in% threshold)

  policy_list <- lapply(
    threshold_idx,
    function(th_idx) {
      ## selecting the stage 1 policy tree associated with the given threshold
      po <- ptl_objects
      po[[1]] <- ptl_objects[[1]][[th_idx]]

      policy <- function(policy_data) {
        if (get_K(policy_data) != K) {
          mes <- paste0(
            "The policy do not have the same ",
            "number of stages as the policy data object."
          )
          stop(mes)
        }

        if (alpha != 0) {
          # evaluating the g-functions:
          g_values <- predict(g_functions, policy_data)
        }

        ## getting the actions recommended by the ptl objects:
        policy_actions <- list()
        for (k in K:1) {
          # getting the policy history:
          policy_history_k <- get_history(policy_data,
            stage = k,
            full_history = full_history
          )

          ## getting the design matrix:
          if (full_history == TRUE) {
            vars <- policy_vars[[k]]
          } else {
            vars <- policy_vars
          }
          H <- get_H(policy_history_k, vars = vars)
          design <- ptl_designs[[k]]
          newdata <- apply_design(design = design, data = H)

          ## getting the stage action set
          stage_action_set <- stage_action_sets[[k]]

          ## policy tree predictions:
          dd <- predict(po[[k]], newdata = newdata)
          d <- stage_action_set[dd]

          ## excluding unrealistic recommendations:
          if (alpha != 0) {
            g_stage_cols <- paste("g_", stage_action_set, sep = "")
            stage <- NULL
            g_values_k <- g_values[stage == k, ]

            d[(g_values_k[, g_stage_cols[1], with = FALSE] < alpha)] <-
              stage_action_set[2]
            d[(g_values_k[, g_stage_cols[2], with = FALSE] < alpha)] <-
              stage_action_set[1]
          }

          pa <- get_id_stage(policy_history_k)
          pa[, d := d]
          policy_actions[[k]] <- pa
          rm(pa, d, dd)
        }
        policy_actions <- rbindlist(policy_actions)
        setkeyv(policy_actions, c("id", "stage"))

        return(policy_actions)
      }
      ## setting class and attributes:
      th <- threshold_selection[th_idx]
      name <- paste0("ptl(eta=", round(th, 3), ")")
      policy <- new_policy(policy, name = name)

      return(policy)
    }
  )

  if (length(policy_list) == 1) {
    out <- policy_list[[1]]
  } else {
    out <- policy_list
  }

  return(out)
}


#' @rdname get_policy_functions
#' @export
get_policy_functions.ptl <- function(object,
                                     stage,
                                     threshold = NULL,
                                     ...) {
  K <- get_element(object, "K")
  check_stage(stage = stage, K = K)
  stage_action_sets <- get_element(object, "stage_action_sets")
  g_functions <- get_element(object, "g_functions", check_name = FALSE)
  if (!is.null(g_functions)) {
    if (length(g_functions) == K) {
      g_function <- g_functions[[stage]]
    } else {
      g_function <- g_functions[[1]]
    }
    rm(g_functions)
  }
  alpha <- get_element(object, "alpha")
  threshold_selection <- get_element(object, "threshold")
  threshold <- set_threshold(
    threshold = threshold,
    selection = threshold_selection,
    overwrite = FALSE
  )
  if (length(threshold) > 1) {
    mes <- paste(threshold, collapse = ", ")
    mes <- paste("Choose threshold: ", mes, ".", sep = "")
    stop(mes)
  }
  threshold_idx <- which(threshold_selection %in% threshold)

  ## selecting the stage 1 policy tree associated with the given threshold:
  ptl_objects <- get_element(object, "ptl_objects")
  ptl_objects[[1]] <- ptl_objects[[1]][[threshold_idx]]

  ## selecting the policy function for the given stage:
  ptl_object <- ptl_objects[[stage]]
  rm(ptl_objects)

  ptl_design <- get_element(object, "ptl_designs")[[stage]]
  full_history <- get_element(object, "full_history")
  if (full_history) {
    policy_vars <- get_element(object, "policy_vars")[[stage]]
  } else {
    policy_vars <- get_element(object, "policy_vars")
  }
  stage_action_set <- stage_action_sets[[stage]]
  rm(stage_action_sets)

  rm(object)
  stage_policy <- function(H) {
    H <- as.data.table(H)
    if (!all(policy_vars %in% colnames(H))) {
      mes <- "H must have column names "
      mes <- paste(mes, paste(policy_vars, collapse = ", "), ".", sep = "")
      stop(mes)
    }
    newdata <- H[, policy_vars, with = FALSE]
    mf <- with(ptl_design, model.frame(terms,
      data = newdata,
      xlev = xlevels,
      drop.unused.levels = FALSE
      ))
    newdata <- model.matrix(mf, data = newdata, xlev = ptl_design$xlevels)
    idx_inter <- which(colnames(newdata) == "(Intercept)")
    if (length(idx_inter) > 0) {
      newdata <- newdata[, -idx_inter, drop = FALSE]
    }

    dd <- predict(ptl_object, newdata = newdata)
    d <- stage_action_set[dd]

    if (alpha > 0) {
      # evaluating the g-function:
      if (!all(g_function$H_names %in% names(H))) {
        mes <- paste(
          "H must contain the columns",
          paste(g_function$H_names, collapse = ","),
          "."
        )
        stop(mes)
      }
      g_values <- predict(g_function$g_model, H)
      colnames(g_values) <- stage_action_set

      # excluding unrealistic recommendations:
      d[(g_values[, stage_action_set[1]] < alpha)] <- stage_action_set[2]
      d[(g_values[, stage_action_set[2]] < alpha)] <- stage_action_set[1]
    }

    return(d)
  }

  return(stage_policy)
}

fit_functions <- function(policy_data,
                          type,
                          policy = NULL, policy_learn = NULL,
                          g_models = NULL, g_functions = NULL, g_full_history,
                          q_models = NULL, q_functions = NULL, q_full_history,
                          c_models = NULL, c_functions = NULL, c_full_history,
                          m_model = NULL, m_function = NULL, m_full_history) {
  ## fitting the g-functions (if NULL and if g_models is not NULL):
  if (is.null(g_functions)) {
    # g-models are not fitted if type is "or".
    if (!is.null(g_models) && (type %in% c("dr", "ipw"))) {
      g_functions <- fit_g_functions(
        policy_data,
        g_models = g_models,
        full_history = g_full_history
      )
    }
  }

  ## fitting the c-functions (if NULL and if c_models is not NULL):
  if (is.null(c_functions)) {
    if (!is.null(c_models)) {
      c_functions <- fit_c_functions(
        policy_data,
        c_models = c_models,
        full_history = c_full_history
      )
    }
  }

  ## fitting m-function:
  if (is.null(m_function)) {
    m_function <- fit_m_function(
      policy_data,
      m_model = m_model,
      full_history = m_full_history
    )
  }

  ## learning the policy:
  policy_object <- NULL
  if (is.null(policy)) {
    policy_object <- policy_learn(
      policy_data = policy_data,
      g_models = g_models, g_functions = g_functions,
      g_full_history = g_full_history,
      q_models = q_models, q_full_history = q_full_history,
      c_models = c_models, c_functions = c_functions, c_full_history = c_full_history,
      m_model = m_model, m_function = m_function, m_full_history = m_full_history
    )
    ## getting the policy associated with the default threshold:
    ## the policy is only used to fit the Q-functions, and
    ## the threshold at stage 1 does not affect the learned
    ## policy at later stages and thus the fitted Q-funcion
    threshold <- get_element(policy_object, "threshold", check_name = FALSE)[1]
    policy <- get_policy(policy_object, threshold = threshold)
  }

  # getting the policy actions:
  policy_actions <- policy(policy_data = policy_data)

  ## the policy actions are needed to fit the Q-functions.
  ## note that the policy actions at the first stage will not
  ## affect the fit.

  # fitting the Q-functions (if NULL and if q_models is not NULL):
  if (is.null(q_functions)) {
    if (!is.null(getElement(policy_object, "q_functions"))) {
      q_functions <- get_element(policy_object, "q_functions")
    } else {
      # q-models are not fitted if type is "ipw".
      if (!is.null(q_models) && (type %in% c("dr", "or"))) {
        q_functions <- fit_Q_functions(
          policy_data,
          policy_actions = policy_actions,
          q_models = q_models, full_history = q_full_history,
          m_function = m_function
        )
      }
    }
  }

  out <- list(
    g_functions = g_functions,
    q_functions = q_functions,
    c_functions = c_functions,
    m_function = m_function,
    policy_object = policy_object
  )
  return(out)
}

#' Cross-fit functions
#'
#' \code{crossfit_function()} is used to cross-fit a list of c-models
#'
#' @param folds List of vectors of indices for each validation fold, see examples.
#' @param policy_data Policy data object created by [policy_data()].
#' @param fun Function used to fit the models, either \code{fit_g_functions()}, \code{fit_c_functions()}
#' @param models single model or list of models like [g_sl()].
#' @param full_history If TRUE, the full history is used to fit each model.
#' If FALSE, the single stage/"Markov type" history is used to fit each model.
#' @param save_models Logical. Should the cross-fitted models be saved.
#' @param future_args arguments passed to [future.apply::future_lapply].
#' @noRd
crossfit_function <- function(folds,
                              policy_data,
                              fun,
                              models,
                              full_history,
                              save_cross_fit_models = FALSE,
                              future_args = list(future.seed = TRUE)){
  id <- get_id(policy_data)
  K <- get_K(policy_data)
  force(fun)
  force(models)
  force(full_history)

  future_args <- append(
    future_args,
    list(X = folds,
         FUN = function(f){
           train_id <- id[-f]
           train_policy_data <- subset_id(policy_data, train_id)
           if (get_K(train_policy_data) != K)
             stop("The number of stages K varies across the training policy data folds.")
           train_functions <- fun(policy_data = train_policy_data,
                                  models,
                                  full_history = full_history)

           valid_id <- id[f]
           valid_policy_data <- subset_id(policy_data, valid_id)
           valid_values <- predict(train_functions, valid_policy_data)

           if (save_cross_fit_models == FALSE)
             train_functions <- NULL

           out <- list(
             train_functions = train_functions,
             valid_values = valid_values,
             valid_id = valid_id
           )
           return(out)
         })
  )

  fit_cf <- do.call(what = future.apply::future_lapply, future_args)
  fit_cf <- simplify2array(fit_cf)

  functions <- fit_cf["train_functions", ]
  values <- fit_cf["valid_values", ]
  valid_ids <- fit_cf["valid_id", ]

  values <- rbindlist(values)
  setkeyv(values, c("id", "stage"))

  out <- list(
    functions = functions,
    values = values,
    valid_ids = valid_ids
  )
  return(out)
}

---
title: "policy_learn"
output:
  rmarkdown::html_vignette:
    fig_caption: true
    toc: true    
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{policy_learn}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib  
---


```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r lib, message = FALSE}
library("data.table")
library("polle")
```

This vignette is a guide to `policy_learn()` and some of the associated S3 methods.
The purpose of `policy_learn` is to specify a policy learning algorithm and estimate an optimal
policy. For details on the methodology, see the associated paper
[@nordland2023policy].

We consider a fixed two-stage problem as a general setup and simulate data using `sim_two_stage()` and create a `policy_data` object using `policy_data()`:

```{r simdata}
d <- sim_two_stage(n = 2e3, seed = 1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = c("B", "BB"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd
```

## Specifying and applying a policy learner

`policy_learn()` specify a policy learning algorithm via
the `type` argument: Q-learning (`ql`), doubly robust Q-learning (`drql`),
doubly robust blip learning (`blip`), policy tree learning (`ptl`), and
outcome weighted learning (`owl`).

Because each policy learning type has varying control arguments, these
are passed as a list using the `control` argument. To help the user
set the required control arguments and to provide documentation, each
type has a helper function `control_type()` which sets the default control
arguments and overwrite values if supplied by the user.

As an example we specify a doubly robust blip learner:

```{r plblip}
pl_blip <- policy_learn(
  type = "blip",
  control = control_blip(
    blip_models = q_glm(formula = ~ BB + L + C)
  )
)
```

For details on the implementation, see Algorithm 3 in [@nordland2023policy].
The only required control argument for blip learning is a model input.
The `blip_models` argument expects a `q_model`. In this case we input a
simple linear model as implemented in `q_glm`.

The output of `policy_learn()` is again a function:

```{r plblipout}
pl_blip
```

In order to apply the policy learner we need to input a `policy_data` object
and nuisance models `g_models` and `q_models` for computing the doubly robust
score.

```{r plblipapply}

(po_blip <- pl_blip(
  pd,
  g_models = list(g_glm(), g_glm()),
  q_models = list(q_glm(), q_glm())
 ))
```

## Cross-fitting the doubly robust score

Like `policy_eval()` is it possible to cross-fit the doubly robust score used as input to the policy model.
The number of folds for the cross-fitting procedure is provided via the `L` argument. As default, the
cross-fitted nuisance models are not saved. The cross-fitted nuisance models can be saved via the
`save_cross_fit_models` argument:

```{r plcross}
pl_blip_cross <- policy_learn(
  type = "blip",
  control = control_blip(
    blip_models = q_glm(formula = ~ BB + L + C)
  ),
  L = 2,
  save_cross_fit_models = TRUE
)
po_blip_cross <- pl_blip_cross(
   pd,
   g_models = list(g_glm(), g_glm()),
   q_models = list(q_glm(), q_glm())
 )
```
From a user perspective, nothing has changed. However, the policy object now contains each of the cross-fitted nuisance models:

```{r plcrossinspect}
po_blip_cross$g_functions_cf
```

## Realistic policy learning

Realistic policy learning is implemented for types `ql`, `drql`, `blip` and `ptl` (for a binary action set).
The `alpha` argument sets the probability threshold for defining the realistic action set. For implementation details,
see Algorithm 5 in [@nordland2023policy]. Here we set a 5\% restriction:

```{r pl_alpha}
pl_blip_alpha <- policy_learn(
  type = "blip",
  control = control_blip(
    blip_models = q_glm(formula = ~ BB + L + C)
  ),
  alpha = 0.05,
  L = 2
)
po_blip_alpha <- pl_blip_alpha(
   pd,
   g_models = list(g_glm(), g_glm()),
   q_models = list(q_glm(), q_glm())
 )
```

The policy object now lists the `alpha` level as well as the g-model used to define the realistic action set:
```{r viewalpha}
po_blip_alpha$alpha
po_blip_alpha$g_functions
```

## Implementation/Simulation and `get_policy_functions()`

A `policy` function is great for evaluating a given policy or
even implementing or simulating from  a single-stage policy.
However, the function is not useful for implementing or simulating
from a learned multi-stage policy. To access the policy function for each stage
we use `get_policy_functions()`. In this case we get the second stage policy function:

```{r }
pf_blip <- get_policy_functions(po_blip, stage = 2)
```

The stage specific policy requires a `data.table` with named columns as input and returns a character vector with the recommended actions:

```{r }
pf_blip(
  H = data.table(BB = c("group2", "group1"),
                 L = c(1, 0),
                 C = c(1, 2))
)
```



## Policy objects and `get_policy()`

Applying the policy learner returns a `policy_object` containing all of the components needed to specify the learned policy. In this the only component of the policy is a model for the blip function:

```{r inspectblip}
po_blip$blip_functions$stage_1$blip_model
```

To access and apply the policy itself use `get_policy()`, which behaves
as a `policy` meaning that we can apply to any (suitable) `policy_data` object to get the policy actions:

```{r }
get_policy(po_blip)(pd) |> head(4)
```

# SessionInfo

```{r sessionInfo}
sessionInfo()
```

# References

---
title: "policy_data"
output:
  rmarkdown::html_vignette:
    fig_caption: true
    toc: true    
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{policy_data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib  
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, message = FALSE}
library(polle)
```

This vignette is a guide to `policy_data()`. As the name suggests, the function creates a `policy_data` object with
a specific data structure making it easy to use in combination with `policy_def()`, `policy_learn()`, and `policy_eval()`.
The vignette is also a guide to some of the associated S3 functions
which transform or access parts of the data, see `?policy_data` and `methods(class="policy_data")`.

We will start by looking at a simple single-stage example, then consider a fixed two-stage example with varying actions sets and data in wide format, and
finally we will look at an example with a stochastic number of stages and data in long format.

# Single-stage: wide data

Consider a simple single-stage problem with covariates/state variables $(Z, L, B)$, binary action variable $A$, and
utility outcome $U$. We use `sim_single_stage()` to simulate data:

```{r single stage data}
(d <- sim_single_stage(n = 5e2, seed=1)) |> head()
```

We give instructions to `policy_data()` which variables define the `action`, the state `covariates`, and the `utility` variable: 

```{r pdss}
pd <- policy_data(d, action="A", covariates=list("Z", "B", "L"), utility="U")
pd
```

In the single-stage case the history $H$ is just $(B, Z, L)$. We access the history and actions using
`get_history()`:

```{r gethistoryss}
get_history(pd)$H |> head()
get_history(pd)$A |> head()
```

Similarly, we  access the utility outcomes $U$:

```{r get}
get_utility(pd) |> head()
```

```{r cleanup, include=FALSE}
rm(list = ls())
```

# Two-stage: wide data

Consider a two-stage problem with observations $O = (B, BB, L_{1}, C_{1}, U_{1},
A_1, L_2, C_{2}, U_{2}, A_2, U_{3})$. Following the general notation introduced
in Section 3.1 of [@nordland2023policy], $(B,BB)$ are the baseline covariates, $S_k =(L_{k, C_{k}})$ are the
state covariates at stage k, $A_{k}$ is the action at stage k, and $U_k$ is the reward at stage $k$.
The utility is the sum of the rewards $U=U_{1}+U_{2}+U_{3}$.

We use `sim_two_stage_multi_actions()` to simulate data:
```{r simtwostage}
d <- sim_two_stage_multi_actions(n=2e3, seed = 1)
colnames(d)
```
Note that the data is in wide format.
The data is transformed using `policy_data()` with instructions on which
variables define the actions, baseline covariates, state covariates, and the rewards:

```{r pdtwostage}
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = c("B", "BB"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd
```

The length of the character vector `action` determines the number of stages `K` (in this case 2).
If the number of stages is 2 or more, the `covariates` argument must be a named list. Each element must be
a character vector with length equal to the number of stages. If a covariate is not available at a
given stage we insert an `NA` value, e.g., `L = c(NA, "L_2")`.

Finally, the `utility` argument must
be a single character string (the utility is observed after stage K) or a character vector
of length K+1 with the names of the rewards.

In this example, the observed action sets vary for each stage. `get_action_set()` returns the
global action set and `get_stage_action_sets()` returns the action set for each stage:

```{r getactionsets}
get_action_set(pd)
get_stage_action_sets(pd)
```

The full histories $H_1 = (B, BB, L_{1}, C_{1})$ and $H_2=(B, BB, L_{1}, C_{1}, A_{1}, L_{2}, C_{2})$ are available using `get_history()` and `full_history = TRUE`:

```{r gethistwostage}
get_history(pd, stage = 1, full_history = TRUE)$H |> head()
get_history(pd, stage = 2, full_history = TRUE)$H |> head()
```
Similarly, we access the associated actions at each stage via list element `A`:

```{r}
get_history(pd, stage = 1, full_history = TRUE)$A |> head()
get_history(pd, stage = 2, full_history = TRUE)$A |> head()
```

Alternatively, the state/Markov type history and actions are available using `full_history = FALSE`:

```{r gethisstate}
get_history(pd, full_history = FALSE)$H |> head()
get_history(pd, full_history = FALSE)$A |> head()
```

Note that `policy_data()` overrides the action variable names to `A_1`, `A_2`, ... in the full history case and
`A` in the state/Markov history case.

As in the single-stage case we access the utility, i.e. the sum of the rewards, using
`get_utility()`:

```{r getutiltwo}
get_utility(pd) |> head()
```

# Multi-stage: long data

In this example we illustrate how `polle` handles decision
processes with a stochastic number of stages, see Section 3.5 in [@nordland2023policy].
The data is simulated using `sim_multi_stage()`.
Detailed information on the simulation is available in `?sim_multi_stage`.
We simulate data from 2000 iid subjects:

```{r sim_data}
d <- sim_multi_stage(2e3, seed = 1)
```
As described, the stage data is in long format:

```{r view_data}
d$stage_data[, -(9:10)] |> head()
```

The `id` variable is important for identifying which rows belong
to each subjects. The baseline data uses the same `id` variable:

```{r view_b_data}
d$baseline_data |> head()
```

The data is transformed using `policy_data()` with `type = "long"`.
The names of the `id`, `stage`, `event`, `action`,
and `utility` variables must be specified. The event variable, inspired by
the event variable in `survival::Surv()`, is `0` whenever an
action occur and `1` for a terminal event.

```{r pd}
pd <- policy_data(data = d$stage_data,
                  baseline_data = d$baseline_data,
                  type = "long",
                  id = "id",
                  stage = "stage",
                  event = "event",
                  action = "A",
                  utility = "U")
pd
```

In some cases we are only interested in analyzing a subset of the decision stages.
`partial()` trims the maximum number of decision stages:

```{r partial}
pd3 <- partial(pd, K = 3)
pd3

```


# SessionInfo

```{r sessionInfo}
sessionInfo()
```

# References

---
title: "Right censoring/monotone coarsening"
output:
  rmarkdown::html_vignette:
    fig_caption: true
    toc: true    
    toc_depth: 2
    mathjax: "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/MathJax.js?config=TeX-AMS_CHTML.js"
vignette: >
  %\VignetteIndexEntry{right_censoring}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib  
---

```{r lib, message = FALSE}
library(polle)
library(data.table)
```

## Target parameter under discrete time right-censoring/monotone coarsening

For simplicity we consider a two-stage case with a binary action set
$\{0,1\}$ at each stage and a final outcome $U$.

A full observation without right censoring is given by the temporally ordered variable $O$:

\begin{align*}
O^F = \{B, X_1, A_1, X_2, A_2, X_3, U_3\},
\end{align*}

where $B$ denotes the baseline variables, $X_1, X_2, X_3$ denote the stage specific variables, $U$ denotes the outcome/utility, and $A_1, A_2$ denote the treatment actions.

Now, introduce non-missingness indicators $\Delta_1, \Delta_2, \Delta_3$. Under right-censoring,
an observation is given by

\begin{align*}
O = \{B, X_1, \Delta_1, \Delta_1 A_1, \Delta_1 X_2,  \Delta_1 \Delta_2,  \Delta_1 \Delta_2 A_2,  \Delta_1 \Delta_2 X_3,  \Delta_1 \Delta_2\Delta_3, \Delta_1 \Delta_2 \Delta_3 U\}.
\end{align*}

For convenience, $H_1 = \{B, X_1\}$, $H_2 = \{B, X_1, A_1, X_2\}$, and $H_3 = \{B, X_1, A_1, X_2, A_2, X_3\}$.

The full data target parameter is still

\begin{aligned}
\mathbb{E}\left[ \left(\prod_{k=1}^2 \frac{I\{A_k = d_k(H_k\} }{g_{0,k}(H_k, A_k)} \right) U \right],
\end{aligned}
We assume the following (strong) sequential randomization for the right-censoring process:

\begin{aligned}
&\Delta_3 \perp U | H_3, \Delta_1 = 1, \Delta_2 = 1\\
&\Delta_2 \perp \{A_2, X_3, \Delta_3, U\}| H_2, \Delta_1 = 1 \\
&\Delta_1 \perp \{A_1, X_2, \Delta_2, A_2, X_3, \Delta_3, U\} | H_1.
\end{aligned}

Under these assumptions it can be shown that our observed data target parameter is equal to the full data target parameter:

\begin{aligned}
&\mathbb{E}\left[  \left(\prod_{k=1}^3 \frac{\Delta_k}{C_{0,k}(H_k)} \right) \left(\prod_{k=1}^2 \frac{I\{A_k = d_k(H_k\} }{g_{0,k}(H_k, A_k)} \right) U \right]\\ =&\mathbb{E}\left[ \left(\prod_{k=1}^2 \frac{I\{A_k = d_k(H_k\} }{g_{0,k}(H_k, A_k)} \right) U \right],
\end{aligned}
where
\begin{aligned}
C_{0,k}(h_k) = \mathbb{P}(\Delta_k = 1| H_k = h_k, \overline{\Delta}_k = 1)
\end{aligned}

and

\begin{aligned}
g_{0,1}(H_1, a_1) &= \mathbb{P}(A_1 = a_1|H_1, \Delta_1 = 1) =  \mathbb{P}(A_1 = a_1|H_1)\\
g_{0,2}(H_2, a_2) &= \mathbb{P}(A_2 = a_2|H_2, \Delta_1 = 1, \Delta_2 = 1) = \mathbb{P}(A_2 = a_2|H_2)
\end{aligned}


Note the observed target parameter is also identified via the g-computation formula
\begin{aligned}
\mathbb{E}[Q^d_{0,1}(H_1, d_1(H_1))],
\end{aligned}
where
\begin{aligned}
Q^d_{0,1}(h_1, d_1(h_1)) &= \mathbb{E}[Q^d_{0,2}(H_2, d_2(H_2))|H_1=h_1, \Delta_1 = 1, A_1 = d(h_1)]\\
Q^d_{0,2}(h_2, d_2(h_2)) &= \mathbb{E}[Q^d_{0,3}(H_3)|H_2 = h_2, \Delta_1 = 1, \Delta_2 = 1, A_2 = d(h_2)]\\
Q^d_{0,3}(h_3) &= \mathbb{E}[U|H_3 = h_3, \Delta_1 = 1, \Delta_2 = 2, \Delta_3 = 1].
\end{aligned}
Note that the observed target parameter can be seen as a 5-step inverse probability weighting expression, and that the associated g-computation formula can be reduced to 3 steps. This would not be the case if additional information was collected between $\Delta_k$ and $A_k$.

## Efficient estimation

For the same reason the (un-centralized) efficient/robust score reduces to

\begin{aligned}
Z_1(d,& g, Q^d, C)(O) \\
&= \,Q_{1}^{d}(H_1 , d_1(H_1))  \\
&+ \sum_{r = 1}^3 \left\{ \prod_{j = 1}^{r} \frac{\Delta_j}{C_j(H_j)} \frac{I\{A_j = d_j(H_{j})\}}{g_{j}(H_j, A_j)} \right\}\left\{Q_{r+1}^{d}(H_{r+1} , d_{r+1}(H_{r+1})) - Q_{r}^{d}(H_r , d_r(H_r))\right\},
\end{aligned}
where we set
\begin{aligned}
\frac{I\{A_3 = d_3(H_{3})\}}{g_{3}(H_3, A_3)} = 1, \qquad Q_{4}^{d}(H_{4} , d_{4}(H_{4})) = U.
\end{aligned}

The above expression for the score by direct calculation of the efficient influence function for the observed target parameter or by
utilizing a monotone coarsening argument, @tsiatis2006semiparametric or @laan2003unified.

A one-step estimator based on this score is implmented in polle via policy_eval().

## Adaption to continuous time right-censoring

Assume that $A_1$, $A_2$, $U$ occur at time points $T_1 < T_2 < T_3$. Let $C$ denote an underlying
censoring time point. Then we observe time points $\tilde T_k = \min(T_k, C)$ and non-misising indicators $\Delta_k = \{T_k \leq C\}$. Note that with this temporal structure $\prod_{j = 1}^k \Delta_j = \Delta_k$.
For notational convenience we include $\tilde T_k$ in the state variable $X_k$.

Now, let $C_k$ denote a survival model for the censoring process:

\begin{aligned}
C_{0,k}(\tilde T_k|H_k) = \mathbb{P}(C > T_k|H_k, C \geq \tilde T_{k-1})
\end{aligned}

\begin{aligned}
Z_1(d,& g, Q^d, C)(O) \\
&= \,Q_{1}^{d}(H_1 , d_1(H_1))  \\
&+ \sum_{r = 1}^3 \left\{ \prod_{j = 1}^{r} \frac{\Delta_j}{C_j(\tilde T_j|H_j)} \frac{I\{A_j = d_j(H_{j})\}}{g_{j}(H_j, A_j)} \right\}\left\{Q_{r+1}^{d}(H_{r+1} , d_{r+1}(H_{r+1})) - Q_{r}^{d}(H_r , d_r(H_r))\right\}.
\end{aligned}

We note that this score adapted to handle continuous right-censoring is not the efficient score which incorporate the induced temporal structure. A study of the second order remainder terms yields that for correctly specified nuisance models for $g_0, Q^d_0, C_0$, under reasonable convergence rate assumptions, the associated one-step estimator has asymptotic variance as induced by the influence function $Z_1(d,g_0, Q^d_0, C_0)(O) - \mathbb{E}[Z_1(d,g_0, Q^d_0, C_0)(O)]$.

## Terminal events

Terminal events may occur, resulting in a stochastic number of stages. However, auxiliary data can be constructed, and the methodology for a fixed-stage problem can be applied, taking into account the degenerate distribution of the new data.

Let $\epsilon_1$ and $\epsilon_2$ be terminal-event indicators included in the state covariates $X_1$
and $X_22$. Recall that $X_1$ and $X_2$ also include utility contributions $U_1$ and $U_2$, and
that the complete utility is $U = U_1 + U_2 + U_3$.

If a terminal event occurs at stage 2 we observe

\begin{align*}
O^\ast = \{B, X_1, \Delta_1, \Delta_1 A_1, \Delta_1 X_2, \Delta_1 U_2\}.
\end{align*}

We then construct auxiliary data

\begin{align*}
\{\Delta_2 = 1,  A_2 = a^\dagger, X_3 = \emptyset,  \Delta_3 = 0,  U_3 = 0\}
\end{align*}

and calculate the (un-centralized) doubly robust score using that if a terminal
event happens at stage 2, i.e., that $\epsilon_2 = 1$:

\begin{align*}
&C_2(H_2, \epsilon_2 = 1) = C_3(H_3, \epsilon_2 = 1) = 1\\
&g_2(H_2, \epsilon_2 = 1) = 1\\
&Q_2^d(H_2, \epsilon_2 = 1, a^\dagger) = Q_3^d(H_3, \epsilon_2 = 1) = Q_4^d(H_4,
\epsilon_2 = 1) = U_1 + U_2 = U
\end{align*}



## Examples

```{r simdata}
sim_single_stage_right_cens <- function(n = 2e3, zeta = c(0.7, 0.2), type = "right"){

  d <- sim_single_stage(n = n)
  pd <- policy_data(data = d,
                    action = "A",
                    covariates = c("Z", "L", "B"),
                    utility = "U")

  ld <- pd$stage_data

  ld[stage == 1, time := 1]
  ld[stage == 2, time := 2]

  ld[stage == 2, Z := d$Z]
  ld[stage == 2, L := d$L]
  ld[stage == 2, B := d$B]

  ## simulating the right censoring time
  ## only depending on the baseline covariate Z:
  C <- c(rexp(n, 1) / exp((-1) * cbind(1, as.numeric(d$Z)) %*% zeta))

  ld[stage == 1, time_c := C]
  ld[stage == 2, time_c := C]

  ld[, delta := time_c >= time]

  ld[delta == FALSE , event := 2]
  ld[delta == FALSE, A := NA]
  ld[delta == FALSE & stage == 2, U := NA]
  ld[delta == FALSE, U_A0 := 0]
  ld[delta == FALSE, U_A1 := 0]

  ld[ , tmp := shift(delta, fill = TRUE), by = list(id)]
  ld <- ld[tmp == TRUE, ]
  ld[ , time := pmin(time, time_c)]
  ld[ , time_c := NULL]
  ld[ , tmp := NULL]
  ld[ , delta := NULL]

  if (type == "interval"){
    ld[, time2 := time]
    ld[, time := shift(time, fill = 0), by = list(id)]
  }

  return(ld)
}
```

```{r simpd}
set.seed(1)
ld <- sim_single_stage_right_cens(n = 5e2, type = "interval")
pd <- policy_data(data = ld, type = "long", action = "A", time = "time", time2 = "time2")
```

```{r pe}
pe <- policy_eval(
  policy_data = pd,
  policy = policy_def(1),
  m_model = q_glm(~ .),
  m_full_history = TRUE,
  c_models = list(c_cox(formula = ~ Z_1),
                  c_cox(formula = ~ Z_2*A_1)),
  c_full_history = TRUE
)
pe
```

```{r pl}
pe <- policy_eval(
  policy_data = pd,
  policy_learn = policy_learn(type = "blip", control = control_blip()),
  m_model = q_glm(~.),
  m_full_history = FALSE,
  c_models = c_cox(formula = ~ Z)
)
pe
```

# References

---
title: "Estimating and Evaluating the Optimal Subgroup"
output:
  rmarkdown::html_vignette:
    fig_caption: true
    toc: true    
    toc_depth: 2
    mathjax: "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/MathJax.js?config=TeX-AMS_CHTML.js"
vignette: >
  %\VignetteIndexEntry{optimal_subgroup}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib  
---

```{r lib, message = FALSE}
library(polle)
library(data.table)
```

This vignette showcase how `policy_learn()` and `policy_eval()` can be
combined to estimate and evaluate the optimal subgroup in the single-stage
case. We refer to [@nordland2023policy] for the syntax and methodological context.

# Setup

From here on we consider the single-stage case with a binary action set
$\{0,1\}$. For a given threshold $\eta > 0$ we can formulate the optimal subgroup
function via the conditional average treatment effect (CATE/blip) as

\begin{align*}
d^\eta_0(v)_ = I\{B_0(v) > \eta\},
\end{align*}
where $B_0$ is the CATE defined as
\begin{align*}
E\left[U^{(1)} - U^{(0)} \big | V = v \right].
\end{align*}

The average treatment effect in the optimal subgroup is now defined as
\begin{align*}
E\left[U^{(1)} - U^{(0)} \big | d^\eta_0(V) = 1 \right],
\end{align*}

which under consistency, positivity and randomization is identified as

\begin{align*}
E\left[Z(1,g_0,Q_0)(O) - Z(0,g_0,Q_0)(O) \big | d^\eta_0(V) = 1 \right],
\end{align*}

where $Z(a,g,Q)(O)$ is the doubly robust score for treatment $a$ and

\begin{align*}
d^\eta_0(v) &= I\{B_0(v) > \eta\}\\
B_0(v) &= E\left[Z(1,g_0,Q_0)(O) - Z(0,g_0,Q_0)(O) \big | V = v \right]
\end{align*}


# Threshold policy learning

In `polle` the threshold policy $d_\eta$ can be estimated using `policy_learn()`
via the `threshold` argument, and the average treatment effect in the subgroup
can be estimated using `policy_eval()` setting `target = subgroup`.

Here we consider an example using simulated data:

```{r}
par0 <- list(a = 1, b = 0, c = 3)
sim_d <- function(n, par=par0, potential_outcomes = FALSE) {
  W <- runif(n = n, min = -1, max = 1)
  L <- runif(n = n, min = -1, max = 1)
  A <- rbinom(n = n, size = 1, prob = 0.5)
  U1 <- W + L + (par$c*W + par$a*L + par$b) # U^1
  U0 <- W + L # U^0
  U <- A * U1 + (1 - A) * U0 + rnorm(n = n)
  out <- data.table(W = W, L = L, A = A, U = U)
  if (potential_outcomes == TRUE) {
    out$U0 <- U0
    out$U1 <- U1
  }
  return(out)
}
```

Note that in this simple case $U^{(1)} - U^{(0)} = cW + aL + b$.

```{r single stage data}
set.seed(1)
d <- sim_d(n = 200)
pd <- policy_data(
    d,
    action = "A",
    covariates = list("W", "L"),
    utility = "U"
)
```

We set a correctly specified policy learner using `policy_learn()` with `type = "blip"`  and set
a threshold of $\eta = 1$:

```{r}
pl1 <- policy_learn(
  type = "blip",
  control = control_blip(blip_models = q_glm(~ W + L)),
  threshold = 1
)
```

When then apply the policy learner based on the correctly
specified nuisance models. Furthermore, we extract
the corresponding policy actions, where $d_N(Z,L) = 1$
identifies the optimal subgroup for $\eta = 1$:

```{r}
po1 <- pl1(
  policy_data = pd,
  g_models = g_glm(~ 1),
  q_models = q_glm(~ A * (W + L))
)
pf1 <- get_policy(po1)
pa <- pf1(pd)
```

In the following plot, the black line indicates the boundary
for the true optimal subgroup. The dots represent the
estimated threshold policy:

```{r pa_plot, echo = FALSE}
library("ggplot2")
plot_data <- data.table(d_N = factor(pa$d), W = d$W, L = d$L)
ggplot(plot_data) +
  geom_point(aes(x = W, y = L, color = d_N)) +
  geom_abline(slope = -3, intercept = 1) +
  theme_bw()
```

Similarly, we can also use `type = "ptl"` to fit a policy tree with
a given threshold for not choosing the reference action
(first action in action set in alphabetical order)

```{r}
get_action_set(pd)[1] # reference action
pl1_ptl <- policy_learn(
    type = "ptl",
    control = control_ptl(policy_var = c("W", "L")),
    threshold = 1
)
po1_ptl <- pl1_ptl(
  policy_data = pd,
  g_models = g_glm(~ 1),
  q_models = q_glm(~ A * (W + L))
)
po1_ptl$ptl_objects
```

```{r pa_plot_ptl, echo = FALSE}
pf1_ptl <- get_policy(po1_ptl)
pa_ptl <- pf1_ptl(pd)
library("ggplot2")
plot_data <- data.table(d_N = factor(pa_ptl$d), W = d$W, L = d$L)
ggplot(plot_data) +
  geom_point(aes(x = W, y = L, color = d_N)) +
  geom_abline(slope = -3, intercept = 1) +
  theme_bw()
```

# Subgroup average treatment effect

The true subgroup average treatment effect is given by:

\begin{align*}
E[cW + aL + b | cW + aL + b \geq \eta ],
\end{align*}

which we can easily approximate:

```{r sate, cache = TRUE}
set.seed(1)
approx <- sim_d(n = 1e7, potential_outcomes = TRUE)
(sate <- with(approx, mean((U1 - U0)[(U1 - U0 >= 1)])))
rm(approx)
```

The subgroup average treatment effect associated with the learned optimal threshold policy
can be directly estimated using `policy_eval()` via the `target` argument:

```{r pe}
(pe <- policy_eval(
  policy_data = pd,
  policy_learn = pl1,
  target = "subgroup"
 ))
```

We can also estimate the subgroup average treatment effect for a set of thresholds at once:

```{r}
pl_set <- policy_learn(
  type = "blip",
  control = control_blip(blip_models = q_glm(~ W + L)),
  threshold = c(0, 1)
)

policy_eval(
  policy_data = pd,
  g_models = g_glm(~ 1),
  q_models = q_glm(~ A * (W + L)),
  policy_learn = pl_set,
  target = "subgroup"
)
```

## Asymptotics

The data adaptive target parameter

\begin{align*}
E[U^{(1)} - U^{(0)}| d_N(V) = 1] = E[Z_0(1,g,Q)(O) - Z_0(0,g,Q)(O)| d_N(V) = 1]
\end{align*}

is asymptotically normal with influence function

\begin{align*}
\frac{1}{P(d'(\cdot) = 1)} I\{d'(\cdot) = 1\}\left\{Z(1,g,Q)(O) - Z(0,g,Q)(O) - E[Z(1,g,Q)(O) - Z(0,g,Q)(O) | d'(\cdot) = 1]\right\},
\end{align*}

where $d'$ is the limiting policy of $d_N$. The fitted influence curve can be extracted using `IC()`:

```{r}
IC(pe) |> head()
```

# References

---
title: "policy_eval"
output:
  rmarkdown::html_vignette:
    fig_caption: true
    toc: true    
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{policy_eval}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib  
---

```{r lib}
library(polle)
```

This vignette is a guide to `policy_eval()` and some of the associated S3 methods.
The purpose of `policy_eval` is to estimate (evaluate) the value of a user-defined policy or a policy learning algorithm.
For details on the methodology, see the associated paper [@nordland2023policy].

We consider a fixed two-stage problem as a general setup and simulate data using `sim_two_stage()` and create a `policy_data` object using `policy_data()`:

```{r simdata}
d <- sim_two_stage(n = 2e3, seed = 1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = c("B", "BB"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd
```

## Evaluating a user-defined policy

User-defined policies are created using `policy_def()`. In this case we define a simple static policy always
selecting action `'1'`:

```{r poldef}
p1 <- policy_def(policy_functions = '1', reuse = TRUE, name = "(A=1)")
```

As we want to apply the same policy function at both stages we set `reuse = TRUE`.

`policy_eval()` implements three types of policy evaluations: Inverse
probability weighting estimation, outcome regression estimation, and doubly
robust (DR) estimation. As doubly robust estimation is a combination of the two
other types, we focus on this approach. For details on the implementation see
Algorithm 1 in [@nordland2023policy]. 

```{r polevalp1}
(pe1 <- policy_eval(policy_data = pd,
                    policy = p1,
                    type = "dr"))
```

`policy_eval()` returns an object of type `policy_eval` which prints like a `lava::estimate` object. The policy value estimate and variance are available via `coef()` and `vcov()`:

```{r coef}
coef(pe1)
vcov(pe1)
```

### Working with `policy_eval` objects

The `policy_eval` object behaves like an `lava::estimate` object, which can also be directly accessed using `estimate()`. 

`estimate` objects makes it easy to work with estimates with an iid decomposition given by the influence curve/function, see the [estimate vignette](https://CRAN.R-project.org/package=lava/vignettes/influencefunction.html).

The influence curve is available via `IC()`:

```{r ic}
IC(pe1) |> head()
```

Merging `estimate` objects allow the user to get inference for transformations of the estimates via the Delta method.
Here we get inference for the average treatment effect, both as a difference and as a ratio:

```{r estimate_merge}
p0 <- policy_def(policy_functions = 0, reuse = TRUE, name = "(A=0)")
pe0 <- policy_eval(policy_data = pd,
                   policy = p0,
                   type = "dr")

(est <- merge(pe0, pe1))
estimate(est, function(x) x[2]-x[1], labels="ATE-difference")
estimate(est, function(x) x[2]/x[1], labels="ATE-ratio")
```

### Nuisance models

So far we have relied on the default generalized linear models for the nuisance g-models and Q-models.
As default, a single g-model trained across all stages using the state/Markov type history, see the
`policy_data` vignette. Use `get_g_functions()` to get access to the
fitted model:

```{r get_nuisance}
(gf <- get_g_functions(pe1))
```

The g-functions can be used as input to a new policy evaluation:
```{r pe_pol}
policy_eval(policy_data = pd,
            g_functions = gf,
            policy = p0,
            type = "dr")
```

or we can get the associated predicted values:

```{r pred_gfun}
predict(gf, new_policy_data = pd) |> head(6)
```

Similarly, we can inspect the Q-functions using `get_q_functions()`:

```{r get_qfun}
get_q_functions(pe1)
```
Note that a model is trained for each stage. Again, we can predict from the Q-models using `predict()`.


Usually, we want to specify the nuisance models ourselves using the `g_models` and `q_models` arguments:

```{r gq_models, cache = TRUE}
pe1 <- policy_eval(pd,
            policy = p1,
            g_models = list(
              g_sl(formula = ~ BB + L_1, SL.library = c("SL.glm", "SL.ranger")),
              g_sl(formula = ~ BB + L_1 + C_2, SL.library = c("SL.glm", "SL.ranger"))
            ),
            g_full_history = TRUE,
            q_models = list(
              q_glm(formula = ~ A * (B + C_1)), # including action interactions
              q_glm(formula = ~ A * (B + C_1 + C_2)) # including action interactions
            ),
            q_full_history = TRUE)
```

Here we train a super learner g-model for each stage using the full available history and a generalized linear model for the Q-models. The `formula` argument is used to construct the model frame passed to the model for training (and prediction).
The valid formula terms depending on `g_full_history` and `q_full_history` are available via `get_history_names()`:

```{r history_names}
get_history_names(pd) # state/Markov history
get_history_names(pd, stage = 1) # full history
get_history_names(pd, stage = 2) # full history
```

Remember that the action variable at the current stage is always named `A`. Some models like `glm` require interactions to be specified via the model frame. Thus, for some models, it is important to include action interaction terms for the Q-models.

## Evaluating a policy learning algorithm

The value of a learned policy is an important performance measure, and `policy_eval()` allow for
direct evaluation of a given policy learning algorithm. For details, see Algorithm 4 in [@nordland2023policy].

In `polle`, policy learning algorithms are specified using `policy_learn()`, see the associated vignette. These functions can be directly evaluated in `policy_eval()`:

```{r pe_pl}
policy_eval(pd,
            policy_learn = policy_learn(type = "ql"))
```

In the above example we evaluate the policy estimated via Q-learning. Alternatively,
we can first learn the policy and then pass it to `policy_eval()`:

```{r pe_manual_pol_eval}
p_ql <- policy_learn(type = "ql")(pd, q_models = q_glm())
policy_eval(pd,
            policy = get_policy(p_ql))
```

## Cross-fitting

A key feature of `policy_eval()` is that it allows for easy cross-fitting of the nuisance models as well the learned policy.
Here we specify two-fold cross-fitting via the `M` argument:

```{r crossfits}
pe_cf <- policy_eval(pd,
                     policy_learn = policy_learn(type = "ql"),
                     M = 2)
```

Specifically, both the nuisance models and the optimal policy are fitted on each training fold.
Subsequently, the doubly robust value score is calculated on the validation folds.

The `policy_eval` object now consists of a list of `policy_eval` objects associated with each fold:

```{r pe_crossfits}
pe_cf$folds$fold_1 |> head()
pe_cf$cross_fits$fold_1
```

In order to save memory, particularly when cross-fitting, it is possible not to save the nuisance models via the
`save_g_functions` and `save_q_functions` arguments.

## Parallel processing via `future.apply`

It is easy to parallelize the cross-fitting procedure via the `future.apply` package:

```{r demo_future, eval = FALSE}
library(future.apply)
plan("multisession") # local parallel procession
library("progressr") # progress bar
handlers(global = TRUE)

policy_eval(pd,
            policy_learn = policy_learn(type = "ql"),
            q_models = q_rf(),
            M = 20)

plan("sequential") # resetting to sequential processing
```

# SessionInfo

```{r sessionInfo}
sessionInfo()
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{get_stage_action_sets}
\alias{get_stage_action_sets}
\title{Get Stage Action Sets}
\usage{
get_stage_action_sets(object)
}
\arguments{
\item{object}{Object of class \link{policy_data}.}
}
\value{
List of character vectors.
}
\description{
\code{get_stage_action_sets} returns the action sets at each stage, i.e.,
the possible actions at each stage for the policy data object.
}
\examples{
### Two stages:
d <- sim_two_stage_multi_actions(5e2, seed=1)
# constructing policy_data object:
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = c("B"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

# getting the stage actions set:
get_stage_action_sets(pd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_object_functions.R
\name{get_policy_object}
\alias{get_policy_object}
\title{Get Policy Object}
\usage{
get_policy_object(object)
}
\arguments{
\item{object}{Object of class \link{policy_eval}.}
}
\value{
Object of class \link{policy_object}.
}
\description{
Extract the fitted policy object.
}
\examples{
library("polle")
### Single stage:
d1 <- sim_single_stage(5e2, seed=1)
pd1 <- policy_data(d1, action="A", covariates=list("Z", "B", "L"), utility="U")
pd1


# evaluating the policy:
pe1 <- policy_eval(policy_data = pd1,
                   policy_learn = policy_learn(type = "drql",
                                               control = control_drql(qv_models = q_glm(~.))),
                   g_models = g_glm(),
                   q_models = q_glm())

# extracting the policy object:
get_policy_object(pe1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation.R
\name{sim_two_stage}
\alias{sim_two_stage}
\title{Simulate Two-Stage Data}
\usage{
sim_two_stage(
  n = 10000,
  par = c(gamma = 0.5, beta = 1),
  seed = NULL,
  action_model_1 = function(C_1, beta, ...) stats::rbinom(n = NROW(C_1), size = 1, prob =
    lava::expit(beta * C_1)),
  action_model_2 = function(C_2, beta, ...) stats::rbinom(n = NROW(C_2), size = 1, prob =
    lava::expit(beta * C_2)),
  deterministic_rewards = FALSE
)
}
\arguments{
\item{n}{Number of observations.}

\item{par}{Named vector with distributional parameters.
\itemize{
\item{} \code{gamma}: \eqn{\gamma}
\item{} \code{beta}: \eqn{\beta}
}}

\item{seed}{Integer.}

\item{action_model_1}{Function used to specify the action/treatment at stage 1.}

\item{action_model_2}{Function used to specify the action/treatment at stage 2.}

\item{deterministic_rewards}{Logical. If TRUE, the deterministic reward
contributions are returned as well (columns U_1_A0, U_1_A1, U_2_A0, U_2_A1).}
}
\value{
\link[data.table:data.table]{data.table::data.table} with n rows and columns B, BB, L_1, C_1, A_1, L_2, C_2,
A_2, L_3, U_1, U_2, U_3 (,U_1_A0, U_1_A1, U_2_A0, U_2_A1).
}
\description{
Simulate Two-Stage Data
}
\details{
\code{sim_two_stage} samples \code{n} iid observation
\eqn{O} with the following distribution:
\eqn{BB} is a random categorical variable with levels \code{group1},
\code{group2}, and \code{group3}. Furthermore,
\deqn{
B \sim \mathcal{N}(0,1)\\
L_{1} \sim \mathcal{N}(0, 1)\\
C_{1} \mid L_{1} \sim \mathcal{N}(L_1, 1)\\
A_1 \mid C_1 \sim Bernoulli(expit(\beta C_1))\\
L_{2} \sim \mathcal{N} (0, 1)\\
C_{2} \mid A_1, L_1 \sim \mathcal{N}(\gamma L_1 + A_1, 1)\\
A_2 \mid C_2 \sim Bernoulli(expit(\beta C_2))\\
L_{3} \sim \mathcal{N} (0, 1)
}
The rewards are calculated as
\deqn{
U_1 = L_1\\
U_2 = A_1\cdot C_1 + L_2 \\
U_3 = A_2\cdot C_2 + L_3.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g_function.R
\name{get_g_functions}
\alias{get_g_functions}
\title{Get g-functions}
\usage{
get_g_functions(object)
}
\arguments{
\item{object}{Object of class \link{policy_eval} or \link{policy_object}.}
}
\value{
List of class \link{nuisance_functions}.
}
\description{
\code{get_g_functions()} returns a list of (fitted) g-functions
associated with each stage.
}
\examples{
### Two stages:
d <- sim_two_stage(5e2, seed=1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = c("B"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

# evaluating the static policy a=1 using inverse propensity weighting
# based on a GLM model at each stage
pe <- policy_eval(type = "ipw",
                  policy_data = pd,
                  policy = policy_def(1, reuse = TRUE, name = "A=1"),
                  g_models = list(g_glm(), g_glm()))
pe

# getting the g-functions
g_functions <- get_g_functions(pe)
g_functions

# getting the fitted g-function values
head(predict(g_functions, pd))
}
\seealso{
\link{predict.nuisance_functions}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{history}
\alias{history}
\alias{get_history}
\title{Get History Object}
\usage{
get_history(object, stage = NULL, full_history = FALSE, event_set = c(0))
}
\arguments{
\item{object}{Object of class \link{policy_data}.}

\item{stage}{Stage number. If NULL, the state/Markov-type history across
all stages is returned.}

\item{full_history}{Logical. If TRUE, the full history is returned
If FALSE, only the state/Markov-type history is returned.}

\item{event_set}{Integer vector. Subset of c(0,1,2).}
}
\value{
Object of class \link{history}. The object is a list
containing the following elements:
\item{\code{H}}{\link[data.table:data.table]{data.table::data.table} with keys id and stage and with variables
\{\eqn{B}, \eqn{X_k}\} (state history) or
\{\eqn{B}, \eqn{X_1}, \eqn{A_1}, ..., \eqn{X_k}\}
(full history), see details.}
\item{\code{A}}{\link[data.table:data.table]{data.table::data.table} with keys id and stage and variable \eqn{A_k}, see
details.}
\item{action_name}{Name of the action variable in \code{A}.}
\item{action_set}{Sorted character vector defining the action set.}
\item{U}{(If \code{stage} is not NULL) \link[data.table:data.table]{data.table::data.table} with keys id and stage
and with variables U_bar and U_Aa for every a in the actions set.
U_bar is the accumulated rewards up till and including the given
stage, i.e., \eqn{\sum_{j=1}^k U_j}. U_Aa is the deterministic
reward of action a.}
}
\description{
\code{get_history} summarizes the history and action at a given stage from a
\link{policy_data} object.
}
\details{
Each observation has the sequential form
\deqn{O= {B, U_1, X_1, A_1, ..., U_K, X_K, A_K, U_{K+1}},}
for a possibly stochastic number of stages K.
\itemize{
\item{} \eqn{B} is a vector of baseline covariates.
\item{} \eqn{U_k} is the reward at stage k (not influenced by the action \eqn{A_k}).
\item{} \eqn{X_k} is a vector of state covariates summarizing the state at stage k.
\item{} \eqn{A_k} is the categorical action at stage k.
}
}
\examples{
library("polle")
### Single stage:
d1 <- sim_single_stage(5e2, seed=1)
# constructing policy_data object:
pd1 <- policy_data(d1, action="A", covariates=list("Z", "B", "L"), utility="U")
pd1

# In the single stage case, set stage = NULL
h1 <- get_history(pd1)
head(h1$H)
head(h1$A)

### Two stages:
d2 <- sim_two_stage(5e2, seed=1)
# constructing policy_data object:
pd2 <- policy_data(d2,
                  action = c("A_1", "A_2"),
                  baseline = c("B"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd2
# getting the state/Markov-type history across all stages:
h2 <- get_history(pd2)
head(h2$H)
head(h2$A)

# getting the full history at stage 2:
h2 <- get_history(pd2, stage = 2, full_history = TRUE)
head(h2$H)
head(h2$A)
head(h2$U)

# getting the state/Markov-type history at stage 2:
h2 <- get_history(pd2, stage = 2, full_history = FALSE)
head(h2$H)
head(h2$A)

### Multiple stages
d3 <- sim_multi_stage(5e2, seed = 1)
# constructing policy_data object:
pd3 <- policy_data(data = d3$stage_data,
                   baseline_data = d3$baseline_data,
                   type = "long",
                   id = "id",
                   stage = "stage",
                   event = "event",
                   action = "A",
                   utility = "U")
pd3

# getting the full history at stage 2:
h3 <- get_history(pd3, stage = 2, full_history = TRUE)
head(h3$H)
# note that not all observations have two stages:
nrow(h3$H) # number of observations with two stages.
get_n(pd3) # number of observations in total.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_eval_functions.R
\name{conditional}
\alias{conditional}
\title{Conditional Policy Evaluation}
\usage{
conditional(object, policy_data, baseline)
}
\arguments{
\item{object}{Policy evaluation object created by \code{\link[=policy_eval]{policy_eval()}}.}

\item{policy_data}{Policy data object created by \code{\link[=policy_data]{policy_data()}}.}

\item{baseline}{Character string.}
}
\value{
object of inherited class 'estimate', see \link[lava:estimate.default]{lava::estimate.default}.
The object is a list with elements 'coef' (policy value estimate for each
group) and 'IC' (influence curve estimate matrix).
}
\description{
\code{conditional()} is used to calculate the
policy value for each group defined by a given baseline variable.
}
\examples{
library("polle")
library("data.table")
setDTthreads(1)
d <- sim_single_stage(n=2e3)
pd <- policy_data(d,
                  action = "A",
                  baseline = c("B"),
                  covariates = c("Z","L"),
                  utility = "U")

# static policy:
p <- policy_def(1)

pe <- policy_eval(pd,
                  policy = p)

# conditional value for each group defined by B
conditional(pe, pd, "B")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{plot.policy_data}
\alias{plot.policy_data}
\title{Plot policy data for given policies}
\usage{
\method{plot}{policy_data}(
  x,
  policy = NULL,
  which = c(1),
  stage = 1,
  history_variables = NULL,
  jitter = 0.05,
  ...
)
}
\arguments{
\item{x}{Object of class \link{policy_data}}

\item{policy}{An object or list of objects of class \link{policy}}

\item{which}{A subset of the numbers 1:2
\itemize{
\item{1} Spaghetti plot of the cumulative rewards
\item{2} Plot of the policy actions for a given stage
}}

\item{stage}{Stage number for plot 2}

\item{history_variables}{character vector of length 2 for plot 2}

\item{jitter}{numeric}

\item{...}{Additional arguments}
}
\description{
Plot policy data for given policies
}
\examples{
library("polle")
library("data.table")
setDTthreads(1)
d3 <- sim_multi_stage(2e2, seed = 1)
pd3 <- policy_data(data = d3$stage_data,
                   baseline_data = d3$baseline_data,
                   type = "long",
                   id = "id",
                   stage = "stage",
                   event = "event",
                   action = "A",
                   utility = "U")
pd3 <- partial(pd3, K = 3)

# specifying two static policies:
p0 <- policy_def(c(1,1,0), name = "p0")
p1 <- policy_def(c(1,0,0), name = "p1")

plot(pd3)
plot(pd3, policy = list(p0, p1))

# learning and plotting a policy:
 pe3 <- policy_eval(pd3,
                    policy_learn = policy_learn(),
                    q_models = q_glm(formula = ~t + X + X_lead))
plot(pd3, list(get_policy(pe3), p0))

# plotting the recommended actions at a specific stage:
plot(pd3, get_policy(pe3),
     which = 2,
     stage = 2,
     history_variables = c("t","X"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{get_action_set}
\alias{get_action_set}
\title{Get Action Set}
\usage{
get_action_set(object)
}
\arguments{
\item{object}{Object of class \link{policy_data}.}
}
\value{
Character vector.
}
\description{
\code{get_action_set} returns the action set, i.e., the possible
actions at each stage for the policy data object.
}
\examples{
### Two stages:
d <- sim_two_stage(5e2, seed=1)
# constructing policy_data object:
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = c("B"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

# getting the actions set:
get_action_set(pd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polle-package.R
\docType{package}
\name{polle-package}
\alias{polle}
\alias{polle-package}
\title{polle: Policy Learning}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Package for learning and evaluating (subgroup) policies via doubly robust loss functions. Policy learning methods include doubly robust blip/conditional average treatment effect learning and sequential policy tree learning. Methods for (subgroup) policy evaluation include doubly robust cross-fitting and online estimation/sequential validation. See Nordland and Holst (2026) \doi{10.18637/jss.v116.i04} for documentation and references.
}
\seealso{
Useful links:
\itemize{
  \item Report bugs at \url{https://github.com/AndreasNordland/polle/issues}
}

}
\author{
\strong{Maintainer}: Andreas Nordland \email{andreasnordland@gmail.com}

Authors:
\itemize{
  \item Klaus Holst \email{klaus@holst.it} (\href{https://orcid.org/0000-0002-1364-6789}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drql.R
\name{control_drql}
\alias{control_drql}
\title{Control arguments for doubly robust Q-learning}
\usage{
control_drql(qv_models = q_glm(~.))
}
\arguments{
\item{qv_models}{Single element or list of V-restricted Q-models created
by \code{\link[=q_glm]{q_glm()}}, \code{\link[=q_rf]{q_rf()}}, \code{\link[=q_sl]{q_sl()}} or similar functions.}
}
\value{
list of (default) control arguments.
}
\description{
\code{control_drql} sets the default control arguments
for doubly robust Q-learning, \code{type = "drql"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_eval_functions.R
\name{get_policy_actions}
\alias{get_policy_actions}
\title{Get Policy Actions}
\usage{
get_policy_actions(object)
}
\arguments{
\item{object}{Object of class \link{policy_eval}.}
}
\value{
\link[data.table:data.table]{data.table::data.table} with keys \code{id} and \code{stage} and action variable
\code{d}.
}
\description{
\code{get_policy_actions()} extract the actions dictated by the
(learned and possibly cross-fitted) policy a every stage.
}
\examples{
### Two stages:
d <- sim_two_stage(5e2, seed=1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

# defining a policy learner based on cross-fitted doubly robust Q-learning:
pl <- policy_learn(type = "drql",
                   control = control_drql(qv_models = list(q_glm(~C_1), q_glm(~C_1+C_2))),
                   full_history = TRUE,
                   L = 2) # number of folds for cross-fitting

# evaluating the policy learner using 2-fold cross fitting:
pe <- policy_eval(type = "dr",
                   policy_data = pd,
                   policy_learn = pl,
                   q_models = q_glm(),
                   g_models = g_glm(),
                   M = 2) # number of folds for cross-fitting

# Getting the cross-fitted actions dictated by the fitted policy:
head(get_policy_actions(pe))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Q_function.R
\name{get_q_functions}
\alias{get_q_functions}
\title{Get Q-functions}
\usage{
get_q_functions(object)
}
\arguments{
\item{object}{Object of class \link{policy_eval} or \link{policy_object}.}
}
\value{
List of class \link{nuisance_functions}.
}
\description{
\code{get_q_functions()} returns a list of (fitted) Q-functions
associated with each stage.
}
\examples{
### Two stages:
d <- sim_two_stage(5e2, seed = 1)
pd <- policy_data(d,
  action = c("A_1", "A_2"),
  baseline = c("B"),
  covariates = list(
    L = c("L_1", "L_2"),
    C = c("C_1", "C_2")
  ),
  utility = c("U_1", "U_2", "U_3")
)
pd

# evaluating the static policy a=1 using outcome regression
# based on a GLM model at each stage.
pe <- policy_eval(
  type = "or",
  policy_data = pd,
  policy = policy_def(1, reuse = TRUE, name = "A=1"),
  q_models = list(q_glm(), q_glm())
)
pe

# getting the Q-functions
q_functions <- get_q_functions(pe)

# getting the fitted g-function values
head(predict(q_functions, pd))
}
\seealso{
\link{predict.nuisance_functions}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_learn.R, R/policy_object_functions.R
\name{policy_learn}
\alias{policy_learn}
\alias{policy_object}
\alias{print.policy_learn}
\alias{print.policy_object}
\title{Create Policy Learner}
\usage{
policy_learn(
  type = "blip",
  control = control_blip(),
  alpha = 0,
  threshold = NULL,
  full_history = FALSE,
  L = 1,
  cross_fit_g_models = TRUE,
  cross_fit_c_models = TRUE,
  save_cross_fit_models = FALSE,
  future_args = list(future.seed = TRUE),
  name = type
)

\method{print}{policy_learn}(x, ...)

\method{print}{policy_object}(x, ...)
}
\arguments{
\item{type}{Type of policy learner method:
\itemize{
\item \code{"ql"}: Quality/Q-learning.
\item \code{"drql"}: Doubly Robust Q-learning.
\item \code{"blip"}:
Doubly Robust blip-learning (only for dichotomous actions).
\item \code{"ptl"}: Policy Tree Learning.
\item \code{"owl"}: Outcome Weighted Learning.
\item \code{"earl"}:
Efficient Augmentation and Relaxation Learning (only single stage).
\item \code{"rwl"}: Residual Weighted Learning (only single stage).
}}

\item{control}{List of control arguments.
Values (and default values) are set using
\code{control_{type}()}. Key arguments include:\cr
\code{\link[=control_drql]{control_drql()}}:\cr
\itemize{
\item \code{qv_models}:
Single element or list of V-restricted Q-models created
by \code{\link[=q_glm]{q_glm()}}, \code{\link[=q_rf]{q_rf()}}, \code{\link[=q_sl]{q_sl()}} or similar functions.
}
\code{\link[=control_blip]{control_blip()}}:\cr
\itemize{
\item \code{blip_models}:
Single element or list of V-restricted blip-models created
by \code{\link[=q_glm]{q_glm()}}, \code{\link[=q_rf]{q_rf()}}, \code{\link[=q_sl]{q_sl()}} or similar functions.
}
\code{\link[=control_ptl]{control_ptl()}}: \cr
\itemize{
\item \code{policy_vars}: Character vector/string or list of character
vectors/strings. Variable names used to construct
the V-restricted policy tree.
The names must be a subset of the history names, see get_history_names().
\item \code{hybrid}: If \code{TRUE},
\code{\link[policytree:hybrid_policy_tree]{policytree::hybrid_policy_tree()}} is used to
fit a policy tree.
\item \code{depth}:
Integer or integer vector. The depth of the fitted policy
tree for each stage.
}
\code{\link[=control_owl]{control_owl()}}: \cr
\itemize{
\item \code{policy_vars}: As in \code{control_ptl()}.
\item \code{loss}: Loss function.
The options are \code{"hinge"}, \code{"ramp"},
\code{"logit"}, \code{"logit.lasso"}, \code{"l2"}, \code{"l2.lasso"}.
\item \code{kernel}: Type of kernel
used by the support vector machine. The
options are \code{"linear"}, \code{"rbf"}.
\item \code{augment}:  If \code{TRUE} the outcomes are augmented.
}
\code{\link[=control_earl]{control_earl()}}/\code{\link[=control_rwl]{control_rwl()}}: \cr
\itemize{
\item \code{moPropen}:
Propensity model of class "ModelObj", see \link[modelObj:modelObj]{modelObj::modelObj}.
\item \code{moMain}: Main effects outcome model of class "ModelObj".
\item \code{moCont} Contrast outcome model of class "ModelObj".
\item \code{regime}:
An object of class \link{formula} specifying the design of the policy.
\item \code{surrogate}:
The surrogate 0-1 loss function. The options are
\code{"logit"}, \code{"exp"},
\code{"hinge"}, \code{"sqhinge"}, \code{"huber"}.
\item \code{kernel}: The options are
\code{"linear"}, \code{"poly"}, \code{"radial"}.
}}

\item{alpha}{Probability threshold for determining realistic actions.}

\item{threshold}{Numeric vector, thresholds for not
choosing the reference action at stage 1.}

\item{full_history}{If \code{TRUE}, the full
history is used to fit each policy function (e.g. QV-model,
policy tree). If FALSE, the single stage/
"Markov type" history is used to fit each policy function.}

\item{L}{Number of folds for cross-fitting nuisance models.}

\item{cross_fit_g_models}{If \code{TRUE}, the g-models will not be
cross-fitted even if L > 1.}

\item{cross_fit_c_models}{If \code{TRUE}, the c-models will not be
cross-fitted even if L > 1.}

\item{save_cross_fit_models}{If \code{TRUE}, the cross-fitted
models will be saved.}

\item{future_args}{Arguments passed to \code{\link[future.apply:future_apply]{future.apply::future_apply()}}.}

\item{name}{Character string.}

\item{x}{Object of class "policy_object" or "policy_learn".}

\item{...}{Additional arguments passed to print.}
}
\value{
Function of inherited class \code{"policy_learn"}.
Evaluating the function on a \link{policy_data} object returns an object of
class \link{policy_object}. A policy object is a list containing all or
some of the following elements:
\describe{
\item{\code{q_functions}}{Fitted Q-functions. Object
of class "nuisance_functions".}
\item{\code{g_functions}}{Fitted g-functions. Object
of class "nuisance_functions".}
\item{\code{action_set}}{Sorted character vector describing
the action set, i.e.,
the possible actions at each stage.}
\item{\code{alpha}}{Numeric. Probability threshold to determine
realistic actions.}
\item{\code{K}}{Integer. Maximal number of stages.}
\item{\code{qv_functions}}{(only if \code{type = "drql"}) Fitted
V-restricted Q-functions. Contains a fitted model for each stage and action.}
\item{\code{ptl_objects}}{(only if \code{type = "ptl"}) Fitted V-restricted
policy trees. Contains a \link[policytree:policy_tree]{policytree::policy_tree} for each stage.}
\item{\code{ptl_designs}}{(only if \code{type = "ptl"}) Specification of the
V-restricted design matrix for each stage}
}
}
\description{
\code{policy_learn()} is used to specify a policy
learning method (Q-learning,
doubly robust Q-learning, policy tree
learning and outcome weighted learning).
Evaluating the policy learner returns a policy object.
}
\section{S3 generics}{

The following S3 generic functions are available for an object of
class "policy_object":
\describe{
\item{\code{\link[=get_g_functions]{get_g_functions()}}}{Extract the fitted g-functions.}
\item{\code{\link[=get_q_functions]{get_q_functions()}}}{Extract the fitted Q-functions.}
\item{\code{\link[=get_policy]{get_policy()}}}{Extract the fitted policy object.}
\item{\code{\link[=get_policy_functions]{get_policy_functions()}}}{Extract the fitted policy function for
a given stage.}
\item{\code{\link[=get_policy_actions]{get_policy_actions()}}}{Extract the (fitted) policy actions.}
}
}

\examples{
library("polle")
### Two stages:
d <- sim_two_stage(5e2, seed=1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = c("BB"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

### V-restricted (Doubly Robust) Q-learning

# specifying the learner:
pl <- policy_learn(
  type = "drql",
  control = control_drql(qv_models = list(q_glm(formula = ~ C_1 + BB),
                                          q_glm(formula = ~ L_1 + BB))),
  full_history = TRUE
)

# evaluating the learned policy
pe <- policy_eval(policy_data = pd,
                  policy_learn = pl,
                  q_models = q_glm(),
                  g_models = g_glm())
pe
# getting the policy object:
po <- get_policy_object(pe)
# inspecting the fitted QV-model for each action strata at stage 1:
po$qv_functions$stage_1
head(get_policy(pe)(pd))
}
\references{
Doubly Robust Q-learning (\code{type = "drql"}): Luedtke, Alexander R., and
Mark J. van der Laan. "Super-learning of an optimal dynamic treatment rule."
The international journal of biostatistics 12.1 (2016): 305-332.
\doi{10.1515/ijb-2015-0052}.\cr
\cr
Policy Tree Learning (\code{type = "ptl"}): Zhou, Zhengyuan, Susan Athey,
and Stefan Wager. "Offline multi-action policy learning: Generalization and
optimization." Operations Research (2022). \doi{10.1287/opre.2022.2271}.\cr
\cr
(Augmented) Outcome Weighted Learning: Liu, Ying, et al. "Augmented
outcome‐weighted learning for estimating optimal dynamic treatment regimens."
Statistics in medicine 37.26 (2018): 3776-3788. \doi{10.1002/sim.7844}.
}
\seealso{
\code{\link[=policy_eval]{policy_eval()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_eval_online.R
\name{policy_eval_online}
\alias{policy_eval_online}
\title{Online/Sequential Policy Evaluation}
\usage{
policy_eval_online(
  policy_data,
  policy = NULL,
  policy_learn = NULL,
  g_functions = NULL,
  g_models = g_glm(),
  g_full_history = FALSE,
  save_g_functions = TRUE,
  q_functions = NULL,
  q_models = q_glm(),
  q_full_history = FALSE,
  save_q_functions = TRUE,
  c_functions = NULL,
  c_models = NULL,
  c_full_history = FALSE,
  save_c_functions = TRUE,
  m_function = NULL,
  m_model = NULL,
  m_full_history = FALSE,
  save_m_function = TRUE,
  target = "value",
  M = 4,
  train_block_size = get_n(policy_data)/5,
  name = NULL,
  min_subgroup_size = 1
)
}
\arguments{
\item{policy_data}{Policy data object created by \code{\link[=policy_data]{policy_data()}}.}

\item{policy}{Policy object created by \code{\link[=policy_def]{policy_def()}}.}

\item{policy_learn}{Policy learner object created by \code{\link[=policy_learn]{policy_learn()}}.}

\item{g_functions}{Fitted g-model objects, see \link{nuisance_functions}.
Preferably, use \code{g_models}.}

\item{g_models}{List of action probability models/g-models for each stage
created by \code{\link[=g_empir]{g_empir()}}, \code{\link[=g_glm]{g_glm()}}, \code{\link[=g_rf]{g_rf()}}, \code{\link[=g_sl]{g_sl()}} or similar functions.
Only used for evaluation if \code{g_functions} is \code{NULL}.
If a single model is provided and \code{g_full_history} is \code{FALSE},
a single g-model is fitted across all stages. If \code{g_full_history} is
\code{TRUE} the model is reused at every stage.}

\item{g_full_history}{If TRUE, the full history is used to fit each g-model.
If FALSE, the state/Markov type history is used to fit each g-model.}

\item{save_g_functions}{If TRUE, the fitted g-functions are saved.}

\item{q_functions}{Fitted Q-model objects, see \link{nuisance_functions}.
Only valid if the Q-functions are fitted using the same policy.
Preferably, use \code{q_models}.}

\item{q_models}{Outcome regression models/Q-models created by
\code{\link[=q_glm]{q_glm()}}, \code{\link[=q_rf]{q_rf()}}, \code{\link[=q_sl]{q_sl()}} or similar functions.
Only used for evaluation if \code{q_functions} is \code{NULL}.
If a single model is provided, the model is reused at every stage.}

\item{q_full_history}{Similar to g_full_history.}

\item{save_q_functions}{Similar to save_g_functions.}

\item{c_functions}{Fitted c-model/censoring probability model objects. Preferably, use \code{c_models}.}

\item{c_models}{List of right-censoring probability models, see \link{c_model}.}

\item{c_full_history}{Similar to g_full_history.}

\item{save_c_functions}{Similar to save_g_functions.}

\item{m_function}{Fitted outcome model object for stage K+1. Preferably, use \code{m_model}.}

\item{m_model}{Outcome model for the utility at stage K+1. Only used if the final utility
contribution is missing/has been right-censored}

\item{m_full_history}{Similar to g_full_history.}

\item{save_m_function}{Similar to save_g_functions.}

\item{target}{Character string. Either "value" or "subgroup". If "value",
the target parameter is the policy value.
If "subgroup", the target parameter
is the subgroup average treatement effect given by the policy, see details.
"subgroup" is only implemented for \code{type = "dr"}
in the single-stage case with a dichotomous action set.}

\item{M}{Number of folds for online estimation/sequential validation excluding the initial training block, see details.}

\item{train_block_size}{Integer. Size of the initial training block only
used for training of the policy and nuisance models, see details.}

\item{name}{Character string.}

\item{min_subgroup_size}{Minimum number of observations in the evaluated subgroup (Only used if target = "subgroup").}
}
\value{
\code{policy_eval_online()} returns an object of inherited class "policy_eval_online", "policy_eval".
The object is a list containing the following elements:
\item{\code{coef}}{Numeric vector. The estimated target parameters:
policy value or subgroup average treatment effect.}
\item{\code{vcov}}{Numeric vector. The estimated squared standard deviation associated with
\code{coef}.}
\item{\code{target}}{Character string. The target parameter ("value" or "subgroup")}
\item{\code{id}}{Character vector. The IDs of the observations.}
\item{\code{name}}{Character vector. Names for the each element in \code{coef}.}
\item{\code{train_sequential_index}}{list of indexes used for training at each step.}
\item{\code{valid_sequential_index}}{list of indexes used for validation at each step.}
}
\description{
\code{policy_eval_online()} is used to estimate
the value of a given fixed policy
or a data adaptive policy (e.g. a policy
learned from the data). \code{policy_eval_online()}
is also used to estimate the subgroup average
treatment effect as defined by the (learned) policy.
The evaluation is based on a online/sequential validation
estimation scheme making the estimation approach valid for a
non-converging policy under no heterogenuous treatment effect
(exceptional law), see details.
}
\details{
\subsection{Setup}{

Each observation has the sequential form
\deqn{O= {B, U_1, X_1, A_1, ..., U_K, X_K, A_K, U_{K+1}},}
for a possibly stochastic number of stages K.
\itemize{
\item \eqn{B} is a vector of baseline covariates.
\item \eqn{U_k} is the reward at stage k
(not influenced by the action \eqn{A_k}).
\item \eqn{X_k} is a vector of state
covariates summarizing the state at stage k.
\item \eqn{A_k} is the categorical action
within the action set \eqn{\mathcal{A}} at stage k.
}
The utility is given by the sum of the rewards, i.e.,
\eqn{U = \sum_{k = 1}^{K+1} U_k}.

A (subgroup) policy is a set of functions
\deqn{d = \{d_1, ..., d_K\},}
where \eqn{d_k} for \eqn{k\in \{1, ..., K\}}
maps a subset or function \eqn{V_1} of \eqn{\{B, X_1, A_1, ..., A_{k-1}, X_k\}} into the
action set (or set of subgroups).

Recursively define the Q-models (\code{q_models}):
\deqn{Q^d_K(h_K, a_K) = \mathbb{E}[U|H_K = h_K, A_K = a_K]}
\deqn{Q^d_k(h_k, a_k) = \mathbb{E}[Q_{k+1}(H_{k+1},
d_{k+1}(V_{k+1}))|H_k = h_k, A_k = a_k].}
If \code{q_full_history = TRUE},
\eqn{H_k = \{B, X_1, A_1, ..., A_{k-1}, X_k\}}, and if
\code{q_full_history = FALSE}, \eqn{H_k = \{B, X_k\}}.

The g-models (\code{g_models}) are defined as
\deqn{g_k(h_k, a_k) = \mathbb{P}(A_k = a_k|H_k = h_k).}
If \code{g_full_history = TRUE},
\eqn{H_k = \{B, X_1, A_1, ..., A_{k-1}, X_k\}}, and if
\code{g_full_history = FALSE}, \eqn{H_k = \{B, X_k\}}.
Furthermore, if \code{g_full_history = FALSE} and \code{g_models} is a
single model, it is assumed that \eqn{g_1(h_1, a_1) = ... = g_K(h_K, a_K)}.
}

\subsection{Target parameters}{

If \code{target = "value"}, \code{policy_eval_online}
returns the estimates of
the value, i.e., the expected potential utility under the policy, (\code{coef}):
\deqn{\mathbb{E}[U^{(d)}]}
If \code{target = "subgroup"}, \code{K = 1}, \eqn{\mathcal{A} = \{0,1\}},
and \eqn{d_1(V_1) \in \{s_1, s_2\} },  \code{policy_eval()}
returns the estimates of the subgroup average
treatment effect (\code{coef}):
\deqn{\mathbb{E}[U^{(1)} - U^{(0)}| d_1(\cdot) = s]\quad s\in \{s_1,s_2\},}
}

\subsection{Online estimation/sequential validation}{

Estimation of the target parameter is based online estimation/sequential
validation using the doubly robust score. The following figure illustrate
online estimation using \code{M = 5} steps and an initial training block of
size \code{train_block_size = }\eqn{l}.

\figure{online-validation.jpg}{Online estimaiton scheme}

Step 1:

The \eqn{n} observations are randomly ordered. In step 1,
the first \eqn{\{1,...,l\}} observations, highlighted in teal/blue, are used to fit the
Q-models, g-models, the policy (if using the \code{policy_learn} argument), and other required models.
We denote the collection of these fitted models as \eqn{P}.
The remaining observations are split into \code{M} blocks of size \eqn{m = (n-l)/M}, which
we for simplicity assume to be a whole number. In step 1, the target
parameter is estimated using the associated doubly robust score \eqn{Z(P)}
evaluated on the first validation fold
highlighted in pink \eqn{\{l+1,...,l+m\}}:

\deqn{
 \frac{\sum_{i = l+1}^{l+m}  {\widehat \sigma_{i}^{-1}}
 Z(\widehat P_i)(O_i)}
 {\sum_{i = l+1}^{l+m} \widehat \sigma_{i}^{-1}},
}
where \eqn{\widehat P_i} for \eqn{i \in \{l+1,...,l+m\}}
refer to the fitted models trained on \eqn{\{1,...,l\}}, and \eqn{\widehat \sigma_i}
is the insample estimate for the standard deviation based on the training observations \eqn{\{1,...,l\}}.
We will later give an exact expression for \eqn{\widehat \sigma_i} for each target parameter.
Note that \eqn{\widehat \sigma_i} is constant for \eqn{i \in \{l+1,...,l+m\}}, but it will be
convenient to keep the same index for \eqn{\widehat \sigma}.

Step 2 to M:

In step 2, observations with index \eqn{\{1,...,l+m\}} are used to fit the model collection \eqn{P},
as well as the insample estimate for the standard deviation. For \eqn{i \in \{l+m+1,...,l+2m\}} these are
denoted as \eqn{\widehat P_i, \widehat \sigma_i}.
This sequential model fitting is repeated for all \code{M}
steps and the updated online estimator is given by

\deqn{
 \frac{\sum_{i = l+1}^{n}  {\widehat \sigma_{i}^{-1}}
 Z(\widehat P_i)(O_i)}
 {\sum_{i = l+1}^n \widehat \sigma_{i}^{-1}},
}

with an associated standard error estimate given by

\deqn{\frac{\left(\frac{1}{n-l}\sum_{i = l+1}^n \widehat \sigma_{i}^{-1}\right)^{-1}}{\sqrt{n-l}}.}
}

\subsection{Doubly robust scores}{

\code{target = "value"}:

For a policy value target the doubly robust score is given by
\deqn{
Z(d, g, Q^d)(O) = Q^d_1(H_1 , d_1(V_1)) +
\sum_{r = 1}^K \prod_{j = 1}^{r}
\frac{I\{A_j = d_j(\cdot)\}}{g_{j}(H_j, A_j)}
\{Q_{r+1}^d(H_{r+1} , d_{r+1}(V_1)) - Q_{r}^d(H_r , d_r(V_1))\}.
}
The influence function(/curve) for the associated onestep etimator is
\deqn{Z(d, g, Q^d)(O) - \mathbb{E}[Z(d,g, Q^d)(O)],}
which is used to estimate the insample stadard deviation. For example,
in step 2, i.e., for \eqn{i \in \{l+m+1,...,l+2m\}}
\deqn{
 \widehat \sigma_i^2 = \frac{1}{l+m}\sum_{j=1}^{l+m} \left(Z(\widehat d_i,\widehat Q_i,\widehat{g}_i)(O_j) - \frac{1}{l+m}\sum_{r=1}^{l+m} Z(\widehat d_i,\widehat Q_i,\widehat{g}_i)(O_r) \right)^2
}

\code{target = "subgroup"}:

For a subgroup average treatment effect target,
where \code{K = 1} (single-stage),
\eqn{\mathcal{A} = \{0,1\}} (binary treatment), and
\eqn{d_1(V_1) \in \{s_1, s_2\}} (dichotomous subgroup policy) the
doubly robust score is given by

\deqn{
Z(d,g,Q, D) = \frac{I\{d_1(\cdot) = s\}}{D}
\Big\{Z_1(1,g,Q)(O) - Z_1(0,g,Q)(O) \Big\}.}
\deqn{
Z_1(a, g, Q)(O) = Q_1(H_1 , a) +
\frac{I\{A = a\}}{g_1(H_1, a)}
\{U - Q_{1}(H_1 , a)\},
}
where \eqn{D} is \eqn{\mathbb{P}(d_1(V_1) = s)}.

The associated onestep/estimating equation estimator has influence function
\deqn{\frac{ I\{d_1(\cdot) = s\}}{D}
\Big\{Z_1(1,g,Q)(O) - Z_1(0,g,Q)(O) - E[Z_1(1,g,Q)(O)
- Z_1(0,g,Q)(O) | d_1(\cdot) = s]\Big\},}
which is used to estimate the standard deviation \eqn{\widehat \sigma}.
}
}
\references{
Luedtke, Alexander R, and Mark J van der Laan.
“STATISTICAL INFERENCE FOR THE MEAN OUTCOME
UNDER A POSSIBLY NON-UNIQUE OPTIMAL TREATMENT STRATEGY.”
Annals of statistics vol. 44,2 (2016): 713-742.
\doi{10.1214/15-AOS1384}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_eval_functions.R
\name{plot.policy_eval}
\alias{plot.policy_eval}
\title{Plot histogram of the influence curve for a \code{policy_eval} object}
\usage{
\method{plot}{policy_eval}(x, ...)
}
\arguments{
\item{x}{Object of class \link{policy_eval}}

\item{...}{Additional arguments}
}
\description{
Plot histogram of the influence curve for a \code{policy_eval} object
}
\examples{
d <- sim_two_stage(2e3, seed=1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = "BB",
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))

pe <- policy_eval(pd,
                  policy_learn = policy_learn())

plot(pe)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g_function.R
\name{fit_g_functions}
\alias{fit_g_functions}
\title{Fit g-functions}
\usage{
fit_g_functions(policy_data, g_models, full_history = FALSE)
}
\arguments{
\item{policy_data}{Policy data object created by \code{\link[=policy_data]{policy_data()}}.}

\item{g_models}{List of action probability models/g-models for each stage
created by \code{\link[=g_empir]{g_empir()}}, \code{\link[=g_glm]{g_glm()}}, \code{\link[=g_rf]{g_rf()}}, \code{\link[=g_sl]{g_sl()}} or similar functions.}

\item{full_history}{If TRUE, the full history is used to fit each g-model.
If FALSE, the single stage/"Markov type" history is used to fit each g-model.}
}
\description{
\code{fit_g_functions} is used to fit a list of g-models.
}
\examples{
library("polle")
### Simulating two-stage policy data
d <- sim_two_stage(2e3, seed=1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

# fitting a single g-model across all stages:
g_functions <- fit_g_functions(policy_data = pd,
                               g_models = g_glm(),
                               full_history = FALSE)
g_functions

# fitting a g-model for each stage:
g_functions <- fit_g_functions(policy_data = pd,
                               g_models = list(g_glm(), g_glm()),
                               full_history = TRUE)
g_functions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{get_history_names}
\alias{get_history_names}
\title{Get history variable names}
\usage{
get_history_names(object, stage)
}
\arguments{
\item{object}{Policy data object created by \code{\link[=policy_data]{policy_data()}}.}

\item{stage}{Stage number. If NULL, the state/Markov-type history variable
names are returned.}
}
\value{
Character vector.
}
\description{
\code{get_history_names()} returns the state covariate names of the history data
table for a given stage. The function is useful when specifying
the design matrix for \link{g_model} and \link{q_model} objects.
}
\examples{
library("polle")
### Multiple stages:
d3 <- sim_multi_stage(5e2, seed = 1)
pd3 <- policy_data(data = d3$stage_data,
                   baseline_data = d3$baseline_data,
                   type = "long",
                   id = "id",
                   stage = "stage",
                   event = "event",
                   action = "A",
                   utility = "U")
pd3
# state/Markov type history variable names (H):
get_history_names(pd3)
# full history variable names (H_k) at stage 2:
get_history_names(pd3, stage = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data.R, R/policy_data_functions.R
\name{policy_data}
\alias{policy_data}
\alias{print.policy_data}
\alias{summary.policy_data}
\title{Create Policy Data Object}
\usage{
policy_data(
  data,
  baseline_data,
  type = "wide",
  action,
  covariates,
  utility,
  baseline = NULL,
  deterministic_rewards = NULL,
  id = NULL,
  stage = NULL,
  event = NULL,
  time = NULL,
  time2 = NULL,
  action_set = NULL,
  verbose = FALSE
)

\method{print}{policy_data}(x, digits = 2, ...)

\method{summary}{policy_data}(object, probs = seq(0, 1, 0.25), ...)
}
\arguments{
\item{data}{\link{data.frame} or \link[data.table:data.table]{data.table::data.table}; see Examples.}

\item{baseline_data}{\link{data.frame} or \link[data.table:data.table]{data.table::data.table}; see Examples.}

\item{type}{Character string. If "wide", \code{data} is considered to be on
wide format. If "long", \code{data} is considered to be on long format; see
Examples.}

\item{action}{Action variable name(s). Character vector or character string.
\itemize{ \item A vector is valid for wide data. The length of the vector
determines the number of stages (K). \item A string is valid for single
stage wide data or long data. }}

\item{covariates}{Stage specific covariate name(s). Character vector or named
list of character vectors. \itemize{ \item A vector is valid for single
stage wide data or long data. \item A named list is valid for multiple
stages wide data. Each element must be a character vector with length K.
Each vector can contain NA elements, if a covariate is not available for
the given stage(s). }}

\item{utility}{Utility/Reward variable name(s). Character string or vector.
\itemize{ \item A string is valid for long data and wide data with a single
final utility. \item A vector is valid for wide data with incremental
rewards. Must have length K+1; see Examples. }}

\item{baseline}{Baseline covariate name(s). Character vector.}

\item{deterministic_rewards}{Deterministic reward variable name(s). Named
list of character vectors of length K. The name of each element must be on
the form "U_Aa" where "a" corresponds to an action in the action set.}

\item{id}{ID variable name. Character string.}

\item{stage}{Stage number variable name.}

\item{event}{Event indicator name.}

\item{time}{Character string}

\item{time2}{Character string}

\item{action_set}{Character string. Action set across all stages.}

\item{verbose}{Logical. If TRUE, formatting comments are printed to the
console.}

\item{x}{Object to be printed.}

\item{digits}{Minimum number of digits to be printed.}

\item{...}{Additional arguments passed to print.}

\item{object}{Object of class \link{policy_data}}

\item{probs}{numeric vector (probabilities)}
}
\value{
\code{policy_data()} returns an object of class "policy_data". The
object is a list containing the following elements:
\item{\code{stage_data}}{\link[data.table:data.table]{data.table::data.table} containing the id, stage
number, event indicator, action (\eqn{A_k}), state covariates (\eqn{X_k}),
reward (\eqn{U_k}), and the deterministic rewards.}
\item{\code{baseline_data}}{\link[data.table:data.table]{data.table::data.table} containing the id and
baseline covariates (\eqn{B}).} \item{\code{colnames}}{List containing the
state covariate names, baseline covariate names, and the deterministic
reward variable names.} \item{\code{action_set}}{Sorted character vector
describing the action set, i.e., the possible actions at all stages.}
\item{\code{stage_action_sets}}{List of sorted character vectors describing
the observed actions at each stage.} \item{\code{dim}}{List containing the
number of observations (n) and the number of stages (K).}
}
\description{
\code{policy_data()} creates a policy data object which
is used as input to \code{\link[=policy_eval]{policy_eval()}} and \code{\link[=policy_learn]{policy_learn()}} for policy
evaluation and data adaptive policy learning.
}
\details{
Each observation has the sequential form \deqn{O= {B, U_1, X_1, A_1,
  ..., U_K, X_K, A_K, U_{K+1}},} for a possibly stochastic number of stages
K. \itemize{ \item \eqn{B} is a vector of baseline covariates. \item
\eqn{U_k} is the reward at stage k (not influenced by the action
\eqn{A_k}). \item \eqn{X_k} is a vector of state covariates summarizing the
state at stage k. \item \eqn{A_k} is the categorical action at stage k. }
The utility is given by the sum of the rewards, i.e., \eqn{U = \sum_{k =
  1}^{K+1} U_k}.
}
\section{S3 generics}{
 The following S3 generic functions are available for an
object of class \code{policy_data}: \describe{ \item{\code{\link[=partial]{partial()}}}{ Trim the
maximum number of stages in a \code{policy_data} object.}
\item{\code{\link[=subset_id]{subset_id()}}}{ Subset a a \code{policy_data} object on ID.}
\item{\code{\link[=get_history]{get_history()}}}{ Summarize the history and action at a given stage.}
\item{\code{\link[=get_history_names]{get_history_names()}}}{ Get history variable names.}
\item{\code{\link[=get_actions]{get_actions()}}}{ Get the action at every stage.}
\item{\code{\link[=get_utility]{get_utility()}}}{Get the utility.} \item{\code{\link[=plot]{plot()}}}{Plot method.} }
}

\examples{
library("polle")
### Single stage: Wide data
d1 <- sim_single_stage(n = 5e2, seed=1)
head(d1, 5)
# constructing policy_data object:
pd1 <- policy_data(d1,
                   action="A",
                   covariates=c("Z", "B", "L"),
                   utility="U")
pd1
# associated S3 methods:
methods(class = "policy_data")
head(get_actions(pd1), 5)
head(get_utility(pd1), 5)
head(get_history(pd1)$H, 5)

### Two stage: Wide data
d2 <- sim_two_stage(5e2, seed=1)
head(d2, 5)
# constructing policy_data object:
pd2 <- policy_data(d2,
                  action = c("A_1", "A_2"),
                  baseline = c("B"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd2
head(get_history(pd2, stage = 2)$H, 5) # state/Markov type history and action, (H_k,A_k).
head(get_history(pd2, stage = 2, full_history = TRUE)$H, 5) # Full history and action, (H_k,A_k).

### Multiple stages: Long data
d3 <- sim_multi_stage(5e2, seed = 1)
head(d3$stage_data, 10)
# constructing policy_data object:
pd3 <- policy_data(data = d3$stage_data,
                   baseline_data = d3$baseline_data,
                   type = "long",
                   id = "id",
                   stage = "stage",
                   event = "event",
                   action = "A",
                   utility = "U")
pd3
head(get_history(pd3, stage = 3)$H, 5) # state/Markov type history and action, (H_k,A_k).
head(get_history(pd3, stage = 2, full_history = TRUE)$H, 5) # Full history and action, (H_k,A_k).
}
\seealso{
\code{\link[=policy_eval]{policy_eval()}}, \code{\link[=policy_learn]{policy_learn()}}, \code{\link[=copy_policy_data]{copy_policy_data()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blip.R, R/drql.R, R/policy_object_functions.R,
%   R/ptl.R, R/rql.R
\name{get_policy_functions.blip}
\alias{get_policy_functions.blip}
\alias{get_policy_functions.drql}
\alias{get_policy_functions}
\alias{get_policy_functions.ptl}
\alias{get_policy_functions.ql}
\title{Get Policy Functions}
\usage{
\method{get_policy_functions}{blip}(
  object,
  stage,
  threshold = NULL,
  include_g_values = FALSE,
  ...
)

\method{get_policy_functions}{drql}(
  object,
  stage,
  threshold = NULL,
  include_g_values = FALSE,
  ...
)

get_policy_functions(object, stage, threshold, ...)

\method{get_policy_functions}{ptl}(object, stage, threshold = NULL, ...)

\method{get_policy_functions}{ql}(
  object,
  stage,
  threshold = NULL,
  include_g_values = FALSE,
  ...
)
}
\arguments{
\item{object}{Object of class "policy_object" or "policy_eval",
see \link{policy_learn} and \link{policy_eval}.}

\item{stage}{Integer. Stage number.}

\item{threshold}{Numeric, threshold for not
choosing the reference action at stage 1.}

\item{include_g_values}{If TRUE, the g-values are included as an attribute.}

\item{...}{Additional arguments.}
}
\value{
Functions with arguments:
\describe{
\item{\code{H}}{\link[data.table:data.table]{data.table::data.table} containing the variables needed to evaluate the policy (and g-function).}
}
}
\description{
\code{get_policy_functions()} returns a function defining the policy at
the given stage. \code{get_policy_functions()} is useful when implementing
the learned policy.
}
\examples{
library("polle")
### Two stages:
d <- sim_two_stage(5e2, seed=1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = "BB",
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

### Realistic V-restricted Policy Tree Learning
# specifying the learner:
pl <- policy_learn(type = "ptl",
                   control = control_ptl(policy_vars = list(c("C_1", "BB"),
                                                            c("L_1", "BB"))),
                   full_history = TRUE,
                   alpha = 0.05)

# evaluating the learner:
pe <- policy_eval(policy_data = pd,
                  policy_learn = pl,
                  q_models = q_glm(),
                  g_models = g_glm())

# getting the policy function at stage 2:
pf2 <- get_policy_functions(pe, stage = 2)
args(pf2)

# applying the policy function to new data:
set.seed(1)
L_1 <- rnorm(n = 10)
new_H <- data.frame(C = rnorm(n = 10),
                    L = L_1,
                    L_1 = L_1,
                    BB = "group1")
d2 <- pf2(H = new_H)
head(d2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rwl.R
\name{control_rwl}
\alias{control_rwl}
\title{Control arguments for Residual Weighted Learning}
\usage{
control_rwl(
  moPropen,
  moMain,
  regime,
  fSet = NULL,
  lambdas = 2,
  cvFolds = 0L,
  kernel = "linear",
  kparam = NULL,
  responseType = "continuous",
  verbose = 2L
)
}
\arguments{
\item{moPropen}{Propensity model of class "ModelObj", see \link[modelObj:modelObj]{modelObj::modelObj}.}

\item{moMain}{Main effects outcome model of class "ModelObj".}

\item{regime}{An object of class \link{formula} specifying the design of the policy/regime.}

\item{fSet}{A function or NULL defining subset structure.}

\item{lambdas}{Numeric or numeric vector. Penalty parameter.}

\item{cvFolds}{Integer. Number of folds for cross-validation of the parameters.
\code{"logit"}, \code{"exp"}, \code{"hinge"}, \code{"sqhinge"}, \code{"huber"}.}

\item{kernel}{The options are \code{"linear"}, \code{"poly"}, \code{"radial"}.}

\item{kparam}{Numeric. Kernel parameter}

\item{responseType}{Character string. Options are \code{"continuous"},
\code{"binary"}, \code{"count"}.}

\item{verbose}{Integer.}
}
\value{
list of (default) control arguments.
}
\description{
\code{control_rwl} sets the default control arguments
for residual learning , \code{type = "rwl"}.
The arguments are passed directly to \code{\link[DynTxRegime:rwl]{DynTxRegime::rwl()}} if not
specified otherwise.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nuisance_functions.R
\docType{class}
\name{nuisance_functions}
\alias{nuisance_functions}
\title{Nuisance Functions}
\description{
The fitted g-functions and Q-functions are stored in an object
of class "nuisance_functions". The object is a list with a fitted model
object for every stage. Information on whether the full history or the
state/Markov-type history is stored as an attribute ("full_history").
}
\section{S3 generics}{

The following S3 generic functions are available for an object of class
\code{nuisance_functions}:
\describe{
\item{\code{predict}}{ Predict the values of the g- or Q-functions based
on a \link{policy_data} object.}
}
}

\examples{
### Two stages:
d <- sim_two_stage(5e2, seed=1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

# evaluating the static policy a=1:
pe <- policy_eval(policy_data = pd,
                  policy = policy_def(1, reuse = TRUE),
                  g_models = g_glm(),
                  q_models = q_glm())

# getting the fitted g-functions:
(g_functions <- get_g_functions(pe))

# getting the fitted Q-functions:
(q_functions <- get_q_functions(pe))

# getting the fitted values:
head(predict(g_functions, pd))
head(predict(q_functions, pd))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{get_id_stage}
\alias{get_id_stage}
\title{Get IDs and Stages}
\usage{
get_id_stage(object, event_set = c(0))
}
\arguments{
\item{object}{Object of class \link{policy_data} or \link{history}.}

\item{event_set}{Integer vector. Subset of c(0,1,2).}
}
\value{
\link[data.table:data.table]{data.table::data.table} with keys id and stage.
}
\description{
\code{get_id_stage} returns the ID and stage number differnt types of events.
}
\examples{
### Two stages:
d <- sim_two_stage(5e2, seed=1)
# constructing policy_data object:
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = c("B"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

# getting the IDs and stages:
head(get_id_stage(pd))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation.R
\name{sim_single_stage_multi_actions}
\alias{sim_single_stage_multi_actions}
\title{Simulate Single-Stage Multi-Action Data}
\usage{
sim_single_stage_multi_actions(n = 1000, seed = NULL)
}
\arguments{
\item{n}{Number of observations.}

\item{seed}{Integer.}
}
\value{
data.frame with n rows and columns z, x, a, and u.
}
\description{
Simulate Single-Stage Multi-Action Data
}
\details{
\code{sim_single_stage_multi_actions} samples \code{n} iid observation
\eqn{O = (z, x, a, u)} with the following distribution:
\deqn{
z, x \sim Uniform([0,1])\\
\tilde a \sim \mathcal{N}(0,1)\\
a \mid \tilde a \sim
\begin{cases}
0 \quad if \quad \tilde a < -1\\
1 \quad if \quad \tilde a -1 \leq a < 0.5\\
2 \quad otherwise
\end{cases}\\
u \mid z, x \sim \mathcal{N}(x + z + I\{a=2\}(x-0.5) + I\{a=1\}(x^2 + z -0.5), 1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_def.R
\docType{class}
\name{policy}
\alias{policy}
\title{Policy-class}
\value{
\link[data.table:data.table]{data.table::data.table} with keys \code{id} and \code{stage} and
action variable \code{d}.
}
\description{
A function of inherited class "policy" takes a policy
data object as input and returns the policy actions for every observation
for every (observed) stage.
}
\details{
A policy can either be defined directly by the user using
\link{policy_def} or a policy can be fitted using \link{policy_learn}
(or \link{policy_eval}). \link{policy_learn} returns a \link{policy_object} from which
the policy can be extracted using \link{get_policy}.
}
\section{S3 generics}{

The following S3 generic functions are available for an object of class
\code{policy}:
\describe{
\item{\code{print}}{Baisc print function}
}
}

\examples{
### Two stages:
d <- sim_two_stage(5e2, seed=1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))

# defining a dynamic policy:
p <- policy_def(
  function(L) (L>0)*1,
  reuse = TRUE
)
p
head(p(pd), 5)

# V-restricted (Doubly Robust) Q-learning:
# specifying the learner:
pl <- policy_learn(type = "drql",
                   control = control_drql(qv_models = q_glm(formula = ~ C)))

# fitting the policy (object):
po <- pl(policy_data = pd,
         q_models = q_glm(),
         g_models = g_glm())

p <- get_policy(po)
p

head(p(pd))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_object_functions.R
\name{get_policy}
\alias{get_policy}
\title{Get Policy}
\usage{
get_policy(object, threshold = NULL)
}
\arguments{
\item{object}{Object of class \link{policy_object} or \link{policy_eval}.}

\item{threshold}{Numeric vector.
Thresholds for the first stage policy function.}
}
\value{
function of class \link{policy}.
}
\description{
\code{get_policy} extracts the policy from a policy object
or a policy evaluation object The policy is a function which take a
policy data object as input and returns the policy actions.
}
\examples{
library("polle")
### Two stages:
d <- sim_two_stage(5e2, seed = 1)
pd <- policy_data(d,
  action = c("A_1", "A_2"),
  baseline = c("BB"),
  covariates = list(
    L = c("L_1", "L_2"),
    C = c("C_1", "C_2")
  ),
  utility = c("U_1", "U_2", "U_3")
)
pd

### V-restricted (Doubly Robust) Q-learning

# specifying the learner:
pl <- policy_learn(
  type = "drql",
  control = control_drql(qv_models = q_glm(formula = ~C))
)

# fitting the policy (object):
po <- pl(
  policy_data = pd,
  q_models = q_glm(),
  g_models = g_glm()
)

# getting and applying the policy:
head(get_policy(po)(pd))

# the policy learner can also be evaluated directly:
pe <- policy_eval(
  policy_data = pd,
  policy_learn = pl,
  q_models = q_glm(),
  g_models = g_glm()
)

# getting and applying the policy again:
head(get_policy(pe)(pd))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation.R
\name{sim_multi_stage}
\alias{sim_multi_stage}
\title{Simulate Multi-Stage Data}
\usage{
sim_multi_stage(
  n,
  par = list(tau = 10, gamma = c(0, -0.2, 0.3), alpha = c(0, 0.5, 0.2, -0.5, 0.4), beta =
    c(3, -0.5, -0.5), psi = 1, xi = 0.3),
  a = function(t, x, beta, ...) {
     prob <- lava::expit(beta[1] + (beta[2] * t^2) +
    (beta[3] * x))
     stats::rbinom(n = 1, size = 1, prob = prob)
 },
  seed = NULL
)
}
\arguments{
\item{n}{Number of observations.}

\item{par}{Named list with distributional parameters.
\itemize{
\item{} \code{tau}: \eqn{\tau}
\item{} \code{gamma}: \eqn{\gamma}
\item{} \code{alpha}: \eqn{\alpha}
\item{} \code{beta}: \eqn{\beta}
\item{} \code{psi}: \eqn{\psi}
\item{} \code{xi}: \eqn{\xi}
}}

\item{a}{Function used to specify the action/treatment at every stage.}

\item{seed}{Integer.}
}
\value{
list with elements \code{stage_data} (\link[data.table:data.table]{data.table::data.table}) and
\code{baseline_data} (\link[data.table:data.table]{data.table::data.table}).
}
\description{
Simulate Multi-Stage Data
}
\details{
\code{sim_multi_stage} samples \code{n} iid observation
\eqn{O} with the following distribution:
\deqn{
W \sim \mathcal{N}(0, 1)\\
B \sim Ber(\xi)
}
For \eqn{k\geq 1} let
\deqn{
(T_k - T_{k-1})| X_{k-1}, A_{k-1}, W \sim
\begin{cases}
Exp\Big\{\exp\left(\gamma^T [1, X_{k-1}, W] \right) \Big\} + \psi \quad A_{k-1} = 1\\
\infty \quad A_{k-1} = 0
\end{cases}\\
X_{k}\mid T_k, X_{k-1}, B \sim
\begin{cases}
\mathcal{N}\left\{ \alpha^T [1, T_k, T^2_k, X_{k-1}, B], 1\right\} \quad T_k < \infty \\
0 \quad T_k = \infty
\end{cases}\\
A_k \mid X_k, T_k \sim
\begin{cases}
Ber\left\{ expit\left(\beta^T[1, T_{k}^2, X_k] \right)\right\} \quad T_k < \infty\\
0 \quad T_k = \infty,
\end{cases}
}
Note that \eqn{\psi} is the minimum increment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation.R
\name{sim_two_stage_multi_actions}
\alias{sim_two_stage_multi_actions}
\title{Simulate Two-Stage Multi-Action Data}
\usage{
sim_two_stage_multi_actions(
  n = 1000,
  par = list(gamma = 0.5, beta = 1, prob = c(0.2, 0.4, 0.4)),
  seed = NULL,
  action_model_1 = function(C_1, beta, ...) stats::rbinom(n = NROW(C_1), size = 1, prob =
    lava::expit(beta * C_1))
)
}
\arguments{
\item{n}{Number of observations.}

\item{par}{Named vector with distributional parameters.
\itemize{
\item{} \code{gamma}: \eqn{\gamma}
\item{} \code{beta}: \eqn{\beta}
\item{} \code{prob}: \eqn{p}
}}

\item{seed}{Integer.}

\item{action_model_1}{Function used to specify the dichotomous
action/treatment at stage 1.}
}
\value{
\link[data.table:data.table]{data.table::data.table} with n rows and columns B, BB, L_1, C_1, A_1, L_2, C_2,
A_2, L_3, U_1, U_2, U_3.
}
\description{
Simulate Two-Stage Multi-Action Data
}
\details{
\code{sim_two_stage_multi_actions} samples \code{n} iid observation
\eqn{O} with the following distribution:
\eqn{BB} is a random categorical variable with levels \code{group1},
\code{group2}, and \code{group3}. Furthermore,
\deqn{
B \sim \mathcal{N}(0,1)\\
L_{1} \sim \mathcal{N}(0, 1)\\
C_{1} \mid L_{1} \sim \mathcal{N}(L_1, 1)\\
P(A_1='yes'\mid C_1) =  expit(\beta C_1)\\
P(A_1='no'\mid C_1) = 1 - P(A_1='yes' \mid C_1)\\
L_{2} \sim \mathcal{N} (0, 1)\\
C_{2} \mid A_1, L_1 \sim \mathcal{N}(\gamma L_1 + A_1, 1)\\
P(A_2='yes') = p_1\\
P(A_2='no') = p_2\\
P(A_2='default') = p_3\\
L_{3} \sim \mathcal{N} (0, 1)
}
The rewards are calculated as
\deqn{
U_1 = L_1\\
U_2 = A_1\cdot C_1 + L_2 \\
U_3 = A_2\cdot C_2 + L_3.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blip.R
\name{control_blip}
\alias{control_blip}
\title{Control arguments for doubly robust blip-learning}
\usage{
control_blip(blip_models = q_glm(~.), quantile_prob_threshold = NULL)
}
\arguments{
\item{blip_models}{Single element or list of V-restricted blip-models created
by \code{\link[=q_glm]{q_glm()}}, \code{\link[=q_rf]{q_rf()}}, \code{\link[=q_sl]{q_sl()}} or similar functions.}

\item{quantile_prob_threshold}{Numeric vector. Quantile probabilities
for adaptively setting the threshold for the conditional average treatment effect.}
}
\value{
list of (default) control arguments.
}
\description{
\code{control_blip} sets the default control arguments
for doubly robust blip-learning, \code{type = "blip"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{partial}
\alias{partial}
\title{Trim Number of Stages}
\usage{
partial(object, K)
}
\arguments{
\item{object}{Object of class \link{policy_data}.}

\item{K}{Maximum number of stages.}
}
\value{
Object of class \link{policy_data}.
}
\description{
\code{partial} creates a partial policy data object by trimming
the maximum number of stages in the policy data object to a fixed
given number.
}
\examples{
library("polle")
### Multiple stage case
d <- sim_multi_stage(5e2, seed = 1)
# constructing policy_data object:
pd <- policy_data(data = d$stage_data,
                   baseline_data = d$baseline_data,
                   type = "long",
                   id = "id",
                   stage = "stage",
                   event = "event",
                   action = "A",
                   utility = "U")
pd
# Creating a partial policy data object with 3 stages
pd3 <- partial(pd, K = 3)
pd3
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_def.R
\name{policy_def}
\alias{policy_def}
\title{Define Policy}
\usage{
policy_def(
  policy_functions,
  full_history = FALSE,
  reuse = FALSE,
  name = NULL,
  natural_action = FALSE,
  stage_number = FALSE
)
}
\arguments{
\item{policy_functions}{A single function/character string or a list of
functions/character strings. The list must have the same length as the number
of stages.}

\item{full_history}{If TRUE, the full history at each stage is used as
input to the policy functions.}

\item{reuse}{If TRUE, the policy function is reused at every stage.}

\item{name}{Character string.}

\item{natural_action}{Logical. If TRUE, the natural/observed action
variable will be passeed to the policy functions.}

\item{stage_number}{Logical. If TRUE, the stage number is passed to the
policy functions.}
}
\value{
Function of class \code{"policy"}. The function takes a
\link{policy_data} object as input and returns a \link[data.table:data.table]{data.table::data.table}
with keys \code{id} and \code{stage} and action variable \code{d}.
}
\description{
\code{policy_def} returns a function of class \link{policy}.
The function input is a \link{policy_data} object and it returns a \link[data.table:data.table]{data.table::data.table}
with keys \code{id} and \code{stage} and action variable \code{d}.
}
\examples{
library("polle")
### Single stage"
d1 <- sim_single_stage(5e2, seed=1)
pd1 <- policy_data(d1, action="A", covariates=list("Z", "B", "L"), utility="U")
pd1

# defining a static policy (A=1):
p1_static <- policy_def(1)

# applying the policy:
p1_static(pd1)

# defining a dynamic policy:
p1_dynamic <- policy_def(
  function(Z, L) ((3*Z + 1*L -2.5)>0)*1
)
p1_dynamic(pd1)

### Two stages:
d2 <- sim_two_stage(5e2, seed = 1)
pd2 <- policy_data(d2,
                  action = c("A_1", "A_2"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))

# defining a static policy (A=0):
p2_static <- policy_def(0,
                        reuse = TRUE)
p2_static(pd2)

# defining a reused dynamic policy:
p2_dynamic_reuse <- policy_def(
  function(L) (L > 0)*1,
  reuse = TRUE
)
p2_dynamic_reuse(pd2)

# defining a dynamic policy for each stage based on the full history:
# available variable names at each stage:
get_history_names(pd2, stage = 1)
get_history_names(pd2, stage = 2)

p2_dynamic <- policy_def(
  policy_functions = list(
    function(L_1) (L_1 > 0)*1,
    function(L_1, L_2) (L_1 + L_2 > 0)*1
  ),
  full_history = TRUE
)
p2_dynamic(pd2)
}
\seealso{
\code{\link[=get_history_names]{get_history_names()}}, \code{\link[=get_history]{get_history()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{get_K}
\alias{get_K}
\title{Get Maximal Stages}
\usage{
get_K(object)
}
\arguments{
\item{object}{Object of class \link{policy_data}.}
}
\value{
Integer.
}
\description{
\code{get_K} returns the maximal number of stages for the observations in
the policy data object.
}
\examples{
d <- sim_multi_stage(5e2, seed = 1)
pd <- policy_data(data = d$stage_data,
                   baseline_data = d$baseline_data,
                   type = "long",
                   id = "id",
                   stage = "stage",
                   event = "event",
                   action = "A",
                   utility = "U")
pd
# getting the maximal number of stages:
get_K(pd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_eval.R, R/policy_eval_functions.R,
%   R/policy_eval_online.R
\name{policy_eval}
\alias{policy_eval}
\alias{coef.policy_eval}
\alias{IC.policy_eval}
\alias{vcov.policy_eval}
\alias{print.policy_eval}
\alias{summary.policy_eval}
\alias{estimate.policy_eval}
\alias{merge.policy_eval}
\alias{+.policy_eval}
\alias{vcov.policy_eval_online}
\title{Policy Evaluation}
\usage{
policy_eval(
  policy_data,
  policy = NULL,
  policy_learn = NULL,
  g_functions = NULL,
  g_models = g_glm(),
  g_full_history = FALSE,
  save_g_functions = TRUE,
  q_functions = NULL,
  q_models = q_glm(),
  q_full_history = FALSE,
  save_q_functions = TRUE,
  c_functions = NULL,
  c_models = NULL,
  c_full_history = FALSE,
  save_c_functions = TRUE,
  m_function = NULL,
  m_model = NULL,
  m_full_history = FALSE,
  save_m_function = TRUE,
  target = "value",
  type = "dr",
  cross_fit_type = "pooled",
  variance_type = "pooled",
  M = 1,
  nrep = 1,
  min_subgroup_size = 1,
  future_args = list(future.seed = TRUE),
  name = NULL
)

\method{coef}{policy_eval}(object, ...)

\method{IC}{policy_eval}(x, ...)

\method{vcov}{policy_eval}(object, ...)

\method{print}{policy_eval}(
  x,
  digits = 4L,
  width = 35L,
  std.error = TRUE,
  level = 0.95,
  p.value = TRUE,
  ...
)

\method{summary}{policy_eval}(object, ...)

\method{estimate}{policy_eval}(
  x,
  labels = get_element(x, "name", check_name = FALSE),
  level = 0.95,
  ...
)

\method{merge}{policy_eval}(x, y, ..., paired = TRUE)

\method{+}{policy_eval}(x, ...)

\method{vcov}{policy_eval_online}(object, ...)
}
\arguments{
\item{policy_data}{Policy data object created by \code{\link[=policy_data]{policy_data()}}.}

\item{policy}{Policy object created by \code{\link[=policy_def]{policy_def()}}.}

\item{policy_learn}{Policy learner object created by \code{\link[=policy_learn]{policy_learn()}}.}

\item{g_functions}{Fitted g-model objects, see \link{nuisance_functions}.
Preferably, use \code{g_models}.}

\item{g_models}{List of action probability models/g-models for each stage
created by \code{\link[=g_empir]{g_empir()}}, \code{\link[=g_glm]{g_glm()}}, \code{\link[=g_rf]{g_rf()}}, \code{\link[=g_sl]{g_sl()}} or similar functions.
Only used for evaluation if \code{g_functions} is \code{NULL}.
If a single model is provided and \code{g_full_history} is \code{FALSE},
a single g-model is fitted across all stages. If \code{g_full_history} is
\code{TRUE} the model is reused at every stage.}

\item{g_full_history}{If TRUE, the full history is used to fit each g-model.
If FALSE, the state/Markov type history is used to fit each g-model.}

\item{save_g_functions}{If TRUE, the fitted g-functions are saved.}

\item{q_functions}{Fitted Q-model objects, see \link{nuisance_functions}.
Only valid if the Q-functions are fitted using the same policy.
Preferably, use \code{q_models}.}

\item{q_models}{Outcome regression models/Q-models created by
\code{\link[=q_glm]{q_glm()}}, \code{\link[=q_rf]{q_rf()}}, \code{\link[=q_sl]{q_sl()}} or similar functions.
Only used for evaluation if \code{q_functions} is \code{NULL}.
If a single model is provided, the model is reused at every stage.}

\item{q_full_history}{Similar to g_full_history.}

\item{save_q_functions}{Similar to save_g_functions.}

\item{c_functions}{Fitted c-model/censoring probability model objects. Preferably, use \code{c_models}.}

\item{c_models}{List of right-censoring probability models, see \link{c_model}.}

\item{c_full_history}{Similar to g_full_history.}

\item{save_c_functions}{Similar to save_g_functions.}

\item{m_function}{Fitted outcome model object for stage K+1. Preferably, use \code{m_model}.}

\item{m_model}{Outcome model for the utility at stage K+1. Only used if the final utility
contribution is missing/has been right-censored}

\item{m_full_history}{Similar to g_full_history.}

\item{save_m_function}{Similar to save_g_functions.}

\item{target}{Character string. Either "value" or "subgroup". If "value",
the target parameter is the policy value.
If "subgroup", the target parameter
is the average treatement effect among
the subgroup of subjects that would receive
treatment under the policy, see details.
"subgroup" is only implemented for \code{type = "dr"}
in the single-stage case with a dichotomous action set.}

\item{type}{Character string. Type of evaluation. Either \code{"dr"}
(doubly robust),
\code{"ipw"} (inverse propensity weighting),
or \code{"or"} (outcome regression).}

\item{cross_fit_type}{Character string.
Either "stacked", or "pooled", see details. (Only used if \code{M > 1} and target = "subgroup")}

\item{variance_type}{Character string. Either "pooled" (default),
"stacked" or "complete", see details. (Only used if \code{M > 1})}

\item{M}{Number of folds for cross-fitting.}

\item{nrep}{Number of repetitions of cross-fitting (estimates averaged over repeated cross-fittings)}

\item{min_subgroup_size}{Minimum number of observations in the evaluated subgroup (Only used if target = "subgroup").}

\item{future_args}{Arguments passed to \code{\link[future.apply:future_apply]{future.apply::future_apply()}}.}

\item{name}{Character string.}

\item{object, x, y}{Objects of class "policy_eval".}

\item{...}{Additional arguments.}

\item{digits}{Integer. Number of printed digits.}

\item{width}{Integer. Width of printed parameter name.}

\item{std.error}{Logical. Should the std.error be printed.}

\item{level}{Numeric. Level of confidence limits.}

\item{p.value}{Logical. Should the p.value for associated confidence level be printed.}

\item{labels}{Name(s) of the estimate(s).}

\item{paired}{\code{TRUE} indicates that the estimates are based on
the same data sample.}
}
\value{
\code{policy_eval()} returns an object of class "policy_eval".
The object is a list containing the following elements:
\item{\code{coef}}{Numeric vector. The estimated target parameter:
policy value or subgroup average treatment effect.}
\item{\code{IC}}{Numeric matrix. Estimated influence curve associated with
\code{coef}.}
\item{\code{type}}{Character string. The type of evaluation ("dr", "ipw",
"or").}
\item{\code{target}}{Character string. The target parameter ("value" or "subgroup")}
\item{\code{id}}{Character vector. The IDs of the observations.}
\item{\code{name}}{Character vector. Names for the each element in \code{coef}.}
\item{\code{coef_ipw}}{(only if \code{type = "dr"}) Numeric vector.
Estimate of \code{coef} based solely on inverse probability weighting.}
\item{\code{coef_or}}{(only if \code{type = "dr"}) Numeric vector.
Estimate of \code{coef} based solely on outcome regression.}
\item{\code{policy_actions}}{\link[data.table:data.table]{data.table::data.table} with keys id and stage. Actions
associated with the policy for every observation and stage.}
\item{\code{policy_object}}{(only if \code{policy = NULL} and \code{M = 1})
The policy object returned by \code{policy_learn}, see \link{policy_learn}.}
\item{\code{g_functions}}{(only if \code{M = 1}) The
fitted g-functions. Object of class "nuisance_functions".}
\item{\code{g_values}}{The fitted g-function values.}
\item{\code{q_functions}}{(only if \code{M = 1}) The
fitted Q-functions. Object of class "nuisance_functions".}
\item{\code{q_values}}{The fitted Q-function values.}
\item{\code{Z}}{(only if \code{target = "subgroup"})
Matrix with the doubly robust stage 1 scores for each action.}
\item{\code{subgroup_indicator}}{(only if \code{target = "subgroup"})
Logical matrix identifying subjects in the subgroup.
Each column represents a different subgroup threshold.}
\item{\code{cross_fits}}{(only if \code{M > 1}) List containing the
"policy_eval" object for every (validation) fold.}
\item{\code{folds}}{(only if \code{M > 1}) The (validation) folds used
for cross-fitting.}
\item{\code{cross_fit_type}}{Character string.}
\item{\code{variance_type}}{Character string.}
}
\description{
\code{policy_eval()} is used to estimate
the value of a given fixed policy
or a data adaptive policy (e.g. a policy
learned from the data). \code{policy_eval()}
is also used to estimate the average
treatment effect among the subjects who would
get the treatment under the policy.
}
\details{
Each observation has the sequential form
\deqn{O= {B, U_1, X_1, A_1, ..., U_K, X_K, A_K, U_{K+1}},}
for a possibly stochastic number of stages K.
\itemize{
\item \eqn{B} is a vector of baseline covariates.
\item \eqn{U_k} is the reward at stage k
(not influenced by the action \eqn{A_k}).
\item \eqn{X_k} is a vector of state
covariates summarizing the state at stage k.
\item \eqn{A_k} is the categorical action
within the action set \eqn{\mathcal{A}} at stage k.
}
The utility is given by the sum of the rewards, i.e.,
\eqn{U = \sum_{k = 1}^{K+1} U_k}.

A policy is a set of functions
\deqn{d = \{d_1, ..., d_K\},}
where \eqn{d_k} for \eqn{k\in \{1, ..., K\}}
maps \eqn{\{B, X_1, A_1, ..., A_{k-1}, X_k\}} into the
action set.

Recursively define the Q-models (\code{q_models}):
\deqn{Q^d_K(h_K, a_K) = E[U|H_K = h_K, A_K = a_K]}
\deqn{Q^d_k(h_k, a_k) = E[Q_{k+1}(H_{k+1},
d_{k+1}(B,X_1, A_1,...,X_{k+1}))|H_k = h_k, A_k = a_k].}
If \code{q_full_history = TRUE},
\eqn{H_k = \{B, X_1, A_1, ..., A_{k-1}, X_k\}}, and if
\code{q_full_history = FALSE}, \eqn{H_k = \{B, X_k\}}.

The g-models (\code{g_models}) are defined as
\deqn{g_k(h_k, a_k) = P(A_k = a_k|H_k = h_k).}
If \code{g_full_history = TRUE},
\eqn{H_k = \{B, X_1, A_1, ..., A_{k-1}, X_k\}}, and if
\code{g_full_history = FALSE}, \eqn{H_k = \{B, X_k\}}.
Furthermore, if \code{g_full_history = FALSE} and \code{g_models} is a
single model, it is assumed that \eqn{g_1(h_1, a_1) = ... = g_K(h_K, a_K)}.

If \code{target = "value"} and \code{type = "or"}
\code{policy_eval()} returns the empirical estimate of
the value (\code{coef}):
\deqn{E\left[Q^d_1(H_1, d_1(\cdot))\right]}

If \code{target = "value"} and \code{type = "ipw"} \code{policy_eval()}
returns the empirical estimates of
the value (\code{coef}) and influence curve (\code{IC}):
\deqn{E\left[\left(\prod_{k=1}^K I\{A_k = d_k(\cdot)\}
g_k(H_k, A_k)^{-1}\right) U\right].}
\deqn{\left(\prod_{k=1}^K I\{A_k =
d_k(\cdot)\} g_k(H_k, A_k)^{-1}\right) U -
E\left[\left(\prod_{k=1}^K
I\{A_k = d_k(\cdot)\} g_k(H_k, A_k)^{-1}\right) U\right].}

If \code{target = "value"} and
\code{type = "dr"} \code{policy_eval} returns the empirical estimates of
the value (\code{coef}) and influence curve (\code{IC}):
\deqn{E[Z_1(d,g,Q^d)(O)],}
\deqn{Z_1(d, g, Q^d)(O) - E[Z_1(d,g, Q^d)(O)],}
where
\deqn{
Z_1(d, g, Q^d)(O) = Q^d_1(H_1 , d_1(\cdot)) +
\sum_{r = 1}^K \prod_{j = 1}^{r}
\frac{I\{A_j = d_j(\cdot)\}}{g_{j}(H_j, A_j)}
\{Q_{r+1}^d(H_{r+1} , d_{r+1}(\cdot)) - Q_{r}^d(H_r , d_r(\cdot))\}.
}

If \code{target = "subgroup"}, \code{type = "dr"}, \code{K = 1},
and \eqn{\mathcal{A} = \{0,1\}}, \code{policy_eval()}
returns the empirical estimates of the subgroup average
treatment effect (\code{coef}) and influence curve (\code{IC}):
\deqn{E[Z_1(1,g,Q)(O) - Z_1(0,g,Q)(O) | d_1(\cdot) = 1],}
\deqn{\frac{1}{P(d_1(\cdot) = 1)} I\{d_1(\cdot) = 1\}
\Big\{Z_1(1,g,Q)(O) - Z_1(0,g,Q)(O) - E[Z_1(1,g,Q)(O)
- Z_1(0,g,Q)(O) | d_1(\cdot) = 1]\Big\}.}

Applying \eqn{M}-fold cross-fitting using the \{M\} argument, let
\deqn{\mathcal{Z}_{1,m}(a) = \{Z_1(a, g_m, Q_m^d)(O): O\in \mathcal{O}_m \}.}

If \code{target = "subgroup"}, \code{type = "dr"}, \code{K = 1},
\eqn{\mathcal{A} = \{0,1\}}, and \code{cross_fit_type = "pooled"},
\code{policy_eval()} returns the estimate \deqn{\frac{1}{{N^{-1} \sum_{i =
1}^N I\{d(H_i) = 1\}}} N^{-1} \sum_{m=1}^M \sum_{(Z, H) \in \mathcal{Z}_{1,m}
\times \mathcal{H}_{1,m}} I\{d_1(H) = 1\} \left\{Z(1)-Z(0)\right\}} If
\code{cross_fit_type = "stacked"} the returned estimate is \deqn{M^{-1}
\sum_{m = 1}^M \frac{1}{{n^{-1} \sum_{h \in \mathcal{H}_{1,m}} I\{d(h) =
1\}}} n^{-1} \sum_{(Z, H) \in \mathcal{Z}_{1,m} \times \mathcal{H}_{1,m}}
I\{d_1(H) = 1\} \left\{Z(1)-Z(0)\right\},} where for ease of notation we let
the integer \eqn{n} be the number of oberservations in each fold.
}
\section{S3 generics}{

The following S3 generic functions are available for an object of
class \code{policy_eval}:
\describe{
\item{\code{\link[=get_g_functions]{get_g_functions()}}}{ Extract the fitted g-functions.}
\item{\code{\link[=get_q_functions]{get_q_functions()}}}{ Extract the fitted Q-functions.}
\item{\code{\link[=get_policy]{get_policy()}}}{ Extract the fitted policy object.}
\item{\code{\link[=get_policy_functions]{get_policy_functions()}}}{Extract the fitted policy function for a given stage.}
\item{\code{\link[=get_policy_actions]{get_policy_actions()}}}{ Extract the (fitted) policy actions.}
\item{\code{\link[=plot.policy_eval]{plot.policy_eval()}}}{Plot diagnostics.}
}
}

\examples{
library("polle")
### Single stage:
d1 <- sim_single_stage(5e2, seed=1)
pd1 <- policy_data(d1,
                   action = "A",
                   covariates = list("Z", "B", "L"),
                   utility = "U")
pd1

# defining a static policy (A=1):
pl1 <- policy_def(1)

# evaluating the policy:
pe1 <- policy_eval(policy_data = pd1,
                   policy = pl1,
                   g_models = g_glm(),
                   q_models = q_glm(),
                   name = "A=1 (glm)")

# summarizing the estimated value of the policy:
# (equivalent to summary(pe1)):
pe1
coef(pe1) # value coefficient
sqrt(vcov(pe1)) # value standard error

# getting the g-function and Q-function values:
head(predict(get_g_functions(pe1), pd1))
head(predict(get_q_functions(pe1), pd1))

# getting the fitted influence curve (IC) for the value:
head(IC(pe1))

# evaluating the policy using random forest nuisance models:
set.seed(1)
pe1_rf <- policy_eval(policy_data = pd1,
                      policy = pl1,
                      g_models = g_rf(),
                      q_models = q_rf(),
                      name = "A=1 (rf)")

# merging the two estimates (equivalent to pe1 + pe1_rf):
(est1 <- merge(pe1, pe1_rf))
coef(est1)
head(IC(est1))

### Two stages:
d2 <- sim_two_stage(5e2, seed=1)
pd2 <- policy_data(d2,
                   action = c("A_1", "A_2"),
                   covariates = list(L = c("L_1", "L_2"),
                                     C = c("C_1", "C_2")),
                   utility = c("U_1", "U_2", "U_3"))
pd2

# defining a policy learner based on cross-fitted doubly robust Q-learning:
pl2 <- policy_learn(
   type = "drql",
   control = control_drql(qv_models = list(q_glm(~C_1),
                                           q_glm(~C_1+C_2))),
   full_history = TRUE,
   L = 2) # number of folds for cross-fitting

# evaluating the policy learner using 2-fold cross fitting:
pe2 <- policy_eval(type = "dr",
                   policy_data = pd2,
                   policy_learn = pl2,
                   q_models = q_glm(),
                   g_models = g_glm(),
                   M = 2, # number of folds for cross-fitting
                   name = "drql")
# summarizing the estimated value of the policy:
pe2

# getting the cross-fitted policy actions:
head(get_policy_actions(pe2))
}
\references{
van der Laan, Mark J., and Alexander R. Luedtke.
"Targeted learning of the mean outcome under an optimal dynamic treatment rule."
Journal of causal inference 3.1 (2015): 61-95.
\doi{10.1515/jci-2013-0022}
\cr \cr
Tsiatis, Anastasios A., et al. Dynamic
treatment regimes: Statistical methods for precision medicine. Chapman and
Hall/CRC, 2019. \doi{10.1201/9780429192692}.
\cr \cr
Victor Chernozhukov, Denis
Chetverikov, Mert Demirer, Esther Duflo, Christian Hansen, Whitney Newey,
James Robins, Double/debiased machine learning for treatment and structural
parameters, The Econometrics Journal, Volume 21, Issue 1, 1 February 2018,
Pages C1–C68, \doi{10.1111/ectj.12097}.
}
\seealso{
\link[lava:IC.default]{lava::IC}, \link[lava:estimate.default]{lava::estimate.default}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nuisance_functions.R
\name{predict.nuisance_functions}
\alias{predict.nuisance_functions}
\title{Predict g-functions and Q-functions}
\usage{
\method{predict}{nuisance_functions}(object, new_policy_data, event_set = c(0), ...)
}
\arguments{
\item{object}{Object of class "nuisance_functions". Either \code{g_functions}
or \code{q_functions} as returned by \code{\link[=policy_eval]{policy_eval()}} or \code{\link[=policy_learn]{policy_learn()}}.}

\item{new_policy_data}{Policy data object created by \code{\link[=policy_data]{policy_data()}}. Action events (0),
terminal events (1), or right-censoring events (2).}

\item{event_set}{Subset of the events considered. Action events (0), terminal events (1), or
right-censoring events (2)}

\item{...}{Additional arguments.}
}
\value{
\link[data.table:data.table]{data.table::data.table} with keys \code{id} and \code{stage} and variables \code{g_a} or \code{Q_a} for
each action a in the actions set.
}
\description{
\code{predict()} returns the fitted values of the g-functions and
Q-functions when applied to a (new) policy data object.
}
\examples{
library("polle")
### Single stage:
d <- sim_single_stage(5e2, seed=1)
pd <- policy_data(d, action="A", covariates=list("Z", "B", "L"), utility="U")
pd
# defining a static policy (A=1):
pl <- policy_def(1, name = "A=1")

# doubly robust evaluation of the policy:
pe <- policy_eval(policy_data = pd,
                  policy = pl,
                  g_models = g_glm(),
                  q_models = q_glm())
# summarizing the estimated value of the policy:
pe

# getting the fitted g-function values:
head(predict(get_g_functions(pe), pd))

# getting the fitted Q-function values:
head(predict(get_q_functions(pe), pd))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ptl.R
\name{control_ptl}
\alias{control_ptl}
\title{Control arguments for Policy Tree Learning}
\usage{
control_ptl(
  policy_vars = NULL,
  hybrid = FALSE,
  depth = 2,
  search.depth = 2,
  split.step = 1,
  min.node.size = 1
)
}
\arguments{
\item{policy_vars}{Character vector/string or list of character
vectors/strings. Variable names used to
construct the V-restricted policy tree.
The names must be a subset of the history names, see get_history_names().
Not passed to \code{policy_tree()}.}

\item{hybrid}{If \code{TRUE}, \code{\link[policytree:hybrid_policy_tree]{policytree::hybrid_policy_tree()}} is used to
fit a policy tree. Not passed to \code{policy_tree()}.}

\item{depth}{Integer or integer vector. The depth of the fitted policy
tree for each stage.}

\item{search.depth}{(only used if \code{hybrid = TRUE}) Integer or integer
vector. Depth to look ahead when splitting at each stage.}

\item{split.step}{Integer or integer vector. The number of possible splits
to consider when performing policy tree search at each stage.}

\item{min.node.size}{Integer or integer vector. The smallest terminal node
size permitted at each stage.}
}
\value{
list of (default) control arguments.
}
\description{
\code{control_ptl} sets the default control arguments
for doubly robust policy tree learning, \code{type = "ptl"}.
The arguments are passed directly to \code{\link[policytree:policy_tree]{policytree::policy_tree()}} (or
\code{\link[policytree:hybrid_policy_tree]{policytree::hybrid_policy_tree()}}) if not specified otherwise.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{subset_id}
\alias{subset_id}
\title{Subset Policy Data on ID}
\usage{
subset_id(object, id, preserve_action_set = TRUE)
}
\arguments{
\item{object}{Object of class \link{policy_data}.}

\item{id}{character vectors of IDs.}

\item{preserve_action_set}{If TRUE, the action sets must be preserved.}
}
\value{
Object of class \link{policy_data}.
}
\description{
\code{subset_id} returns a policy data object containing the given IDs.
}
\examples{
library("polle")
### Single stage:
d <- sim_single_stage(5e2, seed=1)
# constructing policy_data object:
pd <- policy_data(d, action="A", covariates=list("Z", "B", "L"), utility="U")
pd

# getting the observation IDs:
get_id(pd)[1:10]

# subsetting on IDs:
pdsub <- subset_id(pd, id = 250:500)
pdsub
get_id(pdsub)[1:10]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/owl.R
\name{control_owl}
\alias{control_owl}
\title{Control arguments for Outcome Weighted Learning}
\usage{
control_owl(
  policy_vars = NULL,
  reuse_scales = TRUE,
  res.lasso = TRUE,
  loss = "hinge",
  kernel = "linear",
  augment = FALSE,
  c = 2^(-2:2),
  sigma = c(0.03, 0.05, 0.07),
  s = 2^(-2:2),
  m = 4
)
}
\arguments{
\item{policy_vars}{Character vector/string or list of character
vectors/strings. Variable names used to restrict the policy.
The names must be a subset of the history names, see get_history_names().
Not passed to \code{owl()}.}

\item{reuse_scales}{The history matrix passed to \code{owl()} is scaled
using \code{\link[=scale]{scale()}} as advised. If \code{TRUE}, the scales of the history matrix
will be saved and reused when applied to (new) test data.}

\item{res.lasso}{If \code{TRUE} a lasso penalty is applied.}

\item{loss}{Loss function. The options are \code{"hinge"}, \code{"ramp"},
\code{"logit"}, \code{"logit.lasso"}, \code{"l2"}, \code{"l2.lasso"}.}

\item{kernel}{Type of kernel used by the support vector machine. The
options are \code{"linear"}, \code{"rbf"}.}

\item{augment}{If \code{TRUE} the outcomes are augmented.}

\item{c}{Regularization parameter.}

\item{sigma}{Tuning parameter.}

\item{s}{Slope parameter.}

\item{m}{Number of folds for cross-validation of the parameters.}
}
\value{
list of (default) control arguments.
}
\description{
\code{control_owl()} sets the default control arguments
for backwards outcome weighted learning, \code{type = "owl"}.
The arguments are passed directly to \code{\link[DTRlearn2:owl]{DTRlearn2::owl()}} if not
specified otherwise.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{get_event}
\alias{get_event}
\title{Get event indicators}
\usage{
get_event(object)
}
\arguments{
\item{object}{Object of class \link{policy_data} or \link{history}.}
}
\value{
\link[data.table:data.table]{data.table::data.table} with keys id and stage.
}
\description{
\code{get_event} returns the event indicators for given IDs
stages
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_data_functions.R
\name{copy_policy_data}
\alias{copy_policy_data}
\title{Copy Policy Data Object}
\usage{
copy_policy_data(object)
}
\arguments{
\item{object}{Object of class \link{policy_data}.}
}
\value{
Object of class \link{policy_data}.
}
\description{
Objects of class \link{policy_data} contains elements of
class \link[data.table:data.table]{data.table::data.table}.
\code{data.table} provide functions that operate on objects by reference.
Thus, the \code{policy_data} object is not copied when modified by reference,
see examples. An explicit copy can be made by \code{copy_policy_data}. The
function is a wrapper of \code{\link[data.table:copy]{data.table::copy()}}.
}
\examples{
library("polle")
### Single stage case: Wide data
d1 <- sim_single_stage(5e2, seed=1)
head(d1, 5)
# constructing policy_data object:
pd1 <- policy_data(d1,
                   action="A",
                   covariates=c("Z", "B", "L"),
                   utility="U")
pd1

# True copy
pd2 <- copy_policy_data(pd1)
# manipulating the data.table by reference:
pd2$baseline_data[, id := id + 1]
head(pd2$baseline_data$id - pd1$baseline_data$id)

# False copy
pd2 <- pd1
# manipulating the data.table by reference:
pd2$baseline_data[, id := id + 1]
head(pd2$baseline_data$id - pd1$baseline_data$id)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/earl.R
\name{control_earl}
\alias{control_earl}
\title{Control arguments for Efficient Augmentation and Relaxation Learning}
\usage{
control_earl(
  moPropen,
  moMain,
  moCont,
  regime,
  iter = 0L,
  fSet = NULL,
  lambdas = 0.5,
  cvFolds = 0L,
  surrogate = "hinge",
  kernel = "linear",
  kparam = NULL,
  verbose = 0L
)
}
\arguments{
\item{moPropen}{Propensity model of class "ModelObj", see \link[modelObj:modelObj]{modelObj::modelObj}.}

\item{moMain}{Main effects outcome model of class "ModelObj".}

\item{moCont}{Contrast outcome model of class "ModelObj".}

\item{regime}{An object of class \link{formula} specifying the design of the policy/regime.}

\item{iter}{Maximum number of iterations for outcome regression.}

\item{fSet}{A function or NULL defining subset structure.}

\item{lambdas}{Numeric or numeric vector. Penalty parameter.}

\item{cvFolds}{Integer. Number of folds for cross-validation of the parameters.}

\item{surrogate}{The surrogate 0-1 loss function. The options are
\code{"logit"}, \code{"exp"}, \code{"hinge"}, \code{"sqhinge"}, \code{"huber"}.}

\item{kernel}{The options are \code{"linear"}, \code{"poly"}, \code{"radial"}.}

\item{kparam}{Numeric. Kernel parameter}

\item{verbose}{Integer.}
}
\value{
list of (default) control arguments.
}
\description{
\code{control_earl} sets the default control arguments
for efficient augmentation and relaxation learning , \code{type = "earl"}.
The arguments are passed directly to \code{\link[DynTxRegime:earl]{DynTxRegime::earl()}} if not
specified otherwise.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/c_models.R
\docType{class}
\name{c_model}
\alias{c_model}
\alias{c_cox}
\alias{c_no_censoring}
\title{c_model class object}
\usage{
c_cox(formula = ~., offset = NULL, weights = NULL, ...)

c_no_censoring()
}
\arguments{
\item{formula}{An object of class \link{formula} specifying the design matrix for
the right-censoring model. Use \code{\link[=get_history_names]{get_history_names()}} to view the available
variable names.}

\item{offset}{offsets for Cox model, see \code{\link[mets:phreg]{mets::phreg()}}}

\item{weights}{weights for Cox score equations, see \code{\link[mets:phreg]{mets::phreg()}}}

\item{...}{Additional arguments passed to the model.}
}
\value{
A c-model object (function) with arguments:
\itemize{
\item event: censoring events
\item time: start time
\item time2: end time
\item H: history matrix
}
}
\description{
Provides constructors for right-censoring models.
Main constructors include:
\itemize{
\item \code{c_cox()}: Cox proportional hazards model
\item \code{c_no_censoring()}: Model for scenarios without censoring
The constructors are used as input for \code{\link[=policy_eval]{policy_eval()}} and \code{\link[=policy_learn]{policy_learn()}}.
}
}
\details{
\code{c_cox()} is a wrapper of \code{\link[mets:phreg]{mets::phreg()}} (Cox proportional hazard model).\cr
}
\seealso{
\code{\link[=get_history_names]{get_history_names()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/c_function.R
\name{fit_c_functions}
\alias{fit_c_functions}
\title{Fit Censoring Functions}
\usage{
fit_c_functions(policy_data, c_models, full_history = FALSE)
}
\arguments{
\item{policy_data}{Policy data object created by \code{\link[=policy_data]{policy_data()}}}

\item{c_models}{Single c_model or list of K+1 c_models}

\item{full_history}{Logical; use full history (TRUE) or Markov-type history (FALSE)}
}
\value{
List of fitted censoring functions with class "c_functions"
}
\description{
Fits right-censoring models for each stage or a single model across all stages.
}
\details{
The function handles two scenarios:
\itemize{
\item Multiple models: One model per stage (length K+1)
\item Single model: Same model applied across all stages
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/q_models.R
\docType{class}
\name{q_model}
\alias{q_model}
\alias{q_glm}
\alias{q_glmnet}
\alias{q_rf}
\alias{q_sl}
\alias{q_xgboost}
\title{q_model class object}
\usage{
q_glm(
  formula = ~A * .,
  family = gaussian(),
  model = FALSE,
  na.action = na.pass,
  ...
)

q_glmnet(
  formula = ~A * .,
  family = "gaussian",
  alpha = 1,
  s = "lambda.min",
  ...
)

q_rf(
  formula = ~.,
  num.trees = c(250, 500, 750),
  mtry = NULL,
  cv_args = list(nfolds = 3, rep = 1),
  ...
)

q_sl(
  formula = ~.,
  SL.library = c("SL.mean", "SL.glm"),
  env = parent.frame(),
  onlySL = TRUE,
  discreteSL = FALSE,
  ...
)

q_xgboost(
  formula = ~.,
  objective = "reg:squarederror",
  nrounds,
  max_depth = 6,
  learning_rate = NULL,
  nthread = 1,
  cv_args = list(nfolds = 3, rep = 1),
  ...
)
}
\arguments{
\item{formula}{An object of class \link{formula} specifying the design matrix for
the outcome regression model/Q-model at the given stage. The action at the
given stage is always denoted 'A', see examples. Use
\code{\link[=get_history_names]{get_history_names()}} to see the additional
available variable names.}

\item{family}{A description of the error distribution and link function to
be used in the model.}

\item{model}{(Only used by \code{q_glm}) If \code{FALSE} model frame will
not be saved.}

\item{na.action}{(Only used by \code{q_glm}) A function which indicates what
should happen when the data contain NAs, see \link{na.pass}.}

\item{...}{Additional arguments passed to \code{\link[=glm]{glm()}}, \link[glmnet:glmnet]{glmnet::glmnet},
\link[ranger:ranger]{ranger::ranger} or \link[SuperLearner:SuperLearner]{SuperLearner::SuperLearner}.}

\item{alpha}{(Only used by \code{q_glmnet}) The elasticnet mixing parameter
between 0 and 1. alpha equal to 1 is the lasso penalty, and alpha equal
to 0 the ridge penalty.}

\item{s}{(Only used by \code{q_glmnet}) Value(s) of the penalty parameter
lambda at which predictions are required, see \code{\link[glmnet:predict.glmnet]{glmnet::predict.glmnet()}}.}

\item{num.trees}{(Only used by \code{q_rf}) Number of trees.}

\item{mtry}{(Only used by \code{q_rf}) Number of variables to possibly split
at in each node.}

\item{cv_args}{(Only used by \code{q_rf}) Cross-validation parameters.
Only used if multiple hyper-parameters are given. \code{K} is the number
of folds and
\code{rep} is the number of replications.}

\item{SL.library}{(Only used by \code{q_sl}) Either a character vector of
prediction algorithms or a list containing character vectors,
see \link[SuperLearner:SuperLearner]{SuperLearner::SuperLearner}.}

\item{env}{(Only used by \code{q_sl}) Environment containing the learner
functions. Defaults to the calling environment.}

\item{onlySL}{(Only used by \code{q_sl}) Logical. If TRUE, only saves and computes predictions
for algorithms with non-zero coefficients in the super learner object.}

\item{discreteSL}{(Only used by \code{q_sl}) If TRUE, select the model with
the lowest cross-validated risk.}

\item{objective}{(Only used by \code{q_xgboost}) specify the learning
task and the corresponding learning objective, see \link[xgboost:xgboost]{xgboost::xgboost}.}

\item{nrounds}{(Only used by \code{q_xgboost}) max number of boosting iterations.}

\item{max_depth}{(Only used by \code{q_xgboost}) maximum depth of a tree.}

\item{learning_rate}{(Only used by \code{q_xgboost}) learning rate.}

\item{nthread}{(Only used by \code{q_xgboost}) number of threads.}
}
\value{
q_model object: function with arguments 'AH'
(combined action and history matrix) and 'V_res' (residual value/expected
utility).
}
\description{
Use \code{q_glm()}, \code{q_glmnet()}, \code{q_rf()}, and \code{q_sl()} to construct
an outcome regression model/Q-model object.
The constructors are used as input for \code{\link[=policy_eval]{policy_eval()}} and \code{\link[=policy_learn]{policy_learn()}}.
}
\details{
\code{q_glm()} is a wrapper of \code{\link[=glm]{glm()}} (generalized linear model).\cr
\code{q_glmnet()} is a wrapper of \code{\link[glmnet:glmnet]{glmnet::glmnet()}} (generalized linear model via
penalized maximum likelihood).\cr
\code{q_rf()} is a wrapper of \code{\link[ranger:ranger]{ranger::ranger()}} (random forest).
When multiple hyper-parameters are given, the
model with the lowest cross-validation error is selected.\cr
\code{q_sl()} is a wrapper of \link[SuperLearner:SuperLearner]{SuperLearner::SuperLearner} (ensemble model).
\code{q_xgboost()} is a wrapper of \link[xgboost:xgboost]{xgboost::xgboost}.
}
\examples{
library("polle")
### Single stage case
d1 <- sim_single_stage(5e2, seed=1)
pd1 <- policy_data(d1,
                   action="A",
                   covariates=list("Z", "B", "L"),
                   utility="U")
pd1

# available history variable names for the outcome regression:
get_history_names(pd1)

# evaluating the static policy a=1 using inverse
# propensity weighting based on the given Q-model:
pe1 <- policy_eval(type = "or",
                   policy_data = pd1,
                   policy = policy_def(1, name = "A=1"),
                   q_model = q_glm(formula = ~A*.))
pe1

# getting the fitted Q-function values
head(predict(get_q_functions(pe1), pd1))

### Two stages:
d2 <- sim_two_stage(5e2, seed=1)
pd2 <- policy_data(d2,
                  action = c("A_1", "A_2"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd2

# available full history variable names at each stage:
get_history_names(pd2, stage = 1)
get_history_names(pd2, stage = 2)

# evaluating the static policy a=1 using outcome
# regression based on a glm model for each stage:
pe2 <- policy_eval(type = "or",
            policy_data = pd2,
            policy = policy_def(1, reuse = TRUE, name = "A=1"),
            q_model = list(q_glm(~ A * L_1),
                           q_glm(~ A * (L_1 + L_2))),
            q_full_history = TRUE)
pe2

# getting the fitted Q-function values
head(predict(get_q_functions(pe2), pd2))
}
\seealso{
\code{\link[=get_history_names]{get_history_names()}}, \code{\link[=get_q_functions]{get_q_functions()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polle-package.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{IC}
\alias{estimate}
\alias{All}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{lava}{\code{\link[lava:estimate.default]{estimate}}, \code{\link[lava:IC.default]{IC}}}

  \item{SuperLearner}{\code{\link[SuperLearner:write.screen.template]{All}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g_models.R
\docType{class}
\name{g_model}
\alias{g_model}
\alias{g_empir}
\alias{g_glm}
\alias{g_glmnet}
\alias{g_rf}
\alias{g_sl}
\alias{g_xgboost}
\title{g_model class object}
\usage{
g_empir(formula = ~1, ...)

g_glm(
  formula = ~.,
  family = "binomial",
  model = FALSE,
  na.action = na.pass,
  ...
)

g_glmnet(formula = ~., family = "binomial", alpha = 1, s = "lambda.min", ...)

g_rf(
  formula = ~.,
  num.trees = c(500),
  mtry = NULL,
  cv_args = list(nfolds = 5, rep = 1),
  ...
)

g_sl(
  formula = ~.,
  SL.library = c("SL.mean", "SL.glm"),
  family = binomial(),
  env = parent.frame(),
  onlySL = TRUE,
  ...
)

g_xgboost(
  formula = ~.,
  objective = "binary:logistic",
  nrounds,
  max_depth = 6,
  learning_rate = NULL,
  nthread = 1,
  cv_args = list(nfolds = 3, rep = 1),
  ...
)
}
\arguments{
\item{formula}{An object of class \link{formula} specifying the design matrix for
the propensity model/g-model. Use \code{\link[=get_history_names]{get_history_names()}} to view the available
variable names.}

\item{...}{Additional arguments passed to \code{\link[=glm]{glm()}}, \link[glmnet:glmnet]{glmnet::glmnet},
\link[ranger:ranger]{ranger::ranger} or \link[SuperLearner:SuperLearner]{SuperLearner::SuperLearner}.}

\item{family}{A description of the error distribution and link function to
be used in the model.}

\item{model}{(Only used by \code{g_glm}) If \code{FALSE} model frame will
not be saved.}

\item{na.action}{(Only used by \code{g_glm}) A function which indicates what
should happen when the data contain NAs, see \link{na.pass}.}

\item{alpha}{(Only used by \code{g_glmnet}) The elastic net mixing parameter
between 0 and 1. alpha equal to 1 is the lasso penalty, and alpha equal
to 0 the ridge penalty.}

\item{s}{(Only used by \code{g_glmnet}) Value(s) of the penalty parameter
lambda at which predictions are required, see \code{\link[glmnet:predict.glmnet]{glmnet::predict.glmnet()}}.}

\item{num.trees}{(Only used by \code{g_rf}) Number of trees.}

\item{mtry}{(Only used by \code{g_rf}) Number of variables to possibly split
at in each node.}

\item{cv_args}{(Only used by \code{g_rf} and \code{g_xgboost}) Cross-validation parameters.
Only used if multiple hyper-parameters are given. \code{K} is the number
of folds and
\code{rep} is the number of replications.}

\item{SL.library}{(Only used by \code{g_sl}) Either a character vector of prediction algorithms or
a list containing character vectors, see \link[SuperLearner:SuperLearner]{SuperLearner::SuperLearner}.}

\item{env}{(Only used by \code{g_sl}) Environment containing the learner functions. Defaults to the calling environment.}

\item{onlySL}{(Only used by \code{g_sl}) Logical. If TRUE, only saves and computes predictions
for algorithms with non-zero coefficients in the super learner object.}

\item{objective}{(Only used by \code{g_xgboost}) specify the learning
task and the corresponding learning objective, see \link[xgboost:xgboost]{xgboost::xgboost}.}

\item{nrounds}{(Only used by \code{g_xgboost}) max number of boosting iterations.}

\item{max_depth}{(Only used by \code{g_xgboost}) maximum depth of a tree.}

\item{learning_rate}{(Only used by \code{g_xgboost}) learning rate.}

\item{nthread}{(Only used by \code{g_xgboost}) number of threads.}
}
\value{
g-model object: function with arguments 'A'
(action vector), 'H' (history matrix) and 'action_set'.
}
\description{
Use \code{g_glm()}, \code{g_empir()},
\code{g_glmnet()}, \code{g_rf()}, \code{g_sl()}, \code{g_xgboost} to construct
an action probability model/g-model object.
The constructors are used as input for \code{\link[=policy_eval]{policy_eval()}} and \code{\link[=policy_learn]{policy_learn()}}.
}
\details{
\code{g_glm()} is a wrapper of \code{\link[=glm]{glm()}} (generalized linear model).\cr
\code{g_empir()} calculates the empirical probabilities within the groups
defined by the formula.\cr
\code{g_glmnet()} is a wrapper of \code{\link[glmnet:glmnet]{glmnet::glmnet()}} (generalized linear model via
penalized maximum likelihood).\cr
\code{g_rf()} is a wrapper of \code{\link[ranger:ranger]{ranger::ranger()}} (random forest).
When multiple hyper-parameters are given, the
model with the lowest cross-validation error is selected.\cr
\code{g_sl()} is a wrapper of \link[SuperLearner:SuperLearner]{SuperLearner::SuperLearner} (ensemble model).\cr
\code{g_xgboost()} is a wrapper of \link[xgboost:xgboost]{xgboost::xgboost}.
}
\examples{
library("polle")
### Two stages:
d <- sim_two_stage(2e2, seed=1)
pd <- policy_data(d,
                  action = c("A_1", "A_2"),
                  baseline = c("B"),
                  covariates = list(L = c("L_1", "L_2"),
                                    C = c("C_1", "C_2")),
                  utility = c("U_1", "U_2", "U_3"))
pd

# available state history variable names:
get_history_names(pd)
# defining a g-model:
g_model <- g_glm(formula = ~B+C)

# evaluating the static policy (A=1) using inverse propensity weighting
# based on a state glm model across all stages:
pe <- policy_eval(type = "ipw",
                  policy_data = pd,
                  policy = policy_def(1, reuse = TRUE),
                 g_models = g_model)
# inspecting the fitted g-model:
get_g_functions(pe)

# available full history variable names at each stage:
get_history_names(pd, stage = 1)
get_history_names(pd, stage = 2)

# evaluating the same policy based on a full history
# glm model for each stage:
pe <- policy_eval(type = "ipw",
                   policy_data = pd,
                   policy = policy_def(1, reuse = TRUE),
                   g_models = list(g_glm(~ L_1 + B),
                                   g_glm(~ A_1 + L_2 + B)),
                   g_full_history = TRUE)
# inspecting the fitted g-models:
get_g_functions(pe)
}
\seealso{
\code{\link[=get_history_names]{get_history_names()}}, \code{\link[=get_g_functions]{get_g_functions()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation.R
\name{sim_single_stage}
\alias{sim_single_stage}
\title{Simulate Single-Stage Data}
\usage{
sim_single_stage(
  n = 10000,
  par = c(k = 0.1, d = 0.5, a = 1, b = -2.5, c = 3, p = 0.3),
  action_model = function(Z, L, B, k, d) {
     k * (Z + L - 1) * Z^(-2) + d * (B == 1)

    },
  utility_model = function(Z, L, A, a, b, c) {
     Z + L + A * (c * Z + a * L + b)
 },
  seed = NULL,
  return_model = FALSE,
  ...
)
}
\arguments{
\item{n}{Number of observations.}

\item{par}{Named vector with distributional parameters.
\itemize{
\item{} \code{k}: \eqn{\kappa}
\item{} \code{d}: \eqn{\delta}
\item{} \code{a}: \eqn{\alpha}
\item{} \code{b}: \eqn{\beta}
\item{} \code{c}: \eqn{\gamma}
\item{} \code{p}: \eqn{\pi}
}}

\item{action_model}{Function used to specify the action/treatment probability (logit link).}

\item{utility_model}{Function used to specify the conditional mean utility.}

\item{seed}{Integer.}

\item{return_model}{If TRUE, the \link[lava:lvm]{lava::lvm} model is returned.}

\item{...}{Additional arguments passed to \code{\link[lava:lvm]{lava::lvm()}}.}
}
\value{
data.frame with n rows and columns Z, L, B, A, and U.
}
\description{
Simulate Single-Stage Data
}
\details{
\code{sim_single_stage} samples \code{n} iid observation
\eqn{O = (B, Z, L, A, U)} with the following distribution:
\deqn{
B \sim Bernoulli(\pi)\\
Z, L \sim Uniform([0,1])\\
A\mid Z,L,B \sim Bernoulli(expit\{\kappa Z^{-2}(Z+L-1) + \delta B)\\
U\mid Z,L,A \sim \mathcal{N}(Z+L+A\cdot\{\gamma Z + \alpha L + \beta\}, 1)
}
}
